/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Contributor: Ovidiu Serban, University of Reading, www.isr.reading.ac.uk
 * Contributor: Kevin Fuchs, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-10-30
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmComInterface;

import com.google.common.util.concurrent.ThreadFactoryBuilder;
import eu.intuitel.db.access.DbAccessException;
import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.lpmComInterface.communicators.EngineCommunicator;
import eu.intuitel.lpmComInterface.communicators.LmsCommunicator;
import eu.intuitel.lpmComInterface.communicators.SlomLocalCommunicator;
import eu.intuitel.lpmComInterface.communicators.SlomRemoteCommunicator;
import eu.intuitel.lpmComInterface.runnables.MsgHandler;
import eu.intuitel.lpmComInterface.runnables.UserInputHandler;
import eu.intuitel.lpmComInterface.runnables.WorkerStarter;
import eu.intuitel.lpmComInterface.runnables.WorkflowHandler;
import eu.intuitel.lpmComInterface.utils.CmdParser;
import eu.intuitel.lpmComInterface.utils.LpmComInterfaceException;
import eu.intuitel.lpmComInterface.utils.ValueConverterUtil;
import eu.intuitel.lpmGeneralInterfaces.IEngineCommunicator;
import eu.intuitel.lpmGeneralInterfaces.ILmsCommunicator;
import eu.intuitel.lpmGeneralInterfaces.ISlomCommunicator;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Handler;
import org.apache.log4j.Level;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Main class to run the LpmComInterface.
 */
public class Main {

    private static final Logger logger = LoggerFactory.getLogger(Main.class);
    protected static final ExecutorService workerExecutor = Executors.newFixedThreadPool(5,
            (new ThreadFactoryBuilder()).setNameFormat("LpmProcessorPool-Thread-%d").build());
    protected static final CmdParser cmd = new CmdParser();
    protected static MsgHandler msgHandler = null;
    protected static WorkflowHandler wfHandler = null;
    protected static UserInputHandler inputHandler = null;
    protected static Class<? extends ILmsCommunicator> lmsComClass;
    protected static String[] lmsComParameters;
    protected static Class<? extends IEngineCommunicator> engineComClass;
    protected static String[] engineComParameters;
    protected static Class<? extends ISlomCommunicator> slomComClass;
    protected static String[] slomComParameters;
    protected static DbAccessor dbCom;

    /**
     * Returns the MsgHandler of the LPM.
     *
     * @return Null if not yet initialized.
     */
    public static MsgHandler getMsgHandler() {
        return msgHandler;
    }

    /**
     * Returns the WorkflowHandler of the LPM.
     *
     * @return Null if not yet initialized.
     */
    public static WorkflowHandler getWorkflowHandler() {
        return wfHandler;
    }

    /**
     * Main method to run the LpmComInterface.
     *
     * @param args Command line arguments
     */
    public static void main(String[] args) {
        printWelcome();

        //Register hook for safe shutdown
        Runtime.getRuntime().addShutdownHook(new Thread() {
            @Override
            public void run() {
                Thread.currentThread().setName("INTUITEL-Shutdown-Thread");
                shutdown();
            }
        });

        bridgeLogging();
        setLoggingLevels();
        if (!cmd.parseCmdLine(args)) {
            System.exit(1);
        }
        setLoggingLevels(); //update if needed
        if (cmd.isHelpMode()) {
            System.exit(0);
        }
        if (!initComConstructorData()) {
            System.exit(2);
        }
        if (!initDbConnection()) {
            System.exit(3);
        }
        if (!startUserInputHandler()) {
            System.exit(6);
        }
        if (!startMsgHandler()) {
            System.exit(4);
        }
        if (!startWorkflowHandler()) {
            System.exit(5);
        }
        if (cmd.isTestMode()) {
            logger.info("Exit immidiately - testmode ON!");
            System.exit(0);
        }
        logger.info("Finished init. LPM is now functional.");
        keepAlive();
        System.exit(0);
    }

    /**
     * Prints the welcome message to the console.
     */
    private static void printWelcome() {
        StringBuilder sb = new StringBuilder();
        sb.append("\n");
        sb.append("####################################################\n");
        sb.append("\n");
        sb.append("Starting the Learning Progress Model (LPM)\n");
        sb.append("Write 'shutdown' to initiate the shutdown procedure.\n");
        sb.append("(c) INTUITEL consortium, licence to tbd\n");
        sb.append("\n");
        sb.append("####################################################\n");
        sb.append("\n");
        System.out.print(sb.toString());
    }

    /**
     * Bridges the standard java.util.logging to the slf4j logging as it is used in this and the
     * other projects.
     *
     * @return True if the bridgeing was successfull
     */
    protected static boolean bridgeLogging() {
        try {
            java.util.logging.Logger rootLogger = java.util.logging.LogManager.getLogManager().getLogger("");
            java.util.logging.Handler[] handlers = rootLogger.getHandlers();
            for (Handler handler : handlers) {
                rootLogger.removeHandler(handler);
            }
            org.slf4j.bridge.SLF4JBridgeHandler.install();
            return true;
        } catch (SecurityException ex) {
            logger.error("SecurityException while bridgeing HTTP server.", ex);
            return false;
        }
    }

    /**
     * Sets the logging levels to have a clean console output.
     */
    protected static void setLoggingLevels() {
        org.apache.log4j.Logger hibernateLogger = org.apache.log4j.Logger.getLogger("org.hibernate");
        hibernateLogger.setLevel((Level) cmd.getDbLogLvl());
        org.apache.log4j.Logger c3p0Logger = org.apache.log4j.Logger.getLogger("com.mchange.v2");
        c3p0Logger.setLevel((Level) cmd.getDbLogLvl());
        org.apache.log4j.Logger intuitelLogger = org.apache.log4j.Logger.getLogger("eu.intuitel");
        intuitelLogger.setLevel((Level) cmd.getIntuitelLogLvl());
        org.apache.log4j.Logger grizzlyLogger = org.apache.log4j.Logger.getLogger("org.glassfish.grizzly");
        grizzlyLogger.setLevel((Level) cmd.getHttpLogLvl());
        org.apache.log4j.Logger jerseyLogger = org.apache.log4j.Logger.getLogger("com.sun.jersey");
        jerseyLogger.setLevel((Level) cmd.getHttpLogLvl());
        org.apache.log4j.Logger jbossLogger = org.apache.log4j.Logger.getLogger("org.jboss");
        jbossLogger.setLevel((Level) cmd.getHttpLogLvl());
    }

    /**
     * Initiates the communication constructor data by using the commandline information.
     *
     * @return True on success
     */
    protected static boolean initComConstructorData() {
        lmsComClass = LmsCommunicator.class;
        lmsComParameters = new String[]{};
        engineComClass = EngineCommunicator.class;
        engineComParameters = new String[]{};
        if (cmd.getSlomRepoAddress() == null) {
            slomComClass = SlomRemoteCommunicator.class;
            slomComParameters = new String[]{};
        } else {
            slomComClass = SlomLocalCommunicator.class;
            slomComParameters = new String[]{cmd.getSlomRepoAddress().toString()};
        }
        return true;
    }

    /**
     * Tests the connection to the user database.
     *
     * @return False if connection failed.
     */
    private static boolean initDbConnection() {
        try {
            logger.debug("Setting up DB connection.");
            dbCom = DbAccessor.createHibernateDbAccessor();
            logger.debug("DB connection established.");
            return true;
        } catch (DbAccessException ex) {
            logger.error("Error on testing database connection.", ex);
            return false;
        }
    }

    /**
     * Initializes and starts the instance that handles the incoming and outgoing messages of the
     * LPM and triggers the respective reactions. This will be carried out in a daemon thread that
     * runs until terminated.
     *
     * @return True if init and start was successful
     */
    private static boolean startMsgHandler() {
        try {
            logger.debug(String.format("Starting the %s.", MsgHandler.class.getSimpleName()));
            if (cmd.isBaseConnectionDataAvailable()) {
                msgHandler = new MsgHandler(50, false, cmd.getLpmAddress(), cmd.getComLayerAddress());
            } else if (cmd.isAdvancedConnectionDataAvailable()) {
                msgHandler = new MsgHandler(50, false,
                        cmd.getXmppUser(), cmd.getXmppPsw(),
                        cmd.getXmppServerAdr(), ValueConverterUtil.toInteger(cmd.getXmppServerPort()),
                        cmd.getTsbUser());
            } else {
                throw new LpmComInterfaceException("Connection data incomplete.");
            }
            msgHandler.start();
            logger.debug(String.format("%s started.", MsgHandler.class.getSimpleName()));
            return true;
        } catch (URISyntaxException | IOException | LpmComInterfaceException ex) {
            logger.error("Exception on starting the message handler.", ex);
            return false;
        }
    }

    /**
     * Starts the instance that manages the recommendation creation workflow.
     *
     * @return True if init and start was successful
     */
    private static boolean startWorkflowHandler() {
        try {
            logger.debug(String.format("Starting the %s.", WorkflowHandler.class.getSimpleName()));
            wfHandler = new WorkflowHandler(1000, false, dbCom,
                    lmsComClass, lmsComParameters);
            wfHandler.start();
            logger.debug(String.format("%s started.", WorkflowHandler.class.getSimpleName()));
            return true;
        } catch (Exception ex) {
            logger.error("Exception on starting the workflow handler.", ex);
            return false;
        }
    }

    /**
     * Starts the instance that monitors the user's console input.
     *
     * @return True if init and start was successful
     */
    private static boolean startUserInputHandler() {
        try {
            logger.debug(String.format("Starting the %s.", UserInputHandler.class.getSimpleName()));
            inputHandler = new UserInputHandler(1000, true);
            inputHandler.start();
            logger.debug(String.format("%s started.", UserInputHandler.class.getSimpleName()));
            return true;
        } catch (Exception ex) {
            logger.error("Exception on starting the user input handler.", ex);
            return false;
        }
    }

    /**
     * Starts a new LPM processor instance that carries out a pre-processing job and forwards its
     * results to the INTUITEL Engine.
     *
     * @param loId ID of the LO which triggered the job
     * @param learnerId ID of the learner in scope
     * @param mId Message ID of the learner update to which this job is assigned to
     * @param time Time of learner update
     * @return True if start was successfull
     */
    public static boolean startWorker(String loId, String learnerId, String mId, long time) {
        try {
            workerExecutor.submit(new WorkerStarter(loId, learnerId, mId, time,
                    engineComClass, engineComParameters,
                    slomComClass, slomComParameters,
                    lmsComClass, lmsComParameters,
                    cmd.getMergedLsoFlag()));
            return true;
        } catch (Exception ex) {
            logger.error("Exception on starting a new LPM processor.", ex);
            return false;
        }
    }

    /**
     * Calls a System.in.read() that keeps the application alive until the user presses a button to
     * shutdown the LPM.
     */
    @SuppressWarnings("SleepWhileInLoop")
    private static void keepAlive() {
        try {
            boolean keepAlive = true;
            while (keepAlive) {
                if (inputHandler.isAlive() && wfHandler.isAlive() && msgHandler.isAlive()) {
                    Thread.sleep(1000);
                } else {
                    if (!inputHandler.isAlive() && !inputHandler.isShutdown()) {
                        logger.error("Input handler thread died unexpectedly.");
                    }
                    if (!wfHandler.isAlive()) {
                        logger.error("Workflow handler thread died unexpectedly.");
                    }
                    if (!msgHandler.isAlive()) {
                        logger.error("Message handler thread died unexpectedly.");
                    }
                    keepAlive = false;
                }
            }
        } catch (InterruptedException ex) {
            logger.error("Critical Exception at LPM runtime. LPM will be shut down." + ex);
        }
    }

    /**
     * Shuts down the LPM securly.
     */
    public static void shutdown() {
        try {
            if (workerExecutor != null) {
                workerExecutor.shutdown();
                logger.info("Worker executor shut down.");
            }
        } catch (Exception ex) {
            logger.error(String.format("Shutdown of worker executor failed: %s", ex.getMessage()), ex);
        }
        try {
            if (dbCom != null) {
                dbCom.shutdown();
                logger.info("Closed DB connection.");
            }
        } catch (Exception ex) {
            logger.error(String.format("Shutdown of DB connection failed: %s", ex.getMessage()), ex);
        }
        try {
            if (inputHandler != null && inputHandler.isAlive()) {
                inputHandler.shutdown();
                inputHandler.waitForShutdownCompletion();
            }
        } catch (InterruptedException ex) {
            logger.error(String.format("Shutdown of input handler failed: %s", ex.getMessage()), ex);
        }
        try {
            if (wfHandler != null && wfHandler.isAlive()) {
                wfHandler.shutdown();
                wfHandler.waitForShutdownCompletion();
            }
        } catch (InterruptedException ex) {
            logger.error(String.format("Shutdown of workflow handler failed: %s", ex.getMessage()), ex);
        }
        try {
            if (msgHandler != null && msgHandler.isAlive()) {
                msgHandler.shutdown();
                msgHandler.waitForShutdownCompletion();
            }
        } catch (InterruptedException ex) {
            logger.error(String.format("Shutdown of msg handler failed in %s.", ex), ex);
        }
        logger.info("Shutdown complete.");
    }
}
