/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Kevin Fuchs, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-08
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmComInterface.communicators;

import eu.intuitel.comm.entities.exceptions.IntuitelException;
import eu.intuitel.lpmComInterface.Main;
import eu.intuitel.lpmComInterface.wrappers.IIncomingMsg;
import eu.intuitel.lpmComInterface.wrappers.IOutgoingMsg;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Abstract base class for asynchronous communicators.
 *
 * @author fhe
 */
public abstract class BaseCommunicator {

    /**
     * Logger for status information
     */
    private static final Logger logger = LoggerFactory.getLogger(BaseCommunicator.class);
    /**
     * Time in ms after which request is chanceled interally if the message was not received until
     * then.
     */
    protected final static long TIMEOUTTIME = 5000; 
    /**
     * Element which was send as response by the counterpart.
     */
    private IIncomingMsg response;

    /**
     * Sets the response of the request.
     *
     * @param response Response that has been received, structured as an IIntuitelContentElement
     */
    public void setResponse(IIncomingMsg response) {
        this.response = response;
        synchronized (this) {
            this.notify();
        }
    }

    /**
     * Sends the provided {@code msg} and pauses the sending thread until a response is received or
     * the response times out.
     *
     * @param msg Message to be send
     * @return Null if no response was received.
     */
    @SuppressWarnings("ThrowableResultIgnored")
    protected IIncomingMsg sendAndReceiveAsyncResponse(IOutgoingMsg msg) {
        logger.debug("Sending request: " + msg.getClass().getSimpleName());
        try {
            synchronized (this) {
                Main.getMsgHandler().sendMsg(msg, this);
                if(this.response == null) {
                    this.wait(TIMEOUTTIME);
                }
            }
        } catch (InterruptedException ex) {
            logger.warn(String.format("Communicator '%s' interrupted.",
                    this.getClass().getSimpleName()), ex);
        } catch (IntuitelException ex) {
            logger.error(String.format("Sending '%s'-msg failed: %s, %s",
                    msg.getClass().getSimpleName(), ex.getError(), ex.getInnerException().getMessage()), ex);
        }
        if (this.response == null) {
            logger.warn(String.format("No response received for: %s, %s",
                    msg.getClass().getSimpleName(), msg.getMID()));
        }
        return this.response;
    }

    /**
     * Sends the given message without expecting a response.
     *
     * @param msg Message to be send.
     */
    protected void send(IOutgoingMsg msg) {
        logger.debug("Sending message: " + msg.getClass().getSimpleName());
        try {
            Main.getMsgHandler().sendMsg(msg);
        } catch (IntuitelException ex) {
            logger.error(String.format("Sending '%s'-msg failed.",
                    msg.getClass().getSimpleName()), ex);
        }
    }

    @Override
    public String toString() {
        return this.getClass().getName();
    }
}
