/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-08
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmComInterface.communicators;

import eu.intuitel.lpmGeneralInterfaces.ISlomCommunicator;
import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * SLOM communicator instance to access the SLOM data from a local repository.
 *
 * @author fhe
 */
public class SlomLocalCommunicator implements ISlomCommunicator {

    private static final Logger logger = LoggerFactory.getLogger(SlomLocalCommunicator.class);
    private static final String CCM_TEMPLATE = "intui_CCM.owl";
    private static final String CM_TEMPLATE = "intui_CM.owl";
    protected static Map<String, String[]> indices = new HashMap<>();
    protected static List<String> loadedAddresses = new ArrayList<>();
    protected final String LOCALREPO;

    /**
     * Creates a new communicator for a local SLOM repository.
     *
     * @param address Path of the local SLOM repository.
     */
    public SlomLocalCommunicator(String address) {
        this.LOCALREPO = address;
        createIndices(address);
    }

    @Override
    public ISlomRefContainer getSlomDataForLo(String loId) {
        if (StringUtils.isBlank(loId)) {
            logger.warn("Could not retrieve SLOM data. The specified Learning Object ID is blank.");
            return null;
        }
        if (indices.containsKey(loId)) {
            String[] files = indices.get(loId);
            return new SlomFileRefContainer(new File(files[0]), new File(files[1]));
        }
        logger.warn(String.format("No SLOM data for requested loID '%s' available.", loId));
        return null;
    }

    /**
     * Reloads the indices from the local SLOM repo.
     */
    public void reloadIndices() {
        indices.clear();
        createIndices(LOCALREPO);
    }

    /**
     * Loads the SLOM files from the specified local repository and stores the LO ID for indexing.
     *
     * @param path Path of the local repository
     */
    private void createIndices(String path) {
        if(loadedAddresses.contains(path)) {
            //NOP, path already loaded
            return;
        }
        
        List<String> ccmFiles = new ArrayList<>();
        getCcmFilesAt(new File(path), ccmFiles);

        List<String> cmFiles = new ArrayList<>();
        for (String ccmFile : ccmFiles) {
            cmFiles.add(getAccompanyingCm(ccmFile));
        }

        for (int i = 0; i < ccmFiles.size(); i++) {
            String ccm = ccmFiles.get(i);
            String cm = cmFiles.get(i);
            if (!StringUtils.isBlank(cm)) {
                extendIndicesFor(ccm, cm);
            }
        }
        
        loadedAddresses.add(path);
    }

    /**
     * Recursively searches for CCM files under the given {@code path}.
     *
     * @param startPath Path to inspect
     */
    private void getCcmFilesAt(File startPath, List<String> ccmFiles) {
        if (!startPath.exists() || !startPath.isDirectory()) {
            logger.error("Local SLOM repository directory could not be found.");
            return;
        }
        for (File file : startPath.listFiles()) {
            if (file.isFile()) {
                if (file.getName().equalsIgnoreCase(CCM_TEMPLATE)) {
                    ccmFiles.add(file.getPath());
                    logger.debug("CCM found in local repository: " + file);
                }
            } else if (file.isDirectory()) {
                getCcmFilesAt(file, ccmFiles);
            }
        }
    }

    /**
     * Returns the CM that accompanies the given CCM.
     *
     * @param ccmFile CCM file in scope
     * @return Null if accompanying CM could not be found
     */
    private String getAccompanyingCm(String ccmFile) {
        String cmFile = ccmFile.replace(CCM_TEMPLATE, CM_TEMPLATE);
        File file = new File(cmFile);
        if (file.exists()) {
            return cmFile;
        } else {
            logger.warn("No accompanying CM for CCM '" + ccmFile + "'.");
            return null;
        }
    }

    /**
     * Returns the contained KOs as indices for the given CCM-CM-combination.
     *
     * @param ccm CCM in scope
     * @param cm CM that is accompanying the CCM
     * @return Empty map if no KOs could be found
     */
    private void extendIndicesFor(String ccm, String cm) {
        try {
            String ccmContent = (Charset.forName("UTF-8")).decode(ByteBuffer.wrap(
                    Files.readAllBytes(Paths.get(ccm)))).toString();
            for (String ko : getIndices(ccm, ccmContent)) {
                indices.put(ko, new String[]{ccm, cm});
            }
            logger.debug("Added indices for CCM: " + ccm);
        } catch (IOException ex) {
            logger.error("Cannot read CCM at '" + ccm + "'.", ex);
        }
    }

    /**
     * Returns the KO IRIs which are contained in the given CCM.
     *
     * @param ccm CCM in scope
     * @param ccmContent Content of the CCM file
     * @return Empty List if nothing was found
     */
    private List<String> getIndices(String ccm, String ccmContent) {
        List<String> kos = new ArrayList<>();
        String base = getBaseIri(ccmContent);
        if (StringUtils.isBlank(base)) {
            logger.error("Base IRI could not be determined for: " + ccm + ". CCM will be ignored.");
            return kos;
        }
        for (String ko : getKoDefinitions(ccmContent)) {
            kos.add(ko);
        }
        return kos;
    }

    /**
     * Looks for the base IRI of the CCM.
     *
     * @param ccmContent Content of the CCM file
     * @return Null if nothing was found
     */
    protected static String getBaseIri(String ccmContent) {
        //xml:base="http://www.intuitel.eu/public/testcase01/intui_CCM.owl"
        Pattern pattern = Pattern.compile("(xml:base=\")(.*.owl)");
        Matcher matcher = pattern.matcher(ccmContent);
        if (matcher.find()) {
            return matcher.group(2);
        }
        return null;
    }

    /**
     * Looks for KO definitions in the given CCM content.
     *
     * @param ccmContent Content of the CCM file
     * @return Empty list if nothing was found
     */
    protected static List<String> getKoDefinitions(String ccmContent) {
        List<String> kos = new ArrayList<>();
        //<DataProperty abbreviatedIRI="intui_PO:isLinkedWithLmsElement"/>
        //<NamedIndividual IRI="#Ko11"/>
        //<Literal datatypeIRI="&rdf;PlainLiteral">LO4721</Literal>
        Pattern pattern = Pattern.compile(
                "(abbreviatedIRI=\"intui_PO:isLinkedWithLmsElement\"/>\\s*<NamedIndividual IRI=\")"
                + "(#\\w+)"
                + "(\"/>\\s*<Literal datatypeIRI=\"&xsd;string\">)"
                + "(.+)"
                + "(</Literal>)");
        Matcher matcher = pattern.matcher(ccmContent);
        while (matcher.find()) {
            //String koAbbrevIri = matcher.group(2);
            String koLmsId = matcher.group(4);
            kos.add(koLmsId);
        }
        //NOTE: This only looks for URIs, but the LMS provides LMS IDs
        //<Class abbreviatedIRI="intui_PO:KnowledgeObject"/>
        //<NamedIndividual IRI="#Ko02"/>
        //Pattern pattern = Pattern.compile("(abbreviatedIRI=\"intui_PO:KnowledgeObject\"/>\\s*<NamedIndividual IRI=\")(.+)(\")");
        //Matcher matcher = pattern.matcher(ccmContent);
        //while (matcher.find()) {
        //  kos.add(matcher.group(2));
        //}
        return kos;
    }
}
