/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-26
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmComInterface.runnables;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Abstract base class for handlers that encapsulate threading issues.
 *
 * @author fhe
 */
public abstract class BaseHandler implements Runnable {

    private static final Logger logger = LoggerFactory.getLogger(BaseHandler.class);
    /**
     * Flag that, if set to true, triggers the exit from the while-loop and thus the thread to exit.
     */
    private volatile Thread blinker;
    /**
     * Thread in which this instance is running.
     */
    private Thread thread;
    /**
     * Duration of ms for which this handler sleeps between cycles.
     */
    private final long SLEEPDURATION;
    /**
     * Flag to state if the handler Thread is a daemon.
     */
    private final boolean ISDAEMON;

    /**
     * Constructor.
     *
     * @param sleepDuration Duration of ms for which this handler sleeps between cycles
     * @param isDaemon The thread being created is registerd as a daemon thread
     */
    public BaseHandler(long sleepDuration, boolean isDaemon) {
        this.SLEEPDURATION = sleepDuration;
        this.ISDAEMON = isDaemon;
    }

    /**
     * Starts the Thread.
     */
    public void start() {
        thread = new Thread(this);
        thread.setDaemon(ISDAEMON);
        thread.setName("INTUITEL-" + this.getClass().getSimpleName());
        thread.start();
        blinker = thread;
    }

    /**
     * Method which is called in each cycle.
     *
     * @return False if the thread should exit
     */
    protected abstract boolean whileLoop();

    /**
     * Method which is called before the Thread finishes the run()-method and exits.
     */
    protected abstract void beforeShutdown();

    @Override
    @SuppressWarnings("SleepWhileInLoop")
    public void run() {
        try {
            while (blinker == thread) {
                if (!whileLoop()) {
                    shutdown();
                } else {
                    Thread.sleep(SLEEPDURATION);
                }
            }
        } catch (Exception ex) {
            logger.error(String.format("Exception in handler %s. Shutting down thread %s.",
                    this.getClass().getSimpleName(), thread.getName()), ex);
        }
        logger.debug("Executing beforeShutdown-procedure: " + this.getClass().getSimpleName());
        beforeShutdown();
        logger.info(String.format("%s exited.", this.getClass().getSimpleName()));
    }

    /**
     * Terminates the Thread running the handler.
     */
    public void shutdown() {
        logger.debug("Shutting down: " + this.getClass().getName());
        blinker = null;
    }

    /**
     * Waits until handler threads terminates or timeout occurs.
     */
    public void waitForShutdownCompletion() throws InterruptedException {
        thread.join(2500);
        if (!thread.isAlive()) {
            logger.debug("Shutdown complete: " + this.getClass().getName());
            return;
        }
        logger.debug("Shutdown failed, interrupting thread.");
        thread.interrupt();
        if (!thread.isAlive()) {
            logger.debug("Shutdown complete: " + this.getClass().getName());
        } else {
            logger.error("Could not shutdown thread: " + this.getClass().getName());
        }
    }

    /**
     * Checks whether the Thread running this handler is alive or not.
     *
     * @return True if alive
     */
    public boolean isAlive() {
        return thread.isAlive();
    }
}
