/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Kevin Fuchs, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-08
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmComInterface.runnables;

import eu.intuitel.CommLayer.CommLibrary.LpmServiceConnection;
import eu.intuitel.comm.entities.slomRepo.CcmRef;
import eu.intuitel.comm.entities.slomRepo.CmRef;
import eu.intuitel.comm.entities.slomRepo.SlomData;
import eu.intuitel.comm.entities.slomRepo.SlomUpdate;
import eu.intuitel.comm.entities.exceptions.IntuitelException;
import eu.intuitel.comm.entities.lms.Learner;
import eu.intuitel.comm.entities.lms.Learners;
import eu.intuitel.comm.entities.lms.Learners.Learner.VisitedLo;
import eu.intuitel.comm.services.ILpmListener;
import eu.intuitel.lpmComInterface.Main;
import eu.intuitel.lpmComInterface.communicators.BaseCommunicator;
import eu.intuitel.lpmComInterface.wrappers.IIncomingMsg;
import eu.intuitel.lpmComInterface.wrappers.LmsProfileIn;
import eu.intuitel.lpmComInterface.wrappers.UseEnvironmentIn;
import java.util.List;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import eu.intuitel.lpmComInterface.wrappers.IOutgoingMsg;
import eu.intuitel.lpmComInterface.wrappers.SlomRequestIn;
import eu.intuitel.lpmComInterface.wrappers.UsePerformanceIn;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.concurrent.ConcurrentHashMap;
import org.apache.commons.lang.StringUtils;

/**
 * Handles sending and receiving of messages for the LPM.
 *
 * @author fhe
 */
public class MsgHandler extends BaseHandler implements ILpmListener {

    /**
     * Logger interface.
     */
    private static final Logger logger = LoggerFactory.getLogger(MsgHandler.class);
    /**
     * Connection service to send and receive messages.
     */
    private final LpmServiceConnection service;
    /**
     * Map to store which communicator waits for which mID response.
     */
    private final ConcurrentHashMap<String, BaseCommunicator> pendingRequests = new ConcurrentHashMap<>();

    /**
     * Constructor - Initializes msg handler with a connection to a basic commmunication layer.
     *
     * @param sleepDuration Duration of sleep during cycles
     * @param isDaemon Thread is started as a daemon true/false
     * @param lpmAddress Address of the this LPM to receive responses from the CL
     * @param clAddress Address of the basic communication layer
     * @throws URISyntaxException
     * @throws IOException
     */
    public MsgHandler(long sleepDuration, boolean isDaemon, URL lpmAddress, URL clAddress) throws URISyntaxException, IOException {
        this(sleepDuration, isDaemon);
        service.OpenBasic(lpmAddress.toURI(), clAddress.toURI());
    }

    /**
     * Constructor - Initialized msg handler with an advanced connection.
     *
     * @param sleepDuration Duration of sleep during cycles
     * @param isDaemon Thread is started as a daemon true/false
     * @param xmppUser Username for the XMPP server
     * @param xmppPsw Password for the XMPP server
     * @param xmppServer Address of the XMPP server
     * @param xmppServerPort Port of the XMPP server
     * @param tsbUsername Username for the TIE Smart Bridge
     * @throws IOException
     */
    public MsgHandler(long sleepDuration, boolean isDaemon, String xmppUser, String xmppPsw, String xmppServer, int xmppServerPort, String tsbUsername) throws IOException {
        this(sleepDuration, isDaemon);
        service.OpenAdvanced(xmppUser, xmppPsw, xmppServer, xmppServerPort, tsbUsername);
    }

    /**
     * Private constructor for base init.
     *
     * @param sleepDuration Duration of sleep during cycles
     * @param isDaemon Thread is started as a daemon true/false
     */
    @SuppressWarnings("LeakingThisInConstructor")
    private MsgHandler(long sleepDuration, boolean isDaemon) {
        super(sleepDuration, isDaemon);
        service = LpmServiceConnection.getInstance();
        service.setLpmListener(this);
        //NOTE: Leaking 'this' is ok here, because it only registers the MsgHandler as listener for incoming msgs. The connection is established afterwards.
    }

    /**
     * Sends a new message to another INTUITEL component, expecting an asynchronous response.
     *
     * @param msg Message to be send
     * @param sender Sender of the message, which will receive the response afterwards.
     * @throws IntuitelException
     */
    public void sendMsg(IOutgoingMsg msg, BaseCommunicator sender) throws IntuitelException {
        pendingRequests.put(msg.getMID(), sender);
        sendMsg(msg);
    }

    /**
     * Sends a new message to another INTUITEL component, not expecting an asynchronous response.
     *
     * @param msg Message to be send
     * @throws IntuitelException
     */
    public void sendMsg(IOutgoingMsg msg) throws IntuitelException {
        msg.send(service);
    }

    @Override
    protected boolean whileLoop() {
        //Note: Nothing to do here, because the listeners to the work of assigning the values. 
        //The loop is only necessary to keep the service running.
        return true;
    }

    @Override
    protected void beforeShutdown() {
        service.Close();
    }

    @Override
    public void onNewLmsProfile(eu.intuitel.comm.entities.lms.LmsProfile lp) {
        logger.debug("Received LMS profile response.");
        LmsProfileIn msg = new LmsProfileIn(lp);
        if(msg.getMID() == null) {
            //This only works, because for LmsProfile, there always is only one request pending.
            String mId = pendingRequests.keys().nextElement();
            BaseCommunicator communicator = pendingRequests.get(mId);
            pendingRequests.remove(mId);
            communicator.setResponse(msg);
        } else {
            forwardMsg(new LmsProfileIn(lp));
        }
    }

    @Override
    public void onNewPullLearnerUpdate(List<Learner> list) {
        logger.debug("Received LMS learner update pull response.");
        for (Learner learner : list) {
            Main.startWorker(learner.getLoId(), learner.getUId(), learner.getMId(), learner.getTime());
        }
    }

    @Override
    public void onNewPushLearnerUpdate(List<Learner> list) {
        logger.debug("Received LMS learner update push response.");
        for (Learner learner : list) {
            Main.startWorker(learner.getLoId(), learner.getUId(), learner.getMId(), learner.getTime());
        }
    }

    @Override
    public void onNewPushWithLearnerPolling(List<Learners> list) {
        logger.debug("Received LMS learner update push polling response.");
        for (Learners learners : list) {
            for (Learners.Learner learner : learners.getLearner()) {
                if (learner.getVisitedLo().size() > 1) {
                    logger.debug("More than one visited LOs in learner update polling not yet supported.");
                    //TODO: Lowprio - Special case for polling
                    //A learner may have accessed more than two LOs during one polling intervall. 
                    //New LpmProcessor-constructor needed for this case.
                    //However, if the polling intervall is low (which it should be) this can be ignored.
                } else {
                    VisitedLo visitedLo = learner.getVisitedLo().get(0);
                    Main.startWorker(visitedLo.getLoId(), learner.getUId(), learners.getMId(), visitedLo.getTime());
                }
            }
        }
    }

    @Override
    public void onNewUseEnvResponse(List<eu.intuitel.comm.entities.lms.UseEnv> list) {
        logger.debug("Received LMS USE environment response.");
        for (eu.intuitel.comm.entities.lms.UseEnv use : list) {
            forwardMsg(new UseEnvironmentIn(use));
        }
    }

    @Override
    public void onNewUsePerfResponse(List<eu.intuitel.comm.entities.lms.UsePerf> list) {
        logger.debug("Received LMS USE performance response.");
        for (eu.intuitel.comm.entities.lms.UsePerf use : list) {
            forwardMsg(new UsePerformanceIn(use));
        }
    }

    @Override
    public void onNewSlomDataResponse(List<SlomData> slomDatas) {
        logger.debug("Received a SLOM data response.");
        for (SlomData data : slomDatas) {
            forwardMsg(new SlomRequestIn(data));
        }
    }

    @Override
    public void onNewSlomUpdateResponse(List<SlomUpdate> slomUpdates) {
        logger.debug("Received a SLOM update notification.");
        for (SlomUpdate data : slomUpdates) {
            for (CmRef cmRef : data.getCmRef()) {
                Main.getWorkflowHandler().updateCm(cmRef.getIri());
            }
            for (CcmRef ccmRef : data.getCcmRef()) {
                Main.getWorkflowHandler().updateCcm(ccmRef.getIri());
            }
        }
    }

    @Override
    public void onNewRRLearnerUpdate(List<eu.intuitel.comm.entities.rr.Learner> learners) {
        logger.debug("Received RR learner update.");
        for (eu.intuitel.comm.entities.rr.Learner learner : learners) {
            Main.startWorker(learner.getLoId(), learner.getUId(), learner.getMId(), learner.getTime());
        }
    }

    /**
     * Forwards the given msg to the requesting communicator.
     *
     * @param msg Message to be forwarded.
     */
    private void forwardMsg(IIncomingMsg msg) {
        if (msg != null && !StringUtils.isBlank(msg.getMID())) {
            BaseCommunicator communicator = pendingRequests.get(msg.getMID());
            if (communicator == null) {
                logger.warn(String.format("Received response for unlisted request: %s, %s",
                        msg.getClass().getSimpleName(), msg.getMID()));
            } else {
                pendingRequests.remove(msg.getMID());
                communicator.setResponse(msg);
            }
        } else {
            logger.error("Cannot formward a null msg.");
        }
    }
}
