/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Contributor: Ovidiu Serban, University of Reading, www.isr.reading.ac.uk
 * Version: [0.01]
 * Date: 2013-11-08
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmComInterface.runnables;

import eu.intuitel.db.access.DbAccessException;
import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.lpmComInterface.utils.GenericsUtil;
import eu.intuitel.lpmGeneralInterfaces.IEngineCommunicator;
import eu.intuitel.lpmGeneralInterfaces.ILmsCommunicator;
import eu.intuitel.lpmGeneralInterfaces.ISlomCommunicator;
import eu.intuitel.lpmProcessor.LpmProcessor;
import java.lang.reflect.InvocationTargetException;
import org.apache.commons.lang.Validate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Runnable to start workers and their communication interfaces in a separate Thread.
 *
 * @author fhe
 */
public class WorkerStarter implements Runnable {

    private static final Logger logger = LoggerFactory.getLogger(WorkerStarter.class);
    private final Class<? extends IEngineCommunicator> engineComClass;
    private final Class<? extends ISlomCommunicator> slomComClass;
    private final Class<? extends ILmsCommunicator> lmsComClass;
    private final Object[] engineComParameters;
    private final Object[] slomComParameters;
    private final Object[] lmsComParameters;
    private final String loId;
    private final String learnerId;
    private final String mId;
    private final long time;
    private final boolean mergedLso;
    private LpmProcessor worker;

    /**
     * Create a new instance of a WorkerStarter
     *
     * @param loId Unique identifier of the LO which triggered the worker
     * @param learnerId Unique identifier of the learner for whom the worker should be started
     * @param mId Unique identifier of the ID which triggered the worker
     * @param time Time the LO was accessed
     * @param engineComClass Class of the ILpmEngineCommunicator to be used for communicating with
     * the INTUITEL Engine
     * @param engineComParameters Parameters to be used for instantiating the Engine communicator
     * @param slomComClass Class of the ILpmSlomCommunicator to be used for communicating with the
     * SLOM repository
     * @param slomComParameters Parameters to be used for instantiating the SLOM communicator
     * @param lmsComClass LMS communicator class
     * @param lmsComParameters LMS communicator class parameters
     * @param mergedLso Flag indicating if the LSO should be merged or raw
     */
    public WorkerStarter(String loId, String learnerId, String mId, long time,
            Class<? extends IEngineCommunicator> engineComClass, Object[] engineComParameters,
            Class<? extends ISlomCommunicator> slomComClass, Object[] slomComParameters,
            Class<? extends ILmsCommunicator> lmsComClass, Object[] lmsComParameters,
            boolean mergedLso) {
        //Validate.notEmpty(loId); May be empty or null on "session end" or "refresh"
        Validate.notEmpty(learnerId);
        Validate.notEmpty(mId);
        Validate.notNull(engineComClass);
        Validate.notNull(engineComParameters);
        Validate.notNull(slomComClass);
        Validate.notNull(slomComParameters);
        Validate.notNull(lmsComClass);
        Validate.notNull(lmsComParameters);
        this.engineComClass = engineComClass;
        this.engineComParameters = engineComParameters;
        this.slomComClass = slomComClass;
        this.slomComParameters = slomComParameters;
        this.lmsComClass = lmsComClass;
        this.lmsComParameters = lmsComParameters;
        this.loId = loId;
        this.learnerId = learnerId;
        this.mId = mId;
        this.time = time;
        this.mergedLso = mergedLso;
    }

    @Override
    public void run() {
        try {
            //NOTE: Communicators created in the run()-method to assign them to the Thread of the worker
            IEngineCommunicator engineCom = (IEngineCommunicator) GenericsUtil.createWithParameters(engineComClass, engineComParameters);
            if (engineCom == null) {
                logger.error("Creation of Engine communicator failed.");
                return;
            }
            ISlomCommunicator slomCom = (ISlomCommunicator) GenericsUtil.createWithParameters(slomComClass, slomComParameters);
            if (slomCom == null) {
                logger.error("Creation of SLOM communicator failed.");
                return;
            }
            ILmsCommunicator lmsCom = (ILmsCommunicator) GenericsUtil.createWithParameters(lmsComClass, lmsComParameters);
            if (lmsCom == null) {
                logger.error("Creation of LMS communicator failed.");
                return;
            }
            worker = new LpmProcessor(loId, learnerId, mId, time, slomCom, engineCom, lmsCom, DbAccessor.createHibernateDbAccessor(), mergedLso);
            logger.debug(String.format("Starting worker for loId=%s, learnerId=%s, mId=%s", loId, learnerId, mId));
            worker.work();
            logger.debug(String.format("Finished worker for loId=%s, learnerId=%s, mId=%s", loId, learnerId, mId));
        } catch (NoSuchMethodException | InstantiationException | IllegalAccessException | IllegalArgumentException | InvocationTargetException ex) {
            logger.error("Creation of LpmProcessor failed.", ex);
        } catch (DbAccessException ex) {
            logger.error("Creation of LpmProcessor failed. Error on DB connection.", ex);
        }
    }

    /**
     * Returns the worker of this WorkerStarter.
     *
     * @return Null if worker has not yet been created.
     */
    public LpmProcessor getWorker() {
        return this.worker;
    }
}
