/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Ovidiu Serban, University of Reading, www.isr.reading.ac.uk
 * Contributor: Kevin Fuchs, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-26
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmComInterface.runnables;

import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.lpmComInterface.utils.GenericsUtil;
import eu.intuitel.lpmComInterface.utils.LpmComInterfaceException;
import eu.intuitel.lpmComInterface.wrappers.LmsProfileIn;
import eu.intuitel.lpmGeneralInterfaces.ILmsCommunicator;
import eu.intuitel.lpmProcessor.LpmProcessor;
import eu.intuitel.lpmProcessor.utils.SlomMapCache;
import eu.intuitel.types.ComStyle;
import eu.intuitel.types.LoreLvl;
import eu.intuitel.types.MediaLvl;
import eu.intuitel.types.NameFormality;
import eu.intuitel.types.UseLvl;
import java.util.Map;
import org.apache.commons.lang.Validate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Class which manages the LPM workflow.
 *
 * @author fhe
 */
public class WorkflowHandler extends BaseHandler {

    private static final Logger logger = LoggerFactory.getLogger(WorkflowHandler.class);
    private final DbAccessor dbCom;
    private final ILmsCommunicator lmsCom;
    private final SlomMapCache slomCache;
    private boolean polling;

    /**
     * Constructor to initiate the LPM workflow.
     *
     * @param sleepDuration Sleep duration between cycles - Polling of learner updates
     * @param isDaemon Set true to make the handler run as a daemon thread
     * @param dbCom Object to use for accessing the database
     * @param lmsComClass Class of the ILmsCommunicator to be used for communicating with the LMS
     * @param lmsComParameters Parameters to be used for instantiating the LMS communicator
     * @throws java.lang.Exception
     */
    public WorkflowHandler(long sleepDuration, boolean isDaemon, DbAccessor dbCom,
            Class<? extends ILmsCommunicator> lmsComClass, Object[] lmsComParameters) throws Exception {
        super(sleepDuration, isDaemon);
        Validate.notNull(lmsComClass);
        Validate.notNull(lmsComParameters);
        this.dbCom = dbCom;
        this.lmsCom = (ILmsCommunicator) GenericsUtil.createWithParameters(lmsComClass, lmsComParameters);
        initLmsProfileData();
        slomCache = (new LpmProcessor("dummy", "dummy", "dummy", 0, null, null, null, DbAccessor.createCachedDbAccessor(), false)).getCache();
        if (slomCache == null) {
            throw new LpmComInterfaceException("SLOM cache was not initiated properly.");
        }
    }

    /**
     * Requests and stores the LMS profile data.
     *
     * @throws Exception
     */
    private void initLmsProfileData() throws Exception {
        Map<String, String> lmsProfile = lmsCom.getLmsProfile();
        if (lmsProfile == null) {
            throw new LpmComInterfaceException("Could not acquire LMS profile data. Terminating LPM workflow.");
        }
        String lmsId = LmsProfileIn.getLmsId(lmsProfile);
        String lmsName = LmsProfileIn.getLmsName(lmsProfile);
        Integer lmsNameForm = LmsProfileIn.getLearnerNameFormality(lmsProfile);
        String lmsType = LmsProfileIn.getLmsType(lmsProfile);
        String lmsMediaLvl = LmsProfileIn.getLmsMediaLevel(lmsProfile);
        Integer lmsLoreLvl = LmsProfileIn.getLmsLoreLevel(lmsProfile);
        Integer lmsUseLvl = LmsProfileIn.getLmsUseLevel(lmsProfile);
        Integer comStyle = LmsProfileIn.getLmsComStyle(lmsProfile);
        dbCom.storeLmsProperties(
                lmsId,
                lmsName,
                lmsType,
                lmsMediaLvl != null ? MediaLvl.fromString(lmsMediaLvl) : null,
                lmsLoreLvl != null ? LoreLvl.fromInt(lmsLoreLvl) : null,
                lmsUseLvl != null ? UseLvl.fromInt(lmsUseLvl) : null,
                lmsNameForm != null ? NameFormality.fromInt(lmsNameForm) : null,
                comStyle != null ? ComStyle.fromInt(comStyle) : null);
        if (2 == comStyle) {
            polling = true;
        }
        logger.info(String.format("LMS data stored: %s", lmsProfile.toString()));
    }

    @Override
    protected boolean whileLoop() {
        if (polling) {
            //Depending on comStyle, an learner updates need to be requested by the LPM.
            lmsCom.sendLearnerUpdate();
        }
        return true;
    }

    @Override
    protected void beforeShutdown() {
        //NOP
    }

    /**
     * Triggers an update of SLOM data in the cache.
     *
     * @param iri IRI of the CM to be updated
     */
    public void updateCm(String iri) {
        slomCache.updateCm(iri);
    }

    /**
     * Triggers an update of SLOM data in the cache.
     *
     * @param iri IRI of the CCM to be updated
     */
    public void updateCcm(String iri) {
        slomCache.updateCcm(iri);
    }
}
