/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-05-08
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmComInterface.utils;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Level;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Helper class to parse commandline arguments.
 *
 * @author fhe
 */
public class CmdParser {

    private static final Logger logger = LoggerFactory.getLogger(CmdParser.class);
    // <editor-fold defaultstate="collapsed" desc="CMD arguments">
    private static final String LPMADDRESS_ARG = "lpmAddress";
    private static final String CLADDRESS_ARG = "clAddress";
    private static final String SLOMREPOADDRESS_ARG = "slomRepoAddress";
    private static final String TEST_ARG = "test";
    private static final String HELP_ARG = "help";
    private static final String XMPP_USER_ARG = "xmppUser";
    private static final String XMPP_PSW_ARG = "xmppPsw";
    private static final String XMPP_SERVERADR_ARG = "xmppServerAdr";
    private static final String XMPP_SERVERPORT_ARG = "xmppServerPort";
    private static final String TSB_USER_ARG = "tsbUser";
    private static final String LOGGINGLVL_INTUITEL_ARG = "intitelLogLvl";
    private static final String LOGGINGLVL_HTTP_ARG = "httpLogLvl";
    private static final String LOGGINGLVL_DB_ARG = "dbLogLvl";
    private static final String MERGEDLSO_ARG = "mergedLso";
    // </editor-fold>
    // <editor-fold defaultstate="collapsed" desc="CMD argument description texts">
    private static final String LPMADDRESS_ARG_TXT = "URL under which this LPM instance should be started.";
    private static final String CLADDRESS_ARG_TXT = "URL of where the communication layer is running.";
    private static final String SLOMREPOADDRESS_ARG_TXT = "A SLOM repository may be located locally or remotely. "
            + "If using a local repository, please specify the (absolute/relative) path under which the SLOM files are stored. "
            + "If using a remote repository, please specifiy the URL of the reposiotry.";
    private static final String TEST_ARG_TXT = "Boolean value notifying the LPM to run in test mode, i.e. to exit immidiately after init.";
    private static final String HELP_ARG_TXT = "Shows the help text for the command line arguments.";
    private static final String XMPP_USER_TXT = "Username to use for connecting to the XMPP server in context of the advanced CL.";
    private static final String XMPP_PSW_TXT = "Password to use for connecting to the XMPP server in context of the advanced CL.";
    private static final String XMPP_SERVERADR_TXT = "Address to use for connecting to the XMPP server in context of the advanced CL.";
    private static final String XMPP_SERVERPORT_TXT = "Port to use for connecting to the XMPP server in context of the advanced CL.";
    private static final String TSB_USER_TXT = "Username to use for connecting to the TSB in context of the advanced CL.";
    private static final String LOGGINGLVL_INTUITEL_TXT = "The threshold of messages logged from LPM components. Possible values are 'DEBUG', 'INFO', 'WARN' and 'ERROR'. Default is 'INFO'";
    private static final String LOGGINGLVL_HTTP_TXT = "The threshold of messages logged for the HTTP server. Possible values are 'DEBUG', 'INFO', 'WARN' and 'ERROR'. Default is 'ERROR'";
    private static final String LOGGINGLVL_DB_TXT = "The threshold of messages logged for database. Possible values are 'DEBUG', 'INFO', 'WARN' and 'ERROR'. Default is 'ERROR'";
    private static final String MERGEDLSO_ARG_TXT = "If set to true, the LSO created by the LPM will be merged with the ontologies it references on.";
    // </editor-fold>
    private final Options parseOptions;
    protected URL lpmAddress = null;
    protected URL comLayerAddress = null;
    protected File slomRepoAddress = null;
    protected boolean testMode = false;
    protected boolean helpMode = false;
    private String xmppUser = null;
    private String xmppPsw = null;
    private String xmppServerAdr = null;
    private String xmppServerPort = null;
    private String tsbUser = null;
    private String intuitelLogLvl = Level.INFO.toString();
    private String httpLogLvl = Level.ERROR.toString();
    private String dbLogLvl = Level.ERROR.toString();
    private Boolean mergedLso = false;

    /**
     * Constructor.
     */
    public CmdParser() {
        parseOptions = new Options();
        parseOptions.addOption(createArgumentOption(CLADDRESS_ARG, URL.class, CLADDRESS_ARG_TXT));
        parseOptions.addOption(createArgumentOption(LOGGINGLVL_INTUITEL_ARG, String.class, LOGGINGLVL_INTUITEL_TXT));
        parseOptions.addOption(createArgumentOption(LOGGINGLVL_HTTP_ARG, String.class, LOGGINGLVL_HTTP_TXT));
        parseOptions.addOption(createArgumentOption(LOGGINGLVL_DB_ARG, String.class, LOGGINGLVL_DB_TXT));
        parseOptions.addOption(createArgumentOption(LPMADDRESS_ARG, URL.class, LPMADDRESS_ARG_TXT));
        parseOptions.addOption(createArgumentOption(SLOMREPOADDRESS_ARG, File.class, SLOMREPOADDRESS_ARG_TXT));
        parseOptions.addOption(createBooleanOption(HELP_ARG, HELP_ARG_TXT, false));
        parseOptions.addOption(createBooleanOption(TEST_ARG, TEST_ARG_TXT, false));
        parseOptions.addOption(createBooleanOption(MERGEDLSO_ARG, MERGEDLSO_ARG_TXT, false));
        parseOptions.addOption(createArgumentOption(XMPP_USER_ARG, String.class, XMPP_USER_TXT));
        parseOptions.addOption(createArgumentOption(XMPP_PSW_ARG, String.class, XMPP_PSW_TXT));
        parseOptions.addOption(createArgumentOption(XMPP_SERVERADR_ARG, String.class, XMPP_SERVERADR_TXT));
        parseOptions.addOption(createArgumentOption(XMPP_SERVERPORT_ARG, String.class, XMPP_SERVERPORT_TXT));
        parseOptions.addOption(createArgumentOption(TSB_USER_ARG, String.class, TSB_USER_TXT));
    }

    /**
     * Creates a new argument Option for commandline parsing.
     *
     * @param name Name of the argument
     * @param required Argument is required true/false
     * @param argClass Class of the argument
     * @param description Description of the argument for help
     * @return New Option instance as create via the OptionBuilder
     */
    @SuppressWarnings("static-access")
    private Option createArgumentOption(String name, Class argClass, String description) {
        return OptionBuilder.withArgName(name)
                .hasArg()
                .withType(argClass)
                .withDescription(description)
                .create(name);
    }

    /**
     * Create a new boolean Option for commandline parsing.
     *
     * @param name Name of the argument
     * @param description Description of the argument for help
     * @param hasArguments Option takes an argument
     * @return New Option instance
     */
    private Option createBooleanOption(String name, String description, boolean hasArguments) {
        return new Option(name, hasArguments, description);
    }

    /**
     * Parses command line parameters.
     *
     * @param args Command line parameters (given by main(...))
     * @return False on error
     */
    public boolean parseCmdLine(String[] args) {
        CommandLine line = parse(args);
        if (line == null) {
            logger.error("No command line parameters given.");
            return false;
        }
        if (line.hasOption(HELP_ARG)) {
            new HelpFormatter().printHelp("...", parseOptions);
            helpMode = true;
            //Abort here, due to help-cmd, other arguments probably missing.
            return true;
        }
        if (line.hasOption(TEST_ARG)) {
            logger.debug("============= TEST MODE ON ================");
            testMode = true;
        }
        parseValues(line);
        if (mandatoryArgumentMissing()) {
            logger.error("Manadory command line arguments missing.");
            return false;
        }
        if (!argumentsValid()) {
            logger.error("Invalid arguments detected.");
            return false;
        }
        return true;
    }

    /**
     * Parses the commandline with the given options.
     *
     * @param options Options for parsing the commandline
     * @param args Arguments to be parsed
     * @return Null on error
     */
    private CommandLine parse(String[] args) {
        CommandLineParser parser = new BasicParser();
        try {
            CommandLine line = parser.parse(parseOptions, args);
            if (line == null) {
                logger.error("CommandLine could not be parsed.");
                return null;
            }
            return line;
        } catch (ParseException ex) {
            logger.error(String.format("Exception on parsing commandline arguments: %s", ex.getMessage()));
            return null;
        }
    }

    /**
     * Parses the values from the given command line.
     *
     * @param line CMD to be parsed
     */
    private void parseValues(CommandLine line) {
        comLayerAddress = parseValueFromCmdLine(line, CLADDRESS_ARG, URL.class, null);
        lpmAddress = parseValueFromCmdLine(line, LPMADDRESS_ARG, URL.class, "http://localhost:8080/lpm/");
        slomRepoAddress = parseValueFromCmdLine(line, SLOMREPOADDRESS_ARG, File.class, null);
        xmppUser = parseValueFromCmdLine(line, XMPP_USER_ARG, String.class, null);
        xmppPsw = parseValueFromCmdLine(line, XMPP_PSW_ARG, String.class, null);
        xmppServerAdr = parseValueFromCmdLine(line, XMPP_SERVERADR_ARG, String.class, null);
        xmppServerPort = parseValueFromCmdLine(line, XMPP_SERVERPORT_ARG, String.class, null);
        tsbUser = parseValueFromCmdLine(line, TSB_USER_ARG, String.class, null);
        intuitelLogLvl = parseValueFromCmdLine(line, LOGGINGLVL_INTUITEL_ARG, String.class, intuitelLogLvl);
        httpLogLvl = parseValueFromCmdLine(line, LOGGINGLVL_HTTP_ARG, String.class, httpLogLvl);
        dbLogLvl = parseValueFromCmdLine(line, LOGGINGLVL_DB_ARG, String.class, dbLogLvl);
        mergedLso = parseValueFromCmdLine(line, MERGEDLSO_ARG, Boolean.class, mergedLso.toString());
    }

    /**
     * Generically parses an element from the commandline with the given {@code argName}.
     *
     * @param <T> Class of the cmdline element to be parsed
     * @param line Cmdline to be parsed
     * @param argName Name of the argument in the cmdline
     * @param argClass Class of the argument in the cmdline
     * @param defaultValue Default value of the argument to be used if the argument is not part of
     * the cmdline
     * @return Null on error or if not containe
     */
    private <T> T parseValueFromCmdLine(CommandLine line, String argName, Class<T> argClass, String defaultValue) {
        try {
            if (line.hasOption(argName)) {
                if(argClass == Boolean.class) {
                    logger.debug(String.format("Parsed value '%s' for arg '%s'", true, argName));
                    return argClass.cast(Boolean.TRUE);
                }
                Object returnValueObj = line.getParsedOptionValue(argName);
                if (argClass.isInstance(returnValueObj)) {
                    logger.debug(String.format("Parsed value '%s' for arg '%s'", returnValueObj, argName));
                    return argClass.cast(returnValueObj);
                }
            } else if (!StringUtils.isBlank(defaultValue)) {
                logger.debug(String.format("Set default value '%s' for arg '%s'", defaultValue, argName));
                return argClass.getConstructor(new Class[]{String.class}).newInstance(defaultValue);
            }
            return null;
        } catch (ParseException | NoSuchMethodException | SecurityException | InstantiationException |
                IllegalAccessException | IllegalArgumentException | InvocationTargetException ex) {
            logger.error(String.format("Exception on parsing commandline argument %s.", argName), ex);
            return null;
        }
    }

    /**
     * Checks if a mandatory argument is missing.
     *
     * @return True if at least one argument is missing.
     */
    private boolean mandatoryArgumentMissing() {
        boolean missingArgument = false;
        if (basicConnectionDataMissing() && advancedConnectionDataMissing()) {
            logger.error("Data for connecting to the communication layer is missing.");
            missingArgument = true;
        }
        return missingArgument;
    }

    /**
     * Checks whether all data for using the basic CL is available.
     *
     * @return True when data is missing.
     */
    private boolean basicConnectionDataMissing() {
        return comLayerAddress == null;
    }

    /**
     * Checks whether all data for using the advanced CL is available.
     *
     * @return True when data is missing.
     */
    private boolean advancedConnectionDataMissing() {
        return xmppUser == null || xmppPsw == null || xmppServerAdr == null || xmppServerPort == null || tsbUser == null;
    }

    /**
     * Checks if all variables are valid.
     *
     * @return False if at least one argument is invalid
     */
    protected boolean argumentsValid() {
        boolean valid = true;
        if (slomRepoAddress != null) {
            if (!slomRepoAddress.isDirectory()) {
                logger.error("'slomRepoAddress' is no valid local path: " + getSlomRepoAddress());
                valid = false;
            }
        }
        if (valid) {
            logger.debug("Commandline valid.");
        } else {
            logger.error("Commandline invalid.");
        }
        return valid;
    }

    /**
     * @return the lpmAddress
     */
    public URL getLpmAddress() {
        return lpmAddress;
    }

    /**
     * @return the clAddress
     */
    public URL getComLayerAddress() {
        return comLayerAddress;
    }

    /**
     * @return the slomRepoAddress
     */
    public File getSlomRepoAddress() {
        return slomRepoAddress;
    }

    /**
     * @return the testMode
     */
    public boolean isTestMode() {
        return testMode;
    }

    /**
     * @return the helpMode
     */
    public boolean isHelpMode() {
        return helpMode;
    }

    /**
     * @return the xmppUser
     */
    public String getXmppUser() {
        return xmppUser;
    }

    /**
     * @return the xmppPsw
     */
    public String getXmppPsw() {
        return xmppPsw;
    }

    /**
     * @return the xmppServerAdr
     */
    public String getXmppServerAdr() {
        return xmppServerAdr;
    }

    /**
     * @return the xmppServerPort
     */
    public String getXmppServerPort() {
        return xmppServerPort;
    }

    /**
     * @return the tsbUser
     */
    public String getTsbUser() {
        return tsbUser;
    }

    /**
     * @return The INTUITEL log level.
     */
    public Level getIntuitelLogLvl() {
        return Level.toLevel(intuitelLogLvl);
    }

    /**
     * @return The HTTP log level.
     */
    public Level getHttpLogLvl() {
        return Level.toLevel(httpLogLvl);
    }

    /**
     * @return The DB log level.
     */
    public Level getDbLogLvl() {
        return Level.toLevel(dbLogLvl);
    }
        
    /**
     * @return The mergedLso flag.
     */
    public boolean getMergedLsoFlag() {
        return mergedLso;
    }

    /**
     * Checks whether the connection data for the base connection is available.
     *
     * @return True if all needed data is available.
     */
    public boolean isBaseConnectionDataAvailable() {
        return lpmAddress != null && comLayerAddress != null;
    }

    /**
     * Checks whether the connection data for the advanced connection is available.
     *
     * @return True if all needed data is available.
     */
    public boolean isAdvancedConnectionDataAvailable() {
        return !StringUtils.isBlank(xmppUser)
                && !StringUtils.isBlank(xmppPsw)
                && !StringUtils.isBlank(xmppServerAdr)
                && !StringUtils.isBlank(xmppServerPort)
                && !StringUtils.isBlank(tsbUser);
    }
}
