/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-05-08
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmComInterface.wrappers;

import eu.intuitel.comm.entities.lms.Data;
import eu.intuitel.lpmComInterface.utils.ValueConverterUtil;
import java.util.HashMap;
import java.util.Map;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Wrapper class for LmsProfile request, incoming from the LMS.
 *
 * @author fhe
 */
public class LmsProfileIn implements IIncomingMsg {

    private static final Logger logger = LoggerFactory.getLogger(LmsProfileIn.class);
    private final Map<String, String> DATA;
    private final String MID;

    /**
     * Constructor.
     *
     * @param receivedItem LmsProfil data item that has been received by the LPM.
     */
    public LmsProfileIn(eu.intuitel.comm.entities.lms.LmsProfile receivedItem) {
        if(receivedItem == null) {
            logger.error("Received null response.");
            MID = null;
            DATA = new HashMap<>();
        } else {
            MID = receivedItem.getMId();
            if(StringUtils.isBlank(MID)) {
                logger.warn("Cannot process received message - MID is not set.");
            }
            Map<String, String> receivedData = new HashMap<>();
            for (Data dataItem : receivedItem.getData()) {
                receivedData.put(dataItem.getName(), dataItem.getValue());
            }
            DATA = receivedData;
        }
    }

    @Override
    public String getMID() {
        return MID;
    }

    /**
     * Returns the received LMS profile data as a String-map.
     *
     * @return Empty map if nothing was received.
     */
    public Map<String, String> getData() {
        return DATA;
    }

    /**
     * Returns the LMS name.
     *
     * @param data Data as returned by the LMS
     * @return Null if not available.
     */
    public static String getLmsName(Map<String, String> data) {
        return data.get("lmsName");
    }

    /**
     * Returns the name formality with which the learner should be addressed. 0 = Neutral, 1 =
     * Forename, 2 = Surename, 3 = Forename + surename
     *
     * @param data Data as returned by the LMS
     * @return Null if not available.
     */
    public static Integer getLearnerNameFormality(Map<String, String> data) {
        return ValueConverterUtil.toInteger(data.get("lNameFormality"));
    }

    /**
     * Returns the LMS type.
     *
     * @param data Data as returned by the LMS
     * @return Null if not available.
     */
    public static String getLmsType(Map<String, String> data) {
        return data.get("lmsType");
    }

    /**
     * Returns the ID of the LMS.
     *
     * @param data Data as returned by the LMS
     * @return Null if not available.
     */
    public static String getLmsId(Map<String, String> data) {
        return data.get("lmsId");
    }

    /**
     * Returns the media level of the LMS (v=video, a=audio, va or av = audio and video).
     *
     * @param data Data as returned by the LMS
     * @return Null if not available.
     */
    public static String getLmsMediaLevel(Map<String, String> data) {
        return data.get("lmsMediaLevel");
    }

    /**
     * Returns the LORE level with "0" meaning emulation via TUG.
     *
     * @param data Data as returned by the LMS
     * @return Null if not available.
     */
    public static Integer getLmsLoreLevel(Map<String, String> data) {
        return ValueConverterUtil.toInteger(data.get("loreLevel"));
    }

    /**
     * Returns the USE level with "0" meaning emulation via TUG.
     *
     * @param data Data as returned by the LMS
     * @return Null if not available.
     */
    public static Integer getLmsUseLevel(Map<String, String> data) {
        return ValueConverterUtil.toInteger(data.get("useLevel"));
    }

    /**
     * Returns the communication style of the LMS. 0 = PULL, 1 = PUSH, 2 = PUSH WITH POLLING
     *
     * @param data Data as returned by the LMS
     * @return Null if not available.
     */
    public static Integer getLmsComStyle(Map<String, String> data) {
        Integer comStyle = ValueConverterUtil.toInteger(data.get("comStyle"));
        if (comStyle == null) {
            logger.info("No comStyle specified, using polling (default value).");
            comStyle = 2;
        }
        return comStyle;
    }
}
