/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-05-15
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmComInterface.wrappers;

import eu.intuitel.CommLayer.CommLibrary.LpmServiceConnection;
import eu.intuitel.comm.entities.exceptions.IntuitelException;
import eu.intuitel.comm.entities.lpm.LearnerState;
import eu.intuitel.comm.entities.lpm.LearnerStateOntology;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * Sends a message to the Engine containing the LSO for processing.
 *
 * @author fhe
 */
public class LsoOut implements IOutgoingMsg {

    private final LearnerState ITEM;
    private static final Logger logger = LoggerFactory.getLogger(LsoOut.class);

    /**
     * Constructor.
     *
     * @param mId ID of the learner update which lead to this LSO
     * @param lId ID of the learner to whom the LSO belongs
     * @param cId ID of the course in scope
     * @param ccmUpdateNeeded Flag indicating for the Engine that the CCM has changed
     * @param cmUpdateNeeded Flag indicating for the Engine that the CM has changed
     * @param lso Learner State Ontology to be send
     */
    public LsoOut(String mId, String lId, String cId, boolean ccmUpdateNeeded, boolean cmUpdateNeeded, String lso) {
        LearnerState item = new LearnerState();
        item.setmId(mId);
        item.setuId(lId);
        item.setCourseId(cId);
        item.setCcmUpdateNeeded(ccmUpdateNeeded);
        item.setCmUpdateNeeded(cmUpdateNeeded);
        item.setLearnerStateOntology(getLsoObj(lso));
        ITEM = item;
    }

    @Override
    public String getMID() {
        return ITEM.getmId();
    }

    @Override
    public void send(LpmServiceConnection service) throws IntuitelException {
        if (ITEM.getLearnerStateOntology() == null) {
            return;
        }
        service.sendLearnerState(ITEM);
    }

    /**
     * Sets the content of the LearnerState item.
     *
     * @param lso LSO to be integrated
     * @return New LSO object, null on problem.
     */
    public static LearnerStateOntology getLsoObj(String lso) {
        LearnerStateOntology lsoObj = new LearnerStateOntology();
        Element contentElement = getElementFromString(lso);
        if (contentElement == null) {
            return null;
        }
        List<Element> contentList = new ArrayList<>();
        contentList.add(contentElement);
        lsoObj.setContent(contentList);
        return lsoObj;
    }

    /**
     * Creates a DOM Element from the given String.
     *
     * @param value Value to be converted
     * @return Null on error.
     */
    public static Element getElementFromString(String value) {
        try {
            DocumentBuilderFactory b = DocumentBuilderFactory.newInstance();
            b.setNamespaceAware(false);
            DocumentBuilder db = b.newDocumentBuilder();
            Document doc = db.parse(new ByteArrayInputStream(value.getBytes(StandardCharsets.UTF_8)));
            return doc.getDocumentElement();
        } catch (ParserConfigurationException | SAXException | IOException ex) {
            logger.error("Could not create LSO message content element.", ex);
            return null;
        }
    }
}
