/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-08
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmComInterface.communicators;

import eu.intuitel.lpmGeneralInterfaces.ISlomCommunicator;
import eu.intuitel.lpmGeneralInterfaces.ISlomCommunicator.ISlomRefContainer;
import eu.intuitel.lpmGeneralInterfaces.ISlomCommunicator.SlomFileRefContainer;
import java.util.List;
import org.apache.commons.lang3.StringUtils;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 * Tests for loading and accessing of a local SLOM repo.
 *
 * @author fhe
 */
public class SlomLocalCommunicatorTest {

    private static final String PATH = ".\\src\\test\\resources\\slomTestRepository\\";

    @Test
    public void testIndicesMap() {
        try {
            SlomLocalCommunicator localSlomCom = new SlomLocalCommunicator(PATH);

            String[] kos = new String[]{"case01_Ko01", "case01_Ko02", "case01_Ko03", "case01_Ko04"};
            testKoIndices(localSlomCom, kos,
                    "http://www.intuitel.eu/public/testcase01/intui_CCM.owl",
                    "http://www.intuitel.eu/public/testcase01/intui_CM.owl");

            kos = new String[]{"case02_Ko01", "case02_Ko02", "case02_Ko03", "case02_Ko04", 
                "case02_Ko05", "case02_Ko06", "case02_Ko07", "case02_Ko08"};
            testKoIndices(localSlomCom, kos,
                    "http://www.intuitel.eu/public/testcase02/intui_CCM.owl",
                    "http://www.intuitel.eu/public/testcase02/intui_CM.owl");

            kos = new String[]{"case03_Ko01", "case03_Ko02", "case03_Ko03", "case03_Ko04", "case03_Ko05"};
            testKoIndices(localSlomCom, kos,
                    "http://www.intuitel.eu/public/testcase03/intui_CCM.owl",
                    "http://www.intuitel.eu/public/testcase03/intui_CM.owl");

            kos = new String[]{"case04_Ko01", "case04_Ko02", "case04_Ko03", "case04_Ko04", "case04_Ko05", 
                "case04_Ko06", "case04_Ko07", "case04_Ko08", "case04_Ko09", "case04_Ko10", "case04_Ko11"};
            testKoIndices(localSlomCom, kos,
                    "http://www.intuitel.eu/public/testcase04/intui_CCM.owl",
                    "http://www.intuitel.eu/public/testcase04/intui_CM.owl");
        } catch (Exception ex) {
            assertNull("Exception has been thrown on loading local SLOM repo.", ex);
        }
    }

    private void testKoIndices(SlomLocalCommunicator localSlomCom, String[] kos, String ccm, String cm) {
        for (String ko : kos) {
            testKoIndex(localSlomCom, ko, ccm, cm);
        }
    }

    private void testKoIndex(SlomLocalCommunicator localSlomCom, String ko, String ccm, String cm) {
        ISlomRefContainer container = localSlomCom.getSlomDataForLo(ko);
        assertNotNull(container);
        assertTrue("Incorrect SLOM ref container type.", container instanceof SlomFileRefContainer);
        SlomFileRefContainer fileContainer = (SlomFileRefContainer) container;
        assertNotNull("CM file is null.", fileContainer.getCmFile());
        assertNotNull("CCM file is null.", fileContainer.getCcmFile());
        assertNotNull("CM file is not readable.", fileContainer.getCmFile().canRead());
        assertNotNull("CCM file is not readable..", fileContainer.getCcmFile().canRead());
    }

    @Test
    public void testGetBaseIriInvalid() {
        assertNull("Base IRI was returned for an invalid input.",
                SlomLocalCommunicator.getBaseIri("sagkjsagjlkbsg"));
    }

    @Test
    public void testGetBaseIriValid() {
        assertEquals("Base IRI was not returned correctly for a valid input.",
                "http://www.intuitel.eu/public/testcase01/intui_CCM.owl",
                SlomLocalCommunicator.getBaseIri("xml:base=\"http://www.intuitel.eu/public/testcase01/intui_CCM.owl\""));
    }

    @Test
    public void testKosDefinitionsInvalid() {
        List<String> kos = SlomLocalCommunicator.getKoDefinitions(
                "safdgulosafgfg");
        assertNotNull("KO definitions was returned as null.", kos);
        assertEquals("KO definition was returend for an invalid input.", 0, kos.size());
    }

    @Test
    public void testKosDefinitionsValid() {
        List<String> kos = SlomLocalCommunicator.getKoDefinitions(
                "<DataProperty abbreviatedIRI=\"intui_PO:isLinkedWithLmsElement\"/>\n" +
                "        <NamedIndividual IRI=\"#Ko03\"/>\n" +
                "        <Literal datatypeIRI=\"&xsd;string\">case01_Ko03</Literal>");
        assertNotNull("KO definitions was returned as null.", kos);
        assertEquals("KO definition was returend for an invalid input.", 1, kos.size());
        assertEquals("KO definition was not returned correctly for a valid input.", "case01_Ko03", kos.get(0));
    }

    @Test
    public void testKosDefinitionsMultipleValid() {
        List<String> kos = SlomLocalCommunicator.getKoDefinitions(
                "<DataProperty abbreviatedIRI=\"intui_PO:isLinkedWithLmsElement\"/>\n" +
                "        <NamedIndividual IRI=\"#Ko03\"/>\n" +
                "        <Literal datatypeIRI=\"&xsd;string\">case01_Ko03</Literal>" +
                "<DataProperty abbreviatedIRI=\"intui_PO:isLinkedWithLmsElement\"/>\n" +
                "        <NamedIndividual IRI=\"#Ko03\"/>\n" +
                "        <Literal datatypeIRI=\"&xsd;string\">case01_Ko04</Literal>");
        assertNotNull("KO definitions was returned as null.", kos);
        assertEquals("KO definition was returend for an invalid input.", 2, kos.size());
        assertTrue("KO definition does not contain first result", kos.contains("case01_Ko03"));
        assertTrue("KO definition does not contain second result", kos.contains("case01_Ko04"));
    }
}