/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-05-12
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmGeneralInterfaces;

import eu.intuitel.types.*;
import java.util.Collection;
import java.util.List;
import org.apache.commons.lang3.tuple.Pair;

/**
 * General interface for the user database access.
 *
 * @author fhe
 */
public interface IUserDbCommunicator {

    //<editor-fold defaultstate="collapsed" desc="Learning history getters">
    /**
     * Returns the completion status of the CC.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param ccIri CC in scope
     * @return Integer value between 0 and 100.
     */
    public int getCcCompletionStatus(String learnerId, String courseId, String ccIri);

    /**
     * Returns the completion status of the KO.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri KO in scope
     * @return Integer value between 0 and 100.
     */
    public int getKoCompletionStatus(String learnerId, String courseId, String koIri);

    /**
     * Returns all completed CCs of the given learner in the course.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return List of completed CC URIs; empty if no CCs completed yet.
     */
    public List<String> getCompletedCCs(String learnerId, String courseId);

    /**
     * Returns all completed KOs of the given learner in the course.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return List of completed KOs; empty if no KOs completed yet.
     */
    public List<String> getCompletedKOs(String learnerId, String courseId);

    /**
     * Returns the last n elements of the learning history of CCs.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param inspectionSize Number of elements (n) to be returned
     * @return List of last visited CCs; empty if none was visited yet.
     */
    public List<String> getHistoryOfCCs(String learnerId, String courseId, int inspectionSize);

    /**
     * Returns the all elements of the learning history of CCs.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return List of last visited CCs; empty if none was visited yet.
     */
    public List<String> getHistoryOfCCs(String learnerId, String courseId);

    /**
     * Returns the last n elements of the learning history of KOs.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param inspectionSize Number of elements (n) to be returned
     * @return List of last visited KOs; empty if none was visited yet.
     */
    public List<String> getHistoryOfKOs(String learnerId, String courseId, int inspectionSize);

    /**
     * Returns the all elements of the learning history of KOs.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return List of last visited KOs; empty if none was visited yet.
     */
    public List<String> getHistoryOfKOs(String learnerId, String courseId);

    /**
     * Returns the LmsId of the last Knowledge Object the learner has accessed.
     *
     * @param learnerId ID of the learner in scope
     * @return LmsId of the last accessed KO in the course.
     */
    public String getLastKnowledgeObjectLmsId(String learnerId);

    /**
     * Returns the LmsId of the last Knowledge Object the learner has accessed.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return LmsId of the last accessed KO in the course.
     */
    public String getLastKnowledgeObjectLmsId(String learnerId, String courseId);

    /**
     * Returns the IRI of the last Knowledge Object the learner has accessed.
     *
     * @param learnerId ID of the learner in scope
     * @return IRI of the last accessed KO in the course.
     */
    public String getLastKnowledgeObjectIri(String learnerId);

    /**
     * Returns the IRI of the last Knowledge Object the learner has accessed.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return IRI of the last accessed KO in the course.
     */
    public String getLastKnowledgeObjectIri(String learnerId, String courseId);

    /**
     * Returns the last recommendations given to the learner.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param stepsCount Number of learning steps to be returned
     * @return List containing Lists of the past {@code stepsCount} recommendations (IRIs) given to
     * the learner. May never return null, but (nested) empty lists.
     */
    public List<List<Pair<String, Double>>> getLastRecommendations(String learnerId, String courseId, int stepsCount);

    /**
     * Determines the learning time of the specified KO.
     * 
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri IRI of the KO in scope
     * @return Null in case of error or if there is no learning time for the specified KO.
     */
    public Long getLearningTime(String learnerId, String courseId, String koIri);
    
    /**
     * Returns grade the learner has received in the specified KO.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri Knowledge Object in scope
     * @return Grade received by the learner. Null if not specified.
     */
    public Double getKoCompletionGrade(String learnerId, String courseId, String koIri);
    
    /**
     * Retrieves all session durations of the learner.
     * 
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Returns an empty collection if nothing is available or in case of error.
     */
    public Collection<Long> getSessionLengths(String learnerId, String courseId);
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="Learning history setters">
    /**
     * Adds a new entry in the learning history.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri IRI of the KO to be added
     * @param koLmsId LmsId of the KO to be added
     * @param ccIri IRI CC to be added
     * @param time Time of access as long
     * @return Adding the new entry was successfull true/false
     */
    public boolean addLearningHistoryEntry(String learnerId, String courseId, String koIri, String koLmsId, String ccIri, long time);

    /**
     * Sets the session end marker for the specified learner.
     *
     * @param learnerId Learner in scope
     * @param courseId Id of the course the learner is working on
     * @return Setting session end was successful (=true)
     */
    public boolean setSessionEnd(String learnerId, String courseId);

    /**
     * Sets the received grade of the given KO.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri SLOM IRI of the KO in scope
     * @param grade Grade received by the learner
     * @return True if storing the value was successfull.
     */
    public boolean setKoCompletionStatusGrade(String learnerId, String courseId, String koIri, int grade);

    /**
     * Sets the completion status of the given KO.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri SLOM IRI of the KO in scope
     * @param completion Percentage value indicating how much of the KO was actually completed by
     * the learner
     * @return True if storing the value was successfull.
     */
    public boolean setKoCompletionStatusCompletion(String learnerId, String courseId, String koIri, int completion);

    /**
     * Sets the seen percentage of the given KO.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri SLOM IRI of the KO in scope
     * @param seenPercentage Percentage value indicating how much of the KO was actually seen by the
     * learner
     * @return True if storing the value was successfull.
     */
    public boolean setKoCompletionStatusSeenPercentage(String learnerId, String courseId, String koIri,
            int seenPercentage);

    /**
     * Sets the accessed status of the given KO.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri SLOM IRI of the KO in scope
     * @param accessed Boolean value denoting that the learner accessed the KO
     * @return True if storing the value was successfull.
     */
    public boolean setKoCompletionStatusAccessed(String learnerId, String courseId, String koIri, boolean accessed);

    /**
     * Sets the completion value of the given CC.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param ccIri CC in scope
     * @param complete Binaray value denoting whether the CC is complete (true) or not (false)
     * @return True if setting the value was successfull.
     */
    public boolean setCcCompletionStatus(String learnerId, String courseId, String ccIri, int complete);
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="User profile getters">
    /**
     * Returns the current battery level of the device of the learner
     *
     * @param learnerId ID of the learner in scope
     * @return Double value between 0 and 100.
     */
    public Double getBatteryLevel(String learnerId);

    /**
     * Returns the age of the learner.
     *
     * @param learnerId ID of the learner in scope
     * @return Integer value larger or equal to 0.
     */
    public Integer getAge(String learnerId);

    /**
     * Returns true if the current learner is blind.
     *
     * @param learnerId ID of the learner in scope
     * @return True if learner is blind.
     */
    public Boolean isLearnerBlind(String learnerId);

    /**
     * Checks whether the learner is deaf.
     *
     * @param learnerId ID of the learner in scope
     * @return True if the learner is deaf.
     */
    public Boolean isLearnerDeaf(String learnerId);

    /**
     * Checks whether the learner is mute.
     *
     * @param learnerId ID of the learner in scope
     * @return True if the learner is mute.
     */
    public Boolean isLearnerMute(String learnerId);

    /**
     * Returns the horizontal resolution of the learner's access device.
     *
     * @param learnerId ID of the learner in scope
     * @return Horizontal pixel count.
     */
    public Integer getHorizontalDeviceResolution(String learnerId);

    /**
     * Returns the vertical resolution of the learner's access device.
     *
     * @param learnerId ID of the learner in scope
     * @return Vertical pixel count.
     */
    public Integer getVerticalDeviceResolution(String learnerId);

    /**
     * Returns the gender of the learner.
     *
     * @param learnerId ID of the learner in scope
     * @return Gender of the learner as represented by an enum. May never be null.
     */
    public Gender getGender(String learnerId);

    /**
     * Returns the current willingness of the learner to interact with peers.
     *
     * @param learnerId ID of the learner in scope
     * @return True if learner is willing, False if learner is unwilling, Null if value is unknown.
     */
    public Boolean getInteractionWillingness(String learnerId);

    /**
     * Returns the type of the internet connection the learner is using to access the LMS.
     *
     * @param learnerId ID of the learner in scope
     * @return String value describing the connection type, e.g. LTE
     */
    public ConnectionType getConnectionType(String learnerId);

    /**
     * Returns an Integer value indicating the stability of the learner's network connection.
     *
     * @param learnerId ID of the learner in scope
     * @return Value between 0 and 100, Null if unknown.
     */
    public Integer getConnectionStability(String learnerId);

    /**
     * Returns an Integer value indicating the preferred difficulty level of the learner.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return 1=Beginner, 2=Intermediate, 3=Advanced, Null=Unknown
     */
    public DifficultyLevel getPreferredDifficultyLevel(String learnerId, String courseId);

    /**
     * Returns an Integer value indiciating the preferred EQF level of the learner.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Inter value between 1 and 7, Null=Unknown
     */
    public EqfLevel getPreferredEqfLevel(String learnerId, String courseId);

    /**
     * Returns the learning attitude of the learner.
     *
     * @param learnerId ID of the learner in scope
     * @return True if learner is motivated, False if not, LearningAttitude.unknown if unknown.
     */
    public LearningAttitude getLearningAttitude(String learnerId);

    /**
     * Returns the noise level in which the learner is currently working in decibel.
     *
     * @param learnerId ID of the learner in scope
     * @return Decibel value, Null if unknown.
     */
    public Integer getNoise(String learnerId);

    /**
     * Get the timestamp of the current session start.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scipe
     * @return Timestamp as long.
     */
    public Long getCurrentLoginTime(String learnerId, String courseId);
    
    /**
     * Get the timestamp of the last session start.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scipe
     * @return Timestamp as long.
     */
    public Long getLastLoginTime(String learnerId, String courseId);

    /**
     * Gets the languages spoken by the learner.
     *
     * @param learnerId ID of the learner in scope
     * @return Language as spoken be the learner. May be null if nothing was specified.
     */
    public Language[] getSpokenLanguages(String learnerId);
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="User profile setters">
    /**
     * Stores profile related information for a learner. Uses and parses the String values as
     * obtained via USE.
     *
     * @param learnerId ID of the learner in scope
     * @param age Age of the learner in scope
     * @param gender Gender of the learner in scope
     * @param name Name of the learner in scope
     * @param culture The culture of the learner
     * @param deafness Deafness of the learner in scope
     * @param blindness Blindness of the learner in scope
     * @param muteness Muteness of the learner in scope
     * @param languages Comma-separated list of language codes
     * @return True if successfull.
     */
    public boolean setLearnerProfileData(String learnerId, String age, String gender, String name, String culture, String deafness, String blindness, String muteness, String languages);

    /**
     * Stores the session related data in the database. Uses the String values as obtained from USE.
     *
     * @param learnerId ID of the learner in scope
     * @param deviceType Type of device the learner is using
     * @param devicePixelDensity DPI of the device the learner is using
     * @param deviceResolution Resulution of the device the learner is using
     * @param environmentTime Local time as set on the device of the learner
     * @param attitude The current learning attitude of the learner
     * @param interactionWillingness The current interaction willingness of the learner
     * @return True if successfull
     */
    public boolean setSessionData(String learnerId, String deviceType, String devicePixelDensity, String deviceResolution, String environmentTime, String attitude, String interactionWillingness);

    /**
     * Stores the current environmental data of the learning situation. Uses values as obtained via
     * USE.
     *
     * @param learnerId ID of the learner in scope
     * @param batteryLevel Current battery level of the device the learner is using
     * @param conType Type of connection with which the learner is connected to the network
     * @param conStability Stability of connection with which the learner is connected to the
     * network
     * @param noiseLevel Current noise level in which the learning session takes place
     * @return True if successfull.
     */
    public boolean setEnvironmentData(String learnerId, String batteryLevel, String conType, String conStability, String noiseLevel);
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="Learning pathway getters">
    /**
     * Returns the currently selected macro LP of the learner.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Null if no MLP has yet been selected.
     */
    public String getMacroLpOfLearner(String learnerId, String courseId);

    /**
     * Returns the currently selected micro LP of the learner.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Null if no mLP has yet been selected.
     */
    public String getMicroLpOfLearner(String learnerId, String courseId);

    /**
     * Returns the percentage value of learners of the course who selected the given LP.
     *
     * @param lp LP in scope
     * @param courseId ID of the course in scope
     * @return Integer value between 0 and 100.
     */
    public int getPopularityOfLp(String lp, String courseId);

    /**
     * Returns the number of milliseconds since the last time the LPs have been evaluated for the
     * specified learner.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Maxium Long value if no last eval incident available.
     */
    public long getTimeSinceLastLpEvaluation(String learnerId, String courseId);

    /**
     * Returns the number of learning steps since the last time the LPs have been evaluated for the
     * specified learner.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Maximum Integer value if no last eval incident available.
     */
    public int getStepsSinceLastLpEvaluation(String learnerId, String courseId);
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="Learning pathway setters">
    /**
     * Sets now as the last LP evaluation incident for the given learner .
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Setting evaluation incident was successfull true/false.
     */
    public boolean setLpEvaluationIncident(String learnerId, String courseId);
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="LMS profile setters">
    /**
     * Sets the LMS properties.
     *
     * @param id ID of the LMS
     * @param name Name of the LMS
     * @param type Type of the LMS
     * @param mediaLvl Media lvl of the LMS
     * @param loreLvl LORE of the LMS
     * @param useLvl USE of the LMS
     * @param nameFormality Name formality of the LMS
     * @param comStyle Communication style of the LMS
     * @return False if storing failed.
     */
    public boolean storeLmsProperties(String id, String name, String type,
            MediaLvl mediaLvl, LoreLvl loreLvl, UseLvl useLvl, NameFormality nameFormality, ComStyle comStyle);
    //</editor-fold>

    // <editor-fold defaultstate="collapsed" desc="Peer data">    
    /**
     * Retrieves all session durations of the learner's peers.
     * 
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Returns an empty collection if nothing is available or in case of error.
     */
    public Collection<Collection<Long>> getPeerSessionLengths(String learnerId, String courseId);
    
    /**
     * Determines the learnerIds of the peers in the course.
     * 
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return If there are no peers or in case of error, an empty collection will be returned.
     */
    public Collection<String> getPeerIdsInCourse(String learnerId, String courseId);
    // </editor-fold>
    
    /**
     * @param iri IRI of the SLOM element
     * @param courseId ID of the course
     * @param title Title of the element
     * @param description Description of the element
     * @param lmsId ID of the KO in the LMS
     * @return True if successfull, otherwise false.
     */
    public boolean setKoData(String iri, String courseId, String title, String description, String lmsId);
    
    /**
     * @param iri IRI of the SLOM element
     * @param courseId ID of the course
     * @param title Title of the element
     * @param description Description of the element
     * @return True if successfull, otherwise false.
     */
    public boolean setCcData(String iri, String courseId, String title, String description);

    /**
     * @param iri IRI of the SLOM element
     * @param courseId ID of the course
     * @param title Title of the element
     * @param description Description of the element
     * @return True if successfull, otherwise false.
     */
    public boolean setMicroLpData(String iri, String courseId, String title, String description);

    /**
     * @param iri IRI of the SLOM element
     * @param courseId ID of the course
     * @param title Title of the element
     * @param description Description of the element
     * @return True if successfull, otherwise false.
     */
    public boolean setMacroLpData(String iri, String courseId, String title, String description);

    /**
     * Closes the DB connection.
     */
    public void close();    
}
