package eu.intuitel.types;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
 * Author: Ovidiu Serban, o.serban@reading.ac.uk.
 * Version: 1, 25/06/2014
 * Created at Intelligent System Research Laboratory,
 * University of Reading, United Kingdom
 * web: http://www.isr.reading.ac.uk/
 * Copyright may apply to this file
 */
public enum Language {
    English("en", "eng", "dfSpokenLanguage_English"),
    German("de", "deu", "dfSpokenLanguage_German"),
    Spanish("es", "spa", "dfSpokenLanguage_Spanish"),
    Italian("it", "ita", "dfSpokenLanguage_Italian"),
    Hebrew("he", "heb", "dfSpokenLanguage_Hebrew");

    public static final Language Default = English;
    public static final String SPLIT = ",";

    private final String iso1Code; //ISO 639‑1
    private final String iso3Code; //ISO 639‑3
    private final String dfValue;

    Language(String iso1Code, String iso3Code, String dfValue) {
        this.iso1Code = iso1Code;
        this.iso3Code = iso3Code;
        this.dfValue = dfValue;
    }

    public String getIso1Code() {
        return iso1Code;
    }

    public String getIso3Code() {
        return iso3Code;
    }

    public String getDfValue() {
        return dfValue;
    }

    /**
     * Returns the given language String as language object
     *
     * @param code String to parse
     * @return Default if not parsable or unknown code.
     */
    public static Language getLanguage(String code) {
        if (code == null) {
            return Default;
        } else {
            code = code.trim();
            for (Language value : values()) {
                if (code.equalsIgnoreCase(value.iso1Code)
                    || code.equalsIgnoreCase(value.iso3Code)
                    || code.equalsIgnoreCase(value.dfValue)) {
                    return value;
                }
            }
            return Default;
        }
    }

    public static Language[] toArray(String languages) {
        Collection<Language> langSet = split(languages);
        return langSet.toArray(new Language[langSet.size()]);
    }

    public static Collection<Language> split(String languages) {
        if (languages == null) {
            return Collections.emptySet();
        }
        if (languages.trim().length() == 0) {
            return Collections.emptySet();
        }
        Set<Language> langSet = new HashSet<>();
        for (String lang : languages.split(SPLIT)) {
            langSet.add(getLanguage(lang));
        }
        if (langSet.isEmpty()) {
            langSet.add(Default);
        }
        return langSet;
    }

    public static String join(Language... languages) {
        if (languages != null) {
            StringBuilder sb = new StringBuilder();
            for (Language lang : languages) {
                if (sb.length() > 0) {
                    sb.append(SPLIT);
                }
                sb.append(lang.iso3Code);
            }
            return sb.toString();
        } else {
            return null;
        }
    }
}
