/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Contributor: Ovidiu Serban, University of Reading, www.isr.reading.ac.uk
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor;

import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.lpmGeneralInterfaces.IEngineCommunicator;
import eu.intuitel.lpmGeneralInterfaces.ILmsCommunicator;
import eu.intuitel.lpmGeneralInterfaces.ISlomCommunicator;
import eu.intuitel.lpmGeneralInterfaces.ISlomCommunicator.ISlomRefContainer;
import eu.intuitel.lpmGeneralInterfaces.ISlomCommunicator.SlomFileRefContainer;
import eu.intuitel.lpmGeneralInterfaces.ISlomCommunicator.SlomUrlRefContainer;
import eu.intuitel.lpmProcessor.didacticFactors.DidacticFactorApplicator;
import eu.intuitel.lpmProcessor.models.LpEvaluationResult;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.factories.LearnerStateOntologyFactory;
import eu.intuitel.lpmProcessor.models.factories.SlomFactory;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.LearnerStateOntology;
import eu.intuitel.lpmProcessor.utils.CompletionUtil;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import eu.intuitel.lpmProcessor.utils.LpEvalUtil;
import eu.intuitel.lpmProcessor.utils.ReflexTesterUtil;
import eu.intuitel.lpmProcessor.utils.ReflexTesterUtil.ReflexException;
import eu.intuitel.lpmProcessor.utils.SlomMapCache;
import eu.intuitel.lpmProcessor.utils.UserScoreExtractionUtil;
import java.util.Collection;
import java.util.Set;
import org.apache.commons.lang.StringUtils;
import org.semanticweb.owlapi.model.OWLNamedIndividual;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * LpmProcessor implementation to execute the preprocessing for the reasoning process.
 *
 * @author fhe
 */
public class LpmProcessor {

    private static final Logger logger = LoggerFactory.getLogger(LpmProcessor.class);
    private static final SlomMapCache cache = SlomMapCache.getInstance();
    private final ISlomCommunicator slomCom;
    private final IEngineCommunicator engineCom;
    private final ILmsCommunicator lmsCom;
    private final DbAccessor accessor;
    private final String mId;
    private final String lId;
    private final long time;
    private final boolean createMergedLso;
    private String koLmsId;
    private DbAccessUtil db;

    /**
     * Constructor for a new LPM processor instance.
     *
     * @param loId LMS id of the current KO viewed in the LMS
     * @param learnerId ID of the learner currently in scope
     * @param mId ID of the learner update message
     * @param time Point in time the learner accessed the LO
     * @param slomCom Communication interface to the SLOM repository
     * @param engineCom Communication interface to the INTUITEL engine
     * @param lmsCom Communication interface to the LMS
     * @param accessor Instance for accessing the DB
     * @param createMergedLso LSO should be merged with the other ontologies (=true)
     */
    public LpmProcessor(String loId, String learnerId, String mId, long time,
            ISlomCommunicator slomCom,
            IEngineCommunicator engineCom,
            ILmsCommunicator lmsCom,
            DbAccessor accessor,
            boolean createMergedLso) {
        this.koLmsId = loId;
        this.lId = learnerId;
        this.mId = mId;
        this.time = time;
        this.slomCom = slomCom;
        this.engineCom = engineCom;
        this.lmsCom = lmsCom;
        this.accessor = accessor;
        this.createMergedLso = createMergedLso;
    }

    /**
     * Starts a new preprocessing job with the data defined for the LpmProcessor instance.
     */
    @SuppressWarnings({"BroadCatchBlock", "TooBroadCatch"}) //All errors to be catched inside the processor
    public void work() {
        SlomMetadata slom = null;
        try {
            db = new DbAccessUtil(lId, accessor);

            boolean isRefresh = false;
            if (koLmsId == null) {
                //Note: loId null --> refresh recommendation.
                koLmsId = db.getLastKnowledgeObjectLmsId();
                isRefresh = true;
            } else if (StringUtils.isBlank(koLmsId)) {
                //Note: loId empty ("") --> session ends.
                db.setSessionEnd();
                logger.debug("Session end registered for learner: " + lId);
                return;
            }

            //1. Get SLOM data for the LO the learner has just opened
            slom = getSlomMetadataForLo(koLmsId);
            ReflexTesterUtil.testSlom(mId, lId, slom, engineCom);
            
            if(db.isLearnerRegistered()) {
                //Create empty learner data if learner is new to the system
                db.setLearnerProfile();
                db.setLearnerPreferences();
            }

            //2. Get the currently selected LPs
            MacroLearningPathway currentMacroLp = db.getMacroLpOfLearner();
            if (currentMacroLp == null && slom.getMacroLps().size() == 1) {
                currentMacroLp = slom.getMacroLps().iterator().next();
                db.setMacroLearningPathway(currentMacroLp);
            }
            MicroLearningPathway currentMicroLp = db.getMicroLpOfLearner();
            if (currentMicroLp == null && slom.getMicroLps().size() == 1) {
                currentMicroLp = slom.getMicroLps().iterator().next();
                db.setMicroLearningPathway(currentMicroLp);
            }
            ReflexTesterUtil.testLpSelection(mId, lId, currentMacroLp, currentMicroLp, slom, engineCom);

            //3. Determine current KO and CC
            KnowledgeObject currentKo = slom.getKoByLmsId(koLmsId);
            Collection<ConceptContainer> currentCcs = getCurrentCcs(currentKo, currentMacroLp, slom, db);
            ReflexTesterUtil.testCurrentCc(mId, lId, currentCcs, slom, engineCom);
            ConceptContainer currentCc = currentCcs.iterator().next(); //Reflex filters cases where CC count != 1
            if (!isRefresh) {
                db.addLearningHistoryEntry(currentKo, currentCc, time);
            }

            //4. USE performance update
            UserScoreExtractionUtil useUtil = new UserScoreExtractionUtil();
            boolean usePerfSuccess = useUtil.getAndStoreUsePerfData(lId, koLmsId, currentKo, slom, lmsCom, db);
            ReflexTesterUtil.testUsePerformance(mId, lId, usePerfSuccess, slom, engineCom);

            //5. USE environment update
            boolean useEnvSuccess = useUtil.getAndStoreUseEnvData(lId, lmsCom, db);
            ReflexTesterUtil.testUseEnvironment(mId, lId, useEnvSuccess, slom, engineCom);

            //6. Update CC completion status
            CompletionUtil.updateCcCompletionValue(currentKo, db, slom);
            //Note: Strictly speaking, it might be possible that other completion values change too, 
            //as there might be a USE update for a KO that has been graded by a teacher
            //- i.e. completion changed much later than access.

            //7. Check for missing data
            ReflexTesterUtil.testDataAvailability(mId, lId, db, useUtil, slom, engineCom);

            //8. LP evaluation
            Set<LpEvaluationResult> lpEvaluations = LpEvalUtil.getLpEvaluations(slom, lId, db);
            ReflexTesterUtil.testLpEvaluations(mId, lId, lpEvaluations, slom, engineCom);

            //9. DF caluclation
            Collection<OWLNamedIndividual> relevantDfs = DidacticFactorApplicator.applyDidacticFactors(
                    OntoMngr.LMO.getDidacticFactors(), slom, lId, db);

            //10. Creation of LSO
            LearnerStateOntology lso;
            if (createMergedLso) {
                lso = LearnerStateOntologyFactory.createLsoMerged(lId, slom, db, relevantDfs);
            } else {
                lso = LearnerStateOntologyFactory.createLsoRaw(lId, slom, db, relevantDfs);
            }
            logger.debug(String.format("Create LSO for learner %s, containing %s DFs: %s", lId, relevantDfs.size(), relevantDfs));

            String lsoString = lso.toXml();
            engineCom.sendLso(mId, lId, slom.getCourseId(), slom.isCcmUpdateNeeded(), slom.isCmUpdateNeeded(), lsoString);

            //For advanced debugging purposes:
            //logger.debug("Send LSO to QB: \n" + lsoString);
        } catch (ReflexException ex) {
            logger.info(String.format("%s-Reflex triggered for learner '%s' ('%s', '%s', '%s').",
                    ex.getMessage(), lId, koLmsId, slom == null ? "unkownCourse" : slom.getCourseId(), mId));
        } catch (Exception ex) {
            logger.error(String.format("An exception occured while creating LSO (%s, %s, %s, %s).",
                    lId, koLmsId, slom == null ? "unkownCourse" : slom.getCourseId(), mId), ex);
        }
    }

    /**
     * Returns the SLOM cache, which is shared accross all LPM processors.
     *
     * @return Cache instance, which should never be null.
     */
    public SlomMapCache getCache() {
        return cache;
    }

    /**
     * Retrieves a SLOM metadata object for the given loId from the SLOM cache or via the
     * communicator.
     *
     * @param loId ID of the Learning Object in scope
     * @return Null in case of error or if not part of the repository
     */
    protected SlomMetadata getSlomMetadataForLo(String loId) {
        //First attempt to get it from cache
        SlomMetadata cachedSlom = cache.getSlomMetadata(loId);
        if (cachedSlom != null) {
            db = db.extendWithSlom(cachedSlom);
            return cachedSlom;
        }
        //If not in cache, request data from repo...
        ISlomRefContainer slomData = slomCom.getSlomDataForLo(loId);
        if (slomData == null) {
            logger.warn("Retrieved illegal SLOM data for LO {}", loId);
            return null;
        }
        //... and store it in cache before returning
        SlomMetadata slom;
        if (slomData instanceof SlomFileRefContainer) {
            SlomFileRefContainer fileContainer = (SlomFileRefContainer) slomData;
            slom = SlomFactory.createNewInstance(fileContainer.getCcmFile(), fileContainer.getCmFile());
        } else if (slomData instanceof SlomUrlRefContainer) {
            SlomUrlRefContainer urlContainer = (SlomUrlRefContainer) slomData;
            slom = SlomFactory.createNewInstance(urlContainer.getCcmUrl(), urlContainer.getCmUrl());
        } else {
            logger.error(String.format("Unknown SLOM container type: %s", slomData.getClass().getName()));
            return null;
        }
        cache.store(slom);
        db = db.extendWithSlom(slom);
        storeDescriptions(slom, db);
        return slom;
    }

    /**
     * Stores the texts from SLOM data in the database to have it available for feedback messages.
     *
     * @param slom SLOM data to be inspected
     * @param db Database access
     */
    private void storeDescriptions(SlomMetadata slom, DbAccessUtil db) {
        try {
            for (LearningPathway value : slom.getMacroLps()) {
                db.setMacroLpData(value.getIriAsString(), value.getTitle(), value.getDescription());
            }
            for (LearningPathway value : slom.getMicroLps()) {
                db.setMicroLpData(value.getIriAsString(), value.getTitle(), value.getDescription());
            }
            for (ConceptContainer value : slom.getCcs()) {
                db.setCcData(value.getIriAsString(), value.getTitle(), value.getDescription());
            }
            for (KnowledgeObject value : slom.getKos()) {
                db.setKoData(value.getIriAsString(), value.getTitle(), value.getDescription(), value.getLinkedLmsElement());
            }
        } catch (Exception ex) {
            logger.error("Error updating SLOM data descriptions in the DB.", ex);
        }
    }

    /**
     * Determines the current CCs of the learner. In some cases, this is not 100% determinable. If
     * more than one CC is returned, a reflex needs to be triggered which lets the learner choose
     * the current CC.
     *
     * @param currentKo The currently viewed KO
     * @param currentMacroLp The currently selected macro LP
     * @param slom SLOM data of the course
     * @param dbCom DB intergace
     * @return Set of current CCs.
     */
    private Collection<ConceptContainer> getCurrentCcs(KnowledgeObject currentKo, MacroLearningPathway currentMacroLp, SlomMetadata slom, DbAccessUtil db) {
        //1. Get base set - CCs connected to the current KO
        Collection<ConceptContainer> ccSelectionSet = currentKo.getCcReferences();
        if (ccSelectionSet.size() <= 1) {
            return ccSelectionSet;
        }

        //2. Remove all CCs not part of the current MacroLp
        ccSelectionSet.retainAll(currentMacroLp.getAllReferences());
        if (ccSelectionSet.size() <= 1) {
            return ccSelectionSet;
        }

        //3. Remove all completed CCs
        ccSelectionSet.removeAll(db.getCompletedCCs());
        return ccSelectionSet;

        //Note: try to apply the last learning step for the case that the learner has been asked via a reflex
    }
}
