/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.06.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import org.apache.commons.collections.CollectionUtils;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * DF absolute course time exposure
 * The time the learner spent on the course.
 *
 * @author Florian Heberle
 */
public class DfAbsoluteCourseTimeExposure extends DidacticFactor {

    private static final String VERYSHORT = "dfAbsoluteCourseTimeExposure_VeryShort";
    private static final String SHORT = "dfAbsoluteCourseTimeExposure_Short";
    private static final String HALFTIME = "dfAbsoluteCourseTimeExposure_HalfTime";
    private static final String LONG = "dfAbsoluteCourseTimeExposure_Long";
    private static final String VERYLONG = "dfAbsoluteCourseTimeExposure_VeryLong";
    private static final String COMPLETEOROVER = "dfAbsoluteCourseTimeExposure_CompleteOrOver";

    /**
     * Constructor
     *
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    public DfAbsoluteCourseTimeExposure(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        super(loadedIndividuals);
    }

    @Override
    protected Collection<String> getExpectedIndividualFragments() {
        Collection<String> fragments = new HashSet<>();
        fragments.add(VERYSHORT);
        fragments.add(SHORT);
        fragments.add(HALFTIME);
        fragments.add(LONG);
        fragments.add(VERYLONG);
        fragments.add(COMPLETEOROVER);
        return fragments;
    }

    @Override
    public OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db) {
        //Determine how much time the learner has spent on the course on CC-basis
        //ELT of a CC is the sum of all connected KO-ELTs. 
        Long value = getValue(db.getCompletedKOs(), slom);
        if (value == null) {
            return null;
        } else if (value >= 100) {
            return new OWLNamedIndividual[]{individualsMap.get(COMPLETEOROVER)};
        } else if (value >= 90) {
            return new OWLNamedIndividual[]{individualsMap.get(VERYLONG)};
        } else if (value >= 80) {
            return new OWLNamedIndividual[]{individualsMap.get(LONG)};
        } else if (value >= 50) {
            return new OWLNamedIndividual[]{individualsMap.get(HALFTIME)};
        } else if (value >= 20) {
            return new OWLNamedIndividual[]{individualsMap.get(SHORT)};
        } else if (value >= 10) {
            return new OWLNamedIndividual[]{individualsMap.get(VERYSHORT)};
        }
        return null;
    }

    /**
     * Determines the value for the DF.
     *
     * @param completedKos The KOs the learner has already completed
     * @param slom The SLOM data of the current course
     * @return Null in case nothing could be determined
     */
    public static Long getValue(List<KnowledgeObject> completedKos, SlomMetadata slom) {
        if (CollectionUtils.isEmpty(completedKos)) {
            return null;
        }
        long eltSum = 0;
        long eltSumLearner = 0;
        for (ConceptContainer cc : slom.getCcs()) {
            for (KnowledgeObject ko : cc.getReferencedKos(slom)) {
                if (ko.getEstimatedLearningTime() != null) {
                    eltSum += ko.getEstimatedLearningTime();
                    if (completedKos.contains(ko)) {
                        eltSumLearner += ko.getEstimatedLearningTime();
                    }
                }
            }
        }
        if (eltSum == 0) {
            return null;
        }
        return eltSumLearner * 100 / eltSum;
    }
}
