/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.06.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.Collection;
import java.util.HashSet;
import org.apache.commons.collections.CollectionUtils;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * DF comparative course time
 * Comparison between the amount of time the learner spent on the course in contrast to all course
 * participants.
 *
 * @author Florian Heberle
 */
public class DfComparativeCourseTime extends DidacticFactor {

    private static final String LONGER = "dfComparativeCourseTime_Longer";
    private static final String AVERAGE = "dfComparativeCourseTime_Average";
    private static final String SHORTER = "dfComparativeCourseTime_Shorter";

    /**
     * Constructor
     *
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    public DfComparativeCourseTime(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        super(loadedIndividuals);
    }

    @Override
    protected Collection<String> getExpectedIndividualFragments() {
        Collection<String> fragments = new HashSet<>();
        fragments.add(LONGER);
        fragments.add(AVERAGE);
        fragments.add(SHORTER);
        return fragments;
    }

    @Override
    public OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db) {
        Long value = getValue(db.getSessionLengths(), db.getPeerSessionLengths());
        if (value == null) {
            return null;
        } else if (value >= 130) {
            return new OWLNamedIndividual[]{individualsMap.get(LONGER)};
        } else if (value > 70) {
            return new OWLNamedIndividual[]{individualsMap.get(AVERAGE)};
        } else {
            return new OWLNamedIndividual[]{individualsMap.get(SHORTER)};
        }
    }

    /**
     * Determines the value for the DF.
     * 1.: Determines the sum of session lengths for the learner
     * 2.: Determines the average sum of session lengths of all learners
     * 3.: Determines the relation of the two values as result
     *
     * @param learnerSessionLengths The session lengths of the learner
     * @param peerSessionLengths The session lengths of the learner's peers
     * @return Null in case nothing could be determined
     */
    public static Long getValue(Collection<Long> learnerSessionLengths, Collection<Collection<Long>> peerSessionLengths) {
        if (CollectionUtils.isEmpty(learnerSessionLengths)) {
            return null;
        }
        if (CollectionUtils.isEmpty(peerSessionLengths)) {
            return null;
        }
        long learnerSum = 0;
        for (Long slength : learnerSessionLengths) {
            if (slength != null) {
                learnerSum += slength;
            }
        }
        long peerSum = 0;
        long peerCtr = 0;
        for (Collection<Long> p : peerSessionLengths) {
            if (!CollectionUtils.isEmpty(p)) {
                peerCtr++;
                for (Long slength : p) {
                    if (slength != null) {
                        peerSum += slength;
                    }
                }
            }
        }
        if (peerCtr == 0) {
            return null;
        }
        return learnerSum * 100 / (peerSum / peerCtr);
    }
}
