/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.06.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * DF comparative KO time
 * Comparison of the average deviation of learning time of the learner and the other course
 * participants.
 *
 * @author Florian Heberle
 */
public class DfComparativeKoTime extends DidacticFactor {

    private static final String FASTER = "dfComparativeKoTime_Faster";
    private static final String AVERAGE = "dfComparativeKoTime_Average";
    private static final String SLOWER = "dfComparativeKoTime_Slower";

    /**
     * Constructor
     *
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    public DfComparativeKoTime(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        super(loadedIndividuals);
    }

    @Override
    protected Collection<String> getExpectedIndividualFragments() {
        Collection<String> fragments = new HashSet<>();
        fragments.add(FASTER);
        fragments.add(AVERAGE);
        fragments.add(SLOWER);
        return fragments;
    }

    @Override
    public OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db) {
        //Determine the average deviation of the learner from the ELT
        Long avgEltDeviationLearner = getAvgEltDeviation(null, db);
        if (avgEltDeviationLearner == null) {
            return null;
        }

        //Determine the average deviation of all learners from the ELT
        List<Long> avgEltDeviationPeersList = new ArrayList<>();
        for (String peerId : db.getPeerIdsInCourse()) {
            Long avgEltDeviationPeer = getAvgEltDeviation(peerId, db);
            if (avgEltDeviationPeer != null) {
                avgEltDeviationPeersList.add(avgEltDeviationPeer);
            }
        }
        if (avgEltDeviationPeersList.isEmpty()) {
            return null;
        }
        long avgEltDeviationPeers = 0;
        for (Long avgEltDeviationPeer : avgEltDeviationPeersList) {
            avgEltDeviationPeers += avgEltDeviationPeer;
        }

        //Determine the %-difference between the two values
        long difference = avgEltDeviationLearner * 100 / avgEltDeviationPeers;
        if (difference >= 130) {
            return new OWLNamedIndividual[]{individualsMap.get(SLOWER)};
        } else if (difference > 70) {
            return new OWLNamedIndividual[]{individualsMap.get(AVERAGE)};
        } else {
            return new OWLNamedIndividual[]{individualsMap.get(FASTER)};
        }
    }

    private Long getAvgEltDeviation(String learnerId, DbAccessUtil db) {
        List<KnowledgeObject> completedKos = db.getCompletedKOs(learnerId);
        if (completedKos.isEmpty()) {
            return null;
        }
        long eltDeviation = 0;
        int ctr = 0;
        for (KnowledgeObject completedKo : completedKos) {
            Long estimate = completedKo.getEstimatedLearningTime();
            Long actual = db.getLearningTime(learnerId, completedKo);
            if (estimate != null && actual != null) {
                eltDeviation += estimate - actual;
                ctr++;
            }
        }
        return ctr == 0 ? null : eltDeviation / ctr;
    }
}
