/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.06.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.Collection;
import java.util.HashSet;
import org.apache.commons.collections.CollectionUtils;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * DF comparative session length
 * Comparison of the current session length and the average session length of the learner's peers.
 *
 * @author Florian Heberle
 */
public class DfComparativeSessionLength extends DidacticFactor {

    private static final String LONGER = "dfComparativeSessionLength_Longer";
    private static final String AVERAGE = "dfComparativeSessionLength_Average";
    private static final String SHORTER = "dfComparativeSessionLength_Shorter";

    /**
     * Constructor
     *
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    public DfComparativeSessionLength(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        super(loadedIndividuals);
    }

    @Override
    protected Collection<String> getExpectedIndividualFragments() {
        Collection<String> fragments = new HashSet<>();
        fragments.add(LONGER);
        fragments.add(AVERAGE);
        fragments.add(SHORTER);
        return fragments;
    }

    @Override
    public OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db) {
        Long currentSessionLength = db.getCurrentSessionLength();
        if (currentSessionLength == null) {
            return null;
        }
        Collection<Collection<Long>> peersSessionLengths = db.getPeerSessionLengths();
        if (CollectionUtils.isEmpty(peersSessionLengths)) {
            return null;
        }
        long avg = 0;
        long ctr = 0;
        for (Collection<Long> peerSessionLength : peersSessionLengths) {
            if (!CollectionUtils.isEmpty(peerSessionLength)) {
                long sum = 0;
                for (Long length : peerSessionLength) {
                    if (length != null) {
                        sum += length;
                    }
                }
                avg += sum;
                ctr++;
            }
        }
        if (ctr == 0) {
            return null;
        } else {
            avg /= ctr;
        }
        long val = currentSessionLength * 100 / avg;
        if (val >= 130) {
            return new OWLNamedIndividual[]{individualsMap.get(LONGER)};
        } else if (val <= 50) {
            return new OWLNamedIndividual[]{individualsMap.get(SHORTER)};
        }
        return new OWLNamedIndividual[]{individualsMap.get(AVERAGE)};
    }
}
