/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.06.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.types.ConnectionType;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.Collection;
import java.util.HashSet;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * DF connectivity
 * Ranking of the connectivity between the learner's access device and the LMS.
 *
 * @author Florian Heberle
 */
public class DfConnectivity extends DidacticFactor {

    private static final String GOOD = "dfConnectivity_Good";
    private static final String OK = "dfConnectivity_OK";
    private static final String POOR = "dfConnectivity_Poor";

    /**
     * Constructor
     *
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    public DfConnectivity(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        super(loadedIndividuals);
    }

    @Override
    protected Collection<String> getExpectedIndividualFragments() {
        Collection<String> fragments = new HashSet<>();
        fragments.add(GOOD);
        fragments.add(OK);
        fragments.add(POOR);
        return fragments;
    }

    @Override
    public OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db) {
        Integer conStab = db.getConnectionStability();
        ConnectionType conType = db.getConnectionType();
        if (conType == ConnectionType.Unknown && conStab == null) { //No statement possible
            return null;
        } else if (conType == null) { //Only consider conStab
            if (conStab >= 80) {
                return new OWLNamedIndividual[]{individualsMap.get(GOOD)};
            } else if (conStab >= 20) {
                return new OWLNamedIndividual[]{individualsMap.get(OK)};
            } else {
                return new OWLNamedIndividual[]{individualsMap.get(POOR)};
            }
        } else if (conStab == null) { //Only consider conType
            if (conType == ConnectionType.Wire || conType == ConnectionType.Lte || conType == ConnectionType.Hsdpa) {
                return new OWLNamedIndividual[]{individualsMap.get(GOOD)};
            } else if (conType == ConnectionType.Umts) {
                return new OWLNamedIndividual[]{individualsMap.get(OK)};
            } else {
                return new OWLNamedIndividual[]{individualsMap.get(POOR)};
            }
        } else { //Consider both
            if (conType == ConnectionType.Wire) {
                return new OWLNamedIndividual[]{individualsMap.get(GOOD)};
            } else if (conType == ConnectionType.Lte || conType == ConnectionType.Hsdpa) {
                if (conStab >= 40) {
                    return new OWLNamedIndividual[]{individualsMap.get(GOOD)};
                } else if (conStab >= 20) {
                    return new OWLNamedIndividual[]{individualsMap.get(OK)};
                } else {
                    return new OWLNamedIndividual[]{individualsMap.get(POOR)};
                }
            } else if (conType == ConnectionType.Umts) {
                if (conStab >= 80) {
                    return new OWLNamedIndividual[]{individualsMap.get(GOOD)};
                } else if (conStab >= 20) {
                    return new OWLNamedIndividual[]{individualsMap.get(OK)};
                } else {
                    return new OWLNamedIndividual[]{individualsMap.get(POOR)};
                }
            } else {
                if (conStab >= 80) {
                    return new OWLNamedIndividual[]{individualsMap.get(GOOD)};
                } else if (conStab >= 60) {
                    return new OWLNamedIndividual[]{individualsMap.get(OK)};
                } else {
                    return new OWLNamedIndividual[]{individualsMap.get(POOR)};
                }
            }
        }
    }
}
