/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.06.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * DF knowledge type preference
 * Statement about which type of content the learner prefers.
 *
 * @author Florian Heberle
 */
public class DfKnowledgeTypePreference extends DidacticFactor {

    private static final String COOPERATIVE = "dfKnowledgeTypePreference_Cooperative";
    private static final String INTERACTIVE = "dfKnowledgeTypePreference_Interactive";
    private static final String RECEPTIVE = "dfKnowledgeTypePreference_Receptive";

    /**
     * Constructor
     *
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    public DfKnowledgeTypePreference(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        super(loadedIndividuals);
    }

    @Override
    protected Collection<String> getExpectedIndividualFragments() {
        Collection<String> fragments = new HashSet<>();
        fragments.add(COOPERATIVE);
        fragments.add(INTERACTIVE);
        fragments.add(RECEPTIVE);
        return fragments;
    }

    @Override
    public OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db) {
        KT prefKt = getValue(db.getRecomSelections());
        if (prefKt == null) {
            return null;
        } else if (prefKt == KT.Cooperative) {
            return new OWLNamedIndividual[]{individualsMap.get(COOPERATIVE)};
        } else if (prefKt == KT.Interactive) {
            return new OWLNamedIndividual[]{individualsMap.get(INTERACTIVE)};
        } else if (prefKt == KT.Receptive) {
            return new OWLNamedIndividual[]{individualsMap.get(RECEPTIVE)};
        }
        return null;
    }

    enum KT {

        Cooperative,
        Interactive,
        Receptive;
    }

    /**
     * Determines the value of this DF.
     *
     * @param learnerSelections The selections, i.e. the learner updates, of the current learner
     * @param pastRecommendations The recommendations made to the learner
     * @return Null if not determinable or inconclusive.
     */
    static KT getValue(List<Pair<List<KnowledgeObject>, KnowledgeObject>> recomSelections) {
        if (CollectionUtils.isEmpty(recomSelections)) {
            return null;
        }
        int[] count = countKts(recomSelections);
        return compareCtrs(count[0], count[1], count[2]);
    }

    private static int[] countKts(List<Pair<List<KnowledgeObject>, KnowledgeObject>> recomSelections) {
        Integer coopCtr = 0;
        Integer interactiveCtr = 0;
        Integer receptiveCtr = 0;
        for (Pair<List<KnowledgeObject>, KnowledgeObject> recomSelection : recomSelections) {
            List<KnowledgeObject> stepRecom = recomSelection.getLeft();
            KnowledgeObject stepSelection = recomSelection.getRight();
            if (stepRecom != null && stepRecom.size() > 1) {
                if (stepSelection != null) {
                    //Only consider steps in which the learner selected from the recommendations
                    if (stepRecom.contains(stepSelection)) {
                        if (stepSelection.getKnowledgeType().isCooperative()) {
                            coopCtr++;
                        } else if (stepSelection.getKnowledgeType().isInteractive()) {
                            interactiveCtr++;
                        } else if (stepSelection.getKnowledgeType().isReceptive()) {
                            receptiveCtr++;
                        }
                    }
                }
            }
        }
        return new int[]{coopCtr, interactiveCtr, receptiveCtr};
    }

    private static KT compareCtrs(int coopCtr, int interactiveCtr, int receptiveCtr) {
        if (coopCtr > interactiveCtr) {
            if (coopCtr > receptiveCtr) {
                return KT.Cooperative;
            } else if (coopCtr == receptiveCtr) {
                return null;
            } else {
                return KT.Receptive;
            }
        } else if (coopCtr < interactiveCtr) {
            if (interactiveCtr > receptiveCtr) {
                return KT.Interactive;
            } else if (interactiveCtr == receptiveCtr) {
                return null;
            } else {
                return KT.Receptive;
            }
        } else {
            //coopCtr == interactiveCtr
            return null;
        }
    }
}
