/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.06.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.utils.CompletionUtil;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import org.apache.commons.collections.CollectionUtils;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * DF last cc completion
 * Statement about the completion of the last CC based on the completion of the mLP, which is
 * created when the learner switched to a new CC.
 *
 * @author Florian Heberle
 */
public class DfLastCcCompletion extends DidacticFactor {

    private static final String FULL = "dfLastCcCompletion_Full";
    private static final String HIGH = "dfLastCcCompletion_High";
    private static final String MEDIUM = "dfLastCcCompletion_Medium";
    private static final String LOW = "dfLastCcCompletion_Low";
    private static final String VERYLOW = "dfLastCcCompletion_VeryLow";

    /**
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    public DfLastCcCompletion(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        super(loadedIndividuals);
    }

    @Override
    protected Collection<String> getExpectedIndividualFragments() {
        Collection<String> fragments = new HashSet<>();
        fragments.add(FULL);
        fragments.add(HIGH);
        fragments.add(MEDIUM);
        fragments.add(LOW);
        fragments.add(VERYLOW);
        return fragments;
    }

    @Override
    public OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db) {
        Integer value = getValue(slom, db.getHistoryOfCCs(), db.getHistoryOfKOs(), db.getCompletedKOs());
        if (value == null) {
            return null;
        } else if (value >= 100) {
            return new OWLNamedIndividual[]{individualsMap.get(FULL)};
        } else if (value >= 75) {
            return new OWLNamedIndividual[]{individualsMap.get(HIGH)};
        } else if (value >= 50) {
            return new OWLNamedIndividual[]{individualsMap.get(MEDIUM)};
        } else if (value >= 25) {
            return new OWLNamedIndividual[]{individualsMap.get(LOW)};
        } else {
            return new OWLNamedIndividual[]{individualsMap.get(VERYLOW)};
        }
    }

    /**
     * Retrieves the value of this DF.
     *
     * @param slom SLOM data of the course
     * @param historyOfCCs History of CCs the learner has accessed
     * @param historyOfKOs History of KOs the learner has accessed
     * @param completedKOs The KOs the learner has completed
     * @return Maximum percentage completion of the last CC or null if undeterminable.
     */
    public static Integer getValue(SlomMetadata slom, List<ConceptContainer> historyOfCCs, List<KnowledgeObject> historyOfKOs, List<KnowledgeObject> completedKOs) {
        //Determine whether the currentCC != lastCC, i.e. did a CC change occur?
        ConceptContainer lastCc = getLastCc(historyOfCCs, historyOfKOs);
        if (lastCc == null) {
            return null;
        }
        return CompletionUtil.getCcCompletion(lastCc, slom, completedKOs);
    }

    /**
     * Iterates through the KO history and determines the accessed last CC, which is not the current
     * one.
     * This filters KO-refreshs and only returns a value if a switch just occured.
     *
     * @param historyOfCCs The history of accessed CCs
     * @param historyOfKOs The history of accessed KOs
     * @return Null if undeterminable.
     */
    private static ConceptContainer getLastCc(List<ConceptContainer> historyOfCCs, List<KnowledgeObject> historyOfKOs) {
        if (historyOfKOs == null || historyOfKOs.size() < 3) {
            return null;
        }
        KnowledgeObject thisKo = historyOfKOs.get(0);
        Integer stepNo = null;
        for (int i = 1; i < historyOfKOs.size(); i++) {
            KnowledgeObject ko = historyOfKOs.get(i);
            if (ko != thisKo) {
                stepNo = i;
                break;
            }
        }
        if (stepNo == null) {
            return null;
        }
        //KOs can be assigned to multiple CCs. Only take the CC of the step in scope.
        return historyOfCCs.get(stepNo);
    }
}
