/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.06.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * DF media type preference
 * Statement about which type of media the learner prefers.
 *
 * @author Florian Heberle
 */
public class DfMediaTypePreference extends DidacticFactor {

    private static final String COMMUNICATION = "dfMediaTypePreference_Communication";
    private static final String INTERACTION = "dfMediaTypePreference_Interaction";
    private static final String PRESENTATION = "dfMediaTypePreference_Presentation";

    /**
     * Constructor
     *
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    public DfMediaTypePreference(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        super(loadedIndividuals);
    }

    @Override
    protected Collection<String> getExpectedIndividualFragments() {
        Collection<String> fragments = new HashSet<>();
        fragments.add(COMMUNICATION);
        fragments.add(INTERACTION);
        fragments.add(PRESENTATION);
        return fragments;
    }

    @Override
    public OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db) {
        MT prefMt = getValue(db.getRecomSelections());
        if (prefMt == null) {
            return null;
        } else if (prefMt == MT.Communication) {
            return new OWLNamedIndividual[]{individualsMap.get(COMMUNICATION)};
        } else if (prefMt == MT.Interaction) {
            return new OWLNamedIndividual[]{individualsMap.get(INTERACTION)};
        } else if (prefMt == MT.Presentation) {
            return new OWLNamedIndividual[]{individualsMap.get(PRESENTATION)};
        }
        return null;
    }

    enum MT {

        Communication,
        Interaction,
        Presentation;
    }

    /**
     * Determines the value of this DF.
     *
     * @param learnerSelections The selections, i.e. the learner updates, of the current learner
     * @param pastRecommendations The recommendations made to the learner
     * @return Null if not determinable or inconclusive.
     */
    static MT getValue(List<Pair<List<KnowledgeObject>, KnowledgeObject>> recomSelections) {
        if (CollectionUtils.isEmpty(recomSelections)) {
            return null;
        }
        int[] count = countMts(recomSelections);
        return compareCtrs(count[0], count[1], count[2]);
    }

    private static int[] countMts(List<Pair<List<KnowledgeObject>, KnowledgeObject>> recomSelections) {
        Integer communicationCtr = 0;
        Integer interactionCtr = 0;
        Integer presentationCtr = 0;
        for (Pair<List<KnowledgeObject>, KnowledgeObject> recomSelection : recomSelections) {
            List<KnowledgeObject> stepRecom = recomSelection.getLeft();
            KnowledgeObject stepSelection = recomSelection.getRight();
            if (stepRecom != null && stepRecom.size() > 1) {
                if (stepSelection != null) {
                    //Only consider steps in which the learner selected from the recommendations
                    if (stepRecom.contains(stepSelection)) {
                        if (stepSelection.getMediaType().isCommunication()) {
                            communicationCtr++;
                        } else if (stepSelection.getMediaType().isInteraction()) {
                            interactionCtr++;
                        } else if (stepSelection.getMediaType().isPresentation()) {
                            presentationCtr++;
                        }
                    }
                }
            }
        }
        return new int[]{communicationCtr, interactionCtr, presentationCtr};
    }

    private static MT compareCtrs(int communicationCtr, int interactionCtr, int presentationCtr) {
        if (communicationCtr > interactionCtr) {
            if (communicationCtr > presentationCtr) {
                return MT.Communication;
            } else if (communicationCtr == presentationCtr) {
                return null;
            } else {
                return MT.Presentation;
            }
        } else if (communicationCtr < interactionCtr) {
            if (interactionCtr > presentationCtr) {
                return MT.Interaction;
            } else if (interactionCtr == presentationCtr) {
                return null;
            } else {
                return MT.Presentation;
            }
        } else {
            //communicationCtr == interactionCtr
            return null;
        }
    }
}
