/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.06.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.types.DifficultyLevel;
import eu.intuitel.types.LearningAttitude;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.Collection;
import java.util.HashSet;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * DF motivation
 * Statement about the motivation level of the learner in combination of what difficulty level she
 * prefers.
 *
 * @author Florian Heberle
 */
public class DfMotivation extends DidacticFactor {

    private static final String HIGHMOTIVATEDBEGINNER = "dfMotivation_HighMotivatedBeginner";
    private static final String HIGHMOTIVATEDINTERMEDIATE = "dfMotivation_HighMotivatedIntermediate";
    private static final String LOWMOTIVATEDINTERMEDIATE = "dfMotivation_LowMotivatedIntermediate";
    private static final String LOWMOTIVATEDADVANCED = "dfMotivation_LowMotivatedAdvanced";

    /**
     * Constructor
     *
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    public DfMotivation(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        super(loadedIndividuals);
    }

    @Override
    protected Collection<String> getExpectedIndividualFragments() {
        Collection<String> fragments = new HashSet<>();
        fragments.add(HIGHMOTIVATEDBEGINNER);
        fragments.add(HIGHMOTIVATEDINTERMEDIATE);
        fragments.add(LOWMOTIVATEDINTERMEDIATE);
        fragments.add(LOWMOTIVATEDADVANCED);
        return fragments;
    }

    @Override
    public OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db) {
        LearningAttitude motivated = db.getLearningAttitude();
        DifficultyLevel prefDifficulty = db.getPreferredDifficultyLevel();
        if (motivated == null || prefDifficulty == null) {
            return null;
        } else if (motivated == LearningAttitude.motivated && prefDifficulty == DifficultyLevel.Beginner) {
            return new OWLNamedIndividual[]{individualsMap.get(HIGHMOTIVATEDBEGINNER)};
        } else if (motivated == LearningAttitude.motivated && prefDifficulty == DifficultyLevel.Intermediate) {
            return new OWLNamedIndividual[]{individualsMap.get(HIGHMOTIVATEDINTERMEDIATE)};
        } else if (motivated == LearningAttitude.unmotivated && prefDifficulty == DifficultyLevel.Intermediate) {
            return new OWLNamedIndividual[]{individualsMap.get(LOWMOTIVATEDINTERMEDIATE)};
        } else if (motivated == LearningAttitude.unmotivated && prefDifficulty == DifficultyLevel.Advanced) {
            return new OWLNamedIndividual[]{individualsMap.get(LOWMOTIVATEDADVANCED)};
        }
        return null;
    }
}
