/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.06.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import org.apache.commons.collections.CollectionUtils;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * DF recent learning pace
 * Comparison of the actual and estimated learning time of the recent learning steps.
 *
 * @author Florian Heberle
 */
public class DfRecentLearningPace extends DidacticFactor {

    private static final String VERYFAST = "dfRecentLearningPace_VeryFast";
    private static final String FAST = "dfRecentLearningPace_Fast";
    private static final String NORMAL = "dfRecentLearningPace_Normal";
    private static final String SLOW = "dfRecentLearningPace_Slow";
    private static final String VERYSLOW = "dfRecentLearningPace_VerySlow";
    private static final int INSPECTIONSIZE = 10;

    /**
     * Constructor
     *
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    public DfRecentLearningPace(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        super(loadedIndividuals);
    }

    @Override
    protected Collection<String> getExpectedIndividualFragments() {
        Collection<String> fragments = new HashSet<>();
        fragments.add(VERYFAST);
        fragments.add(FAST);
        fragments.add(NORMAL);
        fragments.add(SLOW);
        fragments.add(VERYSLOW);
        return fragments;
    }

    @Override
    public OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db) {
        List<KnowledgeObject> history = db.getHistoryOfKOs();
        if (CollectionUtils.isEmpty(history)) {
            return null;
        }
        if (history.size() <= 3) {
            return null;
            //avoids a "slow learning pace" output if the learner just started learning.
        }
        List<Long> actualLearningTimes = new ArrayList<>();
        List<Long> estLearningTimes = new ArrayList<>();
        for (KnowledgeObject ko : filterKos(history, INSPECTIONSIZE)) {
            actualLearningTimes.add(db.getLearningTime(ko));
            estLearningTimes.add(ko.getEstimatedLearningTime());
        }
        Integer value = getValue(actualLearningTimes, estLearningTimes);
        if (value == null) {
            return null;
        } else if (value >= 150) {
            return new OWLNamedIndividual[]{individualsMap.get(VERYFAST)};
        } else if (value >= 120) {
            return new OWLNamedIndividual[]{individualsMap.get(FAST)};
        } else if (value >= 80) {
            return new OWLNamedIndividual[]{individualsMap.get(NORMAL)};
        } else if (value >= 50) {
            return new OWLNamedIndividual[]{individualsMap.get(SLOW)};
        } else {
            return new OWLNamedIndividual[]{individualsMap.get(VERYSLOW)};
        }
    }

    /**
     * @param kos KOs to be filtered
     * @param maxSize Maximum size of KOs to be returned
     * @return Returns a sub-list of KOs to avoid double entries in the list (i.e. refreshs).
     */
    public static List<KnowledgeObject> filterKos(List<KnowledgeObject> kos, int maxSize) {
        List<KnowledgeObject> filteredKos = new ArrayList<>();
        if (kos == null) {
            return filteredKos;
        }
        for (KnowledgeObject ko : kos) {
            if (!filteredKos.contains(ko)) {
                filteredKos.add(ko);
            }
            if (filteredKos.size() == maxSize) {
                return filteredKos;
            }
        }
        return filteredKos;
    }

    /**
     * Determines the average percentage difference between the two lists of learning times.
     *
     * @param actualLearningTimes The actual learning times the learner needed for a KO
     * @param estLearningTimes The estimated learning time of the KO
     * @return Null in case it could not be determined.
     */
    public static Integer getValue(List<Long> actualLearningTimes, List<Long> estLearningTimes) {
        if (CollectionUtils.isEmpty(actualLearningTimes)) {
            return null;
        }
        if (CollectionUtils.isEmpty(estLearningTimes)) {
            return null;
        }

        //Determine the average learning time of the 10 last learning steps
        long avgALT = 0;
        for (Long actualLT : actualLearningTimes) {
            if (actualLT != null) {
                avgALT += actualLT;
            }
        }
        avgALT /= actualLearningTimes.size();

        //Determine average the estimated learning time of the KOs of the last 10 learning steps
        long avgELT = 0;
        for (Long estLT : estLearningTimes) {
            if (estLT != null) {
                avgELT += estLT;
            }
        }
        avgELT /= estLearningTimes.size();
        if (avgELT == 0) {
            return null;
        }

        //Determine the %-difference between the actual and estimated LT
        return (new Long(avgALT * 100 / avgELT)).intValue();
    }
}
