/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.06.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.Collection;
import java.util.HashSet;
import org.apache.commons.collections.CollectionUtils;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * DF Total Cc Completion
 * Statement about the completion of the course based on the completion of the CCs as measured with
 * the currently selected LPs.
 *
 * @author Florian Heberle
 */
public class DfTotalCcCompletion extends DidacticFactor {

    private static final String FULL = "dfTotalCcCompletion_Full";
    private static final String ALMOSTFULL = "dfTotalCcCompletion_AlmostFull";
    private static final String HALF = "dfTotalCcCompletion_Half";
    private static final String LESSTHANHALF = "dfTotalCcCompletion_LessThanHalf";

    /**
     * Constructor
     *
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    public DfTotalCcCompletion(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        super(loadedIndividuals);
    }

    @Override
    protected Collection<String> getExpectedIndividualFragments() {
        Collection<String> fragments = new HashSet<>();
        fragments.add(FULL);
        fragments.add(ALMOSTFULL);
        fragments.add(HALF);
        fragments.add(LESSTHANHALF);
        return fragments;
    }

    @Override
    public OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db) {
        Integer value = getDfValue(db.getMacroLpOfLearner(), db.getCompletedCCs());
        if(value == null) {
            return null;
        } else if (value >= 100) {
            return new OWLNamedIndividual[]{individualsMap.get(FULL)};
        } else if (value >= 95) {
            return new OWLNamedIndividual[]{individualsMap.get(ALMOSTFULL)};
        } else if (value >= 50) {
            return new OWLNamedIndividual[]{individualsMap.get(HALF)};
        } else if (value >= 0) {
            return new OWLNamedIndividual[]{individualsMap.get(LESSTHANHALF)};
        }
        return null;
    }
    
    /**
     * Determines the value of this DF in terms of a percentage value between 0-100.
     * Inspects each CC on the MLP and determine their completion values (learner's completion of the mLP in the CC). 
     * 
     * @param currentMacroLp The currently selected LP of the learner
     * @param completedCcs The CCs the learner has already completed
     * @return Value between 0-100, or null if not determinable
     */
    public static Integer getDfValue(MacroLearningPathway currentMacroLp, Collection<ConceptContainer> completedCcs) {
        if(currentMacroLp == null || completedCcs == null) {
            return null;
        }
        int totalCcCount = currentMacroLp.getAllReferences().size();
        if (totalCcCount == 0) {
            return null;
        }
        int completedCcsOnMacroLp = 0;
        for (ConceptContainer cc : currentMacroLp.getAllReferences()) {
            if (completedCcs.contains(cc)) {
                completedCcsOnMacroLp++;
            }
        }
        return completedCcsOnMacroLp * 100 / totalCcCount;
    }
}
