/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Contributor: Ovidiu Serban, University of Reading, www.isr.reading.ac.uk
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.didacticFactors;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import org.apache.commons.lang3.Validate;
import org.semanticweb.owlapi.model.OWLNamedIndividual;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Interface for Didactic Factors
 *
 * @author fhe
 */
public abstract class DidacticFactor {

    private static final Logger LOGGER = LoggerFactory.getLogger(DidacticFactor.class);
    /**
     * Map containing DF-Individual-IRI-fragments and their ontological counterpartsm, initialized
     * in the constructor.
     */
    protected Map<String, OWLNamedIndividual> individualsMap = new HashMap<>();

    /**
     * Constructor
     *
     * @param loadedIndividuals Collection of loaded DF-Individuals.
     * @throws NullPointerException Thrown when loadedIndividuals is null
     * @throws IllegalArgumentException Thrown when loadedIndividuals is empty or if there is a
     * mismatch with the expected DF-individuals.
     */
    @SuppressWarnings("OverridableMethodCallInConstructor")
    public DidacticFactor(Collection<OWLNamedIndividual> loadedIndividuals)
            throws NullPointerException, IllegalArgumentException {
        Validate.notEmpty(loadedIndividuals);
        if (getExpectedIndividualFragments().size() != loadedIndividuals.size()) {
            throw new IllegalArgumentException(
                    String.format("Number of expected individuals of DF '%s' does not match the actually loaded count: %d/%d.",
                            this.getClass().getName(), getExpectedIndividualFragments().size(), loadedIndividuals.size()));
        }
        for (OWLNamedIndividual individual : loadedIndividuals) {
            if (!getExpectedIndividualFragments().contains(individual.getIRI().getFragment())) {
                LOGGER.warn(String.format("DfIndividual '%s' was not contained in expected DfIndividual-Set of DF '%s'.",
                        individual.getIRI().getFragment(), this.getClass().getName()));
            } else {
                individualsMap.put(individual.getIRI().getFragment(), individual);
            }
        }
    }

    /**
     * Returns a Collection of IRI fragments of expected DF-Individuals for initialization purposes.
     *
     * @return Collection containing all fragments of the DF.
     */
    protected abstract Collection<String> getExpectedIndividualFragments();

    /**
     * Determines the suitable DF-Individual for the given learning situation.
     *
     * @param slom SLOM metadata in scope
     * @param db Database interface
     * @return Null if no individual applies
     */
    public abstract OWLNamedIndividual[] getDfIndividuals(SlomMetadata slom, DbAccessUtil db);
}
