/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-03-12
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.lpMetric;

import eu.intuitel.lpmProcessor.models.LpEvaluationResult;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.apache.commons.collections.CollectionUtils;

/**
 * LP evaluation metric that inspects the learning steps the learner took and determines which LPs
 * have the clostest match to it.
 *
 * @author fhe
 */
public class LpMetricHistoryWalker implements ILpMetric {

    private static final int INSPECTIONSIZE = 2;

    @Override
    public LpEvaluationResult.EvaluationMethod getMethod() {
        return LpEvaluationResult.EvaluationMethod.HistoryWalker;
    }

    @Override
    public Collection<? extends LearningPathway> getLp(SlomMetadata slom, DbAccessUtil db, Class<? extends LearningPathway> lpClass) {
        List<? extends LearningObject> lHistory = lpClass.equals(MicroLearningPathway.class)
                ? db.getHistoryOfKOs(INSPECTIONSIZE) : db.getHistoryOfCCs(INSPECTIONSIZE);

        List<? extends LearningObject> userPath = getUserPath(db, lpClass);
        if (userPath.isEmpty()) {
            return new HashSet<>();
        }
        Set<? extends LearningPathway> lpInspectionSet = lpClass == MicroLearningPathway.class ? slom.getMicroLps() : slom.getMacroLps();
        int matchCountMax = 0;
        Collection<LearningPathway> resultSet = new HashSet<>();
        for (LearningPathway lpInScope : lpInspectionSet) {
            int matchCount = getMatchCountFor(lpInScope, userPath);
            if (matchCount >= matchCountMax) {
                if (matchCount > matchCountMax) {
                    resultSet.clear();
                }
                resultSet.add(lpInScope);
                matchCountMax = matchCount;
            }
        }
        return resultSet;
    }

    /**
     * Determeins the most suitable LPs for the learner by inspecting her learning history.
     *
     * @param slom SLOM data of the course
     * @param lHistory History of LO selections (KOs for mLPs, CCs for MLPs)
     * @param lpClass Class of LPs in scope
     * @return Never null.
     */
    public Collection<? extends LearningPathway> getLp(SlomMetadata slom, List<? extends LearningObject> lHistory, Class<? extends LearningPathway> lpClass) {
        Collection<LearningPathway> resultSet = new HashSet<>();
        if (slom == null || CollectionUtils.isEmpty(lHistory) || lpClass == null) {
            return resultSet;
        }
        if (lpClass == MicroLearningPathway.class && lHistory.get(0).getClass() == ConceptContainer.class) {
            return resultSet;
        }
        Set<? extends LearningPathway> lpInspectionSet = lpClass == MicroLearningPathway.class
                ? slom.getMicroLps() : slom.getMacroLps();
        int matchCountMax = 0;
        for (LearningPathway lpInScope : lpInspectionSet) {
            int matchCount = getMatchCountFor(lpInScope, lHistory);
            if (matchCount >= matchCountMax) {
                if (matchCount > matchCountMax) {
                    resultSet.clear();
                }
                resultSet.add(lpInScope);
                matchCountMax = matchCount;
            }
        }
        return resultSet;
    }

    /**
     * Determines the subSet of the learner's learning history for the given type of LP.
     *
     * @param dbCom Database interface
     * @param lpClass Class of LPs used to determine if KOs or CCs are needed
     * @return Returns an empty list if no elements are available.
     */
    private List<? extends LearningObject> getUserPath(DbAccessUtil db, Class<? extends LearningPathway> lpClass) {
        List<? extends LearningObject> result;
        if (lpClass == MicroLearningPathway.class) {
            result = db.getHistoryOfKOs(INSPECTIONSIZE);
        } else {
            result = db.getHistoryOfCCs(INSPECTIONSIZE);
        }
        if (result == null) {
            return new ArrayList<>();
        } else {
            return result;
        }

    }

    /**
     * Counts the matches of learning steps of the student and steps defined in the given LP.
     *
     * @param lpInScope LP to be inspected
     * @param path Learning steps of the learner
     * @return Integer value larger or equal to zero.
     */
    private int getMatchCountFor(LearningPathway lpInScope, List<? extends LearningObject> path) {
        int matchCount = 0;
        for (int i = path.size() - 1; i >= 1; i--) {
            LearningObject current = path.get(i);
            LearningObject previous = path.get(i - 1);
            if (lpInScope.getReferencersFor(current).contains(previous)) {
                matchCount++;
            }
        }
        return matchCount;
    }
}
