/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.lpMetric;

import eu.intuitel.lpmProcessor.models.LpEvaluationResult;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Util to determine a fitting Learning Pathway for a learner based on the hypercube method.
 *
 * @author fhe
 */
public class LpMetricHypercube implements ILpMetric {

    private static final Logger logger = LoggerFactory.getLogger(LpMetricHypercube.class);

    @Override
    public LpEvaluationResult.EvaluationMethod getMethod() {
        return LpEvaluationResult.EvaluationMethod.Hypercube;
    }

    /**
     * Returns the most suitable LP for the given learner on the basis of the hypercube approach
     * without considering the order in which the LOs have been visited.
     *
     * @param slom SLOM metadata to be inspected
     * @param db Communication interface for the user database
     * @param lpClass Class of LPs to inspect
     * @return Empty collection if the LP could not be determined
     */
    @Override
    public Collection<? extends LearningPathway> getLp(SlomMetadata slom, DbAccessUtil db, Class<? extends LearningPathway> lpClass) {
        List<? extends LearningObject> completedLos = lpClass.equals(MicroLearningPathway.class) ? db.getCompletedKOs() : db.getCompletedCCs();
        return getLp(slom, completedLos, lpClass);
    }

    /**
     * Returns the most suitable LP for the given learner on the basis of the hypercube approach
     * without considering the order in which the LOs have been visited.
     *
     * @param slom SLOM metadata to be inspected
     * @param completedLos The completed LOs of the learner (KOs when mLP, CCs when MLP)
     * @param lpClass Class of LPs to inspect
     * @return Empty collection if the LP could not be determined
     */
    public Collection<? extends LearningPathway> getLp(SlomMetadata slom, List<? extends LearningObject> completedLos, Class<? extends LearningPathway> lpClass) {
        Collection<LearningPathway> results = new HashSet<>();
        if (slom == null || CollectionUtils.isEmpty(completedLos)) {
            return results;
        }
        if (lpClass == MicroLearningPathway.class && completedLos.get(0).getClass() == ConceptContainer.class) {
            return results;
        }

        //TODO: This class needs to be adjusted to the final hypercube-algorithm as described in the DELFI2014 paper!
        String hcValLearner = getHypercubeValue(completedLos);
        int hcValMinimum = Integer.MAX_VALUE;
        for (LearningPathway lp : lpClass == MicroLearningPathway.class ? slom.getMicroLps() : slom.getMacroLps()) {
            String hcValLp = getHypercubeValue(lp.getAllReferences());
            int hcValCurrent = getMaskedHammingDistanceFor(hcValLp, hcValLearner);
            if (hcValCurrent < hcValMinimum) {
                results.clear();
                results.add(lp);
                hcValMinimum = hcValCurrent;
            } else if (hcValCurrent == hcValMinimum) {
                results.add(lp);
            }
        }
        return results;
    }

    /**
     * Returns the Hypercube coordinate representation the given LO collection.
     * I.e. each LO is converted into a binary containing a single '1' at the position of the LO's
     * numeric ID.
     * All binaries will then be merged to create a single return value.
     * E.g. Collection{LO(id=1), LO(id=4)} -> hcVal=1001
     *
     * @param los Learning Objects to inspect
     * @return Empty string if LO collection is empty or null
     */
    protected static String getHypercubeValue(Collection<? extends LearningObject> los) {
        if (los == null || los.isEmpty()) {
            return StringUtils.EMPTY;
        }
        List<LearningObject> sortedLos = sortLosByNumericId(los);

        StringBuilder builder = new StringBuilder();
        //Backward iteration, because append adds on the right side
        //Iteration only until second-last index, because last one requires to pad right side with zeros
        for (int loIndex = sortedLos.size() - 1; loIndex >= 1; loIndex--) {
            LearningObject currentLo = sortedLos.get(loIndex);
            builder.append("1");
            LearningObject nextLo = sortedLos.get(loIndex - 1);
            if (currentLo.getNumericId() - nextLo.getNumericId() > 1) {
                //pad right with zeros until next index
                builder.append(StringUtils.leftPad("", currentLo.getNumericId() - nextLo.getNumericId() - 1, "0"));
            }
        }
        builder.append("1");
        if (sortedLos.get(0).getNumericId() > 1) {
            //pad right until index 0
            builder.append(StringUtils.leftPad("", sortedLos.get(0).getNumericId() - 1, "0"));
        }

        return builder.toString();
    }

    /**
     * Sorts the given collection of LOs by their numeric Id.
     *
     * @param los Collection of LOs to sort
     * @return Empty list if collection is empty, otherwise a sorted List of the given LOs
     */
    protected static List<LearningObject> sortLosByNumericId(Collection<? extends LearningObject> los) {
        List<LearningObject> sortedLos = new ArrayList<>(los);
        Comparator<? super LearningObject> c = new Comparator<LearningObject>() {
            @Override
            public int compare(LearningObject o1, LearningObject o2) {
                if (o1.getNumericId() < o2.getNumericId()) {
                    return -1;
                } else if (o1.getNumericId() > o2.getNumericId()) {
                    return 1;
                } else {
                    return 0;
                }
            }
        };
        Collections.sort(sortedLos, c);
        return sortedLos;
    }

    /**
     * Calculates the Hamming Distance for the given strings, but masks the Strings to make the
     * value more accurate for LP-selection.
     * I.e. Strings will be padded left and distance will only be increased when a '1' in the
     * learner's String meets a '0' in the LP string.
     * E.g.
     * lpHcVal=1 and learnerHcVal=1 -> 0,
     * lpHcVal=1 and learnerHcVal=0 -> 0,
     * lpHcVal=0 and learnerHcVal=1 -> 1,
     * lpHcVal=0 and learnerHcVal=0 -> 0
     *
     * @param lpHcVal Hypercube representation value of the Learning Pathway in scope
     * @param learnerHcVal Hypercube representation value of the learner's history
     * @return Integer.MAX_VALUE in case no "1" is contained in lpVal
     */
    protected static int getMaskedHammingDistanceFor(String lpHcVal, String learnerHcVal) {
        if (StringUtils.isBlank(lpHcVal) || learnerHcVal == null || isPlain(lpHcVal)) {
            return Integer.MAX_VALUE;
        }
        String formattedLpHcVal = StringUtils.leftPad(lpHcVal, learnerHcVal.length(), "0");
        String formattedLearnerHcVal = StringUtils.leftPad(learnerHcVal, lpHcVal.length(), "0");
        int dist = 0;
        for (int lpHcValIndex = 0; lpHcValIndex < formattedLpHcVal.length(); lpHcValIndex++) {
            if (formattedLearnerHcVal.charAt(lpHcValIndex) == '1' && formattedLpHcVal.charAt(lpHcValIndex) == '0') {
                dist++;
            }
        }
        return dist;
    }

    /**
     * Returns true if the given binary coordinate representation contains KOs.
     * I.e. hcVal at least contains one "1"
     *
     * @param hcVal binary Hypercube coordinate representation to be inspected
     * @return False if count of "1" is zero, otherwise true
     */
    private static boolean isPlain(String hcVal) {
        return StringUtils.countMatches(hcVal, "1") == 0;
    }
}
