/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-19
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.lpReevaluation;

import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.List;
import org.apache.commons.collections.CollectionUtils;

/**
 * Re-evaluation criterion that inspects if the learner deviates from the LP to a given degree.
 *
 * @author fhe
 */
public class LpReevaluationCriterionDeviation implements ILpReevaluationCriterion {

    /**
     * Threshold for deviation after which a re-evaluation is triggered.
     */
    private static final double THRESHOLD = 0.5;
    /**
     * Number of steps to be inspect
     */
    private static final int STEPS = 10;

    @Override
    @SuppressWarnings("element-type-mismatch") //Is filtered in if-statement: isCc
    public boolean fulfillsCriterion(DbAccessUtil db, LearningPathway lp, List<? extends LearningObject> history) {
        return fulfillsCriterion(lp, history, db.getLastRecommendations());
    }

    /**
     * Determines if the criterion is fulfilled.
     *
     * @param lp LP in scope
     * @param history History of LOs
     * @param lastRecommendations The last recommendations of the learner
     * @return True if fulfilled.
     */
    public boolean fulfillsCriterion(LearningPathway lp, List<? extends LearningObject> history, List<List<KnowledgeObject>> lastRecommendations) {
        if (lp == null || CollectionUtils.isEmpty(history) || CollectionUtils.isEmpty(lastRecommendations)) {
            return false; //Nothing set yet, so nothing to check
        }
        if (history.get(0).isCc()) {
            return false; //Deviation criterion only valid for KOs 
        }

        int inspectionSize = Math.min(STEPS, history.size());
        inspectionSize = Math.min(inspectionSize, lastRecommendations.size());
        if (inspectionSize <= 0) {
            //Nothing to inspect, nothing to deviate from
            return false;
        }

        int matchCount = 0;
        for (int i = 0; i < inspectionSize; i++) {
            if (history.get(i) instanceof KnowledgeObject) {
                KnowledgeObject ko = (KnowledgeObject) history.get(i);
                if (lastRecommendations.get(i).contains(ko)) {
                    matchCount++;
                }
            }
        }
        return (double) matchCount / (double) inspectionSize >= THRESHOLD;
    }

    /**
     * Returns the percentage-threshold after which this criterion triggers.
     *
     * @return Value between 0 and 100.
     */
    public double getThreshold() {
        return THRESHOLD;
    }
}
