/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models;

import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import org.apache.commons.lang3.Validate;

/**
 * Class to store the result of a LP selection process.
 *
 * @author fhe
 */
public class LpEvaluationResult {

    private final LearningPathway determinedLp;
    private final EvaluationMethod evaluationMethod;

    /**
     * Enum containing information about why this specific Lp has been selected.
     */
    public enum EvaluationMethod {

        /**
         * LP is the only available one.
         */
        Unique("Uniqueness", "LP is the only available one."),
        /**
         * LP has been selected because it has the highest LP coverage of yet unvisited LOs.
         */
        Coverage("Coverage", "LP has been selected because it has the highest LP coverage of yet unvisited LOs."),
        /**
         * LP has been selected because it has been chosen by the most learners.
         */
        Popularity("Popularity", "LP has been selected because it has been chosen by the most learners."),
        /**
         * LP has been selected by evaluating the learner history via the hypercube approach.
         */
        Hypercube("Cognitive-Space", "LP has been selected by evaluating the learnering history via a hyperdimensional representation of the learning space."),
        /**
         * LP has been selected by evaluating the learner's learning history.
         */
        HistoryWalker("History", "LP has been selected by evaluating the learning history and matching it to the LP with the highest compliance.");

        private final String TITLE;
        private final String DESCRIPTION;

        private EvaluationMethod(String title, String description) {
            this.TITLE = title;
            Validate.notNull(title);
            this.DESCRIPTION = description;
            Validate.notNull(description);
        }

        @Override
        public String toString() {
            return this.TITLE;
        }

        /**
         * Clear-text title of the method.
         *
         * @return May never be null.
         */
        public String getTitle() {
            return this.TITLE;
        }

        /**
         * Clear-text description of the method.
         *
         * @return May never be null.
         */
        public String getDescription() {
            return this.DESCRIPTION;
        }
    }

    /**
     * Constructor
     *
     * @param determinedLp LP that has been evaluated as reasonable selection
     * @param evaluationMethod Type of evaluation, i.e a specification by which method the
     * {@code selection} has been chosen
     */
    public LpEvaluationResult(LearningPathway determinedLp, EvaluationMethod evaluationMethod) {
        this.determinedLp = determinedLp;
        this.evaluationMethod = evaluationMethod;
    }

    @Override
    public String toString() {
        return determinedLp.toString() + "|" + evaluationMethod;
    }

    /**
     * Returns true if the suggested LP is a macro LP.
     *
     * @return True if macro LP
     */
    public boolean isMacro() {
        return determinedLp.isMacro();
    }

    /**
     * Returns true if the suggested LP is a micro LP.
     *
     * @return True if micro LP
     */
    public boolean isMicro() {
        return determinedLp.isMicro();
    }

    /**
     * Returns the evaluation method of this result, i.e. the method which suggested this LP as a
     * suitable one.
     *
     * @return Method that created this result
     */
    public EvaluationMethod getEvaluationMethod() {
        return evaluationMethod;
    }

    /**
     * Returns the LearningPathway that has been determined.
     *
     * @return LP that has been selected
     */
    public LearningPathway getDeterminedLp() {
        return determinedLp;
    }

    /**
     * Clear-text title of the method.
     *
     * @return May never be null.
     */
    public String getTitle() {
        return evaluationMethod.getTitle();
    }

    /**
     * Clear-text description of the method.
     *
     * @return May never be null.
     */
    public String getDescription() {
        return evaluationMethod.getDescription();
    }
}
