/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models;

import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.IOntologyObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.Ontology;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.Validate;

/**
 * Class to store SLOM metadata.
 *
 * @author stc, fhe
 */
public class SlomMetadata implements Cloneable {

    private final Ontology ccm;
    private final Ontology cm;
    private final Set<KnowledgeObject> kos;
    private final Set<ConceptContainer> ccs;
    private final Set<MicroLearningPathway> microLps;
    private final Set<MacroLearningPathway> macroLps;
    private final Map<ConceptContainer, Collection<KnowledgeObject>> ccToKoRefMap = new HashMap<>();
    private final AtomicBoolean ccmUpdateNeeded = new AtomicBoolean(true);
    private final AtomicBoolean cmUpdateNeeded = new AtomicBoolean(true);

    /**
     * Creates an empty SLOM metadata object.
     *
     * @param ccm Cognitive Content Model
     * @param cm Cognitive Model
     * @param kos Set of KOs
     * @param ccs Set of CCs
     * @param microLps Set of mLPs
     * @param macroLps Set of MLPs
     */
    public SlomMetadata(Ontology ccm, Ontology cm,
            Set<KnowledgeObject> kos, Set<ConceptContainer> ccs,
            Set<MicroLearningPathway> microLps, Set<MacroLearningPathway> macroLps) {
        Validate.notNull(ccm);
        this.ccm = ccm;
        Validate.notNull(cm);
        this.cm = cm;
        this.kos = Collections.unmodifiableSet(kos);
        this.ccs = Collections.unmodifiableSet(ccs);
        this.microLps = Collections.unmodifiableSet(microLps);
        this.macroLps = Collections.unmodifiableSet(macroLps);
        fillCcToKoRefMap();
    }

    @Override
    public String toString() {
        return "[" + ccm.getIri() + "|" + cm.getIri() + "]";
    }

    /**
     * Returns the HashCode of this SLOM instance.
     *
     * @return HashCode of the toString value
     */
    public int getHashCode() {
        return toString().hashCode();
    }

    /**
     * Returns the course ID generated by the IRIs of the CM, CCM.
     *
     * @return May never be null.
     */
    public String getCourseId() {
        return Integer.toString(getHashCode());
    }

    /**
     * KO getter
     *
     * @return Set of KOs which are part of this SLOM metadata object
     */
    public Set<KnowledgeObject> getKos() {
        return kos;
    }

    /**
     * CC getter
     *
     * @return Set of CCs which are part of this SLOM metadata object
     */
    public Set<ConceptContainer> getCcs() {
        return ccs;
    }

    /**
     * Micro LP getter
     *
     * @return Set of micro LPs which are part of this SLOM metadata object
     */
    public Set<MicroLearningPathway> getMicroLps() {
        return microLps;
    }

    /**
     * Macro LP getter
     *
     * @return Set of macro LPs which are part of this SLOM metadata object
     */
    public Set<MacroLearningPathway> getMacroLps() {
        return macroLps;
    }

    /**
     * CCM getter
     *
     * @return The Cognitive Content Model of this SLOM object
     */
    public Ontology getCcm() {
        return ccm;
    }

    /**
     * CM getter
     *
     * @return The Cognitive Model of this SLOM object
     */
    public Ontology getCm() {
        return cm;
    }

    /**
     * Returns true if a KO with the given id is part of this object.
     *
     * @param id Id of KO to look for
     * @return True if the ID is part of the KO set
     */
    public boolean containsKo(String id) {
        return containsOntoObj(id, kos);
    }

    /**
     * Returns true if a CC with the given id is part of this object.
     *
     * @param id Id of CC to look for
     * @return True if the ID is part of the CC set
     */
    public boolean containsCc(String id) {
        return containsOntoObj(id, ccs);
    }

    /**
     * Returns the KO with the provided lms ID.
     *
     * @param koLmsId LMS-ID of the KO to look for
     * @return Null if LMS-ID is unknown.
     */
    public KnowledgeObject getKoByLmsId(String koLmsId) {
        for (KnowledgeObject ko : kos) {
            if (ko.getLinkedLmsElement().equals(koLmsId)) {
                return ko;
            }
        }
        return null;
    }

    /**
     * Returns the KO with the provided IRI if part of this SLOM object.
     *
     * @param koIri IRI to look for
     * @return Null if not found
     */
    public KnowledgeObject getKoByIri(String koIri) {
        return getByIri(koIri, kos);
    }

    /**
     * Returns the CC with the provided IRI if part of this SLOM object.
     *
     * @param ccIri IRI to look for
     * @return Null if not found
     */
    public ConceptContainer getCcByIri(String ccIri) {
        return getByIri(ccIri, ccs);
    }

    /**
     * Returns true if this SLOM instance contains a macro LP with the specified ID.
     *
     * @param lpIri IRI of the LP
     * @return True if this SLOM object contains a macro LP with the given {@code lpIri}
     */
    public boolean containsMacroLp(String lpIri) {
        return containsOntoObj(lpIri, macroLps);
    }

    /**
     * Returns true if this SLOM instance contains a micro LP with the specified ID.
     *
     * @param lpIri IRI of the LP
     * @return True if this SLOM object contains a micro LP with the given {@code lpIri}
     */
    public boolean containsMicroLp(String lpIri) {
        return containsOntoObj(lpIri, microLps);
    }

    /**
     * Returns the micro LP of the given IRI if part of this SLOM instance.
     *
     * @param microLpIri IRI of the LP to look for
     * @return Null if not found
     */
    public MicroLearningPathway getMicroLp(String microLpIri) {
        return getByIri(microLpIri, microLps);
    }

    /**
     * Returns the macro LP of the given IRI if part of this SLOM instance.
     *
     * @param macroLpIri IRI of the LP to look for
     * @return Null if not found
     */
    public MacroLearningPathway getMacroLp(String macroLpIri) {
        return getByIri(macroLpIri, macroLps);
    }

    /**
     * Returns the object of the given collection, if it is part of this collection.
     *
     * @param <T> A type of ontology object, KnowledgeObject, ConceptContainer or a sub-type of
     * Learning Pathway
     * @param iri IRI to look for
     * @param collection Collection of LOs to inspect
     * @return Null if no element was found.
     */
    public static <T extends IOntologyObject> T getByIri(String iri, Collection<T> collection) {
        if (StringUtils.isBlank(iri)) {
            return null;
        }
        for (T element : collection) {
            if (element.hasSameIri(iri)) {
                return element;
            }
        }
        return null;
    }

    /**
     * Tests whether a certain IRI string is part of the given set of OntoObjects.
     *
     * @param iri String representation of the IRI to look for
     * @param collection Collection to inspect
     * @return True if the given {@code iri} is part of the given {@code set}
     */
    public static boolean containsOntoObj(String iri, Collection<? extends IOntologyObject> collection) {
        return getByIri(iri, collection) != null;
    }

    /**
     * Returns all KO references for the given CC.
     *
     * @param cc Concept Container to inspect
     * @return Null in case there are no references.
     */
    public Collection<KnowledgeObject> getReferences(ConceptContainer cc) {
        Collection<KnowledgeObject> refs = ccToKoRefMap.get(cc);
        if (refs == null) {
            return new HashSet<>();
        } else {
            return refs;
        }
    }

    /**
     * Returns all CC references for the given KO.
     *
     * @param ko Knowledge Object to inspect
     * @return Null in case there are no references.
     */
    public Collection<ConceptContainer> getReferences(KnowledgeObject ko) {
        return ko.getCcReferences();
    }

    /**
     * Initializes the ccToKoRefMap.
     */
    private void fillCcToKoRefMap() {
        for (KnowledgeObject ko : kos) {
            for (ConceptContainer cc : ko.getCcReferences()) {
                if (!ccToKoRefMap.containsKey(cc)) {
                    ccToKoRefMap.put(cc, new HashSet<KnowledgeObject>());
                }
                ccToKoRefMap.get(cc).add(ko);
            }
        }
    }

    /**
     * Returns a flag indicating that the CCM of this SLOM object was recently updated and needs to
     * be updated in the INTUITEL Engine.
     *
     * @return True if update is needed.
     */
    public boolean isCcmUpdateNeeded() {
        return ccmUpdateNeeded.getAndSet(false);
    }

    /**
     * Returns a flag indicating that the CM of this SLOM object was recently updated and needs to
     * be updated in the INTUITEL Engine.
     *
     * @return True if update is needed.
     */
    public boolean isCmUpdateNeeded() {
        return cmUpdateNeeded.getAndSet(false);
    }
}
