/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-03-04
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.factories;

import eu.intuitel.lpmProcessor.OntoMngr;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeType;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.Location;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MediaType;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.PedagogicalOntology;
import eu.intuitel.lpmProcessor.utils.OwlUtil;
import java.util.Collection;
import java.util.HashSet;
import java.util.Locale;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.LocaleUtils;
import org.apache.commons.lang3.Validate;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLDataProperty;
import org.semanticweb.owlapi.model.OWLIndividual;
import org.semanticweb.owlapi.model.OWLNamedIndividual;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.reasoner.OWLReasoner;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Factory class for the creation of Knowledge Objects.
 *
 * @author fhe
 */
public class KnowledgeObjectFactory {

    private static final Logger logger = LoggerFactory.getLogger(KnowledgeObjectFactory.class);

    /**
     * Determines all KOs from the given CCM and returns them as a Set.
     *
     * @param ccmOnto CCM ontology representation
     * @param ccmReasoner An OWL reasoner for the given CCM
     * @param ccs Collection of ConceptContainers of the respective CM
     * @return Empty Set if no CCs could be found.
     */
    public static Set<KnowledgeObject> createKoSet(OWLOntology ccmOnto, OWLReasoner ccmReasoner, Collection<ConceptContainer> ccs) {
        try {
            Validate.notNull(ccmOnto);
            Validate.notNull(ccmReasoner);
            Validate.notEmpty(ccs);
            Set<KnowledgeObject> koSet = new HashSet<>();
            for (OWLNamedIndividual koIndividual : OwlUtil.getNamedIndividualsOfClass(ccmReasoner, PedagogicalOntology.KO_CLASS, false)) {
                koSet.add(createNewKo(koIndividual, ccmOnto, ccs, koSet.size() + 1));
            }
            return koSet;
        } catch (Exception ex) {
            logger.error("Exception while loading KOs.", ex);
            return new HashSet<>();
        }
    }

    /**
     * Creates a new KO instance.
     *
     * @param koIndividual Individual to be analysed
     * @param ccmOnto CCM of the individual
     * @param ccs CCs of the connected CM
     * @param numericId Unique numeric ID of this particular KO
     * @return New KO based on the given individual.
     */
    private static KnowledgeObject createNewKo(OWLNamedIndividual koIndividual, OWLOntology ccmOnto, Collection<ConceptContainer> ccs, int numericId) {
        //Data properties
        Integer difficultyLevel = getInteger(ccmOnto, koIndividual, PedagogicalOntology.HAS_DIFFICULTY_LVL);
        Integer eqfLevel = getInteger(ccmOnto, koIndividual, PedagogicalOntology.HAS_EQF_LVL);
        Integer minHeight = getInteger(ccmOnto, koIndividual, PedagogicalOntology.HAS_MIN_HEIGHT);
        Integer minWidth = getInteger(ccmOnto, koIndividual, PedagogicalOntology.HAS_MIN_WIDTH);
        Integer recomAge = getInteger(ccmOnto, koIndividual, PedagogicalOntology.HAS_RECOMMENDED_AGE);
        Integer recomHeight = getInteger(ccmOnto, koIndividual, PedagogicalOntology.HAS_RECOMMENDED_HEIGHT);
        Integer recomWidth = getInteger(ccmOnto, koIndividual, PedagogicalOntology.HAS_RECOMMENDED_WIDTH);
        Integer size = getInteger(ccmOnto, koIndividual, PedagogicalOntology.HAS_SIZE);
        Boolean subtitles = getBoolean(ccmOnto, koIndividual, PedagogicalOntology.HAS_SUBTITLES);
        Boolean suitableForBlind = getBoolean(ccmOnto, koIndividual, PedagogicalOntology.IS_SUITABLE_FOR_BLIND);
        Boolean suitableForDeaf = getBoolean(ccmOnto, koIndividual, PedagogicalOntology.IS_SUITABLE_FOR_DEAF);
        Boolean suitableForMute = getBoolean(ccmOnto, koIndividual, PedagogicalOntology.IS_SUITABLE_FOR_MUTE);
        String linkdWithLmsElement = getString(ccmOnto, koIndividual, PedagogicalOntology.IS_LINKED_WITH_LMS_ELEMENT);
        Locale hasLanguage = localeStringToLocale(getString(ccmOnto, koIndividual, PedagogicalOntology.HAS_LANGUAGE));
        Long estimatedLearningTime = timeStringToLong(getString(ccmOnto, koIndividual, PedagogicalOntology.HAS_ESTIMATED_LEARNING_TIME));

        //Object properties
        KnowledgeType kt = determineKoKnowledgeType(koIndividual, ccmOnto);
        MediaType mt = determineKoMediaType(koIndividual, ccmOnto);
        Location location = determineKoLocation(koIndividual, ccmOnto);
        Set<ConceptContainer> ccConnections = determineCcConnections(koIndividual, ccmOnto, ccs);

        //Annotation properties
        String title = OwlUtil.getAnnotationPropertyValue(ccmOnto, koIndividual,
                OwlUtil.getAnnotationProperty("http://dublincore.org/documents/dces/", "title", OntoMngr.DATAFAC));
        String description = OwlUtil.getAnnotationPropertyValue(ccmOnto, koIndividual,
                OwlUtil.getAnnotationProperty("http://dublincore.org/documents/dces/", "description", OntoMngr.DATAFAC));

        return new KnowledgeObject(koIndividual, numericId, ccConnections, linkdWithLmsElement,
                difficultyLevel, eqfLevel, minHeight, minWidth, recomAge, recomHeight, recomWidth, size,
                estimatedLearningTime, hasLanguage,
                subtitles, suitableForBlind, suitableForDeaf, suitableForMute,
                location, kt, mt,
                title, description);
    }

    /**
     * Returns the Integer value of the given data property.
     *
     * @param ccmOnto OWLOntology representing the CCM of the KO
     * @param koIndividual Individual containing the relevant information
     * @param prop OWLDataProperty to inspect
     * @return Null on error or if no value.
     */
    private static Integer getInteger(OWLOntology ccmOnto, OWLNamedIndividual koIndividual, OWLDataProperty prop) {
        return OwlUtil.getDataPropertyValueAsInteger(prop, koIndividual, ccmOnto);
    }

    /**
     * Returns the String value of the given data property.
     *
     * @param ccmOnto OWLOntology representing the CCM of the KO
     * @param koIndividual Individual containing the relevant information
     * @param prop OWLDataProperty to inspect
     * @return Null on error or if no value.
     */
    private static String getString(OWLOntology ccmOnto, OWLNamedIndividual koIndividual, OWLDataProperty prop) {
        return OwlUtil.getDataPropertyValue(prop, koIndividual, ccmOnto);
    }

    /**
     * Returns the Boolean value of the given data property.
     *
     * @param ccmOnto OWLOntology representing the CCM of the KO
     * @param koIndividual Individual containing the relevant information
     * @param prop OWLDataProperty to inspect
     * @return Null on error or if no value.
     */
    private static Boolean getBoolean(OWLOntology ccmOnto, OWLNamedIndividual koIndividual, OWLDataProperty prop) {
        return OwlUtil.getDataPropertyValueAsBoolean(prop, koIndividual, ccmOnto);
    }

    /**
     * Returns the Double value of the given data property.
     *
     * @param ccmOnto OWLOntology representing the CCM of the KO
     * @param koIndividual Individual containing the relevant information
     * @param prop OWLDataProperty to inspect
     * @return Null on error or if no value.
     */
    private static Double getDouble(OWLOntology ccmOnto, OWLNamedIndividual koIndividual, OWLDataProperty prop) {
        return OwlUtil.getDataPropertyValueAsDouble(prop, koIndividual, ccmOnto);
    }

    /**
     * Determines the KnowledgeType of the given {@code ko}.
     *
     * @param koIndividual Individual containing the relevant information
     * @param ccmOnto OWLOntology representing the CCM of the KO
     * @return Null in case of exception or if none or more than two values are specified.
     */
    private static KnowledgeType determineKoKnowledgeType(OWLNamedIndividual koIndividual, OWLOntology ccmOnto) {
        Set<OWLIndividual> ktIndividuals = OwlUtil.getObjectPropertyValue(koIndividual, PedagogicalOntology.HAS_KT, ccmOnto);
        if (ktIndividuals != null && ktIndividuals.size() == 1) {
            return KnowledgeType.createInstance(ktIndividuals.iterator().next());
        }
        return null;
    }

    /**
     * Determines the MediaType of the given {@code ko}.
     *
     * @param individual Individual containing the relevant information
     * @param ccm OWLOntology representing the CCM of the KO
     * @return Null in case of exception or if none or more than two values are specified.
     */
    private static MediaType determineKoMediaType(OWLNamedIndividual koIndividual, OWLOntology ccmOnto) {
        Set<OWLIndividual> mtIndividuals = OwlUtil.getObjectPropertyValue(koIndividual, PedagogicalOntology.HAS_MT, ccmOnto);
        if (mtIndividuals != null && mtIndividuals.size() == 1) {
            return MediaType.createInstance(mtIndividuals.iterator().next());
        }
        return null;
    }

    /**
     * Determines the Location of the given {@code ko}.
     *
     * @param individual Individual containing the relevant information
     * @param ccm OWLOntology representing the CCM of the KO
     * @param poUri URI of the Pedagogical Ontology
     * @param fac OWLDataFactory as created from an OWLManager
     * @return Null in case of exception or if none or more than two values are specified.
     */
    private static Location determineKoLocation(OWLNamedIndividual individual, OWLOntology ccm) {
        Set<OWLIndividual> locationIndividuals = OwlUtil.getObjectPropertyValue(individual, PedagogicalOntology.HAS_PREFERRED_LOCATION, ccm);
        if (locationIndividuals == null || locationIndividuals.size() != 1) {
            return null;
        }

        OWLNamedIndividual locationIndividual = (OWLNamedIndividual) locationIndividuals.iterator().next();
        if (locationIndividual == null) {
            return null;
        }

        Double longitude = getDouble(ccm, locationIndividual, PedagogicalOntology.HAS_PREFERRED_LOCATION_LONGITUDE);
        if (longitude == null) {
            logger.warn(String.format("Longitude not specified for KO '%s'.", OwlUtil.getIriAsString(individual)));
            return null;
        }

        Double latitude = getDouble(ccm, locationIndividual, PedagogicalOntology.HAS_PREFERRED_LOCATION_LATITUDE);
        if (latitude == null) {
            logger.warn(String.format("Latitude not specified for KO '%s'.", OwlUtil.getIriAsString(individual)));
            return null;
        }

        Integer margin = getInteger(ccm, locationIndividual, PedagogicalOntology.HAS_PREFERRED_LOCATION_MARGIN);
        if (margin == null) {
            logger.warn(String.format("Margin not specified for KO '%s'.", OwlUtil.getIriAsString(individual)));
            return null;
        }

        return new Location(locationIndividual, latitude, longitude, margin);
    }

    /**
     * Determines the connections from the given KO to CCs.
     *
     * @param koIndividual KO to inspect
     * @param ccm CCM to inspect
     * @param ccs List of CCs of the connected CM
     * @return Empty Set in case no connections are available.
     */
    private static Set<ConceptContainer> determineCcConnections(OWLNamedIndividual koIndividual, OWLOntology ccm, Collection<ConceptContainer> ccs) {
        Set<ConceptContainer> connectedCcs = new HashSet<>();
        for (OWLIndividual ccIndividual : OwlUtil.getObjectPropertyValue(koIndividual, PedagogicalOntology.IS_CONTAINED_BY_CC, ccm)) {
            ConceptContainer cc = getCcForIri(OwlUtil.getIri(ccIndividual), ccs);
            if (cc != null) {
                connectedCcs.add(cc);
            } else {
                logger.warn(String.format("KO '%s' references on undefined CC '%s'.",
                        OwlUtil.getIriAsString(koIndividual), OwlUtil.getIriAsString(ccIndividual)));
            }
        }
        return connectedCcs;
    }

    /**
     * Returns the ConceptContainer with the given IRI from the given Collection.
     *
     * @param iri IRI of the CC to look for
     * @param ccs Collection of CCs to inspect
     * @return Null if CC was not part of the Collection.
     */
    private static ConceptContainer getCcForIri(IRI iri, Collection<ConceptContainer> ccs) {
        for (ConceptContainer cc : ccs) {
            if (cc.hasSameIri(iri)) {
                return cc;
            }
        }
        return null;
    }

    /**
     * Converts a given time-String in the form of 00:00:00 to its long representation.
     *
     * @param time String to be converted
     * @return Null if not convertable.
     */
    protected static Long timeStringToLong(String time) {
        if (StringUtils.isBlank(time)) {
            return null;
        }
        if (time.length() != 8) {
            logger.debug(String.format("Invalid number of characters for time-string '%s'.", time));
            return null;
        }

        Matcher matcher = Pattern.compile("(\\d\\d)(:)(\\d\\d)(:)(\\d\\d)").matcher(time);
        if (!matcher.find()) {
            return null;
        }
        if (matcher.groupCount() != 5) {
            logger.debug(String.format("Invalid number format for time-string '%s'.", time));
            return null;
        }
        String hourString = matcher.group(1);
        String minString = matcher.group(3);
        String secString = matcher.group(5);

        try {
            long hours = Long.valueOf(hourString);
            long minutes = Long.valueOf(minString);
            long seconds = Long.valueOf(secString);
            return hours * 1000 * 60 * 60 + minutes * 1000 * 60 + seconds * 1000;
        } catch (NumberFormatException ex) {
            logger.debug(String.format("Invalid number format for time-string '%s'.", time), ex);
            return null;
        }
    }

    /**
     * Converts a given locale-String to a Locale object.
     *
     * @param locale String to be converted
     * @return Null in case conversion was not possible.
     */
    protected static Locale localeStringToLocale(String locale) {
        try {
            if (!StringUtils.isBlank(locale) && locale.length() >= 2) {
                return LocaleUtils.toLocale(locale.substring(0, 2));
            }
            return null;
        } catch (IllegalArgumentException ex) {
            logger.debug(String.format("Invalid locale format for locale-string '%s'.", locale), ex.getCause());
            return null;
        }
    }
}
