/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Ovidiu Serban, University of Reading, www.isr.reading.ac.uk
 * Version: [0.01]
 * Date: 2014-03-14
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.factories;

import eu.intuitel.lpmProcessor.OntoMngr;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.LearnerStateOntology;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.LearningModelOntology;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.PedagogicalOntology;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import eu.intuitel.lpmProcessor.utils.OwlUtil;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.apache.commons.lang3.Validate;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Factory class for the creation of Learner State Ontologies.
 *
 * @author fhe
 */
public class LearnerStateOntologyFactory {

    private static final Logger logger = LoggerFactory.getLogger(LearnerStateOntologyFactory.class);
    private static final String LSO_IRI_TEMPLATE = "http://www.intuitel.eu/lsos/%s.owl";
    private static final Map<String, String> PREFIXES;

    static {
        PREFIXES = new HashMap<>();
        PREFIXES.put("intui_PO", PedagogicalOntology.BASE_IRI);
        PREFIXES.put("intui_LMO", LearningModelOntology.BASE_IRI);
    }

    /**
     * Creates a LSO based on an ontology that contains the PO, LMO, CM and CCM.
     *
     * @param learnerId ID of the learner in scope
     * @param slom SLOM metadata for the current course
     * @param db Database interface
     * @param validDfIndividuals Collection of valid DF individuals
     * @return New LSO or null in case of error.
     */
    public static LearnerStateOntology createLsoMerged(String learnerId, SlomMetadata slom, DbAccessUtil db,
            Collection<OWLNamedIndividual> validDfIndividuals) {
        try {
            OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
            OWLDataFactory datafac = manager.getOWLDataFactory();
            OWLOntology lso = createLsoBase(manager, createLsoIri(learnerId),
                    OntoMngr.PO.getOntoObj(),
                    OntoMngr.LMO.getOntoObj(),
                    slom.getCm().getOntoObj(),
                    slom.getCcm().getOntoObj());
            return createLso(manager, datafac, lso, slom, db, validDfIndividuals);
        } catch (OWLOntologyCreationException ex) {
            logger.error(String.format("Exception during merged LSO creation for learner '%s'", learnerId), ex);
            return null;
        }
    }

    /**
     * Creates a LSO, but only includes new axioms.
     *
     * @param learnerId ID of the learner in scope
     * @param slom SLOM metadata for the current course
     * @param db Database interface
     * @param validDfIndividuals Collection of valid DF individuals
     * @return New LSO or null in case of error.
     */
    public static LearnerStateOntology createLsoRaw(String learnerId, SlomMetadata slom, DbAccessUtil db,
            Collection<OWLNamedIndividual> validDfIndividuals) {
        try {
            OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
            OWLDataFactory datafac = manager.getOWLDataFactory();
            OWLOntology lso = createLsoBase(manager, createLsoIri(learnerId));
            return createLso(manager, datafac, lso, slom, db, validDfIndividuals);
        } catch (OWLOntologyCreationException ex) {
            logger.error(String.format("Exception during raw LSO creation for learner '%s'", learnerId), ex);
            return null;
        }
    }

    /**
     * Creates the IRI of the LSO based on the ID of the learner.
     *
     * @param learnerId ID of the learner
     * @return IRI as String
     */
    private static String createLsoIri(String learnerId) {
        return String.format(LSO_IRI_TEMPLATE, learnerId.hashCode());
    }

    /**
     * Creates the base ontology for the LSO, based on the given ontologies.
     *
     * @param manager OWL ontology manager to store the axioms.
     * @param datafac Datafactory with whic hthe axioms are being created
     * @param iri IRI of the ontology to be created
     * @param ontos Array of ontologies to be included
     * @return New ontology instance.
     * @throws OWLOntologyCreationException
     */
    private static OWLOntology createLsoBase(OWLOntologyManager manager, String iri, OWLOntology... ontos)
            throws OWLOntologyCreationException {
        IRI lsoIri = IRI.create(iri);
        if (manager.contains(lsoIri)) {
            manager.removeOntology(manager.getOntology(lsoIri));
        }
        if (ontos.length == 0) {
            return manager.createOntology(lsoIri);
        } else if (ontos.length == 1) {
            Set<OWLOntology> ontoSet = new HashSet<>();
            ontoSet.add(ontos[0]);
            return manager.createOntology(lsoIri, ontoSet);
        } else {
            return OwlUtil.mergeOntologies(manager, lsoIri, ontos);
        }
    }

    /**
     * Creates the a new LSO.
     *
     * @param manager OWL ontology manager to store the axioms.
     * @param datafac Datafactory with whic hthe axioms are being created
     * @param lso Base ontology to be extended
     * @param learnerId ID of the learner
     * @param slom SLOM metadata of the course
     * @param dbCom Database interface
     * @param dfIndividuals Collection of valid DF individuals
     * @return New LSO
     */
    private static LearnerStateOntology createLso(OWLOntologyManager manager, OWLDataFactory datafac, OWLOntology lso, SlomMetadata slom,
            DbAccessUtil db, Collection<OWLNamedIndividual> dfIndividuals) {
        OWLIndividual learnerState = addLearnerStateIndividual(manager, datafac, lso);
        addDfReferences(manager, datafac, lso, learnerState, dfIndividuals);
        markSelectedLp(manager, datafac, lso, db);
        markLoPositions(manager, datafac, lso, db);
        markCompletionStates(manager, datafac, lso, slom, db);
        return new LearnerStateOntology(lso, PREFIXES);
    }

    /**
     * Adds a LS-individual to the given LSO ontology.
     *
     * @param manager OWL ontology manager to store the axioms.
     * @param datafac Datafactory with whic hthe axioms are being created
     * @param lso LSO to alter
     * @return The newly created LS-individual.
     */
    private static OWLIndividual addLearnerStateIndividual(OWLOntologyManager manager, OWLDataFactory datafac, OWLOntology lso) {
        OWLIndividual lsIndividual = datafac.getOWLNamedIndividual(IRI.create(LearningModelOntology.BASE_IRI
                + "learnerState"));
        OWLClassAssertionAxiom classAssertion = datafac.getOWLClassAssertionAxiom(LearningModelOntology.LEARNER_STATE_CLASS, lsIndividual);
        manager.addAxiom(lso, classAssertion);
        return lsIndividual;
    }

    /**
     * Adds the references to the given DF-individuals to the LS-individual.
     *
     * @param manager OWL ontology manager to store the axioms.
     * @param datafac Datafactory with whic hthe axioms are being created
     * @param lso LSO to extend
     * @param learnerState LS to reference from
     * @param dfIndividuals DF-individuals to reference to
     */
    private static void addDfReferences(OWLOntologyManager manager, OWLDataFactory datafac, OWLOntology lso, OWLIndividual learnerState,
            Collection<OWLNamedIndividual> dfIndividuals) {
        for (OWLNamedIndividual dfIndividual : dfIndividuals) {
            manager.addAxiom(lso, datafac.getOWLObjectPropertyAssertionAxiom(LearningModelOntology.HAS_DF_VALUE, learnerState, dfIndividual));
        }
    }

    /**
     * Marks the selected LPs for the given learner.
     *
     * @param manager OWL ontology manager to store the axioms.
     * @param datafac Datafactory with whic hthe axioms are being created
     * @param lso LSO to be extended
     * @param learnerId ID of the learner in scope
     * @param slom SLOM Metadata of the course
     * @param dbCom Database interface
     */
    private static void markSelectedLp(OWLOntologyManager manager, OWLDataFactory datafac, OWLOntology lso, DbAccessUtil db) {
        markSelectedLp(manager, datafac, lso, db.getMacroLpOfLearner());
        markSelectedLp(manager, datafac, lso, db.getMicroLpOfLearner());
    }

    /**
     * Marks the selected LPs for the given learner.
     *
     * @param manager OWL ontology manager to store the axioms.
     * @param datafac Datafactory with whic hthe axioms are being created
     * @param lso LSO to be extended
     * @param lp LP to be marked
     */
    private static void markSelectedLp(OWLOntologyManager manager, OWLDataFactory datafac, OWLOntology lso, LearningPathway lp) {
        Validate.notNull(lp, "The selected LP was null.");
        OWLSubClassOfAxiom subClassAssertion = datafac.getOWLSubClassOfAxiom(lp.getOntoObj(), LearningModelOntology.CURRENT_LP);
        manager.addAxiom(lso, subClassAssertion);
    }

    /**
     * Marks the LO positions (current and previous).
     *
     * @param manager OWL ontology manager to store the axioms.
     * @param datafac Datafactory with whic hthe axioms are being created
     * @param lso LSO to be extended
     * @param learnerId ID of the learner in scope
     * @param slom SLOM Metadata of the course
     * @param dbCom Database interface
     */
    private static void markLoPositions(OWLOntologyManager manager, OWLDataFactory datafac, OWLOntology lso, DbAccessUtil db) {
        KnowledgeObject[] kos = new KnowledgeObject[2];
        kos = db.getHistoryOfKOs(2).toArray(kos);
        Validate.notEmpty(kos, "The history of KOs was empty.");
        markCurrentLo(manager, datafac, lso, kos[0]);
        markPreviousLo(manager, datafac, lso, kos[1]);
        ConceptContainer[] ccs = new ConceptContainer[2];
        ccs = db.getHistoryOfCCs(2).toArray(ccs);
        Validate.notEmpty(kos, "The history of CCs was empty.");
        markCurrentLo(manager, datafac, lso, ccs[0]);
        markPreviousLo(manager, datafac, lso, ccs[1]);
    }

    /**
     * Marks the given LO as current LO.
     *
     * @param manager OWL ontology manager to store the axioms.
     * @param datafac Datafactory with whic hthe axioms are being created
     * @param lso LSO to be extended
     * @param lo LO to be marked
     */
    private static void markCurrentLo(OWLOntologyManager manager, OWLDataFactory datafac, OWLOntology lso, LearningObject lo) {
        if (lo == null) {
            return;
        }
        OWLClassAssertionAxiom classAssertion = datafac.getOWLClassAssertionAxiom(LearningModelOntology.CURRENT_LO, lo.getOntoObj());
        manager.addAxiom(lso, classAssertion);
    }

    /**
     * Marks the given LO as previous LO.
     *
     * @param manager OWL ontology manager to store the axioms.
     * @param datafac Datafactory with whic hthe axioms are being created
     * @param lso LSO to be extended
     * @param lo LO to be marked
     */
    private static void markPreviousLo(OWLOntologyManager manager, OWLDataFactory datafac, OWLOntology lso, LearningObject lo) {
        if (lo != null) {
            OWLClassAssertionAxiom classAssertion = datafac.getOWLClassAssertionAxiom(LearningModelOntology.PREVIOUS_LO, lo.getOntoObj());
            manager.addAxiom(lso, classAssertion);
        }
    }

    /**
     * Adds the completion states to the LOs of the SLOM object.
     *
     * @param manager OWL ontology manager to store the axioms.
     * @param datafac Datafactory with whic hthe axioms are being created
     * @param lso LSO to be extended
     * @param learnerId ID of the learner in scope
     * @param slom SLOM Metadata of the course
     * @param dbCom Database interface
     */
    private static void markCompletionStates(OWLOntologyManager manager, OWLDataFactory datafac, OWLOntology lso, SlomMetadata slom, DbAccessUtil db) {
        for (KnowledgeObject ko : slom.getKos()) {
            Integer grade = db.getKoCompletionGrade(ko.getIriAsString());
            if (grade != null) {
                markCompletionStatus(manager, datafac, lso, ko, grade);
            } else {
                markCompletionStatus(manager, datafac, lso, ko, db.getCompletionStatus(ko));
            }
        }
        for (ConceptContainer cc : slom.getCcs()) {
            markCompletionStatus(manager, datafac, lso, cc, db.getCompletionStatus(cc) ? 100 : 0);
        }
    }

    /**
     * Adds the completion status to the given LO.
     *
     * @param manager OWL ontology manager to store the axioms.
     * @param datafac Datafactory with whic hthe axioms are being created
     * @param lso LSO to be extended
     * @param lo LO to be marked
     * @param completionStatus Completion value to be used as value for categorizing the LO.
     */
    private static void markCompletionStatus(OWLOntologyManager manager, OWLDataFactory datafac, OWLOntology lso, LearningObject lo, Integer completionStatus) {
        OWLClassAssertionAxiom classAssertion;
        if (completionStatus == null || completionStatus <= 10) {
            classAssertion = datafac.getOWLClassAssertionAxiom(LearningModelOntology.UNSEEN_LO, lo.getOntoObj());
        } else if (completionStatus <= 75) {
            classAssertion = datafac.getOWLClassAssertionAxiom(LearningModelOntology.PARTIALLY_COMPLETE_LO, lo.getOntoObj());
        } else {
            classAssertion = datafac.getOWLClassAssertionAxiom(LearningModelOntology.COMPLETE_LO, lo.getOntoObj());
        }
        manager.addAxiom(lso, classAssertion);
    }
}
