/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-03-04
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.factories;

import eu.intuitel.lpmProcessor.didacticFactors.DidacticFactor;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.LearningModelOntology;
import eu.intuitel.lpmProcessor.utils.OwlUtil;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.URISyntaxException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import org.apache.commons.lang3.StringUtils;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLAnnotation;
import org.semanticweb.owlapi.model.OWLAnnotationProperty;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLLiteral;
import org.semanticweb.owlapi.model.OWLNamedIndividual;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.semanticweb.owlapi.reasoner.OWLReasoner;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Factory class to load the Learner Model Ontology from the project resources.
 *
 * @author fhe
 */
public class LearningModelOntologyFactory {

    private static final Logger logger = LoggerFactory.getLogger(LearningModelOntologyFactory.class);
    private static final String LOCAL_LMO_FILENAME = "intui_LM.owl";

    /**
     * Loads the Pedagogical Ontology from project resources via the given ontology objects.
     *
     * @param manager OWLOntologyManager to be used for loading the PO
     * @return Null in case of exception.
     */
    public static LearningModelOntology createNewInstance(OWLOntologyManager manager) {
        try {
            IRI localIri = IRI.create(Thread.currentThread().getContextClassLoader().getResource(LOCAL_LMO_FILENAME));
            OWLOntology onto = manager.loadOntologyFromOntologyDocument(localIri);
            OWLReasoner reasoner = OwlUtil.getOwlReasoner(onto);

            OWLAnnotationProperty versionProp = OwlUtil.getAnnotationProperty("http://www.w3.org/2002/07/owl#", "versionInfo", manager.getOWLDataFactory());
            String versionPropValue = OwlUtil.getAnnotationPropertyValue(onto, versionProp);

            Set<DidacticFactor> dfs = loadDfs(onto, reasoner);
            return new LearningModelOntology(onto, dfs, versionPropValue);
        } catch (URISyntaxException ex) {
            logger.error("URI-Exception while loading the Learning Model Ontology", ex);
        } catch (OWLOntologyCreationException ex) {
            logger.error("Ontology-Exception while loading the Learning Model Ontology", ex);
        }
        return null;
    }

    /**
     * Loads DFs from the given ontology.
     *
     * @param onto LMO to be loaded
     * @param reasoner Reasoner to analyse the given LMO
     * @return Empty set if none was loaded.
     */
    private static Set<DidacticFactor> loadDfs(OWLOntology onto, OWLReasoner reasoner) {
        Set<DidacticFactor> dfs = new HashSet<>();
        for (OWLClass dfSubClass : OwlUtil.getSubClassesOf(LearningModelOntology.DIDACTIC_FACTOR, onto, true)) {
            Collection<OWLNamedIndividual> members = OwlUtil.getNamedIndividualsOfClass(reasoner, dfSubClass, true);
            if (!members.isEmpty()) {
                for (OWLAnnotation rule : dfSubClass.getAnnotations(onto, LearningModelOntology.HAS_RULE)) {
                    OWLLiteral ruleValue = (OWLLiteral) rule.getValue();
                    DidacticFactor df = getDfFromRuleRef(ruleValue.getLiteral(), members);
                    if (df != null) {
                        dfs.add(df);
                    }
                }
            }
        }
        return dfs;
    }

    /**
     * Returns a new DF instance from the rule reference.
     *
     * @param ruleRef Rule reference from the LMO
     * @param members The individuals of the DF in scope
     * @return Null in case the DF could not be initialized.
     */
    private static DidacticFactor getDfFromRuleRef(String ruleRef, Collection<OWLNamedIndividual> members) {
        if (StringUtils.isBlank(ruleRef)) {
            return null;
        }
        DidacticFactor df = null;
        try {
            Constructor c = Class.forName(ruleRef).getConstructor(Collection.class);
            if (c == null) {
                logger.warn("Could not create DF rule, could not find constructor: " + ruleRef);
                return null;
            }
            df = (DidacticFactor) c.newInstance(members);
        } catch (ClassNotFoundException ex) {
            logger.warn(String.format("DF rule class for '%s' could not be found: %s", ruleRef, ex.getMessage()));
        } catch (InstantiationException ex) {
            logger.warn(String.format("DF rule class for '%s' could not be instantiated: %s", ruleRef, ex.getMessage()));
        } catch (IllegalAccessException ex) {
            logger.warn(String.format("DF rule class for '%s' could not be accessed: %s", ruleRef, ex.getMessage()));
        } catch (NoSuchMethodException ex) {
            logger.warn(String.format("DF rule class for '%s' has no such method: %s", ruleRef, ex.getMessage()));
        } catch (InvocationTargetException ex) {
            logger.warn(String.format("DF rule class for '%s' cannot be called: %s", ruleRef, ex.getMessage()));
        }
        return df;
    }
}
