/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Contributor: Ovidiu Serban, University of Reading, www.isr.reading.ac.uk
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.factories;

import eu.intuitel.lpmProcessor.OntoMngr;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.PedagogicalOntology;
import eu.intuitel.lpmProcessor.utils.OwlUtil;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLIndividual;
import org.semanticweb.owlapi.model.OWLNamedIndividual;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.reasoner.OWLReasoner;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Factory class to create new Learning Pathway objects
 *
 * @author fhe
 */
public class LearningPathwayFactory {

    private static final Logger logger = LoggerFactory.getLogger(LearningPathwayFactory.class);

    /**
     * Creates a set of macro LP instances from the CM and determines the CC-couples for this LP.
     *
     * @param cmOnto CM ontology to be inspected
     * @param cmReasoner Reasoner for the given CM ontology
     * @param ccs Collection of CCs in the SLOM data
     * @return Empty set if nothing can be determined.
     */
    public static Set<MacroLearningPathway> createNewMacroLpInstances(OWLOntology cmOnto, OWLReasoner cmReasoner, Collection<ConceptContainer> ccs) {
        try {
            Set<MacroLearningPathway> lpSet = new HashSet<>();
            for (OWLClass lpSubClass : OwlUtil.getSubClassesOf(PedagogicalOntology.MACRO_LP_CLASS, cmOnto, true)) {
                String title = OwlUtil.getAnnotationPropertyValue(cmOnto, lpSubClass,
                        OwlUtil.getAnnotationProperty("http://dublincore.org/documents/dces/", "title", OntoMngr.DATAFAC));
                String description = OwlUtil.getAnnotationPropertyValue(cmOnto, lpSubClass,
                        OwlUtil.getAnnotationProperty("http://dublincore.org/documents/dces/", "description", OntoMngr.DATAFAC));
                MacroLearningPathway lp = new MacroLearningPathway(lpSubClass, title, description);
                Set<LearningObject> predecessors = new HashSet<>();
                Set<LearningObject> successors = new HashSet<>();
                for (OWLNamedIndividual lpIndividual : OwlUtil.getNamedIndividualsOfClass(cmReasoner, lpSubClass, true)) {
                    Set<ConceptContainer> stepPredecessors = new HashSet<>();
                    for (OWLIndividual predecessorIndividual : OwlUtil.getObjectPropertyValue(lpIndividual, PedagogicalOntology.HAS_MACRO_PREDECESSOR, cmOnto)) {
                        ConceptContainer cc = SlomMetadata.getByIri(OwlUtil.getIriAsString(predecessorIndividual), ccs);
                        if (cc != null) {
                            stepPredecessors.add(cc);
                            predecessors.add(cc);
                        }
                    }
                    Set<ConceptContainer> stepSuccessors = new HashSet<>();
                    for (OWLIndividual successorIndividual : OwlUtil.getObjectPropertyValue(lpIndividual, PedagogicalOntology.HAS_MACRO_SUCCESSOR, cmOnto)) {
                        ConceptContainer cc = SlomMetadata.getByIri(OwlUtil.getIriAsString(successorIndividual), ccs);
                        if (cc != null) {
                            stepSuccessors.add(cc);
                            successors.add(cc);
                        }
                    }
                    if (stepPredecessors.size() != 1) {
                        logger.error(String.format("Invalid count of predecessors for %s.", lpIndividual.getIRI().getFragment()));
                        return new HashSet<>();
                    }
                    if (stepSuccessors.size() != 1) {
                        logger.error(String.format("Invalid count of successors for %s.", lpIndividual.getIRI().getFragment()));
                        return new HashSet<>();
                    }
                    for (ConceptContainer predecessorCc : stepPredecessors) {
                        for (ConceptContainer successorCc : stepSuccessors) {
                            lp.addLearningObjectRelation(predecessorCc, successorCc);
                            lp.addLpStep(lpIndividual, predecessorCc, successorCc);
                        }
                    }
                }
                setStarters(predecessors, successors, lp);
                setEnders(predecessors, successors, lp);
                lpSet.add(lp);
            }
            return lpSet;
        } catch (Exception ex) {
            logger.error("Exception while creating new micro LP instances.", ex);
            return new HashSet<>();
        }
    }

    /**
     * Creates a set of micro LP instances from the CCM and determines the KO-couples for this LP.
     *
     * @param ccmOnto CCM ontology to be inspected
     * @param ccmReasoner Reasoner for the given CCM ontology
     * @param kos Collection of KOs of this CCM
     * @return Empty set if nothing can be determined.
     */
    public static Set<MicroLearningPathway> createNewMicroLpInstances(OWLOntology ccmOnto, OWLReasoner ccmReasoner, Collection<KnowledgeObject> kos) {
        try {
            Set<MicroLearningPathway> lpSet = new HashSet<>();
            for (OWLClass lpSubClass : OwlUtil.getSubClassesOf(PedagogicalOntology.MICRO_LP_CLASS, ccmOnto, true)) {
                String title = OwlUtil.getAnnotationPropertyValue(ccmOnto, lpSubClass,
                        OwlUtil.getAnnotationProperty("http://dublincore.org/documents/dces/", "title", OntoMngr.DATAFAC));
                String description = OwlUtil.getAnnotationPropertyValue(ccmOnto, lpSubClass,
                        OwlUtil.getAnnotationProperty("http://dublincore.org/documents/dces/", "description", OntoMngr.DATAFAC));
                MicroLearningPathway lp = new MicroLearningPathway(lpSubClass, title, description);
                Set<LearningObject> predecessors = new HashSet<>();
                Set<LearningObject> successors = new HashSet<>();
                for (OWLNamedIndividual lpIndividual : OwlUtil.getNamedIndividualsOfClass(ccmReasoner, lpSubClass, true)) {
                    Set<KnowledgeObject> stepPredecessors = new HashSet<>();
                    for (OWLIndividual predecessorIndividual : OwlUtil.getObjectPropertyValue(lpIndividual, PedagogicalOntology.HAS_MICRO_PREDECESSOR, ccmOnto)) {
                        KnowledgeObject ko = SlomMetadata.getByIri(OwlUtil.getIriAsString(predecessorIndividual), kos);
                        if (ko != null) {
                            stepPredecessors.add(ko);
                            predecessors.add(ko);
                        }
                    }
                    Set<KnowledgeObject> stepSuccessors = new HashSet<>();
                    for (OWLIndividual successorIndividual : OwlUtil.getObjectPropertyValue(lpIndividual, PedagogicalOntology.HAS_MICRO_SUCCESSOR, ccmOnto)) {
                        KnowledgeObject ko = SlomMetadata.getByIri(OwlUtil.getIriAsString(successorIndividual), kos);
                        if (ko != null) {
                            stepSuccessors.add(ko);
                            successors.add(ko);
                        }
                    }
                    if (stepPredecessors.size() != 1) {
                        logger.error(String.format("Invalid count of predecessors for %s.", lpIndividual.getIRI().getFragment()));
                        return new HashSet<>();
                    }
                    if (stepSuccessors.size() != 1) {
                        logger.error(String.format("Invalid count of successors for %s.", lpIndividual.getIRI().getFragment()));
                        return new HashSet<>();
                    }
                    for (KnowledgeObject predecessorKo : stepPredecessors) {
                        for (KnowledgeObject successorKo : stepSuccessors) {
                            lp.addLearningObjectRelation(predecessorKo, successorKo);
                            lp.addLpStep(lpIndividual, predecessorKo, successorKo);
                        }
                    }
                }
                setStarters(predecessors, successors, lp);
                setEnders(predecessors, successors, lp);
                lpSet.add(lp);
            }
            return lpSet;
        } catch (Exception ex) {
            logger.error("Exception while creating new micro LP instances.", ex);
            return new HashSet<>();
        }
    }

    /**
     * Sets the starters for a given LP.
     *
     * @param predecessors Set of predecessors in the given LP
     * @param successors Set of successors in the given LP
     * @param lp LP in scope
     */
    private static void setStarters(Set<LearningObject> predecessors, Set<LearningObject> successors, LearningPathway lp) {
        Set<LearningObject> starters = new HashSet<>(predecessors);
        starters.removeAll(successors);
        lp.setStarters(starters);
    }

    /**
     * Sets the enders for a given LP.
     *
     * @param predecessors Set of predecessors in the given LP
     * @param successors Set of successors in the given LP
     * @param lp LP in scope
     */
    private static void setEnders(Set<LearningObject> predecessors, Set<LearningObject> successors, LearningPathway lp) {
        Set<LearningObject> enders = new HashSet<>(successors);
        enders.removeAll(predecessors);
        lp.setEnders(enders);
    }
}
