/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.factories;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.Ontology;
import eu.intuitel.lpmProcessor.utils.OwlUtil;
import java.io.File;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Set;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.reasoner.OWLReasoner;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Factory class to create new SLOM objects.
 *
 * @author fhe
 */
public class SlomFactory {

    private static final Logger logger = LoggerFactory.getLogger(SlomFactory.class);

    /**
     * Creates a new SLOM metadata instance from the specified files.
     *
     * @param ccmFile File containing the CCM
     * @param cmFile File containing the CM
     * @return Null on error.
     */
    public static SlomMetadata createNewInstance(File ccmFile, File cmFile) {
        try {
            logger.debug(String.format("Starting to load SLOM data:\n\t'%s'\n\t'%s'",
                    ccmFile.getAbsolutePath(), cmFile.getAbsolutePath()));
            Ontology cm = Ontology.createFromFile(cmFile);
            if (cm == null) {
                logger.error(String.format("CM '%s' was not loaded", cmFile.getAbsolutePath()));
                return null;
            }
            Ontology ccm = Ontology.createFromFile(ccmFile);
            if (ccm == null) {
                logger.error(String.format("CCM '%s' was not loaded", ccmFile.getAbsolutePath()));
                return null;
            }
            SlomMetadata slom = createNewInstance(ccm, cm);
            logger.debug(String.format("Finished loading SLOM data:\n\t'%s'\n\t'%s'.",
                    ccmFile.getAbsolutePath(), cmFile.getAbsolutePath()));
            return slom;
        } catch (OWLOntologyCreationException ex) {
            logger.error(String.format("Exception on reading SLOM data:\n\t'%s'\n\t'%s'",
                    ccmFile.getAbsolutePath(), cmFile.getAbsolutePath()), ex);
            return null;
        }
    }

    /**
     * Creates a new SLOM metadata instance from the specified URLs.
     *
     * @param ccmUrl URL under which the CCM is stored
     * @param cmUrl URL under which the CM is stored
     * @return Null on error.
     */
    public static SlomMetadata createNewInstance(URL ccmUrl, URL cmUrl) {
        try {
            logger.debug(String.format("Starting to load SLOM data:\n\t'%s'\n\t'%s'", ccmUrl, cmUrl));
            Ontology cm = Ontology.createFromUrl(cmUrl);
            if (cm == null) {
                logger.error(String.format("CM '%s' was not loaded", cmUrl));
                return null;
            }
            Ontology ccm = Ontology.createFromUrl(ccmUrl);
            if (ccm == null) {
                logger.error(String.format("CCM '%s' was not loaded", ccmUrl));
                return null;
            }
            SlomMetadata slom = createNewInstance(ccm, cm);
            logger.debug(String.format("Finished loading SLOM data:\n\t'%s'\n\t'%s'", ccmUrl, cmUrl));
            return slom;
        } catch (OWLOntologyCreationException | URISyntaxException ex) {
            logger.error(String.format("Exception on reading SLOM data:\n\t'%s'\n\t'%s'", ccmUrl, cmUrl), ex);
            return null;
        }
    }

    /**
     * Creates a new SLOM metadata instance from the given CM & CCM.
     *
     * @param ccm CCM to use for loading the SLOM data
     * @param cm CM to use for loading the SLOM data
     * @return Null on error.
     */
    private static SlomMetadata createNewInstance(Ontology ccm, Ontology cm) {
        OWLReasoner cmReasoner = OwlUtil.getOwlReasoner(cm.getOntoObj());
        Set<ConceptContainer> ccSet = ConceptContainerFactory.createCcSet(cm.getOntoObj(), cmReasoner);
        if (ccSet.isEmpty()) {
            logger.error("Set of loaded Concept Containers is null.");
            return null;
        } else {
            logger.debug(String.format("%d Concept Containers loaded.", ccSet.size()));
        }

        OWLReasoner ccmReasoner = OwlUtil.getOwlReasoner(ccm.getOntoObj());
        Set<KnowledgeObject> koSet = KnowledgeObjectFactory.createKoSet(ccm.getOntoObj(), ccmReasoner, ccSet);
        if (koSet.isEmpty()) {
            logger.error("Set of loaded Knowledge Objects is null.");
            return null;
        } else {
            logger.debug(String.format("%d Knowledge Objects loaded.", koSet.size()));
        }

        Set<MacroLearningPathway> macroLps = LearningPathwayFactory.createNewMacroLpInstances(cm.getOntoObj(), cmReasoner, ccSet);
        if (macroLps == null || macroLps.isEmpty()) {
            logger.error("No macro LPs found in CCM.");
            return null;
        } else {
            logger.debug(String.format("%d macro LPs loaded.", macroLps.size()));
        }

        Set<MicroLearningPathway> microLps = LearningPathwayFactory.createNewMicroLpInstances(ccm.getOntoObj(), ccmReasoner, koSet);
        if (microLps == null || microLps.isEmpty()) {
            logger.error("No micro LPs found in CCM.");
            return null;
        } else {
            logger.debug(String.format("%d micro LPs loaded.", microLps.size()));
        }

        return new SlomMetadata(ccm, cm, koSet, ccSet, microLps, macroLps);
    }
}
