/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes;

import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.utils.OwlUtil;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.apache.commons.lang3.Validate;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * Wrapper class for Learning Objects.
 *
 * @author fhe
 */
public abstract class LearningObject implements IOntologyObject {

    /**
     * The original OWL representation of this Learning Object.
     */
    protected final OWLNamedIndividual ONTOOBJ;
    /**
     * A number uniquely representing this LO in the SLOM object.
     * This is always a number which only has none or one "1" in binary.
     * I.e.: 0, 1, 2, 4, 8, 16, 32, ...
     */
    protected int numericId;
    /**
     * Map containing the successors of this LO per LP.
     */
    protected final Map<LearningPathway, Set<LearningObject>> successors = new HashMap<>();
    /**
     * Map containing the predecessors of this LO per LP.
     */
    protected final Map<LearningPathway, Set<LearningObject>> predecessor = new HashMap<>();
    /**
     * Set containing the LPs referenced from this LP.
     */
    protected final Set<LearningPathway> referencedLps = new HashSet<>();
    /**
     * The dc:title annotation of this LO.
     */
    protected final String TITLE;
    /**
     * The dc:description annotation of this LO.
     */
    protected final String DESCRIPTION;

    /**
     * Creates a new Learning Object.
     *
     * @param ontoObj Ontology object of this LO
     * @param numericId SLOM-wide unique numeric ID of this LO
     * @param title The dc:title of this element
     * @param description The dc:decription of this element
     */
    public LearningObject(OWLNamedIndividual ontoObj, int numericId, String title, String description) {
        Validate.notNull(ontoObj);
        this.ONTOOBJ = ontoObj;
        Validate.notNull(numericId);
        this.numericId = numericId;
        this.TITLE = title;
        this.DESCRIPTION = description;
    }

    @Override
    public OWLNamedIndividual getOntoObj() {
        return ONTOOBJ;
    }

    @Override
    public boolean hasSameIri(String iri) {
        return iri.equalsIgnoreCase(getIriAsString());
    }

    @Override
    public boolean hasSameIri(IRI iri) {
        return hasSameIri(iri.toString());
    }

    @Override
    public IRI getIri() {
        return OwlUtil.getIri(ONTOOBJ);
    }

    @Override
    public String getIriAsString() {
        return OwlUtil.getIriAsString(ONTOOBJ);
    }

    @Override
    public String toString() {
        return OwlUtil.getIri(ONTOOBJ).getFragment();
    }

    /**
     * Returns the numeric ID of this LO.
     * This is a number that distinctly identfies this LO inside its CM/CCM.
     *
     * @return Unique numeric ID of this LO
     */
    public int getNumericId() {
        return numericId;
    }

    /**
     * Sets the {@code numericId}.
     * This has to be a CM/CCM unique number to distinctly identify this LO.
     *
     * @param numericId ID to be set
     */
    public void setNumericId(int numericId) {
        this.numericId = numericId;
    }

    /**
     * Returns true if this is a KO.
     *
     * @return True if this LO is instantiated as a KnowledgeObject
     */
    public boolean isKo() {
        return this instanceof KnowledgeObject;
    }

    /**
     * Returns true if this is CC.
     *
     * @return True if this LO is instantiated as a ConceptContainer
     */
    public boolean isCc() {
        return this instanceof ConceptContainer;
    }

    /**
     * Adds the specified {@code lo} as successor of this LO for the given {@code lp}.
     *
     * @param <LP> Subtype of LP
     * @param <LO> Subtype of LO
     * @param lp LP on which this LO has the specified successor
     * @param lo Successor to add
     */
    public <LP extends LearningPathway, LO extends LearningObject> void addSuccessor(LP lp, LO lo) {
        addLpRelation(lp, lo, successors);
    }

    /**
     * Adds the specified {@code lo} as predecessor of this LO for the given {@code lp}.
     *
     * @param <LP> Subtype of LP
     * @param <LO> Subtype of LO
     * @param lp LP on which this LO has the specified predecessor
     * @param lo Predecessor to add
     */
    public <LP extends LearningPathway, LO extends LearningObject> void addPredecessor(LP lp, LO lo) {
        addLpRelation(lp, lo, predecessor);
    }

    /**
     * Adds a LP relation (i.e. successor or predecessor) to the given {@code map} for the given
     * {@code lo} and {@code lp}.
     *
     * @param <LP> Subtype of LP
     * @param <LO> Subtype of LO
     * @param lp LP on which this LO has the specified successor/predecessor
     * @param lo Successor/Predecessor to add
     * @param map Map to add the specified LO
     */
    private <LP extends LearningPathway, LO extends LearningObject> void addLpRelation(LP lp, LO lo, Map<LP, Set<LO>> map) {
        if (!map.containsKey(lp)) {
            map.put(lp, new HashSet<LO>());
        }
        if (!referencedLps.contains(lp)) {
            referencedLps.add(lp);
        }
        map.get(lp).add(lo);
    }

    /**
     * Returns true if the LO has a successor for the specified {@code lp}.
     *
     * @param <LP> Subtype of LP
     * @param lp Learning Pathway to inspect
     * @return True if this LO is references on another LO via this LP
     */
    public <LP extends LearningPathway> boolean hasSuccessorFor(LP lp) {
        return successors.containsKey(lp);
    }

    /**
     * Returns true if the LO has a predecessor for the specified {@code lp}.
     *
     * @param <LP> Subtype of LP
     * @param lp Learning Pathway to inspect
     * @return True if this LO is referenced on by another LO via this LP
     */
    public <LP extends LearningPathway> boolean hasPredecessorFor(LP lp) {
        return predecessor.containsKey(lp);
    }

    /**
     * Returns the Set of LOs that are successors of this LO for the given {@code lp}.
     *
     * @param <LP> Subtype of LP
     * @param lp Learning Pathway to inspect
     * @return Emtpy Set if no successors available
     */
    public <LP extends LearningPathway> Set<? extends LearningObject> getDirectSucessorsFor(LP lp) {
        return getLpReferences(lp, successors);
    }

    /**
     * Returns the Set of LOs that are predecessors of this LO for the given {@code lp}.
     *
     * @param <LP> Subtype of LP
     * @param lp Learning Pathway to inspect
     * @return Emtpy Set if no predecessors available
     */
    public <LP extends LearningPathway> Set<? extends LearningObject> getDirectPredecessorsFor(LP lp) {
        return getLpReferences(lp, predecessor);
    }

    /**
     * Returns the Set of LOs that are referenced for the given {@code lp} in the provided
     * {@code map}.
     *
     * @param <LP> Subtype of LP
     * @param <LO> Subtype of LO
     * @param lp Learning Pathway to inspect
     * @param map Map to inspect
     * @return Empty Set if no references available
     */
    protected <LP extends LearningPathway, LO extends LearningObject> Set<? extends LearningObject> getLpReferences(LP lp, Map<LP, Set<LO>> map) {
        if (map.containsKey(lp)) {
            return map.get(lp);
        } else {
            return new HashSet<>();
        }
    }

    /**
     * Returns a subtyped Set of the given Set.
     *
     * @param <LO> Type of LO
     * @param set Set containing the elements to be inspected
     * @param classOfLo Class of target LO
     * @return Empty Set if nothing matches the given class.
     */
    public static <LO extends LearningObject> Set<LO> getAsSubtypedSet(Set<? extends LearningObject> set, Class<LO> classOfLo) {
        Set<LO> newSet = new HashSet<>();
        if (set == null) {
            return newSet;
        }
        if (set.isEmpty()) {
            return newSet;
        }
        for (LearningObject lo : set) {
            if (classOfLo.isAssignableFrom(lo.getClass())) {
                newSet.add(classOfLo.cast(lo));
            }
        }
        return newSet;
    }

    /**
     * Returns the dc:title info of this LO.
     *
     * @return Null if annotation not available.
     */
    public String getTitle() {
        return TITLE;
    }

    /**
     * Returns the dc:description info of this LO.
     *
     * @return Null if annotation is not available.
     */
    public String getDescription() {
        return DESCRIPTION;
    }
}
