/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.apache.commons.lang3.Validate;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * Wrapper class for a Learning Pathway.
 *
 * @author fhe
 */
public abstract class LearningPathway implements IOntologyObject {

    /**
     * The original OWL representation of this Learning Object.
     */
    protected final OWLClass ontoObj;
    /**
     * Map of LOs that reference on other LOs via this LP
     */
    protected Map<LearningObject, Set<LearningObject>> references = new HashMap<>();
    /**
     * Set containing all LOs that are either referenced on by this LP or use this LP to reference
     * on another LO.
     */
    protected Set<LearningObject> usages = new HashSet<>();
    /**
     * Coverage of LOs of this LP as measured on the LO count.
     */
    protected int coverage = -1;
    /**
     * Set of LOs which have successors, but not predecessors.
     */
    protected Set<LearningObject> starters = new HashSet<>();
    /**
     * Set of LOs which have predecessors, but not successors.
     */
    protected Set<LearningObject> enders = new HashSet<>();
    /**
     * Map containing the steps of this LP.
     */
    protected Map<OWLNamedIndividual, LearningObject[]> lpSteps = new HashMap<>();
    /**
     * The dc:title annotation of this LP.
     */
    protected final String TITLE;
    /**
     * The dc:description annotation of this LP.
     */
    protected final String DESCRIPTION;

    /**
     * Creates a new Learning Pathway.
     *
     * @param ontoObj OWLObjectProperty which represents this LP
     * @param title The dc:title of this element
     * @param description The dc:decription of this element
     */
    public LearningPathway(OWLClass ontoObj, String title, String description) {
        Validate.notNull(ontoObj);
        this.ontoObj = ontoObj;
        this.TITLE = title;
        this.DESCRIPTION = description;
    }

    @Override
    public OWLClass getOntoObj() {
        return ontoObj;
    }

    @Override
    public boolean hasSameIri(String iri) {
        return iri.equalsIgnoreCase(getIriAsString());
    }

    @Override
    public boolean hasSameIri(IRI iri) {
        return hasSameIri(iri.toString());
    }

    @Override
    public IRI getIri() {
        return ontoObj.getIRI();
    }

    @Override
    public String getIriAsString() {
        return ontoObj.getIRI().toString();
    }

    @Override
    public String toString() {
        return ontoObj.getIRI().getFragment();
    }

    /**
     * Returns true if this is a micro LP.
     *
     * @return True if this LP is instantiated as a MicroLearningPathway
     */
    public boolean isMicro() {
        return this instanceof MicroLearningPathway;
    }

    /**
     * Returns true if this is a macro LP.
     *
     * @return True if this LP is instantiated as a MacroLearningPathway
     */
    public boolean isMacro() {
        return this instanceof MacroLearningPathway;
    }

    /**
     * Returns the KO/CC coverage of this LP in contrast to total KO/CC count in percent.
     *
     * @param slom The SLOM object of this LP
     * @return Value between 0 and 100 that represents that percentage coverage of CCs/KOs of this
     * LP as measured by the total of CCs/KOs of the respective SLOM object
     */
    public int getCoverage(SlomMetadata slom) {
        if (coverage == -1) {
            int c = 0;
            boolean isMacro = isMacro();
            for (LearningObject lo : isMacro ? slom.getCcs() : slom.getKos()) {
                if (lo.hasSuccessorFor(this)) {
                    c++;
                }
            }
            int total = isMacro ? slom.getCcs().size() : slom.getKos().size();
            coverage = c * 100 / total;
        }
        return coverage;
    }

    /**
     * Adds a LO relation to this LP.
     * i.e. {@code referencer} links on {@code referencee} with this LP.
     *
     * @param <LO> LO Subtype
     * @param referencer LO that references with this LP
     * @param referencee LO that is referenced with this LP
     */
    public <LO extends LearningObject> void addLearningObjectRelation(LO referencer, LO referencee) {
        if (!references.containsKey(referencer)) {
            Set<LearningObject> newSet = new HashSet<>();
            newSet.add(referencee);
            references.put(referencer, newSet);
        } else {
            references.get(referencer).add(referencee);
        }

        usages.add(referencer);
        usages.add(referencee);
        referencer.addSuccessor(this, referencee);
        referencee.addPredecessor(this, referencer);
    }

    /**
     * Returns the a map containing the assignment of LOs with this LP.
     *
     * @return Empty set if this LP is not used by LOs
     */
    public Map<LearningObject, Set<LearningObject>> getReferenceMap() {
        return Collections.unmodifiableMap(references);
    }

    /**
     * Returns a Set of LOs that are referenced on by the specified {@code referencer} by using this
     * LP.
     * I.e. LO1 -> LO2 and LO1 -> LO3, return value for LO1 is Set{LO2, LO3}
     *
     * @param <LO> LO Subtype
     * @param referencer LO that uses this LP to reference on another LO
     * @return Empty set if this LP is not used by the referencer
     */
    public <LO extends LearningObject> Set<? extends LearningObject> getReferenceesFor(LO referencer) {
        return references.containsKey(referencer)
                ? references.get(referencer)
                : new HashSet<LearningObject>();
    }

    /**
     * Returns a Set of LOs that are referencing on the {@code referencee} by using this LP.
     * I.e. LO1 -> LO3 and LO2 -> LO3, return value for LO3 is Set{LO1, LO2}
     *
     * @param <LO> LO Subtype
     * @param referencee LO that is referenced on by this LP
     * @return Empty set if this LP is not used to reference on the referencee
     */
    public <LO extends LearningObject> Set<? extends LearningObject> getReferencersFor(LO referencee) {
        Set<LearningObject> referencers = new HashSet<>();
        for (LearningObject key : references.keySet()) {
            if (references.get(key).contains(referencee)) {
                referencers.add(key);
            }
        }
        return referencers;
    }

    /**
     * Returns a Set of LOs that are either referenced on or are referencing on another LO via this
     * LP.
     * I.e. LO1 -> LO2, return value is Set{LO1, LO2}
     *
     * @return Empty Set if no LOs
     */
    public Set<? extends LearningObject> getAllReferences() {
        return Collections.unmodifiableSet(usages);
    }

    /**
     * Sets the Learning Objects that are starters for this LP. These are all LOs in that are used
     * in this LP which have successors, but no predecessors.
     *
     * @param starters Set of Learning Objects to be set as starters
     */
    public void setStarters(Set<? extends LearningObject> starters) {
        this.starters = new HashSet<>(starters);
    }

    /**
     * Returns all starters of this LP, meaning all LOs that are used in this LP which have
     * successors, but no predecessors.
     *
     * @return Empty set if no starters available. This can only occur, when the LP has not properly
     * been instantiated.
     */
    public Set<? extends LearningObject> getStarters() {
        return Collections.unmodifiableSet(starters);
    }

    /**
     * Sets the Learning Objects that are enders for this LP. These are all LOs in that are used
     * in this LP which have predecessors, but no successors.
     *
     * @param enders Set of Learning Objects to be set as starters
     */
    public void setEnders(Set<? extends LearningObject> enders) {
        this.enders = new HashSet<>(enders);
    }

    /**
     * Returns all enders of this LP, meaning all LOs that are used in this LP which have
     * predecessors, but no successors.
     *
     * @return Empty set if no enders available. This can only occur, when the LP has not properly
     * been instantiated.
     */
    public Set<? extends LearningObject> getEnders() {
        return Collections.unmodifiableSet(enders);
    }

    /**
     * Adds a new LP step to this LP.
     *
     * @param <LO> LO Subtype
     * @param lpIndividual LP-step individual in scope
     * @param predecessor Predecessor as defined in this step
     * @param successor Successor as defined in this step
     */
    public <LO extends LearningObject> void addLpStep(OWLNamedIndividual lpIndividual, LO predecessor, LO successor) {
        lpSteps.put(lpIndividual, new LearningObject[]{predecessor, successor});
    }

    /**
     * Returns the LP-steps of this LP.
     *
     * @return Empty Map if no steps available.
     */
    public Map<OWLNamedIndividual, LearningObject[]> getLpSteps() {
        return Collections.unmodifiableMap(lpSteps);
    }

    /**
     * Returns the dc:title info of this LO.
     *
     * @return Null if annotation not available.
     */
    public String getTitle() {
        return TITLE;
    }

    /**
     * Returns the dc:description info of this LO.
     *
     * @return Null if annotation is not available.
     */
    public String getDescription() {
        return DESCRIPTION;
    }

    /**
     * Returns the provided set of LPs as a subtyped set.
     *
     * @param <LP> Subtype of LP
     * @param set Set to be converted
     * @param classOfLp Subclass of lps to convert to
     * @return Empty set if params are invalid or if content unassignable to specified class.
     */
    public static <LP extends LearningPathway> Set<LP> getAsSubtypedSet(Set<? extends LearningPathway> set, Class<LP> classOfLp) {
        Set<LP> newSet = new HashSet<>();
        if (set == null) {
            return newSet;
        }
        if (set.isEmpty()) {
            return newSet;
        }
        for (LearningPathway lp : set) {
            if (classOfLp.isAssignableFrom(lp.getClass())) {
                newSet.add(classOfLp.cast(lp));
            }
        }
        return newSet;
    }

}
