/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes;

import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import java.util.Collections;
import java.util.Locale;
import java.util.Set;
import org.semanticweb.owlapi.model.OWLNamedIndividual;

/**
 * Wrapper class for a Knowledge Object.
 *
 * @author fhe
 */
public class KnowledgeObject extends LearningObject {

    private final Set<ConceptContainer> references;
    private final Integer difficultyLevel;
    private final Integer eqfLevel;
    private final Integer minimalHeight;
    private final Integer minimalWidth;
    private final Integer recommendedAge;
    private final Integer recommendedHeight;
    private final Integer recommendedWidth;
    private final Integer digitalSize;
    private final Long estimatedLearningTime;
    private final Locale language;
    private final Boolean subtiles;
    private final Boolean suitableForBlind;
    private final Boolean suitableForDeaf;
    private final Boolean suitableForMute;
    private final String linkedLmsElement;
    private final KnowledgeType knowledgeType;
    private final Location location;
    private final MediaType mediaType;

    /**
     * Creates a new Knowledge Object.
     *
     * @param ko Ontology object representing this KO
     * @param numericId SLOM-unique numeric identifier of this KO
     * @param references The isContainedBy references of this KO to CCs
     * @param linkedLmsElement ID of the KO in the LMS
     * @param difficultyLevel Recommended level of diffiulty
     * @param eqfLevel Recommended EQF level
     * @param minimalHeight Minimal height (pixel)
     * @param minimalWidth Minimal width (pixel)
     * @param recommendedAge Recommended age
     * @param recommendedHeight Recommended height (pixel)
     * @param recommendedWidth Recommended width (pxiel)
     * @param digitalSize Size in KB
     * @param estimatedLearningTime Estimate learing time
     * @param language The language in which the KO is given
     * @param subtiles KO contains subtitles (if video/audio)
     * @param suitableForBlind KO is suitable for blind people
     * @param suitableForDeaf KO is suitable for deaf people
     * @param suitableForMute KO is suitable for mute people
     * @param location Location in which the KO should best be learned
     * @param knowledgeType The KT of this KO
     * @param mediaType The MT of this KO
     * @param title The dc:title of this element
     * @param description The dc:description of this element
     */
    public KnowledgeObject(OWLNamedIndividual ko, int numericId, Set<ConceptContainer> references, String linkedLmsElement,
            Integer difficultyLevel, Integer eqfLevel, Integer minimalHeight, Integer minimalWidth,
            Integer recommendedAge, Integer recommendedHeight, Integer recommendedWidth, Integer digitalSize,
            Long estimatedLearningTime, Locale language, Boolean subtiles, Boolean suitableForBlind,
            Boolean suitableForDeaf, Boolean suitableForMute, Location location,
            KnowledgeType knowledgeType, MediaType mediaType,
            String title, String description) {
        super(ko, numericId, title, description);
        this.references = references;
        this.difficultyLevel = difficultyLevel;
        this.eqfLevel = eqfLevel;
        this.minimalHeight = minimalHeight;
        this.minimalWidth = minimalWidth;
        this.recommendedAge = recommendedAge;
        this.recommendedHeight = recommendedHeight;
        this.recommendedWidth = recommendedWidth;
        this.digitalSize = digitalSize;
        this.estimatedLearningTime = estimatedLearningTime;
        this.language = language;
        this.subtiles = subtiles;
        this.suitableForBlind = suitableForBlind;
        this.suitableForDeaf = suitableForDeaf;
        this.suitableForMute = suitableForMute;
        this.linkedLmsElement = linkedLmsElement;
        this.knowledgeType = knowledgeType;
        this.location = location;
        this.mediaType = mediaType;
    }

    /**
     * Returns a Set containing the CCs that are connected to this KO.
     *
     * @return Set of CCs that are connected to this KO
     */
    public Set<ConceptContainer> getCcReferences() {
        return Collections.unmodifiableSet(references);
    }

    /**
     * Returns true when this KO is contained in the given {@code cc}.
     *
     * @param cc CC in scope
     * @return True if the given {@code cc} is contained in the CC references
     */
    public boolean isContainedIn(ConceptContainer cc) {
        return references.contains(cc);
    }

    /**
     * Returns the difficulty level of the KO.
     * This is can have the values 1, 2 and 3.
     *
     * @return Null in case the value is not available.
     */
    public Integer getDifficultyLevel() {
        return difficultyLevel;
    }

    /**
     * Returns the EQF level of the KO.
     * This is can have the values 1-8.
     *
     * @return Null in case the value is not available.
     */
    public Integer getEqfLevel() {
        return eqfLevel;
    }

    /**
     * Returns the minimal height of the KO.
     *
     * @return Null in case the value is not available.
     */
    public Integer getMinimalHeight() {
        return minimalHeight;
    }

    /**
     * Returns the minimal width of the KO.
     *
     * @return Null in case the value is not available.
     */
    public Integer getMinimalWidth() {
        return minimalWidth;
    }

    /**
     * Returns the recommended age of the KO.
     *
     * @return Null in case the value is not available.
     */
    public Integer getRecommendedAge() {
        return recommendedAge;
    }

    /**
     * Returns the recommended height of the KO.
     *
     * @return Null in case the value is not available.
     */
    public Integer getRecommededHeight() {
        return recommendedHeight;
    }

    /**
     * Returns the recommended width of the KO.
     *
     * @return Null in case the value is not available.
     */
    public Integer getRecommendedWidth() {
        return recommendedWidth;
    }

    /**
     * Returns the digital size (i.e. KB) of the KO.
     *
     * @return Null in case the value is not available.
     */
    public Integer getDigitalSize() {
        return digitalSize;
    }

    /**
     * Returns the estimated learning time of the KO.
     * This is the specified time, but converted to long.
     *
     * @return Null in case the value is not available.
     */
    public Long getEstimatedLearningTime() {
        return estimatedLearningTime;
    }

    /**
     * Returns the language of the KO.
     *
     * @return Null in case the value is not available.
     */
    public Locale getLanguage() {
        return language;
    }

    /**
     * Returns true if the KO has subtitles.
     *
     * @return Null in case the value is not available.
     */
    public Boolean getSubtiles() {
        return subtiles;
    }

    /**
     * Returns true if the KO is suitable for blind learners.
     *
     * @return Null in case the value is not available.
     */
    public Boolean getSuitableForBlind() {
        return suitableForBlind;
    }

    /**
     * Returns true if the KO is suitable for deaf learners.
     *
     * @return Null in case the value is not available.
     */
    public Boolean getSuitableForDeaf() {
        return suitableForDeaf;
    }

    /**
     * Returns true if the KO is suitable for mute learners.
     *
     * @return Null in case the value is not available.
     */
    public Boolean getSuitableForMute() {
        return suitableForMute;
    }

    /**
     * Returns the String identifier of the element with which this KO is linked.
     *
     * @return Null in case the value is not available. (Must never be the case!)
     */
    public String getLinkedLmsElement() {
        return linkedLmsElement;
    }

    /**
     * Returns the KnowledgeType of this KO.
     *
     * @return Null if not available
     */
    public KnowledgeType getKnowledgeType() {
        return knowledgeType;
    }

    /**
     * Returns the Location of this KO.
     *
     * @return Null if not available
     */
    public Location getLocation() {
        return location;
    }

    /**
     * Returns the MediaType of this KO.
     *
     * @return Null if not available
     */
    public MediaType getMediaType() {
        return mediaType;
    }

    /**
     * Checks whether the KO is categorized as an automatically assessed test.
     *
     * @return True if KO is a test.
     */
    public boolean isAutomaticallyAssessedTest() {
        if (knowledgeType == null) {
            return false;
        }
        return knowledgeType.isAutomaticallyAssessedTest();
    }

    @Override
    public Set<KnowledgeObject> getDirectSucessorsFor(LearningPathway lp) {
        return LearningObject.getAsSubtypedSet(getLpReferences(lp, successors), KnowledgeObject.class);
    }

    @Override
    public Set<KnowledgeObject> getDirectPredecessorsFor(LearningPathway lp) {
        return LearningObject.getAsSubtypedSet(getLpReferences(lp, predecessor), KnowledgeObject.class);
    }

    /**
     * @return A set of LPs that are used to reference from or on this KO.
     */
    public Set<MicroLearningPathway> getLearningPathwayReferences() {
        return LearningPathway.getAsSubtypedSet(referencedLps, MicroLearningPathway.class);
    }
}
