/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-20
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes;

import eu.intuitel.lpmProcessor.OntoMngr;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.IOntologyObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.PedagogicalOntology;
import eu.intuitel.lpmProcessor.utils.OwlUtil;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.apache.commons.lang3.Validate;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLIndividual;
import org.semanticweb.owlapi.model.OWLObject;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.PrefixManager;
import org.semanticweb.owlapi.util.DefaultPrefixManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Class to wrap the Knowledge Types of Knowledge Objects.
 *
 * @author fhe
 */
public class KnowledgeType implements IOntologyObject {

    private static final Logger logger = LoggerFactory.getLogger(KnowledgeType.class);
    /**
     * KT cooperative knowledge
     */
    public static final OWLClass COOPERATIVEKNOWLEDGE;
    /**
     * KT Pique curiosity cooperation
     */
    public static final OWLClass PIQUECURIOSITYCOOPERATION;
    /**
     * KT planned cooperation
     */
    public static final OWLClass PLANNEDCOOPERATION;
    /**
     * KT conference planned
     */
    public static final OWLClass CONFERENCEPLANNED;
    /**
     * KT counselling planned
     */
    public static final OWLClass COUNSELLINGPLANNED;
    /**
     * KT human feedback planned
     */
    public static final OWLClass HUMANFEEDBACKPLANNED;
    /**
     * KT roleplay planned
     */
    public static final OWLClass ROLEPLAYPLANNED;
    /**
     * KT talk planned
     */
    public static final OWLClass TALKPLANNED;
    /**
     * KT workgroup planned
     */
    public static final OWLClass WORKGROUPPLANNED;
    /**
     * KT spontaneous cooperation
     */
    public static final OWLClass SPONTANEOUSCOOPERATION;
    /**
     * KT community communications spontaneous
     */
    public static final OWLClass COMMUNITYCOMMUNICATIONSPONANTEOUS;
    /**
     * KT counselling on learning methods spontaneous
     */
    public static final OWLClass COUNSELLINGONLEARNINGMETHODSPONTANEOUS;
    /**
     * KT question to counsellor spontaneous
     */
    public static final OWLClass QUESTIONTOCOUNSELLORSPONTANEOUS;
    /**
     * KT interactive knowledge
     */
    public static final OWLClass INTERACTIVEKNOWLEDGE;
    /**
     * KT assignment interactive automatic assessment
     */
    public static final OWLClass ASSIGNMENTINTERACTIVEAUTOMATICASSESSMENT;
    /**
     * KT arranging assignment
     */
    public static final OWLClass ARRANGINGASSIGNMENT;
    /**
     * KT articulation assignment
     */
    public static final OWLClass ARTICULATIONASSIGNMENT;
    /**
     * KT detecting assignment
     */
    public static final OWLClass DETECTINGASSIGNMENT;
    /**
     * KT discirimiation assignment
     */
    public static final OWLClass DISCRIMINATIONASSIGNMENT;
    /**
     * KT educational game assignment
     */
    public static final OWLClass EDUCATIONALGAMEASSIGNMENT;
    /**
     * KT fill in the blanks assignment
     */
    public static final OWLClass FILLINTHEBLANKASSIGNMENT;
    /**
     * KT mapping assignment
     */
    public static final OWLClass MAPPINGASSIGNMENT;
    /**
     * KT repetition assignment
     */
    public static final OWLClass REPETITIONASSIGNMENT;
    /**
     * KT delayed repeating
     */
    public static final OWLClass DELAYEDREPEATING;
    /**
     * KT drill and practice repeating
     */
    public static final OWLClass DRILLANDPRACTICEREPEATING;
    /**
     * KT sequencing assignment
     */
    public static final OWLClass SEQUENCINGASSIGNMENT;
    /**
     * KT build sentence sequencing
     */
    public static final OWLClass BUILDSENTENCESEQUENCING;
    /**
     * KT technical construction sequencing
     */
    public static final OWLClass TECHNICALCONSTRUCTIONSEQUENCING;
    /**
     * KT short answer assignment
     */
    public static final OWLClass SHORTANSWERASSIGNMENT;
    /**
     * KT simulation assignment
     */
    public static final OWLClass SIMULATIONASSIGNMENT;
    /**
     * KT spelling assignment
     */
    public static final OWLClass SPELLINGASSIGNMENT;
    /**
     * KT ticking assignment
     */
    public static final OWLClass TICKINGASSIGNMENT;
    /**
     * KT multiple choice assignment
     */
    public static final OWLClass MULTIPLECHOICEASSIGNMENT;
    /**
     * KT single choice assignment
     */
    public static final OWLClass SINGLECHOICEASSIGNMENT;
    /**
     * KT choose methods single choice
     */
    public static final OWLClass CHOOSEMETHODSINGLECHOICE;
    /**
     * KT choose question single choice
     */
    public static final OWLClass CHOOSEQUESTIONSINGLECHOICE;
    /**
     * KT true false assignment
     */
    public static final OWLClass TRUEFALSEASSIGNMENT;
    /**
     * KT assignment interactive no assessment
     */
    public static final OWLClass ASSIGNMENTINTERACTIVENOASSESSMENT;
    /**
     * KT assignment interactive self assessment
     */
    public static final OWLClass ASSIGNMENTINTERACTIVESELFASSESSMENT;
    /**
     * KT hand in assignment
     */
    public static final OWLClass HANDINASSIGNMENT;
    /**
     * KT formulate question hand in
     */
    public static final OWLClass FORMULATEQUESTIONHANDIN;
    /**
     * KT gather evidence hand in
     */
    public static final OWLClass GATHEREVIDENCEHANDIN;
    /**
     * KT plan investigation hand in
     */
    public static final OWLClass PLANINVESTIGATIONHANDIN;
    /**
     * KT present evidence hand in
     */
    public static final OWLClass PRESENTEVIDENCEHANDIN;
    /**
     * KT propose hypothesis hand in
     */
    public static final OWLClass PROPOSEHYPOTHESISHANDIN;
    /**
     * KT provide epxlanation hand in
     */
    public static final OWLClass PROVIDEEXPLANATIONHANDIN;
    /**
     * KT reflect on evidence hand in
     */
    public static final OWLClass REFLECTONEVIDENCEHANDIN;
    /**
     * KT simulation
     */
    public static final OWLClass SIMULATION;
    /**
     * KT good practice explanation
     */
    public static final OWLClass GOODPRACTICEEXPLANATION;
    /**
     * KT guided simulation explanation
     */
    public static final OWLClass GUIDEDSIMULATIONEXPLANATION;
    /**
     * KT interactive video
     */
    public static final OWLClass INTERACTIVEVIDEO;
    /**
     * KT investiation simulation
     */
    public static final OWLClass INVESTIGATIONSIMULATION;
    /**
     * KT mini cbt
     */
    public static final OWLClass MINICBT;
    /**
     * KT step by step good practice
     */
    public static final OWLClass STEPBYSTEPGOODPRACTICE;
    /**
     * KT receptive knowledge
     */
    public static final OWLClass RECEPTIVEKNOWLEDGE;
    /**
     * KT action receptive
     */
    public static final OWLClass ACTIONRECEPTIVE;
    /**
     * KT checklist
     */
    public static final OWLClass CHECKLIST;
    /**
     * KT instructions
     */
    public static final OWLClass INSTRUCTIONS;
    /**
     * KT law
     */
    public static final OWLClass LAW;
    /**
     * KT principle strategy
     */
    public static final OWLClass PRINCIPLESTRATEGY;
    /**
     * KT formulate question principle strategy
     */
    public static final OWLClass FORMULATEQUESTIONPRINCIPLESTRATEGY;
    /**
     * KT plan investigation principle strategy
     */
    public static final OWLClass PLANINVESTIGATIONPRINCIPLESTRATEGY;
    /**
     * KT present evidence principle strategy
     */
    public static final OWLClass PRESENTEVIDENCEPRINCIPLESTRATEGY;
    /**
     * KT rule action
     */
    public static final OWLClass RULEACTION;
    /**
     * KT explanation receptive
     */
    public static final OWLClass EXPLANATIONRECEPTIVE;
    /**
     * KT argument explanation
     */
    public static final OWLClass ARGUMENTEXPLANATION;
    /**
     * KT comment explanation
     */
    public static final OWLClass COMMENTEXPLANATION;
    /**
     * KT definition explanation
     */
    public static final OWLClass DEFINITIONEXPLANATION;
    /**
     * KT description explanation
     */
    public static final OWLClass DESCRIPTIONEXPLANATION;
    /**
     * KT example explanation
     */
    public static final OWLClass EXAMPLEEXPLANATION;
    /**
     * KT explanation example
     */
    public static final OWLClass EXPLANATIONEXAMPLE;
    /**
     * KT further explanation example
     */
    public static final OWLClass FURTHEREXPLANATIONEXAMPLE;
    /**
     * KT hypothesis example
     */
    public static final OWLClass HYPOTHESISEXAMPLE;
    /**
     * KT investigation explanation
     */
    public static final OWLClass INVESTIGATIONEXPLANATION;
    /**
     * KT present evidence example
     */
    public static final OWLClass PRESENTEVIDENCEEXAMPLE;
    /**
     * KT exemplification explanation
     */
    public static final OWLClass EXEMPLIFICATIONEXPLANATION;
    /**
     * KT idea tip explanation
     */
    public static final OWLClass IDEATIPEXPLANATION;
    /**
     * KT interview explanation
     */
    public static final OWLClass INTERVIEWVEXPLANATION;
    /**
     * KT simulation explanation
     */
    public static final OWLClass SIMULATIONEXPLANATION;
    /**
     * KT orientation receptive
     */
    public static final OWLClass ORIENTATIONRECEPTIVE;
    /**
     * KT abstract orientation
     */
    public static final OWLClass ABSTRACTORIENTATION;
    /**
     * KT fact orientation
     */
    public static final OWLClass FACTORIENTATION;
    /**
     * KT history orientation
     */
    public static final OWLClass HISTORYORIENTATION;
    /**
     * KT knowledge map orientation
     */
    public static final OWLClass KNOWLEDGEMAPORIENTATION;
    /**
     * KT log orientation
     */
    public static final OWLClass LOGORIENTATION;
    /**
     * KT method orientation
     */
    public static final OWLClass METHODORIENTATION;
    /**
     * KT news orientation
     */
    public static final OWLClass NEWSORIENTATION;
    /**
     * KT ovweview orientation
     */
    public static final OWLClass OVERVIEWORIENTATION;
    /**
     * KT pique curiosity orientation
     */
    public static final OWLClass PIQUECURIOSITYORIENTATION;
    /**
     * KT question orientation
     */
    public static final OWLClass QUESTIONORIENTATION;
    /**
     * KT scenario orientation
     */
    public static final OWLClass SCENARIOORIENTATION;
    /**
     * KT story scenario
     */
    public static final OWLClass STORYSCENARIO;
    /**
     * KT virtual world scenario
     */
    public static final OWLClass VIRTUALWORLDSCENARIO;
    /**
     * KT source receptive
     */
    public static final OWLClass SOURCERECEPTIVE;
    /**
     * KT address source
     */
    public static final OWLClass ADDRESSSOURCE;
    /**
     * KT archive address
     */
    public static final OWLClass ARCHIVEADDRESS;
    /**
     * KT link list source
     */
    public static final OWLClass LINKLISTSOURCE;
    /**
     * KT literature list source
     */
    public static final OWLClass LITERATURELISTSOURCE;
    private static final Map<OWLClass, Collection<OWLClass>> SUPERMAP;
    private static final Map<OWLClass, Collection<OWLClass>> SUBMAP;
    private static final List<OWLClass> KTLIST;
    private final OWLClass CURRENTTYPE;
    private final OWLIndividual INDIVIDUAL;

    static {
        PrefixManager PREFIX_MNGR = new DefaultPrefixManager(PedagogicalOntology.BASE_IRI);
        COOPERATIVEKNOWLEDGE = OntoMngr.DATAFAC.getOWLClass(":CooperativeKnowledge", PREFIX_MNGR);
        PIQUECURIOSITYCOOPERATION = OntoMngr.DATAFAC.getOWLClass(":PiqueCuriosityCooperation", PREFIX_MNGR);
        PLANNEDCOOPERATION = OntoMngr.DATAFAC.getOWLClass(":PlannedCooperation", PREFIX_MNGR);
        CONFERENCEPLANNED = OntoMngr.DATAFAC.getOWLClass(":ConferencePlanned", PREFIX_MNGR);
        COUNSELLINGPLANNED = OntoMngr.DATAFAC.getOWLClass(":CounsellingPlanned", PREFIX_MNGR);
        HUMANFEEDBACKPLANNED = OntoMngr.DATAFAC.getOWLClass(":HumanFeedbackPlanned", PREFIX_MNGR);
        ROLEPLAYPLANNED = OntoMngr.DATAFAC.getOWLClass(":RoleplayPlanned", PREFIX_MNGR);
        TALKPLANNED = OntoMngr.DATAFAC.getOWLClass(":TalkPlanned", PREFIX_MNGR);
        WORKGROUPPLANNED = OntoMngr.DATAFAC.getOWLClass(":WorkgroupPlanned", PREFIX_MNGR);
        SPONTANEOUSCOOPERATION = OntoMngr.DATAFAC.getOWLClass(":SpontaneousCooperation", PREFIX_MNGR);
        COMMUNITYCOMMUNICATIONSPONANTEOUS = OntoMngr.DATAFAC.getOWLClass(":CommunityCommunicationSpontaneous", PREFIX_MNGR);
        COUNSELLINGONLEARNINGMETHODSPONTANEOUS = OntoMngr.DATAFAC.getOWLClass(":CounsellingOnLearningMethodSpontaneous", PREFIX_MNGR);
        QUESTIONTOCOUNSELLORSPONTANEOUS = OntoMngr.DATAFAC.getOWLClass(":QuestionToCounsellorSpontaneous", PREFIX_MNGR);
        INTERACTIVEKNOWLEDGE = OntoMngr.DATAFAC.getOWLClass(":InteractiveKnowledge", PREFIX_MNGR);
        ASSIGNMENTINTERACTIVEAUTOMATICASSESSMENT = OntoMngr.DATAFAC.getOWLClass(":AssignmentInteractiveAutomaticAssessment", PREFIX_MNGR);
        ARRANGINGASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":ArrangingAssignment", PREFIX_MNGR);
        ARTICULATIONASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":ArticulationAssignment", PREFIX_MNGR);
        DETECTINGASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":DetectingAssignment", PREFIX_MNGR);
        DISCRIMINATIONASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":DiscriminationAssignment", PREFIX_MNGR);
        EDUCATIONALGAMEASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":EducationalGameAssignment", PREFIX_MNGR);
        FILLINTHEBLANKASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":FillInTheBlankAssignment", PREFIX_MNGR);
        MAPPINGASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":MappingAssignment", PREFIX_MNGR);
        REPETITIONASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":RepetitionAssignment", PREFIX_MNGR);
        DELAYEDREPEATING = OntoMngr.DATAFAC.getOWLClass(":DelayedRepeating", PREFIX_MNGR);
        DRILLANDPRACTICEREPEATING = OntoMngr.DATAFAC.getOWLClass(":DrillAndPracticeRepeating", PREFIX_MNGR);
        SEQUENCINGASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":SequencingAssignment", PREFIX_MNGR);
        BUILDSENTENCESEQUENCING = OntoMngr.DATAFAC.getOWLClass(":BuildSentenceSequencing", PREFIX_MNGR);
        TECHNICALCONSTRUCTIONSEQUENCING = OntoMngr.DATAFAC.getOWLClass(":TechnicalConstructionSequencing", PREFIX_MNGR);
        SHORTANSWERASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":ShortAnswerAssignment", PREFIX_MNGR);
        SIMULATIONASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":SimulationAssignment", PREFIX_MNGR);
        SPELLINGASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":SpellingAssignment", PREFIX_MNGR);
        TICKINGASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":TickingAssignment", PREFIX_MNGR);
        MULTIPLECHOICEASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":MultipleChoiceAssignment", PREFIX_MNGR);
        SINGLECHOICEASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":SingleChoiceAssignment", PREFIX_MNGR);
        CHOOSEMETHODSINGLECHOICE = OntoMngr.DATAFAC.getOWLClass(":ChooseMethodSingleChoice", PREFIX_MNGR);
        CHOOSEQUESTIONSINGLECHOICE = OntoMngr.DATAFAC.getOWLClass(":ChooseQuestionSingleChoice", PREFIX_MNGR);
        TRUEFALSEASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":TrueFalseAssignment", PREFIX_MNGR);
        ASSIGNMENTINTERACTIVENOASSESSMENT = OntoMngr.DATAFAC.getOWLClass(":AssignmentInteractiveNoAssessment", PREFIX_MNGR);
        ASSIGNMENTINTERACTIVESELFASSESSMENT = OntoMngr.DATAFAC.getOWLClass(":AssignmentInteractiveSelfAssessment", PREFIX_MNGR);
        HANDINASSIGNMENT = OntoMngr.DATAFAC.getOWLClass(":HandInAssignment", PREFIX_MNGR);
        FORMULATEQUESTIONHANDIN = OntoMngr.DATAFAC.getOWLClass(":FormulateQuestionHandIn", PREFIX_MNGR);
        GATHEREVIDENCEHANDIN = OntoMngr.DATAFAC.getOWLClass(":GatherEvidenceHandIn", PREFIX_MNGR);
        PLANINVESTIGATIONHANDIN = OntoMngr.DATAFAC.getOWLClass(":PlanInvestigationHandIn", PREFIX_MNGR);
        PRESENTEVIDENCEHANDIN = OntoMngr.DATAFAC.getOWLClass(":PresentEvidenceHandIn", PREFIX_MNGR);
        PROPOSEHYPOTHESISHANDIN = OntoMngr.DATAFAC.getOWLClass(":ProposeHypothesisHandIn", PREFIX_MNGR);
        PROVIDEEXPLANATIONHANDIN = OntoMngr.DATAFAC.getOWLClass(":ProvideExplanationHandIn", PREFIX_MNGR);
        REFLECTONEVIDENCEHANDIN = OntoMngr.DATAFAC.getOWLClass(":ReflectOnEvidenceHandIn", PREFIX_MNGR);
        SIMULATION = OntoMngr.DATAFAC.getOWLClass(":Simulation", PREFIX_MNGR);
        GOODPRACTICEEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":GoodPracticeExplanation", PREFIX_MNGR);
        GUIDEDSIMULATIONEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":GuidedSimulationExplanation", PREFIX_MNGR);
        INTERACTIVEVIDEO = OntoMngr.DATAFAC.getOWLClass(":InteractiveVideo", PREFIX_MNGR);
        INVESTIGATIONSIMULATION = OntoMngr.DATAFAC.getOWLClass(":InvestigationSimulation", PREFIX_MNGR);
        MINICBT = OntoMngr.DATAFAC.getOWLClass(":MiniCBT", PREFIX_MNGR);
        STEPBYSTEPGOODPRACTICE = OntoMngr.DATAFAC.getOWLClass(":StepByStepGoodPractice", PREFIX_MNGR);
        RECEPTIVEKNOWLEDGE = OntoMngr.DATAFAC.getOWLClass(":ReceptiveKnowledge", PREFIX_MNGR);
        ACTIONRECEPTIVE = OntoMngr.DATAFAC.getOWLClass(":ActionReceptive", PREFIX_MNGR);
        CHECKLIST = OntoMngr.DATAFAC.getOWLClass(":Checklist", PREFIX_MNGR);
        INSTRUCTIONS = OntoMngr.DATAFAC.getOWLClass(":Instructions", PREFIX_MNGR);
        LAW = OntoMngr.DATAFAC.getOWLClass(":Law", PREFIX_MNGR);
        PRINCIPLESTRATEGY = OntoMngr.DATAFAC.getOWLClass(":PrincipleStrategy", PREFIX_MNGR);
        FORMULATEQUESTIONPRINCIPLESTRATEGY = OntoMngr.DATAFAC.getOWLClass(":FormulateQuestionPrincipleStrategy", PREFIX_MNGR);
        PLANINVESTIGATIONPRINCIPLESTRATEGY = OntoMngr.DATAFAC.getOWLClass(":PlanInvestigationPrincipleStrategy", PREFIX_MNGR);
        PRESENTEVIDENCEPRINCIPLESTRATEGY = OntoMngr.DATAFAC.getOWLClass(":PresentEvidencePrincipleStrategy", PREFIX_MNGR);
        RULEACTION = OntoMngr.DATAFAC.getOWLClass(":RuleAction", PREFIX_MNGR);
        EXPLANATIONRECEPTIVE = OntoMngr.DATAFAC.getOWLClass(":ExplanationReceptive", PREFIX_MNGR);
        ARGUMENTEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":ArgumentExplanation", PREFIX_MNGR);
        COMMENTEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":CommentExplanation", PREFIX_MNGR);
        DEFINITIONEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":DefinitionExplanation", PREFIX_MNGR);
        DESCRIPTIONEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":DescriptionExplanation", PREFIX_MNGR);
        EXAMPLEEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":ExampleExplanation", PREFIX_MNGR);
        EXPLANATIONEXAMPLE = OntoMngr.DATAFAC.getOWLClass(":ExplanationExample", PREFIX_MNGR);
        FURTHEREXPLANATIONEXAMPLE = OntoMngr.DATAFAC.getOWLClass(":FurtherExplanationExample", PREFIX_MNGR);
        HYPOTHESISEXAMPLE = OntoMngr.DATAFAC.getOWLClass(":HypothesisExample", PREFIX_MNGR);
        INVESTIGATIONEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":InvestigationExample", PREFIX_MNGR);
        PRESENTEVIDENCEEXAMPLE = OntoMngr.DATAFAC.getOWLClass(":PresentEvidenceExample", PREFIX_MNGR);
        EXEMPLIFICATIONEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":ExemplificationExplanation", PREFIX_MNGR);
        IDEATIPEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":IdeaTipExplanation", PREFIX_MNGR);
        INTERVIEWVEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":InterviewExplanation", PREFIX_MNGR);
        SIMULATIONEXPLANATION = OntoMngr.DATAFAC.getOWLClass(":SimulationExplanation", PREFIX_MNGR);
        ORIENTATIONRECEPTIVE = OntoMngr.DATAFAC.getOWLClass(":OrientationReceptive", PREFIX_MNGR);
        ABSTRACTORIENTATION = OntoMngr.DATAFAC.getOWLClass(":AbstractOrientation", PREFIX_MNGR);
        FACTORIENTATION = OntoMngr.DATAFAC.getOWLClass(":FactOrientation", PREFIX_MNGR);
        HISTORYORIENTATION = OntoMngr.DATAFAC.getOWLClass(":HistoryOrientation", PREFIX_MNGR);
        KNOWLEDGEMAPORIENTATION = OntoMngr.DATAFAC.getOWLClass(":KnowledgeMapOrientation", PREFIX_MNGR);
        LOGORIENTATION = OntoMngr.DATAFAC.getOWLClass(":LogOrientation", PREFIX_MNGR);
        METHODORIENTATION = OntoMngr.DATAFAC.getOWLClass(":MethodOrientation", PREFIX_MNGR);
        NEWSORIENTATION = OntoMngr.DATAFAC.getOWLClass(":NewsOrientation", PREFIX_MNGR);
        OVERVIEWORIENTATION = OntoMngr.DATAFAC.getOWLClass(":OverviewOrientation", PREFIX_MNGR);
        PIQUECURIOSITYORIENTATION = OntoMngr.DATAFAC.getOWLClass(":PiqueCuriosityOrientation", PREFIX_MNGR);
        QUESTIONORIENTATION = OntoMngr.DATAFAC.getOWLClass(":QuestionOrientation", PREFIX_MNGR);
        SCENARIOORIENTATION = OntoMngr.DATAFAC.getOWLClass(":ScenarioOrientation", PREFIX_MNGR);
        STORYSCENARIO = OntoMngr.DATAFAC.getOWLClass(":StoryScenario", PREFIX_MNGR);
        VIRTUALWORLDSCENARIO = OntoMngr.DATAFAC.getOWLClass(":VirtualWorldScenario", PREFIX_MNGR);
        SOURCERECEPTIVE = OntoMngr.DATAFAC.getOWLClass(":SourceReceptive", PREFIX_MNGR);
        ADDRESSSOURCE = OntoMngr.DATAFAC.getOWLClass(":AddressSource", PREFIX_MNGR);
        ARCHIVEADDRESS = OntoMngr.DATAFAC.getOWLClass(":ArchiveAddress", PREFIX_MNGR);
        LINKLISTSOURCE = OntoMngr.DATAFAC.getOWLClass(":LinkListSource", PREFIX_MNGR);
        LITERATURELISTSOURCE = OntoMngr.DATAFAC.getOWLClass(":LiteratureListSource", PREFIX_MNGR);

        KTLIST = new ArrayList<>();
        KTLIST.add(COOPERATIVEKNOWLEDGE);
        KTLIST.add(PIQUECURIOSITYCOOPERATION);
        KTLIST.add(PLANNEDCOOPERATION);
        KTLIST.add(CONFERENCEPLANNED);
        KTLIST.add(COUNSELLINGPLANNED);
        KTLIST.add(HUMANFEEDBACKPLANNED);
        KTLIST.add(ROLEPLAYPLANNED);
        KTLIST.add(TALKPLANNED);
        KTLIST.add(WORKGROUPPLANNED);
        KTLIST.add(SPONTANEOUSCOOPERATION);
        KTLIST.add(COMMUNITYCOMMUNICATIONSPONANTEOUS);
        KTLIST.add(COUNSELLINGONLEARNINGMETHODSPONTANEOUS);
        KTLIST.add(QUESTIONTOCOUNSELLORSPONTANEOUS);
        KTLIST.add(INTERACTIVEKNOWLEDGE);
        KTLIST.add(ASSIGNMENTINTERACTIVEAUTOMATICASSESSMENT);
        KTLIST.add(ARRANGINGASSIGNMENT);
        KTLIST.add(ARTICULATIONASSIGNMENT);
        KTLIST.add(DETECTINGASSIGNMENT);
        KTLIST.add(DISCRIMINATIONASSIGNMENT);
        KTLIST.add(EDUCATIONALGAMEASSIGNMENT);
        KTLIST.add(FILLINTHEBLANKASSIGNMENT);
        KTLIST.add(MAPPINGASSIGNMENT);
        KTLIST.add(REPETITIONASSIGNMENT);
        KTLIST.add(DELAYEDREPEATING);
        KTLIST.add(DRILLANDPRACTICEREPEATING);
        KTLIST.add(SEQUENCINGASSIGNMENT);
        KTLIST.add(BUILDSENTENCESEQUENCING);
        KTLIST.add(TECHNICALCONSTRUCTIONSEQUENCING);
        KTLIST.add(SHORTANSWERASSIGNMENT);
        KTLIST.add(SIMULATIONASSIGNMENT);
        KTLIST.add(SPELLINGASSIGNMENT);
        KTLIST.add(TICKINGASSIGNMENT);
        KTLIST.add(MULTIPLECHOICEASSIGNMENT);
        KTLIST.add(SINGLECHOICEASSIGNMENT);
        KTLIST.add(CHOOSEMETHODSINGLECHOICE);
        KTLIST.add(CHOOSEQUESTIONSINGLECHOICE);
        KTLIST.add(TRUEFALSEASSIGNMENT);
        KTLIST.add(ASSIGNMENTINTERACTIVENOASSESSMENT);
        KTLIST.add(ASSIGNMENTINTERACTIVESELFASSESSMENT);
        KTLIST.add(HANDINASSIGNMENT);
        KTLIST.add(FORMULATEQUESTIONHANDIN);
        KTLIST.add(GATHEREVIDENCEHANDIN);
        KTLIST.add(PLANINVESTIGATIONHANDIN);
        KTLIST.add(PRESENTEVIDENCEHANDIN);
        KTLIST.add(PROPOSEHYPOTHESISHANDIN);
        KTLIST.add(PROVIDEEXPLANATIONHANDIN);
        KTLIST.add(REFLECTONEVIDENCEHANDIN);
        KTLIST.add(SIMULATION);
        KTLIST.add(GOODPRACTICEEXPLANATION);
        KTLIST.add(GUIDEDSIMULATIONEXPLANATION);
        KTLIST.add(INTERACTIVEVIDEO);
        KTLIST.add(INVESTIGATIONSIMULATION);
        KTLIST.add(MINICBT);
        KTLIST.add(STEPBYSTEPGOODPRACTICE);
        KTLIST.add(RECEPTIVEKNOWLEDGE);
        KTLIST.add(ACTIONRECEPTIVE);
        KTLIST.add(CHECKLIST);
        KTLIST.add(INSTRUCTIONS);
        KTLIST.add(LAW);
        KTLIST.add(PRINCIPLESTRATEGY);
        KTLIST.add(FORMULATEQUESTIONPRINCIPLESTRATEGY);
        KTLIST.add(PLANINVESTIGATIONPRINCIPLESTRATEGY);
        KTLIST.add(PRESENTEVIDENCEPRINCIPLESTRATEGY);
        KTLIST.add(RULEACTION);
        KTLIST.add(EXPLANATIONRECEPTIVE);
        KTLIST.add(ARGUMENTEXPLANATION);
        KTLIST.add(COMMENTEXPLANATION);
        KTLIST.add(DEFINITIONEXPLANATION);
        KTLIST.add(DESCRIPTIONEXPLANATION);
        KTLIST.add(EXAMPLEEXPLANATION);
        KTLIST.add(EXPLANATIONEXAMPLE);
        KTLIST.add(FURTHEREXPLANATIONEXAMPLE);
        KTLIST.add(HYPOTHESISEXAMPLE);
        KTLIST.add(INVESTIGATIONEXPLANATION);
        KTLIST.add(PRESENTEVIDENCEEXAMPLE);
        KTLIST.add(EXEMPLIFICATIONEXPLANATION);
        KTLIST.add(IDEATIPEXPLANATION);
        KTLIST.add(INTERVIEWVEXPLANATION);
        KTLIST.add(SIMULATIONEXPLANATION);
        KTLIST.add(ORIENTATIONRECEPTIVE);
        KTLIST.add(ABSTRACTORIENTATION);
        KTLIST.add(FACTORIENTATION);
        KTLIST.add(HISTORYORIENTATION);
        KTLIST.add(KNOWLEDGEMAPORIENTATION);
        KTLIST.add(LOGORIENTATION);
        KTLIST.add(METHODORIENTATION);
        KTLIST.add(NEWSORIENTATION);
        KTLIST.add(OVERVIEWORIENTATION);
        KTLIST.add(PIQUECURIOSITYORIENTATION);
        KTLIST.add(QUESTIONORIENTATION);
        KTLIST.add(SCENARIOORIENTATION);
        KTLIST.add(STORYSCENARIO);
        KTLIST.add(VIRTUALWORLDSCENARIO);
        KTLIST.add(SOURCERECEPTIVE);
        KTLIST.add(ADDRESSSOURCE);
        KTLIST.add(ARCHIVEADDRESS);
        KTLIST.add(LINKLISTSOURCE);
        KTLIST.add(LITERATURELISTSOURCE);

        OWLClass topClass = OntoMngr.DATAFAC.getOWLClass(":KnowledgeType", PREFIX_MNGR);
        SUPERMAP = new HashMap<>();
        SUBMAP = new HashMap<>();
        try {
            for (OWLClass owlClass : KTLIST) {
                Validate.isTrue(OntoMngr.PO.getOntoObj().containsClassInSignature(owlClass.getIRI()),
                        String.format("Specified KT is missing in PO: %s", owlClass.getIRI().toString()));
                SUPERMAP.put(owlClass, getSuperClasses(owlClass, OntoMngr.PO.getOntoObj(), topClass));
                SUBMAP.put(owlClass, getSubClasses(owlClass, OntoMngr.PO.getOntoObj(), topClass));
            }

            for (Collection<OWLClass> values : SUPERMAP.values()) {
                for (OWLClass value : values) {
                    if (!SUPERMAP.keySet().contains(value)) {
                        logger.warn(String.format("Detected missing KT type: %s", value.getIRI().toString()));
                    }
                }
            }

            for (Collection<OWLClass> values : SUBMAP.values()) {
                for (OWLClass value : values) {
                    if (!SUBMAP.keySet().contains(value)) {
                        logger.warn(String.format("Detected missing KT type: %s", value.getIRI().toString()));
                    }
                }
            }
        } catch (Exception ex) {
            logger.warn("Error on KnowledgeType class init.", ex);
        }
    }

    /**
     * Private constructor
     *
     * @param type OWLClass identifying the KT
     * @param individual The KT individual
     */
    protected KnowledgeType(OWLClass type, OWLIndividual individual) {
        this.CURRENTTYPE = type;
        this.INDIVIDUAL = individual;
    }

    /**
     * Creates a new instance of this class using the given individual.
     *
     * @param individual KT individiual in scope
     * @return Null in case this is an unknown type.
     */
    public static KnowledgeType createInstance(OWLIndividual individual) {
        for (OWLClassExpression owlClassExpression : individual.getTypes(OntoMngr.PO.getOntoObj())) {
            if (KTLIST.contains(owlClassExpression.asOWLClass())) {
                return new KnowledgeType(owlClassExpression.asOWLClass(), individual);
            }
        }
        return null;
    }

    /**
     * Determins the super classes of the given owl class.
     *
     * @param owlClass OWL class in scope
     * @param onto Ontology to inspect
     * @return List of super classes.
     */
    private static List<OWLClass> getSuperClasses(OWLClass owlClass, OWLOntology onto, OWLClass topClass) {
        List<OWLClass> superClasses = new ArrayList<>();
        for (OWLClassExpression ce : owlClass.getSuperClasses(onto)) {
            if (!ce.asOWLClass().equals(topClass)) {
                superClasses.add(ce.asOWLClass());
            }
        }
        return superClasses;
    }

    /**
     * Determins the sub classes of the given owl class.
     *
     * @param owlClass OWL class in scope
     * @param onto Ontology to inspect
     * @return List of sub classes.
     */
    private static List<OWLClass> getSubClasses(OWLClass owlClass, OWLOntology onto, OWLClass topClass) {
        List<OWLClass> subClasses = new ArrayList<>();
        for (OWLClassExpression ce : owlClass.getSubClasses(onto)) {
            if (!ce.asOWLClass().equals(topClass)) {
                subClasses.add(ce.asOWLClass());
            }
        }
        return subClasses;
    }

    /**
     * Returns the KT class object.
     *
     * @return Can never be null.
     */
    public OWLClass getType() {
        return CURRENTTYPE;
    }

    /**
     * Returns the super type of this KT.
     *
     * @param onlyDirectSuperTypes Only considers direct super types
     * @return Empty collection if there are no super types for this KT.
     */
    public Collection<OWLClass> getSuperTypes(boolean onlyDirectSuperTypes) {
        return getTypes(onlyDirectSuperTypes, CURRENTTYPE, SUPERMAP);
    }

    /**
     * Returns the sub type of this KT.
     *
     * @param onlyDirectSubTypes Only considers direct sub types
     * @return Empty collection if there are no ssub types for this KT.
     */
    public Collection<OWLClass> getSubTypes(boolean onlyDirectSubTypes) {
        return getTypes(onlyDirectSubTypes, CURRENTTYPE, SUBMAP);
    }

    /**
     * Retrieves the connected types for the {@code typeInScope}.
     *
     * @param onlyDirectTypes Only considers direct relations, i.e. not recursive
     * @param typeInScope The type currently in scope
     * @param map Map to be inspected
     * @return Empty Collection if nothing was found.
     */
    private Collection<OWLClass> getTypes(boolean onlyDirectTypes, OWLClass typeInScope, Map<OWLClass, Collection<OWLClass>> map) {
        Set<OWLClass> values = new HashSet<>();
        values.addAll(map.get(typeInScope));
        if (!onlyDirectTypes) {
            for (OWLClass nextTypeInScope : map.get(typeInScope)) {
                values.addAll(getTypes(onlyDirectTypes, nextTypeInScope, map));
            }
        }
        return values;
    }

    /**
     * Checks whether this KT is categorized as an automatically assessed test.
     *
     * @return True if test.
     */
    public boolean isAutomaticallyAssessedTest() {
        Collection<OWLClass> inspectionCollection = getSuperTypes(false);
        inspectionCollection.add(this.getType());
        return inspectionCollection.contains(ASSIGNMENTINTERACTIVEAUTOMATICASSESSMENT);
    }

    @Override
    public OWLObject getOntoObj() {
        return INDIVIDUAL;
    }

    @Override
    public boolean hasSameIri(String iri) {
        return OwlUtil.getIriAsString(INDIVIDUAL).equals(iri);
    }

    @Override
    public boolean hasSameIri(IRI iri) {
        return OwlUtil.getIriAsString(INDIVIDUAL).equals(OwlUtil.getIriAsString(iri));
    }

    @Override
    public String getIriAsString() {
        return OwlUtil.getIriAsString(INDIVIDUAL);
    }

    @Override
    public IRI getIri() {
        return OwlUtil.getIri(INDIVIDUAL);
    }

    /**
     * @return True if the KT is a (sub-)type of cooperative knowlwedge
     */
    public boolean isCooperative() {
        return this.CURRENTTYPE == COOPERATIVEKNOWLEDGE || getSuperTypes(false).contains(COOPERATIVEKNOWLEDGE);
    }

    /**
     * @return True if the KT is a (sub-)type of interactve knowlwedge
     */
    public boolean isInteractive() {
        return this.CURRENTTYPE == INTERACTIVEKNOWLEDGE || getSuperTypes(false).contains(INTERACTIVEKNOWLEDGE);
    }

    /**
     * @return True if the KT is a (sub-)type of receptive knowlwedge
     */
    public boolean isReceptive() {
        return this.CURRENTTYPE == RECEPTIVEKNOWLEDGE || getSuperTypes(false).contains(RECEPTIVEKNOWLEDGE);
    }
}
