/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-20
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes;

import eu.intuitel.lpmProcessor.OntoMngr;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.IOntologyObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.PedagogicalOntology;
import eu.intuitel.lpmProcessor.utils.OwlUtil;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.apache.commons.lang3.Validate;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLIndividual;
import org.semanticweb.owlapi.model.OWLObject;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.PrefixManager;
import org.semanticweb.owlapi.util.DefaultPrefixManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Class to wrap the Media Types of Knowledge Objects.
 *
 * @author fhe
 */
public class MediaType implements IOntologyObject {

    private static final Logger logger = LoggerFactory.getLogger(KnowledgeType.class);
    /**
     * MT Communication media
     */
    public static final OWLClass COMMUNICATIONMEDIA;
    /**
     * MT Asynchronous communicaton
     */
    public static final OWLClass ASYNCHRONOUSCOMMUNICATION;
    /**
     * MT Blog
     */
    public static final OWLClass BLOGASYNCHRONOUS;
    /**
     * MT Email
     */
    public static final OWLClass EMAILASYNCHRONOUS;
    /**
     * MT Mailinglist
     */
    public static final OWLClass MAILINGLISTASYNCHRONOUS;
    /**
     * MT Newsgroup
     */
    public static final OWLClass NEWSGROUPASYNCHRONOUS;
    /**
     * MT Combined communication
     */
    public static final OWLClass COMBINEDCOMMUNICATION;
    /**
     * MT Social media
     */
    public static final OWLClass SOCIALMEDIACOMBINED;
    /**
     * MT Synchronous communication
     */
    public static final OWLClass SYNCHRONOUSCOMMUNICATION;
    /**
     * MT Audio conference
     */
    public static final OWLClass AUDIOCONFERENCESYNCHRONOUS;
    /**
     * MT Chat
     */
    public static final OWLClass CHATSYNCHRONOUS;
    /**
     * MT Shared application
     */
    public static final OWLClass SHAREDAPPLICATIONSYNCHRONOUS;
    /**
     * MT Video conference
     */
    public static final OWLClass VIDEOCONFERENCESYNCHRONOUS;
    /**
     * MT Interaction media
     */
    public static final OWLClass INTERACTIONMEDIA;
    /**
     * MT Form interaction
     */
    public static final OWLClass FORMINTERACTION;
    /**
     * MT Interactive video
     */
    public static final OWLClass INTERACTIVEVIDEOINTERACTION;
    /**
     * MT Presentation media
     */
    public static final OWLClass PRESENTATIONMEDIA;
    /**
     * MT Animation presentation
     */
    public static final OWLClass ANIMATIONPRESENTATION;
    /**
     * MT Audop presentation
     */
    public static final OWLClass AUDIOPRESENTATION;
    /**
     * MT Music presentation
     */
    public static final OWLClass MUSICAUDIO;
    /**
     * MT Sound
     */
    public static final OWLClass SOUNDAUDIO;
    /**
     * MT Speech
     */
    public static final OWLClass SPEECHAUDIO;
    /**
     * MT Drawing presentation
     */
    public static final OWLClass DRAWINGPRESENTATION;
    /**
     * MT Photo presentation
     */
    public static final OWLClass PHOTOPRESENTATION;
    /**
     * MT Table presentation
     */
    public static final OWLClass TABLEPRESENTATION;
    /**
     * MT Text presentation
     */
    public static final OWLClass TEXTPRESENTATION;
    /**
     * MT Video presentation
     */
    public static final OWLClass VIDEOPRESENTATION;
    private static final Map<OWLClass, Collection<OWLClass>> SUPERMAP;
    private static final Map<OWLClass, Collection<OWLClass>> SUBMAP;
    private static final List<OWLClass> MTLIST;
    private final OWLClass TYPE;
    private final OWLIndividual INDIVIDUAL;

    static {
        PrefixManager PREFIX_MNGR = new DefaultPrefixManager(PedagogicalOntology.BASE_IRI);
        COMMUNICATIONMEDIA = OntoMngr.DATAFAC.getOWLClass(":CommunicationMedia", PREFIX_MNGR);
        ASYNCHRONOUSCOMMUNICATION = OntoMngr.DATAFAC.getOWLClass(":AsynchronousCommunication", PREFIX_MNGR);
        BLOGASYNCHRONOUS = OntoMngr.DATAFAC.getOWLClass(":BlogAsynchronous", PREFIX_MNGR);
        EMAILASYNCHRONOUS = OntoMngr.DATAFAC.getOWLClass(":EmailAsynchronous", PREFIX_MNGR);
        MAILINGLISTASYNCHRONOUS = OntoMngr.DATAFAC.getOWLClass(":MailingListAsynchronous", PREFIX_MNGR);
        NEWSGROUPASYNCHRONOUS = OntoMngr.DATAFAC.getOWLClass(":NewsgroupAsynchronous", PREFIX_MNGR);
        COMBINEDCOMMUNICATION = OntoMngr.DATAFAC.getOWLClass(":CombinedCommunication", PREFIX_MNGR);
        SOCIALMEDIACOMBINED = OntoMngr.DATAFAC.getOWLClass(":SocialMediaCombined", PREFIX_MNGR);
        SYNCHRONOUSCOMMUNICATION = OntoMngr.DATAFAC.getOWLClass(":SynchronousCommunication", PREFIX_MNGR);
        AUDIOCONFERENCESYNCHRONOUS = OntoMngr.DATAFAC.getOWLClass(":AudioConferenceSynchronous", PREFIX_MNGR);
        CHATSYNCHRONOUS = OntoMngr.DATAFAC.getOWLClass(":ChatSynchronous", PREFIX_MNGR);
        SHAREDAPPLICATIONSYNCHRONOUS = OntoMngr.DATAFAC.getOWLClass(":SharedApplicationSynchronous", PREFIX_MNGR);
        VIDEOCONFERENCESYNCHRONOUS = OntoMngr.DATAFAC.getOWLClass(":VideoConferenceSynchronous", PREFIX_MNGR);
        INTERACTIONMEDIA = OntoMngr.DATAFAC.getOWLClass(":InteractionMedia", PREFIX_MNGR);
        FORMINTERACTION = OntoMngr.DATAFAC.getOWLClass(":FormInteraction", PREFIX_MNGR);
        INTERACTIVEVIDEOINTERACTION = OntoMngr.DATAFAC.getOWLClass(":InteractiveVideoInteraction", PREFIX_MNGR);
        PRESENTATIONMEDIA = OntoMngr.DATAFAC.getOWLClass(":PresentationMedia", PREFIX_MNGR);
        ANIMATIONPRESENTATION = OntoMngr.DATAFAC.getOWLClass(":AnimationPresentation", PREFIX_MNGR);
        AUDIOPRESENTATION = OntoMngr.DATAFAC.getOWLClass(":AudioPresentation", PREFIX_MNGR);
        MUSICAUDIO = OntoMngr.DATAFAC.getOWLClass(":MusicAudio", PREFIX_MNGR);
        SOUNDAUDIO = OntoMngr.DATAFAC.getOWLClass(":SoundAudio", PREFIX_MNGR);
        SPEECHAUDIO = OntoMngr.DATAFAC.getOWLClass(":SpeechAudio", PREFIX_MNGR);
        DRAWINGPRESENTATION = OntoMngr.DATAFAC.getOWLClass(":DrawingPresentation", PREFIX_MNGR);
        PHOTOPRESENTATION = OntoMngr.DATAFAC.getOWLClass(":PhotoPresentation", PREFIX_MNGR);
        TABLEPRESENTATION = OntoMngr.DATAFAC.getOWLClass(":TablePresentation", PREFIX_MNGR);
        TEXTPRESENTATION = OntoMngr.DATAFAC.getOWLClass(":TextPresentation", PREFIX_MNGR);
        VIDEOPRESENTATION = OntoMngr.DATAFAC.getOWLClass(":VideoPresentation", PREFIX_MNGR);

        MTLIST = new ArrayList<>();
        MTLIST.add(COMMUNICATIONMEDIA);
        MTLIST.add(ASYNCHRONOUSCOMMUNICATION);
        MTLIST.add(BLOGASYNCHRONOUS);
        MTLIST.add(EMAILASYNCHRONOUS);
        MTLIST.add(MAILINGLISTASYNCHRONOUS);
        MTLIST.add(NEWSGROUPASYNCHRONOUS);
        MTLIST.add(COMBINEDCOMMUNICATION);
        MTLIST.add(SOCIALMEDIACOMBINED);
        MTLIST.add(SYNCHRONOUSCOMMUNICATION);
        MTLIST.add(AUDIOCONFERENCESYNCHRONOUS);
        MTLIST.add(CHATSYNCHRONOUS);
        MTLIST.add(SHAREDAPPLICATIONSYNCHRONOUS);
        MTLIST.add(VIDEOCONFERENCESYNCHRONOUS);
        MTLIST.add(INTERACTIONMEDIA);
        MTLIST.add(FORMINTERACTION);
        MTLIST.add(INTERACTIVEVIDEOINTERACTION);
        MTLIST.add(PRESENTATIONMEDIA);
        MTLIST.add(ANIMATIONPRESENTATION);
        MTLIST.add(AUDIOPRESENTATION);
        MTLIST.add(MUSICAUDIO);
        MTLIST.add(SOUNDAUDIO);
        MTLIST.add(SPEECHAUDIO);
        MTLIST.add(DRAWINGPRESENTATION);
        MTLIST.add(PHOTOPRESENTATION);
        MTLIST.add(TABLEPRESENTATION);
        MTLIST.add(TEXTPRESENTATION);
        MTLIST.add(VIDEOPRESENTATION);

        OWLClass topClass = OntoMngr.DATAFAC.getOWLClass(":MediaType", PREFIX_MNGR);
        SUPERMAP = new HashMap<>();
        SUBMAP = new HashMap<>();
        try {
            for (OWLClass owlClass : MTLIST) {
                Validate.isTrue(OntoMngr.PO.getOntoObj().containsClassInSignature(owlClass.getIRI()));
                SUPERMAP.put(owlClass, getSuperClasses(owlClass, OntoMngr.PO.getOntoObj(), topClass));
                SUBMAP.put(owlClass, getSubClasses(owlClass, OntoMngr.PO.getOntoObj(), topClass));
            }

            for (Collection<OWLClass> values : SUPERMAP.values()) {
                for (OWLClass value : values) {
                    if (!SUPERMAP.keySet().contains(value)) {
                        logger.warn(String.format("Detected missing MT type: %s", value.getIRI().toString()));
                    }
                }
            }

            for (Collection<OWLClass> values : SUBMAP.values()) {
                for (OWLClass value : values) {
                    if (!SUBMAP.keySet().contains(value)) {
                        logger.warn(String.format("Detected missing MT type: %s", value.getIRI().toString()));
                    }
                }
            }
        } catch (Exception ex) {
            logger.warn("Error on MediaType class init.", ex);
        }
    }

    /**
     * Private constructor
     *
     * @param type OWLClass identifying the MT
     */
    protected MediaType(OWLClass type, OWLIndividual individual) {
        this.TYPE = type;
        this.INDIVIDUAL = individual;
    }

    /**
     * Creates a new instance of this class using the given individual.
     *
     * @param individual MT individiual in scope
     * @return Null in case this is an unknown type.
     */
    public static MediaType createInstance(OWLIndividual individual) {
        for (OWLClassExpression owlClassExpression : individual.getTypes(OntoMngr.PO.getOntoObj())) {
            if (MTLIST.contains(owlClassExpression.asOWLClass())) {
                return new MediaType(owlClassExpression.asOWLClass(), individual);
            }
        }
        return null;
    }

    /**
     * Determins the super classes of the given owl class.
     *
     * @param owlClass OWL class in scope
     * @param onto Ontology to inspect
     * @return List of super classes.
     */
    private static List<OWLClass> getSuperClasses(OWLClass owlClass, OWLOntology onto, OWLClass topClass) {
        List<OWLClass> superClasses = new ArrayList<>();
        for (OWLClassExpression ce : owlClass.getSuperClasses(onto)) {
            if (!ce.asOWLClass().equals(topClass)) {
                superClasses.add(ce.asOWLClass());
            }
        }
        return superClasses;
    }

    /**
     * Determins the sub classes of the given owl class.
     *
     * @param owlClass OWL class in scope
     * @param onto Ontology to inspect
     * @return List of sub classes.
     */
    private static List<OWLClass> getSubClasses(OWLClass owlClass, OWLOntology onto, OWLClass topClass) {
        List<OWLClass> subClasses = new ArrayList<>();
        for (OWLClassExpression ce : owlClass.getSubClasses(onto)) {
            if (!ce.asOWLClass().equals(topClass)) {
                subClasses.add(ce.asOWLClass());
            }
        }
        return subClasses;
    }

    /**
     * Returns the MT class object.
     *
     * @return Can never be null.
     */
    public OWLClass getType() {
        return TYPE;
    }

    /**
     * Returns the super type of this MT.
     *
     * @param onlyDirectSuperTypes Only considers direct super types
     * @return Empty collection if there are no super types for this MT.
     */
    public Collection<OWLClass> getSuperTypes(boolean onlyDirectSuperTypes) {
        return getTypes(onlyDirectSuperTypes, TYPE, SUPERMAP);
    }

    /**
     * Returns the sub type of this MT.
     *
     * @param onlyDirectSubTypes Only considers direct sub types
     * @return Empty collection if there are no ssub types for this MT.
     */
    public Collection<OWLClass> getSubTypes(boolean onlyDirectSubTypes) {
        return getTypes(onlyDirectSubTypes, TYPE, SUBMAP);
    }

    /**
     * Retrieves the connected types for the {@code typeInScope}.
     *
     * @param onlyDirectTypes Only considers direct relations, i.e. not recursive
     * @param typeInScope The type currently in scope
     * @param map Map to be inspected
     * @return Empty Collection if nothing was found.
     */
    private Collection<OWLClass> getTypes(boolean onlyDirectTypes, OWLClass typeInScope, Map<OWLClass, Collection<OWLClass>> map) {
        Set<OWLClass> values = new HashSet<>();
        values.addAll(map.get(typeInScope));
        if (!onlyDirectTypes) {
            for (OWLClass nextTypeInScope : map.get(typeInScope)) {
                values.addAll(getTypes(onlyDirectTypes, nextTypeInScope, map));
            }
        }
        return values;
    }

    @Override
    public OWLObject getOntoObj() {
        return INDIVIDUAL;
    }

    @Override
    public boolean hasSameIri(String iri) {
        return OwlUtil.getIriAsString(INDIVIDUAL).equals(iri);
    }

    @Override
    public boolean hasSameIri(IRI iri) {
        return OwlUtil.getIriAsString(INDIVIDUAL).equals(OwlUtil.getIriAsString(iri));
    }

    @Override
    public String getIriAsString() {
        return OwlUtil.getIriAsString(INDIVIDUAL);
    }

    @Override
    public IRI getIri() {
        return OwlUtil.getIri(INDIVIDUAL);
    }

    /**
     * @return True if the MT is a (sub-)type of communication MT
     */
    public boolean isCommunication() {
        return this.TYPE == COMMUNICATIONMEDIA || getSuperTypes(false).contains(COMMUNICATIONMEDIA);
    }

    /**
     * @return True if the MT is a (sub-)type of interaction MT
     */
    public boolean isInteraction() {
        return this.TYPE == INTERACTIONMEDIA || getSuperTypes(false).contains(INTERACTIONMEDIA);
    }

    /**
     * @return True if the MT is a (sub-)type of presentation MT
     */
    public boolean isPresentation() {
        return this.TYPE == PRESENTATIONMEDIA || getSuperTypes(false).contains(PRESENTATIONMEDIA);
    }
}
