/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.owlWrapper.ontologies;

import eu.intuitel.lpmProcessor.utils.OwlUtil;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.io.StringDocumentSource;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Class to store Learner-State-Ontologies.
 *
 * @author fhe
 */
public class LearnerStateOntology extends Ontology {

    private static final Logger logger = LoggerFactory.getLogger(LearnerStateOntology.class);
    private final Map<String, String> prefixes;

    /**
     * Creates a new Learner-State-Ontology object.
     *
     * @param lsoRaw OWL representation of the LSO
     * @param prefixes Map containing prefixes and the URLs they point to
     */
    public LearnerStateOntology(OWLOntology lsoRaw, Map<String, String> prefixes) {
        super(lsoRaw);
        if (prefixes == null) {
            this.prefixes = new HashMap<>();
        } else {
            this.prefixes = prefixes;
        }
    }

    /**
     * Private constructor for LSO creation from deserialization.
     *
     * @param lsoRaw Loaded ontology.
     */
    private LearnerStateOntology(OWLOntology lsoRaw) {
        super(lsoRaw);
        this.prefixes = new HashMap<>();
    }

    /**
     * Returns a String representation of this LSO.
     *
     * @return XML serialization of this LSO
     * @throws java.lang.Exception
     */
    public String serialize() throws Exception {
        return toXml();
    }

    /**
     * De-Serializes the LSO which is given as String.
     *
     * @param ontologyString Ontology to de-serialize
     * @return Null on exception
     */
    public static LearnerStateOntology deserialize(String ontologyString) {
        try {
            OWLOntology onto = OwlUtil.fromXml(ontologyString);
            return new LearnerStateOntology(onto);
        } catch (OWLOntologyCreationException ex) {
            logger.error("Exception on deserializing LearnerStateOntology.", ex);
            return null;
        }
    }

    /**
     * Converts the LSO to its XML representation and returns it as String.
     *
     * @return Null on error.
     * @throws org.semanticweb.owlapi.model.OWLOntologyCreationException
     */
    public String toXml() throws OWLOntologyCreationException, Exception {
        //XML needs to be trimmed so that the <?xml... is the start of the string
        String out = OwlUtil.toXml(this.getOntoObj(), prefixes).trim();
        //Replaces all unnecessary comments and empty parts
        Matcher matcher = Pattern.compile("\\<\\!--.*?-\\>", Pattern.DOTALL).matcher(out);
        out = matcher.replaceAll("");
        matcher = Pattern.compile(">\\s*<").matcher(out);
        out = matcher.replaceAll(">\n<");

        //For testing purposes, check if the created LSO is valid
        OWLOntologyManager mngr = OWLManager.createOWLOntologyManager();
        OWLOntology onto = mngr.loadOntologyFromOntologyDocument(new StringDocumentSource(out));
        if (onto == null) {
            throw new Exception("LSO creation failed at optimization.");
        }

        return out;
    }

    /**
     * Converts the LSO to its Turtle representation and returns it as String.
     *
     * @return Null on error.
     */
    public String toTurtle() {
        return OwlUtil.toTurtle(this.getOntoObj(), prefixes).trim();
    }
}
