/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.owlWrapper.ontologies;

import eu.intuitel.lpmProcessor.OntoMngr;
import eu.intuitel.lpmProcessor.didacticFactors.DidacticFactor;
import java.util.Collections;
import java.util.Set;
import org.apache.commons.lang3.Validate;
import org.semanticweb.owlapi.model.OWLAnnotationProperty;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLObjectProperty;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.PrefixManager;
import org.semanticweb.owlapi.util.DefaultPrefixManager;

/**
 * Class to store LM Ontology related information.
 *
 * @author fhe
 */
public class LearningModelOntology extends Ontology {

    public static final String BASE_IRI = "http://www.intuitel.eu/public/intui_LM.owl#";
    private static final PrefixManager PREFIX_MNGR;
    // <editor-fold defaultstate="collapsed" desc="IRI abbreviations">
    private static final String DIDACTIC_FACTOR_ABBREV_IRI = ":DidacticFactor";
    private static final String LEARNER_STATE_CLASS_ABBREV_IRI = ":LearnerState";
    private static final String CURRENT_LP_ABBREV_IRI = ":CurrentLearningPathway";
    private static final String PREVIOUS_LO_ABBREV_IRI = ":PreviousLo";
    private static final String CURRENT_LO_ABBREV_IRI = ":CurrentLo";
    private static final String COMPLETE_LO_ABBREV_IRI = ":CompleteLo";
    private static final String PARTIALLY_COMPLETE_LO_ABBREV_IRI = ":PartiallyCompleteLo";
    private static final String UNSEEN_LO_ABBREV_IRI = ":UnseenLo";
    private static final String HAS_DF_VALUE_ABBREV_IRI = ":hasDidacticFactorValue";
    private static final String HAS_RULE_ABBREV_IRI = ":hasDeterminingRule";
    // </editor-fold>
    // <editor-fold defaultstate="collapsed" desc="Ontology objects">
    public static final OWLClass DIDACTIC_FACTOR;
    public static final OWLClass LEARNER_STATE_CLASS;
    public static final OWLClass CURRENT_LP;
    public static final OWLClass PREVIOUS_LO;
    public static final OWLClass CURRENT_LO;
    public static final OWLClass COMPLETE_LO;
    public static final OWLClass PARTIALLY_COMPLETE_LO;
    public static final OWLClass UNSEEN_LO;
    public static final OWLObjectProperty HAS_DF_VALUE;
    public static final OWLAnnotationProperty HAS_RULE;
    // </editor-fold>
    private final Set<DidacticFactor> dfs;
    private final String version;

    static {
        PREFIX_MNGR = new DefaultPrefixManager(BASE_IRI);
        Validate.notNull(PREFIX_MNGR);
        DIDACTIC_FACTOR = OntoMngr.DATAFAC.getOWLClass(DIDACTIC_FACTOR_ABBREV_IRI, PREFIX_MNGR);
        Validate.notNull(DIDACTIC_FACTOR);
        LEARNER_STATE_CLASS = OntoMngr.DATAFAC.getOWLClass(LEARNER_STATE_CLASS_ABBREV_IRI, PREFIX_MNGR);
        Validate.notNull(LEARNER_STATE_CLASS);
        CURRENT_LP = OntoMngr.DATAFAC.getOWLClass(CURRENT_LP_ABBREV_IRI, PREFIX_MNGR);
        Validate.notNull(CURRENT_LP);
        PREVIOUS_LO = OntoMngr.DATAFAC.getOWLClass(PREVIOUS_LO_ABBREV_IRI, PREFIX_MNGR);
        Validate.notNull(PREVIOUS_LO);
        CURRENT_LO = OntoMngr.DATAFAC.getOWLClass(CURRENT_LO_ABBREV_IRI, PREFIX_MNGR);
        Validate.notNull(CURRENT_LO);
        COMPLETE_LO = OntoMngr.DATAFAC.getOWLClass(COMPLETE_LO_ABBREV_IRI, PREFIX_MNGR);
        Validate.notNull(COMPLETE_LO);
        PARTIALLY_COMPLETE_LO = OntoMngr.DATAFAC.getOWLClass(PARTIALLY_COMPLETE_LO_ABBREV_IRI, PREFIX_MNGR);
        Validate.notNull(PARTIALLY_COMPLETE_LO);
        UNSEEN_LO = OntoMngr.DATAFAC.getOWLClass(UNSEEN_LO_ABBREV_IRI, PREFIX_MNGR);
        Validate.notNull(UNSEEN_LO);
        HAS_DF_VALUE = OntoMngr.DATAFAC.getOWLObjectProperty(HAS_DF_VALUE_ABBREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_DF_VALUE);
        HAS_RULE = OntoMngr.DATAFAC.getOWLAnnotationProperty(HAS_RULE_ABBREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_RULE);
    }

    /**
     * Creates a new instance of and LMO with the given Set of {@code dfs}.
     *
     * @param onto Original OWL representation of the LMO
     * @param dfs Set of Didactic Factors as part of the LMO
     * @param version Version number of the LMO
     */
    public LearningModelOntology(OWLOntology onto, Set<DidacticFactor> dfs, String version) {
        super(onto);
        Validate.notNull(dfs);
        Validate.notEmpty(dfs);
        this.version = version;
        this.dfs = dfs;

        Validate.isTrue(onto.containsClassInSignature(DIDACTIC_FACTOR.getIRI()), String.format("LMO does not contain class '%s'", DIDACTIC_FACTOR));
        Validate.isTrue(onto.containsClassInSignature(LEARNER_STATE_CLASS.getIRI()), String.format("LMO does not contain class '%s'", LEARNER_STATE_CLASS));
        Validate.isTrue(onto.containsClassInSignature(CURRENT_LP.getIRI()), String.format("LMO does not contain class '%s'", CURRENT_LP));
        Validate.isTrue(onto.containsClassInSignature(PREVIOUS_LO.getIRI()), String.format("LMO does not contain class '%s'", PREVIOUS_LO));
        Validate.isTrue(onto.containsClassInSignature(CURRENT_LO.getIRI()), String.format("LMO does not contain class '%s'", CURRENT_LO));
        Validate.isTrue(onto.containsClassInSignature(COMPLETE_LO.getIRI()), String.format("LMO does not contain class '%s'", COMPLETE_LO));
        Validate.isTrue(onto.containsClassInSignature(PARTIALLY_COMPLETE_LO.getIRI()), String.format("LMO does not contain class '%s'", PARTIALLY_COMPLETE_LO));
        Validate.isTrue(onto.containsClassInSignature(UNSEEN_LO.getIRI()), String.format("LMO does not contain class '%s'", UNSEEN_LO));
        Validate.isTrue(onto.containsObjectPropertyInSignature(HAS_DF_VALUE.getIRI()), String.format("LMO does not contain object property '%s'", HAS_DF_VALUE));
        Validate.isTrue(onto.containsAnnotationPropertyInSignature(HAS_RULE.getIRI()), String.format("LMO does not contain annotation property '%s'", HAS_RULE));
    }

    /**
     * Returns the Set of Didactic Factors of this LMO.
     *
     * @return The DFs that are part of this LMO
     */
    public Set<DidacticFactor> getDidacticFactors() {
        return Collections.unmodifiableSet(dfs);
    }

    /**
     * Returns the version info of the LMO.
     *
     * @return Null in case version info could not be loaded.
     */
    public String getVersion() {
        return version;
    }
}
