/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.owlWrapper.ontologies;

import eu.intuitel.lpmProcessor.OntoMngr;
import eu.intuitel.lpmProcessor.models.factories.OntologyFactory;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.IOntologyObject;
import eu.intuitel.lpmProcessor.utils.OwlUtil;
import java.io.File;
import java.net.URISyntaxException;
import java.net.URL;
import org.apache.commons.lang3.Validate;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyAlreadyExistsException;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;

/**
 * General wrapper class for ontologies.
 *
 * @author fhe
 */
public class Ontology implements IOntologyObject {

    protected final OWLOntology ontoObj;

    /**
     * Creates a new Ontology object .
     *
     * @param ontoObj Original OWL representation of the ontology
     */
    public Ontology(OWLOntology ontoObj) {
        Validate.notNull(ontoObj);
        this.ontoObj = ontoObj;
    }

    @Override
    public OWLOntology getOntoObj() {
        return ontoObj;
    }

    @Override
    public boolean hasSameIri(String iri) {
        return iri.equalsIgnoreCase(getIriAsString());
    }

    @Override
    public boolean hasSameIri(IRI iri) {
        return hasSameIri(iri.toString());
    }

    @Override
    public IRI getIri() {
        return OwlUtil.getIri(ontoObj);
    }

    @Override
    public String getIriAsString() {
        return OwlUtil.getIriAsString(ontoObj);
    }

    @Override
    public String toString() {
        return OwlUtil.getIri(ontoObj).getFragment();
    }

    /**
     * Creates an Ontology instance from a given URL.
     *
     * @param url URL of the ontology to be loaded.
     * @return Should never be null.
     * @throws OWLOntologyCreationException Thrown on error while loading the ontology via the OWL
     * API
     * @throws URISyntaxException Syntax exception of the given URL
     */
    public static Ontology createFromUrl(URL url) throws OWLOntologyCreationException, URISyntaxException {
        OWLOntology onto;
        try {
            onto = OntoMngr.MANAGER.loadOntology(IRI.create(url));
        } catch (OWLOntologyAlreadyExistsException ex) {
            onto = OntoMngr.MANAGER.getOntology(ex.getOntologyID());
        }
        return OntologyFactory.createNewInstance(onto);
    }

    /**
     * Creates an Ontology instance from a given URL.
     *
     * @param file File of the ontology to be loaded.
     * @return Should never be null.
     * @throws OWLOntologyCreationException Thrown on error while loading the ontology via the OWL
     * API
     */
    public static Ontology createFromFile(File file) throws OWLOntologyCreationException {
        if (null == file) {
            return null;
        }
        if (!file.canRead()) {
            return null;
        }
        OWLOntology onto;
        try {
            onto = OntoMngr.MANAGER.loadOntologyFromOntologyDocument(file);
        } catch (OWLOntologyAlreadyExistsException ex) {
            onto = OntoMngr.MANAGER.getOntology(ex.getOntologyID());
        }
        return OntologyFactory.createNewInstance(onto);
    }
}
