/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-03-04
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.owlWrapper.ontologies;

import eu.intuitel.lpmProcessor.OntoMngr;
import org.apache.commons.lang3.Validate;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLDataProperty;
import org.semanticweb.owlapi.model.OWLObjectProperty;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.PrefixManager;
import org.semanticweb.owlapi.util.DefaultPrefixManager;

/**
 * Implementation of the Pedagogical Ontology.
 *
 * @author fhe
 */
public class PedagogicalOntology extends Ontology {

    public static final String BASE_IRI = "http://www.intuitel.eu/public/intui_PO.owl#";
    private static final PrefixManager PREFIX_MNGR;
    private final String version;
    // <editor-fold defaultstate="collapsed" desc="Classes - IRI abbreviations">
    private static final String KO_ABREV_IRI = ":KnowledgeObject";
    private static final String CC_ABREV_IRI = ":ConceptContainer";
    private static final String LP_ABREV_IRI = ":LearningPathway";
    private static final String MICRO_LP_ABREV_IRI = ":MicroLearningPathway";
    private static final String MACRO_LP_ABREV_IRI = ":MacroLearningPathway";
    private static final String KT_ABREV_IRI = ":KnowledgeType";
    private static final String MT_ABREV_IRI = ":MediaType";
    // </editor-fold>
    // <editor-fold defaultstate="collapsed" desc="Classes - Ontology objects">
    public static final OWLClass KO_CLASS;
    public static final OWLClass CC_CLASS;
    public static final OWLClass LP_CLASS;
    public static final OWLClass MICRO_LP_CLASS;
    public static final OWLClass MACRO_LP_CLASS;
    public static final OWLClass KT_CLASS;
    public static final OWLClass MT_CLASS;
    // </editor-fold>
    // <editor-fold defaultstate="collapsed" desc="LP properties - IRI abbreviations">
    private static final String HAS_MICRO_PREDECESSOR_ABREV_IRI = ":hasPredecessorInMicroLp";
    private static final String HAS_MACRO_PREDECESSOR_ABREV_IRI = ":hasPredecessorInMacroLp";
    private static final String HAS_MICRO_SUCCESSOR_ABREV_IRI = ":hasSuccessorInMicroLp";
    private static final String HAS_MACRO_SUCCESSOR_ABREV_IRI = ":hasSuccessorInMacroLp";
    // </editor-fold>
    // <editor-fold defaultstate="collapsed" desc="LP properties - Ontology objects">
    public static final OWLObjectProperty HAS_MICRO_PREDECESSOR;
    public static final OWLObjectProperty HAS_MACRO_PREDECESSOR;
    public static final OWLObjectProperty HAS_MICRO_SUCCESSOR;
    public static final OWLObjectProperty HAS_MACRO_SUCCESSOR;
    // </editor-fold>
    // <editor-fold defaultstate="collapsed" desc="KO properties - IRI abbreviations">
    private static final String HAS_KT_ABREV_IRI = ":hasKnowledgeType";
    private static final String HAS_MT_ABREV_IRI = ":hasMediaType";
    private static final String HAS_PREFFERED_LOCATION_ABREV_IRI = ":hasPreferredLocation";
    private static final String IS_CONTAINED_BY_CC_ABREV_IRI = ":isContainedByConceptContainer";
    private static final String HAS_EQF_LVL_ABREV_IRI = ":hasEqfLevel";
    private static final String HAS_DIFFICULTY_LVL_ABREV_IRI = ":hasEqfLevel";
    private static final String HAS_MIN_HEIGHT_ABREV_IRI = ":hasMinimalHeight";
    private static final String HAS_MIN_WIDTH_ABREV_IRI = ":hasMinimalWidth";
    private static final String HAS_RECOMMENDED_AGE_ABREV_IRI = ":hasRecommendedAge";
    private static final String HAS_RECOMMENDED_HEIGHT_ABREV_IRI = ":hasRecommendedHeight";
    private static final String HAS_RECOMMENDED_WIDTH_ABREV_IRI = ":hasRecommendedWidth";
    private static final String HAS_SIZE_ABREV_IRI = ":hasSize";
    private static final String HAS_SUBTITLES_ABREV_IRI = ":hasSubtitles";
    private static final String HAS_LANGUAGE_ABREV_IRI = ":hasLanguage";
    private static final String HAS_ESTIMATED_LEARNING_TIME_ABREV_IRI = ":hasEstimatedLearningTime";
    private static final String HAS_PREFERRED_LOCATION_LONGITUDE_ABREV_IRI = ":hasPreferredLocationLongitude";
    private static final String HAS_PREFERRED_LOCATION_LATITUDE_ABREV_IRI = ":hasPreferredLocationLatitude";
    private static final String HAS_PREFERRED_LOCATION_MARGIN_ABREV_IRI = ":hasPreferredLocationMargin";
    private static final String IS_LINKED_WITH_LMS_ELEMENT_ABREV_IRI = ":isLinkedWithLmsElement";
    private static final String IS_SUITABLE_FOR_BLIND_ABREV_IRI = ":isSuitableForBlind";
    private static final String IS_SUITABLE_FOR_DEAF_ABREV_IRI = ":isSuitableForDeaf";
    private static final String IS_SUITABLE_FOR_MUTE_ABREV_IRI = ":isSuitableForMute";
    // </editor-fold>
    // <editor-fold defaultstate="collapsed" desc="KO properties  - Ontology objects">
    public static final OWLObjectProperty HAS_KT;
    public static final OWLObjectProperty HAS_MT;
    public static final OWLObjectProperty HAS_PREFERRED_LOCATION;
    public static final OWLObjectProperty IS_CONTAINED_BY_CC;
    public static final OWLDataProperty HAS_DIFFICULTY_LVL;
    public static final OWLDataProperty HAS_EQF_LVL;
    public static final OWLDataProperty HAS_MIN_HEIGHT;
    public static final OWLDataProperty HAS_MIN_WIDTH;
    public static final OWLDataProperty HAS_RECOMMENDED_AGE;
    public static final OWLDataProperty HAS_RECOMMENDED_HEIGHT;
    public static final OWLDataProperty HAS_RECOMMENDED_WIDTH;
    public static final OWLDataProperty HAS_SIZE;
    public static final OWLDataProperty HAS_SUBTITLES;
    public static final OWLDataProperty HAS_LANGUAGE;
    public static final OWLDataProperty HAS_ESTIMATED_LEARNING_TIME;
    public static final OWLDataProperty HAS_PREFERRED_LOCATION_LONGITUDE;
    public static final OWLDataProperty HAS_PREFERRED_LOCATION_LATITUDE;
    public static final OWLDataProperty HAS_PREFERRED_LOCATION_MARGIN;
    public static final OWLDataProperty IS_LINKED_WITH_LMS_ELEMENT;
    public static final OWLDataProperty IS_SUITABLE_FOR_BLIND;
    public static final OWLDataProperty IS_SUITABLE_FOR_DEAF;
    public static final OWLDataProperty IS_SUITABLE_FOR_MUTE;
    // </editor-fold>

    static {
        PREFIX_MNGR = new DefaultPrefixManager(BASE_IRI);

        KO_CLASS = OntoMngr.DATAFAC.getOWLClass(KO_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(KO_CLASS);
        CC_CLASS = OntoMngr.DATAFAC.getOWLClass(CC_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(CC_CLASS);
        LP_CLASS = OntoMngr.DATAFAC.getOWLClass(LP_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(LP_CLASS);
        MICRO_LP_CLASS = OntoMngr.DATAFAC.getOWLClass(MICRO_LP_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(MICRO_LP_CLASS);
        MACRO_LP_CLASS = OntoMngr.DATAFAC.getOWLClass(MACRO_LP_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(MACRO_LP_CLASS);
        KT_CLASS = OntoMngr.DATAFAC.getOWLClass(KT_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(KT_CLASS);
        MT_CLASS = OntoMngr.DATAFAC.getOWLClass(MT_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(MT_CLASS);

        HAS_MICRO_PREDECESSOR = OntoMngr.DATAFAC.getOWLObjectProperty(HAS_MICRO_PREDECESSOR_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_MICRO_PREDECESSOR);
        HAS_MACRO_PREDECESSOR = OntoMngr.DATAFAC.getOWLObjectProperty(HAS_MACRO_PREDECESSOR_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_MACRO_PREDECESSOR);
        HAS_MICRO_SUCCESSOR = OntoMngr.DATAFAC.getOWLObjectProperty(HAS_MICRO_SUCCESSOR_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_MICRO_SUCCESSOR);
        HAS_MACRO_SUCCESSOR = OntoMngr.DATAFAC.getOWLObjectProperty(HAS_MACRO_SUCCESSOR_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_MACRO_SUCCESSOR);

        HAS_KT = OntoMngr.DATAFAC.getOWLObjectProperty(HAS_KT_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_KT);
        HAS_MT = OntoMngr.DATAFAC.getOWLObjectProperty(HAS_MT_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_MT);
        HAS_PREFERRED_LOCATION = OntoMngr.DATAFAC.getOWLObjectProperty(HAS_PREFFERED_LOCATION_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_PREFERRED_LOCATION);
        IS_CONTAINED_BY_CC = OntoMngr.DATAFAC.getOWLObjectProperty(IS_CONTAINED_BY_CC_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(IS_CONTAINED_BY_CC);

        HAS_DIFFICULTY_LVL = OntoMngr.DATAFAC.getOWLDataProperty(HAS_DIFFICULTY_LVL_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_DIFFICULTY_LVL);
        HAS_EQF_LVL = OntoMngr.DATAFAC.getOWLDataProperty(HAS_EQF_LVL_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_EQF_LVL);
        HAS_MIN_HEIGHT = OntoMngr.DATAFAC.getOWLDataProperty(HAS_MIN_HEIGHT_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_MIN_HEIGHT);
        HAS_MIN_WIDTH = OntoMngr.DATAFAC.getOWLDataProperty(HAS_MIN_WIDTH_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_MIN_WIDTH);
        HAS_RECOMMENDED_AGE = OntoMngr.DATAFAC.getOWLDataProperty(HAS_RECOMMENDED_AGE_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_RECOMMENDED_AGE);
        HAS_RECOMMENDED_HEIGHT = OntoMngr.DATAFAC.getOWLDataProperty(HAS_RECOMMENDED_HEIGHT_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_RECOMMENDED_HEIGHT);
        HAS_RECOMMENDED_WIDTH = OntoMngr.DATAFAC.getOWLDataProperty(HAS_RECOMMENDED_WIDTH_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_RECOMMENDED_WIDTH);
        HAS_SIZE = OntoMngr.DATAFAC.getOWLDataProperty(HAS_SIZE_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_SIZE);
        HAS_SUBTITLES = OntoMngr.DATAFAC.getOWLDataProperty(HAS_SUBTITLES_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_SUBTITLES);
        HAS_LANGUAGE = OntoMngr.DATAFAC.getOWLDataProperty(HAS_LANGUAGE_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_LANGUAGE);
        HAS_ESTIMATED_LEARNING_TIME = OntoMngr.DATAFAC.getOWLDataProperty(HAS_ESTIMATED_LEARNING_TIME_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_ESTIMATED_LEARNING_TIME);
        HAS_PREFERRED_LOCATION_LONGITUDE = OntoMngr.DATAFAC.getOWLDataProperty(HAS_PREFERRED_LOCATION_LONGITUDE_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_PREFERRED_LOCATION_LONGITUDE);
        HAS_PREFERRED_LOCATION_LATITUDE = OntoMngr.DATAFAC.getOWLDataProperty(HAS_PREFERRED_LOCATION_LATITUDE_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_PREFERRED_LOCATION_LATITUDE);
        HAS_PREFERRED_LOCATION_MARGIN = OntoMngr.DATAFAC.getOWLDataProperty(HAS_PREFERRED_LOCATION_MARGIN_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(HAS_PREFERRED_LOCATION_MARGIN);
        IS_LINKED_WITH_LMS_ELEMENT = OntoMngr.DATAFAC.getOWLDataProperty(IS_LINKED_WITH_LMS_ELEMENT_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(IS_LINKED_WITH_LMS_ELEMENT);
        IS_SUITABLE_FOR_BLIND = OntoMngr.DATAFAC.getOWLDataProperty(IS_SUITABLE_FOR_BLIND_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(IS_SUITABLE_FOR_BLIND);
        IS_SUITABLE_FOR_DEAF = OntoMngr.DATAFAC.getOWLDataProperty(IS_SUITABLE_FOR_DEAF_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(IS_SUITABLE_FOR_DEAF);
        IS_SUITABLE_FOR_MUTE = OntoMngr.DATAFAC.getOWLDataProperty(IS_SUITABLE_FOR_MUTE_ABREV_IRI, PREFIX_MNGR);
        Validate.notNull(IS_SUITABLE_FOR_MUTE);
    }

    /**
     * Constructor
     *
     * @param onto OWL representation of the PO
     * @param version Version number of the PO
     */
    public PedagogicalOntology(OWLOntology onto, String version) {
        super(onto);
        this.version = version;

        Validate.isTrue(onto.containsClassInSignature(KO_CLASS.getIRI()),
                String.format("PO does not contain class '%s'", KO_CLASS));
        Validate.isTrue(onto.containsClassInSignature(CC_CLASS.getIRI()),
                String.format("PO does not contain class '%s'", CC_CLASS));
        Validate.isTrue(onto.containsClassInSignature(LP_CLASS.getIRI()),
                String.format("PO does not contain class '%s'", LP_CLASS));
        Validate.isTrue(onto.containsClassInSignature(MICRO_LP_CLASS.getIRI()),
                String.format("PO does not contain class '%s'", MICRO_LP_CLASS));
        Validate.isTrue(onto.containsClassInSignature(MACRO_LP_CLASS.getIRI()),
                String.format("PO does not contain class '%s'", MACRO_LP_CLASS));
        Validate.isTrue(onto.containsClassInSignature(KT_CLASS.getIRI()),
                String.format("PO does not contain class '%s'", KT_CLASS));
        Validate.isTrue(onto.containsClassInSignature(MT_CLASS.getIRI()),
                String.format("PO does not contain class '%s'", MT_CLASS));

        Validate.isTrue(onto.containsObjectPropertyInSignature(HAS_MICRO_PREDECESSOR.getIRI()),
                String.format("PO does not contain object property '%s'", HAS_MICRO_PREDECESSOR));
        Validate.isTrue(onto.containsObjectPropertyInSignature(HAS_MACRO_PREDECESSOR.getIRI()),
                String.format("PO does not contain object property '%s'", HAS_MACRO_PREDECESSOR));
        Validate.isTrue(onto.containsObjectPropertyInSignature(HAS_MICRO_SUCCESSOR.getIRI()),
                String.format("PO does not contain object property '%s'", HAS_MICRO_SUCCESSOR));
        Validate.isTrue(onto.containsObjectPropertyInSignature(HAS_MACRO_SUCCESSOR.getIRI()),
                String.format("PO does not contain object property '%s'", HAS_MACRO_SUCCESSOR));
        Validate.isTrue(onto.containsObjectPropertyInSignature(HAS_KT.getIRI()),
                String.format("PO does not contain object property '%s'", HAS_KT));
        Validate.isTrue(onto.containsObjectPropertyInSignature(HAS_MT.getIRI()),
                String.format("PO does not contain object property '%s'", HAS_MT));
        Validate.isTrue(onto.containsObjectPropertyInSignature(HAS_PREFERRED_LOCATION.getIRI()),
                String.format("PO does not contain object property '%s'", HAS_PREFERRED_LOCATION));
        Validate.isTrue(onto.containsObjectPropertyInSignature(IS_CONTAINED_BY_CC.getIRI()),
                String.format("PO does not contain object property '%s'", IS_CONTAINED_BY_CC));

        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_DIFFICULTY_LVL.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_DIFFICULTY_LVL));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_EQF_LVL.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_EQF_LVL));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_MIN_HEIGHT.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_MIN_HEIGHT));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_MIN_WIDTH.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_MIN_WIDTH));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_RECOMMENDED_AGE.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_RECOMMENDED_AGE));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_RECOMMENDED_HEIGHT.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_RECOMMENDED_HEIGHT));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_RECOMMENDED_WIDTH.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_RECOMMENDED_WIDTH));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_SIZE.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_SIZE));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_SUBTITLES.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_SUBTITLES));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_LANGUAGE.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_LANGUAGE));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_ESTIMATED_LEARNING_TIME.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_ESTIMATED_LEARNING_TIME));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_PREFERRED_LOCATION_LONGITUDE.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_PREFERRED_LOCATION_LONGITUDE));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_PREFERRED_LOCATION_LATITUDE.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_PREFERRED_LOCATION_LATITUDE));
        Validate.isTrue(onto.containsDataPropertyInSignature(HAS_PREFERRED_LOCATION_MARGIN.getIRI()),
                String.format("PO does not contain data property '%s'", HAS_PREFERRED_LOCATION_MARGIN));
        Validate.isTrue(onto.containsDataPropertyInSignature(IS_LINKED_WITH_LMS_ELEMENT.getIRI()),
                String.format("PO does not contain data property '%s'", IS_LINKED_WITH_LMS_ELEMENT));
        Validate.isTrue(onto.containsDataPropertyInSignature(IS_SUITABLE_FOR_BLIND.getIRI()),
                String.format("PO does not contain data property '%s'", IS_SUITABLE_FOR_BLIND));
        Validate.isTrue(onto.containsDataPropertyInSignature(IS_SUITABLE_FOR_DEAF.getIRI()),
                String.format("PO does not contain data property '%s'", IS_SUITABLE_FOR_DEAF));
        Validate.isTrue(onto.containsDataPropertyInSignature(IS_SUITABLE_FOR_MUTE.getIRI()),
                String.format("PO does not contain data property '%s'", IS_SUITABLE_FOR_MUTE));
    }

    /**
     * Returns the version info of the LMO.
     *
     * @return Null in case version info could not be loaded.
     */
    public String getVersion() {
        return version;
    }
}
