/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 04.08.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.reflex;

import eu.intuitel.lpmGeneralInterfaces.IEngineCommunicator.Option;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import org.apache.commons.collections.CollectionUtils;

/**
 * Reflex message being triggered when data about the learner is missing.
 *
 * @author Florian Heberle
 */
public class ReflexDataIncomplete extends Reflex implements ILearnerReflex {

    private static final Map<String, Long> TIMESTAMPS = new HashMap<>();
    private static final Long THRESHOLD = (long) 480000; 
    //For testing:
    //private static final Long THRESHOLD = (long) 1; 

    private ReflexDataIncomplete(String mId, String lId, String cId) {
        super(mId, lId, cId, false);
    }

    /**
     * Creates Reflex messages for incomplete data items for a given learner.
     *
     * @param mId Message ID of the learner update that triggered the reflex
     * @param lId Learner ID who triggered the learner update
     * @param cId Course ID of the slom data
     * @param db DB accessor
     * @return Reflex to be triggered or null.
     */
    public static Reflex getReflexes(String mId, String lId, String cId, DbAccessUtil db) {
        //Only trigger this reflex from time to time.
        long now = new Date().getTime();
        if (!TIMESTAMPS.containsKey(lId)) {
            TIMESTAMPS.put(lId, now);
            return null;
        } else {
            if (now - TIMESTAMPS.get(lId) < THRESHOLD) {
                return null;
            } else {
                TIMESTAMPS.put(lId, now);
            }
        }
        //Check the data in the DB. If items are missing, send reflex. If no items are missing, return null.
        Reflex profileReflex = getProfileReflex(mId, lId, cId, db);
        if (profileReflex != null) {
            return profileReflex;
        }
        Reflex prefReflex = getPreferencesReflex(mId, lId, cId, db);
        if (prefReflex != null) {
            return prefReflex;
        }
        Reflex disabilitiesReflex = getDisabilitiesReflex(mId, lId, cId, db);
        if (disabilitiesReflex != null) {
            return disabilitiesReflex;
        }
        return null;
    }

    private static ReflexDataIncomplete getProfileReflex(String mId, String lId, String cId, DbAccessUtil db) {
        ReflexDataIncomplete reflex = new ReflexDataIncomplete(mId, lId, cId);
        if (null == db.getLearnerAge()) {
            reflex.OPTIONS.add(new Option("age", null, null, null));
        }
        if (null == db.getLearnerCulture()) {
            reflex.OPTIONS.add(new Option("culture", null, null, null));
        }
        if (null == db.getLearnerGender()) {
            reflex.OPTIONS.add(new Option("gender", null, null, null));
        }
        if (CollectionUtils.isEmpty(db.getLearnerLanguages())) {
            reflex.OPTIONS.add(new Option("languages", null, null, null));
        }
        if (reflex.OPTIONS.isEmpty()) {
            return null;
        }
        return reflex;
    }

    private static Reflex getDisabilitiesReflex(String mId, String lId, String cId, DbAccessUtil db) {
        ReflexDataIncomplete reflex = new ReflexDataIncomplete(mId, lId, cId);
        if (null == db.getPreferredDifficultyLevel()) {
            reflex.OPTIONS.add(new Option("preferredDifficultyLevel", null, null, null));
        }
        if (null == db.getPreferredEqfLevel()) {
            reflex.OPTIONS.add(new Option("preferredEqfLevel", null, null, null));
        }
        if (reflex.OPTIONS.isEmpty()) {
            return null;
        }
        return reflex;
    }

    private static Reflex getPreferencesReflex(String mId, String lId, String cId, DbAccessUtil db) {
        ReflexDataIncomplete reflex = new ReflexDataIncomplete(mId, lId, cId);
        if (null == db.isLearnerMute()) {
            reflex.OPTIONS.add(new Option("muteness", null, null, null));
        }
        if (null == db.isLearnerDeaf()) {
            reflex.OPTIONS.add(new Option("deafness", null, null, null));
        }
        if (reflex.OPTIONS.isEmpty()) {
            return null;
        }
        return reflex;
    }
}
