/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 14.07.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import org.apache.commons.collections.CollectionUtils;

/**
 * Util for storing, updating and determining completion values.
 *
 * @author Florian Heberle
 */
public class CompletionUtil {

    /**
     * Updates the completion value of the specified KO.
     *
     * @param ko Knowledge Object to be updated
     * @param accessed KO was accessed
     * @param grade Received grade for the specified KO
     * @param seenPercentage The seen percentage (video/audio)
     * @param textCompletion The read amount of text
     * @param db DB accessor object
     */
    public static void updateKoCompletionValue(KnowledgeObject ko, Boolean accessed, Integer grade, Integer seenPercentage, Integer textCompletion, DbAccessUtil db) {
        updateKoCompletionValue(ko.getIriAsString(), accessed, grade, seenPercentage, textCompletion, db);
    }

    /**
     * Updates the completion value of the specified KO.
     *
     * @param koIri IRI of the Knowledge Object in scope.
     * @param accessed KO was accessed
     * @param grade Received grade for the specified KO
     * @param seenPercentage The seen percentage (video/audio)
     * @param textCompletion The read amount of text
     * @param db DB accessor object
     */
    public static void updateKoCompletionValue(String koIri, Boolean accessed, Integer grade, Integer seenPercentage, Integer textCompletion, DbAccessUtil db) {
        Boolean accessedStored = db.getKoCompletionAccessed(koIri);
        Integer gradeStored = db.getKoCompletionGrade(koIri);
        Integer seenPercentageStored = db.getKoCompletionSeenPercentage(koIri);
        Integer textCompletionStored = db.getKoCompletionTextualCompletion(koIri);
        if (!Objects.equals(accessedStored, accessed)) {
            return;
        }
        if (!Objects.equals(gradeStored, grade)) {
            return;
        }
        if (!Objects.equals(seenPercentageStored, seenPercentage)) {
            return;
        }
        if (!Objects.equals(textCompletionStored, textCompletion)) {
            return;
        }
        db.setKoCompletionStatus(koIri, accessed, grade, seenPercentage, textCompletion);
    }

    /**
     * Updates the completion values of the current/last CCs of the current course.
     *
     * @param ko The KO for which the completion values shall be changed
     * @param db DB accessor object
     * @param slom SLOM data of the current course
     */
    public static void updateCcCompletionValue(KnowledgeObject ko, DbAccessUtil db, SlomMetadata slom) {
        List<KnowledgeObject> completedKos = db.getCompletedKOs();
        List<ConceptContainer> completedCcs = db.getCompletedCCs();
        for (ConceptContainer cc : ko.getCcReferences()) {
            boolean isComplete = isCompletedCc(cc, slom, completedKos);
            boolean isStoredAsComplete = completedCcs.contains(cc);
            if (isComplete != isStoredAsComplete) {
                db.setCcCompletionStatus(cc, isComplete);
            }
        }
    }

    /**
     * Determines whether the given CC is completed.
     * A CC is completed when it contains a mLP with a cohesive chain of completed KOs.
     *
     * @param cc CC to inspect
     * @param slom SLOM data of containing the given CC
     * @param completedKos Collection of completed KOs
     * @return True if completed.
     */
    protected static boolean isCompletedCc(ConceptContainer cc, SlomMetadata slom, Collection<KnowledgeObject> completedKos) {
        Integer value = getCcCompletion(cc, slom, completedKos);
        if (value != null) {
            if (value == 100) {
                return true;
            }
        }
        return false;
    }

    /**
     * Determines the percentage of completion of the specified CC.
     *
     * @param cc The CC to be inspected
     * @param slom The SLOM data of the course
     * @param completedKos KOs the learner has completed
     * @return Null if undeterminable.
     */
    public static Integer getCcCompletion(ConceptContainer cc, SlomMetadata slom,
            Collection<KnowledgeObject> completedKos) {
        List<MicroLearningPathway> mLPsToInspect = getMicroLpsToInspect(cc, slom);
        if (CollectionUtils.isEmpty(mLPsToInspect)) {
            return null;
        }
        List<Integer> completionValues = new ArrayList<>();
        for (MicroLearningPathway microLp : mLPsToInspect) {
            Set<KnowledgeObject> starters = microLp.getStarters();
            starters.retainAll(cc.getReferencedKos(slom));
            for (KnowledgeObject starter : starters) {
                getPercentageCompletionOfRoutes(completionValues, starter, microLp, completedKos, 0, 0);
            }
        }
        if (completionValues.isEmpty()) {
            return null;
        }
        int max = 0;
        for (Integer completionValue : completionValues) {
            if (completionValue != null) {
                if (completionValue > max) {
                    max = completionValue;
                }
            }
        }
        return max;
    }

    /**
     * Retrieves the mLPs to inspect, i.e. the mLPs that are used to link KOs assigned to the
     * specified CC.
     *
     * @param cc CC to inspect
     * @param slom SLOM data of the course
     * @return Empty list if nothing was determined; never null.
     */
    private static List getMicroLpsToInspect(ConceptContainer cc, SlomMetadata slom) {
        List<MicroLearningPathway> lps = new ArrayList<>();
        for (KnowledgeObject ko : cc.getReferencedKos(slom)) {
            for (MicroLearningPathway lp : ko.getLearningPathwayReferences()) {
                if (!lps.contains(lp)) {
                    lps.add(lp);
                }
            }
        }
        return lps;
    }

    /**
     * Recursively determines the completion values of mLP routes.
     *
     * @param completionValues The result produced by this method
     * @param ko KO to inspect
     * @param microLp Micro LP to inspect
     * @param completedKos The KOs the learner has completed
     * @param completionCtr Ctr for completed KOs on the route
     * @param totalCtr Ctr for the length of the route
     */
    private static void getPercentageCompletionOfRoutes(Collection<Integer> completionValues, KnowledgeObject ko, MicroLearningPathway microLp,
            Collection<KnowledgeObject> completedKos, Integer completionCtr, Integer totalCtr) {
        if (completedKos.contains(ko)) {
            completionCtr++;
        }
        totalCtr++;
        if (!ko.hasSuccessorFor(microLp)) {
            completionValues.add(completionCtr * 100 / totalCtr);
        } else {
            for (KnowledgeObject successor : ko.getDirectSucessorsFor(microLp)) {
                getPercentageCompletionOfRoutes(completionValues, successor, microLp, completedKos, completionCtr, totalCtr);
            }
        }
    }

    /**
     * Tests wheter the given LP was already completed by the learner.
     *
     * @param db Communication interface to the user database
     * @param lp Learning Pathway to test
     * @return True if finished
     */
    public static boolean isCompletedLp(DbAccessUtil db, LearningPathway lp) {
        List<? extends LearningObject> completedCcs = lp instanceof MicroLearningPathway ? db.getCompletedKOs() : db.getCompletedCCs();
        return containsStarterOfCompletedChain(lp.getStarters(), completedCcs, lp);
    }

    /**
     * Recursively traverses the given collection of LOs and checks if there is a cohesive chain of
     * completed LOs.
     *
     * @param losToBeInspected Collection of LOs to be inspeted
     * @param completedLos Collection of completed LOs to validate the {@code losToBeInspected}
     * @param lpInScope LP currently in scope
     * @return True if a cohesive chain of completed LOs was found.
     */
    protected static boolean containsStarterOfCompletedChain(Collection<? extends LearningObject> losToBeInspected,
            Collection<? extends LearningObject> completedLos,
            LearningPathway lpInScope) {
        if (losToBeInspected.isEmpty() || completedLos.isEmpty() || lpInScope == null) {
            return false;
        }
        for (LearningObject lo : losToBeInspected) {
            if (completedLos.contains(lo)) {
                if (!lo.hasSuccessorFor(lpInScope)) {
                    return true;
                }
                boolean reachedEnd = containsStarterOfCompletedChain(lo.getDirectSucessorsFor(lpInScope), completedLos, lpInScope);
                if (reachedEnd) {
                    return true;
                }
            }
        }
        return false;
    }
}
