/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-05-12
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.types.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang.Validate;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Utility class to access data from the database and receive SLOM-specific return values.
 *
 * @author fhe
 */
public class DbAccessUtil {

    private static final Logger LOGGER = LoggerFactory.getLogger(DbAccessUtil.class);
    private final DbAccessor db;
    private final String learnerId;
    private final SlomMetadata slom;
    private final String courseId;

    /**
     * Constructor - Creates a new utility object to access the database.
     *
     * @param learnerId user id
     * @param accessor Accessor to use for DB data exchange
     */
    public DbAccessUtil(String learnerId, DbAccessor accessor) {
        Validate.notNull(learnerId);
        Validate.notNull(accessor);
        this.learnerId = learnerId;
        this.db = accessor;
        this.slom = null;
        this.courseId = null;
    }

    /**
     * Constructor - Creates a new utility object to access the database.
     *
     * @param learnerId user id
     * @param accessor Accessor to use for DB data exchange
     * @param slom Slom data of the course
     */
    public DbAccessUtil(String learnerId, DbAccessor accessor, SlomMetadata slom) {
        Validate.notNull(learnerId);
        Validate.notNull(accessor);
        Validate.notNull(slom);
        this.learnerId = learnerId;
        this.db = accessor;
        this.courseId = slom.getCourseId();
        this.slom = slom;
    }

    /**
     * Generates a new AccessUtil on basis of this DbAccessUtil using the provided SLOM data.
     *
     * @param slom SLOM data of the course in scope
     * @return Never null.
     */
    public DbAccessUtil extendWithSlom(SlomMetadata slom) {
        return new DbAccessUtil(learnerId, db, slom);
    }

    /**
     * @return The DB accessor used be this util
     */
    public DbAccessor getDbAccessor() {
        return db;
    }

    // <editor-fold defaultstate="collapsed" desc="Learning history getters">
    /**
     * Returns the completion status of the given CC.
     *
     * @param cc CC in scope
     * @return True if CC was marked as complete
     */
    public Boolean getCompletionStatus(ConceptContainer cc) {
        return db.getCcCompletionStatus(learnerId, cc.getIriAsString());
    }

    /**
     * Returns the completion status of the given KO.
     *
     * @param ko KO in scope
     * @return Integer between 0 and 100.
     */
    public Integer getCompletionStatus(KnowledgeObject ko) {
        return db.getKoCompletionStatus(learnerId, ko.getIriAsString());
    }

    /**
     * Returns the completed CCs of the learner.
     *
     * @return Empty list if no CC was completed yet.
     */
    public List<ConceptContainer> getCompletedCCs() {
        List<ConceptContainer> result = new ArrayList<>();
        for (String iri : db.getCompletedCcs(learnerId, courseId)) {
            ConceptContainer lo = slom.getCcByIri(iri);
            if (lo != null) {
                result.add(lo);
            } else {
                LOGGER.debug("Unknown CC in DB: " + iri);
            }
        }
        return result;
    }

    /**
     * Returns the completed KOs of the learner.
     *
     * @return Empty list if no KO was completed yet.
     */
    public List<KnowledgeObject> getCompletedKOs() {
        return getCompletedKOs(learnerId);
    }

    /**
     * Determines the completed KOs of the specified learner in the current course.
     *
     * @param learnerId Id of the learner in scope. If null, the ID of the learner currently in
     * scope is used.
     * @return Returns an empty List if there are no completed KOs or in case of error.
     */
    public List<KnowledgeObject> getCompletedKOs(String learnerId) {
        String learnerInScope = learnerId;
        if (StringUtils.isEmpty(learnerId)) {
            learnerInScope = learnerId;
        }
        List<String> completedKoIris = db.getCompletedKOs(learnerInScope, courseId);
        if (CollectionUtils.isEmpty(completedKoIris)) {
            return new ArrayList<>();
        }
        List<KnowledgeObject> completedKos = new ArrayList<>();
        for (String completedKoIri : completedKoIris) {
            KnowledgeObject completedKo = slom.getKoByIri(completedKoIri);
            if (completedKo != null) {
                completedKos.add(completedKo);
            }
        }
        return completedKos;
    }

    /**
     * Returns the CC history of the learner.
     *
     * @return Empty list if nothing was accessed yet.
     */
    public List<ConceptContainer> getHistoryOfCCs() {
        List<ConceptContainer> result = new ArrayList<>();
        for (String iri : db.getHistoryOfCCs(learnerId, courseId)) {
            ConceptContainer lo = slom.getCcByIri(iri);
            if (lo != null) {
                result.add(lo);
            } else {
                LOGGER.debug("Unknown CC in DB: " + iri);
            }
        }
        return result;
    }

    /**
     * Returns the KO history of the learner.
     *
     * @return Empty list if nothing was accessed yet.
     */
    public List<KnowledgeObject> getHistoryOfKOs() {
        List<KnowledgeObject> result = new ArrayList<>();
        for (String iri : db.getHistoryOfKOs(learnerId, courseId)) {
            KnowledgeObject lo = slom.getKoByIri(iri);
            if (lo != null) {
                result.add(lo);
            } else {
                LOGGER.debug("Unknown KO in DB: " + iri);
            }
        }
        return result;
    }

    /**
     * Returns the CC history of the learner.
     *
     * @param inspectionSize Number of items to be included
     * @return Empty list if nothing was accessed yet.
     */
    public List<ConceptContainer> getHistoryOfCCs(int inspectionSize) {
        List<ConceptContainer> result = new ArrayList<>();
        for (String iri : db.getHistoryOfCCs(learnerId, courseId, inspectionSize)) {
            ConceptContainer lo = slom.getCcByIri(iri);
            if (lo != null) {
                result.add(lo);
            } else {
                LOGGER.debug("Unknown CC in DB: " + iri);
            }
        }
        return result;
    }

    /**
     * Returns the KO history of the learner.
     *
     * @param inspectionSize Number of items to be included
     * @return Empty list if nothing was accessed yet.
     */
    public List<KnowledgeObject> getHistoryOfKOs(int inspectionSize) {
        List<KnowledgeObject> result = new ArrayList<>();
        for (String iri : db.getHistoryOfKOs(learnerId, courseId, inspectionSize)) {
            KnowledgeObject lo = slom.getKoByIri(iri);
            if (lo != null) {
                result.add(lo);
            } else {
                LOGGER.debug("Unknown KO in DB: " + iri);
            }
        }
        return result;
    }

    /**
     * Returns the LmsId of last KO the learner has accessed.
     *
     * @return Null if nothing was accessed yet.
     */
    public String getLastKnowledgeObjectLmsId() {
        return db.getLastKnowledgeObjectLmsId(learnerId);
    }

    /**
     * Returns the IRI of last KO the learner has accessed.
     *
     * @return Null if nothing was accessed yet.
     */
    public String getLastKnowledgeObjectIri() {
        return db.getLastKnowledgeObjectIri(learnerId);
    }

    /**
     * Returns the last KO the learner has accessed.
     *
     * @return Null if nothing was accessed yet.
     */
    public KnowledgeObject getLastKnowledgeObject() {
        if (StringUtils.isBlank(courseId) || slom == null) {
            return null;
        }
        return slom.getKoByIri(db.getLastKnowledgeObjectIri(learnerId, courseId));
    }

    /**
     * Returns the last CC the learner has accessed.
     *
     * @return Null if nothing was accessed yet.
     */
    public ConceptContainer getLastConceptContainer() {
        if (StringUtils.isBlank(courseId) || slom == null) {
            return null;
        }
        return slom.getCcByIri(db.getLastConceptContainerIri(learnerId));
    }

    /**
     * Returns the last recommendations given to the learner.
     *
     * @return List containing Lists of the past {@code stepsCount} recommendations given to the
     * learner. May never return null, but (nested) empty lists.
     */
    public List<List<KnowledgeObject>> getLastRecommendations() {
        List<List<KnowledgeObject>> result = new ArrayList<>();
        List<List<Pair<String, Double>>> recommendations = db.getLearningRecommendations(learnerId, courseId, Integer.MAX_VALUE);
        for (List<Pair<String, Double>> recommendedKos : recommendations) {
            List<KnowledgeObject> innerResult = new ArrayList<>();
            for (Pair<String, Double> recommendedKo : recommendedKos) {
                KnowledgeObject ko = slom.getKoByIri(recommendedKo.getLeft());
                if (ko == null) {
                    LOGGER.error("Error retrieving past recommendation: " + recommendedKo);
                    return Collections.emptyList();
                }
                innerResult.add(ko);
            }
            result.add(innerResult);
        }
        return result;
    }

    /**
     * Returns the learning time the learner required to complete the specified KO.
     *
     * @param ko Knowledge Object in scope.
     * @return The amount of millisends the learner needed to complete the KO. Null in case of error
     * or if there is no learning time for the specified KO.
     */
    public Long getLearningTime(KnowledgeObject ko) {
        return getLearningTime(learnerId, ko);
    }

    /**
     * Determines the learning time of the specified KO.
     *
     * @param learnerId Id of the learner in scope. If null, the ID of the learner currently in
     * scope is used.
     * @param ko KO in scope
     * @return Null in case of error or if there is no learning time for the specified KO.
     */
    public Long getLearningTime(String learnerId, KnowledgeObject ko) {
        String learnerInScope = learnerId;
        if (StringUtils.isEmpty(learnerId)) {
            learnerInScope = learnerId;
        }
        if (ko == null) {
            return null;
        }
        return db.getLearningTime(learnerInScope, ko.getIriAsString());
    }

    /**
     * @param koIri IRI of the Knowledge Object in scope.
     * @return True if the learner has accessed the KO, false if not, null on error.
     */
    public Boolean getKoCompletionAccessed(String koIri) {
        return db.getKoCompletionAccessed(learnerId, koIri);
    }

    /**
     * Returns grade the learner has received in the specified KO.
     *
     * @param koIri IRI of the Knowledge Object in scope.
     * @return Grade received by the learner. Null if unknown.
     */
    public Integer getKoCompletionGrade(String koIri) {
        return db.getKoCompletionGrade(learnerId, koIri);
    }

    /**
     * Returns the percentage value of how much the learner has seen from specified KO
     * (audio/video).
     *
     * @param koIri IRI of the Knowledge Object in scope.
     * @return Value 0-100, null in unknown.
     */
    public Integer getKoCompletionSeenPercentage(String koIri) {
        return db.getKoCompletionSeenPercentage(learnerId, koIri);
    }

    /**
     * Returns the percentage value of how much the learner has read of the specified KO.
     *
     * @param koIri IRI of the Knowledge Object in scope.
     * @return Value 0-100, null in unknown.
     */
    public Integer getKoCompletionTextualCompletion(String koIri) {
        return db.getKoCompletionTextualCompletion(learnerId, koIri);
    }

    /**
     * Retrieves all session durations of the learner.
     *
     * @return Returns an empty collection if nothing is available or in case of error.
     */
    public Collection<Long> getSessionLengths() {
        return db.getSessionDurations(learnerId, courseId);
    }

    /**
     * Retrieves the current duration of the session.
     *
     * @return Null if unavailable or on error.
     */
    public Long getCurrentSessionLength() {
        Collection<Long> durations = getSessionLengths();
        return CollectionUtils.isEmpty(durations) ? null : durations.iterator().next();
    }
    // </editor-fold>

    //<editor-fold defaultstate="collapsed" desc="Learning history setters">
    /**
     * Adds a new entry in the learning history for the learner.
     *
     * @param currentKo KO currently accessed
     * @param currentCc CC assigned to the currently accessed KO
     * @param time Time of the access
     * @return True if adding the entry was successfull.
     */
    public boolean addLearningHistoryEntry(KnowledgeObject currentKo, ConceptContainer currentCc, long time) {
        if (!db.isLoggedIn(learnerId)) {
            db.loginUser(learnerId, courseId);
        }
        return db.storeLearningStep(learnerId,
                courseId,
                currentKo.getIriAsString(),
                currentCc.getIriAsString(),
                time);
    }

    /**
     * Marks the end of the current learning session for the learner.
     */
    public void setSessionEnd() {
        db.logoutUser(learnerId);
    }

    /**
     * Sets the completion status (grade) for the given KO.
     *
     * @param koIRI IRI of the KO
     * @param accessed
     * @param grade The received grade (1-6)
     * @param seenPercentage
     * @param textCompletion
     * @return True if setting the value was successfull.
     */
    public boolean setKoCompletionStatus(String koIRI, Boolean accessed, Integer grade, Integer seenPercentage, Integer textCompletion) {
        return db.storeKoCompletion(learnerId, courseId, koIRI, accessed, grade, seenPercentage, textCompletion);
    }

    /**
     * Sets the completion status (grade) for the given KO.
     *
     * @param koIRI IRI of the KO
     * @param grade The received grade (1-6)
     * @return True if setting the value was successfull.
     */
    public boolean setKoCompletionStatusGrade(String koIRI, int grade) {
        return db.storeKoCompletionGrade(learnerId, courseId, koIRI, grade);
    }

    /**
     * Sets the completion status for the given KO (how much was accessed by the learner).
     *
     * @param koIri IRI of the KO
     * @param completion Completion value of the KO
     * @return True if setting the value was successfull.
     */
    public boolean setKoCompletionStatusCompletion(String koIri, int completion) {
        return db.storeKoCompletionTextualCompletion(learnerId, courseId, koIri, completion);
    }

    /**
     * Sets the completion status of the given KO (% of seen/heard video or audio of the KO).
     *
     * @param koIri IRI of the KO
     * @param seenPercentage the seen percentage
     * @return True if setting the value was successfull.
     */
    public boolean setKoCompletionStatusSeenPercentage(String koIri, int seenPercentage) {
        return db.storeKoCompletionSeenPercentage(learnerId, courseId, koIri, seenPercentage);
    }

    /**
     * Sets the whether a given KO was accessed by the learner or not.
     *
     * @param koIri IRI of the KO
     * @param accessed True if the KO was accessed
     * @return True if setting the value was successful.
     */
    public boolean setKoCompletionStatusAccessed(String koIri, boolean accessed) {
        return db.storeKoCompletionAccessed(learnerId, courseId, koIri, accessed);
    }

    /**
     * Sets the completion value of the given CC.
     *
     * @param cc CC in scope
     * @param complete Value denoting whether the CC is complete (true) or not (false)
     * @return True if setting the value was successful.
     */
    public boolean setCcCompletionStatus(ConceptContainer cc, boolean complete) {
        return db.storeCcCompletion(learnerId, courseId, cc.getIriAsString(), complete);
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="User profile setters">
    /**
     * Stores profile related information for a learner. Uses and parses the String values as
     * obtained via USE.
     *
     * @param lpData The profile data to be stored
     * @return True if successfull.
     */
    public boolean setLearnerProfileData(UserScoreExtractionUtil.LearnerProfileData lpData) {
        return db.storeLearner(learnerId,
                lpData.name,
                lpData.gender,
                lpData.age,
                lpData.culture,
                lpData.deafness,
                lpData.blindness,
                lpData.muteness,
                lpData.languages);
    }

    /**
     * Stores the session related data in the database. Uses the String values as obtained from USE.
     *
     * @param sData Sessiondata to be stored.
     * @return True if successfull.
     */
    public boolean setSessionData(UserScoreExtractionUtil.SessionData sData) {
        return db.storeStaticSessionData(learnerId,
                sData.deviceType,
                sData.deviceDpi,
                sData.deviceTime,
                sData.width,
                sData.height,
                sData.willingness,
                sData.attitude);
    }

    /**
     * Stores the current environmental data of the learning situation. Uses values as obtained via
     * USE.
     *
     * @param envData EnvData to be stored.
     * @return True if successfull.
     */
    public boolean setEnvironmentData(UserScoreExtractionUtil.EnvData envData) {
        return db.storeTimedependentSessionData(learnerId,
                envData.battery,
                envData.conType,
                envData.conStab,
                envData.noise);
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="User profile getters">
    /**
     * Returns the last known battery level of the learner.
     *
     * @return Double value betweem 0 and 100. Default is 100.
     */
    public Integer getBatteryLevel() {
        return db.getBatteryLevel(learnerId);
    }

    /**
     * Returns the age of the learner.
     *
     * @return If age is unkown, -1 is returned.
     */
    public Integer getLearnerAge() {
        return db.getLearnerAge(learnerId);
    }

    /**
     * Checks whether the learner is blind.
     *
     * @return True if learner is blind.
     */
    public Boolean isLearnerBlind() {
        return db.isLearnerBlind(learnerId);
    }

    /**
     * Checks whether the learner is deaf.
     *
     * @return True if the learner is deaf.
     */
    public Boolean isLearnerDeaf() {
        return db.isLearnerDeaf(learnerId);
    }

    /**
     * Checks whether the learner is mute.
     *
     * @return True if the learner is mute.
     */
    public Boolean isLearnerMute() {
        return db.isLearnerMute(learnerId);
    }

    /**
     * Returns the horizontal resolution of the learner's access device.
     *
     * @return Horizontal pixel count.
     */
    public Integer getHorizontalDeviceResolution() {
        return db.getDeviceHorizontalResolution(learnerId);
    }

    /**
     * Returns the vertical resolution of the learner's access device.
     *
     * @return Vertical pixel count.
     */
    public Integer getVerticalDeviceResolution() {
        return db.getDeviceVerticalResolution(learnerId);
    }

    /**
     * Returns the gender of the learner.
     *
     * @return Gender of the learner as defined by the given enum. May never be null.
     */
    public Gender getLearnerGender() {
        return db.getLearnerGender(learnerId);
    }

    /**
     * Returns the current willingness of the learner to interact with peers.
     *
     * @return True if learner is willing, False if learner is unwilling, Null if value is unknown.
     */
    public InteractionWillingness getInteractionWillingness() {
        return db.getInteractionWillingness(learnerId);
    }

    /**
     * Returns the type of the internet connection the learner is using to access the LMS.
     *
     * @return Connection type as defined by the given enum. May never be null.
     */
    public ConnectionType getConnectionType() {
        ConnectionType value = db.getConnectionType(learnerId);
        return value == null ? ConnectionType.Unknown : value;
    }

    /**
     * Returns an Integer value indicating the stability of the learner's network connection.
     *
     * @return Value between 0 and 100, Null if unknown.
     */
    public Integer getConnectionStability() {
        return db.getConnectionStability(learnerId);
    }

    /**
     * Returns an Integer value indicating the preferred difficulty level of the learner.
     *
     * @return Difficulty level as defined by the given enum. May never be null.
     */
    public DifficultyLevel getPreferredDifficultyLevel() {
        return db.getPreferredDifficultyLevel(learnerId, courseId);
    }

    /**
     * Returns an Integer value indiciating the preferred EQF level of the learner.
     *
     * @return Inter value between 1 and 7, Null=Unknown
     */
    public EqfLevel getPreferredEqfLevel() {
        return db.getPreferredEqfLevel(learnerId, courseId);
    }

    /**
     * Returns the motivation of the learner as Boolean.
     *
     * @return LearningAttitude.unknown if no specific value is stored.
     */
    public LearningAttitude getLearningAttitude() {
        return db.getLearningAttitude(learnerId);
    }

    /**
     * Returns the noise level in which the learner is currently working in decibel.
     *
     * @return Decibel value, Null if unknown.
     */
    public Integer getNoise() {
        return db.getNoiseLevel(learnerId);
    }

    /**
     * Get the timestamp of the last session start.
     *
     * @return Timestamp as long.
     */
    public Long getLastLoginTime() {
        return db.getPreviousLoginTime(learnerId, courseId);
    }

    /**
     * Get the timestamp of the current session start.
     *
     * @return Timestamp as long.
     */
    public Long getCurrentLoginTime() {
        return db.getLoginTime(learnerId, courseId);
    }

    /**
     * Gets the languages spoken by the learner.
     *
     * @return Language as defined by the given enum. May never be null. May be null if nothing was
     * specified.
     */
    public Collection<Language> getLearnerLanguages() {
        return db.getLearnerLanguages(learnerId);
    }

    /**
     * Gets the culture of the learner.
     *
     * @return Null in case nothing is specified.
     */
    public Culture getLearnerCulture() {
        return db.getLearnerCulture(learnerId);
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="Learning pathway getters">
    /**
     * Returns the currently selected macro LP of the learner.
     *
     * @return Null if no MLP has yet been selected.
     */
    public MacroLearningPathway getMacroLpOfLearner() {
        String lpURI = db.getMacroLpOfLearner(learnerId, courseId);
        return slom.getMacroLp(lpURI);
    }

    /**
     * Returns the currently selected micro LP of the learner.
     *
     * @return Null if no mLP has yet been selected.
     */
    public MicroLearningPathway getMicroLpOfLearner() {
        String lpURI = db.getMicroLpOfLearner(learnerId, courseId);
        return slom.getMicroLp(lpURI);
    }

    /**
     * Returns the percentage value of learners of the course who selected the given LP.
     *
     * @param lp LP in scope
     * @return Integer value between 0 and 100.
     */
    public Integer getPopularityOfLp(LearningPathway lp) {
        if (lp instanceof MicroLearningPathway) {
            return db.getPopularityOfMicroLp(lp.getIriAsString(), courseId);
        } else if (lp instanceof MacroLearningPathway) {
            return db.getPopularityOfMacroLp(lp.getIriAsString(), courseId);
        } else {
            return 0;
        }
    }

    /**
     * Returns the number of milliseconds since the last time the LPs have been evaluated for the
     * specified learner.
     *
     * @return Maxium Long value if no last eval incident available.
     */
    public Long getTimeSinceLastLpEvaluation() {
        return db.getTimeSinceLastLpEvaluation(learnerId, courseId);
    }

    /**
     * Returns the number of learning steps since the last time the LPs have been evaluated for the
     * specified learner.
     *
     * @return Maximum Integer value if no last eval incident available.
     */
    public Integer getStepsSinceLastLpEvaluation() {
        return db.getStepsSinceLastLpEvaluation(learnerId, courseId);
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="Learning pathway setters">
    /**
     * Sets now as the last LP evaluation incident for the given learner .
     */
    public void setLpEvaluationIncident() {
        db.storeLpEvaluationIncident(learnerId, courseId);
    }

    /**
     * Sets the current macro LP for the learner.
     *
     * @param currentMacroLp Macro LP to be set
     */
    public void setMacroLearningPathway(MacroLearningPathway currentMacroLp) {
        db.storeMacroLpSelection(learnerId, currentMacroLp.getIriAsString(), courseId);
    }

    /**
     * Sets the current micro LP for the learner.
     *
     * @param currentMicroLp Micro LP to be set
     */
    public void setMicroLearningPathway(MicroLearningPathway currentMicroLp) {
        db.storeMicroLpSelection(learnerId, currentMicroLp.getIriAsString(), courseId);
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="Peer data">    
    /**
     * Retrieves all session durations of the learner's peers.
     *
     * @return Returns an empty collection if nothing is available or in case of error.
     */
    public Collection<Collection<Long>> getPeerSessionLengths() {
        return db.getPeerSessionLengths(learnerId, courseId);
    }

    /**
     * Determines the learnerIds of the peers in the course.
     *
     * @return If there are no peers or in case of error, an empty collection will be returned.
     */
    public Collection<String> getPeerIdsInCourse() {
        return db.getPeerIds(learnerId, courseId);
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="SLOM data">    
    /**
     * @param iri IRI of the SLOM element
     * @param title Title of the element
     * @param description Description of the element
     * @param lmsId ID of the KO in the LMS
     * @return True if successfull, otherwise false.
     */
    public boolean setKoData(String iri, String title, String description, String lmsId) {
        return db.storeKoData(iri, courseId, title, description, lmsId);
    }

    /**
     * @param iri IRI of the SLOM element
     * @param title Title of the element
     * @param description Description of the element
     * @return True if successfull, otherwise false.
     */
    public boolean setCcData(String iri, String title, String description) {
        return db.storeCcData(iri, courseId, title, description);
    }

    /**
     * @param iri IRI of the SLOM element
     * @param title Title of the element
     * @param description Description of the element
     * @return True if successfull, otherwise false.
     */
    public boolean setMicroLpData(String iri, String title, String description) {
        return db.storeMicroLpData(iri, courseId, title, description);
    }

    /**
     * @param iri IRI of the SLOM element
     * @param title Title of the element
     * @param description Description of the element
     * @return True if successfull, otherwise false.
     */
    public boolean setMacroLpData(String iri, String title, String description) {
        return db.storeMacroLpData(iri, courseId, title, description);
    }
    // </editor-fold>

    /**
     * Parses the given String and tries to convert it to Integer.
     *
     * @param string String to be parsed
     * @return Null in case of error.
     */
    public static Integer saveParseInt(String string) {
        try {
            return Integer.parseInt(string);
        } catch (NumberFormatException ex) {
            return null;
        }
    }

    /**
     * Parses the given String and tries to convert it to Integer.
     *
     * @param string String to be parsed
     * @return Null in case of error.
     */
    public static Long saveParseLong(String string) {
        try {
            return Long.parseLong(string);
        } catch (NumberFormatException ex) {
            return null;
        }
    }

    /**
     * @return List containing the the past recommendations (Pair.left) and the selection
     * (Pair.right) the learner made. The returned list ignores eventually existing empty learner
     * updates, e.g. when a reflex aborted the recommendation process. <br>
     * The returned List is never null, but could be empty. The contained pair might contain null
     * values.
     */
    public List<Pair<List<KnowledgeObject>, KnowledgeObject>> getRecomSelections() {
        List<Pair<List<String>, String>> recomSelections = db.getRecommendationSelections(learnerId, courseId);
        List<Pair<List<KnowledgeObject>, KnowledgeObject>> result = new ArrayList<>();
        for (Pair<List<String>, String> recomSelection : recomSelections) {
            List<String> recoms = recomSelection.getLeft();
            List<KnowledgeObject> recomKos = new ArrayList<>();
            for (String recom : recoms) {
                recomKos.add(slom.getKoByIri(recom));
            }
            KnowledgeObject selectionKo = slom.getKoByIri(recomSelection.getRight());
            result.add(Pair.of(recomKos, selectionKo));
        }
        return result;
    }
    
    /**
     * Tests whether the learner with the given ID is already registered in the system.
     * @return 
     */
    public boolean isLearnerRegistered() {
        return db.isLearnerRegistered(learnerId);
    }
    
    /**
     * Creates an empty learner profile for the learner.
     */
    public void setLearnerProfile() {
        db.storeLearner(learnerId, null, Gender.Unknown, null, Culture.Unknown, null, null, null, new Language[0]);
    }
    
    /**
     * Creates an empty learner preferences entry for the learner.
     */
    public void setLearnerPreferences() {
        db.storeLearnerPreference(learnerId, courseId, EqfLevel.Unknown, DifficultyLevel.Unknown);
    }
}
