/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.lpmProcessor.lpMetric.ILpMetric;
import eu.intuitel.lpmProcessor.lpMetric.LpMetricAvailability;
import eu.intuitel.lpmProcessor.lpMetric.LpMetricCoverage;
import eu.intuitel.lpmProcessor.lpMetric.LpMetricHistoryWalker;
import eu.intuitel.lpmProcessor.lpMetric.LpMetricHypercube;
import eu.intuitel.lpmProcessor.lpMetric.LpMetricPopularity;
import eu.intuitel.lpmProcessor.lpReevaluation.ILpReevaluationCriterion;
import eu.intuitel.lpmProcessor.lpReevaluation.LpReevaluationCriterionDeviation;
import eu.intuitel.lpmProcessor.lpReevaluation.LpReevaluationCriterionSteps;
import eu.intuitel.lpmProcessor.lpReevaluation.LpReevaluationCriterionTime;
import eu.intuitel.lpmProcessor.models.LpEvaluationResult;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * Evaluator util to test LPs for a given learner.
 *
 * @author fhe
 */
public class LpEvalUtil {

    private static final List<ILpMetric> lpMetrics;
    private static final List<ILpReevaluationCriterion> lpReevaluationCriteria;

    static {
        lpMetrics = new ArrayList<>();
        lpMetrics.add(new LpMetricAvailability());
        lpMetrics.add(new LpMetricCoverage());
        lpMetrics.add(new LpMetricHistoryWalker());
        lpMetrics.add(new LpMetricHypercube());
        lpMetrics.add(new LpMetricPopularity());

        lpReevaluationCriteria = new ArrayList<>();
        lpReevaluationCriteria.add(new LpReevaluationCriterionTime());
        lpReevaluationCriteria.add(new LpReevaluationCriterionSteps());
        lpReevaluationCriteria.add(new LpReevaluationCriterionDeviation());
    }

    /**
     * Private constructor - only static methods
     */
    private LpEvalUtil() {
    }

    /**
     * Evaluates which LPs are suitable for a given learner.
     *
     * @param slom SLOM metadata of the course in scope
     * @param learnerId ID of the learner in scope
     * @param db Communication interface with the user database
     * @return Empty Set if no new LPs are found for suggestion.
     */
    public static Set<LpEvaluationResult> getLpEvaluations(SlomMetadata slom, String learnerId, DbAccessUtil db) {
        Set<LpEvaluationResult> resultSet = new HashSet<>();
        resultSet.addAll(getLpEvaluations(slom, learnerId, db,
                db.getMicroLpOfLearner(), //Micro
                db.getHistoryOfKOs()));
        resultSet.addAll(getLpEvaluations(slom, learnerId, db,
                db.getMacroLpOfLearner(), //Macro
                db.getHistoryOfCCs()));
        return resultSet;
    }

    /**
     * Determines the LP evaluations for the given type of LPs.
     *
     * @param slom SLOM metadata currently in scope
     * @param learnerId ID of the learner currently in scope
     * @param db Communication interface to the user database
     * @param currentLp Currently selected lp of a given type
     * @param loHistory Learning history of the learner of the correct level
     * @return Set of new selections or the one that has previously been selected.
     */
    private static Collection<LpEvaluationResult> getLpEvaluations(SlomMetadata slom, String learnerId, DbAccessUtil db,
            LearningPathway currentLp, List<? extends LearningObject> loHistory) {
        Set<LpEvaluationResult> results = new HashSet<>();
        if (isReevaluationNecessary(db, currentLp, loHistory)) {
            results.addAll(getReevaluationResults(slom, db, currentLp));
        }
        return results;
    }

    /**
     * Determines if a re-evaluation of the LP is neccessary.
     *
     * @param slom SLOM metadata currently in scope
     * @param learnerId ID of the learner currently in scope
     * @param db Communication interface to the user database
     * @param lp Currently selected lp of a given type
     * @param loHistory Learning history of the learner of the correct level
     * @return True if re-evaluation is necessary
     */
    private static boolean isReevaluationNecessary(DbAccessUtil db, LearningPathway lp, List<? extends LearningObject> loHistory) {
        for (ILpReevaluationCriterion criterion : lpReevaluationCriteria) {
            if (criterion.fulfillsCriterion(db, lp, loHistory)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Calls the available LP metrics and returns the selection results.
     *
     * @param slom SLOM metadata currently in scope
     * @param learnerId ID of the learner currently in scope
     * @param db Communication interface to the user database
     * @param currentLp Currently selected lp of a given type
     * @param loHistory Learning history of the learner of the correct level
     * @return Empty set if no new suggestions are available
     */
    private static Set<LpEvaluationResult> getReevaluationResults(SlomMetadata slom, DbAccessUtil db,
            LearningPathway currentLp) {
        Set<LpEvaluationResult> results = new HashSet<>();
        for (ILpMetric metric : lpMetrics) {
            for (LearningPathway determinedLp : metric.getLp(slom, db, currentLp.getClass())) {
                if (determinedLp != currentLp && !CompletionUtil.isCompletedLp(db, determinedLp)) {
                    results.add(new LpEvaluationResult(determinedLp, metric.getMethod()));
                }
            }
        }
        return results;
    }
}
