/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-03-12
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.apache.commons.lang.StringUtils;
import org.semanticweb.owlapi.model.OWLNamedIndividual;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Debugger class for LPs to create GraphViz files for visualization purposes.
 *
 * @author fhe
 */
public class LpGrapherUtil {

    private static final Logger logger = LoggerFactory.getLogger(LpGrapherUtil.class);

    /**
     * Creates a graphical output of the given LP.
     *
     * @param lp LP to be visualized
     * @param directory Directory where the output will be saved
     * @return Returns true if successfull.
     */
    public static boolean createGraph(LearningPathway lp, File directory) {
        return createGraph(createGraphCode(lp), lp.getIri().getFragment(), directory);
    }

    /**
     * Creates a graphical output of the given SLOM metadata.
     *
     * @param slom SLOM metadata to be visualized
     * @param directory Directory where the output will be saved
     * @return Returns true if successfull.
     */
    public static boolean createGraph(SlomMetadata slom, File directory) {
        //Note: The graph code for complete SLOMs is messy! This needs improvement. 
        return createGraph(createGraphCode(slom), slomToName(slom), directory);
    }

    /**
     * Creates a String from the given SLOM object that can serve as a file name.
     *
     * @param slom SLOM object in scope
     * @return String based on CCM IRI
     */
    private static String slomToName(SlomMetadata slom) {
        return slom.getCcm().getIri().getFragment().replaceAll("[^a-zA-Z0-9.-]", "_").replace(".owl", "");
    }

    /**
     * Creates a graph from the given {@code graphCode} at the specified location.
     *
     * @param graphCode Code of the graph
     * @param filename Name of the file to store the graph code in
     * @param directory Directory of the file file to store
     * @return True if successfull.
     */
    private static boolean createGraph(String graphCode, String filename, File directory) {
        try {
            if (StringUtils.isBlank(graphCode)) {
                logger.error("Graph to visualize is blank.");
                return false;
            }
            if (StringUtils.isBlank(filename)) {
                logger.error("Filename is blank.");
                return false;
            }
            if (!directory.exists()) {
                directory.mkdirs();
            }
            File gvFile = new File(directory + "\\" + filename + ".gv");
            File batFile = new File(directory + "\\" + filename + ".bat");
            if (!createFileFromString(gvFile, graphCode)) {
                return false;
            }
            if (!createFileFromString(batFile, createBatch("svg", gvFile))) {
                return false;
            }
            logger.debug("Created LP-graph code: " + gvFile);
            return true;
        } catch (Exception ex) {
            logger.error("Exception while creating graph files.", ex);
            return false;
        }
    }

    /**
     * Adds the starter nodes of the graph to the graph code.
     *
     * @param starters Learning Objects that only have leaving LP-relations
     * @param sb StringBuilder to which the code will be added
     */
    private static void addStarterNodes(Collection<? extends LearningObject> starters, StringBuilder sb) {
        sb.append(" node [shape=circle, fillcolor=\"#B5E61D\", style=filled]; ");
        for (LearningObject starter : starters) {
            sb.append(" ").append(starter.getIri().getFragment()).append(" ");
        }
        sb.append("; /*List of start-states*/\n");
    }

    /**
     * Adds the ender nodes of the graph to the graph code.
     *
     * @param enders Learning Objects that only have incoming LP-relations
     * @param sb StringBuilder to which the code will be added
     */
    private static void addEnderNodes(Collection<? extends LearningObject> enders, StringBuilder sb) {
        sb.append(" node [shape=circle, fillcolor=\"#04D3FF\", style=filled]; ");
        for (LearningObject ender : enders) {
            sb.append(" ").append(ender.getIri().getFragment()).append(" ");
        }
        sb.append("; /*List of end-states*/\n");
    }

    /**
     * Adds the intermediary nodes of the graph to the graph code.
     *
     * @param intermediates Learning Objects that have incoming and leaving LP-relations
     * @param sb StringBuilder to which the code will be added
     */
    private static void addIntermediateNodes(Collection<LearningObject> intermediates, StringBuilder sb) {
        if (intermediates.isEmpty()) {
            return;
        }
        sb.append(" node [shape=circle, fillcolor=\"#EEEEEE\", style=filled]; ");
        for (LearningObject enders : intermediates) {
            sb.append(" ").append(enders.getIri().getFragment()).append(" ");
        }
        sb.append("; /*List of intermediate states*/\n");
    }

    /**
     * Adds the transitions between the graph nodes.
     *
     * @param transitions Map containing the LP-steps and an LO-Array containing at 0 the left and
     * at 1 the right side of the LP-relation
     * @param sb StringBuilder to which the code will be added
     */
    private static void addTransitions(Map<OWLNamedIndividual, LearningObject[]> transitions, StringBuilder sb) {
        for (OWLNamedIndividual step : transitions.keySet()) {
            LearningObject[] los = transitions.get(step);
            sb.append(" ");
            sb.append(los[0].getIri().getFragment());
            sb.append(" -> ");
            sb.append(los[1].getIri().getFragment());
            sb.append(" [ label = \"");
            sb.append(step.getIRI().getFragment());
            sb.append("\" ];\n");
        }
    }

    /**
     * Creates a file with the given String as content.
     *
     * @param file File to be created
     * @param content String to be stored in the file
     * @return True if saving file was successfull.
     */
    private static boolean createFileFromString(File file, String content) {
        PrintWriter out = null;
        try {
            if (!file.exists()) {
                file.getParentFile().mkdir();
                file.createNewFile();
            }
            out = new PrintWriter(file);
            out.println(content);
            return true;
        } catch (IOException ex) {
            logger.error(String.format("Exception while creating file %s.", file), ex);
            return false;
        } finally {
            if (out != null) {
                out.close();
            }
        }
    }

    /**
     * Creates the graph code for the given LP.
     *
     * @param lp LP for which the graph should be created
     * @return String representing the graph code to create the visualization.
     */
    private static String createGraphCode(LearningPathway lp) {
        StringBuilder sb = new StringBuilder();
        sb.append("digraph ");
        sb.append(lp.getIri().getFragment());
        sb.append(" { \n rankdir=LR; \n ");
        addStarterNodes(lp.getStarters(), sb);
        addEnderNodes(lp.getEnders(), sb);
        addIntermediateNodes(getIntermediates(lp), sb);
        addTransitions(lp.getLpSteps(), sb);
        sb.append("}");
        return sb.toString();
    }

    /**
     * Creates the graph code of the given SLOM metadata.
     *
     * @param slom SLOM for which the graph should be created
     * @return String representing the graph code to create the visualization.
     */
    private static String createGraphCode(SlomMetadata slom) {
        StringBuilder sb = new StringBuilder();
        sb.append("digraph ");
        sb.append(slomToName(slom));
        sb.append(" { \n rankdir=LR; \n ");

        Collection<LearningObject> starters = new HashSet<>();
        Collection<LearningObject> enders = new HashSet<>();
        Collection<LearningObject> intermediates = new HashSet<>();
        Map<OWLNamedIndividual, LearningObject[]> transitions = new HashMap<>();
        for (LearningPathway lp : slom.getMacroLps()) {
            populateCollections(lp, starters, enders, intermediates, transitions);
        }
        for (LearningPathway lp : slom.getMicroLps()) {
            populateCollections(lp, starters, enders, intermediates, transitions);
        }
        addStarterNodes(starters, sb);
        addEnderNodes(enders, sb);
        addIntermediateNodes(intermediates, sb);
        addTransitions(transitions, sb);

        Map<KnowledgeObject, Set<ConceptContainer>> koCcRelations = new HashMap<>();
        for (KnowledgeObject ko : slom.getKos()) {
            koCcRelations.put(ko, ko.getCcReferences());
        }
        addKoCcRelations(koCcRelations, sb);

        sb.append("}");
        return sb.toString();
    }

    /**
     * Creates a batch to easily execute the graph code.
     *
     * @param outputType Type of graphic to be created (e.g. svg)
     * @param gvFile File containing the graph code
     * @return Batch code
     */
    private static String createBatch(String outputType, File gvFile) {
        StringBuilder sb = new StringBuilder();
        sb.append("@echo off\n");
        String outputFile = gvFile.getName().replace(".gv", "." + outputType);
        sb.append(String.format("dot.exe -T%s %s > %s \n", outputType, gvFile.getName(), gvFile.getName().replace(".gv", "." + outputType)));
        sb.append("@echo Created graph at %CD%");
        sb.append(outputFile);
        sb.append("\n pause");
        return sb.toString();
    }

    /**
     * Analyzes the given LP and populates the given collections.
     *
     * @param lp Learning Pathway to inspect
     * @param starters All LOs with only leaving LP-relations
     * @param enders All LOs wich only incoming LP-relations
     * @param intermediates All LOs with leaving and incoming LP-relations
     * @param transitions Map containing the LP-steps with their start- and end-points
     */
    private static void populateCollections(LearningPathway lp,
            Collection<LearningObject> starters,
            Collection<LearningObject> enders,
            Collection<LearningObject> intermediates,
            Map<OWLNamedIndividual, LearningObject[]> transitions) {
        starters.addAll(lp.getStarters());
        enders.addAll(lp.getEnders());
        intermediates.addAll(getIntermediates(lp));
        transitions.putAll(lp.getLpSteps());
    }

    /**
     * Determines the intermediary nodes of the given LP.
     *
     * @param lp LP to inspect
     * @return The intermediate nodes of the LP, i.e. all nodes having incoming and leaving
     * LP-relations.
     */
    private static Collection<LearningObject> getIntermediates(LearningPathway lp) {
        Set<LearningObject> intermediates = new HashSet<>(lp.getAllReferences());
        intermediates.removeAll(lp.getStarters());
        intermediates.removeAll(lp.getEnders());
        return intermediates;
    }

    /**
     * Adds all KO-CC-relations (isContainedInCC) to the graph code.
     *
     * @param koCcConnections Map containing the connections between KOs and their connected CCs
     * @param sb StringBuilder containing the graph code
     */
    private static void addKoCcRelations(Map<KnowledgeObject, Set<ConceptContainer>> koCcConnections, StringBuilder sb) {
        for (KnowledgeObject ko : koCcConnections.keySet()) {
            for (ConceptContainer cc : koCcConnections.get(ko)) {
                sb.append(" ");
                sb.append(ko.getIri().getFragment());
                sb.append(" -> ");
                sb.append(cc.getIri().getFragment());
                sb.append(" [ color=\"#31CEF0\" ];\n");
            }
        }
    }
}
