/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import org.apache.commons.lang3.StringUtils;
import org.coode.owlapi.turtle.TurtleOntologyFormat;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.io.OWLXMLOntologyFormat;
import org.semanticweb.owlapi.io.RDFXMLOntologyFormat;
import org.semanticweb.owlapi.io.StringDocumentSource;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLAnnotation;
import org.semanticweb.owlapi.model.OWLAnnotationProperty;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLDataFactory;
import org.semanticweb.owlapi.model.OWLDataProperty;
import org.semanticweb.owlapi.model.OWLDataPropertyExpression;
import org.semanticweb.owlapi.model.OWLIndividual;
import org.semanticweb.owlapi.model.OWLLiteral;
import org.semanticweb.owlapi.model.OWLNamedIndividual;
import org.semanticweb.owlapi.model.OWLObjectProperty;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyFormat;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.semanticweb.owlapi.model.OWLOntologySetProvider;
import org.semanticweb.owlapi.model.OWLOntologyStorageException;
import org.semanticweb.owlapi.model.UnknownOWLOntologyException;
import org.semanticweb.owlapi.reasoner.OWLReasoner;
import org.semanticweb.owlapi.reasoner.OWLReasonerFactory;
import org.semanticweb.owlapi.reasoner.structural.StructuralReasonerFactory;
import org.semanticweb.owlapi.util.OWLOntologyMerger;
import org.semanticweb.owlapi.util.SimpleIRIMapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Collection of methods using OwlApi to provide easy access to relevant data.
 *
 * @author fhe
 */
public class OwlUtil {

    private static final Logger logger = LoggerFactory.getLogger(OwlUtil.class);

    /**
     * Private constructor - static class
     */
    private OwlUtil() {
    }

    /**
     * Returns the IRI of the given {@code ontology}.
     *
     * @param ontology
     * @return IRI of the given OWL object
     */
    public static IRI getIri(OWLOntology ontology) {
        return ontology.getOntologyID().getOntologyIRI();
    }

    /**
     * Returns the IRI of the given {@code individual}.
     *
     * @param individual
     * @return IRI of the given OWL object
     */
    public static IRI getIri(OWLIndividual individual) {
        return ((OWLNamedIndividual) individual).getIRI();
    }

    /**
     * Returns the IRI of the given {@code objectProperty}.
     *
     * @param objectProperty
     * @return IRI of the given OWL object
     */
    public static IRI getIri(OWLObjectProperty objectProperty) {
        return objectProperty.getIRI();
    }

    /**
     * Returns the given {@code iri} as String.
     *
     * @param iri IRI to return as String
     * @return String representation of the given IRI
     */
    public static String getIriAsString(IRI iri) {
        return iri.toString();
    }

    /**
     * Returns the IRI of the given {@code objectProperty} as String
     *
     * @param objectProperty OWLObjectProperty whose IRI should be returned
     * @return String representation of the IRI of the given OWL object
     */
    public static String getIriAsString(OWLObjectProperty objectProperty) {
        return getIri(objectProperty).toString();
    }

    /**
     * Returns the IRI of the given {@code individual} as String
     *
     * @param individual OWLIndividual whose IRI should be returned
     * @return String representation of the IRI of the given OWL object
     */
    public static String getIriAsString(OWLIndividual individual) {
        return getIri(individual).toString();
    }

    /**
     * Returns the IRI of the given {@code ontology} as String
     *
     * @param ontology OWLOntology whise IRI should be returned
     * @return String representation of the IRI of the given OWL object
     */
    public static String getIriAsString(OWLOntology ontology) {
        return getIri(ontology).toString();
    }

    /**
     * Returns an OWLReasoner instance for the given {@code ontology}.
     *
     * @param ontology Ontology for which a reasoner should be created
     * @return Reasoner for the given ontology
     */
    public static OWLReasoner getOwlReasoner(OWLOntology ontology) {
        OWLReasonerFactory reasonerFactory = new StructuralReasonerFactory();
        // De-comment for progress output to console
        //ConsoleProgressMonitor progressMonitor = new ConsoleProgressMonitor();
        //OWLReasonerConfiguration config = new SimpleConfiguration(progressMonitor);
        //OWLReasoner reasoner = reasonerFactory.createReasoner(onto, config);
        OWLReasoner reasoner = reasonerFactory.createReasoner(ontology);
        reasoner.precomputeInferences();
        return reasoner;
    }

    /**
     * Returns an instance of OWLClass of the given {@code className} and {@code baseURI}.
     *
     * @param baseURI Base URI of the class to create
     * @param className Name of the class to create (starting with "#")
     * @param fac OWLDataFactory to create class with
     * @return OWLClass representation as generated by the given input parameters
     */
    public static OWLClass getOwlClass(String baseURI, String className, OWLDataFactory fac) {
        return fac.getOWLClass(IRI.create(baseURI + className));
    }

    /**
     * Returns an instance of OWLObjectProperty of the given {@code propertyName} and
     * {@code baseURI}.
     *
     * @param baseURI Base URI of the property to create
     * @param propertyName Name of the property to create
     * @param fac OWLDataFactory to create class with
     * @return OWLObjectProperty representation as created via joining {@code baseURI} and
     * {@code propertyName}
     */
    public static OWLObjectProperty getObjectProperty(String baseURI, String propertyName, OWLDataFactory fac) {
        return fac.getOWLObjectProperty(IRI.create(baseURI + propertyName));
    }

    /**
     * Returns an instance of OWLDataProperty for the given {@code propertyName} and
     * {@code baseURI}.
     *
     * @param baseURI Base URI of the property to create
     * @param propertyName Name of the property to create
     * @param fac OWLDataFactory to create class with
     * @return OWLDataProperty representation as created via joining {@code baseURI} and
     * {@code propertyName}
     */
    public static OWLDataProperty getDataProperty(String baseURI, String propertyName, OWLDataFactory fac) {
        return fac.getOWLDataProperty(IRI.create(baseURI + propertyName));
    }

    /**
     * Returns an instance of OWLAnnotationProperty for the given {@code propertyName} and
     * {@code baseURI}.
     *
     * @param baseURI Base URI of the property to create
     * @param propertyName Name of the property to create
     * @param fac OWLDataFactory to create class with
     * @return OWLAnnotationProperty representation as created via joining {@code baseURI} and
     * {@code propertyName}
     */
    public static OWLAnnotationProperty getAnnotationProperty(String baseURI, String propertyName, OWLDataFactory fac) {
        return fac.getOWLAnnotationProperty(IRI.create(baseURI + propertyName));
    }

    /**
     * Returns a set of NamedIndividuals of the given {@code owlClass} as provided via the
     * {@code reasoner}.
     *
     * @param reasoner Reasoner that contains of the ontology to inspect
     * @param owlClass Class to look for
     * @param onlyDirectSubs Should only look for DirectClassAssertions
     * @return Empty set if nothing is found
     */
    public static Set<OWLNamedIndividual> getNamedIndividualsOfClass(OWLReasoner reasoner, OWLClass owlClass, boolean onlyDirectSubs) {
        return reasoner.getInstances(owlClass, onlyDirectSubs).getFlattened();
    }

    /**
     * Returns a the subproperties of the given {@code objectProperty} in a specified
     * {@code ontology}.
     *
     * @param objectProperty Property to get the subproperties from
     * @param onto Ontology to inspect
     * @return Set of OWLObjectProperties that are listed of sub-properties of the given
     * {@code objectProperty}
     */
    public static Set<OWLObjectProperty> getSubPropertiesOfProperty(OWLObjectProperty objectProperty, OWLOntology onto) {
        Set<OWLObjectProperty> propSet = new HashSet<>();
        for (OWLObjectPropertyExpression prop : objectProperty.getSubProperties(onto)) {
            propSet.add((OWLObjectProperty) prop);
        }
        return propSet;
    }

    /**
     * Returns a map of OWLObjectPropertyExpressions referencing on a Set of individuals of the
     * given {@code namedIndividual} in the specified {@code ontology}.
     *
     * @param namedIndividual Individual in scope
     * @param ontology Ontology in scope
     * @return Map containing the ObjectProperties and their values of the given
     * {@code namedIndividual}
     */
    public static Map<OWLObjectPropertyExpression, Set<OWLIndividual>> getObjectPropertiesOfNamedIndividual(OWLNamedIndividual namedIndividual, OWLOntology ontology) {
        return namedIndividual.getObjectPropertyValues(ontology);
    }

    /**
     * Returns the values that have been assigned to the given {@code namedIndividual} via the
     * specified
     * {@code objectProperty}.
     *
     * @param namedIndividual Individual in scope
     * @param onto Ontology in scope
     * @param objectProperty
     * @return Set of OWLIndividuals that are referenced on by the specified {@code objectProperty}
     * by the given {@code namedIndividual}
     */
    public static Set<OWLIndividual> getObjectPropertyValuesOfNamedIndividual(OWLNamedIndividual namedIndividual, OWLOntology onto, OWLObjectProperty objectProperty) {
        return namedIndividual.getObjectPropertyValues(objectProperty, onto);
    }

    /**
     * Returns a map of OWLDataPropertyExpressions referencing on a Set of OWLLiterals of the given
     * {@code namedIndividual} in the specified {@code ontology}.
     *
     * @param namedIndividual Individual in scope
     * @param ontology Ontology in scope
     * @return Map containing the data property values of the given {@code namedIndividual}
     */
    public static Map<OWLDataPropertyExpression, Set<OWLLiteral>> getDataPropertiesOfNamedIndividual(OWLNamedIndividual namedIndividual, OWLOntology ontology) {
        return namedIndividual.getDataPropertyValues(ontology);
    }

    /**
     * Returns the values that have been assigned to the given {@code namedIndividual} via the given
     * {@code dataProperty}.
     *
     * @param namedIndividual Individual in scope
     * @param onto Ontology in scope
     * @param dataProperty IRI of the property to have a look at
     * @return Set of data property values for the given {@code dataProperty} of the specified
     * {@code namedIndividual}
     */
    public static Set<OWLLiteral> getDataPropertyValuesOfNamedIndividual(OWLNamedIndividual namedIndividual, OWLOntology onto, OWLDataPropertyExpression dataProperty) {
        return namedIndividual.getDataPropertyValues(dataProperty, onto);
    }

    /**
     * Returns the value of the annotation property of the ontology as String.
     *
     * @param annotationProp Property to be loaded
     * @param onto Ontology to load the property from
     * @return Null in case nothing was loaded.
     */
    public static String getAnnotationPropertyValue(OWLOntology onto, OWLAnnotationProperty annotationProp) {
        for (OWLAnnotation annotation : onto.getAnnotations()) {
            if (annotation.getProperty().equals(annotationProp)) {
                return annotation.getValue().toString().replace("\"", "");
            }
        }
        return null;
    }

    /**
     * Returns the value of the annotation property of the given individual as String.
     *
     * @param annotationProp Property to be loaded
     * @param onto Ontology to load the property from
     * @param individual Individual to get the annotation from
     * @return Null in case nothing was loaded.
     */
    public static String getAnnotationPropertyValue(OWLOntology onto, OWLNamedIndividual individual, OWLAnnotationProperty annotationProp) {
        for (OWLAnnotation annotation : individual.getAnnotations(onto, annotationProp)) {
            if (annotation.getProperty().equals(annotationProp)) {
                return annotation.getValue().toString().replace("\"", "");
            }
        }
        return null;
    }

    /**
     * Returns the value of the annotation property of the given individual as String.
     *
     * @param annotationProp Property to be loaded
     * @param onto Ontology to load the property from
     * @param owlClass OwlClass to get the annotation from
     * @return Null in case nothing was loaded.
     */
    public static String getAnnotationPropertyValue(OWLOntology onto, OWLClass owlClass, OWLAnnotationProperty annotationProp) {
        for (OWLAnnotation annotation : owlClass.getAnnotations(onto, annotationProp)) {
            if (annotation.getProperty().equals(annotationProp)) {
                return annotation.getValue().toString().replace("\"", "");
            }
        }
        return null;
    }

    /**
     * Returns the subproperties of the given {@code objectProperty} in the given
     * {@code ontologySet}.
     *
     * @param objectProperty Property to analyse
     * @param ontologySet Set of ontologies to inspect
     * @param recursive If true, also the subproperties will be analysed
     * @return Set of sub-object-properties of the specified {@code objectProperty}
     */
    public static Set<OWLObjectProperty> getSubObjectPropertiesOf(OWLObjectProperty objectProperty, Set<OWLOntology> ontologySet, boolean recursive) {
        Set<OWLObjectProperty> properties = new HashSet<>();
        for (OWLObjectPropertyExpression subProp : objectProperty.getSubProperties(ontologySet)) {
            properties.add(subProp.asOWLObjectProperty());
            if (recursive) {
                properties.addAll(getSubObjectPropertiesOf(subProp.asOWLObjectProperty(), ontologySet, true));
            }
        }
        return properties;
    }

    /**
     * Returns the subclasses of the given {@code owlClass} the given Set of ontologies.
     *
     * @param owlClass Class to be inspected
     * @param ontos Ontologies to investigate
     * @param recursive Recursive search for more decendents
     * @return Set of OWLClasses, which are all decendent of the given {@code owlClass}.
     */
    public static Set<OWLClass> getSubClassesOf(OWLClass owlClass, Set<OWLOntology> ontos, boolean recursive) {
        Set<OWLClass> subClasses = new HashSet<>();
        for (OWLClassExpression subClass : owlClass.getSubClasses(ontos)) {
            subClasses.add(subClass.asOWLClass());
            if (recursive) {
                subClasses.addAll(getSubClassesOf(subClass.asOWLClass(), ontos, recursive));
            }
        }
        return subClasses;
    }

    /**
     * Returns the subclasses of the given {@code owlClass} the given Set of ontologies.
     *
     * @param owlClass Class to be inspected
     * @param onto Ontology to investigate
     * @param recursive Recursive search for more decendents
     * @return Set of OWLClasses, which are all decendent of the given {@code owlClass}.
     */
    public static Set<OWLClass> getSubClassesOf(OWLClass owlClass, OWLOntology onto, boolean recursive) {
        Set<OWLClass> subClasses = new HashSet<>();
        for (OWLClassExpression subClass : owlClass.getSubClasses(onto)) {
            subClasses.add(subClass.asOWLClass());
            if (recursive) {
                subClasses.addAll(getSubClassesOf(subClass.asOWLClass(), onto, recursive));
            }
        }
        return subClasses;
    }

    /**
     * Adds the given IRI mapping to the OWLOntologyManager.
     *
     * @param manager Manager to which the mapping will be added
     * @param mapping Map with IRI (key) to map on IRI (value)
     */
    public static void addIriMapping(OWLOntologyManager manager, Map<IRI, IRI> mapping) {
        for (Map.Entry<IRI, IRI> entry : mapping.entrySet()) {
            manager.addIRIMapper(new SimpleIRIMapper(entry.getKey(), entry.getValue()));
        }
    }

    /**
     * Adds the given IRI mapping to the OWLOntologyManager.
     *
     * @param manager Manager to which the mapping will be added
     * @param originalIri Original IRI to be replaced
     * @param replacementIRI New IRI to be used as replacement
     */
    public static void addIriMapping(OWLOntologyManager manager, IRI originalIri, IRI replacementIRI) {
        manager.addIRIMapper(new SimpleIRIMapper(originalIri, replacementIRI));
    }

    /**
     * Returns the OWLOntology object with the given filename, which is imported in the
     * {@code sourceOntology}.
     *
     * @param sourceOntotlogy Ontology to inspect
     * @param importOntologyFileName Filename of the imported ontology to look for (not complete
     * IRI)
     * @return Null if not found
     */
    public static OWLOntology getOntologyFromImports(OWLOntology sourceOntotlogy, String importOntologyFileName) {
        for (OWLOntology onto : sourceOntotlogy.getImports()) {
            if (importOntologyFileName.equals(onto.getOntologyID().getOntologyIRI().getFragment())) {
                return onto;
            }
        }
        return null;
    }

    /**
     * Returns the values of {@code objectProp} if defined for {@code individual} in the
     * {@code onto}.
     *
     * @param individual Individual to inspect
     * @param objectProp Property to inspect
     * @param ontology Ontology containing the individual
     */
    public static Set<OWLIndividual> getObjectPropertyValue(OWLNamedIndividual individual, OWLObjectProperty objectProp, OWLOntology ontology) {
        return individual.getObjectPropertyValues(objectProp, ontology);
    }

    /**
     * Returns a new OWLOntologyManager.
     *
     * @return New instance of an OWLOntologyManager
     */
    public static OWLOntologyManager createOwlOntologyManager() {
        return OWLManager.createOWLOntologyManager();
    }

    /**
     * Returns a new OWLDataFactory as created by the given {@code manager}.
     *
     * @param manager OWLOntologyManager instance which is used to create the OWLDataFactory
     * @return New instance of an OWLDataFactory
     */
    public static OWLDataFactory createOwlDataFactory(OWLOntologyManager manager) {
        return manager.getOWLDataFactory();
    }

    /**
     * Merges the given {@code ontologies} using the {@code manager} and assigns it the specified
     * {@code iriName}.
     *
     * @param ontologies Ontologies to be merged.
     * @param iri IRI of the ontology to be created.
     * @param manager OWLOntologyManager to be used for merging.
     * @return New merged OWLOntology
     */
    public static OWLOntology mergeOntologies(OWLOntologyManager manager, IRI iri, OWLOntology... ontologies) {
        if (ontologies.length <= 0) {
            logger.error("The given ontology collection to merge is emtpy.");
            return null;
        }
        if (ontologies.length == 1) {
            logger.warn("The given ontology collection to merge only containes one item.");
            return null;
        }
        try {
            /**
             * Custom OWLOntologySetProvider for merging ontologies.
             */
            class CustomOwlOntologySetProvider implements OWLOntologySetProvider {

                /**
                 * Set of ontologies to be mergered.
                 */
                OWLOntology[] ontologies;

                /**
                 * Constructor
                 */
                public CustomOwlOntologySetProvider(OWLOntology... ontologies) {
                    this.ontologies = ontologies;
                }

                @Override
                public Set<OWLOntology> getOntologies() {
                    return new HashSet<>(Arrays.asList(ontologies));
                }
            }
            OWLOntologyMerger merger = new OWLOntologyMerger(new CustomOwlOntologySetProvider(ontologies));
            return merger.createMergedOntology(manager, iri);
        } catch (OWLOntologyCreationException ex) {
            logger.error("Ontology creation exception while merging ontologies.", ex);
            return null;
        }
    }

    /**
     * Returns the given {@code ontology} as its Turtle representation.
     *
     * @param ontology Ontology to be converted into its String representation
     * @param additionalPrefixes Map containing additional prefixes to be add to the output.
     * (prefixName, prefixIRI#)
     * @return Null in case of exception
     */
    public static String toTurtle(OWLOntology ontology, Map<String, String> additionalPrefixes) {
        if (ontology == null) {
            logger.error("The given ontology is null.");
            return null;
        }
        if (additionalPrefixes == null) {
            logger.error("The given ontology prefixes are null.");
            return null;
        }
        try {
            OWLOntologyManager manager = ontology.getOWLOntologyManager();
            OWLOntologyFormat format = manager.getOntologyFormat(ontology);
            TurtleOntologyFormat turtleFormat = new TurtleOntologyFormat();
            turtleFormat.copyPrefixesFrom(format.asPrefixOWLOntologyFormat());
            for (Map.Entry<String, String> entry : additionalPrefixes.entrySet()) {
                turtleFormat.setPrefix(entry.getKey(), entry.getValue());
            }
            turtleFormat.setDefaultPrefix(getIriAsString(ontology));
            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            manager.saveOntology(ontology, turtleFormat, stream);
            return stream.toString();
        } catch (UnknownOWLOntologyException | OWLOntologyStorageException ex) {
            logger.error("Exception on converting OWLOntology object to Turtle.", ex);
            return null;
        }
    }

    /**
     * Returns the given {@code ontology} as its XML representation.
     *
     * @param ontology Ontology to be converted into its String representation
     * @param additionalPrefixes Map containing additional prefixes to be add to the output.
     * (prefixName, prefixIRI#)
     * @return Null in case of exception
     */
    public static String toXml(OWLOntology ontology, Map<String, String> additionalPrefixes) {
        if (ontology == null) {
            logger.error("The given ontology is null.");
            return null;
        }
        if (additionalPrefixes == null) {
            logger.error("The given ontology prefixes are null.");
            return null;
        }
        try {
            OWLOntologyManager manager = ontology.getOWLOntologyManager();
            OWLOntologyFormat format = manager.getOntologyFormat(ontology);
            RDFXMLOntologyFormat xmlFormat = new RDFXMLOntologyFormat();
            xmlFormat.copyPrefixesFrom(format.asPrefixOWLOntologyFormat());
            for (Map.Entry<String, String> entry : additionalPrefixes.entrySet()) {
                xmlFormat.setPrefix(entry.getKey(), entry.getValue());
            }
            xmlFormat.setDefaultPrefix(getIriAsString(ontology));
            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            manager.saveOntology(ontology, xmlFormat, stream);
            return stream.toString("UTF-8");
        } catch (UnknownOWLOntologyException | OWLOntologyStorageException | UnsupportedEncodingException ex) {
            logger.error("Exception on converting OWLOntology object to XML.", ex);
            return null;
        }
    }

    /**
     * Loads the given String representation of an ontology to an OWLOntology object.
     *
     * @param ontology String representation of the ontology to be loaded
     * @return The loaded ontology
     * @throws OWLOntologyCreationException
     */
    public static OWLOntology fromXml(String ontology) throws OWLOntologyCreationException {
        OWLOntologyManager manager = createOwlOntologyManager();
        return manager.loadOntologyFromOntologyDocument(new StringDocumentSource(ontology));
    }

    /**
     * Reads and returns the value of the specified data property as String.
     *
     * @param prop OWL object-property to inspect
     * @param individual OWL individual to inspect
     * @param onto Ontology of the individual
     * @return Null in case more than one or no results could be obtained.
     */
    public static String getDataPropertyValue(OWLDataProperty prop, OWLNamedIndividual individual, OWLOntology onto) {
        Set<OWLLiteral> values = OwlUtil.getDataPropertyValuesOfNamedIndividual(individual, onto, prop);
        if (values.size() == 1) {
            return values.iterator().next().getLiteral();
        }
        if (values.size() >= 2) {
            logger.warn(String.format("DataProperty %s could not be added for '%s', because more than one value has been defined.",
                    prop, individual.getIRI().getFragment()));
        }
        return null;
    }

    /**
     * Reads and returns the value of the specified data property as Boolean.
     *
     * @param prop OWL object-property to inspect
     * @param individual OWL individual to inspect
     * @param onto Ontology of the individual
     * @return Null in case the datatype doesn't fit or if more than one or no results could be
     * obtained.
     */
    public static Boolean getDataPropertyValueAsBoolean(OWLDataProperty prop, OWLNamedIndividual individual, OWLOntology onto) {
        String valueString = getDataPropertyValue(prop, individual, onto);
        try {
            if (!StringUtils.isBlank(valueString)) {
                return Boolean.valueOf(valueString);
            }
        } catch (NumberFormatException ex) {
            logger.warn(String.format("DataProperty %s could not be added for '%s', because the value '%s' could not be converted to Integer.",
                    prop, individual.getIRI().getFragment(), valueString), ex);
        }
        return null;
    }

    /**
     * Reads and returns the value of the specified data property as Integer.
     *
     * @param prop OWL object-property to inspect
     * @param individual OWL individual to inspect
     * @param onto Ontology of the individual
     * @return Null in case the datatype doesn't fit or if more than one or no results could be
     * obtained.
     */
    public static Integer getDataPropertyValueAsInteger(OWLDataProperty prop, OWLNamedIndividual individual, OWLOntology onto) {
        String valueString = getDataPropertyValue(prop, individual, onto);
        try {
            if (!StringUtils.isBlank(valueString)) {
                return Integer.valueOf(valueString);
            }
        } catch (NumberFormatException ex) {
            logger.warn(String.format("DataProperty %s could not be added for '%s', because the value '%s' could not be converted to Integer.",
                    prop, individual.getIRI().getFragment(), valueString), ex);
        }
        return null;
    }

    /**
     * Reads and returns the value of the specified data property as Double.
     *
     * @param prop OWL object-property to inspect
     * @param individual OWL individual to inspect
     * @param onto Ontology of the individual
     * @return Null in case the datatype doesn't fit or if more than one or no results could be
     * obtained.
     */
    public static Double getDataPropertyValueAsDouble(OWLDataProperty prop, OWLNamedIndividual individual, OWLOntology onto) {
        String valueString = getDataPropertyValue(prop, individual, onto);
        try {
            if (!StringUtils.isBlank(valueString)) {
                return Double.valueOf(valueString);
            }
        } catch (NumberFormatException ex) {
            logger.warn(String.format("DataProperty %s could not be added for '%s', because the value '%s' could not be converted to Integer.",
                    prop, individual.getIRI().getFragment(), valueString), ex);
        }
        return null;
    }
}
