/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.semanticweb.owlapi.model.IRI;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Util to load easily load and cache property data.
 *
 * @author fhe
 */
public class PropertiesLoaderUtil {

    private static final Logger logger = LoggerFactory.getLogger(PropertiesLoaderUtil.class);
    /**
     * Relative path to the PO version mappings properties file.
     */
    public static final String PROPERTIES_PO = "intui_PO.properties";
    /**
     * Relative path to the LMO properties file
     */
    public static final String PROPERTIES_LMO = "intui_LM.properties";
    /**
     * Property cache per file
     */
    private static final Map<String, Map<String, String>> propCache = new HashMap<>();
    /**
     * IRI cache per file
     */
    private static final Map<String, Map<IRI, IRI>> iriCache = new HashMap<>();

    /**
     * Private constructor - Static class
     */
    private PropertiesLoaderUtil() {
    }

    /**
     * Loads the properties from the specified properties {@code fileToLoad}.
     *
     * @param fileToLoad Path of the properties file to be loaded
     * @return Map containing key-value-pairs that are contained in the properties file
     */
    public static Map<String, String> loadPropertiesFrom(String fileToLoad) {
        if (propCache.containsKey(fileToLoad)) {
            return propCache.get(fileToLoad);
        }
        ClassLoader classloader = Thread.currentThread().getContextClassLoader();
        URL resourceToLoad = classloader.getResource(fileToLoad);
        Map<String, String> map = new HashMap<>();
        try {
            Configuration config = new PropertiesConfiguration(resourceToLoad);
            Iterator configEntriesIterator = config.getKeys();
            while (configEntriesIterator.hasNext()) {
                String key = configEntriesIterator.next().toString();
                String value = config.getString(key);
                map.put(key, value);
            }
        } catch (ConfigurationException ex) {
            logger.error("Could not load properties file: " + fileToLoad, ex);
        }
        propCache.put(fileToLoad, map);
        return map;
    }

    /**
     * Reads and loads the IRI mappings from a properties file.
     * Properties value must be present in the form: "OriginalIri">"ReplacementIri"
     *
     * @param fileToLoad Properties file to be loaded
     * @return IRI mappings as specified in the properties file or and empty set if nothing could be
     * loaded
     */
    public static Map<IRI, IRI> loadIrisFromProperties(String fileToLoad) {
        if (iriCache.containsKey(fileToLoad)) {
            return iriCache.get(fileToLoad);
        }
        Map<String, String> properties = loadPropertiesFrom(fileToLoad);
        Map<IRI, IRI> iriMappings = new HashMap<>();
        ClassLoader classloader = Thread.currentThread().getContextClassLoader();
        for (Map.Entry<String, String> entry : properties.entrySet()) {
            String propertyValue = entry.getValue();
            int seperatorIndex = propertyValue.lastIndexOf('>');
            if (seperatorIndex > 1) {
                try {
                    IRI left = IRI.create(propertyValue.substring(1, seperatorIndex - 1));
                    URL url = classloader.getResource(propertyValue.substring(seperatorIndex + 2, propertyValue.length() - 1));
                    IRI right = IRI.create(url);
                    iriMappings.put(left, right);
                } catch (URISyntaxException ex) {
                    logger.warn("Could not create IRI mapping for: " + propertyValue, ex);
                }
            } else {
                logger.warn("Found IRI mapping in an invalid format: {}={}", entry.getKey(), propertyValue);
            }
        }
        iriCache.put(fileToLoad, iriMappings);
        return iriMappings;
    }
}
