/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.lpmGeneralInterfaces.IEngineCommunicator;
import eu.intuitel.lpmProcessor.models.LpEvaluationResult;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.models.reflex.IErrorReflex;
import eu.intuitel.lpmProcessor.models.reflex.ILearnerReflex;
import eu.intuitel.lpmProcessor.models.reflex.Reflex;
import eu.intuitel.lpmProcessor.models.reflex.ReflexDataIncomplete;
import eu.intuitel.lpmProcessor.models.reflex.ReflexMacroLpEval;
import eu.intuitel.lpmProcessor.models.reflex.ReflexMacroLpNotSet;
import eu.intuitel.lpmProcessor.models.reflex.ReflexMicroLpEval;
import eu.intuitel.lpmProcessor.models.reflex.ReflexMicroLpNotSet;
import eu.intuitel.lpmProcessor.models.reflex.ReflexMultipleCurrentCc;
import eu.intuitel.lpmProcessor.models.reflex.ReflexNoCurrentCc;
import eu.intuitel.lpmProcessor.models.reflex.ReflexNoSlom;
import eu.intuitel.lpmProcessor.models.reflex.ReflexUseEnvFailed;
import eu.intuitel.lpmProcessor.models.reflex.ReflexUsePerfFailed;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.apache.commons.collections.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Static module to create ReflexMsgs based on the stati of preprocessed data items.
 *
 * @author stc, fhe
 */
public abstract class ReflexTesterUtil {

    private static final Logger LOGGER = LoggerFactory.getLogger(ReflexTesterUtil.class);

    /**
     * Exception triggered by a critical reflex.
     */
    public static class ReflexException extends Exception {

        private final Reflex MSG;

        /**
         * Creates a new reflex exception.
         *
         * @param msg The reflex being triggered.
         */
        public ReflexException(Reflex msg) {
            super(msg.toString());
            MSG = msg;
        }

        /**
         * Returns the reflex triggering this exception.
         *
         * @return May never be null.
         */
        public Reflex getReflex() {
            return MSG;
        }
    }

    /**
     * Handles the given Reflex.
     *
     * @param reflex Reflex to be handled
     * @param com Communicator instance to use for sending the reflex
     * @throws ReflexException Exception being thrown when the reflex aborts the recommendation.
     */
    private static void handleReflex(Reflex reflex, IEngineCommunicator com) throws ReflexException {

        //Note: Theoretically, not all reflexes must abort the recommendation process, however, 
        //LMSs don't accept multiple incoming information (e.g. more that two TUG messages).
        //Thus, all reflexes now abort the process to get a singular response from the back-end.
        if (reflex == null) {
            LOGGER.error("Received a null-reflex.");
        } else if (reflex instanceof IErrorReflex) {
            //Note: Until there is some kind of error handling instance, these reflexes are only logged.
            LOGGER.debug("Registered error reflex: " + reflex.toString());
            //if(reflex.getAbortsFlag()) {
            throw new ReflexException(reflex);
            //}
        } else if (reflex instanceof ILearnerReflex) {
            reflex.send(com);
            LOGGER.debug("Registered learner reflex: " + reflex.toString());
            //if (reflex.getAbortsFlag()) {
            throw new ReflexException(reflex);
            //}
        } else {
            LOGGER.debug("Unknown reflex type: " + reflex.toString());
        }
    }

    /**
     * Determines the validity of SLOM data.
     *
     * @param mId Message ID to use for the reflex
     * @param lId ID of the learner in scope
     * @param slom SLOM metadata to test
     * @param com Communicator to use for sending the reflex
     * @throws ReflexException Exception being thrown when the reflex aborts the recommendation.
     */
    public static void testSlom(String mId, String lId,
            SlomMetadata slom, IEngineCommunicator com) throws ReflexException {
        if (slom == null) {
            handleReflex(new ReflexNoSlom(mId, lId, null), com);
        }
    }

    /**
     * Determines the validity of the LP selection.
     *
     * @param mId Message ID to use for the reflex
     * @param lId ID of the learner in scope
     * @param macroLp The currently selected macro LP
     * @param microLp The currently selected micro LP
     * @param slom SLOM metadata to test
     * @param com Communicator to use for sending the reflex
     * @throws ReflexException Exception being thrown when the reflex aborts the recommendation.
     */
    public static void testLpSelection(String mId, String lId, MacroLearningPathway macroLp, MicroLearningPathway microLp,
            SlomMetadata slom, IEngineCommunicator com) throws ReflexException {
        if (macroLp == null) {
            handleReflex(new ReflexMacroLpNotSet(mId, lId, slom.getCourseId(), slom), com);
        }
        if (microLp == null) {
            handleReflex(new ReflexMicroLpNotSet(mId, lId, slom.getCourseId(), slom), com);
        }
    }

    /**
     * Triggers a reflex message when the current CC is not determinable.
     *
     * @param mId Message ID to use for the reflex
     * @param lId ID of the learner in scope
     * @param currentCcs CCs that have been determined as current CCs
     * @param slom SLOM metadata to test
     * @param com Communicator to use for sending the reflex
     * @throws ReflexException Exception being thrown when the reflex aborts the recommendation.
     */
    public static void testCurrentCc(String mId, String lId, Collection<ConceptContainer> currentCcs,
            SlomMetadata slom, IEngineCommunicator com) throws ReflexException {
        if (CollectionUtils.isEmpty(currentCcs)) {
            handleReflex(new ReflexNoCurrentCc(mId, lId, slom.getCourseId()), com);
        }
        if (currentCcs.size() >= 2) {
            handleReflex(new ReflexMultipleCurrentCc(mId, lId, slom.getCourseId(), currentCcs), com);
        }
    }

    /**
     * Triggers a reflex message if the USE performance request has failed.
     *
     * @param mId Message ID to use for the reflex
     * @param lId ID of the learner in scope
     * @param usePerfSuccess Boolean value indicating that the USE performance processing was
     * successful.
     * @param slom SLOM metadata to test
     * @param com Communicator to use for sending the reflex
     * @throws ReflexException Exception being thrown when the reflex aborts the recommendation.
     */
    public static void testUsePerformance(String mId, String lId, boolean usePerfSuccess,
            SlomMetadata slom, IEngineCommunicator com) throws ReflexException {
        if (!usePerfSuccess) {
            handleReflex(new ReflexUsePerfFailed(mId, lId, slom.getCourseId()), com);
        }
    }

    /**
     * Triggers a reflex message if the USE environment request has failed.
     *
     * @param mId Message ID to use for the reflex
     * @param lId ID of the learner in scope
     * @param useEnvSuccess Boolean value indicating that the USE environment processing was
     * successful.
     * @param slom SLOM metadata to test
     * @param com Communicator to use for sending the reflex
     * @throws ReflexException Exception being thrown when the reflex aborts the recommendation.
     */
    public static void testUseEnvironment(String mId, String lId, boolean useEnvSuccess,
            SlomMetadata slom, IEngineCommunicator com) throws ReflexException {
        if (!useEnvSuccess) {
            handleReflex(new ReflexUseEnvFailed(mId, lId, slom.getCourseId()), com);
        }
    }

    /**
     * Triggers a reflex message if there are missing data items.
     *
     * @param mId Message ID to use for the reflex
     * @param lId ID of the learner in scope
     * @param db Util to access the database
     * @param useUtil The UseUtil that determined the USE env data
     * @param slom SLOM metadata to test
     * @param com Communicator to use for sending the reflex
     * @throws ReflexException Exception being thrown when the reflex aborts the recommendation.
     */
    public static void testDataAvailability(String mId, String lId, DbAccessUtil db,
            UserScoreExtractionUtil useUtil, SlomMetadata slom, IEngineCommunicator com) throws ReflexException {
        Reflex reflex = ReflexDataIncomplete.getReflexes(mId, lId, slom.getCourseId(), db);
        if (reflex != null) {
            handleReflex(reflex, com);
        }
    }

    /**
     * Triggers a reflex message if there are missing data items.
     *
     * @param mId Message ID to use for the reflex
     * @param lId ID of the learner in scope
     * @param lps LP evaluation results
     * @param slom SLOM metadata to test
     * @param com Communicator to use for sending the reflex
     * @throws ReflexException Exception being thrown when the reflex aborts the recommendation.
     */
    public static void testLpEvaluations(String mId, String lId, Set<LpEvaluationResult> lps,
            SlomMetadata slom, IEngineCommunicator com) throws ReflexException {
        if (!CollectionUtils.isEmpty(lps)) {
            Set<LpEvaluationResult> microLps = new HashSet<>();
            Set<LpEvaluationResult> macroLps = new HashSet<>();
            for (LpEvaluationResult lp : lps) {
                if (lp.isMacro()) {
                    macroLps.add(lp);
                } else if (lp.isMicro()) {
                    microLps.add(lp);
                }
            }
            handleReflex(new ReflexMicroLpEval(mId, lId, slom.getCourseId(), microLps), com);
            handleReflex(new ReflexMacroLpEval(mId, lId, slom.getCourseId(), macroLps), com);
        }
    }
}
