/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import java.util.Iterator;
import java.util.concurrent.CopyOnWriteArrayList;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Caching instance for SLOM to avoid unneccessary reloading of SLOM metadata.
 *
 * @author fhe
 * @deprecated Replaced by SlomMapCache, because storing/updating operations are seldom and reading
 * operations are quantitative.
 */
public class SlomCache {

    private static final Logger logger = LoggerFactory.getLogger(SlomCache.class);
    private static final SlomCache instance = new SlomCache();
    /**
     * Threadsafe list to store metadata elements. Provides a failsafe iterator to traverse cache
     * content. See
     * http://javarevisited.blogspot.de/2013/02/concurrent-collections-from-jdk-56-java-example-tutorial.html
     * http://java67.blogspot.de/2012/09/what-is-copyonwritearraylist-in-java-example-vs-arraylist.html
     */
    private final CopyOnWriteArrayList<SlomMetadata> cache;

    /**
     * Private constructor for singleton instantiation initializing cache.
     */
    private SlomCache() {
        cache = new CopyOnWriteArrayList<>();
    }

    /**
     * Returns singleton instance of the cache to share ressources across different worker
     * instances.
     *
     * @return Singleton instance of this class
     */
    public static SlomCache getInstance() {
        return instance;
    }

    /**
     * Returns the SLOM metadata for the given LO ID, if present in the cache.
     *
     * @param loId ID of the LO in scope
     * @return Null if not present in cache
     */
    public SlomMetadata getSlomMetadata(String loId) {
        Iterator<SlomMetadata> cacheIterator = cache.listIterator();
        while (cacheIterator.hasNext()) {
            SlomMetadata current = cacheIterator.next();
            for (KnowledgeObject ko : current.getKos()) {
                if (ko.getLinkedLmsElement().equals(loId)) {
                    return current;
                }
            }
        }
        return null;
    }

    /**
     * Stores the given SLOM object if not Null.
     *
     * @param slom SLOM object to store in cache
     */
    public void store(SlomMetadata slom) {
        if (slom != null) {
            if (cache.add(slom)) {
                logger.debug("Stored metadata in SLOM cache: {}", slom.getCcm().getIriAsString());
            }
        }
    }

    /**
     * Returns the current cache size.
     *
     * @return Integer value representing the number of SLOM objects in the cache
     */
    public int getSize() {
        return cache.size();
    }

    /**
     * Clears the cache.
     */
    public void clear() {
        cache.clear();
        logger.debug("Cleared metadata from SLOM cache.");
    }

    /**
     * Removes the given SLOM object from the cache.
     *
     * @param slom SLOM object to remove
     */
    public void remove(SlomMetadata slom) {
        if (cache.remove(slom)) {
            logger.debug("Removed metadata from SLOM cache: {}", slom.getCcm().getIriAsString());
        }
    }

    /**
     * Returns an Iterator to access the stores SLOM metadata.
     *
     * @return Cache Iterator
     */
    public Iterator<SlomMetadata> getSlomMetadata() {
        return cache.iterator();
    }
}
