/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 29.07.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.lpmProcessor.OntoMngr;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Collection;
import java.util.Map;
import org.apache.commons.lang3.tuple.Pair;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLDataFactory;
import org.semanticweb.owlapi.model.OWLNamedIndividual;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Util to extend SLOM data with specific axioms for an easier conversion of existing models.
 *
 * @author Florian Heberle
 */
public class SlomExtenderUtil {

    private static final Logger LOGGER = LoggerFactory.getLogger(SlomExtenderUtil.class);

    /**
     * Extends the CM and CCM of the given SLOM data with LP starters,
     * i.e. LOs that are first elements in a LP are defined as subclass of "FirstPathwayElement".
     *
     * @param cm CM to be extended
     * @param macroLps Macro LPs of the CM
     * @param ccm CCM to be extended
     * @param microLps Micro LPs of the CCM
     * @return Extended pair of CM (left) and CCM (right), null on error.
     */
    public static Pair<OWLOntology, OWLOntology> extendWithLpStarters(OWLOntology cm, Collection<MacroLearningPathway> macroLps,
            OWLOntology ccm, Collection<MicroLearningPathway> microLps) {
        try {
            if (cm == null || macroLps == null || ccm == null || microLps == null) {
                return null;
            }
            return Pair.of(
                    extendWithLpStarters(cm, macroLps),
                    extendWithLpStarters(ccm, microLps));
        } catch (MalformedURLException | URISyntaxException ex) {
            LOGGER.error("Failed to extend SLOM data with LP starters.", ex);
            return null;
        }
    }

    private static OWLOntology extendWithLpStarters(OWLOntology onto, Collection<? extends LearningPathway> lps) throws MalformedURLException, URISyntaxException {
        OWLOntologyManager manager = OwlUtil.createOwlOntologyManager();
        OWLDataFactory datafac = OwlUtil.createOwlDataFactory(manager);
        OWLClassExpression classExp = datafac.getOWLClass(IRI.create(new URL(OntoMngr.PO.getIriAsString() + "#FirstPathwayElement")));
        for (LearningPathway lp : lps) {
            for (LearningObject lo : lp.getStarters()) {
                for (Map.Entry<OWLNamedIndividual, LearningObject[]> stepEntry : lp.getLpSteps().entrySet()) {
                    if (stepEntry.getValue()[0].equals(lo)) {
                        manager.addAxiom(onto, datafac.getOWLClassAssertionAxiom(classExp, stepEntry.getKey()));
                        break;
                    }
                }
            }
        }
        return onto;
    }

    /**
     * Extends the CM and CCM of the given SLOM data with LP enders,
     * i.e. LOs that are first elements in a LP are defined as subclass of "LastPathwayElement".
     *
     * @param cm CM to be extended
     * @param macroLps Macro LPs of the CM
     * @param ccm CCM to be extended
     * @param microLps Micro LPs of the CCM
     * @return Extended pair of CM (left) and CCM (right), null on error.
     */
    public static Pair<OWLOntology, OWLOntology> extendWithLpEnders(OWLOntology cm, Collection<MacroLearningPathway> macroLps,
            OWLOntology ccm, Collection<MicroLearningPathway> microLps) {
        try {
            if (cm == null || macroLps == null || ccm == null || microLps == null) {
                return null;
            }
            return Pair.of(
                    extendWithLpEnders(cm, macroLps),
                    extendWithLpEnders(ccm, microLps));
        } catch (MalformedURLException | URISyntaxException ex) {
            LOGGER.error("Failed to extend SLOM data with LP starters.", ex);
            return null;
        }
    }

    private static OWLOntology extendWithLpEnders(OWLOntology onto, Collection<? extends LearningPathway> lps) throws MalformedURLException, URISyntaxException {
        OWLOntologyManager manager = OwlUtil.createOwlOntologyManager();
        OWLDataFactory datafac = OwlUtil.createOwlDataFactory(manager);
        OWLClassExpression classExp = datafac.getOWLClass(IRI.create(new URL(OntoMngr.PO.getIriAsString() + "#LastPathwayElement")));
        for (LearningPathway lp : lps) {
            for (LearningObject lo : lp.getEnders()) {
                for (Map.Entry<OWLNamedIndividual, LearningObject[]> stepEntry : lp.getLpSteps().entrySet()) {
                    if (stepEntry.getValue()[1].equals(lo)) {
                        manager.addAxiom(onto, datafac.getOWLClassAssertionAxiom(classExp, stepEntry.getKey()));
                        break;
                    }
                }
            }
        }
        return onto;
    }
}
