/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.factories.SlomFactory;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Caching instance for SLOM to avoid unneccessary reloading of SLOM metadata.
 *
 * @author fhe
 */
public class SlomMapCache {

    private static final Logger logger = LoggerFactory.getLogger(SlomMapCache.class);
    private static final SlomMapCache instance = new SlomMapCache();
    private final ConcurrentHashMap<String, SlomMetadata> map = new ConcurrentHashMap<>();

    /**
     * Constructor for singleton instantiation initializing cache.
     */
    protected SlomMapCache() {
    }

    /**
     * Returns singleton instance of the cache to share ressources across different worker
     * instances.
     *
     * @return Singleton instance of this class
     */
    public static SlomMapCache getInstance() {
        return instance;
    }

    /**
     * Returns the SLOM metadata for the given LO ID, if present in the cache.
     *
     * @param identifier ID of the LO in scope
     * @return Null if not present in cache
     */
    public SlomMetadata getSlomMetadata(String identifier) {
        if (StringUtils.isBlank(identifier)) {
            return null;
        }
        if (map.containsKey(identifier)) {
            return map.get(identifier);
        }
        return null;
    }

    /**
     * Tries to store the given SLOM object in the SLOM cache.
     *
     * @param slom SLOM object to store in cache
     */
    public synchronized void store(SlomMetadata slom) {
        if (slom == null) {
            logger.error("SLOM object to store may never be null.");
        }
        @SuppressWarnings({"null", "ConstantConditions"})
        String testIri = slom.getKos().iterator().next().getIriAsString();
        if (null != getSlomMetadata(testIri)) {
            //Already listed
            return;
        }
        Map<String, SlomMetadata> tmpMap = new HashMap<>();
        for (KnowledgeObject ko : slom.getKos()) {
            tmpMap.put(ko.getLinkedLmsElement(), slom);
            tmpMap.put(ko.getIriAsString(), slom);
        }
        extendIndexMap(tmpMap);
    }

    /**
     * Extends the indexes with the given map. This has been encapsulated for testing purposes.
     *
     * @param extension Map to extend the index map with
     */
    protected void extendIndexMap(Map<String, SlomMetadata> extension) {
        map.putAll(extension);
    }

    /**
     * Clears the cache.
     */
    public void clear() {
        map.clear();
    }

    /**
     * Returns the current cache size.
     *
     * @return Integer value representing the number of SLOM objects in the cache
     */
    public int getSize() {
        return (new HashSet<>(map.values())).size();
    }

    /**
     * Update SLOM data in cache.
     *
     * @param iri CM to be updated
     */
    public synchronized void updateCm(String iri) {
        try {
            List<SlomMetadata> slomList = getSlomFromCmIri(iri);
            for (SlomMetadata slom : slomList) {
                String cmIri = iri;
                String ccmIri = slom.getCcm().getIriAsString();
                remove(slom);
                store(SlomFactory.createNewInstance(new URL(ccmIri), new URL(cmIri)));
            }
        } catch (MalformedURLException ex) {
            logger.error(String.format("Update of SLOM cache for CM %s failed.", iri), ex);
        }
    }

    /**
     * Update SLOM data in cache.
     *
     * @param iri CCM to be updated
     */
    public synchronized void updateCcm(String iri) {
        try {
            SlomMetadata slom = getSlomFromCcmIri(iri);
            if (slom == null) {
                return;
            }
            String ccmIri = iri;
            String cmIri = slom.getCm().getIriAsString();
            remove(slom);
            store(SlomFactory.createNewInstance(new URL(ccmIri), new URL(cmIri)));
        } catch (MalformedURLException ex) {
            logger.error(String.format("Update of SLOM cache for CCM %s failed.", iri), ex);
        }
    }

    /**
     * Retrieves the SLOM object for the given CCM iri.
     *
     * @param iri IRI of the CCM in scope
     * @return Null in case CCM is not yet listed.
     */
    private SlomMetadata getSlomFromCcmIri(String iri) {
        if (StringUtils.isBlank(iri)) {
            return null;
        }
        for (SlomMetadata slom : new HashSet<>(map.values())) {
            if (slom.getCcm().getIriAsString().equals(iri)) {
                return slom;
            }
        }
        return null;
    }

    /**
     * Retrieves the SLOM objects for the given CM iri.
     *
     * @param iri IRI of the CM in scope
     * @return Null in case CM is not yet listed.
     */
    private List<SlomMetadata> getSlomFromCmIri(String iri) {
        if (StringUtils.isBlank(iri)) {
            return null;
        }
        List<SlomMetadata> slomList = new ArrayList<>();
        for (SlomMetadata slom : new HashSet<>(map.values())) {
            if (slom.getCm().getIriAsString().equals(iri)) {
                slomList.add(slom);
            }
        }
        return slomList;
    }

    /**
     * Removes the given SLOM object from the cache.
     *
     * @param slom SLOM object to be removed.
     */
    private void remove(SlomMetadata slom) {
        for (KnowledgeObject ko : slom.getKos()) {
            map.remove(ko.getLinkedLmsElement());
            map.remove(ko.getIriAsString());
        }
    }
}
