/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-04-03
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.lpmGeneralInterfaces.ILmsCommunicator;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import static eu.intuitel.lpmProcessor.utils.DbAccessUtil.saveParseInt;

import eu.intuitel.types.*;

import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.PatternSyntaxException;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Util to get USE data from the LMS and store it in the DB.
 *
 * @author fhe
 */
public class UserScoreExtractionUtil {

    private static final Logger LOGGER = LoggerFactory.getLogger(UserScoreExtractionUtil.class);
    private static final long UPDATEINTERVAL = 480000;
    private static final Map<String, Long> UPDATEMAP = new HashMap<>();
    private static final String PERF_TYPE_GRADE = "grade";
    private static final String PERF_TYPE_COMPLETION = "completion";
    private static final String PERF_TYPE_SEENPERCENTAGE = "seenPercentage";
    private static final String PERF_TYPE_ACCESSED = "accessed";
    private static final String PERF_TYPE_INTERNAL = "internal";
    private static final String PERF_VALUE_LOID_UNKOWN = "loIdUnknown";
    private static final String PERF_VALUE_LEARNER_UNKOWN = "learnerUnknown";
    private static final String ENV_DATA_LEARNER_NAME = "lName";
    private static final String ENV_DATA_LEARNER_GENDER = "lGender";
    private static final String ENV_DATA_LEARNER_AGE = "lAge";
    private static final String ENV_DATA_LEARNER_CULTURE = "lCulture";
    private static final String ENV_DATA_LEARNER_ATTITUDE = "lAttitude"; //TODO: To be added to USE
    private static final String ENV_DATA_LEARNER_INTERACTION = "lInteraction"; //TODO: To be added to USE
    private static final String ENV_DATA_LEARNER_DEAFNESS = "lDeafness"; //TODO: To be added to USE
    private static final String ENV_DATA_LEARNER_BLINDNESS = "lBlindness"; //TODO: To be added to USE
    private static final String ENV_DATA_LEARNER_MUTENESS = "lMuteness"; //TODO: To be added to USE
    private static final String ENV_DATA_LEARNER_LANGUAGES = "lLanguages"; //TODO: To be added to USE
    private static final String ENV_DATA_ENVIRONMENT_NOISE = "eNoise";
    private static final String ENV_DATA_ENVIRONMENT_TIME = "eTime";
    private static final String ENV_DATA_DEVICE_TYPE = "dType";
    private static final String ENV_DATA_DEVICE_CONNECTION_TYPE = "dConType";
    private static final String ENV_DATA_DEVICE_CONNECTION_STABILITY = "dConStab";
    private static final String ENV_DATA_DEVICE_RESOLUTION = "dRes";
    private static final String ENV_DATA_DEVICE_PIXEL_DENSITY = "dDensity"; //TODO: To be added to USE
    private static final String ENV_DATA_DEVICE_BATTERY = "dBattery";
    private LearnerProfileData lpData;
    private SessionData sData;
    private EnvData envData;

    /**
     * Retrieves and stores USE performance data.
     *
     * @param learnerId ID of the learner in scope
     * @param currentLoId LMS-ID of the KO in scope
     * @param currentKo KO representation as contained in the respective SLOM object
     * @param slom SLOM data of the course
     * @param lmsCom LMS communication interface
     * @param db DB communication interface
     * @return False if something goes wrong.
     */
    public boolean getAndStoreUsePerfData(String learnerId, String currentLoId, KnowledgeObject currentKo, SlomMetadata slom,
            ILmsCommunicator lmsCom, DbAccessUtil db) {
        try {
            Set<String> loIdsToInspect = new HashSet<>();
            String previousLoId = getLoIdOfPreviousKo(db);
            if (!StringUtils.isBlank(previousLoId)) {
                loIdsToInspect.add(previousLoId);
            }
            if (isTest(currentKo)) {
                loIdsToInspect.add(currentLoId);
            }
            for (String loIdToInspect : loIdsToInspect) {
                Map<String, Map<String, String>> usePerf = lmsCom.getUsePerformanceData(learnerId, loIdToInspect);
                if (MapUtils.isNotEmpty(usePerf)) {
                    for (String loId : usePerf.keySet()) {
                        if (null == slom.getKoByLmsId(loId)) {
                            LOGGER.debug(String.format("Received UsePerf data for unknown KO with lmsId %s.", loId));
                        } else {
                            Map<String, String> usePerfData = usePerf.get(loId);
                            String iri = slom.getKoByLmsId(loId).getIriAsString();
                            checkInternalMessages(usePerfData);
                            Integer grade = getGrade(usePerfData);
                            Integer seenPercentage = getSeenPercentage(usePerfData);
                            Integer textCompletion = getTextCompletion(usePerfData);
                            Boolean accessed = getAccessed(usePerfData);
                            CompletionUtil.updateKoCompletionValue(iri, accessed, grade, seenPercentage, textCompletion, db);
                        }
                    }
                }
            }
            return true;
        } catch (Exception ex) {
            LOGGER.error(String.format("Error while requesting USE performance data for LO '%s'.", currentLoId), ex);
            return false;
        }
    }

    /**
     * Determines the ID of the KO that has been accessed before the current one.
     *
     * @param db DB communication interface
     * @return ID of the previously accessed KO.
     */
    private String getLoIdOfPreviousKo(DbAccessUtil db) {
        List<KnowledgeObject> koList = db.getHistoryOfKOs(2);
        if (CollectionUtils.isEmpty(koList)) {
            return null;
        }
        if (koList.size() == 2) {
            return koList.get(1).getLinkedLmsElement();
        }
        return null;
    }

    /**
     * Determines whether the currently viewed KO is a test.
     *
     * @param ko Currently viewed KO
     * @return True if the current KO is a test.
     */
    private boolean isTest(KnowledgeObject ko) {
        return ko.isAutomaticallyAssessedTest();
    }

    private void checkInternalMessages(Map<String, String> usePerfData) {
        String value = usePerfData.get(PERF_TYPE_INTERNAL);
        if (!StringUtils.isBlank(value)) {
            switch (value) {
                case PERF_VALUE_LOID_UNKOWN:
                    LOGGER.warn("Requested USE performance data for unknown LoID.");
                    break;
                case PERF_VALUE_LEARNER_UNKOWN:
                    LOGGER.warn("Requested USE performance data for unknown learner.");
                    break;
                default:
                    LOGGER.warn("Unknown USE performance internal value: " + value);
                    break;
            }
        }
    }

    private Integer getGrade(Map<String, String> usePerfData) {
        String value = usePerfData.get(PERF_TYPE_GRADE);
        if (!StringUtils.isBlank(value)) {
            return Integer.valueOf(value);
        }
        return null;
    }

    private Integer getSeenPercentage(Map<String, String> usePerfData) {
        String value = usePerfData.get(PERF_TYPE_SEENPERCENTAGE);
        if (!StringUtils.isBlank(value)) {
            return Integer.valueOf(value);
        }
        return null;
    }

    private Integer getTextCompletion(Map<String, String> usePerfData) {
        String value = usePerfData.get(PERF_TYPE_COMPLETION);
        if (!StringUtils.isBlank(value)) {
            return Integer.valueOf(value);
        }
        return null;
    }

    private Boolean getAccessed(Map<String, String> usePerfData) {
        String value = usePerfData.get(PERF_TYPE_ACCESSED);
        if (!StringUtils.isBlank(value)) {
            return Boolean.valueOf(value);
        }
        return null;
    }

    /**
     * Retrieves and stores USE environment data.
     *
     * @param learnerId ID of the learner in scope
     * @param lmsCom LMS communication interface
     * @param db DB communication interface
     * @return False if something goes wrong.
     */
    public boolean getAndStoreUseEnvData(String learnerId, ILmsCommunicator lmsCom, DbAccessUtil db) {
        if (!useEnvUpdateNeeded(learnerId)) {
            return true;
        }
        Map<String, String> useEnv = lmsCom.getUseEnvironmentData(learnerId);

        if (useEnv == null) {
            return false;
        } else if (useEnv.isEmpty()) {
            return true;
        }
        UPDATEMAP.put(learnerId, new Date().getTime());

        lpData = new LearnerProfileData(
                useEnv.get(ENV_DATA_LEARNER_NAME),
                useEnv.get(ENV_DATA_LEARNER_GENDER),
                useEnv.get(ENV_DATA_LEARNER_AGE),
                useEnv.get(ENV_DATA_LEARNER_CULTURE),
                useEnv.get(ENV_DATA_LEARNER_DEAFNESS),
                useEnv.get(ENV_DATA_LEARNER_BLINDNESS),
                useEnv.get(ENV_DATA_LEARNER_MUTENESS),
                useEnv.get(ENV_DATA_LEARNER_LANGUAGES));
        if (!db.setLearnerProfileData(lpData)) {
            return false;
        }

        sData = new SessionData(
                useEnv.get(ENV_DATA_DEVICE_TYPE),
                useEnv.get(ENV_DATA_DEVICE_PIXEL_DENSITY),
                useEnv.get(ENV_DATA_DEVICE_RESOLUTION),
                useEnv.get(ENV_DATA_ENVIRONMENT_TIME),
                useEnv.get(ENV_DATA_LEARNER_ATTITUDE),
                useEnv.get(ENV_DATA_LEARNER_INTERACTION));
        if (!db.setSessionData(sData)) {
            return false;
        }

        envData = new EnvData(
                useEnv.get(ENV_DATA_DEVICE_BATTERY),
                useEnv.get(ENV_DATA_DEVICE_CONNECTION_TYPE),
                useEnv.get(ENV_DATA_DEVICE_CONNECTION_STABILITY),
                useEnv.get(ENV_DATA_ENVIRONMENT_NOISE));
        return db.setEnvironmentData(envData);
    }

    public class LearnerProfileData {

        public final String name;
        public final Gender gender;
        public final Integer age;
        public final Culture culture;
        public final Boolean deafness;
        public final Boolean blindness;
        public final Boolean muteness;
        public final Language[] languages;

        /**
         * Constructor.
         *
         * @param age Age of the learner in scope
         * @param gender Gender of the learner in scope
         * @param name Name of the learner in scope
         * @param culture The culture of the learner
         * @param deafness Deafness of the learner in scope
         * @param blindness Blindness of the learner in scope
         * @param muteness Muteness of the learner in scope
         * @param languages Comma-separated list of language codes
         */
        public LearnerProfileData(String name, String gender, String age, String culture, String deafness, String blindness, String muteness, String languages) {
            this.name = name;
            this.culture = Culture.getCulture(culture);
            this.age = DbAccessUtil.saveParseInt(age);
            this.gender = Gender.getGender(gender);
            this.deafness = "true".equalsIgnoreCase(deafness) || "1".equals(deafness);
            this.blindness = "true".equalsIgnoreCase(blindness) || "1".equals(blindness);
            this.muteness = "true".equalsIgnoreCase(muteness) || "1".equals(muteness);
            this.languages = Language.toArray(languages);
        }
    }

    public class SessionData {

        public final String deviceType;
        public final String deviceDpi;
        public final Integer width;
        public final Integer height;
        public final Long deviceTime;
        public final InteractionWillingness willingness;
        public final LearningAttitude attitude;

        /**
         * Constructor.
         *
         * @param deviceType Type of device the learner is using
         * @param deviceDpi DPI of the device the learner is using
         * @param deviceResolution Resulution of the device the learner is using
         * @param environmentTime Local time as set on the device of the learner
         * @param attitude The current learning attitude of the learner
         * @param interactionWillingness The current interaction willingness of the learner
         */
        public SessionData(String deviceType, String deviceDpi, String deviceResolution, String environmentTime, String interactionWillingness, String attitude) {
            this.deviceType = deviceType;
            this.deviceDpi = deviceDpi;
            Integer tmpWidth = null;
            Integer tmpHeight = null;
            try {
                if (!StringUtils.isBlank(deviceResolution) && deviceResolution.contains("x")) {
                    String[] splitResolution = deviceResolution.split("x");
                    if (splitResolution.length == 2) {
                        tmpWidth = saveParseInt(splitResolution[0]);
                        tmpHeight = saveParseInt(splitResolution[1]);
                    }
                }
            } catch (PatternSyntaxException ex) {
            }
            width = tmpWidth;
            height = tmpHeight;
            deviceTime = DbAccessUtil.saveParseLong(environmentTime);
            willingness = InteractionWillingness.convertToEnum(interactionWillingness);
            this.attitude = LearningAttitude.getLearningAttitude(attitude);
        }
    }

    public class EnvData {

        public final Integer battery;
        public final ConnectionType conType;
        public final Integer conStab;
        public final Integer noise;

        /**
         * Constructor.
         *
         * @param battery Current battery level of the device the learner is using
         * @param conType Type of connection with which the learner is connected to the network
         * @param conStab Stability of connection with which the learner is connected to the
         * network
         * @param noise Current noise level in which the learning session takes place
         */
        public EnvData(String battery, String conType, String conStab, String noise) {
            this.battery = saveParseInt(battery);
            this.conType = ConnectionType.getConnectionType(conType);
            this.conStab = saveParseInt(conStab);
            this.noise = saveParseInt(noise);
        }

    }

    /**
     * Checks whether a USE environment update is needed. This is triggered when between the current
     * and last USE environment request is a timespan larger than a fixed amount of time.
     *
     * @param learnerId ID of the learner in scope
     * @return True if an update is needed.
     */
    private boolean useEnvUpdateNeeded(String learnerId) {
        //NOTE: Concurrency is here less relevant. 
        //1) We assume that the learner only works on one course at a time
        //2) Learner updates should (in a normal learning situation) not be send so fast that there is an overlap for one learner
        //3) Even if USE requests are send twice, there are no problems resulting from that

        Long storedTime = UPDATEMAP.get(learnerId);
        if (storedTime == null) {
            return true;
        }
        return new Date().getTime() - storedTime >= UPDATEINTERVAL;
    }

    /**
     * @return The received learner profile data - i.e. static learner data
     */
    public LearnerProfileData getLearnerProfileData() {
        return lpData;
    }

    /**
     * @return The recevied learner session data - i.e. static session data
     */
    public SessionData getSessionData() {
        return sData;
    }

    /**
     * @return The received learner session data - i.e. time-dependent session data
     */
    public EnvData getEnvData() {
        return envData;
    }
}
