/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Contributor: Ovidiu Serban, University of Reading, www.isr.reading.ac.uk
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor;

import eu.intuitel.db.access.DbAccessException;
import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.LearnerStateOntology;
import eu.intuitel.lpmProcessor.test.mocks.ComEngineMock;
import eu.intuitel.lpmProcessor.test.mocks.ComLmsMock;
import eu.intuitel.lpmProcessor.test.mocks.ComSlomMock;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData01;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import org.apache.commons.lang3.StringUtils;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.io.StringDocumentSource;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.semanticweb.owlapi.reasoner.OWLReasoner;
import org.semanticweb.owlapi.reasoner.OWLReasonerFactory;
import org.semanticweb.owlapi.reasoner.ReasonerInterruptedException;
import org.semanticweb.owlapi.reasoner.TimeOutException;
import org.semanticweb.owlapi.reasoner.structural.StructuralReasonerFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Tests for LPM Worker functionality with communication interface mockups.
 *
 * @author fhe
 */
public class LpmProcessorTest {

    private static final Logger LOGGER = LoggerFactory.getLogger(LpmProcessorTest.class);
    private static final String MID = "42";
    private SlomMetadata slom;
    private Integer lIdCtr;
    private String koLmsId;
    private MacroLearningPathway macroLp;
    private MicroLearningPathway microLp;
    private ComSlomMock slomMock;
    private List<String> lIds;
    private Map<String, ComLmsMock> lmsMocks;
    private Map<String, ComEngineMock> engineMocks;

    @BeforeClass
    public static void beforeClass() {
        LOGGER.info("Starting tests.");
    }

    @AfterClass
    public static void afterClass() {
        LOGGER.info("Finished tests.");
        TestCommons.clearDataStorage();
    }

    @Before
    public void beforeTest() {
        lIdCtr = 0;
        lIds = new ArrayList<>();

        SlomTestData01 slomData = SlomTestData01.getInstance();
        slom = slomData.getSlom();
        koLmsId = slomData.getKo01().getLinkedLmsElement();
        microLp = slomData.getMicroLp01();
        macroLp = slomData.getMacroLp01();

        slomMock = new ComSlomMock(slomData.getCcmFilePath(), slomData.getCmFilePath());
        lmsMocks = new HashMap<>();
        engineMocks = new HashMap<>();

        DbAccessor accessor = TestCommons.getAccessor();
        accessor.storeMicroLpData(microLp.getIriAsString(), slom.getCourseId(), null, null);
        accessor.storeMacroLpData(macroLp.getIriAsString(), slom.getCourseId(), null, null);
    }

    @Test
    public void test01() throws DbAccessException, Exception {
// For debugging of cached ontologies:        
//        try {
//            OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
//            org.semanticweb.owlapi.model.IRI localIri = IRI.create(new java.io.File("C:\\Users\\Florian Heberle\\Desktop\\lso_input4.txt"));
//            OWLOntology onto = manager.loadOntologyFromOntologyDocument(localIri);
//        } catch(OWLOntologyCreationException ex) {
//            System.out.println(ex);
//            
//        }
        testLpmProcessing(1, 1, "01_");
        
        //For DataIncomplete reflex tests:
        //lIdCtr--;
        //testLpmProcessing(1, 1, "01_");
    }

    @Test
    public void test02() throws DbAccessException, Exception {
        testLpmProcessing(1, 2, "02_");
    }

    @Test
    public void test03() throws DbAccessException, Exception {
        testLpmProcessing(1, 10, "03_");
    }

    @Test
    public void test04() throws DbAccessException, Exception {
        testLpmProcessing(5, 1, "04_");
    }

    @Test
    public void test05() throws DbAccessException, Exception {
        testLpmProcessing(5, 5, "05_");
    }

    @Test
    public void test06() throws DbAccessException, Exception {
        testLpmProcessing(5, 10, "06_");
    }

    @Test
    public void test07() throws DbAccessException, Exception {
        testLpmProcessing(5, 100, "07_");
    }

//    @Test
//    public void test08() throws DbAccessException {
//        testLpmProcessing(25, 500, "08_");
//    }
    /**
     * Carries out LPM processing in a given number of threads with a fixed number of threads to
     * execute in parallel and tests the results.
     *
     * @param threadPoolSize Count of threads to be executed simultaneously
     * @param numberOfLpmProcessors Count of LpmProcessors to execute
     * @param prefix Prefix to be used for learner IDs
     * @throws eu.intuitel.db.access.DbAccessException
     */
    public void testLpmProcessing(int threadPoolSize, int numberOfLpmProcessors, String prefix) throws DbAccessException, Exception {
        Long start = System.currentTimeMillis();
        LOGGER.debug(String.format("Start lpm parallel processing test, poolsize=%s, processorCount=%s", threadPoolSize, numberOfLpmProcessors));

        List<Callable<TestRunner>> runners = new ArrayList<>();
        ExecutorService service = Executors.newFixedThreadPool(threadPoolSize);
        for (int i = 0; i < numberOfLpmProcessors; i++) {
            addNewProcessor(runners, prefix);
        }

        List<Future<TestRunner>> fs = new ArrayList<>();
        try {
            fs = service.invokeAll(runners, runners.size(), TimeUnit.MINUTES);
            assertFalse("List of thread results is empty.", fs.isEmpty());
        } catch (InterruptedException ex) {
            assertNull("Threads have been interrupted before processors finished.", ex);
        }
        for (Future<TestRunner> f : fs) {
            assertTrue("Preprocessing job incomplete.", f.isDone());
        }

        assertEquals("Generated number of runners is invalid.", numberOfLpmProcessors, lIds.size());
        for (String lId : lIds) {
            ComEngineMock eMock = engineMocks.get(lId);
            assertNotNull("Engine mock was null for learner " + lId);
            testProcessingResult(eMock.getLso(), lId);
        }

        service.shutdown();
        LOGGER.debug(String.format("Finished parallel processing test after %s ms.", System.currentTimeMillis() - start));
    }

    /**
     * Adds new instances to the given lists.
     *
     * @param runners List of TestRunners to be extended
     * @param prefix Prefix to be used for learner IDs
     */
    private void addNewProcessor(List<Callable<TestRunner>> runners, String prefix) throws DbAccessException {
        String lId = prefix + lIdCtr.toString();
        lIds.add(lId);
        ComEngineMock engineMock = new ComEngineMock();
        engineMocks.put(lId, engineMock);
        ComLmsMock lmsMock = new ComLmsMock();
        lmsMocks.put(lId, lmsMock);

        DbAccessor dbAccessor = TestCommons.getAccessor();
        dbAccessor.storeMicroLpSelection(lId, microLp.getIriAsString(), slom.getCourseId());
        dbAccessor.storeMacroLpSelection(lId, macroLp.getIriAsString(), slom.getCourseId());

        LpmProcessor processor = new LpmProcessor(koLmsId, lId, MID, new Date().getTime(),
                slomMock, engineMock, lmsMock, dbAccessor, false);

        runners.add(new TestRunner(processor));
        lIdCtr++;
    }

    /**
     * Helper-class for threaded testsing.
     */
    private class TestRunner implements Callable<TestRunner> {

        public final LpmProcessor processor;

        public TestRunner(LpmProcessor processor) {
            this.processor = processor;
        }

        @Override
        public TestRunner call() {
            processor.work();
            return this;
        }
    }

    /**
     * Tests the create {@code lso}
     *
     * @param lso Learner-State-Ontology to be tested
     */
    private void testProcessingResult(LearnerStateOntology lso, String number) throws Exception {
        assertNotNull(String.format("LSO %s has not been created.", number), lso);
        assertNotNull(String.format("LSO %s has not been created.", number), lso.getOntoObj());
        assertNotNull(String.format("LSO %s could not be created in Turtle format.", number), lso.toTurtle());
        assertNotNull(String.format("LSO %s could not be created in XML format.", number), lso.toXml());
        assertFalse(String.format("LSO %s String representation is blank.", number), StringUtils.isBlank(lso.serialize()));
        try {
            OWLReasonerFactory reasonerFactory = new StructuralReasonerFactory();
            OWLReasoner reasoner = reasonerFactory.createReasoner(lso.getOntoObj());
            assertTrue(String.format("LSO %s is inconsistent.", number), reasoner.isConsistent());
        } catch (ReasonerInterruptedException | TimeOutException ex) {
            assertNull(String.format("An exception has been thrown while checking LSO %s consistency.", number), ex);
        }

        try {
            OWLOntologyManager mngr = OWLManager.createOWLOntologyManager();
            OWLOntology onto = mngr.loadOntologyFromOntologyDocument(new StringDocumentSource(lso.toXml()));
            assertNotNull(onto);
        } catch (OWLOntologyCreationException ex) {
            assertEquals(null, ex.toString());
        }
    }
}
