/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.lpMetric;

import eu.intuitel.db.access.DbAccessException;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.factories.SlomFactoryTest;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData02;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData03;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Tests for the LP selection via the hypercube approach.
 *
 * @author fhe
 */
public class LpMetricHypercubeTest {

    private static final Logger LOGGER = LoggerFactory.getLogger(LpMetricHypercubeTest.class);
    private static LpMetricHypercube metric;
    private static SlomTestData02 slomData02;
    private static SlomTestData03 slomData03;

    @BeforeClass
    public static void setUpBeforeClass() {
        LOGGER.info("Starting tests.");

        slomData02 = SlomTestData02.getInstance();
        slomData03 = SlomTestData03.getInstance();

        //This test requires that the LOs have ids in a certain order.
        //I.e. Ko01 needs to have id 1, Ko02 needs to have id 2...
        slomData02.getKo01().setNumericId(1);
        slomData02.getKo02().setNumericId(2);
        slomData02.getKo03().setNumericId(3);
        slomData02.getKo04().setNumericId(4);
        slomData02.getKo05().setNumericId(5);
        slomData02.getKo06().setNumericId(6);
        slomData02.getKo07().setNumericId(7);
        slomData02.getKo08().setNumericId(8);

        slomData03.getKo01().setNumericId(1);
        slomData03.getKo02().setNumericId(2);
        slomData03.getKo03().setNumericId(3);
        slomData03.getKo04().setNumericId(4);
        slomData03.getKo05().setNumericId(5);
    }

    @AfterClass
    public static void afterClass() {
        LOGGER.info("Finished tests.");
    }

    @Before
    public void beforeTest() {
        metric = new LpMetricHypercube();
    }

    @Test
    public void testMaskedHammingDistance() {
        assertEquals(0, LpMetricHypercube.getMaskedHammingDistanceFor("1", "1"));
        assertEquals(Integer.MAX_VALUE, LpMetricHypercube.getMaskedHammingDistanceFor("0", "1"));
        assertEquals(0, LpMetricHypercube.getMaskedHammingDistanceFor("1", "0"));
        assertEquals(Integer.MAX_VALUE, LpMetricHypercube.getMaskedHammingDistanceFor("0", "0"));
        assertEquals(0, LpMetricHypercube.getMaskedHammingDistanceFor("01", "1"));
        assertEquals(Integer.MAX_VALUE, LpMetricHypercube.getMaskedHammingDistanceFor("00", "1"));
        assertEquals(1, LpMetricHypercube.getMaskedHammingDistanceFor("10", "1"));
        assertEquals(0, LpMetricHypercube.getMaskedHammingDistanceFor("11", "1"));
        assertEquals(4, LpMetricHypercube.getMaskedHammingDistanceFor("1100100111", "1111111111"));
        assertEquals(2, LpMetricHypercube.getMaskedHammingDistanceFor("1100100111", "0000111111"));
    }

    @Test
    public void testMaskedHammingDistanceParemeterExtremes() {
        assertEquals(Integer.MAX_VALUE, LpMetricHypercube.getMaskedHammingDistanceFor(null, null));
        assertEquals(Integer.MAX_VALUE, LpMetricHypercube.getMaskedHammingDistanceFor(null, ""));
        assertEquals(Integer.MAX_VALUE, LpMetricHypercube.getMaskedHammingDistanceFor(null, "1"));
        assertEquals(Integer.MAX_VALUE, LpMetricHypercube.getMaskedHammingDistanceFor("1", null));
        assertEquals(Integer.MAX_VALUE, LpMetricHypercube.getMaskedHammingDistanceFor("", null));
        assertEquals(Integer.MAX_VALUE, LpMetricHypercube.getMaskedHammingDistanceFor("", ""));
    }

    @Test
    public void testMaskedHammingDistanceForLpWithDifferentLengths() {
        int distLp1 = LpMetricHypercube.getMaskedHammingDistanceFor("111111", "111000");
        int distLp2 = LpMetricHypercube.getMaskedHammingDistanceFor("111100", "111000");
        assertEquals(distLp1, distLp2);

        distLp1 = LpMetricHypercube.getMaskedHammingDistanceFor("101111", "111000");
        distLp2 = LpMetricHypercube.getMaskedHammingDistanceFor("111100", "111000");
        assertEquals(1, distLp1);
        assertEquals(0, distLp2);

        distLp1 = LpMetricHypercube.getMaskedHammingDistanceFor("111111", "111000");
        distLp2 = LpMetricHypercube.getMaskedHammingDistanceFor("101100", "111000");
        assertEquals(0, distLp1);
        assertEquals(1, distLp2);
    }

    @Test
    public void testHcValuesWithSingleLos() {
        List<KnowledgeObject> testVals = new ArrayList<>();
        testVals.add(slomData02.getKo01());
        assertEquals("1", LpMetricHypercube.getHypercubeValue(testVals));
        testVals.clear();
        testVals.add(slomData02.getKo02());
        assertEquals("10", LpMetricHypercube.getHypercubeValue(testVals));
        testVals.clear();
        testVals.add(slomData02.getKo03());
        assertEquals("100", LpMetricHypercube.getHypercubeValue(testVals));
        testVals.clear();
        testVals.add(slomData02.getKo04());
        assertEquals("1000", LpMetricHypercube.getHypercubeValue(testVals));
        testVals.clear();
        testVals.add(slomData02.getKo08());
        assertEquals("10000000", LpMetricHypercube.getHypercubeValue(testVals));
    }

    @Test
    public void testHcValuesWithMultipleLos() {
        List<KnowledgeObject> testVals = new ArrayList<>();
        testVals.add(slomData02.getKo01());
        assertEquals("1", LpMetricHypercube.getHypercubeValue(testVals));
        testVals.add(slomData02.getKo02());
        assertEquals("11", LpMetricHypercube.getHypercubeValue(testVals));
        testVals.add(slomData02.getKo03());
        assertEquals("111", LpMetricHypercube.getHypercubeValue(testVals));
        testVals.add(slomData02.getKo04());
        assertEquals("1111", LpMetricHypercube.getHypercubeValue(testVals));
    }

    @Test
    public void testHcValueWithGaps() {
        List<KnowledgeObject> testVals = new ArrayList<>();
        testVals.add(slomData02.getKo01());
        testVals.add(slomData02.getKo04());
        assertEquals("1001", LpMetricHypercube.getHypercubeValue(testVals));

        testVals.clear();
        testVals.add(slomData02.getKo04());
        testVals.add(slomData02.getKo03());
        testVals.add(slomData02.getKo02());
        assertEquals("1110", LpMetricHypercube.getHypercubeValue(testVals));

        testVals.clear();
        testVals.add(slomData02.getKo01());
        testVals.add(slomData02.getKo02());
        testVals.add(slomData02.getKo03());
        assertEquals("111", LpMetricHypercube.getHypercubeValue(testVals));
    }

    @Test
    public void testHcValUnordered() {
        List<KnowledgeObject> testVals = new ArrayList<>();
        testVals.add(slomData02.getKo04());
        testVals.add(slomData02.getKo01());
        assertEquals("1001", LpMetricHypercube.getHypercubeValue(testVals));

        testVals.clear();
        testVals.add(slomData02.getKo01());
        testVals.add(slomData02.getKo03());
        testVals.add(slomData02.getKo02());
        assertEquals("111", LpMetricHypercube.getHypercubeValue(testVals));

        testVals.clear();
        testVals.add(slomData02.getKo02());
        testVals.add(slomData02.getKo03());
        testVals.add(slomData02.getKo01());
        assertEquals("111", LpMetricHypercube.getHypercubeValue(testVals));

        testVals.clear();
        testVals.add(slomData02.getKo03());
        testVals.add(slomData02.getKo02());
        testVals.add(slomData02.getKo01());
        assertEquals("111", LpMetricHypercube.getHypercubeValue(testVals));
    }

    @Test
    public void testHcValParameterExtremes() {
        List<KnowledgeObject> testVals = new ArrayList<>();
        assertEquals("", LpMetricHypercube.getHypercubeValue(testVals));
        assertEquals("", LpMetricHypercube.getHypercubeValue(null));
    }

    @Test
    public void testGetLp01() throws DbAccessException {
        List<LearningObject> completedLos = new ArrayList<>();
        completedLos.add(slomData02.getKo01());
        completedLos.add(slomData02.getKo02());
        Collection<? extends LearningPathway> resultLps = metric.getLp(slomData02.getSlom(), completedLos, MicroLearningPathway.class);
        assertNotNull("LP collection has falsely been returned as null.", resultLps);
        assertFalse("LP has falsely been returned with no elements.", resultLps.isEmpty());
        testForLpsInResult(slomData02.getSlom(), MicroLearningPathway.class, resultLps, new String[]{"#HasMicroLpRelation1"});
    }

    @Test
    public void testGetLp02() throws DbAccessException {
        List<LearningObject> completedLos = new ArrayList<>();
        completedLos.add(slomData02.getKo01());
        completedLos.add(slomData02.getKo03());
        Collection<? extends LearningPathway> resultLps = metric.getLp(slomData02.getSlom(), completedLos, MicroLearningPathway.class);
        assertNotNull("LP has falsely been returned as null.", resultLps);
        assertFalse("LP has falsely been returned with no elements.", resultLps.isEmpty());
        testForLpsInResult(slomData02.getSlom(), MicroLearningPathway.class, resultLps, new String[]{"#HasMicroLpRelation2"});
    }

    @Test
    public void testGetLpWithAlternative01() throws DbAccessException {
        List<LearningObject> completedLos = new ArrayList<>();
        completedLos.add(slomData03.getKo01());
        completedLos.add(slomData03.getKo02());
        Collection<? extends LearningPathway> resultLps = metric.getLp(slomData03.getSlom(), completedLos, MicroLearningPathway.class);
        assertNotNull("LP has falsely been returned as null.", resultLps);
        assertFalse("LP has falsely been returned with no elements.", resultLps.isEmpty());
        testForLpsInResult(slomData03.getSlom(), MicroLearningPathway.class, resultLps, new String[]{"#HasMicroLpRelation1"});
    }

    @Test
    public void testGetLpWithAlternative02() throws DbAccessException {
        List<LearningObject> completedLos = new ArrayList<>();
        completedLos.add(slomData03.getKo01());
        completedLos.add(slomData03.getKo03());
        Collection<? extends LearningPathway> resultLps = metric.getLp(slomData03.getSlom(), completedLos, MicroLearningPathway.class);
        assertNotNull("LP has falsely been returned as null.", resultLps);
        assertFalse("LP has falsely been returned with no elements.", resultLps.isEmpty());
        testForLpsInResult(slomData03.getSlom(), MicroLearningPathway.class, resultLps, new String[]{"#HasMicroLpRelation1"});
    }

    @Test
    public void testGetLpWithAlternative03() throws DbAccessException {
        List<LearningObject> completedLos = new ArrayList<>();
        completedLos.add(slomData03.getKo01());
        completedLos.add(slomData03.getKo04());
        Collection<? extends LearningPathway> resultLps = metric.getLp(slomData03.getSlom(), completedLos, MicroLearningPathway.class);
        assertNotNull("LP has falsely been returned as null.", resultLps);
        assertFalse("LP has falsely been returned with no elements.", resultLps.isEmpty());
        testForLpsInResult(slomData03.getSlom(), MicroLearningPathway.class, resultLps, new String[]{"#HasMicroLpRelation2"});
    }

    /**
     * Tests if the correct number and instances of LPs are contained in the given
     * {@code resultLps}.
     *
     * @param slom SLOM data in containing the raw data
     * @param lpClass Class of LPs in scope
     * @param resultLps Set of results from the LP-Metric
     * @param lpFragments URI-Fragments of expected LPs
     */
    public static void testForLpsInResult(SlomMetadata slom, Class<? extends LearningPathway> lpClass,
            Collection<? extends LearningPathway> resultLps, String[] lpFragments) {
        assertEquals("An incorrect number of LPs has been returned.", lpFragments.length, resultLps.size());
        for (String lpFragment : lpFragments) {
            assertTrue("An expected LP is missing in resultset.", resultLps.contains(SlomFactoryTest.getLp(slom, lpFragment, lpClass)));
        }
    }
}
