/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-03-14
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.factories;

import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeType;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MediaType;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData01;
import java.util.Locale;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.BeforeClass;
import org.junit.Test;
import org.semanticweb.owlapi.model.OWLClass;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Tests for the conversion methods of the KO factory class.
 *
 * @author fhe
 */
public class KnowledgeObjectFactoryTest {

    private static final Logger LOGGER = LoggerFactory.getLogger(KnowledgeObjectFactoryTest.class);
    private final TestFactory factory = new TestFactory();

    @BeforeClass
    public static void beforeClass() {
        LOGGER.info("Starting tests.");
    }

    @AfterClass
    public static void afterClass() {
        LOGGER.info("Finished tests.");
    }

    /**
     * Test for propper KO property extraction.
     */
    @Test
    public void testProperties() {
        SlomTestData01 slomData = SlomTestData01.getInstance();
        testObjectProperties(slomData.getKo01(), KnowledgeType.ABSTRACTORIENTATION, MediaType.TEXTPRESENTATION, 49.01572, 8.38913, 100);
        assertEquals("KO property 'hasDifficultyLevel' was incorrect.", new Integer(2), slomData.getKo01().getDifficultyLevel());
        assertEquals("KO property 'hasEqfLevel' was incorrect.", new Integer(2), slomData.getKo01().getEqfLevel());
        assertEquals("KO property 'hasMinimalHeight' was incorrect.", new Integer(2), slomData.getKo01().getMinimalHeight());
        assertEquals("KO property 'hasMinimalWidth' was incorrect.", new Integer(2), slomData.getKo01().getMinimalWidth());
        assertEquals("KO property 'hasRecommendedAge' was incorrect.", new Integer(2), slomData.getKo01().getRecommendedAge());
        assertEquals("KO property 'hasRecommendedHeight' was incorrect.", new Integer(2), slomData.getKo01().getRecommededHeight());
        assertEquals("KO property 'hasRecommendedWidth' was incorrect.", new Integer(2), slomData.getKo01().getRecommendedWidth());
        assertEquals("KO property 'hasSize' was incorrect.", new Integer(2), slomData.getKo01().getDigitalSize());
        assertEquals("KO property 'hasSubtitles' was incorrect.", Boolean.TRUE, slomData.getKo01().getSubtiles());
        assertEquals("KO property 'isSuitableForBlind' was incorrect.", Boolean.TRUE, slomData.getKo01().getSuitableForBlind());
        assertEquals("KO property 'isSuitableForDeaf' was incorrect.", Boolean.TRUE, slomData.getKo01().getSuitableForDeaf());
        assertEquals("KO property 'isSuitableForMute' was incorrect.", Boolean.TRUE, slomData.getKo01().getSuitableForMute());
        assertEquals("KO property 'isLinkedWithLmsElement' was incorrect.", "case01_Ko01", slomData.getKo01().getLinkedLmsElement());
        assertEquals("KO property 'hasLanguage' was incorrect.", Locale.ENGLISH, slomData.getKo01().getLanguage());
        assertEquals("KO property 'hasEstimatedLearningTime' was incorrect.", new Long(42000), slomData.getKo01().getEstimatedLearningTime());
    }

    /**
     * Tests if the given object properties are assigend to the given KO.
     *
     * @param ko KO to be tested
     * @param kt Expected KT class of the KO
     * @param mt Expected MT class of the KO
     * @param longitude Longitude of the preferred location
     * @param latitude Latitude of the preferred location
     * @param margin Margin of the preferred location
     */
    private void testObjectProperties(KnowledgeObject ko, OWLClass kt, OWLClass mt, double longitude, double latitude, int margin) {
        assertEquals(String.format("Test for KT '%s' of KO '%s' failed.", kt, ko.toString()), kt, ko.getKnowledgeType().getType());
        assertEquals(String.format("Test for MT '%s' of KO '%s' failed.", mt, ko.toString()), mt, ko.getMediaType().getType());
        assertEquals(String.format("Test for longitude '%s' of KO '%s' failed.", longitude, ko.toString()), longitude, ko.getLocation().getLongitude(), 0.0000001);
        assertEquals(String.format("Test for latitude '%s' of KO '%s' failed.", latitude, ko.toString()), latitude, ko.getLocation().getLatitude(), 0.0000001);
        assertEquals(String.format("Test for location margin '%s' of KO '%s' failed.", margin, ko.toString()), margin, ko.getLocation().getMargin());
    }

    /**
     * Test for valid time string conversion to long.
     */
    @Test
    public void testTimeStringConversion_Valid() {
        testTimeStringConversion(null, null);
        testTimeStringConversion("00:00:00", new Long(0));
        testTimeStringConversion("01:01:01", new Long(3661000));
        testTimeStringConversion("20:30:30", new Long(73830000));
        testTimeStringConversion("24:60:61", new Long(90061000));
        testTimeStringConversion("25:61:61", new Long(93721000));
    }

    /**
     * Test for valid time string conversion to long, invalid input.
     */
    @Test
    public void testTimeStringConversion_Invalid() {
        testTimeStringConversion("", null);
        testTimeStringConversion("foobar", null);
        testTimeStringConversion("12345678", null);
        testTimeStringConversion("aa:bb:cc", null);
        testTimeStringConversion("-1:00:00", null);
    }

    private void testTimeStringConversion(String valueToConvert, Long expected) {
        try {
            Long timeAsLong = factory.public_timeStringToLong(valueToConvert);
            assertEquals("Expected value did not match actual value.", expected, timeAsLong);
        } catch (Exception ex) {
            assertNull(String.format("Exception on value conversion '%s' -> '%s'.", valueToConvert, expected.toString()), ex);
        }
    }

    /**
     * Test for valid locale conversion.
     */
    @Test
    public void testLocaleStringConversion_Valid() {
        testLocaleStringConversion(null, null);
        testLocaleStringConversion("en", Locale.ENGLISH);
        testLocaleStringConversion("en_EN", Locale.ENGLISH);
        //conversion only considers the first two letters.
        testLocaleStringConversion("en151515", Locale.ENGLISH);
        testLocaleStringConversion("en-en", Locale.ENGLISH);
        testLocaleStringConversion("english", Locale.ENGLISH);
    }

    /**
     * Test fpr valid locale conversion, invalid input.
     */
    @Test
    public void testLocaleStringConversion_Invalid() {
        testLocaleStringConversion("", null);
        testLocaleStringConversion("1515", null);
    }

    private void testLocaleStringConversion(String valueToConvert, Locale expected) {
        try {
            Locale convertedValue = factory.public_localeStringToLocale(valueToConvert);
            if (expected == null) {
                assertNull(convertedValue);
            } else {
                assertNotNull(convertedValue);
                assertEquals("Expected value did not match actual value.", expected.getLanguage(), convertedValue.getLanguage());
            }
        } catch (Exception ex) {
            assertNull(String.format("Exception on value conversion '%s' -> '%s'.", valueToConvert, expected), ex);
        }
    }

    /**
     * TestFactory to give access to protected methods to be tested.
     */
    private class TestFactory extends KnowledgeObjectFactory {

        public TestFactory() {
        }

        public Long public_timeStringToLong(String time) {
            return timeStringToLong(time);
        }

        public Locale public_localeStringToLocale(String locale) {
            return localeStringToLocale(locale);
        }
    }
}
