/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Kevin Fuchs, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Ovidiu Serban, University of Reading, www.isr.reading.ac.uk
 * Version: [0.01]
 * Date: 2014-03-14
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.factories;

import eu.intuitel.db.access.DbAccessException;
import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.lpmProcessor.TestCommons;
import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.LearnerStateOntology;
import eu.intuitel.lpmProcessor.models.owlWrapper.ontologies.LearningModelOntology;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData01;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestDataBase;
import eu.intuitel.lpmProcessor.utils.DbAccessUtil;
import java.io.File;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.BeforeClass;
import org.junit.Test;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLDataFactory;
import org.semanticweb.owlapi.model.OWLNamedIndividual;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Tests to see whether the LSO Factory works as intended.
 *
 * @author fhe
 */
public class LearnerStateOntologyFactoryTest {

    private static final Logger LOGGER = LoggerFactory.getLogger(LearnerStateOntologyFactoryTest.class);
    private static final String LID = "LSOLID";

    @BeforeClass
    public static void beforeClass() {
        LOGGER.info("Starting tests.");
    }

    @AfterClass
    public static void afterClass() {
        LOGGER.info("Finished tests.");
        TestCommons.clearDataStorage();
    }

    @Test
    public void testCase01() throws DbAccessException, Exception {
        SlomTestData01 slomData = SlomTestData01.getInstance();
        OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
        OWLDataFactory fac = manager.getOWLDataFactory();

        DbAccessor accessor = TestCommons.getAccessor();
        DbAccessUtil db = new DbAccessUtil(LID, accessor, slomData.getSlom());

        accessor.storeMicroLpData(slomData.getMicroLp01().getIriAsString(), slomData.getSlom().getCourseId(), null, null);
        accessor.storeMacroLpData(slomData.getMacroLp01().getIriAsString(), slomData.getSlom().getCourseId(), null, null);
        accessor.storeKoData(slomData.getKo01().getIriAsString(), slomData.getSlom().getCourseId(), null, null, slomData.getKo01().getLinkedLmsElement());
        accessor.storeCcData(slomData.getCc01().getIriAsString(), slomData.getSlom().getCourseId(), null, null);

        db.addLearningHistoryEntry(slomData.getKo01(), slomData.getCc01(), new Date().getTime());
        accessor.storeCcCompletion(LID, slomData.getSlom().getCourseId(), slomData.getCc01().getIriAsString(), true);
        accessor.storeMicroLpSelection(LID, slomData.getMicroLp01().getIriAsString(), slomData.getSlom().getCourseId());
        accessor.storeMacroLpSelection(LID, slomData.getMacroLp01().getIriAsString(), slomData.getSlom().getCourseId());

        Collection<OWLNamedIndividual> dfs = new HashSet<>();
        dfs.add(fac.getOWLNamedIndividual(IRI.create("http://www.intuitel.eu/public/intui_LM.owl#dfValAge_Adult")));

        LearnerStateOntology lso = LearnerStateOntologyFactory.createLsoRaw(LID, slomData.getSlom(), db, dfs);
        assertNotNull("LSO was returned as null.", lso);
        String xml = lso.toXml();
        assertNotNull("LSO output is blank.", xml);
        assertTrue("LSO output is empty.", xml.length() > 0);

        isCurrentLp(xml, new String[]{
            slomData.getSlom().getCcm().getIriAsString() + "#HasMicroLpRelation1",
            slomData.getSlom().getCm().getIriAsString() + "#HasMacroLpRelation1"
        });
        namedIndividualIsTypeOf(xml, LearningModelOntology.COMPLETE_LO.getIRI().toString(), new String[]{
            slomData.getKo01().getIriAsString(),
            slomData.getCc01().getIriAsString()
        });
        namedIndividualIsTypeOf(xml, LearningModelOntology.UNSEEN_LO.getIRI().toString(), new String[]{
            slomData.getKo02().getIriAsString(),
            slomData.getKo03().getIriAsString(),
            slomData.getKo04().getIriAsString(),
            slomData.getCc01().getIriAsString(),
            slomData.getCc02().getIriAsString()
        });
        namedIndividualIsTypeOf(xml, LearningModelOntology.CURRENT_LO.getIRI().toString(), new String[]{
            slomData.getKo01().getIriAsString(),
            slomData.getCc01().getIriAsString()
        });
        containsDf(xml, "http://www.intuitel.eu/public/intui_LM.owl#dfValAge_Adult");
    }

    private void isCurrentLp(String xml, String... iris) {
        //<owl:Class rdf:about="http://www.intuitel.eu/public/testcase01/intui_CCM.owl#HasMicroLpRelation1">
        //    <rdfs:subClassOf rdf:resource="http://www.intuitel.eu/public/intui_LM.owl#CurrentLearningPathway"/>
        //</owl:Class>
        for (String iri : iris) {
            Pattern pattern = Pattern.compile("owl:Class rdf:about=\""
                    + iri
                    + "\">\\s*<rdfs:subClassOf rdf:resource=\""
                    + LearningModelOntology.CURRENT_LP.getIRI().toString()
                    + "\"/>");
            Matcher matcher = pattern.matcher(xml);
            assertTrue(String.format("CurrentLp statement missing for %s", iri), matcher.find());
        }
    }

    private void containsDf(String xml, String... iris) {
        //<owl:NamedIndividual rdf:about="http://www.intuitel.eu/public/intui_LM.owl#learnerState">
        //    <rdf:type rdf:resource="http://www.intuitel.eu/public/intui_LM.owl#LearnerState"/>
        //    <intui_LMO:hasDidacticFactorValue rdf:resource="http://www.intuitel.eu/public/intui_LM.owl#dfValAge_Adult"/>
        //</owl:NamedIndividual>
        for (String iri : iris) {
            Pattern pattern = Pattern.compile(
                    "<intui_LMO:hasDidacticFactorValue rdf:resource=\""
                    + iri
                    + "\"/>");
            Matcher matcher = pattern.matcher(xml);
            assertTrue(String.format("DF value missing for %s", iri), matcher.find());
        }
    }

    private void namedIndividualIsTypeOf(String xml, String type, String... iris) {
        for (String iri : iris) {
            assertTrue(String.format("%s was not listed as type of %s", iri, type), namedIndividualIsTypeOf(xml, type, iri));
        }
    }

    private boolean namedIndividualIsTypeOf(String xml, String type, String iri) {
        String outerPatternString = String.format("(<owl:NamedIndividual rdf:about=\"%s\">)(.*)(</owl:NamedIndividual>)", iri);
        Pattern outerPattern = Pattern.compile(outerPatternString, Pattern.DOTALL);
        Matcher outerMatcher = outerPattern.matcher(xml);
        while (outerMatcher.find()) {
            Pattern innerPattern = Pattern.compile("(<rdf:type rdf:resource=\")(" + type + ")(\"/>)");
            Matcher innerMatcher = innerPattern.matcher(outerMatcher.group(2));
            if (innerMatcher.find()) {
                return true;
            }
        }
        return false;
    }

    @Test
    public void testVeggie() throws DbAccessException, Exception {
        OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
        OWLDataFactory fac = manager.getOWLDataFactory();

        SlomMetadata slom = SlomFactory.createNewInstance(
                new File(SlomTestDataBase.BASEPATH + "valid/case_VegetableSoup/intui_CCM.owl"),
                new File(SlomTestDataBase.BASEPATH + "valid/case_VegetableSoup/intui_CM.owl"));
        KnowledgeObject ko = slom.getKoByIri(slom.getCcm().getIriAsString() + "#ko001");
        ConceptContainer cc = slom.getCcByIri(slom.getCm().getIriAsString() + "#ccIngredientsShoppingList");
        MicroLearningPathway microLp = slom.getMicroLp(slom.getCcm().getIriAsString() + "#InquiryBasedMicroLP");
        MacroLearningPathway macroLp = slom.getMacroLp(slom.getCm().getIriAsString() + "#ChronologicalMLP");

        DbAccessor accessor = TestCommons.getAccessor();
        DbAccessUtil db = new DbAccessUtil(LID, accessor, slom);

        accessor.storeKoData(ko.getIriAsString(), slom.getCourseId(), null, null, "dummyLmsId");
        accessor.storeCcData(cc.getIriAsString(), slom.getCourseId(), null, null);
        accessor.storeMicroLpData(microLp.getIriAsString(), slom.getCourseId(), null, null);
        accessor.storeMacroLpData(macroLp.getIriAsString(), slom.getCourseId(), null, null);

        db.addLearningHistoryEntry(ko, cc, new Date().getTime());
        accessor.storeMicroLpSelection(LID, microLp.getIriAsString(), slom.getCourseId());
        accessor.storeMacroLpSelection(LID, macroLp.getIriAsString(), slom.getCourseId());

        Collection<OWLNamedIndividual> dfs = new HashSet<>();
        dfs.add(fac.getOWLNamedIndividual(IRI.create("http://www.intuitel.eu/public/intui_LM.owl#dfValAge_Adult")));

        LearnerStateOntology lso = LearnerStateOntologyFactory.createLsoRaw(LID, slom, db, dfs);
        assertNotNull("LSO was returned as null.", lso);
        assertNotNull("LSO turtle output is null.", lso.toTurtle());
        assertNotNull("LSO xml output is null.", lso.toXml());
        //System.out.println("\n\n\n" + lso_1.toTurtle());
    }
}
