/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.models.factories;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData01;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData02;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData03;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData04;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestDataUvie;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestDataVeggieSoup;
import java.util.Set;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.BeforeClass;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Test of the SLOM Factory (and attached factories) by loading the test-CCM.
 *
 * @author fhe
 */
public class SlomFactoryTest {

    private static final Logger LOGGER = LoggerFactory.getLogger(SlomFactoryTest.class);

    @BeforeClass
    public static void beforeClass() {
        LOGGER.info("Starting tests.");
    }

    @AfterClass
    public static void afterClass() {
        LOGGER.info("Finished tests.");
    }

    /**
     * Tests UVIE Slom
     */
    @Test
    public void testUvieSlom() {
        try {
            SlomTestDataUvie slomData = SlomTestDataUvie.getInstance();
            assertNotNull(slomData);
            assertNotNull(slomData.getSlom());
        } catch (Exception ex) {
            assertEquals(null, ex);
        }
    }

    /**
     * Test of the veggetable soup toy example.
     */
    @Test
    public void testVeggieSoupSlom() {
        try {
            SlomTestDataVeggieSoup slomData = SlomTestDataVeggieSoup.getInstance();
            assertNotNull(slomData);
            assertNotNull(slomData.getSlom());
        } catch (Exception ex) {
            assertEquals(null, ex);
        }
    }

    /**
     * Tests whether the data is correctly loaded from the valid SLOM testcase 01
     */
    @Test
    public void test01() {
        SlomTestData01 slomData = SlomTestData01.getInstance();

        testLoRelations(slomData.getKo01(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo02(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo03(), slomData.getCc02(), slomData.getSlom());
        testLoRelations(slomData.getKo04(), slomData.getCc02(), slomData.getSlom());

        testLpCompleteness(slomData.getMicroLp01(), slomData.getKo01(), slomData.getKo02(), slomData.getKo03(), slomData.getKo04());
        testLpCompleteness(slomData.getMicroLp02(), slomData.getKo01(), slomData.getKo02(), slomData.getKo04(), slomData.getKo03());
        testLpCompleteness(slomData.getMacroLp01(), slomData.getCc01(), slomData.getCc02());

        testLpRelations(slomData.getKo01(), slomData.getKo02(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo03(), slomData.getKo04(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo02(), slomData.getKo01(), slomData.getMicroLp02());
        testLpRelations(slomData.getKo04(), slomData.getKo03(), slomData.getMicroLp02());
        testLpRelations(slomData.getCc01(), slomData.getCc02(), slomData.getMacroLp01());

        testLpStarters(slomData.getMicroLp01(), slomData.getKo01(), slomData.getKo03());
        testLpStarters(slomData.getMicroLp02(), slomData.getKo02(), slomData.getKo04());
        testLpStarters(slomData.getMacroLp01(), slomData.getCc01());

        testLpEnders(slomData.getMicroLp01(), slomData.getKo02(), slomData.getKo04());
        testLpEnders(slomData.getMicroLp02(), slomData.getKo01(), slomData.getKo03());
        testLpEnders(slomData.getMacroLp01(), slomData.getCc02());

        assertEquals("Wrong dc:title loaded for Ko01.", "test title", slomData.getKo01().getTitle());
        assertEquals("Wrong dc:description loaded for Ko01.", "test description", slomData.getKo01().getDescription());
        assertEquals("Wrong dc:title loaded for microLp1.", "test title", slomData.getMicroLp01().getTitle());
        assertEquals("Wrong dc:description loaded for microLp1.", "test description", slomData.getMicroLp01().getDescription());
        assertEquals("Wrong dc:title loaded for Cc01.", "test title", slomData.getCc01().getTitle());
        assertEquals("Wrong dc:description loaded for Cc01.", "test description", slomData.getCc01().getDescription());
        assertEquals("Wrong dc:title loaded for macroLp1.", "test title", slomData.getMacroLp01().getTitle());
        assertEquals("Wrong dc:description loaded for macroLp1.", "test description", slomData.getMacroLp01().getDescription());
    }

    /**
     * Tests whether the data is correctly loaded from the valid SLOM testcase 02
     */
    @Test
    public void test02() {
        //microLp01: 0>1>3>5>7
        //microLp02: 0>2>4>6>7
        //   1-3-5
        //  /     \
        // 0       7
        //  \     /
        //   2-4-6
        //macroLp01: 0>1
        SlomTestData02 slomData = SlomTestData02.getInstance();

        testLoRelations(slomData.getKo01(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo02(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo03(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo04(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo05(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo06(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo07(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo08(), slomData.getCc01(), slomData.getSlom());

        testLpCompleteness(slomData.getMicroLp01(), slomData.getKo01(), slomData.getKo02(), slomData.getKo04(), slomData.getKo06(), slomData.getKo08());
        testLpCompleteness(slomData.getMicroLp02(), slomData.getKo01(), slomData.getKo03(), slomData.getKo05(), slomData.getKo07(), slomData.getKo08());
        testLpCompleteness(slomData.getMacroLp01(), slomData.getCc01(), slomData.getCc02());

        testLpStarters(slomData.getMicroLp01(), slomData.getKo01());
        testLpStarters(slomData.getMicroLp02(), slomData.getKo01());
        testLpStarters(slomData.getMacroLp01(), slomData.getCc01());

        testLpEnders(slomData.getMicroLp01(), slomData.getKo08());
        testLpEnders(slomData.getMicroLp02(), slomData.getKo08());
        testLpEnders(slomData.getMacroLp01(), slomData.getCc02());

        testLpRelations(slomData.getKo01(), slomData.getKo02(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo02(), slomData.getKo04(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo04(), slomData.getKo06(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo06(), slomData.getKo08(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo01(), slomData.getKo03(), slomData.getMicroLp02());
        testLpRelations(slomData.getKo03(), slomData.getKo05(), slomData.getMicroLp02());
        testLpRelations(slomData.getKo05(), slomData.getKo07(), slomData.getMicroLp02());
        testLpRelations(slomData.getKo07(), slomData.getKo08(), slomData.getMicroLp02());
    }

    /**
     * Tests whether the data is correctly loaded from the valid SLOM testcase 03
     */
    @Test
    public void test03() {
        //microLp01-1: 0>1>4
        //microLp01-2: 0>2>4
        //microLp02: 0>3>4
        //  1
        // / \
        //0-2-4
        // \ /
        //  3
        //macroLp01: 0>1
        SlomTestData03 slomData = SlomTestData03.getInstance();

        testLoRelations(slomData.getKo01(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo02(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo03(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo04(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo05(), slomData.getCc01(), slomData.getSlom());

        testLpCompleteness(slomData.getMicroLp01(), slomData.getKo01(), slomData.getKo02(), slomData.getKo03(), slomData.getKo05());
        testLpCompleteness(slomData.getMicroLp02(), slomData.getKo01(), slomData.getKo04(), slomData.getKo05());
        testLpCompleteness(slomData.getMacroLp01(), slomData.getCc01(), slomData.getCc02());

        testLpStarters(slomData.getMicroLp01(), slomData.getKo01());
        testLpStarters(slomData.getMicroLp02(), slomData.getKo01());
        testLpStarters(slomData.getMacroLp01(), slomData.getCc01());

        testLpEnders(slomData.getMicroLp01(), slomData.getKo05());
        testLpEnders(slomData.getMicroLp02(), slomData.getKo05());
        testLpEnders(slomData.getMacroLp01(), slomData.getCc02());

        testLpRelations(slomData.getKo01(), slomData.getKo02(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo01(), slomData.getKo03(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo02(), slomData.getKo05(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo03(), slomData.getKo05(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo01(), slomData.getKo04(), slomData.getMicroLp02());
        testLpRelations(slomData.getKo04(), slomData.getKo05(), slomData.getMicroLp02());
    }

    /**
     * Tests whether the data is correctly loaded from the valid SLOM testcase 04
     */
    @Test
    public void test04() {
        SlomTestData04 slomData = SlomTestData04.getInstance();
        //ccs:
        //  1
        // / \
        //0-2-4
        // \ /
        //  3
        //macroLp01-1: 0>1>4
        //macroLp01-2: 0>2>4
        //macroLp02: 0>3>4

        testLoRelations(slomData.getKo01(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo02(), slomData.getCc01(), slomData.getSlom());
        testLoRelations(slomData.getKo03(), slomData.getCc02(), slomData.getSlom());
        testLoRelations(slomData.getKo04(), slomData.getCc02(), slomData.getSlom());
        testLoRelations(slomData.getKo05(), slomData.getCc03(), slomData.getSlom());
        testLoRelations(slomData.getKo06(), slomData.getCc03(), slomData.getSlom());
        testLoRelations(slomData.getKo07(), slomData.getCc04(), slomData.getSlom());
        testLoRelations(slomData.getKo08(), slomData.getCc04(), slomData.getSlom());
        testLoRelations(slomData.getKo09(), slomData.getCc04(), slomData.getSlom());
        testLoRelations(slomData.getKo10(), slomData.getCc05(), slomData.getSlom());
        testLoRelations(slomData.getKo11(), slomData.getCc05(), slomData.getSlom());

        testLpCompleteness(slomData.getMicroLp01(), slomData.getKo01(), slomData.getKo02(), slomData.getKo03(),
                slomData.getKo04(), slomData.getKo05(), slomData.getKo06(), slomData.getKo07(),
                slomData.getKo08(), slomData.getKo09(), slomData.getKo10(), slomData.getKo11());
        testLpCompleteness(slomData.getMacroLp01(), slomData.getCc01(), slomData.getCc02(), slomData.getCc03(), slomData.getCc05());
        testLpCompleteness(slomData.getMacroLp02(), slomData.getCc01(), slomData.getCc04(), slomData.getCc05());

        testLpStarters(slomData.getMicroLp01(), slomData.getKo01(), slomData.getKo03(), slomData.getKo05(), slomData.getKo07(), slomData.getKo10());
        testLpStarters(slomData.getMacroLp01(), slomData.getCc01());
        testLpStarters(slomData.getMacroLp02(), slomData.getCc01());

        testLpEnders(slomData.getMicroLp01(), slomData.getKo02(), slomData.getKo04(), slomData.getKo06(), slomData.getKo08(), slomData.getKo09(), slomData.getKo11());
        testLpEnders(slomData.getMacroLp01(), slomData.getCc05());
        testLpEnders(slomData.getMacroLp02(), slomData.getCc05());

        testLpRelations(slomData.getCc01(), slomData.getCc02(), slomData.getMacroLp01());
        testLpRelations(slomData.getCc01(), slomData.getCc03(), slomData.getMacroLp01());
        testLpRelations(slomData.getCc02(), slomData.getCc05(), slomData.getMacroLp01());
        testLpRelations(slomData.getCc03(), slomData.getCc05(), slomData.getMacroLp01());
        testLpRelations(slomData.getCc01(), slomData.getCc04(), slomData.getMacroLp02());
        testLpRelations(slomData.getCc04(), slomData.getCc05(), slomData.getMacroLp02());
        testLpRelations(slomData.getKo01(), slomData.getKo02(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo03(), slomData.getKo04(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo05(), slomData.getKo06(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo07(), slomData.getKo08(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo07(), slomData.getKo09(), slomData.getMicroLp01());
        testLpRelations(slomData.getKo10(), slomData.getKo11(), slomData.getMicroLp01());
    }

    /**
     * Tests whether the KO and CC have been assigned to each other.
     *
     * @param ko KO in scope
     * @param cc CC in scope
     */
    private void testLoRelations(KnowledgeObject ko, ConceptContainer cc, SlomMetadata slom) {
        assertTrue(String.format("Missing relation between KO '%s' to CC '%s'.", ko, cc), ko.isContainedIn(cc));
        assertTrue(String.format("Missing relation between KO '%s' to CC '%s'.", ko, cc), slom.getReferences(ko).contains(cc));
        assertTrue(String.format("Missing relation between CC '%s' to KO '%s'.", ko, cc), slom.getReferences(cc).contains(ko));
    }

    /**
     * Seeks the LP with the given ID in the SLOM object, tests it and returns it afterwards
     *
     * @param slom SLOM object to inspect
     * @param id ID of the LP in scope
     * @param lpClass Class of the LP in scope
     * @return
     */
    public static LearningPathway getLp(SlomMetadata slom, String id, Class<? extends LearningPathway> lpClass) {
        assertTrue(id + " has not been loaded.", lpClass == MacroLearningPathway.class
                ? slom.containsMacroLp(slom.getCm().getIriAsString() + id)
                : slom.containsMicroLp(slom.getCcm().getIriAsString() + id));
        LearningPathway lp = lpClass == MacroLearningPathway.class
                ? slom.getMacroLp(slom.getCm().getIriAsString() + id)
                : slom.getMicroLp(slom.getCcm().getIriAsString() + id);
        assertNotNull(id + " has been returned as null.", lp);
        return lp;
    }

    /**
     * Tests whether the given {@code lp} contains the specified {@code los}.
     *
     * @param lp LP to inspect
     * @param los LOs to be tested if part of the given LP
     */
    private void testLpCompleteness(LearningPathway lp, LearningObject... los) {
        Set<? extends LearningObject> lpLos = lp.getAllReferences();
        assertEquals(String.format("Loaded number of LO on %s is incorrect.", lp), los.length, lpLos.size());
        for (LearningObject lo : los) {
            assertTrue(String.format("LP %s does not contain LO %s.", lp, lo), lpLos.contains(lo));
        }
    }

    /**
     * Tests whether the given {@code lp} contains the specified {@code starters}.
     *
     * @param lp LP to inspect
     * @param starters LOs to be tested if starters of the given LP
     */
    private void testLpStarters(LearningPathway lp, LearningObject... starters) {
        assertEquals(String.format("Loaded number of LP starters on %s is incorrect.", lp), starters.length, lp.getStarters().size());
        for (LearningObject lo : starters) {
            assertTrue(String.format("LP %s does not contain LO %s as starter.", lp, lo), lp.getStarters().contains(lo));
        }
    }

    /**
     * Tests whether the given {@code lp} contains the specified {@code enders}.
     *
     * @param lp LP to inspect
     * @param enders LOs to be tested if starters of the given LP
     */
    private void testLpEnders(LearningPathway lp, LearningObject... enders) {
        assertEquals(String.format("Loaded number of LP starters on %s is incorrect.", lp), enders.length, lp.getEnders().size());
        for (LearningObject lo : enders) {
            assertTrue(String.format("LP %s does not contain LO %s as ender.", lp, lo), lp.getEnders().contains(lo));
        }
    }

    /**
     * Tests the whether the realtions between the two specified LOs and the given LP have been
     * loaded correctly.
     *
     * @param lo01 LO that is the referencing entitiy in the SLOM object
     * @param lo02 LO that is the referencee in the SLOM object
     * @param lp LP to test
     */
    private void testLpRelations(LearningObject lo01, LearningObject lo02, LearningPathway lp) {
        assertTrue(lo01.getIri().getFragment() + " does not have successor for " + lp.getIri().getFragment(),
                lo01.hasSuccessorFor(lp));
        assertTrue(lo02.getIri().getFragment() + " does not have predecessor for " + lp.getIri().getFragment(),
                lo02.hasPredecessorFor(lp));
        assertTrue(lo01.getIri().getFragment() + " does not have " + lo02.getIri().getFragment() + " as successor.",
                lo01.getDirectSucessorsFor(lp).contains(lo02));
        assertTrue(lo02.getIri().getFragment() + " does not have " + lo01.getIri().getFragment() + " as predecessor.",
                lo02.getDirectPredecessorsFor(lp).contains(lo01));
        assertTrue("LP does not contain a ref between " + lo01.getIri().getFragment() + " and " + lo02.getIri().getFragment(),
                lp.getReferenceesFor(lo01).contains(lo02));
        assertTrue("LP does not contain a backward ref between " + lo01.getIri().getFragment() + " and " + lo02.getIri().getFragment(),
                lp.getReferencersFor(lo02).contains(lo01));
    }
}
