/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-07-14
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.lpmProcessor.TestCommons;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.abstractTypes.LearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.ConceptContainer;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.KnowledgeObject;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MacroLearningPathway;
import eu.intuitel.lpmProcessor.models.owlWrapper.concreteTypes.MicroLearningPathway;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData01;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.BeforeClass;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Tests for the CompletionUtil.
 *
 * @author Florian Heberle
 */
public class CompletionUtilTest {

    private static final Logger LOGGER = LoggerFactory.getLogger(CompletionUtilTest.class);

    @BeforeClass
    public static void beforeClass() {
        LOGGER.info("Starting tests.");
    }

    @AfterClass
    public static void afterClass() {
        LOGGER.info("Finished tests.");
    }

    @Test
    public void testTraversalMicro() {
        SlomTestData01 slomData = SlomTestData01.getInstance();
        LearningPathway lpToInspect = slomData.getMicroLp01();
        assertNotNull(lpToInspect);
        testTraversal(lpToInspect);
    }

    @Test
    public void testTraversalMacro() {
        SlomTestData01 slomData = SlomTestData01.getInstance();
        LearningPathway lpToInspect = slomData.getMacroLp01();
        assertNotNull(lpToInspect);
        testTraversal(lpToInspect);
    }

    private void testTraversal(LearningPathway lpToInspect) {
        List<LearningObject> losToInspect = new ArrayList<>(lpToInspect.getStarters());
        assertNotNull(losToInspect);
        List<LearningObject> completedLos = new ArrayList<>();
        LearningObject completeLo = losToInspect.get(losToInspect.size() - 1);
        while (completeLo != null) {
            completedLos.add(completeLo);
            if (completeLo.getDirectSucessorsFor(lpToInspect).isEmpty()) {
                completeLo = null;
            } else {
                completeLo = new ArrayList<>(completeLo.getDirectSucessorsFor(lpToInspect)).get(completeLo.getDirectSucessorsFor(lpToInspect).size() - 1);
            }
        }
        assertTrue(CompletionUtil.containsStarterOfCompletedChain(losToInspect, completedLos, lpToInspect));
        completedLos.remove(completedLos.size() - 1);
        assertFalse(CompletionUtil.containsStarterOfCompletedChain(losToInspect, completedLos, lpToInspect));
    }

    @Test
    public void testCcCompletion() {
        //Get test data
        SlomTestData01 slomData = SlomTestData01.getInstance();
        MacroLearningPathway macroLpToInspect = slomData.getMacroLp01();
        //Only one MLP in this SLOM
        assertNotNull(macroLpToInspect);

        ConceptContainer lastStarter = getLastElement(macroLpToInspect.getStarters());
        //Only one starter in this MLP
        assertNotNull(lastStarter);

        MicroLearningPathway microLpToInspect = slomData.getMicroLp01();
        //Specific selection needed, there are two mLPs and the second is not suitable here
        assertNotNull(microLpToInspect);

        //Test empty assumption
        Collection<KnowledgeObject> completedKos = new ArrayList<>();
        assertFalse(CompletionUtil.isCompletedCc(lastStarter, slomData.getSlom(), completedKos));

        //Test positive data
        KnowledgeObject ko1 = slomData.getKo01();
        assertNotNull(ko1);
        KnowledgeObject ko2 = slomData.getKo02();
        assertNotNull(ko2);
        completedKos.add(ko1);
        completedKos.add(ko2);
        assertTrue(CompletionUtil.isCompletedCc(lastStarter, slomData.getSlom(), completedKos));

        //Make completion values incomplete - should fail
        completedKos.remove(ko2);
        assertFalse(CompletionUtil.isCompletedCc(lastStarter, slomData.getSlom(), completedKos));
        completedKos.remove(ko1);
        completedKos.add(ko2);
        assertFalse(CompletionUtil.isCompletedCc(lastStarter, slomData.getSlom(), completedKos));
    }

    private <T> T getLastElement(final Collection<T> c) {
        final Iterator<T> itr = c.iterator();
        T lastElement = itr.next();
        while (itr.hasNext()) {
            lastElement = itr.next();
        }
        return lastElement;
    }

    @Test
    public void testCcCompletionStorage() {
        DbAccessor accessor = TestCommons.getAccessor();
        try {
            //Start with empty tables
            accessor.clearDataStorage();

            SlomTestData01 slomData = SlomTestData01.getInstance();

            String learnerId = "CompUtil_MrTest";
            DbAccessUtil dbUtil = new DbAccessUtil(learnerId, accessor, slomData.getSlom());
            List<ConceptContainer> completedCcs = dbUtil.getCompletedCCs();
            assertNotNull(completedCcs);
            assertTrue(completedCcs.isEmpty());

            //Initial state ok, now add first learning step
            //Ko01 and Ko02 both belong to Cc01. Cc01 is first complete when Ko01 and Ko02 are complete.
            KnowledgeObject ko1 = slomData.getKo01();
            KnowledgeObject ko2 = slomData.getKo02();
            ConceptContainer cc = slomData.getCc01();
            dbUtil.setKoData(ko1.getIriAsString(), "", "", "");
            dbUtil.setKoData(ko2.getIriAsString(), "", "", "");
            dbUtil.setCcData(cc.getIriAsString(), "", "");

            //Add first stepp
            dbUtil.addLearningHistoryEntry(ko1, cc, System.currentTimeMillis() - 100);
            dbUtil.setKoCompletionStatus(ko1.getIriAsString(), Boolean.TRUE, 1, null, null);
            Thread.sleep(100);

            //Now, test if there is a score yet
            completedCcs = dbUtil.getCompletedCCs();
            assertNotNull(completedCcs);
            assertTrue(completedCcs.isEmpty());

            //Update and test again
            CompletionUtil.updateCcCompletionValue(ko1, dbUtil, slomData.getSlom());
            Thread.sleep(100);
            completedCcs = dbUtil.getCompletedCCs();
            assertNotNull(completedCcs);
            assertTrue(completedCcs.isEmpty());

            //Next learning step
            dbUtil.addLearningHistoryEntry(ko2, cc, System.currentTimeMillis() + 100);
            dbUtil.setKoCompletionStatus(ko2.getIriAsString(), Boolean.TRUE, 1, null, null);
            Thread.sleep(100);

            //Test again
            completedCcs = dbUtil.getCompletedCCs();
            assertNotNull(completedCcs);
            assertTrue(completedCcs.isEmpty());

            //Update and test again - since both KOs have scores now, Cc should be completed
            CompletionUtil.updateCcCompletionValue(ko2, dbUtil, slomData.getSlom());
            Thread.sleep(100);
            completedCcs = dbUtil.getCompletedCCs();
            assertNotNull(completedCcs);
            assertFalse(completedCcs.isEmpty());
            assertEquals(1, completedCcs.size());
            assertTrue(completedCcs.contains(cc));

            //Reset Ko, update and test again - should be empty again
            //This does not happen in normal learning scnearios - executed here for testing purposes only.
            dbUtil.setKoCompletionStatus(ko2.getIriAsString(), Boolean.FALSE, null, null, null);
            Thread.sleep(100);
            CompletionUtil.updateCcCompletionValue(ko2, dbUtil, slomData.getSlom());
            Thread.sleep(100);
            completedCcs = dbUtil.getCompletedCCs();
            assertNotNull(completedCcs);
            assertTrue(completedCcs.isEmpty());
        } catch (InterruptedException ex) {
            assertEquals(null, ex);
        } finally {
            //cleanup data
            accessor.clearDataStorage();
        }
    }
}
