/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Alexander Streicher, Fraunhofer IOSB, www.iosb.fraunhofer.de
 * Version: [0.01]
 * Date: 2013-11-05
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData01;
import org.apache.commons.lang3.StringUtils;
import org.junit.After;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Tests for the threadsaftey access of cleared SLOM metdata in cache.
 *
 * @author fhe
 */
@SuppressWarnings("SleepWhileInLoop")
public class SlomCacheTest {

    private static final Logger LOGGER = LoggerFactory.getLogger(SlomCacheTest.class);
    //    private static SlomCache cache;
    private static SlomMapCache cache;
    private static SlomMetadata slom;
    private static String koLmsId;

    @BeforeClass
    public static void setUpBeforeClass() throws CloneNotSupportedException {
        LOGGER.info("Starting tests.");

        SlomTestData01 slomData = SlomTestData01.getInstance();
        slom = slomData.getSlom();
        assertNotNull("Could not load SLOM test data.", slom);
        assertFalse("KO collection is empty.", slom.getKos().isEmpty());
        koLmsId = slomData.getKo01().getLinkedLmsElement();
        assertFalse("LMS-id of test ko is blank.", StringUtils.isBlank(koLmsId));
        //        cache = SlomCache.getInstance();
        cache = SlomMapCache.getInstance();
        assertNotNull("Cache is null.", cache);
        cache.clear();
    }

    @AfterClass
    public static void afterClass() {
        LOGGER.info("Finished tests.");
    }

    @Before
    public void beforeTest() {
        cache.store(slom);
    }

    @After
    public void afterTest() {
        cache.clear();
    }

    @Test
    public void test1() {
        test(1, 500);
    }

    @Test
    public void test100() {
        test(100, 500);
    }

    @Test
    public void test1000c500() {
        test(1000, 500);
    }

    @Test
    public void test1000c10() {
        test(1000, 10);
    }

    @Test
    public void test1000c1000() {
        test(1000, 1000);
    }

    @Test
    public void test10000() {
        test(10000, 500);
    }

    /**
     * Parametrizable test method that endlessly reads the cache with multiple threads until cache
     * is cleared. Asserts true when all threads finished without an error.
     *
     * @param threadCount Number of reader threads
     * @param clearWaitingTime Time till a thread is called to clear the cache
     */
    private void test(final int threadCount, final int clearWaitingTime) {
        Long start = System.currentTimeMillis();
        LOGGER.debug(String.format("Start cache test, threadCount=%s", threadCount));

        CacheReader[] readers = new CacheReader[threadCount];
        for (int i = 0; i < threadCount; i++) {
            readers[i] = new CacheReader();
            readers[i].start();
        }

        CacheClearer clearer = new CacheClearer(clearWaitingTime);
        clearer.start();
        try {
            clearer.join();
        } catch (InterruptedException ex) {
            assertEquals(null, ex);
        }
        assertTrue("Cache clearer thread has exited with error.", clearer.hasExitedWithoutError());

        for (CacheReader reader : readers) {
            try {
                reader.join();
            } catch (InterruptedException ex) {
                assertEquals(null, ex);
            }
            assertTrue("Cache reader thread has exited with error.", reader.hasExitedWithoutError());
        }

        LOGGER.debug(String.format("Finished cache test after %s ms.", System.currentTimeMillis() - start));
    }

    /**
     * Simple Runnable that clears the cache after a given amount of ms.
     */
    private static class CacheClearer extends Thread {

        private final long sleepTime;
        private boolean exitedWithoutError = true;

        public CacheClearer(long sleepTime) {
            this.sleepTime = sleepTime;
        }

        @Override
        public void run() {
            try {
                Thread.sleep(sleepTime);
                cache.clear();
            } catch (InterruptedException ex) {
                LOGGER.error("Thread interrupted", ex);
                exitedWithoutError = false;
            }
        }

        public boolean hasExitedWithoutError() {
            return exitedWithoutError;
        }
    }

    /**
     * Simple Thread that accesses SLOM cache in randomized intervalls as long as the cache
     * contains data.
     */
    private static class CacheReader extends Thread {

        private boolean exitedWithoutError = true;

        public CacheReader() {
        }

        @Override
        public void run() {
            try {
                while (cache.getSize() > 0) {
                    cache.getSlomMetadata(koLmsId);
                    Thread.sleep((long) (Math.random() * 10));
                }
            } catch (InterruptedException ex) {
                LOGGER.error("Thread interrupted", ex);
                exitedWithoutError = false;
            }
        }

        public boolean hasExitedWithoutError() {
            return exitedWithoutError;
        }
    }
}
