/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 2014-03-21
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.lpmProcessor.utils;

import eu.intuitel.lpmProcessor.models.SlomMetadata;
import eu.intuitel.lpmProcessor.test.mocks.SlomTestData01;
import java.util.Map;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Tests of the thread-safeness of the Map-Cache.
 *
 * @author fhe
 */
public class SlomMapCacheTest {

    private static final Logger LOGGER = LoggerFactory.getLogger(SlomMapCacheTest.class);
    private SlomMapCacheTester cache;
    private SlomMetadata slom;

    @BeforeClass
    public static void beforeClass() {
        LOGGER.info("Starting tests.");
    }

    @AfterClass
    public static void afterClass() {
        LOGGER.info("Finished tests.");
    }

    @Before
    public void beforeTest() {
        cache = new SlomMapCacheTester();
        slom = SlomTestData01.getInstance().getSlom();
    }

    @Test
    public void testSimultaneousStoring_2() {
        testSimultaneousStoring(2);
    }

    @Test
    public void testSimultaneousStoring_100() {
        testSimultaneousStoring(100);
    }

    @Test
    public void testSimultaneousStoring_1000() {
        testSimultaneousStoring(1000);
    }

    /**
     * Runs the given {@code count} of threads that try to store an element in the cache
     * simultaneously.
     *
     * @param count Count of storing threads
     */
    private void testSimultaneousStoring(int count) {
        try {
            CyclicBarrier gate = new CyclicBarrier(count + 1);
            CacheStorer[] storers = new CacheStorer[count];
            for (int i = 0; i < storers.length; i++) {
                storers[i] = new CacheStorer(i, gate);
            }

            for (CacheStorer storer : storers) {
                storer.start();
            }
            LOGGER.debug(String.format("Started cache storer threads (%s). Simultaneous storing starts now.", count));
            gate.await();

            for (CacheStorer storer : storers) {
                storer.join();
            }
            LOGGER.debug("Cache storer threads finished.");

            for (CacheStorer storer : storers) {
                assertFalse(String.format("Cache storer %s encountered an error.", storer.id), storer.error);
            }
            assertEquals("Cache was updated too often.", 1, cache.updateCtr);
        } catch (InterruptedException | BrokenBarrierException ex) {
            assertNull("Exception during test. " + ex, ex);
        }
    }

    /**
     * Helper class to store an element in the SLOM cache.
     */
    private class CacheStorer extends Thread {

        public boolean error = true;
        public int id;
        private final CyclicBarrier gate;

        public CacheStorer(int id, CyclicBarrier gate) {
            this.id = id;
            this.gate = gate;
        }

        @Override
        public void run() {
            try {
                gate.await();
                cache.store(slom);
                error = false;
            } catch (InterruptedException ex) {
                LOGGER.error("Interrupted thread exception.", ex);
            } catch (BrokenBarrierException ex) {
                LOGGER.error("Broken barrier exception.", ex);
            }
        }
    }

    /**
     * Subclass of SlomMapCache to avoid the singleton instantiation.
     */
    private class SlomMapCacheTester extends SlomMapCache {

        public int updateCtr = 0;

        @Override
        protected void extendIndexMap(Map<String, SlomMetadata> extension) {
            super.extendIndexMap(extension);
            updateCtr++;
        }
    }
}
