/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, University of Reading, www.reading.ac.uk
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 16.07.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.db.access;

import eu.intuitel.db.model.*;
import static eu.intuitel.db.model.ConceptContainerScore.conceptContainerScore;
import static eu.intuitel.db.model.KnowledgeObjectScore.knowledgeObjectScore;
import static eu.intuitel.db.model.LearningRecommendation.learningRecommendation;
import static eu.intuitel.db.model.LearningStep.learningStep;
import static eu.intuitel.db.model.MacroLearningPathwaySelection.macroLearningPathwaySelection;
import static eu.intuitel.db.model.MicroLearningPathwaySelection.microLearningPathwaySelection;
import static eu.intuitel.db.model.TimedependentSessionData.timedependentSessionData;
import eu.intuitel.types.*;
import java.sql.Date;
import java.util.*;

import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class DbAccessor {

    private static final Logger logger = LoggerFactory.getLogger(DbAccessor.class);

    private final IDbAccessPoint DB;

    /**
     * Constructor.
     *
     * @param db Type of connection to be used.
     * @throws DbAccessException To be thrown if an exception occurs during connection init.
     */
    private DbAccessor(IDbAccessPoint db) throws DbAccessException {
        this.DB = db;
    }

    /**
     * @return A new accessor that loads and stores data in the local cache.
     * @throws DbAccessException To be thrown if an exception occurs during connection init.
     */
    public static DbAccessor createCachedDbAccessor() throws DbAccessException {
        return new DbAccessor(new LocalDbCache());
    }

    /**
     * @return A new accessor that loads and stores data in the local database.
     * @throws DbAccessException To be thrown if an exception occurs during connection init.
     */
    public static DbAccessor createHibernateDbAccessor() throws DbAccessException {
        return new DbAccessor(new HibernateDbAccessPoint());
    }

    /**
     * Shuts the complete DB connection down.
     * This methods needs to be called when no the complete system is being shut down.
     */
    public void shutdown() {
        if (DB != null) {
            DB.shutdown();
        }
    }

    /**
     * Determines if the database connection was established correctly.
     *
     * @return True if DB connection is available.
     */
    public boolean isConnectionEstablished() {
        return DB != null && DB.isConnectionEstablished();
    }

    /**
     * @return The current time as java.sql.Date object.
     */
    private Date getNow() {
        return new Date(new java.util.Date().getTime());
    }

    /**
     * Sets the LMS properties.
     *
     * @param id ID of the LMS
     * @param name Name of the LMS
     * @param type Type of the LMS
     * @param mediaLvl Media lvl of the LMS
     * @param loreLvl LORE of the LMS
     * @param useLvl USE of the LMS
     * @param nameFormality Name formality of the LMS
     * @param comStyle Communication style of the LMS
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeLmsProperties(String id, String name, String type, MediaLvl mediaLvl, LoreLvl loreLvl,
            UseLvl useLvl, NameFormality nameFormality, ComStyle comStyle) {
        try {
            LmsParameter lms = new LmsParameter(getNow(), id, name, type,
                    MediaLvl.toString(mediaLvl),
                    LoreLvl.toInt(loreLvl),
                    UseLvl.toInt(useLvl),
                    ComStyle.toInt(comStyle),
                    NameFormality.toInt(nameFormality));
            DB.storeOrUpdate(lms);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @return Returns the ID of the LMS. May be null.
     */
    public String getLmsId() {
        try {
            LmsParameter lms = DB.getLastLmsParameter();
            return lms.getId();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @return Returns the name of the LMS. May be null.
     */
    public String getLmsName() {
        try {
            LmsParameter lms = DB.getLastLmsParameter();
            return lms.getName();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @return Returns the type of the LMS. May be null.
     */
    public String getLmsType() {
        try {
            LmsParameter lms = DB.getLastLmsParameter();
            return lms.getType();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @return Returns the mediaLvl of the LMS. May be null.
     */
    public MediaLvl getLmsMediaLvl() {
        try {
            LmsParameter lms = DB.getLastLmsParameter();
            return MediaLvl.fromString(lms.getMediaLevel());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @return Returns the LORE lvl of the LMS. May be null.
     */
    public LoreLvl getLmsLoreLvl() {
        try {
            LmsParameter lms = DB.getLastLmsParameter();
            return LoreLvl.fromInt(lms.getLoreLevel());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @return Returns the USE lvl of the LMS. May be null.
     */
    public UseLvl getLmsUseLvl() {
        try {
            LmsParameter lms = DB.getLastLmsParameter();
            return UseLvl.fromInt(lms.getUseLevel());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @return Returns the name formality of the LMS. May be null.
     */
    public NameFormality getLmsNameFormality() {
        try {
            LmsParameter lms = DB.getLastLmsParameter();
            return NameFormality.fromInt(lms.getNameFormality());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @return Returns the communication style of the LMS. May be null.
     */
    public ComStyle getLmsComStyle() {
        try {
            LmsParameter lms = DB.getLastLmsParameter();
            return ComStyle.fromInt(lms.getComStyle());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * Adds a new learner to the data storage or updates an existing one.
     *
     * @param learnerId ID of the learner
     * @param name Name of the learner
     * @param gender Gender of the learner
     * @param age Age of the learner
     * @param culture Culture of the learner
     * @param blindness Blindness of the learner
     * @param muteness Muteness of the learner
     * @param deafness Deafness of the learner
     * @param languages Languages spoken by the learner
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeLearner(String learnerId, String name, Gender gender, Integer age, Culture culture,
            Boolean blindness, Boolean muteness, Boolean deafness, Language... languages) {
        try {
            Learner l = new Learner(learnerId,
                    name,
                    gender == null ? null : gender.getLabel(),
                    age,
                    culture == null ? null : culture.getLabel(),
                    blindness,
                    muteness,
                    deafness,
                    Language.join(languages));
            DB.storeOrUpdate(l);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return The name of the learner. May be null.
     */
    public String getLearnerName(String learnerId) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                return null;
            }
            return l.getName();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return The gender of the learner. May be null.
     */
    public Gender getLearnerGender(String learnerId) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                return null;
            }
            return Gender.getGender(l.getGender());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param gender of the learner
     */
    public void setLearnerGender(String learnerId, Gender gender) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                logger.error("Unable to retrieve the learner ...");
            } else {
                l.setGender(gender.getLabel());
                DB.storeOrUpdate(l);
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return The age of the learner. May be null.
     */
    public Integer getLearnerAge(String learnerId) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                return null;
            }
            return l.getAge();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param age The age of the learner
     */
    public void setLearnerAge(String learnerId, int age) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                logger.error("Unable to retrieve the learner ...");
            } else {
                l.setAge(age);
                DB.storeOrUpdate(l);
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return The culture of the learner. May be null.
     */
    public Culture getLearnerCulture(String learnerId) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                return null;
            }
            return Culture.getCulture(l.getCulture());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    public void setLearnerCulture(String learnerId, Culture culture) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                logger.error("Unable to retrieve the learner ...");
            } else {
                l.setCulture(culture.getLabel());
                DB.storeOrUpdate(l);
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return True if the learner is blind. May be null.
     */
    public Boolean isLearnerBlind(String learnerId) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                return null;
            }
            return l.getBlindness();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    public void setLearnerBlindState(String learnerId, Boolean state) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                logger.error("Unable to retrieve the learner ...");
            } else {
                l.setBlindness(state);
                DB.storeOrUpdate(l);
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Trie if the learner is mute. May be null.
     */
    public Boolean isLearnerMute(String learnerId) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                return null;
            }
            return l.getMuteness();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    public void setLearnerMuteState(String learnerId, Boolean state) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                logger.error("Unable to retrieve the learner ...");
            } else {
                l.setMuteness(state);
                DB.storeOrUpdate(l);
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return True if the learner is deaf. May be null.
     */
    public Boolean isLearnerDeaf(String learnerId) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                return null;
            }
            return l.getDeafness();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    public void setLearnerDeafState(String learnerId, Boolean state) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                logger.error("Unable to retrieve the learner ...");
            } else {
                l.setDeafness(state);
                DB.storeOrUpdate(l);
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Retrieves the languages the learner is able to speak. Empty collection if no data
     * available or on error.
     */
    public Collection<Language> getLearnerLanguages(String learnerId) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                return Collections.emptySet();
            } else {
                return Language.split(l.getLanguages());
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return new HashSet<>();
        }
    }

    public void setLearnerLanguages(String learnerId, Language... languages) {
        try {
            Learner l = DB.getLearner(learnerId);
            if (l == null) {
                logger.error("Unable to retrieve the learner ...");
            } else {
                l.setLanguages(Language.join(languages));
                DB.storeOrUpdate(l);
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param ccIri IRI of the Concept Container in scope
     * @return True if completed, False if not completed, Null on error.
     */
    public Boolean getCcCompletionStatus(String learnerId, String ccIri) {
        try {
            ConceptContainerScore ccScore = DB.getCcLastScore(learnerId, ccIri);
            return ccScore != null && ccScore.isCompleted();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * Determines the completion status of the specified KO by evaluating the different completion
     * values.
     *
     * @param learnerId ID of the learner in scope
     * @param koIri IRI of the ko in scope
     * @return The completion status (0-100) or null on error or if undeterminable.
     */
    public Integer getKoCompletionStatus(String learnerId, String koIri) {
        try {
            KnowledgeObjectScore koScore = DB.getLastKoScore(learnerId, koIri);
            if (koScore == null) {
                return null;
            }
            if (koScore.getGrade() != null) {
                switch (koScore.getGrade()) {
                    case 1:
                        return 100;
                    case 2:
                        return 80;
                    case 3:
                        return 60;
                    case 4:
                        return 40;
                    case 5:
                        return 20;
                    case 6:
                        return 0;
                }
                return koScore.getGrade();
            }
            if (koScore.getSeenPercentage() != null) {
                return koScore.getSeenPercentage();
            }
            if (koScore.getTextualCompletion() != null) {
                return koScore.getTextualCompletion();
            }
            if (koScore.isAccessed()) {
                return 100;
            }
            return null;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param koIri IRI of the ko in scope
     * @return True if accessed, false if not. Null on error.
     */
    public Boolean getKoCompletionAccessed(String learnerId, String koIri) {
        try {
            KnowledgeObjectScore koScore = DB.getLastKoScore(learnerId, koIri);
            if (koScore == null) {
                return null;
            }
            return koScore.isAccessed();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param koIri IRI of the ko in scope
     * @return The completion status (1-6), null if unavailable or null on error.
     */
    public Integer getKoCompletionGrade(String learnerId, String koIri) {
        try {
            KnowledgeObjectScore koScore = DB.getLastKoScore(learnerId, koIri);
            if (koScore == null) {
                return null;
            }
            return koScore.getGrade();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param koIri IRI of the ko in scope
     * @return The completion status (0-100), null if unavailable or null on error.
     */
    public Integer getKoCompletionSeenPercentage(String learnerId, String koIri) {
        try {
            KnowledgeObjectScore koScore = DB.getLastKoScore(learnerId, koIri);
            if (koScore == null) {
                return null;
            }
            return koScore.getSeenPercentage();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param koIri IRI of the ko in scope
     * @return The completion status (0-100), null if unavailable or null on error.
     */
    public Integer getKoCompletionTextualCompletion(String learnerId, String koIri) {
        try {
            KnowledgeObjectScore koScore = DB.getLastKoScore(learnerId, koIri);
            if (koScore == null) {
                return null;
            }
            return koScore.getTextualCompletion();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return List of completed CCs (iris), empty list on error.
     */
    public List<String> getCompletedCcs(String learnerId, String courseId) {
        try {
            List<ConceptContainerScore> ccScores = DB.getCcScores(learnerId, courseId);
            Set<String> checkedIris = new HashSet<>();
            List<String> resultSet = new ArrayList<>();
            for (ConceptContainerScore ccScore : ccScores) {
                //Entries are sorted by timestamp. Only the newest is interesting.
                if (!checkedIris.contains(ccScore.getId().getConceptContainerIri())) {
                    if (ccScore.isCompleted()) {
                        resultSet.add(ccScore.getId().getConceptContainerIri());
                    }
                }
                checkedIris.add(ccScore.getId().getConceptContainerIri());
            }
            return resultSet;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return Collections.emptyList();
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return List of completed KOs (iris), empty list on error.
     */
    public List<String> getCompletedKOs(String learnerId, String courseId) {
        try {
            List<KnowledgeObjectScore> koScores = DB.getKoScores(learnerId, courseId);
            Set<String> checkedIris = new HashSet<>();
            List<String> resultSet = new ArrayList<>();
            for (KnowledgeObjectScore koScore : koScores) {
                //Entries are sorted by timestamp. Only the newest is interesting.
                if (!checkedIris.contains(koScore.getId().getKnowledgeObjectIri())) {
                    if (koScore.getGrade() != null && koScore.getGrade() <= 4) {
                        resultSet.add(koScore.getId().getKnowledgeObjectIri());
                    } else if (koScore.getSeenPercentage() != null && koScore.getSeenPercentage() >= 70) {
                        resultSet.add(koScore.getId().getKnowledgeObjectIri());
                    } else if (koScore.getTextualCompletion() != null && koScore.getTextualCompletion() >= 70) {
                        resultSet.add(koScore.getId().getKnowledgeObjectIri());
                    } else if (koScore.isAccessed()) {
                        resultSet.add(koScore.getId().getKnowledgeObjectIri());
                    }
                }
                checkedIris.add(koScore.getId().getKnowledgeObjectIri());
            }
            return resultSet;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return Collections.emptyList();
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri IRI of the KO in scope
     * @param accessed True if accessed
     * @param grade Grade to be set. Value between 1 and 6.
     * @param seenPercentage Value to be set (0-100) - percentage of how much content the learner
     * has seen (audio/video)
     * @param textCompletion Value to be set (0-100) - percentage of how much content the learner
     * has read
     * @return True if successfull
     */
    public boolean storeKoCompletion(String learnerId, String courseId, String koIri, Boolean accessed, Integer grade,
            Integer seenPercentage, Integer textCompletion) {
        try {
            KnowledgeObjectScore koScore = knowledgeObjectScore(learnerId, courseId, koIri, getNow(),
                    accessed, seenPercentage, textCompletion, grade);
            DB.storeOrUpdate(koScore);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return false;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri IRI of the KO in scope
     * @param grade Grade to be set. Value between 1 and 6.
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeKoCompletionGrade(String learnerId, String courseId, String koIri, int grade) {
        try {
            KnowledgeObjectScore koScore = knowledgeObjectScore(learnerId, courseId, koIri, getNow(),
                    true, null, null, grade);
            DB.storeOrUpdate(koScore);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri IRI of the KO in scope
     * @param value Value to be set (0-100)
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeKoCompletionTextualCompletion(String learnerId, String courseId, String koIri, int value) {
        try {
            KnowledgeObjectScore koScore = knowledgeObjectScore(learnerId, courseId, koIri, getNow(),
                    true, null, value, null);
            DB.storeOrUpdate(koScore);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri IRI of the KO in scope
     * @param value Value to be store (0-100)
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeKoCompletionSeenPercentage(String learnerId, String courseId, String koIri, int value) {
        try {
            KnowledgeObjectScore koScore = knowledgeObjectScore(learnerId, courseId, koIri, getNow(),
                    true, value, null, null);
            DB.storeOrUpdate(koScore);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri IRI of the KO in scope
     * @param accessed True if accessed
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeKoCompletionAccessed(String learnerId, String courseId, String koIri, boolean accessed) {
        try {
            KnowledgeObjectScore koScore = knowledgeObjectScore(learnerId, courseId, koIri, getNow(), accessed);
            DB.storeOrUpdate(koScore);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param ccIri IRI of the CC in scope
     * @param complete True if complete
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeCcCompletion(String learnerId, String courseId, String ccIri, boolean complete) {
        try {
            ConceptContainerScore ccScore = conceptContainerScore(learnerId, ccIri, courseId, getNow(), complete);
            DB.storeOrUpdate(ccScore);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * Marks all KOs of the current CC as complete.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return True if successfull, false if unsuccessfull.
     */
    public Boolean markAllKoComplete(String learnerId, String courseId) {
        String currentCC = getLastConceptContainerIri(learnerId);
        if (currentCC != null) {
            try {
                for (LearningStep learningStep : DB.getLearningStepsWithCcIri(learnerId, currentCC)) {
                    storeKoCompletionSeenPercentage(learnerId, courseId, learningStep.getKnowledgeObjectIri(), 100);
                    learningStep.getKnowledgeObjectIri();
                }
                return true;
            } catch (DbAccessException e) {
                logger.error(e.getMessage(), e);
            }
        }
        return false;
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param inspectionSize Number of elements to load
     * @return List of CCs the learner has accessed (0=newest), or empty List on error.
     */
    public List<String> getHistoryOfCCs(String learnerId, String courseId, int inspectionSize) {
        try {
            List<String> ccIris = new ArrayList<>();
            for (LearningStep lStep : DB.getLearningSteps(learnerId, courseId, inspectionSize)) {
                ccIris.add(lStep.getConceptContainerIri());
            }
            return ccIris;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return List of CCs the learner has accessed (0=newest), or empty List on error.
     */
    public List<String> getHistoryOfCCs(String learnerId, String courseId) {
        try {
            List<String> ccIris = new ArrayList<>();
            for (LearningStep lStep : DB.getLearningSteps(learnerId, courseId, Integer.MAX_VALUE)) {
                ccIris.add(lStep.getConceptContainerIri());
            }
            return ccIris;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param inspectionSize Number of elements to load
     * @return List of CCs the learner has accessed (0=newest), or empty List on error.
     */
    public List<String> getHistoryOfKOs(String learnerId, String courseId, int inspectionSize) {
        try {
            List<String> koIris = new ArrayList<>();
            for (LearningStep lStep : DB.getLearningSteps(learnerId, courseId, inspectionSize)) {
                koIris.add(lStep.getKnowledgeObjectIri());
            }
            return koIris;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return List of CCs the learner has accessed (0=newest), or empty List on error.
     */
    public List<String> getHistoryOfKOs(String learnerId, String courseId) {
        try {
            List<String> koIris = new ArrayList<>();
            for (LearningStep lStep : DB.getLearningSteps(learnerId, courseId, Integer.MAX_VALUE)) {
                koIris.add(lStep.getKnowledgeObjectIri());
            }
            return koIris;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope.
     * @return The LmsId of the last KO the learner has accessed. Null on error or if not available.
     */
    public String getLastKnowledgeObjectLmsId(String learnerId) {
        try {
            List<LearningStep> lSteps = DB.getLearningSteps(learnerId, 1);
            if (lSteps.isEmpty()) {
                return null;
            }
            KnowledgeObject ko = DB.getKo(lSteps.get(0).getKnowledgeObjectIri());
            if (ko == null) {
                return null;
            }
            return ko.getLmsId();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope.
     * @param courseId ID of the course in scope
     * @return The LmsId of the last KO the learner has accessed. Null on error or if not available.
     */
    public String getLastKnowledgeObjectLmsId(String learnerId, String courseId) {
        try {
            List<LearningStep> lSteps = DB.getLearningSteps(learnerId, courseId, 1);
            if (lSteps.isEmpty()) {
                return null;
            }
            KnowledgeObject ko = DB.getKo(lSteps.get(0).getKnowledgeObjectIri());
            if (ko == null) {
                return null;
            }
            return ko.getLmsId();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope.
     * @return The IRI of the last KO the learner has accessed. Null on error or if not available.
     */
    public String getLastKnowledgeObjectIri(String learnerId) {
        try {
            List<LearningStep> lSteps = DB.getLearningSteps(learnerId, 1);
            if (lSteps.isEmpty()) {
                return null;
            }
            return lSteps.get(0).getKnowledgeObjectIri();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope.
     * @return The IRI of the last KO the learner has accessed. Null on error or if not available.
     */
    public String getLastConceptContainerIri(String learnerId) {
        try {
            List<LearningStep> lSteps = DB.getLearningSteps(learnerId, 1);
            if (lSteps.isEmpty()) {
                return null;
            }
            return lSteps.get(0).getConceptContainerIri();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return The IRI of the last KO the learner has accessed. Null on error or if not available.
     */
    public String getLastKnowledgeObjectIri(String learnerId, String courseId) {
        try {
            List<LearningStep> lSteps = DB.getLearningSteps(learnerId, courseId, 1);
            if (lSteps.isEmpty()) {
                return null;
            }
            return lSteps.get(0).getKnowledgeObjectIri();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * Determines the summed up time the learner has spent on the KO.
     *
     * @param learnerId ID of the learner in scope
     * @param koIri ID of the KO in scope
     * @return 0 if no time was spent yet. Null in case of error.
     */
    public Long getLearningTime(String learnerId, String koIri) {
        try {
            long sum = 0;
            for (LearningStep lStep : DB.getLearningStepsWithKoIri(learnerId, koIri)) {
                if (lStep.getStopTime() != null) {
                    Long start = lStep.getId().getStartTime().getTime();
                    Long stop = lStep.getStopTime().getTime();
                    sum += stop - start;
                }
            }
            return sum;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param koIri IRI of the current KO
     * @param ccIri IRI of the current CC
     * @param time Time when the learning step was registered
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeLearningStep(String learnerId, String courseId, String koIri, String ccIri, long time) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                logger.error("Could not store new learning step. There are no learning sessions registered for the "
                        + "learner.");
                return false;
            }
            if (s.getLogoutTime() != null) {
                logger.error("Could not store new learning step. There is no running learning session registered for "
                        + "the learner.");
                return false;
            }

            LearningStep lStep = learningStep(s.getId().getLearnerId(), s.getId().getLoginTime(),
                    s.getId().getCourseId(), new Date(time), false, koIri, ccIri);

            //Finish previous step
            LearningStep previousStep = DB.getLastLearningStep(learnerId, courseId);
            if (previousStep != null) {
                previousStep.setStopTime(lStep.getId().getStartTime());
                DB.storeOrUpdate(previousStep);
            }

            //Store current step
            DB.storeOrUpdate(lStep);

            //Create empty scores if not accessed before
            if (DB.getCcLastScore(learnerId, ccIri) == null) {
                storeCcCompletion(learnerId, courseId, ccIri, false);
            }
            if (DB.getLastKoScore(learnerId, koIri) == null) {
                storeKoCompletionAccessed(learnerId, courseId, koIri, true);
            }

            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Null in case there is no last learning step or if an error occured.
     */
    private LearningStep getLastLearningStep(String learnerId) {
        try {
            List<LearningStep> steps = DB.getLearningSteps(learnerId, 1);
            if (steps == null || steps.isEmpty()) {
                return null;
            }
            return steps.get(0);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Null in case of error or if unknown.
     */
    public Long getTimeSinceLastLpEvaluation(String learnerId, String courseId) {
        try {
            for (LearningStep step : DB.getLearningSteps(learnerId, courseId, Integer.MAX_VALUE)) {
                if (step.isLearningPathwayEval()) {
                    return System.currentTimeMillis() - step.getId().getStartTime().getTime();
                }
            }
            return null;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Null in case of error or if unknown.
     */
    public Integer getStepsSinceLastLpEvaluation(String learnerId, String courseId) {
        try {
            int ctr = 0;
            for (LearningStep step : DB.getLearningSteps(learnerId, courseId, Integer.MAX_VALUE)) {
                if (step.isLearningPathwayEval()) {
                    return ctr;
                } else {
                    ctr++;
                }
            }
            return null;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeLpEvaluationIncident(String learnerId, String courseId) {
        try {
            LearningStep step = DB.getLastLearningStep(learnerId, courseId);
            step.setLearningPathwayEval(true);
            DB.storeOrUpdate(step);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param stepsCount Number of learnig steps to be inspected
     * @return A list of learning steps containing a list of learner recommendations for that step
     * (KoIri, RecomValue). Returns an empty list if no data available or if an error occured.
     */
    public List<List<Pair<String, Double>>> getLearningRecommendations(String learnerId, String courseId,
            int stepsCount) {
        try {
            List<List<Pair<String, Double>>> steps = new ArrayList<>();
            for (LearningStep learningStep : DB.getLearningSteps(learnerId, courseId, stepsCount)) {
                List<Pair<String, Double>> recoms = new ArrayList<>();
                for (LearningRecommendation learningRecom : DB.getLearningRecommendations(learningStep.getId())) {
                    Pair<String, Double> recom = Pair.of(learningRecom.getId().getKnowledgeObjectIri(),
                            learningRecom.getRecommendationValue());
                    recoms.add(recom);
                }
                steps.add(recoms);
            }
            return steps;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param koIri IRI of the KO for which a recommendation was given
     * @param recomValue Recommendation value of the KO
     * @return True if successfull.
     */
    public boolean storeLearningRecommendation(String learnerId, String koIri, Double recomValue) {
        try {
            LearningStep lStep = getLastLearningStep(learnerId);
            if (lStep == null) {
                logger.error("Could not store learning recommendation, because there is no learning step for the "
                        + "learner.");
                return false;
            }
            LearningRecommendation lRecom = learningRecommendation(lStep.getId().getSessionLearnerId(),
                    lStep.getId().getSessionLoginTime(),
                    lStep.getId().getSessionCourseId(),
                    lStep.getId().getStartTime(), koIri, recomValue);
            DB.storeOrUpdate(lRecom);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return false;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return List containing the the past recommendations (Pair.left) and the selection
     * (Pair.right) the learner made. The returned list ignores eventually existing empty learner
     * updates, e.g. when a reflex aborted the recommendation process. <br>
     * The returned List is never null, but could be empty. The contained Pair might contain null
     * values.
     */
    public List<Pair<List<String>, String>> getRecommendationSelections(String learnerId, String courseId) {
        try {
            List<Pair<List<String>, String>> recomSelections = new ArrayList<>();
            List<LearningStep> lSteps = DB.getLearningSteps(learnerId, courseId, Integer.MAX_VALUE);

            //Each LearningStep can have 0-* LearningRecommendations
            //Relevant for this method are 
            //1) LearningRecommendations and 
            //2) what the learner selected in the next step
            for (int i = 1; i < lSteps.size(); i++) {
                //i=1, because the current step is irrelevant - there is no recom for it                
                LearningStep step = lSteps.get(i);
                List<LearningRecommendation> lRecoms = DB.getLearningRecommendations(step.getId());
                if (!lRecoms.isEmpty()) {
                    LearningStep nextStep = lSteps.get(i - 1);
                    List<String> lRecomIris = new ArrayList<>();
                    for (LearningRecommendation lRecom : lRecoms) {
                        lRecomIris.add(lRecom.getId().getKnowledgeObjectIri());
                    }
                    recomSelections.add(Pair.of(lRecomIris, nextStep.getKnowledgeObjectIri()));
                }
            }

            return recomSelections;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return new ArrayList<>();
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Empty list on error or on no data.
     */
    public List<Long> getSessionDurations(String learnerId, String courseId) {
        try {
            List<Long> durations = new ArrayList<>();
            for (Session s : DB.getSessions(learnerId, courseId)) {
                if (s.getLogoutTime() != null) {
                    Long logout = s.getLogoutTime().getTime();
                    Long login = s.getId().getLoginTime().getTime();
                    durations.add(logout - login);
                }
            }
            return durations;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * Determines if there is a running session for the learner.
     *
     * @param learnerId ID of the learner in scope.
     * @return False if there is no running session, otherwise true.
     */
    public boolean isLoggedIn(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            return s != null && s.getLogoutTime() == null;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return false;
        }
    }

    /**
     * Creates a new session for the specified learner.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course the learner is working on
     * @return False in case login failed.
     */
    public boolean loginUser(String learnerId, String courseId) {
        try {
            if (isLoggedIn(learnerId)) {
                logger.error("Could not login learner. There already is a running session.");
                return false;
            }
            Session newS = new Session(new SessionId(learnerId, getNow(), courseId));
            DB.storeOrUpdate(newS);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return false;
        }
    }

    /**
     * Ends the currently active session of the learner.
     *
     * @param learnerId ID of the learner in scope
     * @return False in case the current session could not be closed or in case of error.
     */
    public boolean logoutUser(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                logger.error("Could not logout learner. There is no session.");
                return false;
            }
            if (s.getLogoutTime() != null) {
                logger.error("Could not logout learner. There is no running session.");
                return false;
            }
            s.setLogoutTime(getNow());

            LearningStep currentStep = DB.getLastLearningStep(learnerId, s.getId().getCourseId());
            if (currentStep != null) {
                currentStep.setStopTime(s.getLogoutTime());
                DB.storeOrUpdate(currentStep);
            }

            DB.storeOrUpdate(s);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return false;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param deviceType Type of devive used for learning
     * @param deviceDpi DPI of the device used for learning
     * @param deviceTime Current time on the device
     * @param deviceVerticalResolution Vertical resolution of the device used for learning
     * @param deviceHorizontalResolution Horizontal resolution of the device used for learning
     * @param interactionWillingness Current willingness of the learner to interact with peers
     * @param learningAttitude Current learning attitude of the learner
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeStaticSessionData(String learnerId, String deviceType, String deviceDpi, Long deviceTime,
            Integer deviceVerticalResolution, Integer deviceHorizontalResolution,
            InteractionWillingness interactionWillingness,
            LearningAttitude learningAttitude) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                logger.error("Could not set session data for the learner. There is no session.");
                return false;
            }
            if (s.getLogoutTime() != null) {
                logger.error("Could not set session data for the learner. There is no running session.");
                return false;
            }
            s.setDeviceType(deviceType);
            s.setDeviceDpi(deviceDpi);
            s.setDeviceVerticalResolution(deviceVerticalResolution);
            s.setDeviceHorizontalResolution(deviceHorizontalResolution);
            s.setDeviceTime(deviceTime == null ? null : new Date(deviceTime));
            s.setInteractionWillingness(InteractionWillingness.convertToString(interactionWillingness));
            s.setLearningAttitude(LearningAttitude.getLearningAttitude(learningAttitude));
            DB.storeOrUpdate(s);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Null on error or if no data available.
     */
    public Integer getDeviceHorizontalResolution(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                return null;
            }
            if (s.getLogoutTime() != null) {
                return null;
            }
            return s.getDeviceHorizontalResolution();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Null on error or if no data available.
     */
    public Integer getDeviceVerticalResolution(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                return null;
            }
            if (s.getLogoutTime() != null) {
                return null;
            }
            return s.getDeviceVerticalResolution();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Null on error or if no data available.
     */
    public String getDeviceType(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                return null;
            }
            if (s.getLogoutTime() != null) {
                return null;
            }
            return s.getDeviceType();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Null on error or if no data available.
     */
    public String getDeviceDpi(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                return null;
            }
            if (s.getLogoutTime() != null) {
                return null;
            }
            return s.getDeviceDpi();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Null on error or if no data available.
     */
    public InteractionWillingness getInteractionWillingness(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                return null;
            }
            if (s.getLogoutTime() != null) {
                return null;
            }
            return InteractionWillingness.convertToEnum(s.getInteractionWillingness());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Null on error or if no data available.
     */
    public LearningAttitude getLearningAttitude(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                return null;
            }
            if (s.getLogoutTime() != null) {
                return null;
            }
            return LearningAttitude.getLearningAttitude(s.getLearningAttitude());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Time of current login or null on exception.
     */
    public Long getLoginTime(String learnerId, String courseId) {
        try {
            List<Session> sList = DB.getSessions(learnerId, courseId, 1);
            if (sList == null) {
                return null;
            }
            if (sList.isEmpty()) {
                return null;
            }
            return sList.get(0).getId().getLoginTime().getTime();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Time of previous login or null on exception.
     */
    public Long getPreviousLoginTime(String learnerId, String courseId) {
        try {
            List<Session> sList = DB.getSessions(learnerId, courseId, 2);
            if (sList == null) {
                return null;
            }
            if (sList.size() < 2) {
                return null;
            }
            return sList.get(1).getId().getLoginTime().getTime();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param batteryLvl The current battery level (0-100)
     * @param connectionType The current type of connection the learner uses
     * @param connectionStability The current stability of the connection (0-100)
     * @param noiseLvl The current level of environment noise around the learner (decibel value)
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeTimedependentSessionData(String learnerId, Integer batteryLvl,
            ConnectionType connectionType, Integer connectionStability,
            Integer noiseLvl) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                logger.error("Could not set time-dependent session data for the learner. There is no session.");
                return false;
            }
            if (s.getLogoutTime() != null) {
                logger.error("Could not set time-dependent session data for the learner. There is no running session.");
                return false;
            }
            TimedependentSessionData data = timedependentSessionData(s.getId().getLearnerId(),
                    s.getId().getLoginTime(),
                    s.getId().getCourseId(),
                    getNow(), connectionType.getValue(),
                    connectionStability, noiseLvl, batteryLvl);
            DB.storeOrUpdate(data);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Null in case of error or if no data is available
     */
    public Integer getBatteryLevel(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                logger.error("Could not get time-dependent session data for the learner. There is no session.");
                return null;
            }
            if (s.getLogoutTime() != null) {
                logger.error("Could not get time-dependent session data for the learner. There is no running session.");
                return null;
            }
            TimedependentSessionData data = DB.getLastTimedependentSessionData(s.getId());
            if (data == null) {
                return null;
            }
            return data.getBatteryLevel();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Null in case of error or if no data is available
     */
    public Integer getNoiseLevel(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                logger.error("Could not get time-dependent session data for the learner. There is no session.");
                return null;
            }
            if (s.getLogoutTime() != null) {
                logger.error("Could not get time-dependent session data for the learner. There is no running session.");
                return null;
            }
            TimedependentSessionData data = DB.getLastTimedependentSessionData(s.getId());
            if (data == null) {
                return null;
            }
            return data.getNoiseLevel();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Null in case of error or if no data is available
     */
    public ConnectionType getConnectionType(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                logger.error("Could not get time-dependent session data for the learner. There is no session.");
                return null;
            }
            if (s.getLogoutTime() != null) {
                logger.error("Could not get time-dependent session data for the learner. There is no running session.");
                return null;
            }
            TimedependentSessionData data = DB.getLastTimedependentSessionData(s.getId());
            if (data == null) {
                return null;
            }
            return ConnectionType.getConnectionType(data.getConnectionType());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @return Null in case of error or if no data is available
     */
    public Integer getConnectionStability(String learnerId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                logger.error("Could not get time-dependent session data for the learner. There is no session.");
                return null;
            }
            if (s.getLogoutTime() != null) {
                logger.error("Could not get time-dependent session data for the learner. There is no running session.");
                return null;
            }
            TimedependentSessionData data = DB.getLastTimedependentSessionData(s.getId());
            if (data == null) {
                return null;
            }
            return data.getConnectionStability();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param eqfLevel Preferred EQF level to set
     * @param difficultyLevel Preferred ifficulty level to set
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeLearnerPreference(String learnerId, String courseId, EqfLevel eqfLevel,
            DifficultyLevel difficultyLevel) {
        try {
            LearnerPreference lPref = DB.getLearnerPreference(learnerId, courseId);
            if (lPref == null) {
                lPref = new LearnerPreference(new LearnerPreferenceId(learnerId, courseId),
                        eqfLevel.getLevel(), difficultyLevel.getLevel());
            }
            DB.storeOrUpdate(lPref);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Null on error or on no data
     */
    public DifficultyLevel getPreferredDifficultyLevel(String learnerId, String courseId) {
        try {
            LearnerPreference lPref = DB.getLearnerPreference(learnerId, courseId);
            if (lPref == null) {
                return null;
            }
            return DifficultyLevel.getLevel(lPref.getDifficultyLevel());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    public void setPreferredDifficultyLevel(String learnerId, String courseId, DifficultyLevel lvl) {
        try {
            LearnerPreference l = DB.getLearnerPreference(learnerId, courseId);
            if (l == null) {
                logger.error("Unable to retrieve the learner prefference...");
            } else {
                l.setDifficultyLevel(lvl.getLevel());
                DB.storeOrUpdate(l);
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return true if the operation is successful
     */
    public boolean increasePreferredDifficulty(String learnerId, String courseId) {
        try {
            LearnerPreference lPref = DB.getLearnerPreference(learnerId, courseId);
            if (lPref == null) {
                logger.error("Could not retrieve the learner preference for uid:" + learnerId);
                return false;
            } else {
                DifficultyLevel cLvl = DifficultyLevel.getLevel(lPref.getDifficultyLevel());
                DifficultyLevel dLvl = DifficultyLevel.increase(cLvl);
                if (cLvl != dLvl) {
                    lPref.setDifficultyLevel(dLvl.getLevel());
                    DB.storeOrUpdate(lPref);
                    return true;
                }
                return false;
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return false;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return true if the operation is successful
     */
    public boolean decreasePreferredDifficulty(String learnerId, String courseId) {
        try {
            LearnerPreference lPref = DB.getLearnerPreference(learnerId, courseId);
            if (lPref == null) {
                logger.error("Could not retrieve the learner preference for uid:" + learnerId);
                return false;
            } else {
                DifficultyLevel cLvl = DifficultyLevel.getLevel(lPref.getDifficultyLevel());
                DifficultyLevel dLvl = DifficultyLevel.decrease(cLvl);
                if (cLvl != dLvl) {
                    lPref.setDifficultyLevel(dLvl.getLevel());
                    DB.storeOrUpdate(lPref);
                    return true;
                }
                return false;
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return false;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Null on error or on no data
     */
    public EqfLevel getPreferredEqfLevel(String learnerId, String courseId) {
        try {
            LearnerPreference lPref = DB.getLearnerPreference(learnerId, courseId);
            if (lPref == null) {
                return null;
            }
            return EqfLevel.getLevel(lPref.getEqfLevel());
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    public void setPreferredEqfLevel(String learnerId, String courseId, EqfLevel lvl) {
        try {
            LearnerPreference l = DB.getLearnerPreference(learnerId, courseId);
            if (l == null) {
                logger.error("Unable to retrieve the learner prefference...");
            } else {
                l.setEqfLevel(lvl.getLevel());
                DB.storeOrUpdate(l);
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    public Boolean restartMacroLpSelection(String learnerId, String courseId) {
        return storeMacroLpSelection(learnerId, null, courseId) && storeMicroLpSelection(learnerId, null, courseId);
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param lpIri IRI of the LP in scope
     * @param courseId ID of the course in scope
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeMacroLpSelection(String learnerId, String lpIri, String courseId) {
        try {
            MacroLearningPathwaySelection item = macroLearningPathwaySelection(lpIri, learnerId, courseId, getNow());
            DB.storeOrUpdate(item);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param lpIri IRI of the LP in scope
     * @param courseId ID of the course in scope
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeMicroLpSelection(String learnerId, String lpIri, String courseId) {
        try {
            MicroLearningPathwaySelection item = microLearningPathwaySelection(lpIri, learnerId, courseId, getNow());
            DB.storeOrUpdate(item);
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return IRI of the macro LP of the leanrer. Null if unknown or error.
     */
    public String getMacroLpOfLearner(String learnerId, String courseId) {
        try {
            MacroLearningPathwaySelection selection = DB.getLastMacroLpSelection(learnerId, courseId);
            if (selection == null) {
                return null;
            }
            return selection.getId().getMacroLearningPathwayIri();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return IRI of the micro LP of the leanrer. Null if unknown or error.
     */
    public String getMicroLpOfLearner(String learnerId, String courseId) {
        try {
            MicroLearningPathwaySelection selection = DB.getLastMicroLpSelection(learnerId, courseId);
            if (selection == null) {
                return null;
            }
            return selection.getId().getMicroLearningPathwayIri();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param lpIri IRI of the LP in scope
     * @param courseId ID of the course of the LP in scope
     * @return Percentage value (0-100) or null if unknown or error.
     */
    public Integer getPopularityOfMicroLp(String lpIri, String courseId) {
        try {
            Map<String, java.util.Date> dateMap = new HashMap<>();
            Map<String, Boolean> lpMap = new HashMap<>();
            for (MicroLearningPathwaySelection selection : DB.getMicroLpSelections(courseId)) {
                java.util.Date storedDate = dateMap.get(selection.getId().getLearnerId());
                java.util.Date current = selection.getId().getSelectionTime();
                if (storedDate == null || storedDate.before(current)) {
                    dateMap.put(selection.getId().getLearnerId(), current);
                    if (selection.getId().getMicroLearningPathwayIri().equals(lpIri)) {
                        lpMap.put(selection.getId().getLearnerId(), true);
                    } else {
                        lpMap.put(selection.getId().getLearnerId(), false);
                    }
                }
            }
            Integer total = 0;
            Integer ctr = 0;
            for (Map.Entry<String, Boolean> entry : lpMap.entrySet()) {
                total++;
                if (entry.getValue()) {
                    ctr++;
                }
            }
            if (total == 0) {
                return 0;
            }
            return ctr * 100 / total;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param lpIri IRI of the LP in scope
     * @param courseId ID of the course of the LP in scope
     * @return Percentage value (0-100) or null if unknown or error.
     */
    public Integer getPopularityOfMacroLp(String lpIri, String courseId) {
        try {
            Map<String, java.util.Date> dateMap = new HashMap<>();
            Map<String, Boolean> lpMap = new HashMap<>();
            for (MacroLearningPathwaySelection selection : DB.getMacroLpSelections(courseId)) {
                java.util.Date storedDate = dateMap.get(selection.getId().getLearnerId());
                java.util.Date current = selection.getId().getSelectionTime();
                if (storedDate == null || storedDate.before(current)) {
                    dateMap.put(selection.getId().getLearnerId(), current);
                    if (selection.getId().getMacroLearningPathwayIri().equals(lpIri)) {
                        lpMap.put(selection.getId().getLearnerId(), true);
                    } else {
                        lpMap.put(selection.getId().getLearnerId(), false);
                    }
                }
            }
            Integer total = 0;
            Integer ctr = 0;
            for (Map.Entry<String, Boolean> entry : lpMap.entrySet()) {
                total++;
                if (entry.getValue()) {
                    ctr++;
                }
            }
            if (total == 0) {
                return 0;
            }
            return ctr * 100 / total;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * Retrives the session lenghts of the learner's peers in the specified course.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Empty collection is nothing was found. Null on error.
     */
    public Collection<Collection<Long>> getPeerSessionLengths(String learnerId, String courseId) {
        try {
            Map<String, Collection<Long>> sessionLenghts = new HashMap<>();
            for (Session s : DB.getSessions(courseId)) {
                if (s.getLogoutTime() != null) {
                    Long duration = s.getLogoutTime().getTime() - s.getId().getLoginTime().getTime();
                    if (sessionLenghts.containsKey(s.getId().getLearnerId())) {
                        sessionLenghts.get(s.getId().getLearnerId()).add(duration);
                    } else {
                        List<Long> durations = new ArrayList<>();
                        durations.add(duration);
                        sessionLenghts.put(s.getId().getLearnerId(), durations);
                    }
                }
            }
            Collection<Collection<Long>> result = new HashSet<>();
            for (Map.Entry<String, Collection<Long>> entry : sessionLenghts.entrySet()) {
                if (!entry.getKey().equalsIgnoreCase(learnerId)) {
                    result.add(entry.getValue());
                }
            }
            return result;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * Retrieves the learnerIds of the course-peers of the specified learner.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Empty collection is nothing was found. Null on error.
     */
    public Collection<String> getPeerIds(String learnerId, String courseId) {
        try {
            Collection<String> result = new HashSet<>();
            for (Session s : DB.getSessions(courseId)) {
                if (!s.getId().getLearnerId().equals(learnerId)) {
                    result.add(s.getId().getLearnerId());
                }
            }
            return result;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param iri IRI of the SLOM element
     * @param courseId ID of the course
     * @param title Title of the element
     * @param description Description of the element
     * @param lmsId ID of the KO in the LMS
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeKoData(String iri, String courseId, String title, String description, String lmsId) {
        try {
            DB.storeOrUpdate(new KnowledgeObject(iri, courseId, lmsId, title, description));
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param iri IRI of the SLOM element
     * @param courseId ID of the course
     * @param title Title of the element
     * @param description Description of the element
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeCcData(String iri, String courseId, String title, String description) {
        try {
            DB.storeOrUpdate(new ConceptContainer(iri, courseId, title, description));
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param iri IRI of the SLOM element
     * @param courseId ID of the course
     * @param title Title of the element
     * @param description Description of the element
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeMicroLpData(String iri, String courseId, String title, String description) {
        try {
            DB.storeOrUpdate(new MicroLearningPathway(iri, courseId, title, description));
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param iri IRI of the SLOM element
     * @param courseId ID of the course
     * @param title Title of the element
     * @param description Description of the element
     * @return Null in case of error, false if failed, true if successfull.
     */
    public Boolean storeMacroLpData(String iri, String courseId, String title, String description) {
        try {
            DB.storeOrUpdate(new MacroLearningPathway(iri, courseId, title, description));
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param iri IRI of the KO in scope
     * @return Null in case nothing was found or an error occured.
     */
    public KnowledgeObject getKoData(String iri) {
        try {
            return DB.getKo(iri);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param iri IRI of the CC in scope
     * @return Null in case nothing was found or an error occured.
     */
    public ConceptContainer getCcData(String iri) {
        try {
            return DB.getCc(iri);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param iri IRI of the mLP in scope
     * @return Null in case nothing was found or an error occured.
     */
    public MicroLearningPathway getMicroLpData(String iri) {
        try {
            return DB.getMicroLp(iri);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * @param iri IRI of the MLP in scope
     * @return Null in case nothing was found or an error occured.
     */
    public MacroLearningPathway getMacroLpData(String iri) {
        try {
            return DB.getMacroLp(iri);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    public String getCcTitle(String iri) {
        try {
            ConceptContainer cc = DB.getCc(iri);
            if (cc == null) {
                return null;
            } else {
                return cc.getTitle();
            }
        } catch (DbAccessException e) {
            logger.error(e.getMessage(), e);
            return null;
        }
    }

    public String getKoTitle(String iri) {
        try {
            KnowledgeObject ko = DB.getKo(iri);
            if (ko == null) {
                return null;
            } else {
                return ko.getTitle();
            }
        } catch (DbAccessException e) {
            logger.error(e.getMessage(), e);
            return null;
        }
    }

    /**
     * Resets the score values of the learner for the specified CC.
     *
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param ccIri IRI of the CC in scope
     * @return True if successfull, false if not successfull, null on error.
     */
    public Boolean resetCcScore(String learnerId, String courseId, String ccIri) {
        try {
            //To reset a CC, you have to add a new CC score stating that it was not completed.
            //Additionally, all conected KOs have to be reset.
            //A KO is reset by adding a new KO score stating that it has not been accessed.

            List<LearningStep> lSteps = DB.getLearningStepsWithCcIri(learnerId, ccIri);
            if (lSteps.isEmpty()) {
                return false;
            }
            ConceptContainerScore ccScore = conceptContainerScore(learnerId, ccIri, courseId, getNow(), false);
            DB.storeOrUpdate(ccScore);
            for (int i = lSteps.size() - 1; i >= 0; i--) {
                LearningStep lStep = lSteps.get(i);
                String koIri = lStep.getKnowledgeObjectIri();
                KnowledgeObjectScore koScore = knowledgeObjectScore(learnerId, courseId, koIri, getNow(), false);
                DB.storeOrUpdate(koScore);
            }
            return true;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
            return null;
        }
    }

    /**
     * Clears all DB tables.
     * Only for test cleanup!
     */
    public void clearDataStorage() {
        try {
            DB.clear();
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     */
    void removeLastSession(String learnerId, String courseId) {
        try {
            Session s = DB.getLastSession(learnerId);
            if (s == null) {
                return;
            }
            for (TimedependentSessionData sDat : DB.getTimedependentSessionData(s.getId())) {
                DB.remove(sDat);
            }
            DB.remove(s);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     */
    void removeLastLearningStep(String learnerId, String courseId) {
        try {
            LearningStep step = DB.getLastLearningStep(learnerId, courseId);
            if (step == null) {
                return;
            }
            removeKoScore(learnerId, step.getKnowledgeObjectIri());
            removeCcScore(learnerId, step.getConceptContainerIri());
            DB.remove(step);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param iri IRI in scope
     */
    void removeKoData(String iri) {
        try {
            KnowledgeObject ko = DB.getKo(iri);
            DB.remove(ko);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param iri IRI in scope
     */
    void removeCcData(String iri) {
        try {
            ConceptContainer cc = DB.getCc(iri);
            DB.remove(cc);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param iri IRI in scope
     */
    void removeMicroLpData(String iri) {
        try {
            MicroLearningPathway item = DB.getMicroLp(iri);
            DB.remove(item);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param iri IRI in scope
     */
    void removeMacroLpData(String iri) {
        try {
            MacroLearningPathway item = DB.getMacroLp(iri);
            DB.remove(item);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner
     * @param iri IRI in scope
     */
    void removeKoScore(String learnerId, String iri) {
        try {
            for (KnowledgeObjectScore score : DB.getKoScores(learnerId, iri)) {
                DB.remove(score);
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner
     * @param iri IRI in scope
     */
    void removeCcScore(String learnerId, String iri) {
        try {
            for (ConceptContainerScore score : DB.getCcScores(learnerId, iri)) {
                DB.remove(score);
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner
     * @param courseId ID of the course
     */
    void removeLastMicroLpSelection(String learnerId, String courseId) {
        try {
            MicroLearningPathwaySelection val = DB.getLastMicroLpSelection(learnerId, courseId);
            DB.remove(val);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner
     * @param courseId ID of the course
     */
    void removeLastMacroLpSelection(String learnerId, String courseId) {
        try {
            MacroLearningPathwaySelection val = DB.getLastMacroLpSelection(learnerId, courseId);
            DB.remove(val);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner
     * @param courseId ID of the course
     */
    void removeLearnerPreference(String learnerId, String courseId) {
        try {
            LearnerPreference val = DB.getLearnerPreference(learnerId, courseId);
            DB.remove(val);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner
     */
    void removeLearner(String learnerId) {
        try {
            Learner val = DB.getLearner(learnerId);
            DB.remove(val);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * Removes the last LMS parameters that have been stored.
     */
    void removeLastLmsParameters() {
        try {
            LmsParameter val = DB.getLastLmsParameter();
            DB.remove(val);
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * @param learnerId ID of the learner in scope
     */
    void removeLastLearningRecommendation(String learnerId) {
        try {
            LearningStep lStep = getLastLearningStep(learnerId);
            if (lStep != null) {
                List<LearningRecommendation> lRecoms = DB.getLearningRecommendations(lStep.getId());
                if (!lRecoms.isEmpty()) {
                    DB.remove(lRecoms.get(0));
                }
            }
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
    }

    /**
     * Checks whether the specified learner is registered in the learner table.
     *
     * @param learnerId
     * @return True if registered
     */
    public boolean isLearnerRegistered(String learnerId) {
        try {
            return DB.getLearner(learnerId) == null;
        } catch (DbAccessException ex) {
            logger.error(ex.getMessage(), ex);
        }
        return false;
    }
}
