/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Ovidiu Serban, University of Reading, www.reading.ac.uk
 * Contributor: Jaime Costa, University of Reading, www.reading.ac.uk
 * Version: [0.01]
 * Date: 05.11.2013
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.db.access;

import eu.intuitel.db.model.*;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.hibernate.*;
import org.hibernate.Session;
import org.hibernate.boot.registry.StandardServiceRegistryBuilder;
import org.hibernate.cfg.Configuration;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.Projections;
import org.hibernate.criterion.Restrictions;
import org.hibernate.exception.JDBCConnectionException;
import org.hibernate.service.ServiceRegistry;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

public class HibernateDbAccessPoint implements IDbAccessPoint {
    private static final Logger LOGGER = LoggerFactory.getLogger(HibernateDbAccessPoint.class);
    private static final SessionFactory SESSION_FACTORY; //Factory is threadsafe, i.e. static final is OK, do not use it directly

    static {
        SessionFactory tmpFactory = null;
        try {
            Configuration configuration = new Configuration().configure();
            ServiceRegistry serviceRegistry = new StandardServiceRegistryBuilder()
                                                      .applySettings(configuration.getProperties())
                                                      .build();
            tmpFactory = configuration.buildSessionFactory(serviceRegistry);
        } catch (JDBCConnectionException ex) {
            if (ex.getMessage().equalsIgnoreCase("error calling driver#connect")) {
                LOGGER.error("Could not establish connection to database: " + ex.getMessage());
            } else {
                LOGGER.error("Connection error on setting up session factory.", ex);
            }
        } catch (HibernateException ex) {
            LOGGER.error("Hibernate exception on setting up session factory.", ex);
        }
        SESSION_FACTORY = tmpFactory;
    }

    private static SessionFactory getSession() throws DbAccessException {
        if (SESSION_FACTORY == null) {
            throw new DbAccessException("The HIBERNATE connection failed ...");
        }
        return SESSION_FACTORY;
    }

    public HibernateDbAccessPoint() {
    }

    @Override
    public boolean isConnectionEstablished() {
        return SESSION_FACTORY != null;
    }

    @Override
    public void shutdown() {
        try {
            SessionFactory sessionFactory = getSession();
            if (!sessionFactory.isClosed()) {
                sessionFactory.close();
            }
        } catch (HibernateException | DbAccessException ex) {
            LOGGER.error("Error shutting down DB connection.", ex);
        }
    }

    @Override
    public void storeOrUpdate(IDbItem item) throws DbAccessException {
        if (item != null) {
            updateOrSaveData(item);
        } else {
            LOGGER.error("Cannot store item, because it is null.");
        }
    }

    /**
     * Updates or saves the given items in the database.
     *
     * @param dataList Items in scope
     * @throws DbAccessException
     */
    public void updateOrSaveData(IDbItem... dataList) throws DbAccessException {
        Session s = getSession().openSession();
        try {
            s.beginTransaction();
            for (IDbItem item : dataList) {
                s.saveOrUpdate(item);
            }
            s.getTransaction().commit();
        } catch (HibernateException ex) {
            throw new DbAccessException(ex.getMessage(), ex);
        } finally {
            if (s != null) {
                s.close();
            }
        }
    }

    /**
     * Executes a query and retrieves the data that satisfies the specified parameters.
     *
     * @param <T>                    The sub-class of IDbItem being boxed
     * @param clazz                  The sub-class of IDbItem being boxed
     * @param restrictionColumn      Names of the columns to use for restricting the result. May be null
     *                               or empty.
     * @param restrictionColumnValue Values of the columns to use for restricting the result. May be
     *                               null or empty.
     * @param oderColumn             Name of the column to order the result
     * @param orderAscending         If true, the result will be given in ascending order. If false or null,
     *                               result will be in descending order.
     * @param queryLimit             Maximum number of returned items, may be null
     * @param distinctColumn         Name of the column for DISTINCT filtering
     * @return Empty list if the query returned no data.
     * @throws DbAccessException To be thrown in case of exception during query execution.
     */

    @SuppressWarnings("unchecked")
    public <T extends IDbItem> List<T> queryDbData(Class<T> clazz,
                                                   String[] restrictionColumn, Object[] restrictionColumnValue,
                                                   String oderColumn, Boolean orderAscending,
                                                   Integer queryLimit, String distinctColumn) throws DbAccessException {
        testRestrictionData(restrictionColumn, restrictionColumnValue);
        Session s = getSession().openSession();
        try {
            Criteria criteria = s.createCriteria(clazz);
            if (restrictionColumn != null) {
                for (int i = 0; i < restrictionColumn.length; i++) {
                    criteria.add(Restrictions.eq(restrictionColumn[i], restrictionColumnValue[i]));
                }
            }
            if (!StringUtils.isBlank(oderColumn)) {
                if (orderAscending == null ? false : orderAscending) {
                    criteria.addOrder(Order.asc(oderColumn));
                } else {
                    criteria.addOrder(Order.desc(oderColumn));
                }
            }
            if (!StringUtils.isBlank(distinctColumn)) {
                criteria.setProjection(Projections.distinct(Projections.property(distinctColumn)));
            }

            if (queryLimit != null) {
                criteria.setMaxResults(queryLimit);
            }

            return (List<T>) criteria.list();
        } catch (HibernateException ex) {
            throw new DbAccessException(ex.getMessage(), ex);
        } finally {
            if (s != null) {
                s.close();
            }
        }
    }

    /**
     * Executes a query and retrieves the data that satisfies the specified parameters.
     *
     * @param <T>                    The sub-class of IDbItem being boxed
     * @param clazz                  The sub-class of IDbItem being boxed
     * @param restrictionColumn      Names of the columns to use for restricting the result. May be null
     *                               or empty.
     * @param restrictionColumnValue Values of the columns to use for restricting the result. May be
     *                               null or empty.
     * @param oderColumn             Name of the column to order the result
     * @param orderAscending         If true, the result will be given in ascending order. If false or null,
     *                               result will be in descending order.
     * @param queryLimit             Maximum number of returned items, may be null
     * @return Empty list if the query returned no data.
     * @throws DbAccessException To be thrown in case of exception during query execution.
     */
    public <T extends IDbItem> List<T> queryDbData(Class<T> clazz,
                                                   String[] restrictionColumn, Object[] restrictionColumnValue,
                                                   String oderColumn, Boolean orderAscending,
                                                   Integer queryLimit) throws DbAccessException {
        return queryDbData(clazz, restrictionColumn, restrictionColumnValue,
                           oderColumn, orderAscending, queryLimit, null);
    }

    /**
     * Executes a query and retrieves the data that satisfies the specified parameters.
     *
     * @param <T>                    The sub-class of IDbItem being boxed
     * @param clazz                  The sub-class of IDbItem being boxed
     * @param restrictionColumn      Names of the columns to use for restricting the result. May be null
     *                               or empty.
     * @param restrictionColumnValue Values of the columns to use for restricting the result. May be
     *                               null or empty.
     * @param oderColumn             Name of the column to order the result
     * @param orderAscending         If true, the result will be given in ascending order. If false or null,
     *                               result will be in descending order.
     * @return Empty list if the query returned no data.
     * @throws DbAccessException To be thrown in case of exception during query execution.
     */

    public <T extends IDbItem> List<T> queryDbData(Class<T> clazz,
                                                   String[] restrictionColumn, Object[] restrictionColumnValue,
                                                   String oderColumn, Boolean orderAscending) throws DbAccessException {
        return queryDbData(clazz, restrictionColumn, restrictionColumnValue, oderColumn, orderAscending, null);
    }

    /**
     * Tests if the given criteria data is OK.
     *
     * @param restrictionColumn      Names of the columns to use for restricting the result. May be null
     *                               or empty.
     * @param restrictionColumnValue Values of the columns to use for restricting the result. May be
     *                               null or empty.
     * @throws IllegalArgumentException To be thrown when there is a criteria missmatch.
     */
    protected static void testRestrictionData(String[] restrictionColumn, Object[] restrictionColumnValue)
            throws IllegalArgumentException {
        boolean invalid = false;
        if (restrictionColumn == null) {
            if (restrictionColumnValue != null) {
                invalid = true;
            }
        } else {
            if (restrictionColumnValue == null) {
                invalid = true;
            } else if (restrictionColumn.length != restrictionColumnValue.length) {
                invalid = true;
            }
        }
        if (invalid) {
            throw new IllegalArgumentException(String.format("Criteria specification does not match: %s vs %s",
                                                             Arrays.toString(restrictionColumn),
                                                             Arrays.toString(restrictionColumnValue)));
        }
    }

    /**
     * Executes a query and retrieves the data that satisfies the specified parameters.
     *
     * @param <T>                    The sub-class of IDbItem being boxed
     * @param clazz                  The sub-class of IDbItem being boxed
     * @param restrictionColumn      Names of the columns to use for restricting the result. May be null
     *                               or empty.
     * @param restrictionColumnValue Values of the columns to use for restricting the result. May be
     *                               null or empty.
     * @param oderColumn             Name of the column to order the result
     * @param orderAscending         If true, the result will be given in ascending order. If false or null,
     *                               result will be in descending order.
     * @return Null if no items available
     * @throws DbAccessException To be thrown in case of exception during query execution.
     */
    public <T extends IDbItem> T queryDbDataLast(Class<T> clazz,
                                                 String[] restrictionColumn, Object[] restrictionColumnValue,
                                                 String oderColumn, Boolean orderAscending) throws DbAccessException {
        List<T> result = queryDbData(clazz, restrictionColumn, restrictionColumnValue, oderColumn, orderAscending, 1);
        if (result != null) {
            if (!result.isEmpty()) {
                return result.get(0);
            }
        }
        return null;
    }

    @Override
    public void remove(IDbItem item) throws DbAccessException {
        if (item == null) {
            LOGGER.warn("Could not delete provided item, because it is null.");
            return;
        }
        removeItems(item);
    }

    /**
     * Deletes the given items in the database.
     *
     * @param itemsToDelete Items that shall be deleted in the database
     */
    private void removeItems(IDbItem... itemsToDelete) throws DbAccessException {
        Session s = getSession().openSession();
        try {
            if (ArrayUtils.isEmpty(itemsToDelete)) {
                return;
            }
            Transaction t = s.beginTransaction();
            for (IDbItem item : itemsToDelete) {
                s.delete(item);
            }
            t.commit();
        } catch (HibernateException ex) {
            LOGGER.error(ex.getMessage(), ex);
        } finally {
            if (s != null) {
                s.close();
            }
        }
    }

    @Override
    public void clear() throws DbAccessException {
        Session s = getSession().openSession();
        try {
            s.createQuery(String.format("delete from %s", ConceptContainer.class.getCanonicalName())).executeUpdate();
            s.createQuery(String.format("delete from %s", ConceptContainerScore.class.getCanonicalName()))
             .executeUpdate();
            s.createQuery(String.format("delete from %s", KnowledgeObject.class.getCanonicalName())).executeUpdate();
            s.createQuery(String.format("delete from %s", KnowledgeObjectScore.class.getCanonicalName()))
             .executeUpdate();
            s.createQuery(String.format("delete from %s", Learner.class.getCanonicalName())).executeUpdate();
            s.createQuery(String.format("delete from %s", LearnerPreference.class.getCanonicalName())).executeUpdate();
            s.createQuery(String.format("delete from %s", LearningRecommendation.class.getCanonicalName()))
             .executeUpdate();
            s.createQuery(String.format("delete from %s", LearningStep.class.getCanonicalName())).executeUpdate();
            s.createQuery(String.format("delete from %s", LmsParameter.class.getCanonicalName())).executeUpdate();
            s.createQuery(String.format("delete from %s", MacroLearningPathway.class.getCanonicalName()))
             .executeUpdate();
            s.createQuery(String.format("delete from %s", MacroLearningPathwaySelection.class.getCanonicalName()))
             .executeUpdate();
            s.createQuery(String.format("delete from %s", MicroLearningPathway.class.getCanonicalName()))
             .executeUpdate();
            s.createQuery(String.format("delete from %s", MicroLearningPathwaySelection.class.getCanonicalName()))
             .executeUpdate();
            s.createQuery(String.format("delete from %s", TimedependentSessionData.class.getCanonicalName()))
             .executeUpdate();
            s.createQuery(String.format("delete from %s", eu.intuitel.db.model.Session.class.getCanonicalName()))
             .executeUpdate();
        } catch (HibernateException ex) {
            LOGGER.error(ex.getMessage(), ex);
        } finally {
            if (s != null) {
                s.close();
            }
        }
    }

    @Override
    public LmsParameter getLastLmsParameter() throws DbAccessException {
        return queryDbDataLast(LmsParameter.class,
                               null, null,
                               "updateTime", Boolean.FALSE);
    }

    @Override
    public Learner getLearner(String learnerId) throws DbAccessException {
        return queryDbDataLast(Learner.class,
                               new String[]{"id"},
                               new String[]{learnerId},
                               null, Boolean.FALSE);
    }

    @Override
    public ConceptContainerScore getCcLastScore(String learnerId, String ccIri) throws DbAccessException {
        return queryDbDataLast(ConceptContainerScore.class,
                               new String[]{"id.learnerId", "id.conceptContainerIri"},
                               new String[]{learnerId, ccIri},
                               "id.time", Boolean.FALSE);
    }

    @Override
    public KnowledgeObjectScore getLastKoScore(String learnerId, String koIri) throws DbAccessException {
        return queryDbDataLast(KnowledgeObjectScore.class,
                               new String[]{"id.learnerId", "id.knowledgeObjectIri"},
                               new String[]{learnerId, koIri},
                               "id.time", Boolean.FALSE);
    }

    @Override
    public List<ConceptContainerScore> getCcScores(String learnerId, String courseId) throws DbAccessException {
        return queryDbData(ConceptContainerScore.class,
                           new String[]{"id.learnerId", "id.courseId"},
                           new String[]{learnerId, courseId},
                           "id.time", Boolean.FALSE);
    }

    @Override
    public List<KnowledgeObjectScore> getKoScores(String learnerId, String courseId) throws DbAccessException {
        return queryDbData(KnowledgeObjectScore.class,
                           new String[]{"id.learnerId", "id.courseId"},
                           new String[]{learnerId, courseId},
                           "id.time", Boolean.FALSE);
    }

    @Override
    public List<LearningStep> getLearningSteps(String learnerId, String courseId, int numberOfSteps)
            throws DbAccessException {
        return queryDbData(LearningStep.class,
                           new String[]{"id.sessionLearnerId", "id.sessionCourseId"},
                           new String[]{learnerId, courseId},
                           "id.startTime", Boolean.FALSE, numberOfSteps);
    }

    @Override
    public List<LearningStep> getLearningSteps(String learnerId, int numberOfSteps) throws DbAccessException {
        return queryDbData(LearningStep.class,
                           new String[]{"id.sessionLearnerId"},
                           new String[]{learnerId},
                           "id.startTime", Boolean.FALSE, numberOfSteps);
    }

    @Override
    public List<LearningStep> getLearningStepsWithKoIri(String learnerId, String koIri) throws DbAccessException {
        return queryDbData(LearningStep.class,
                           new String[]{"id.sessionLearnerId", "knowledgeObjectIri"},
                           new String[]{learnerId, koIri},
                           "id.startTime", Boolean.FALSE);
    }

    @Override
    public List<LearningStep> getLearningStepsWithCcIri(String learnerId, String ccIri) throws DbAccessException {
        return queryDbData(LearningStep.class,
                           new String[]{"id.sessionLearnerId", "conceptContainerIri"},
                           new String[]{learnerId, ccIri},
                           "id.startTime", Boolean.FALSE);
    }

    @Override
    public LearningStep getLastLearningStep(String learnerId, String courseId) throws DbAccessException {
        List<LearningStep> steps = getLearningSteps(learnerId, courseId, 1);
        return steps.isEmpty() ? null : steps.get(0);
    }

    @Override
    public List<LearningRecommendation> getLearningRecommendations(LearningStepId id) throws DbAccessException {
        return queryDbData(LearningRecommendation.class,
                           new String[]{"id.learningStepSessionLearnerId", "id.learningStepSessionLoginTime",
                                               "id.learningStepSessionCourseId", "id.learningStepStartTime"},
                           new Object[]{id.getSessionLearnerId(), id.getSessionLoginTime(), id.getSessionCourseId(),
                                               id.getStartTime()},
                           "recommendationValue", Boolean.FALSE);
    }

    @Override
    public List<eu.intuitel.db.model.Session> getSessions(String courseId) throws DbAccessException {
        return queryDbData(eu.intuitel.db.model.Session.class,
                           new String[]{"id.courseId"},
                           new String[]{courseId},
                           "id.loginTime", Boolean.FALSE);
    }

    @Override
    public List<eu.intuitel.db.model.Session> getSessions(String learnerId, String courseId) throws DbAccessException {
        return queryDbData(eu.intuitel.db.model.Session.class,
                           new String[]{"id.learnerId", "id.courseId"},
                           new String[]{learnerId, courseId},
                           "id.loginTime", Boolean.FALSE);
    }

    @Override
    public List<eu.intuitel.db.model.Session> getSessions(String learnerId, String courseId,
                                                          int numberOfSessions) throws DbAccessException {
        return queryDbData(eu.intuitel.db.model.Session.class,
                           new String[]{"id.learnerId", "id.courseId"},
                           new String[]{learnerId, courseId},
                           "id.loginTime", Boolean.FALSE, numberOfSessions);
    }

    @Override
    public eu.intuitel.db.model.Session getLastSession(String learnerId) throws DbAccessException {
        return queryDbDataLast(eu.intuitel.db.model.Session.class,
                               new String[]{"id.learnerId"},
                               new String[]{learnerId},
                               "id.loginTime", Boolean.FALSE);
    }

    @Override
    public TimedependentSessionData getLastTimedependentSessionData(SessionId id) throws DbAccessException {
        return queryDbDataLast(TimedependentSessionData.class,
                               new String[]{"id.sessionLearnerId", "id.sessionLoginTime", "id.sessionCourseId"},
                               new Object[]{id.getLearnerId(), id.getLoginTime(), id.getCourseId()},
                               "id.updateTime", Boolean.FALSE);
    }

    @Override
    public Collection<TimedependentSessionData> getTimedependentSessionData(SessionId id) throws DbAccessException {
        return queryDbData(TimedependentSessionData.class,
                           new String[]{"id.sessionLearnerId", "id.sessionLoginTime", "id.sessionCourseId"},
                           new Object[]{id.getLearnerId(), id.getLoginTime(), id.getCourseId()},
                           "id.updateTime", Boolean.FALSE);
    }

    @Override
    public LearnerPreference getLearnerPreference(String learnerId, String courseId) throws DbAccessException {
        return queryDbDataLast(LearnerPreference.class,
                               new String[]{"id.learnerId", "id.courseId"},
                               new String[]{learnerId, courseId},
                               null, Boolean.FALSE);
    }

    @Override
    public MacroLearningPathwaySelection getLastMacroLpSelection(String learnerId, String courseId)
            throws DbAccessException {
        return queryDbDataLast(MacroLearningPathwaySelection.class,
                               new String[]{"id.learnerId", "id.courseId"},
                               new String[]{learnerId, courseId},
                               "id.selectionTime", false);
    }

    @Override
    public Collection<MacroLearningPathwaySelection> getMacroLpSelections(String courseId) throws DbAccessException {
        return queryDbData(MacroLearningPathwaySelection.class,
                           new String[]{"id.courseId"},
                           new String[]{courseId},
                           "id.selectionTime", false, null);
    }

    @Override
    public MicroLearningPathwaySelection getLastMicroLpSelection(String learnerId, String courseId)
            throws DbAccessException {
        return queryDbDataLast(MicroLearningPathwaySelection.class,
                               new String[]{"id.learnerId", "id.courseId"},
                               new String[]{learnerId, courseId},
                               "id.selectionTime", false);
    }

    @Override
    public Collection<MicroLearningPathwaySelection> getMicroLpSelections(String courseId) throws DbAccessException {
        return queryDbData(MicroLearningPathwaySelection.class,
                           new String[]{"id.courseId"},
                           new String[]{courseId},
                           "id.selectionTime", false, null);
    }

    @Override
    public KnowledgeObject getKo(String iri) throws DbAccessException {
        return queryDbDataLast(KnowledgeObject.class,
                               new String[]{"iri"},
                               new String[]{iri},
                               null, Boolean.FALSE);
    }

    @Override
    public ConceptContainer getCc(String iri) throws DbAccessException {
        return queryDbDataLast(ConceptContainer.class,
                               new String[]{"iri"},
                               new String[]{iri},
                               null, Boolean.FALSE);
    }

    @Override
    public MicroLearningPathway getMicroLp(String iri) throws DbAccessException {
        return queryDbDataLast(MicroLearningPathway.class,
                               new String[]{"iri"},
                               new String[]{iri},
                               null, Boolean.FALSE);
    }

    @Override
    public MacroLearningPathway getMacroLp(String iri) throws DbAccessException {
        return queryDbDataLast(MacroLearningPathway.class,
                               new String[]{"iri"},
                               new String[]{iri},
                               null, Boolean.FALSE);
    }
}
