/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, University of Reading, United Kingdom, http://www.isr.reading.ac.uk/
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 03.02.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.db.access;

import eu.intuitel.db.model.*;
import java.util.Collection;
import java.util.List;

/**
 * Interface containing accessor methods for the data stored in the database.
 */
public interface IDbAccessPoint {
    /**
     * Determines if the database connection was established correctly.
     *
     * @return True if DB connection is available.
     */
    public boolean isConnectionEstablished();


    /**
     * Shuts the complete DB connection down.
     * This methods needs to be called when no the complete system is being shut down.
     */
    public void shutdown();
    
    //<editor-fold defaultstate="collapsed" desc="Store/Remove content">
    /**
     * Stores or updates the provided IDbItem.
     *
     * @param item Item to be stored.
     * @throws DbAccessException To be thrown when setting the data fails.
     */
    public void storeOrUpdate(IDbItem item) throws DbAccessException;

    /**
     * Removes the specified IDbItem from the data storage.
     *
     * @param item Item to be stored.
     * @throws DbAccessException To be thrown when setting the data fails.
     */
    public void remove(IDbItem item) throws DbAccessException;
    
    /**
     * Clears all DB tables. 
     * Only for test cleanup!
     * @throws eu.intuitel.db.access.DbAccessException
     */
    public void clear() throws DbAccessException;
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="LMS data">
    /**
     * @return Null on no result
     * @throws DbAccessException
     */
    public LmsParameter getLastLmsParameter() throws DbAccessException;
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="Learner">
    /**
     * Retrieves user model data for a specific user.
     *
     * @param learnerId ID of the learner in scope
     * @return Null on no result
     * @throws DbAccessException
     */
    public Learner getLearner(String learnerId) throws DbAccessException;
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="LO completion">
    /**
     * @param learnerId ID of the learner in scope
     * @param ccIri IRI of the Concept Container in scope
     * @return Null on no result
     * @throws DbAccessException
     */
    public ConceptContainerScore getCcLastScore(String learnerId, String ccIri) throws DbAccessException;

    /**
     * @param learnerId ID of the learner in scope
     * @param koIri IRI of the Concept Container in scope
     * @return Null on no result
     * @throws DbAccessException
     */
    public KnowledgeObjectScore getLastKoScore(String learnerId, String koIri) throws DbAccessException;

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Empty list on no result.
     * @throws DbAccessException
     */
    public List<ConceptContainerScore> getCcScores(String learnerId, String courseId) throws DbAccessException;

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Empty list on no result.
     * @throws DbAccessException
     */
    public List<KnowledgeObjectScore> getKoScores(String learnerId, String courseId) throws DbAccessException;
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="LO history">
    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param numberOfSteps Maximum number of entries to retrieve
     * @return Empty list on no result or on no data.
     * @throws DbAccessException
     */
    public List<LearningStep> getLearningSteps(String learnerId, String courseId, int numberOfSteps) throws DbAccessException;

    /**
     * @param learnerId ID of the learner in scope
     * @param numberOfSteps The maximum number of steps to be returned
     * @return Empty list on error or on no data.
     * @throws DbAccessException
     */
    public List<LearningStep> getLearningSteps(String learnerId, int numberOfSteps) throws DbAccessException;

    /**
     * @param learnerId ID of the learner in scope
     * @param koIri IRI of the KO in scope.
     * @return Empty list on no result or on no data.
     * @throws DbAccessException
     */
    public List<LearningStep> getLearningStepsWithKoIri(String learnerId, String koIri) throws DbAccessException;
    
    /**
     * @param learnerId ID of the learner in scope
     * @param ccIri IRI of the KO in scope.
     * @return Empty list on no result or on no data.
     * @throws DbAccessException
     */
    public List<LearningStep> getLearningStepsWithCcIri(String learnerId, String ccIri) throws DbAccessException;

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Null if no data available.
     * @throws DbAccessException
     */
    public LearningStep getLastLearningStep(String learnerId, String courseId) throws DbAccessException;
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="LO recommendations">
    /**
     * @param id ID of the learning step to which the recommendations are connected.
     * @return Empty List on error or on no data.
     * @throws DbAccessException
     */
    public List<LearningRecommendation> getLearningRecommendations(LearningStepId id) throws DbAccessException;
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="Session">
    /**
     * Retrieves all sessions connected to the specified course.
     *
     * @param courseId ID of the course in scope
     * @return Empty list in case nothing was found.
     * @throws DbAccessException
     */
    public List<Session> getSessions(String courseId) throws DbAccessException;

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Empty list on error or on no data.
     * @throws DbAccessException
     */
    public List<Session> getSessions(String learnerId, String courseId) throws DbAccessException;

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @param numberOfSessions Number of sessions to retrieve
     * @return Empty list on error or on no data.
     * @throws DbAccessException
     */
    public List<Session> getSessions(String learnerId, String courseId, int numberOfSessions) throws DbAccessException;

    /**
     * @param learnerId ID of the learner in scope
     * @return Null on no data
     * @throws DbAccessException
     */
    public Session getLastSession(String learnerId) throws DbAccessException;

//    /**
//     * Retrives the last Sessions that have been connected with the specified course. This only
//     * returns the last sessions, i.e. the result will only contain the latest session for each
//     * learner.
//     *
//     * @param courseId ID of the course in scope
//     * @return Empty collection if there are no session.
//     * @throws eu.intuitel.db.access.DbAccessException
//     */
//    public List<Session> getLastSessions(String courseId) throws DbAccessException;
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="Session - time-dependent data">
    /**
     * @param id ID of the session in scope
     * @return Null in case nothing was found
     * @throws DbAccessException
     */
    public TimedependentSessionData getLastTimedependentSessionData(SessionId id) throws DbAccessException;

    /**
     * @param id ID of the session in scope
     * @return Empty list if nothing was found or in case of error.
     * @throws DbAccessException
     */
    public Collection<TimedependentSessionData> getTimedependentSessionData(SessionId id) throws DbAccessException;
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="Learner preference">
    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return Null if unknown
     * @throws DbAccessException
     */
    public LearnerPreference getLearnerPreference(String learnerId, String courseId) throws DbAccessException;
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="Learning pathway">
    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return IRI of the macro LP of the leanrer. Null if unknown.
     * @throws DbAccessException
     */
    public MacroLearningPathwaySelection getLastMacroLpSelection(String learnerId, String courseId) throws DbAccessException;

    /**
     * @param learnerId ID of the learner in scope
     * @param courseId ID of the course in scope
     * @return IRI of the micro LP of the leanrer. Null if unknown.
     * @throws DbAccessException
     */
    public MicroLearningPathwaySelection getLastMicroLpSelection(String learnerId, String courseId) throws DbAccessException;

    /**
     * @param courseId ID of the course in scope
     * @return Empty collection on no data.
     * @throws DbAccessException
     */
    public Collection<MicroLearningPathwaySelection> getMicroLpSelections(String courseId) throws DbAccessException;

    /**
     * @param courseId ID of the course in scope
     * @return Empty collection on no data.
     * @throws DbAccessException
     */
    public Collection<MacroLearningPathwaySelection> getMacroLpSelections(String courseId) throws DbAccessException;
    //</editor-fold>

    //<editor-fold defaultstate="collapsed" desc="SLOM data">
    /**
     * @param iri IRI of the KO in scope
     * @return Null if no data available.
     * @throws DbAccessException
     */
    public KnowledgeObject getKo(String iri) throws DbAccessException;

    /**
     * @param iri IRI of the CC in scope
     * @return Null if no data available.
     * @throws DbAccessException
     */
    public ConceptContainer getCc(String iri) throws DbAccessException;

    /**
     * @param iri IRI of the mLP in scope
     * @return Null if no data available.
     * @throws DbAccessException
     */
    public MicroLearningPathway getMicroLp(String iri) throws DbAccessException;

    /**
     * @param iri IRI of the MLP in scope
     * @return Null if no data available.
     * @throws DbAccessException
     */
    public MacroLearningPathway getMacroLp(String iri) throws DbAccessException;
    //</editor-fold>
}
