/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Ovidiu Serban, University of Reading, www.reading.ac.uk
 * Contributor: Jaime Costa, University of Reading, www.reading.ac.uk
 * Version: [0.01]
 * Date: 05.11.2013
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.db.access;

import eu.intuitel.db.model.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

/**
 * Mock class for testing the user DB communication.
 * All return values are accessible via public variables.
 */
public class LocalDbCache implements IDbAccessPoint {
    private static final List<ConceptContainer> ccs = new ArrayList<>();
    private static final List<ConceptContainerScore> ccScores = new ArrayList<>();
    private static final List<KnowledgeObject> kos = new ArrayList<>();
    private static final List<KnowledgeObjectScore> koScores = new ArrayList<>();
    private static final List<Learner> learners = new ArrayList<>();
    private static final List<LearnerPreference> lPrefs = new ArrayList<>();
    private static final List<LearningRecommendation> lRecoms = new ArrayList<>();
    private static final List<LearningStep> lSteps = new ArrayList<>();
    private static final List<LmsParameter> lmsParams = new ArrayList<>();
    private static final List<MacroLearningPathway> macroLps = new ArrayList<>();
    private static final List<MacroLearningPathwaySelection> macroLpSelections = new ArrayList<>();
    private static final List<MicroLearningPathway> microLps = new ArrayList<>();
    private static final List<MicroLearningPathwaySelection> microLpSelections = new ArrayList<>();
    private static final List<Session> sessions = new ArrayList<>();
    private static final List<TimedependentSessionData> timedependentSessionData = new ArrayList<>();

    @Override
    public void storeOrUpdate(IDbItem item) throws DbAccessException {
        if (item == null) {
            return;
        }
        if (item instanceof ConceptContainer) {
            storeOrUpdate((ConceptContainer) item, ccs);
        } else if (item instanceof ConceptContainerScore) {
            storeOrUpdate((ConceptContainerScore) item, ccScores);
        } else if (item instanceof KnowledgeObject) {
            storeOrUpdate((KnowledgeObject) item, kos);
        } else if (item instanceof KnowledgeObjectScore) {
            storeOrUpdate((KnowledgeObjectScore) item, koScores);
        } else if (item instanceof Learner) {
            storeOrUpdate((Learner) item, learners);
        } else if (item instanceof LearnerPreference) {
            storeOrUpdate((LearnerPreference) item, lPrefs);
        } else if (item instanceof LearningRecommendation) {
            storeOrUpdate((LearningRecommendation) item, lRecoms);
        } else if (item instanceof LearningStep) {
            storeOrUpdate((LearningStep) item, lSteps);
        } else if (item instanceof LmsParameter) {
            storeOrUpdate((LmsParameter) item, lmsParams);
        } else if (item instanceof MacroLearningPathway) {
            storeOrUpdate((MacroLearningPathway) item, macroLps);
        } else if (item instanceof MacroLearningPathwaySelection) {
            storeOrUpdate((MacroLearningPathwaySelection) item, macroLpSelections);
        } else if (item instanceof MicroLearningPathway) {
            storeOrUpdate((MicroLearningPathway) item, microLps);
        } else if (item instanceof MicroLearningPathwaySelection) {
            storeOrUpdate((MicroLearningPathwaySelection) item, microLpSelections);
        } else if (item instanceof Session) {
            storeOrUpdate((Session) item, sessions);
        } else if (item instanceof TimedependentSessionData) {
            storeOrUpdate((TimedependentSessionData) item, timedependentSessionData);
        }
    }

    @SuppressWarnings("all")
    private <T extends IDbItem> void storeOrUpdate(T newItem, final List<T> storedItemSet) {
        synchronized (storedItemSet) {
            T itemToReplace = null;
            for (T storedItem : storedItemSet) {
                if (storedItem.isPkEqualTo(newItem)) {
                    itemToReplace = storedItem;
                    break;
                }
            }
            if (itemToReplace != null) {
                storedItemSet.remove(itemToReplace);
            }
            storedItemSet.add(0, newItem);
        }
    }

    @Override
    public void remove(IDbItem item) throws DbAccessException {
        if (item == null) {
            return;
        }
        if (item instanceof ConceptContainer) {
            removeItem((ConceptContainer) item, ccs);
        } else if (item instanceof ConceptContainerScore) {
            removeItem((ConceptContainerScore) item, ccScores);
        } else if (item instanceof KnowledgeObject) {
            removeItem((KnowledgeObject) item, kos);
        } else if (item instanceof KnowledgeObjectScore) {
            removeItem((KnowledgeObjectScore) item, koScores);
        } else if (item instanceof Learner) {
            removeItem((Learner) item, learners);
        } else if (item instanceof LearnerPreference) {
            removeItem((LearnerPreference) item, lPrefs);
        } else if (item instanceof LearningRecommendation) {
            removeItem((LearningRecommendation) item, lRecoms);
        } else if (item instanceof LearningStep) {
            removeItem((LearningStep) item, lSteps);
        } else if (item instanceof LmsParameter) {
            removeItem((LmsParameter) item, lmsParams);
        } else if (item instanceof MacroLearningPathway) {
            removeItem((MacroLearningPathway) item, macroLps);
        } else if (item instanceof MacroLearningPathwaySelection) {
            removeItem((MacroLearningPathwaySelection) item, macroLpSelections);
        } else if (item instanceof MicroLearningPathway) {
            removeItem((MicroLearningPathway) item, microLps);
        } else if (item instanceof MicroLearningPathwaySelection) {
            removeItem((MicroLearningPathwaySelection) item, microLpSelections);
        } else if (item instanceof Session) {
            removeItem((Session) item, sessions);
        } else if (item instanceof TimedependentSessionData) {
            removeItem((TimedependentSessionData) item, timedependentSessionData);
        }
    }

    @SuppressWarnings("all")
    private <T extends IDbItem> void removeItem(T newItem, List<T> storedItemSet) {
        synchronized (storedItemSet) {
            if (storedItemSet.contains(newItem)) {
                storedItemSet.remove(newItem);
            }
        }
    }

    @Override
    public void clear() throws DbAccessException {
        synchronized (ccs) {
            ccs.clear();
        }
        synchronized (ccScores) {
            ccScores.clear();
        }
        synchronized (kos) {
            kos.clear();
        }
        synchronized (koScores) {
            koScores.clear();
        }
        synchronized (learners) {
            learners.clear();
        }
        synchronized (lPrefs) {
            lPrefs.clear();
        }
        synchronized (lRecoms) {
            lRecoms.clear();
        }
        synchronized (lSteps) {
            lSteps.clear();
        }
        synchronized (lmsParams) {
            lmsParams.clear();
        }
        synchronized (macroLps) {
            macroLps.clear();
        }
        synchronized (macroLpSelections) {
            macroLpSelections.clear();
        }
        synchronized (microLps) {
            microLps.clear();
        }
        synchronized (microLpSelections) {
            microLpSelections.clear();
        }
        synchronized (sessions) {
            sessions.clear();
        }
        synchronized (timedependentSessionData) {
            timedependentSessionData.clear();
        }
    }

    @Override
    public void shutdown() {
    }

    @Override
    public boolean isConnectionEstablished() {
        return true;
    }

    @Override
    public LmsParameter getLastLmsParameter() throws DbAccessException {
        synchronized (lmsParams) {
            return lmsParams.isEmpty() ? null : lmsParams.iterator().next();
        }
    }

    @Override
    public Learner getLearner(String learnerId) throws DbAccessException {
        synchronized (learners) {
            for (Learner l : learners) {
                if (l.getId().equals(learnerId)) {
                    return l;
                }
            }
            return null;
        }
    }

    @Override
    public ConceptContainerScore getCcLastScore(String learnerId, String ccIri) throws DbAccessException {
        synchronized (ccScores) {
            for (ConceptContainerScore score : ccScores) {
                if (score.getId().getLearnerId().equals(learnerId)
                    && score.getId().getConceptContainerIri().equals(ccIri)) {
                    return score;
                }
            }
            return null;
        }
    }

    @Override
    public KnowledgeObjectScore getLastKoScore(String learnerId, String koIri) throws DbAccessException {
        synchronized (koScores) {
            for (KnowledgeObjectScore score : koScores) {
                if (score.getId().getLearnerId().equals(learnerId)
                    && score.getId().getKnowledgeObjectIri().equals(koIri)) {
                    return score;
                }
            }
            return null;
        }
    }

    @Override
    public List<ConceptContainerScore> getCcScores(String learnerId, String courseId) throws DbAccessException {
        synchronized (ccScores) {
            List<ConceptContainerScore> returnVal = new ArrayList<>();
            for (ConceptContainerScore item : ccScores) {
                if (item.getId().getLearnerId().equals(learnerId) && item.getId().getCourseId().equals(courseId)) {
                    returnVal.add(item);
                }
            }
            return returnVal;
        }
    }

    @Override
    public List<KnowledgeObjectScore> getKoScores(String learnerId, String courseId) throws DbAccessException {
        synchronized (koScores) {
            List<KnowledgeObjectScore> returnVal = new ArrayList<>();
            for (KnowledgeObjectScore item : koScores) {
                if (item.getId().getLearnerId().equals(learnerId) && item.getId().getCourseId().equals(courseId)) {
                    returnVal.add(item);
                }
            }
            return returnVal;
        }
    }

    @Override
    public List<LearningStep> getLearningSteps(String learnerId, String courseId, int numberOfSteps)
            throws DbAccessException {
        synchronized (lSteps) {
            List<LearningStep> returnVal = new ArrayList<>();
            for (LearningStep lStep : lSteps) {
                if (returnVal.size() == numberOfSteps) {
                    return returnVal;
                }
                if (lStep.getId().getSessionLearnerId().equals(learnerId)
                    && lStep.getId().getSessionCourseId().equals(courseId)) {
                    returnVal.add(lStep);
                }
            }
            return returnVal;
        }
    }

    @Override
    public List<LearningStep> getLearningSteps(String learnerId, int numberOfSteps) throws DbAccessException {
        synchronized (lSteps) {
            List<LearningStep> returnVal = new ArrayList<>();
            for (LearningStep lStep : lSteps) {
                if (lStep.getId().getSessionLearnerId().equals(learnerId)) {
                    returnVal.add(lStep);
                    if (returnVal.size() == numberOfSteps) {
                        return returnVal;
                    }
                }
            }
            return returnVal;
        }
    }

    @Override
    public List<LearningStep> getLearningStepsWithKoIri(String learnerId, String koIri) throws DbAccessException {
        synchronized (lSteps) {
            List<LearningStep> returnVal = new ArrayList<>();
            for (LearningStep lStep : lSteps) {
                if (lStep.getId().getSessionLearnerId().equals(learnerId)
                    && lStep.getKnowledgeObjectIri().equals(koIri)) {
                    returnVal.add(lStep);
                }
            }
            return returnVal;
        }
    }

    @Override
    public List<LearningStep> getLearningStepsWithCcIri(String learnerId, String ccIri) throws DbAccessException {
        synchronized (lSteps) {
            List<LearningStep> returnVal = new ArrayList<>();
            for (LearningStep lStep : lSteps) {
                if (lStep.getId().getSessionLearnerId().equals(learnerId)
                    && lStep.getConceptContainerIri().equals(ccIri)) {
                    returnVal.add(lStep);
                }
            }
            return returnVal;
        }
    }

    @Override
    public LearningStep getLastLearningStep(String learnerId, String courseId) throws DbAccessException {
        synchronized (lSteps) {
            for (LearningStep lStep : lSteps) {
                if (lStep.getId().getSessionLearnerId().equals(learnerId)
                    && lStep.getId().getSessionCourseId().equals(courseId)) {
                    return lStep;
                }
            }
            return null;
        }
    }

    @Override
    public List<LearningRecommendation> getLearningRecommendations(LearningStepId id) throws DbAccessException {
        synchronized (lRecoms) {
            List<LearningRecommendation> recoms = new ArrayList<>();
            for (LearningRecommendation lRecom : lRecoms) {
                if (id.getSessionCourseId().equals(lRecom.getId().getLearningStepSessionCourseId())) {
                    if (id.getSessionLearnerId().equals(lRecom.getId().getLearningStepSessionLearnerId())) {
                        if (id.getSessionLoginTime().compareTo(lRecom.getId().getLearningStepSessionLoginTime()) == 0) {
                            if (id.getStartTime().compareTo(lRecom.getId().getLearningStepStartTime()) == 0) {
                                recoms.add(lRecom);
                            }
                        }
                    }
                }
            }
            return recoms;
        }
    }

    @Override
    public List<Session> getSessions(String courseId) throws DbAccessException {
        synchronized (sessions) {
            List<Session> returnVal = new ArrayList<>();
            for (Session s : sessions) {
                if (s.getId().getCourseId().equals(courseId)) {
                    returnVal.add(s);
                }
            }
            return returnVal;
        }
    }

    @Override
    public List<Session> getSessions(String learnerId, String courseId) throws DbAccessException {
        synchronized (sessions) {
            List<Session> returnVal = new ArrayList<>();
            for (Session s : sessions) {
                if (s.getId().getCourseId().equals(courseId)
                    && s.getId().getLearnerId().equals(learnerId)) {
                    returnVal.add(s);
                }
            }
            return returnVal;
        }
    }

    @Override
    public List<Session> getSessions(String learnerId, String courseId, int numberOfSessions) throws DbAccessException {
        synchronized (sessions) {
            List<Session> returnVal = new ArrayList<>();
            for (Session s : sessions) {
                if (numberOfSessions == returnVal.size()) {
                    return returnVal;
                }
                if (s.getId().getCourseId().equals(courseId) && s.getId().getLearnerId().equals(learnerId)) {
                    returnVal.add(s);
                }
            }
            return returnVal;
        }
    }

    @Override
    public Session getLastSession(String learnerId) throws DbAccessException {
        synchronized (sessions) {
            for (Session s : sessions) {
                if (s.getId().getLearnerId().equals(learnerId)) {
                    return s;
                }
            }
            return null;
        }
    }

    @Override
    public TimedependentSessionData getLastTimedependentSessionData(SessionId id) throws DbAccessException {
        synchronized (timedependentSessionData) {
            for (TimedependentSessionData dat : timedependentSessionData) {
                if (id.getCourseId().equals(dat.getId().getSessionCourseId())) {
                    if (id.getLearnerId().equals(dat.getId().getSessionLearnerId())) {
                        if (id.getLoginTime().compareTo(dat.getId().getSessionLoginTime()) == 0) {
                            return dat;
                        }
                    }
                }
            }
            return null;
        }
    }

    @Override
    public Collection<TimedependentSessionData> getTimedependentSessionData(SessionId id) throws DbAccessException {
        synchronized (timedependentSessionData) {
            List<TimedependentSessionData> returnVal = new ArrayList<>();
            for (TimedependentSessionData dat : timedependentSessionData) {
                if (id.getCourseId().equals(dat.getId().getSessionCourseId())) {
                    if (id.getLearnerId().equals(dat.getId().getSessionLearnerId())) {
                        if (id.getLoginTime().compareTo(dat.getId().getSessionLoginTime()) == 0) {
                            returnVal.add(dat);
                        }
                    }
                }
            }
            return returnVal;
        }
    }

    @Override
    public LearnerPreference getLearnerPreference(String learnerId, String courseId) throws DbAccessException {
        synchronized (lPrefs) {
            for (LearnerPreference pref : lPrefs) {
                if (pref.getId().getLearnerId().equals(learnerId)
                    && pref.getId().getCourseId().equals(courseId)) {
                    return pref;
                }
            }
            return null;
        }
    }

    @Override
    public MacroLearningPathwaySelection getLastMacroLpSelection(String learnerId, String courseId) {
        synchronized (macroLpSelections) {
            for (MacroLearningPathwaySelection item : macroLpSelections) {
                if (item.getId().getLearnerId().equals(learnerId) && item.getId().getCourseId().equals(courseId)) {
                    return item;
                }
            }
        }
        return null;
    }

    @Override
    public MicroLearningPathwaySelection getLastMicroLpSelection(String learnerId, String courseId) {
        synchronized (microLpSelections) {
            for (MicroLearningPathwaySelection item : microLpSelections) {
                if (item.getId().getLearnerId().equals(learnerId) && item.getId().getCourseId().equals(courseId)) {
                    return item;
                }
            }
        }
        return null;
    }

    @Override
    public Collection<MacroLearningPathwaySelection> getMacroLpSelections(String courseId) {
        List<MacroLearningPathwaySelection> returnVal = new LinkedList<>();
        synchronized (macroLpSelections) {
            for (MacroLearningPathwaySelection item : macroLpSelections) {
                if (item.getId().getCourseId().equals(courseId)) {
                    returnVal.add(item);
                }
            }
        }
        return returnVal;
    }

    @Override
    public Collection<MicroLearningPathwaySelection> getMicroLpSelections(String courseId) {
        List<MicroLearningPathwaySelection> returnVal = new LinkedList<>();
        synchronized (microLpSelections) {
            for (MicroLearningPathwaySelection item : microLpSelections) {
                if (item.getId().getCourseId().equals(courseId)) {
                    returnVal.add(item);
                }
            }
        }
        return returnVal;
    }

    @Override
    public KnowledgeObject getKo(String iri) throws DbAccessException {
        synchronized (kos) {
            for (KnowledgeObject item : kos) {
                if (item.getIri().equals(iri)) {
                    return item;
                }
            }
            return null;
        }
    }

    @Override
    public ConceptContainer getCc(String iri) throws DbAccessException {
        synchronized (ccs) {
            for (ConceptContainer item : ccs) {
                if (item.getIri().equals(iri)) {
                    return item;
                }
            }
            return null;
        }
    }

    @Override
    public MicroLearningPathway getMicroLp(String iri) throws DbAccessException {
        synchronized (microLps) {
            for (MicroLearningPathway item : microLps) {
                if (item.getIri().equals(iri)) {
                    return item;
                }
            }
            return null;
        }
    }

    @Override
    public MacroLearningPathway getMacroLp(String iri) throws DbAccessException {
        synchronized (macroLps) {
            for (MacroLearningPathway item : macroLps) {
                if (item.getIri().equals(iri)) {
                    return item;
                }
            }
            return null;
        }
    }
}
