/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, University of Reading, United Kingdom, http://www.isr.reading.ac.uk/
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 02.07.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.db.access;

import org.junit.BeforeClass;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.Test;

/**
 * Various LP-related tests.
 *
 * @author Ovidiu Serban
 * @author Florian Heberle
 */
public class LearningPathwayTest {

    private static final String L1 = "TEST_LP_learner1";
    private static final String L2 = "TEST_LP_learner2";
    private static final String L3 = "TEST_LP_learner3";
    private static final String COURSE = "LPC";
    private static final String MICRO1 = "TEST_LP_mlp1";
    private static final String MACRO1 = "TEST_LP_Malp1";
    private static final String MICRO2 = "TEST_LP_mlp2";
    private static final String MACRO2 = "TEST_LP_Malp2";
    private static final String MICRO3 = "TEST_LP_mlp3";
    private static final String MACRO3 = "TEST_LP_Malp3";
    private static DbAccessor db;

    @BeforeClass
    public static void setup() throws DbAccessException {
        db = TestCommons.getAccessor();
        assertNotNull(db);

        addNewLpSelection(L1, COURSE, MACRO1, MICRO1);
        addNewLpSelection(L2, COURSE, MACRO1, MICRO1);
        addNewLpSelection(L3, COURSE, MACRO1, MICRO1);
        addNewLpSelection(L1, COURSE, MACRO2, MICRO2);
        addNewLpSelection(L3, COURSE, MACRO2, MICRO2);
        addNewLpSelection(L3, COURSE, MACRO3, MICRO2);
    }

    private static void addNewLpSelection(String learner, String course, String macro, String micro) {
        assertTrue(db.storeMacroLpData(macro, course, "", ""));
        assertTrue(db.storeMacroLpSelection(learner, macro, course));
        assertTrue(db.storeMicroLpData(micro, course, "", ""));
        assertTrue(db.storeMicroLpSelection(learner, micro, course));
        try {
            // the sleeps are important because add an LP Selection sets a timestamp for the selection
            // Adding rapidly multiple objects create pseudo-equal timestamp -> last LP selection would not work correctly
            Thread.sleep(50);
        } catch (InterruptedException e) {
            //-- ignore
        }
    }

    private static void addNewLpEvalIncident(String learner, String course) {
        assertTrue(db.storeLpEvaluationIncident(learner, course));
        try {
            // the sleeps are important because add an LP Selection sets a timestamp for the selection
            // Adding rapidly multiple objects create pseudo-equal timestamp -> last LP selection would not work correctly
            Thread.sleep(50);
        } catch (InterruptedException e) {
            //-- ignore
        }
    }

    @AfterClass
    public static void cleanup() throws DbAccessException {
        db.removeLastMicroLpSelection(L1, COURSE);
        db.removeLastMicroLpSelection(L2, COURSE);
        db.removeLastMicroLpSelection(L3, COURSE);
        db.removeLastMicroLpSelection(L1, COURSE);
        db.removeLastMicroLpSelection(L3, COURSE);
        db.removeLastMicroLpSelection(L3, COURSE);

        db.removeLastMacroLpSelection(L1, COURSE);
        db.removeLastMacroLpSelection(L2, COURSE);
        db.removeLastMacroLpSelection(L3, COURSE);
        db.removeLastMacroLpSelection(L1, COURSE);
        db.removeLastMacroLpSelection(L3, COURSE);
        db.removeLastMacroLpSelection(L3, COURSE);
        
        db.removeMacroLpData(MACRO1);
        db.removeMicroLpData(MICRO1);
        db.removeMacroLpData(MACRO2);
        db.removeMicroLpData(MICRO2);
        db.removeMacroLpData(MACRO3);
        //db.removeMicroLpData(MICRO3); not necessary, never stored.
    }

    @Test
    public void testPopularity() {
        assertEquals("LP did not have the expected populairy value.",
                new Integer(0),
                db.getPopularityOfMicroLp(Double.toString(Math.random()), COURSE));
        assertEquals("LP did not have the expected populairy value.",
                new Integer(33),
                db.getPopularityOfMacroLp(MACRO1, COURSE));
        assertEquals("LP did not have the expected populairy value.",
                new Integer(33),
                db.getPopularityOfMacroLp(MACRO2, COURSE));
        assertEquals("LP did not have the expected populairy value.",
                new Integer(33),
                db.getPopularityOfMacroLp(MACRO3, COURSE));
        assertEquals("LP did not have the expected populairy value.",
                new Integer(33),
                db.getPopularityOfMicroLp(MICRO1, COURSE));
        assertEquals("LP did not have the expected populairy value.",
                new Integer(66),
                db.getPopularityOfMicroLp(MICRO2, COURSE));
        assertEquals("LP did not have the expected populairy value.",
                new Integer(0),
                db.getPopularityOfMicroLp(MICRO3, COURSE));
    }

    @Test
    public void testSelection() {
        assertEquals("Wrong macro LP for L1", MACRO2, db.getMacroLpOfLearner(L1, COURSE));
        assertEquals("Wrong micro LP for L1", MICRO2, db.getMicroLpOfLearner(L1, COURSE));
        assertEquals("Wrong macro LP for L2", MACRO1, db.getMacroLpOfLearner(L2, COURSE));
        assertEquals("Wrong micro LP for L2", MICRO1, db.getMicroLpOfLearner(L2, COURSE));
        assertEquals("Wrong macro LP for L3", MACRO3, db.getMacroLpOfLearner(L3, COURSE));
        assertEquals("Wrong micro LP for L3", MICRO2, db.getMicroLpOfLearner(L3, COURSE));
    }

    @Test
    public void testInvalidLp() {
        assertNull(db.getMacroLpOfLearner(null, null));
        assertNull(db.getMacroLpOfLearner("dsgasagasg", null));
        assertNull(db.getMacroLpOfLearner(null, "asgdsagssf"));
        assertNull(db.getMacroLpOfLearner("dsgasagasg", "asgdsagssf"));

        assertNull(db.getMicroLpOfLearner(null, null));
        assertNull(db.getMicroLpOfLearner("dsgasagasg", null));
        assertNull(db.getMicroLpOfLearner(null, "asgdsagssf"));
        assertNull(db.getMicroLpOfLearner("dsgasagasg", "asgdsagssf"));
    }
}
