/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Contributor: Ovidiu Serban, University of Reading, www.reading.ac.uk
 * Version: [0.01]
 * Date: 21.07.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.db.access;

import java.util.List;
import org.apache.commons.lang3.tuple.Pair;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.BeforeClass;
import org.junit.Test;

/**
 * Learning recommendation related tests.
 *
 * @author Florian Heberle
 */
public class LearningRecommendationTest {

    private static final String LID = "LearningRecom_LID";
    private static final String LID2 = "LearningRecom_LID2";
    private static final Double VAL1 = 0.5d;
    private static final Double VAL2 = 0.6d;
    private static final String KOIRI1 = "LearningRecom_KO1";
    private static final String KOIRI2 = "LearningRecom_KO2";
    private static final String KOIRI3 = "LearningRecom_KO3";
    private static final String KOIRI4 = "LearningRecom_KO4";
    private static final String KOIRI5 = "LearningRecom_KO5";
    private static final String KOIRI6 = "LearningRecom_KO6";
    private static final String COURSE = "LRECOMC";
    private static final String CCIRI = "LearningRecom_CCIRI";
    private static DbAccessor db;

    @BeforeClass
    public static void setup() {
        db = TestCommons.getAccessor();
        assertFalse(db.storeLearningRecommendation(LID, KOIRI1, VAL1));
        //should not work, because there is no session and learning step

        assertTrue(db.loginUser(LID, COURSE));

        assertTrue(db.storeLearningStep(LID, COURSE, KOIRI1, CCIRI, System.currentTimeMillis()));
        assertTrue(db.storeLearningRecommendation(LID, KOIRI2, VAL1));

        assertTrue(db.storeLearningStep(LID, COURSE, KOIRI2, CCIRI, System.currentTimeMillis() + 100));
        //No recom - aborted by reflex

        assertTrue(db.storeLearningStep(LID, COURSE, KOIRI2, CCIRI, System.currentTimeMillis() + 200));
        //Refresh of previous KO
        assertTrue(db.storeLearningRecommendation(LID, KOIRI3, VAL1));
        assertTrue(db.storeLearningRecommendation(LID, KOIRI4, VAL2));

        assertTrue(db.storeLearningStep(LID, COURSE, KOIRI5, CCIRI, System.currentTimeMillis() + 300));
        assertTrue(db.storeLearningRecommendation(LID, KOIRI6, VAL1));
    }

    @AfterClass
    public static void cleanUp() {
        db.removeLastLearningRecommendation(LID);
        db.removeLastLearningRecommendation(LID);
        db.removeLastLearningRecommendation(LID);
        db.removeLastLearningRecommendation(LID);
        db.removeLastLearningStep(LID, COURSE);
        db.removeLastLearningStep(LID, COURSE);
        db.removeLastLearningStep(LID, COURSE);
        db.removeLastLearningStep(LID, COURSE);
        db.removeLastSession(LID, COURSE);
    }

    @Test
    public void testLearningRecommendation() {
        List<List<Pair<String, Double>>> stepRecoms = db.getLearningRecommendations(LID, COURSE, Integer.MAX_VALUE);
        assertFalse(stepRecoms.isEmpty());
        assertEquals(4, stepRecoms.size()); //4 steps, 4 recomLists
        //Should be ordered from new to old

        List<Pair<String, Double>> recoms0 = stepRecoms.get(0);
        assertEquals(1, recoms0.size());
        List<Pair<String, Double>> recoms1 = stepRecoms.get(1);
        assertEquals(2, recoms1.size());
        List<Pair<String, Double>> recoms2 = stepRecoms.get(2);
        assertEquals(0, recoms2.size());
        List<Pair<String, Double>> recoms3 = stepRecoms.get(3);
        assertEquals(1, recoms3.size());

        assertEquals(KOIRI6, recoms0.get(0).getLeft());
        assertEquals(VAL1, recoms0.get(0).getRight());

        assertEquals(KOIRI4, recoms1.get(0).getLeft());
        assertEquals(VAL2, recoms1.get(0).getRight());
        assertEquals(KOIRI3, recoms1.get(1).getLeft());
        assertEquals(VAL1, recoms1.get(1).getRight());

        assertEquals(KOIRI2, recoms3.get(0).getLeft());
        assertEquals(VAL1, recoms3.get(0).getRight());
    }

    @Test
    public void testGetRecommendationSelections() {
        List<Pair<List<String>, String>> recomSelections = db.getRecommendationSelections(LID, COURSE);
        assertEquals(2, recomSelections.size());
        //There should only be two, because there are only two recommendations with selections afterwards

        Pair<List<String>, String> recomSelection0 = recomSelections.get(0);
        assertEquals(2, recomSelection0.getLeft().size());
        assertEquals(KOIRI4, recomSelection0.getLeft().get(0));
        assertEquals(KOIRI3, recomSelection0.getLeft().get(1));
        assertEquals(KOIRI5, recomSelection0.getRight());

        Pair<List<String>, String> recomSelection1 = recomSelections.get(1);
        assertEquals(1, recomSelection1.getLeft().size());
        assertEquals(KOIRI2, recomSelection1.getLeft().get(0));
        assertEquals(KOIRI2, recomSelection1.getRight());
    }

    @Test
    public void testInvalidLearningRecommendation() {
        List<List<Pair<String, Double>>> stepRecoms = db.getLearningRecommendations(LID2, COURSE, Integer.MAX_VALUE);
        assertTrue(stepRecoms.isEmpty());
    }
}
