/**
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, University of Reading, United Kingdom, http://www.isr.reading.ac.uk/
 * Contributor: Florian Heberle, University of Applied Sciences Karlsruhe, www.hs-karlsruhe.de
 * Version: [0.01]
 * Date: 11.07.2014
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */
package eu.intuitel.db.access;

import eu.intuitel.types.InteractionWillingness;
import eu.intuitel.types.LearningAttitude;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.BeforeClass;
import org.junit.Test;

/**
 * Various USE data related tests.
 */
public class StaticSessionDataTest {

    private static final String LID = "TESTLEARNER_STATICSESSIONDATA";
    private static final String LID2 = "TESTLEARNER_STATICSESSIONDATA2";
    private static final String LID3 = "TESTLEARNER_STATICSESSIONDATA3";
    private static final Integer RESOLUTIONX = 800;
    private static final Integer RESOLUTIONY = 600;
    private static final String DEVICETYPE = "Tablet";
    private static final String DEVICEDPI = "128";
    private static final Long DEVICETIME = System.currentTimeMillis();
    private static final InteractionWillingness INTERACTION = InteractionWillingness.Willing;
    private static final LearningAttitude ATTITUDE = LearningAttitude.motivated;
    private static final Integer RESOLUTIONX_NEW = 999;
    private static final Integer RESOLUTIONY_NEW = 999;
    private static final String DEVICETYPE_NEW = "PC";
    private static final String DEVICEDPI_NEW = "500";
    private static final Long DEVICETIME_NEW = System.currentTimeMillis() + 100;
    private static final InteractionWillingness INTERACTION_NEW = InteractionWillingness.Unwilling;
    private static final LearningAttitude ATTITUDE_NEW = LearningAttitude.unmotivated;
    private static final String COURSEID = "STATSC";
    private static DbAccessor db;

    @BeforeClass
    public static void setup() {
        db = TestCommons.getAccessor();
    }

    @AfterClass
    public static void cleanUp() {
        db.removeLastSession(LID, COURSEID);
        db.removeLastSession(LID, COURSEID);
        db.removeLastSession(LID2, COURSEID);
        db.removeLastSession(LID3, COURSEID);
        db.removeLastSession(LID3, COURSEID);
    }

    @Test
    public void testSessionData() throws InterruptedException {
        assertTrue(db.loginUser(LID, COURSEID));
        Thread.sleep(100);
        assertTrue(db.storeStaticSessionData(LID, DEVICETYPE, DEVICEDPI, DEVICETIME,
                RESOLUTIONY, RESOLUTIONX, INTERACTION, ATTITUDE));
        assertEquals(DEVICETYPE, db.getDeviceType(LID));
        assertEquals(DEVICEDPI, db.getDeviceDpi(LID));
        assertEquals(RESOLUTIONY, db.getDeviceVerticalResolution(LID));
        assertEquals(RESOLUTIONX, db.getDeviceHorizontalResolution(LID));
        assertTrue(INTERACTION.equals(db.getInteractionWillingness(LID)));
        assertTrue(ATTITUDE.equals(db.getLearningAttitude(LID)));
        Long s1LoginTime = db.getLoginTime(LID, COURSEID);
        assertTrue(s1LoginTime <= new java.util.Date().getTime());
        assertTrue(s1LoginTime >= new Date(
                Calendar.getInstance().get(Calendar.YEAR) - 1900,
                Calendar.getInstance().get(Calendar.MONTH),
                Calendar.getInstance().get(Calendar.DATE),
                0, 0, 0).getTime());

        assertTrue(db.isLoggedIn(LID));
        assertFalse(db.loginUser(LID, COURSEID));
        assertTrue(db.logoutUser(LID));
        Thread.sleep(100);
        assertTrue(db.loginUser(LID, COURSEID));

        List<Long> durations = db.getSessionDurations(LID, COURSEID);
        assertNotNull(durations);
        assertFalse(durations.isEmpty());
        assertTrue(durations.size() == 1);
        assertTrue(durations.get(0) > 0);

        Long s2LoginTime = db.getLoginTime(LID, COURSEID);
        assertTrue(s2LoginTime >= s1LoginTime);

        assertTrue(s2LoginTime > db.getPreviousLoginTime(LID, COURSEID));
    }

    @Test
    public void testInvalidSessionData() {
        assertFalse(db.storeStaticSessionData(LID2, DEVICETYPE, DEVICEDPI, DEVICETIME,
                RESOLUTIONY, RESOLUTIONX, INTERACTION, ATTITUDE));
        assertNull(db.getDeviceType(LID2));
        assertNull(db.getDeviceDpi(LID2));
        assertNull(db.getDeviceVerticalResolution(LID2));
        assertNull(db.getDeviceHorizontalResolution(LID2));
        assertNull(db.getInteractionWillingness(LID2));
        assertNull(db.getLearningAttitude(LID2));
    }

    @Test
    public void testLastSessionsWithCourseId() throws InterruptedException {
        assertTrue(db.loginUser(LID3, COURSEID));
        assertTrue(db.storeStaticSessionData(LID3, DEVICETYPE, DEVICEDPI, DEVICETIME,
                RESOLUTIONY, RESOLUTIONX, INTERACTION, ATTITUDE));
        Thread.sleep(100);
        assertTrue(db.logoutUser(LID3));
        Thread.sleep(100);
        assertTrue(db.loginUser(LID3, COURSEID));
        assertTrue(db.storeStaticSessionData(LID3, DEVICETYPE_NEW, DEVICEDPI_NEW, DEVICETIME_NEW,
                RESOLUTIONY_NEW, RESOLUTIONX_NEW, INTERACTION_NEW, ATTITUDE_NEW));

        assertEquals(DEVICETYPE_NEW, db.getDeviceType(LID3));
        assertEquals(DEVICEDPI_NEW, db.getDeviceDpi(LID3));
        assertEquals(RESOLUTIONY_NEW, db.getDeviceVerticalResolution(LID3));
        assertEquals(RESOLUTIONX_NEW, db.getDeviceHorizontalResolution(LID3));
        assertTrue(INTERACTION_NEW.equals(db.getInteractionWillingness(LID3)));
        assertTrue(ATTITUDE_NEW.equals(db.getLearningAttitude(LID3)));
    }
}
