package eu.intuitel.querybuilder.core;

import eu.intuitel.comm.entities.lpm.LearnerState;
import eu.intuitel.comm.entities.lpm.LearnerStateOntology;
import eu.intuitel.comm.entities.qb.*;
import eu.intuitel.querybuilder.util.EntityHelper;
import eu.intuitel.querybuilder.util.QueryBuilderUtil;
import eu.intuitel.querybuilder.util.QueryBuilderUtilException;
import org.semanticweb.HermiT.Configuration;
import org.semanticweb.HermiT.Reasoner;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.io.OWLOntologyDocumentSource;
import org.semanticweb.owlapi.io.StringDocumentSource;
import org.semanticweb.owlapi.model.*;
import org.semanticweb.owlapi.reasoner.*;

import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.LogManager;

import org.semanticweb.owlapi.util.OWLOntologyMerger;
import org.w3c.dom.Element;
import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSSerializer;

/**
 * QueryBuilder.
 *
 * @author Matthias Stumpp (stumpp@fzi.de)
 */
public class QueryBuilder {

    /*
     * Logger.
     */
    private static final Logger logger = LogManager.getLogger(QueryBuilder.class.getName());


    /**
     * QueryBuilder instance.
     */
    private static QueryBuilder instance;


    /*
     * Number of QueryBuilder workers.
     */
    private static int DEFAULT_NUM_THREADS = 5;


    /**
     * ExecutorService.
     */
    private ThreadPoolExecutor executor;


    /**
     * Constructor QueryBuilder.
     *
     * @return QueryBuilder.
     * @throws QueryBuilderException If something goes wrong.
     */
    private QueryBuilder(int numThreads) throws QueryBuilderException {
        ArrayDeque<QueryBuilderWorker> queryBuilderWorkerQueue = new ArrayDeque<>();
        for (int t=0; t<numThreads; t++) {
            QueryBuilderWorker worker = new QueryBuilderWorker();
            queryBuilderWorkerQueue.add(worker);
            logger.debug("Instantiated QueryBuilderWorker");
        }

        executor = new ThreadPoolExecutorCustom(numThreads, numThreads, 9999,
                TimeUnit.DAYS, new LinkedBlockingQueue<Runnable>(),
                new ThreadFactoryCustom(queryBuilderWorkerQueue), queryBuilderWorkerQueue);
        executor.allowCoreThreadTimeOut(false);
    }


    /**
     * Instantiates and/or returns a singleton instance of QueryBuilder.
     *
     * @return QueryBuilder instance.
     * @throws QueryBuilderException If something goes wrong.
     */
    public static QueryBuilder getInstance() throws QueryBuilderException {
        if (instance == null) {
            return getInstance(DEFAULT_NUM_THREADS);
        }
        return instance;
    }


    /**
     * Instantiates and/or returns a singleton instance of QueryBuilder.
     *
     * @param numThreads The number of threads (= number of QueryBuilder workers).
     * @return QueryBuilder instance.
     * @throws QueryBuilderException If something goes wrong.
     */
    public static QueryBuilder getInstance(int numThreads) throws QueryBuilderException {
        if (instance == null) {
            if (numThreads < 1) {
                throw new QueryBuilderException("Number of threads must be greater or equal to 1.");
            }
            instance = new QueryBuilder(numThreads);
        }
        return instance;
    }


    /**
     * Computes the QueryBuilderOutput.
     *
     * @param learnerState The LearnerState object.
     * @return QueryBuilderOutput The output computed by the QueryBuilder.
     * @throws QueryBuilderException If something goes wrong.
     */
    public QueryBuilderOutput build(LearnerState learnerState)
            throws QueryBuilderException {

        logger.debug("build(LearnerState learnerState) called. Processing...");

        if (learnerState == null) {
            String msg = "LearnerState is null.";
            logger.error(msg);
            throw new QueryBuilderException(msg);
        }

        Future<QueryBuilderOutput> future = executor.submit(new QueryBuilderTask(learnerState));

        if (future.isCancelled()) {
            String msg = "No workers available! Try again later or increase number of workers.";
            logger.error(msg);
            throw new QueryBuilderException(msg);
        }

        try {
            return future.get();
        } catch (Exception e) {
            throw new QueryBuilderException(e.getMessage());
        }
    }


    /**
     * Computes the QueryBuilderOutput.
     *
     * @param learnerStateOntSource The OWLOntologyDocumentSource object. Primarily used for testing.
     * @return QueryBuilderOutput The output computed by the QueryBuilder.
     * @throws QueryBuilderException If something goes wrong.
     */
    public QueryBuilderOutput build(OWLOntologyDocumentSource learnerStateOntSource)
            throws QueryBuilderException{

        logger.debug("build(OWLOntologyDocumentSource learnerStateOntSource) called. Processing...");

        // check if learner state is provided
        if (learnerStateOntSource == null) {
            String msg = "OWLOntologyDocumentSource is null";
            logger.error(msg);
            throw new QueryBuilderException(msg);
        }

        Future<QueryBuilderOutput> future = executor.submit(new QueryBuilderTask(learnerStateOntSource));

        if (future.isCancelled()) {
            String msg = "No workers available! Try again later or increase number of workers.";
            logger.error(msg);
            throw new QueryBuilderException(msg);
        }

        try {
            return future.get();
        } catch (Exception e) {
            throw new QueryBuilderException(e.getMessage());
        }
    }


    /**
     * ThreadFactoryCustom.
     */
    private class ThreadFactoryCustom implements ThreadFactory {

        /**
         * Logger.
         */
        private final Logger logger = LogManager.getLogger(ThreadFactoryCustom.class.getName());

        final AtomicInteger poolNumber = new AtomicInteger(1);
        final ThreadGroup group;
        final AtomicInteger threadNumber = new AtomicInteger(1);
        final String namePrefix;
        final ArrayDeque<QueryBuilderWorker> queryBuilderWorkerQueue;

        ThreadFactoryCustom(ArrayDeque<QueryBuilderWorker> queryBuilderWorkerQueue) {
            SecurityManager s = System.getSecurityManager();
            group = (s != null)? s.getThreadGroup() :
                    Thread.currentThread().getThreadGroup();
            namePrefix = "pool-" +
                    poolNumber.getAndIncrement() +
                    "-thread-";
            this.queryBuilderWorkerQueue = queryBuilderWorkerQueue;
        }

        @Override
        public Thread newThread(Runnable r) {
            if (queryBuilderWorkerQueue.isEmpty())
                return null;
            Thread t = new ThreadCustom(group, r,
                    namePrefix + threadNumber.getAndIncrement(),
                    0, queryBuilderWorkerQueue.pollFirst());
            logger.debug("New Thread: " + t.getName() + " Reference: " +
                    t.getClass().getName() + "@" + Integer.toHexString(t.hashCode()));
            if (t.isDaemon())
                t.setDaemon(false);
            if (t.getPriority() != Thread.NORM_PRIORITY)
                t.setPriority(Thread.NORM_PRIORITY);
            return t;
        }
    }


    /**
     * ThreadCustom.
     */
    private class ThreadCustom extends Thread {

        /**
         * Logger.
         */
        private final Logger logger = LogManager.getLogger(ThreadCustom.class.getName());

        private QueryBuilderWorker worker;

        ThreadCustom(ThreadGroup group, Runnable target, String name, long stackSize, QueryBuilderWorker worker) {
            super(group, target, name, stackSize);
            this.worker = worker;
            logger.debug("New Thread: " + name);
        }

        private QueryBuilderWorker getWorker() {
            return worker;
        }
    }


    /**
     * ThreadPoolExecutorCustom.
     */
    private class ThreadPoolExecutorCustom extends ThreadPoolExecutor {

        /**
         * Logger.
         */
        private final Logger logger = LogManager.getLogger(ThreadPoolExecutorCustom.class.getName());

        private ArrayDeque<QueryBuilderWorker>queryBuilderWorkerQueue;

        public ThreadPoolExecutorCustom(int corePoolSize, int maximumPoolSize, long keepAliveTime,
                                        TimeUnit unit, BlockingQueue<Runnable> workQueue, ThreadFactoryCustom threadFactory,
                                        ArrayDeque<QueryBuilderWorker> queryBuilderWorkerQueue) {
            super(corePoolSize, maximumPoolSize, keepAliveTime, unit, workQueue, threadFactory);
            this.queryBuilderWorkerQueue = queryBuilderWorkerQueue;
        }

        protected void afterExecute(Runnable r, Throwable t) {
            QueryBuilderWorker worker = ((ThreadCustom) Thread.currentThread()).getWorker();
            logger.debug("afterExecute(): Thread: " + Thread.currentThread().getName() + " Reference: " +
                    Thread.currentThread().getClass().getName() + "@" + Integer.toHexString(Thread.currentThread().hashCode()) +
                    " Reference: "
                    + worker.getClass().getName() + "@" + Integer.toHexString(worker.hashCode()));
            if (worker == null) {
                logger.error("QueryBuilderWorker is null in ThreadPoolExecutorCustom.");
            }
            //queryBuilderWorkerQueue.add(worker);
        }
    }


    /**
     * Task for QueryBuilderWorker.
     */
    private class QueryBuilderTask implements Callable<QueryBuilderOutput> {

        /**
         * Logger.
         */
        private final Logger logger = LogManager.getLogger(QueryBuilderTask.class.getName());

        private LearnerState learnerState;

        private OWLOntologyDocumentSource learnerStateOntSource;

        QueryBuilderTask(LearnerState learnerState) {
            this.learnerState = learnerState;
        }

        QueryBuilderTask(OWLOntologyDocumentSource learnerStateOntSource) {
            this.learnerStateOntSource = learnerStateOntSource;
        }

        @Override
        public QueryBuilderOutput call() throws Exception {
            QueryBuilderWorker worker =
                    ((ThreadCustom) Thread.currentThread()).getWorker();
            logger.debug("call(): Thread: " + Thread.currentThread().getName() + " Reference: " +
                    Thread.currentThread().getClass().getName() + "@" + Integer.toHexString(Thread.currentThread().hashCode()) +
                    " Reference: "
                    + worker.getClass().getName() + "@" + Integer.toHexString(worker.hashCode()));
            if (worker == null) {
                throw new QueryBuilderException(
                        "QueryBuilderWorker is null in QueryBuilderTask");
            }
            QueryBuilderOutput output = null;
            if (learnerState != null) {
                output = worker.computeQueryBuilderOutput(learnerState);
            }
            if (learnerStateOntSource != null) {
                output = worker.computeQueryBuilderOutput(learnerStateOntSource);
            }
            return output;
        }
    }


    /**
     * QueryBuilderWorker.
     */
    private class QueryBuilderWorker {

        /**
         * Logger.
         */
        private final Logger logger = LogManager.getLogger(QueryBuilderWorker.class.getName());


        /*
         * Properties containing ontology specific configuration
         */
        private Properties p;


        /**
         * QueryBuilderWorker.
         */
        public QueryBuilderWorker() throws QueryBuilderException {

            InputStream in = QueryBuilder.class.getResourceAsStream("config.properties");
            p = new Properties();
            try {
                p.load(in);
            } catch (IOException e) {
                e.printStackTrace();
            } finally {
                if (in != null) {
                    try {
                        in.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }

            // check if config properties are provided
            if (p == null) {
                String msg = "Unable to load config properties.";
                logger.error("Unable to load config properties.");
                throw new QueryBuilderException(msg);
            }
        }


        /**
         * Computes a QueryBuilderOutput.
         *
         * @param learnerStateOntSource OWLOntologyDocumentSource object.
         * @return QueryBuilderOutput.
         */
        public QueryBuilderOutput computeQueryBuilderOutput(OWLOntologyDocumentSource learnerStateOntSource)
                throws QueryBuilderException{

            logger.debug("build(OWLOntologyDocumentSource learnerStateOntSource) called. Processing...");

            // check if learner state is provided
            if (learnerStateOntSource == null) {
                String msg = "OWLOntologyDocumentSource is null";
                logger.error(msg);
                throw new QueryBuilderException(msg);
            }

            // prepare response
            QueryBuilderOutput queryBuilderOutput = new QueryBuilderOutput();

            logger.debug("QueryBuilder main processing started...");
            QueryBuilderOutput output = build(learnerStateOntSource, queryBuilderOutput);
            logger.info("QueryBuilder main processing finished.");
            return output;
        }


        /**
         * Computes a QueryBuilderOutput.
         *
         * @param learnerState Learnerstate object.
         * @return QueryBuilderOutput.
         */
        public QueryBuilderOutput computeQueryBuilderOutput(LearnerState learnerState)
                throws QueryBuilderException {

            logger.debug("build(LearnerState learnerState) called. Processing...");

            // check if learner state is provided
            if (learnerState == null) {
                String msg = "LearnerState is null";
                logger.error(msg);
                throw new QueryBuilderException(msg);
            }

            if (logger.isDebugEnabled()) {
                // output string to console
                logger.debug("Serializing LearnerState object");
                String theXML = null;
                try {
                    theXML = QueryBuilderUtil.asString(learnerState, LearnerState.class, true);
                } catch (QueryBuilderUtilException e) {
                    e.printStackTrace();
                }
                try {
                    PrintStream out = new PrintStream(new FileOutputStream("ls_object_input.txt"));
                    out.print(theXML);
                } catch (FileNotFoundException e) {
                    e.printStackTrace();
                    logger.debug("Exception while serializing LearnerState object");
                }
            }

            logger.debug("Retrieve LearnerStateOntology...");
            LearnerStateOntology learnerStateOnt = learnerState.getLearnerStateOntology();

            if (learnerStateOnt == null) {
                String msg = "LearnerStateOntology is null!";
                logger.error(msg);
                throw new QueryBuilderException(msg);
            }

            if (logger.isDebugEnabled()) {
                // output string to console
                logger.debug("Serializing LearnerStateOntology object");
                String theXML = null;
                try {
                    theXML = QueryBuilderUtil.asString(learnerStateOnt, LearnerStateOntology.class, true);
                } catch (QueryBuilderUtilException e) {
                    e.printStackTrace();
                }
                try {
                    PrintStream out = new PrintStream(new FileOutputStream("lso_object_input.txt"));
                    out.print(theXML);
                } catch (FileNotFoundException e) {
                    e.printStackTrace();
                    logger.debug("Exception while serializing LearnerStateOntology object");
                }
            }

            List<Element> learnerStateOntStr = learnerStateOnt.getContent();
            if (learnerStateOntStr == null || learnerStateOntStr.size() != 1) {
                String msg = "List of elements in LearnerStateOntology is null, empty or greater than 1. Must be exactly 1.";
                logger.error(msg);
                throw new QueryBuilderException(msg);
            }

            // get the first elem
            Element ontElem = learnerStateOntStr.get(0);

            // serialize xml element containing learner state ontology
            // http://stackoverflow.com/questions/1219596/how-to-i-output-org-w3c-dom-element-to-string-format-in-java
            DOMImplementationLS lsImpl = (DOMImplementationLS) ontElem.getOwnerDocument().getImplementation().getFeature("LS", "3.0");
            LSSerializer serializer = lsImpl.createLSSerializer();
            serializer.getDomConfig().setParameter("xml-declaration", false);
            String str = serializer.writeToString(ontElem);

            if (str == null || str.isEmpty()) {
                String msg = "LearnerStateOntology is null and/or empty";
                logger.error(msg);
                throw new QueryBuilderException(msg);
            }

            if (logger.isDebugEnabled()) {
                try {
                    PrintStream out = new PrintStream(new FileOutputStream("lso_content_input.txt"));
                    out.print(str);
                } catch (FileNotFoundException e) {
                    String msg = "Unable to write.";
                    logger.debug(msg);
                    throw new QueryBuilderException(msg);
                }
            }

            // instantiate a document source for the reasoner
            StringDocumentSource learnerStateOntSource = new StringDocumentSource(str);

            // prepare response
            logger.debug("Initialize QueryBuilderOutput...");
            QueryBuilderOutput queryBuilderOutput = new QueryBuilderOutput();
            queryBuilderOutput.setmId(learnerState.getmId());
            queryBuilderOutput.setuId(learnerState.getuId());
            queryBuilderOutput.setCourseId(learnerState.getCourseId());
            logger.debug("QueryBuilderOutput initialized.");

            logger.debug("QueryBuilder main processing started...");
            QueryBuilderOutput output = build(learnerStateOntSource, queryBuilderOutput);
            logger.info("QueryBuilder main processing finished.");

            if (output != null && logger.isDebugEnabled()) {
                // output string to console
                logger.debug("Serializing QueryBuilderOutput object");
                String theXML = null;
                try {
                    theXML = QueryBuilderUtil.asString(output, QueryBuilderOutput.class, true);
                } catch (QueryBuilderUtilException e) {
                    e.printStackTrace();
                }
                try {
                    PrintStream out = new PrintStream(new FileOutputStream("qbo_object_output.txt"));
                    out.print(theXML);
                } catch (FileNotFoundException e) {
                    e.printStackTrace();
                    logger.debug("Exception while serializing QueryBuilderOutput object");
                }
            }

            return output;
        }


        /**
         * Returns a QueryBuilderOutput object containing the computed KOs according to XSD schema
         * given a LearnerStateOntology.
         *
         * @param ontDocument OWLOntologyDocumentSource containing the learner state ontology.
         * @param queryBuilderOutput Initialized QueryBuilderOutput object.
         * @return QueryBuilderOutput Containing the computed KOs according to XSD schema.
         * @throws QueryBuilderException If something goew wrong internally.
         */
        private QueryBuilderOutput build(OWLOntologyDocumentSource ontDocument,
                                         QueryBuilderOutput queryBuilderOutput) throws QueryBuilderException {

            if (ontDocument == null) {
                throw new QueryBuilderException("LearnerState Ontology is null.");
            }

            OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
            OWLDataFactory dataFactory = manager.getOWLDataFactory();

            InputStream is = QueryBuilder.class.getResourceAsStream(p.getProperty("queryBuilderHelperOnt"));
            OWLOntology queryBuilderHelperOnt = null;
            try {
                queryBuilderHelperOnt = manager.loadOntologyFromOntologyDocument(is);
                logger.debug("QueryBuilder helper ontology loaded.");
            } catch (OWLOntologyCreationException e) {
                e.printStackTrace();
                throw new QueryBuilderException("Cannot load QueryBuilder helper ontology.");
            }

            if (queryBuilderHelperOnt == null) {
                throw new QueryBuilderException("Cannot load QueryBuilder helper ontology.");
            }

            if (logger.isDebugEnabled()) {
                try {
                    PrintStream out = new PrintStream(new FileOutputStream("helper.txt"));
                    out.print(queryBuilderHelperOnt.toString());
                } catch (FileNotFoundException e) {
                    e.printStackTrace();
                }
            }

            OWLOntology learnerStateOnt = null;
            try {
                learnerStateOnt = manager.loadOntologyFromOntologyDocument(ontDocument);
                logger.debug("QueryBuilder LearnerState ontology loaded.");
            } catch (OWLOntologyCreationException e) {
                e.printStackTrace();
                throw new QueryBuilderException("Cannot load QueryBuilder LearnerState ontology.");
            }

            if (learnerStateOnt == null) {
                throw new QueryBuilderException("Cannot load QueryBuilder LearnerState ontology.");
            }

            if (logger.isDebugEnabled()) {
                try {
                    PrintStream out = new PrintStream(new FileOutputStream("lso_loaded.txt"));
                    out.print(learnerStateOnt.toString());
                } catch (FileNotFoundException e) {
                    e.printStackTrace();
                }
            }

            OWLOntologyMerger merger = new OWLOntologyMerger(manager);
            IRI mergedOntologyIRI = IRI.create("http://www.semanticweb.com/queryBuilderReasoningOnt");
            OWLOntology queryBuilderReasoningOnt = null;
            try {
                queryBuilderReasoningOnt = merger.createMergedOntology(manager, mergedOntologyIRI);
            } catch (OWLOntologyCreationException e) {
                System.out.println("OWLOntologyCreationException");
                throw new QueryBuilderException("Unable to merge QueryBuilder helper ontology with LearnerState ontology.");
            }

            if (queryBuilderReasoningOnt == null) {
                throw new QueryBuilderException("Unable to merge QueryBuilder helper ontology with LearnerState ontology.");
            }

            logger.debug("LearnerStateOntology and QueryBuilder helper ontology merged.");

            OWLReasonerFactory factory = new Reasoner.ReasonerFactory();
            /*Configuration configuration = new Configuration();
            configuration.throwInconsistentOntologyException = false;
            OWLReasoner reasoner = factory.createNonBufferingReasoner(queryBuilderReasoningOnt, configuration);*/
            OWLReasoner reasoner = factory.createNonBufferingReasoner(queryBuilderReasoningOnt);

            if (!reasoner.isConsistent()) {
                String msg = "LearnerStateOntology is inconsistent.";
                logger.debug(msg);
                throw new QueryBuilderException(msg);
            }

            OWLClass directSuccessorsOfCurrentKOClazz = dataFactory.getOWLClass(IRI.create(p.getProperty("directSuccessorsOfCurrentKOStr")));
            logger.debug(directSuccessorsOfCurrentKOClazz.toString());
            OWLClass allSuccessorsOfCurrentKOClazz = dataFactory.getOWLClass(IRI.create(p.getProperty("allSuccessorsOfCurrentKOStr")));
            logger.debug(allSuccessorsOfCurrentKOClazz.toString());
            OWLClass allPredecessorsOfCurrentKOClazz = dataFactory.getOWLClass(IRI.create(p.getProperty("allPredecessorsOfCurrentKOStr")));
            logger.debug(allPredecessorsOfCurrentKOClazz.toString());
            OWLClass directSuccessorsOfPreviousKOClazz = dataFactory.getOWLClass(IRI.create(p.getProperty("directSuccessorsOfPreviousKOStr")));
            logger.debug(directSuccessorsOfPreviousKOClazz.toString());

            KnowledgeObject ko;
            OWLNamedIndividual currentKO = getCurrentKO(dataFactory, reasoner);
            OWLNamedIndividual previousKO = getPreviousKO(dataFactory, reasoner);

            ///////////
            // CurrentKnowledgeObject
            ///////////
            CurrentKnowledgeObject currentKnowledgeObject = null;

            ///////////
            // DirectSuccessorsOfCurrentKO

            DirectSuccessors directSuccessorsObj = null;
            List<KnowledgeObject> directSuccessorsList = null;

            NodeSet<OWLNamedIndividual> directSuccessorsNamedIndividuals = reasoner.getInstances(directSuccessorsOfCurrentKOClazz, false);
            Set<OWLNamedIndividual> directSuccessorsNamedIndividuals2 = directSuccessorsNamedIndividuals.getFlattened();
            logger.debug(directSuccessorsNamedIndividuals2.toString());
            if (directSuccessorsNamedIndividuals2.size() != 0) {
                logger.debug("Direct Successors: Successors for current KO exists.");
                currentKnowledgeObject = new CurrentKnowledgeObject();
                queryBuilderOutput.setCurrentKnowledgeObject(currentKnowledgeObject);
                // set current knowledge object
                if (currentKO != null) {
                    currentKnowledgeObject.setLmsID(getLmsID(queryBuilderReasoningOnt, dataFactory, currentKO));
                }
                directSuccessorsObj = new DirectSuccessors();
                directSuccessorsList = new ArrayList<KnowledgeObject>();
                directSuccessorsObj.setKnowledgeObject(directSuccessorsList);
                currentKnowledgeObject.setDirectSuccessors(directSuccessorsObj);
            }
            for(OWLNamedIndividual b : directSuccessorsNamedIndividuals2) {
                assert directSuccessorsList != null;
                ko = new KnowledgeObject();
                ko.setLmsID(getLmsID(queryBuilderReasoningOnt, dataFactory, b));
                directSuccessorsList.add(ko);
                logger.debug("Direct Successors: Successor for current KO added to QueryBuilder output: " + b.toString());
            }

            ///////////
            // AllSuccessorsOfCurrentKO

            AllSuccessors allSuccessorsObj = null;
            List<KnowledgeObject> allSuccessorsList = null;

            NodeSet<OWLNamedIndividual> allSuccessorsNamedIndividuals = reasoner.getInstances(allSuccessorsOfCurrentKOClazz, false);
            Set<OWLNamedIndividual> allSuccessorsNamedIndividuals2 = allSuccessorsNamedIndividuals.getFlattened();
            if (allSuccessorsNamedIndividuals2.size() != 0) {
                logger.debug("AllSuccessors: Successors for current KO exists.");
                if (currentKnowledgeObject == null) {
                    currentKnowledgeObject = new CurrentKnowledgeObject();
                    queryBuilderOutput.setCurrentKnowledgeObject(currentKnowledgeObject);
                    // set current knowledge object
                    if (currentKO != null) {
                        currentKnowledgeObject.setLmsID(getLmsID(queryBuilderReasoningOnt, dataFactory, currentKO));
                    }
                }
                allSuccessorsObj = new AllSuccessors();
                allSuccessorsList = new ArrayList<KnowledgeObject>();
                allSuccessorsObj.setKnowledgeObject(allSuccessorsList);
                currentKnowledgeObject.setAllSuccessors(allSuccessorsObj);
            }
            for(OWLNamedIndividual b : allSuccessorsNamedIndividuals2) {
                assert allSuccessorsList != null;
                ko = new KnowledgeObject();
                ko.setLmsID(getLmsID(queryBuilderReasoningOnt, dataFactory, b));
                allSuccessorsList.add(ko);
                logger.debug("AllSuccessors: Successor for current KO added to QueryBuilder output: " + b.toString());
            }

            //////////
            // AllSuccessorsOfCurrentKO

            AllPredecessors allPredecessorsObj = null;
            List<KnowledgeObject> allPredecessorsList = null;

            NodeSet<OWLNamedIndividual> allPredecessorsNamedIndividuals = reasoner.getInstances(allPredecessorsOfCurrentKOClazz, false);
            Set<OWLNamedIndividual> allPredecessorsNamedIndividuals2 = allPredecessorsNamedIndividuals.getFlattened();
            if (allPredecessorsNamedIndividuals2.size() != 0) {
                logger.debug("AllPredecessors: Predecessors for current KO exists.");
                if (currentKnowledgeObject == null) {
                    currentKnowledgeObject = new CurrentKnowledgeObject();
                    queryBuilderOutput.setCurrentKnowledgeObject(currentKnowledgeObject);
                    // set current knowledge object
                    if (currentKO != null) {
                        currentKnowledgeObject.setLmsID(getLmsID(queryBuilderReasoningOnt, dataFactory, currentKO));
                    }
                }
                allPredecessorsObj = new AllPredecessors();
                allPredecessorsList = new ArrayList<KnowledgeObject>();
                allPredecessorsObj.setKnowledgeObject(allPredecessorsList);
                currentKnowledgeObject.setAllPredecessors(allPredecessorsObj);
            }
            for(OWLNamedIndividual b : allPredecessorsNamedIndividuals2) {
                assert allPredecessorsList != null;
                ko = new KnowledgeObject();
                ko.setLmsID(getLmsID(queryBuilderReasoningOnt, dataFactory, b));
                allPredecessorsList.add(ko);
                logger.debug("AllPredecessors: Predecessors for current KO added to QueryBuilder output: " + b.toString());
            }

            ///////////
            // PreviousKnowledgeObject
            ///////////
            PreviousKnowledgeObject previousKnowledgeObject = null;

            //////////
            // DirectSuccessorsOfPreviousKO

            directSuccessorsObj = null;
            directSuccessorsList = null;

            NodeSet<OWLNamedIndividual> directSuccessorsNamedIndividualsOfPreviousKO = reasoner.getInstances(directSuccessorsOfPreviousKOClazz, false);
            Set<OWLNamedIndividual> directSuccessorsNamedIndividualsOfPreviousKO2 = directSuccessorsNamedIndividualsOfPreviousKO.getFlattened();
            if (directSuccessorsNamedIndividualsOfPreviousKO2.size() != 0) {
                logger.debug("Direct successors for previous KO exists.");
                previousKnowledgeObject = new PreviousKnowledgeObject();
                queryBuilderOutput.setPreviousKnowledgeObject(previousKnowledgeObject);
                if (previousKO != null) {
                    previousKnowledgeObject.setLmsID(getLmsID(queryBuilderReasoningOnt, dataFactory, previousKO));
                }
                directSuccessorsObj = new DirectSuccessors();
                directSuccessorsList = new ArrayList<KnowledgeObject>();
                directSuccessorsObj.setKnowledgeObject(directSuccessorsList);
                previousKnowledgeObject.setDirectSuccessors(directSuccessorsObj);
            }
            for(OWLNamedIndividual b : directSuccessorsNamedIndividualsOfPreviousKO2) {
                assert directSuccessorsList != null;
                ko = new KnowledgeObject();
                ko.setLmsID(getLmsID(queryBuilderReasoningOnt, dataFactory, b));
                directSuccessorsList.add(ko);
                logger.debug("Direct successor for previous KO added to QueryBuilder output: " + b.toString());
            }

            // get complete kos
            Set<OWLNamedIndividual> completeLos = getCompleteLos(queryBuilderReasoningOnt, dataFactory);
            logger.debug(completeLos.toString());

            // get all didactic factor values
            OWLClass didacticFactorClazz = dataFactory.getOWLClass(IRI.create(p.getProperty("didacticFactorsStr")));
            NodeSet<OWLNamedIndividual> didacticFactors = reasoner.getInstances(didacticFactorClazz, false);
            Set<OWLNamedIndividual> didacticFactors2 = didacticFactors.getFlattened();
            logger.debug(didacticFactors2.toString());

            // get didactic factor value for learnerstate
            Set<OWLNamedIndividual> didacticValsForLearnerState = getDidacticValsForLearnerState(queryBuilderReasoningOnt, dataFactory);
            logger.debug(didacticValsForLearnerState.toString());

            ///////
            // didactic factor values for learner state
            ///////

            ///////////
            // DidacticFactors
            ///////////

            DidacticFactors didacticFactorsObj = null;
            List<DidacticFactor> didacticFactorsList = null;
            DidacticFactor dFactor = null;

            if (didacticValsForLearnerState.size() != 0) {
                logger.debug("Didactic factor values for learner state exists.");
                didacticFactorsObj = new DidacticFactors();
                queryBuilderOutput.setDidacticFactors(didacticFactorsObj);
                didacticFactorsList = new ArrayList<DidacticFactor>();
                didacticFactorsObj.setDidacticFactor(didacticFactorsList);
            }
            for(OWLNamedIndividual b : didacticValsForLearnerState) {
                assert didacticFactorsList != null;
                if (b.getIRI() != null && !b.getIRI().toString().isEmpty()) {
                    String res = b.getIRI().toString();
                    String[] str = b.getIRI().toString().split("_");
                    String toRemove = str[str.length-1];
                    dFactor = new DidacticFactor();
                    dFactor.setName(res.replace("_"+toRemove,""));
                    dFactor.setValue(res);
                    didacticFactorsList.add(dFactor);
                }
                logger.debug("Didactic factor values for learner stateO added to QueryBuilder output: " + b.toString());
            }

            ///////////
            // Criteria
            ///////////

            Criteria criteria = null;
            List<KnowledgeObject> criterionKOList = null;

            ////////
            // Hard criteria

            HardCriterion hCriterionObj = null;
            List<HardCriterion> hCriterionList = null;

            OWLClassExpression hardCriteria = null;
            hardCriteria = dataFactory.getOWLClass(IRI.create(p.getProperty("hardCriterionStr")));
            Set<OWLEntity> hardCriterias = EntityHelper.getSubClasses(queryBuilderReasoningOnt, hardCriteria.asOWLClass(), false, false);
            logger.debug(hardCriterias.toString());

            NodeSet<OWLNamedIndividual> kos;
            Set<OWLNamedIndividual> kos2;
            for (OWLEntity clazz : hardCriterias) {

                if (criteria == null) {
                    criteria = new Criteria();
                    queryBuilderOutput.setCriteria(criteria);
                }

                if (hCriterionList == null) {
                    assert criteria != null;
                    hCriterionList = new ArrayList<HardCriterion>();
                    criteria.setHardCriterion(hCriterionList);
                }

                assert hCriterionList != null;
                hCriterionObj = new HardCriterion();
                hCriterionObj.setIri(clazz.asOWLClass().getIRI().toString());
                hCriterionList.add(hCriterionObj);
                criterionKOList = new ArrayList<KnowledgeObject>();
                hCriterionObj.setKnowledgeObject(criterionKOList);

                kos = reasoner.getInstances(clazz.asOWLClass(), false);
                kos2 = kos.getFlattened();

                if (kos2 != null && kos2.size() > 0) {
                    logger.debug("KOs for hard criteria exists.");
                    for (OWLNamedIndividual ko2 : kos2) {
                        assert criterionKOList != null;
                        ko = new KnowledgeObject();
                        ko.setLmsID(getLmsID(queryBuilderReasoningOnt, dataFactory, ko2));
                        criterionKOList.add(ko);
                    }
                }
            }

            ////////
            // Soft criteria

            SoftCriterion sCriterionObj = null;
            List<SoftCriterion> sCriterionList = null;

            OWLClassExpression softCriteria = null;
            softCriteria = dataFactory.getOWLClass(IRI.create(p.getProperty("softCriterionStr")));
            Set<OWLEntity> softCriterias = EntityHelper.getSubClasses(queryBuilderReasoningOnt, softCriteria.asOWLClass(), false, false);
            logger.debug(softCriterias.toString());

            for (OWLEntity clazz : softCriterias) {

                if (criteria == null) {
                    criteria = new Criteria();
                    queryBuilderOutput.setCriteria(criteria);
                }

                if (sCriterionList == null) {
                    assert criteria != null;
                    sCriterionList = new ArrayList<SoftCriterion>();
                    criteria.setSoftCriterion(sCriterionList);
                }

                assert sCriterionList != null;
                sCriterionObj = new SoftCriterion();
                sCriterionObj.setIri(clazz.asOWLClass().getIRI().toString());
                sCriterionList.add(sCriterionObj);
                criterionKOList = new ArrayList<KnowledgeObject>();
                sCriterionObj.setKnowledgeObject(criterionKOList);

                kos = reasoner.getInstances(clazz.asOWLClass(), false);
                kos2 = kos.getFlattened();

                if (kos2 != null && kos2.size() > 0) {

                    logger.debug("KOs for soft criteria exists.");
                    for (OWLNamedIndividual ko2 : kos2) {
                        assert criterionKOList != null;
                        ko = new KnowledgeObject();
                        ko.setLmsID(getLmsID(queryBuilderReasoningOnt, dataFactory, ko2));
                        criterionKOList.add(ko);
                    }
                }
            }

            if (logger.isDebugEnabled()) {
                try {
                    logger.debug(QueryBuilderUtil.asString(queryBuilderOutput, QueryBuilderOutput.class, true));
                } catch (QueryBuilderUtilException e) {
                    new QueryBuilderException(e.getMessage());
                }
            }

            return queryBuilderOutput;
        }


        /**
         * Helpers.
         */
        private Set<OWLNamedIndividual> getDidacticValsForLearnerState(OWLOntology ont, OWLDataFactory dataFactory) {

            OWLNamedIndividual learnerState = dataFactory.getOWLNamedIndividual(IRI.create(p.getProperty("learnerStateStr")));
            OWLObjectProperty hasDidacticFactorValue = dataFactory.getOWLObjectProperty(IRI.create(p.getProperty("hasDidacticFactorValueStr")));

            Set<OWLNamedIndividual> individuals = new HashSet<OWLNamedIndividual>();
            for (OWLIndividual ind : learnerState.getObjectPropertyValues(hasDidacticFactorValue, ont)) {
                individuals.add((OWLNamedIndividual) ind);
            }

            return individuals;
        }


        private Set<OWLClassExpression> getDidacticFactorForDidacticVal(OWLOntology ont, OWLDataFactory dataFactory, OWLNamedIndividual val) {

            Set<OWLClassExpression> clazzes = new HashSet<OWLClassExpression>();
            for (OWLClassExpression clazz : val.getTypes(ont)) {
                clazzes.add(clazz);
            }

            return clazzes;
        }


        private Set<OWLNamedIndividual> getCompleteLos(OWLOntology ont, OWLDataFactory dataFactory) {

            OWLClass completeLos = dataFactory.getOWLClass(IRI.create(p.getProperty("completeLoStr")));

            Set<OWLNamedIndividual> individuals = new HashSet<>();
            for (OWLIndividual ind : completeLos.getIndividuals(ont)) {
                individuals.add((OWLNamedIndividual) ind);
            }

            return individuals;
        }


        private OWLNamedIndividual getCurrentKO(OWLDataFactory dataFactory, OWLReasoner reasoner) {

            OWLClass koClazz = dataFactory.getOWLClass(IRI.create(p.getProperty("koStr")));
            NodeSet<OWLNamedIndividual> koInds = reasoner.getInstances(koClazz, false);
            Set<OWLNamedIndividual> koInds2 = koInds.getFlattened();

            OWLClass currentKO = dataFactory.getOWLClass(IRI.create(p.getProperty("currentLOStr")));
            NodeSet<OWLNamedIndividual> currentKOInds = reasoner.getInstances(currentKO, false);
            Set<OWLNamedIndividual> currentKOInds2 = currentKOInds.getFlattened();
            for (OWLNamedIndividual ind : currentKOInds2) {
                if (koInds2.contains(ind)) {
                    return ind;
                }
            }

            return null;
        }


        private OWLNamedIndividual getPreviousKO(OWLDataFactory dataFactory, OWLReasoner reasoner) {

            OWLClass koClazz = dataFactory.getOWLClass(IRI.create(p.getProperty("koStr")));
            NodeSet<OWLNamedIndividual> koInds = reasoner.getInstances(koClazz, false);
            Set<OWLNamedIndividual> koInds2 = koInds.getFlattened();

            OWLClass previousKO = dataFactory.getOWLClass(IRI.create(p.getProperty("previousLOStr")));
            NodeSet<OWLNamedIndividual> previousKOInds = reasoner.getInstances(previousKO, false);
            Set<OWLNamedIndividual> previousKOInds2 = previousKOInds.getFlattened();
            for (OWLNamedIndividual ind : previousKOInds2) {
                if (koInds2.contains(ind)) {
                    return ind;
                }
            }

            return null;
        }


        // SoftCriteria helper

        private List<OWLNamedIndividual> getSCsForDidacticVal(OWLOntology ont, OWLDataFactory dataFactory, OWLNamedIndividual didacticFactorVal) {

            OWLObjectProperty hasSoftCriterionProp = dataFactory.getOWLObjectProperty(IRI.create(p.getProperty("hasSoftCriterionStr")));

            List<OWLNamedIndividual> individuals = new ArrayList<OWLNamedIndividual>();
            for (OWLIndividual ind : didacticFactorVal.getObjectPropertyValues(hasSoftCriterionProp, ont)) {
                individuals.add((OWLNamedIndividual) ind);
            }

            return individuals;
        }


        private List<OWLClassExpression> getKOsForSoftCriterion(OWLOntology ont, OWLDataFactory dataFactory, OWLNamedIndividual softCriterion) {

            OWLAnnotationProperty hasSoftCriterionClassExpressionProp = dataFactory.
                    getOWLAnnotationProperty(IRI.create(p.getProperty("hasSoftCriterionClassExpressionStr")));

            List<OWLClassExpression> clazzExpressions = new ArrayList<OWLClassExpression>();
            for (OWLAnnotation annotation : softCriterion.getAnnotations(ont, hasSoftCriterionClassExpressionProp)) {
                clazzExpressions.add(dataFactory.getOWLClass(IRI.create(annotation.getValue().toString())));
            }

            return clazzExpressions;
        }


        // HardCriteria helper

        private Set<OWLClassExpression> getKOsForDidacticVal(OWLOntology ont, OWLDataFactory dataFactory, OWLNamedIndividual didacticFactorVal) {

            OWLAnnotationProperty hasHardCriterionClassExpressionProp = dataFactory.
                    getOWLAnnotationProperty(IRI.create(p.getProperty("hasHardCriterionClassExpressionStr")));

            Set<OWLClassExpression> clazzExpressions = new HashSet<OWLClassExpression>();
            for (OWLAnnotation annotation : didacticFactorVal.getAnnotations(ont, hasHardCriterionClassExpressionProp)) {
                clazzExpressions.add(dataFactory.getOWLClass(IRI.create(annotation.getValue().toString())));
            }

            return clazzExpressions;
        }


        private Set<OWLClassExpression> getHardCriteria(OWLOntology ont, OWLDataFactory dataFactory, OWLReasoner reasoner) {

            OWLClassExpression hardCriteria = dataFactory.getOWLClass(IRI.create(p.getProperty("hardCriterionStr")));

            Set<OWLClassExpression> clazzExpressions = new HashSet<OWLClassExpression>();
            for (Node<OWLClass> subClazz : reasoner.getSubClasses(hardCriteria, false)) {
                for (OWLClass clazz : subClazz.getEntities())
                    clazzExpressions.add((OWLClassExpression) clazz);
            }

            return clazzExpressions;
        }


        private Set<OWLClassExpression> getSoftCriteria(OWLOntology ont, OWLDataFactory dataFactory, OWLReasoner reasoner) {

            OWLClassExpression hardCriteria = dataFactory.getOWLClass(IRI.create(p.getProperty("softCriterionStr")));

            Set<OWLClassExpression> clazzExpressions = new HashSet<OWLClassExpression>();
            for (Node<OWLClass> subClazz : reasoner.getSubClasses(hardCriteria, false)) {
                for (OWLClass clazz : subClazz.getEntities())
                    clazzExpressions.add((OWLClassExpression) clazz);
            }

            return clazzExpressions;
        }

        private String getLmsID(OWLOntology ont, OWLDataFactory dataFactory, OWLNamedIndividual koInd) {

            OWLDataProperty isLinkedWithLmsElementProp = dataFactory.getOWLDataProperty(IRI.create(p.getProperty("isLinkedWithLmsElement")));

            List<OWLLiteral> literals = new ArrayList<>();
            for (OWLLiteral lit : koInd.getDataPropertyValues(isLinkedWithLmsElementProp, ont)) {
                literals.add((OWLLiteral) lit);
            }

            if (literals == null || literals.size() != 1) {
                return "";
            }

            return literals.get(0).getLiteral();
        }
    }

}
