package eu.intuitel.querybuilder.service;

import eu.intuitel.comm.entities.exceptions.IntuitelException;
import eu.intuitel.comm.entities.lpm.LearnerState;
import eu.intuitel.comm.entities.qb.QueryBuilderOutput;
import eu.intuitel.comm.services.IQueryBuilderListener;
import eu.intuitel.querybuilder.core.QueryBuilder;
import eu.intuitel.querybuilder.core.QueryBuilderException;
import eu.intuitel.querybuilder.service.mockup.MockupConnectionListenerImpl;
import eu.intuitel.querybuilder.service.mockup.MockupQBServiceConnection;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.List;
import java.util.Properties;

import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.LogManager;

/**
 * QueryBuilderService.
 *
 * @author Matthias Stumpp (stumpp@fzi.de)
 */
public class QueryBuilderService {

    private static final String FILENAME_PROPERTIES = "setup.properties";

    // Properties for the Advanced Communication Layer:
    private static final String PROP_COMM_USER = "username";
    private static final String PROP_COMM_PASS = "password";
    private static final String PROP_COMM_XMPP_SERVER = "xmppServer";
    private static final String PROP_COMM_XMPP_PORT = "xmppServerPort";
    private static final String PROP_COMM_TSB_USER = "tsbUsername";

    // Properties for the Basic Communication Layer:
    private static final String PROP_COMM_URI = "uri";
    private static final String PROP_COMM_GATEWAY = "intuitelGateway";

    private static final String PROP_CONNECTION_TYPE = "connectionType";
    private static final String PROP_CONNECTION_TYPE_COMM_LAYER = "commLayer";

    // Specifies the communication layer type to be used (basic or advanced):
    private static final String PROP_COMM_LAYER_TYPE = "commLayerType";

    // Property values for the communication layer type:
    private static final String PROP_VALUE_COMM_LAYER_TYPE_BASIC = "basic";
    private static final String PROP_VALUE_COMM_LAYER_TYPE_ADVANCED = "advanced";

    private static final String PROP_VALUE_NUM_WORKERS = "numberWorkers";
    private static final String PROP_VALUE_HERAKLES_URI = "heraklesURI";

    private static final Logger logger = LogManager.getLogger(QueryBuilderService.class.getName());

    public static void main(String[] args) throws Exception {

        Properties properties = loadProperties();
        if (properties == null) {
            logger.error("No properties file found. System exit.");
            System.exit(1);
        }

        logger.debug("Setting up communication layer...");

        AbstractCommunicationLayer communicationLayer;
        if (PROP_CONNECTION_TYPE_COMM_LAYER.equals(properties.getProperty(PROP_CONNECTION_TYPE))) {

            logger.debug("Using remote communication layer. Setting up...");
            communicationLayer = CommunicationLayerDelegate.instance;

            try {
                if (properties.getProperty(PROP_COMM_LAYER_TYPE).equals(PROP_VALUE_COMM_LAYER_TYPE_ADVANCED)) {
                    logger.debug("Using advanced communication layer. Setting up...");
                    communicationLayer.OpenAdvanced(
                            properties.getProperty(PROP_COMM_USER),
                            properties.getProperty(PROP_COMM_PASS),
                            properties.getProperty(PROP_COMM_XMPP_SERVER),
                            Integer.parseInt(properties.getProperty(PROP_COMM_XMPP_PORT)),
                            properties.getProperty(PROP_COMM_TSB_USER));
                    logger.debug("Advanced communication layer ready.");

                } else if (properties.getProperty(PROP_COMM_LAYER_TYPE).equals(PROP_VALUE_COMM_LAYER_TYPE_BASIC)) {
                    logger.debug("Using basic communication layer. Setting up...");
                    communicationLayer.OpenBasic(
                            URI.create(properties.getProperty(PROP_COMM_URI)),
                            URI.create(properties.getProperty(PROP_COMM_GATEWAY)));
                    logger.debug("Basic communication layer ready.");

                } else {
                    logger.error("Illegal or unknown communication layer type.");
                    System.exit(1);
                }

                logger.debug("Remote communication layer ready.");

            } catch (IOException e) {
                logger.error(e.getMessage());
                System.exit(1);
            }

        } else {
            logger.debug("Using mockup communication layer. Setting up...");
            communicationLayer = MockupQBServiceConnection.instance;
            MockupQBServiceConnection conn = (MockupQBServiceConnection) communicationLayer;
            conn.setMockupConnectionListener(new MockupConnectionListenerImpl());
            logger.debug("Mockup communication layer ready.");
        }

        logger.debug("Initializing QueryBuilderService...");
        int numWorkers = Integer.parseInt(properties.getProperty(PROP_VALUE_NUM_WORKERS));
        String heraklesURI = properties.getProperty(PROP_VALUE_HERAKLES_URI);
        QueryBuilderServiceRunner myRunnable = new QueryBuilderServiceRunner(communicationLayer, numWorkers, heraklesURI);
        Thread myThread = new Thread(myRunnable);
        myThread.setDaemon(true);
        myThread.start();
        logger.info("QueryBuilderService initialized and ready.");

        // prevent the service from terminating
        while (true) {
            Thread.sleep(1000);
        }
    }

    private static Properties loadProperties() {

        logger.debug("Properties loading...");

        InputStream in = null;
        if (new File(FILENAME_PROPERTIES).exists()) {
            try {
                in = new FileInputStream(FILENAME_PROPERTIES);
            } catch (FileNotFoundException e1s) {
                logger.error("Properties file " + FILENAME_PROPERTIES + " not found in current directory. " +
                        "Will look for it in the classpath.Searching classpath.");
            }
        } else {
            in = QueryBuilder.class.getResourceAsStream(FILENAME_PROPERTIES);
        }

        if (in == null) {
            logger.error("Properties file " + FILENAME_PROPERTIES + " not found.");
            logger.error("Unable to load config properties. System exit.");
            System.exit(1);
        }

        Properties p = new Properties();
        try {
            p.load(in);
        } catch (IOException e) {
            logger.error(e.getMessage());
            logger.error("Unable to load config properties. System exit.");
            System.exit(1);
        }

        try {
            in.close();
        } catch (IOException e) {
            logger.error(e.getMessage());
            logger.error("Unable to load config properties. System exit.");
            System.exit(1);
        }

        logger.info("Properties loaded.");
        return p;
    }

    private static class QueryBuilderServiceRunner implements IQueryBuilderListener,
            CommunicationEventListener, Runnable {

        private AbstractCommunicationLayer connection;

        private int numWorkers;

        private String heraklesURI;

        private static final Logger logger = LogManager.getLogger(QueryBuilderServiceRunner.class.getName());

        public QueryBuilderServiceRunner(AbstractCommunicationLayer connection, int numWorkers, String heraklesURI) {
            this.connection = connection;
            this.numWorkers = numWorkers;
            this.heraklesURI = heraklesURI;
        }

        @Override
        public void onClose(AbstractCommunicationLayer communicationLayer) {
            logger.debug("onClose called. terminating.");
            terminate();
        }

        @Override
        public void onOpen(AbstractCommunicationLayer communicationLayer) {
            //do nothing when communication layer opens
            logger.debug("onOpen called. nothing to do.");
        }

        public void onNewLearnerStates(List<eu.intuitel.comm.entities.lpm.LearnerState> learnerStates) {
            logger.debug("onNewLearnerStates called. Processing request...");

            QueryBuilderOutput output = null;

            // check if learner state is of size 1
            if (learnerStates == null || learnerStates.size() != 1) {
                logger.error("QueryBuilderException: List of learner states is null, empty or has more than 1 one element.");

            } else {

                LearnerState learnerState = learnerStates.get(0);
                try {
                    logger.debug("QueryBuilder called. Waiting for response...");
                    QueryBuilder instance = QueryBuilder.getInstance(5);
                    if (instance != null) {
                        output = instance.build(learnerState);
                        logger.debug("QueryBuilder responded.");
                    } else {
                        logger.error("QueryBuilderException: QueryBuilder master not available!");
                    }

                } catch (QueryBuilderException e) {
                    logger.error("QueryBuilderException: " + e.getMessage());
                }
            }

            // send output, even if it's null
            try {
                logger.debug("Hand QueryBuilderOutput over to connection...");
                connection.sendQueryBuilderOutput(output);
                logger.debug("QueryBuilderOutput handed over.");

            } catch (IntuitelException e) {
                logger.error("IntuitelException: " + e.getMessage());
            }
        }

        @Override
        public void run() {
            connection.setQueryBuilderListener(this);
            connection.addCommunicationLayerEventListener(this);
        }

        public void terminate() {
            connection.setQueryBuilderListener(this);
            connection.Close();
            connection.removeCommunicationLayerEventListener(this);
        }
    }
}