package eu.intuitel.querybuilder.util;

import java.util.HashSet;
import java.util.Set;

import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLDataProperty;
import org.semanticweb.owlapi.model.OWLEntity;
import org.semanticweb.owlapi.model.OWLObjectProperty;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLSubClassOfAxiom;
import org.semanticweb.owlapi.model.OWLSubDataPropertyOfAxiom;
import org.semanticweb.owlapi.model.OWLSubObjectPropertyOfAxiom;

/**
 * A helper class for base matcher that use annotation properties.
 *
 * @author Matthias Stumpp (stumpp@fzi.de)
 *
 */
public class EntityHelper {

    /**
     * Returns super classes of a given class.
     *
     * @param clazz The OWLClass the super classes to be computed of
     * @param includeItself Whether to include the class itself in the returned set of super classes (true) or not (false)
     * @param direct Whether to include only direct super class of the class (true) or all super classes (false)
     * @return Set<OWLClass> Set of super classes
     */
    public static Set<OWLEntity> getSuperClasses(OWLOntology ontology, OWLClass clazz, boolean direct, boolean includeItself) {

        if (ontology == null || clazz == null) {
            final String errMsg = "ontology == null || clazz == null";
            throw new NullPointerException(errMsg);
        }

        Set<OWLEntity> classes = new HashSet<OWLEntity>();

        // Add itself if requested
        if (includeItself)
            classes.add(clazz);

        // If direct only, use corresponding methods
        if (direct) {
            Set<OWLClassExpression> superClasses = clazz.getSuperClasses( ontology );
            for( OWLClassExpression currentSuperClass : superClasses )
                if( currentSuperClass instanceof OWLClass )
                    classes.add( currentSuperClass.asOWLClass() );

            // Otherwise, recursively iterate over all sub class axioms
        } else {

            getSuperClassesHelper(ontology, ontology.getSubClassAxiomsForSubClass(clazz), classes);

            // Add owl:Thing if not contained
            OWLClass thing = ontology.getOWLOntologyManager().getOWLDataFactory().getOWLThing();
            if (!classes.contains(thing))
                classes.add(thing);
        }

        return classes;
    }


    /**
     * Helper to recursively determine sub classes.
     *
     * @param axioms Set of OWLSubClassOfAxiom representing the super/sub class relationship
     * @param clazzes Stores determined sub classes
     * @return Set<OWLClass> Set of sub classes
     */
    private static void getSuperClassesHelper(OWLOntology ontology, Set<OWLSubClassOfAxiom> axioms, Set<OWLEntity> clazzes) {
        OWLClass clazz;
        Set<OWLSubClassOfAxiom> currentAxioms;
        for (OWLSubClassOfAxiom axiom : axioms) {
            if (!(axiom.getSuperClass() instanceof OWLClass))
                continue;

            clazz = (OWLClass) axiom.getSuperClass();

            // Add class to set of super classes
            if (!clazzes.contains(clazz))
                clazzes.add(clazz);

            // Compute new set of super class axioms
            currentAxioms = ontology.getSubClassAxiomsForSubClass(clazz);

            // Continue if no more super class axioms
            if (currentAxioms.size() == 0)
                continue;

            // Further iterate only if flag direct set to false
            getSuperClassesHelper(ontology, currentAxioms, clazzes);
        }
    }


    /**
     * Returns sub classes of a given class.
     *
     * @param clazz The OWLClass the sub classes to be computed of
     * @param includeItself Whether to include the class itself in the returned set of sub classes (true) or not (false)
     * @param direct Whether to include only direct sub class of the class (true) or all sub classes (false)
     * @return Set<OWLClass> Set of sub classes
     */
    public static Set<OWLEntity> getSubClasses(OWLOntology ontology, OWLClass clazz, boolean direct, boolean includeItself) {

        if (ontology == null || clazz == null) {
            final String errMsg = "ontology == null || clazz == null";
            throw new NullPointerException(errMsg);
        }

        Set<OWLEntity> classes = new HashSet<OWLEntity>();

        // Add itself if requested
        if (includeItself)
            classes.add(clazz);

        // If direct only, use corresponding methods
        if (direct) {
            Set<OWLClassExpression> subClasses = clazz.getSubClasses( ontology );
            for( OWLClassExpression currentSubClass : subClasses )
                if( currentSubClass instanceof OWLClass )
                    classes.add( currentSubClass.asOWLClass() );

            // Otherwise, recursively iterate over all sub class axioms
        } else {

            // Recursively compute all sub classes
            getSubClassesHelper(ontology, ontology.getSubClassAxiomsForSuperClass(clazz), classes);

            // Add owl:Nothing if not contained
            //OWLClass nothing = ontology.getOWLOntologyManager().getOWLDataFactory().getOWLNothing();
            //if (!classes.contains(nothing))
            //	classes.add(nothing);

        }

        return classes;
    }


    /**
     * Helper to recursively determine sub classes.
     *
     * @param axioms Set of OWLSubClassOfAxiom representing the super/sub class relationship
     * @param clazzes Stores determined sub classes
     * @return Set<OWLClass> Set of sub classes
     */
    private static void getSubClassesHelper(OWLOntology ontology, Set<OWLSubClassOfAxiom> axioms, Set<OWLEntity> clazzes) {
        OWLClass clazz;
        Set<OWLSubClassOfAxiom> currentAxioms;
        for (OWLSubClassOfAxiom axiom : axioms) {
            if (!(axiom.getSubClass() instanceof OWLClass))
                continue;

            clazz = (OWLClass) axiom.getSubClass();

            // Add class to set of sub classes
            if (!clazzes.contains(clazz))
                clazzes.add(clazz);

            // Compute new set of sub class axioms
            currentAxioms = ontology.getSubClassAxiomsForSuperClass(clazz);

            // Continue if no more sub class axioms
            if (currentAxioms.size() == 0)
                continue;

            getSubClassesHelper(ontology, currentAxioms, clazzes);
        }
    }


    /**
     * Returns super object properties of a given object property.
     *
     * @param prop The OWLObjectProperty the super object properties to be computed of
     * @param includeItself Whether to include the object property itself in the returned set of super object properties (true) or not (false)
     * @param direct Whether to include only direct super object properties of the class (true) or all super object properties (false)
     * @return Set<OWLObjectProperty> Set of super object property
     */
    public static Set<OWLEntity> getSuperObjectProperties(OWLOntology ontology, OWLObjectProperty prop, boolean direct, boolean includeItself) {

        if (ontology == null || prop == null) {
            final String errMsg = "ontology == null || prop == null";
            throw new NullPointerException(errMsg);
        }

        Set<OWLEntity> props = new HashSet<OWLEntity>();

        // Add itself if requested
        if (includeItself)
            props.add(prop);

        // If direct only, use corresponding methods
        if (direct) {

            Set<?> superProperties = prop.getSuperProperties( ontology );
            for( Object currentSuperProperty : superProperties )
                if( currentSuperProperty instanceof OWLObjectProperty )
                    props.add( ((OWLObjectProperty) currentSuperProperty) );

            // Otherwise, recursively iterate over all sub object property axioms
        } else {

            getSuperObjectPropertiesHelper(ontology, ontology.getObjectSubPropertyAxiomsForSubProperty(prop), props);

            // Add topObjectProperty if not contained
            OWLObjectProperty topObjectProperty = ontology.getOWLOntologyManager().getOWLDataFactory().getOWLTopObjectProperty();
            if (!props.contains(topObjectProperty))
                props.add(topObjectProperty);
        }

        return props;
    }


    /**
     * Helper to recursively determine super object properties.
     *
     * @param axioms Set of OWLSubObjectPropertyOfAxiom representing the super/sub object property relationship
     * @return Set<OWLObjectProperty> Set of super object properties
     */
    private static void getSuperObjectPropertiesHelper(OWLOntology ontology, Set<OWLSubObjectPropertyOfAxiom> axioms, Set<OWLEntity> props){
        OWLObjectProperty property;
        Set<OWLSubObjectPropertyOfAxiom> currentAxioms;
        for (OWLSubObjectPropertyOfAxiom axiom : axioms) {
            if (!(axiom.getSuperProperty() instanceof OWLObjectProperty))
                continue;

            property = (OWLObjectProperty) axiom.getSuperProperty();

            // Add class to set of super properties
            if (!props.contains(property))
                props.add(property);

            // Compute new set of super property axioms
            currentAxioms = ontology.getObjectSubPropertyAxiomsForSubProperty(property);

            // Skip if no more super property axioms
            if (currentAxioms.size() == 0)
                continue;

            // Further iterate only of flag direct set to false
            getSuperObjectPropertiesHelper(ontology, currentAxioms, props);
        }
    }


    /**
     * Returns sub object properties of a given object property.
     *
     * @param prop The OWLObjectProperty the sub object properties to be computed of
     * @param includeItself Whether to include the object property itself in the returned set of sub object properties (true) or not (false)
     * @param direct Whether to include only direct sub object properties of the class (true) or all sub object properties (false)
     * @return Set<OWLObjectProperty> Set of sub object property
     */
    public static Set<OWLEntity> getSubObjectProperties(OWLOntology ontology, OWLObjectProperty prop, boolean direct, boolean includeItself) {

        if (ontology == null || prop == null) {
            final String errMsg = "ontology == null || prop == null";
            throw new NullPointerException(errMsg);
        }

        Set<OWLEntity> props = new HashSet<OWLEntity>();

        // Add itself if requested
        if (includeItself)
            props.add(prop);

        // If direct only, use corresponding methods
        if (direct) {

            Set<?> subProperties = prop.getSubProperties( ontology );
            for( Object currentSubProperty : subProperties )
                if( currentSubProperty instanceof OWLObjectProperty )
                    props.add( ((OWLObjectProperty) currentSubProperty) );

        } else {

            getSubObjectPropertiesHelper(ontology, ontology.getObjectSubPropertyAxiomsForSuperProperty(prop), props);

            // Add bottomObjectProperty if not contained
            OWLObjectProperty bottomObjectProperty = ontology.getOWLOntologyManager().getOWLDataFactory().getOWLBottomObjectProperty();
            if (!props.contains(bottomObjectProperty))
                props.add(bottomObjectProperty);
        }

        return props;
    }


    /**
     * Helper to recursively determine sub object properties.
     *
     * @param axioms Set of OWLSubObjectPropertyOfAxiom representing the super/sub object property relationship
     * @return Set<OWLObjectProperty> Set of sub object properties
     */
    private static void getSubObjectPropertiesHelper(OWLOntology ontology, Set<OWLSubObjectPropertyOfAxiom> axioms, Set<OWLEntity> props){
        OWLObjectProperty property;
        Set<OWLSubObjectPropertyOfAxiom> currentAxioms;
        for (OWLSubObjectPropertyOfAxiom axiom : axioms) {
            if (!(axiom.getSubProperty() instanceof OWLObjectProperty))
                continue;

            property = (OWLObjectProperty) axiom.getSubProperty();

            // Add class to set of sub properties
            if (!props.contains(property))
                props.add(property);

            // Compute new set of sub property axioms
            currentAxioms = ontology.getObjectSubPropertyAxiomsForSuperProperty(property);

            // Skip if no more sub property axioms
            if (currentAxioms.size() == 0)
                continue;

            // Further iterate only of flag direct set to false
            getSubObjectPropertiesHelper(ontology, currentAxioms, props);
        }
    }


    /**
     * Returns super data properties of a given data property.
     *
     * @param prop The OWLDataProperty the super data properties to be computed of
     * @param includeItself Whether to include the data property itself in the returned set of super data properties (true) or not (false)
     * @param direct Whether to include only direct super data properties of the class (true) or all super data properties (false)
     * @return Set<OWLDataProperty> Set of super data property
     */
    public static Set<OWLEntity> getSuperDataProperties(OWLOntology ontology, OWLDataProperty prop, boolean direct, boolean includeItself) {

        if (ontology == null || prop == null) {
            final String errMsg = "ontology == null || prop == null";
            throw new NullPointerException(errMsg);
        }

        Set<OWLEntity> props = new HashSet<OWLEntity>();

        // Add itself if requested
        if (includeItself)
            props.add(prop);

        // If direct only, use corresponding methods
        if (direct) {

            Set<?> superProperties = prop.getSuperProperties( ontology );
            for( Object currentSuperProperty : superProperties )
                if( currentSuperProperty instanceof OWLDataProperty )
                    props.add( ((OWLDataProperty) currentSuperProperty) );

        } else {

            getSuperDataPropertiesHelper(ontology, ontology.getDataSubPropertyAxiomsForSubProperty(prop), props);

            // Add topDataProperty if not contained
            OWLDataProperty topDataProperty = ontology.getOWLOntologyManager().getOWLDataFactory().getOWLTopDataProperty();
            if (!props.contains(topDataProperty))
                props.add(topDataProperty);
        }

        return props;
    }


    /**
     * Helper to recursively determine super data properties.
     *
     * @param axioms Set of OWLSubDataPropertyOfAxiom representing the super/sub data property relationship
     * @return Set<OWLDataProperty> Set of super data properties
     */
    private static void getSuperDataPropertiesHelper(OWLOntology ontology, Set<OWLSubDataPropertyOfAxiom> axioms, Set<OWLEntity> props){
        OWLDataProperty property;
        Set<OWLSubDataPropertyOfAxiom> currentAxioms;
        for (OWLSubDataPropertyOfAxiom axiom : axioms) {
            if (!(axiom.getSuperProperty() instanceof OWLDataProperty))
                continue;

            property = (OWLDataProperty) axiom.getSuperProperty();

            // Add class to set of super properties
            if (!props.contains(property))
                props.add(property);

            // Compute new set of super property axioms
            currentAxioms = ontology.getDataSubPropertyAxiomsForSubProperty(property);

            // Skip if no more super property axioms
            if (currentAxioms.size() == 0)
                continue;

            // Further iterate only of flag direct set to false
            getSuperDataPropertiesHelper(ontology, currentAxioms, props);
        }
    }


    /**
     * Returns sub data properties of a given data property.
     *
     * @param prop The OWLDataProperty the sub data properties to be computed of
     * @param includeItself Whether to include the data property itself in the returned set of sub data properties (true) or not (false)
     * @param direct Whether to include only direct sub data properties of the class (true) or all sub data properties (false)
     * @return Set<OWLDataProperty> Set of sub data property
     */
    public static Set<OWLEntity> getSubDataProperties(OWLOntology ontology, OWLDataProperty prop, boolean direct, boolean includeItself) {

        if (ontology == null || prop == null) {
            final String errMsg = "ontology == null || prop == null";
            throw new NullPointerException(errMsg);
        }

        Set<OWLEntity> props = new HashSet<OWLEntity>();

        // Add itself if requested
        if (includeItself)
            props.add(prop);

        // If direct only, use corresponding methods
        if (direct) {

            Set<?> subProperties = prop.getSubProperties( ontology );
            for( Object currentSubProperty : subProperties )
                if( currentSubProperty instanceof OWLDataProperty )
                    props.add( ((OWLDataProperty) currentSubProperty) );

        } else {

            getSubDataPropertiesHelper(ontology, ontology.getDataSubPropertyAxiomsForSuperProperty(prop), props);

            // Add bottomDataProperty if not contained
            OWLDataProperty bottomDataProperty = ontology.getOWLOntologyManager().getOWLDataFactory().getOWLBottomDataProperty();
            if (!props.contains(bottomDataProperty))
                props.add(bottomDataProperty);

        }

        return props;
    }


    /**
     * Helper to recursively determine sub data properties.
     *
     * @param axioms Set of OWLSubDataPropertyOfAxiom representing the super/sub data property relationship
     * @return Set<OWLDataProperty> Set of sub data properties
     */
    private static void getSubDataPropertiesHelper(OWLOntology ontology, Set<OWLSubDataPropertyOfAxiom> axioms, Set<OWLEntity> props){
        OWLDataProperty property;
        Set<OWLSubDataPropertyOfAxiom> currentAxioms;
        for (OWLSubDataPropertyOfAxiom axiom : axioms) {
            if (!(axiom.getSuperProperty() instanceof OWLDataProperty))
                continue;

            property = (OWLDataProperty) axiom.getSubProperty();

            // Add class to set of sub classes
            if (!props.contains(property))
                props.add(property);

            // Compute new set of sub class axioms
            currentAxioms = ontology.getDataSubPropertyAxiomsForSuperProperty(property);

            // Skip if no more sub class axioms
            if (currentAxioms.size() == 0)
                continue;

            // Further iterate only of flag direct set to false
            getSubDataPropertiesHelper(ontology, currentAxioms, props);
        }
    }

}
