package eu.intuitel.querybuilder.util;

import eu.intuitel.comm.entities.qb.QueryBuilderOutput;
import eu.intuitel.querybuilder.core.QueryBuilder;
import org.apache.commons.lang.StringUtils;
import org.dom4j.DocumentException;
import org.dom4j.DocumentHelper;
import org.dom4j.io.OutputFormat;
import org.dom4j.io.SAXReader;
import org.dom4j.io.XMLWriter;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.SchemaFactory;
import java.io.*;
import java.net.URISyntaxException;
import java.util.Properties;

import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.LogManager;

/**
 * QueryBuilderUtil.
 *
 * @author Matthias Stumpp (stumpp@fzi.de)
 */
public class QueryBuilderUtil {

    /*
     * Properties containing ontology specific configuration
     */
    private static Properties p;

    /*
     * logger
     */
    private static final Logger logger = LogManager.getLogger(QueryBuilderUtil.class.getName());


    /*
     * Validate a given QueryBuilderOutput
     */
    public static boolean isValid(QueryBuilderOutput output) throws QueryBuilderUtilException {

        InputStream in = QueryBuilder.class.getResourceAsStream("config.properties");
        p = new Properties();
        try {
            p.load(in);
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }

        // check if config properties are provided
        if (p == null) {
            String msg = "Unable to load config properties.";
            logger.error(msg);
            throw new QueryBuilderUtilException(msg);
        }

        File queryBuilderOutputXSDFile = null;
        try {
            queryBuilderOutputXSDFile = new File(QueryBuilder.class.getResource(p.getProperty("queryBuilderOutputXSD")).toURI());
            logger.debug("QueryBuilder XSD file found.");
        } catch (URISyntaxException e) {
            String msg = "No validation, because XSD does not exist at given resource: " + p.getProperty("queryBuilderOutputXSD");
            logger.error(msg);
            throw new QueryBuilderUtilException(msg);
        }

        if (queryBuilderOutputXSDFile == null) {
            logger.error("XML validation skipped.");
            throw new QueryBuilderUtilException("XML validation skipped.");
        }

        logger.debug("XML validation starting.");

        try {
            //File file = new File(System.getProperty("java.io.tmpdir") + "/querybuilder_output_schema_test.xml");
            //FileWriter fileWriter = new FileWriter(file);
            //XMLWriter writer = new XMLWriter(fileWriter);
            //writer.write(output);
            //writer.close();

            // serialise to xml
            File file = new File(System.getProperty("java.io.tmpdir") + "/querybuilder_output_schema_test.xml");
            FileWriter fileWriter = new FileWriter(file);
            XMLWriter writer = new XMLWriter(fileWriter);
            JAXBContext context = null;
            try {
                context = JAXBContext.newInstance(QueryBuilderOutput.class);
                Marshaller m = context.createMarshaller();
                m.marshal(output, writer);
            } catch (JAXBException e) {
                e.printStackTrace();
            }

            // output string to console
            // String theXML = writer.toString();
            writer.write(output);
            writer.close();

            SAXParserFactory parserFactory = SAXParserFactory.newInstance();
            SchemaFactory schemaFactory = SchemaFactory.newInstance("http://www.w3.org/2001/XMLSchema");
            parserFactory.setSchema(schemaFactory.newSchema(new Source[]{new StreamSource(queryBuilderOutputXSDFile)}));
            SAXParser parser = parserFactory.newSAXParser();
            SAXReader reader = new SAXReader(parser.getXMLReader());

            reader.setValidation(true);
            reader.setErrorHandler(new SimpleErrorHandler());
            reader.read(file);

        } catch (SAXException e) {
            throw new QueryBuilderUtilException("SAXException");
        } catch (ParserConfigurationException e) {
            throw new QueryBuilderUtilException("ParserConfigurationException");
        } catch (DocumentException e) {
            throw new QueryBuilderUtilException("DocumentException");
        } catch (IOException e) {
            throw new QueryBuilderUtilException("IOException");
        }
        logger.debug("XML validation finished.");

        return true;
    }

    public static String asString(Object output, Class clazz, boolean prettyPrint)
            throws QueryBuilderUtilException {
        // serialise to xml
        StringWriter writer = new StringWriter();
        JAXBContext context = null;
        try {
            context = JAXBContext.newInstance(clazz);
            Marshaller m = context.createMarshaller();
            m.marshal(output, writer);
        } catch (JAXBException e) {
            throw new QueryBuilderUtilException(e.getMessage());
        }

        // output string to console
        String theXML = writer.toString();
        if (prettyPrint) {
            return prettyPrint(theXML);
        } else {
            return theXML;
        }
    }

    public static QueryBuilderOutput fromFile(File file, Class clazz) {
        // deserialise to object
        FileReader reader = null;
        try {
            reader = new FileReader(file);
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
        JAXBContext context = null;
        QueryBuilderOutput output = null;
        try {
            context = JAXBContext.newInstance(clazz);
            Unmarshaller um = context.createUnmarshaller();
            output = (QueryBuilderOutput) um.unmarshal(reader);
        } catch (JAXBException e) {
            e.printStackTrace();
        }

        return output;
    }

    public static QueryBuilderOutput fromString(String string, Class clazz) {
        // deserialise to object
        StringReader reader = new StringReader(string);
        JAXBContext context = null;
        QueryBuilderOutput output = null;
        try {
            context = JAXBContext.newInstance(clazz);
            Unmarshaller um = context.createUnmarshaller();
            output =  (QueryBuilderOutput) um.unmarshal(reader);
        } catch (JAXBException e) {
            e.printStackTrace();
        }

        return output;
    }

    /**
     * SimpleErrorHandler.
     */
    public static class SimpleErrorHandler implements ErrorHandler {
        public void warning(SAXParseException e) throws SAXException {
            System.out.println(e.getMessage());
        }

        public void error(SAXParseException e) throws SAXException {
            System.out.println(e.getMessage());
        }

        public void fatalError(SAXParseException e) throws SAXException {
            System.out.println(e.getMessage());
        }
    }

    private static String prettyPrint(String xml) {

        if (StringUtils.isBlank(xml)) {
            throw new RuntimeException("xml was null or blank in prettyPrint()");
        }

        final StringWriter sw;

        try {
            final OutputFormat format = OutputFormat.createPrettyPrint();
            final org.dom4j.Document document = DocumentHelper.parseText(xml);
            sw = new StringWriter();
            final XMLWriter writer = new XMLWriter(sw, format);
            writer.write(document);
        }
        catch (Exception e) {
            throw new RuntimeException("Error pretty printing xml:\n" + xml, e);
        }
        return sw.toString();
    }

}
