/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Version: 1.00
 * Date: 2014/7/3
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.communication;

import edu.uk.reading.communication.connection.AbstractCommunicationLayer;
import edu.uk.reading.communication.connection.CommunicationEventListener;
import edu.uk.reading.communication.connection.CommunicationLayerDelegate;
import edu.uk.reading.communication.connection.mockup.MockupRRServiceConnection;
import edu.uk.reading.db.VariableFiller;
import edu.uk.reading.dm.common.LanguageCode;
import edu.uk.reading.dm.manager.ReplyMessageListener;
import edu.uk.reading.dm.manager.UserManager;
import edu.uk.reading.dm.messages.dialogue.ErrorMessage;
import edu.uk.reading.dm.messages.dialogue.ReplyMessage;
import edu.uk.reading.dm.messages.dialogue.UpdateMessage;
import edu.uk.reading.dm.model.ModelLoader;
import edu.uk.reading.dm.model.ParseException;
import edu.uk.reading.dm.nlg.NlgManager;
import edu.uk.reading.ranking.LoreRankingService;
import edu.uk.reading.ranking.RankingListener;
import edu.uk.reading.ranking.RankingResult;
import eu.intuitel.comm.entities.exceptions.IntuitelException;
import eu.intuitel.comm.entities.lpm.Reflex;
import eu.intuitel.comm.entities.qb.QueryBuilderOutput;
import eu.intuitel.comm.entities.qb.ReasoningResults;
import eu.intuitel.comm.services.IRecommendationRewriterListener;
import eu.intuitel.config.ConfigurationLoader;
import eu.intuitel.db.access.DbAccessException;
import eu.intuitel.db.access.DbAccessor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import static edu.uk.reading.tools.CommonUtils.files;
import static edu.uk.reading.transformer.CommunicationUtils.*;

public class RecommendedRewriteService implements IRecommendationRewriterListener, ReplyMessageListener,
                                                  RankingListener, CommunicationEventListener {
    private static final String FILENAME_PROPERTIES = "recommendedRewrite.properties";

    private static final String PROP_DIALOGUE_MODEL = "dialogueModel";
    private static final String PROP_NLG_LANGUAGE = "nlgLanguage";
    private static final String PROP_RANKING_MODEL = "rankingModel"; //"Ranking.properties"
    private static final String PROP_FILLER_MODEL = "fillerModel";
    private static final String PROP_WORKER_QUEUE_SIZE = "workerQueueSize";

    private static final String PROP_COMM_BASIC_URI = "basicUri";
    private static final String PROP_COMM_BASIC_GATEWAY = "basicGateway";

    private static final String PROP_COMM_USER = "username";
    private static final String PROP_COMM_PASS = "password";
    private static final String PROP_COMM_XMPP_SERVER = "xmppServer";
    private static final String PROP_COMM_XMPP_PORT = "xmppServerPort";
    private static final String PROP_COMM_TSB_USER = "tsbUsername";

    private static final String PROP_DB_TYPE = "dbType";
    private static final String PROP_DB_TYPE_HIBERNATE = "hibernate";

    private static final String PROP_CONNECTION_TYPE = "connectionType";
    private static final String PROP_CONNECTION_TYPE_COMM_LAYER = "commLayer";

    private static VariableFiller variableFiller;

    private UserManager userManager;
    private LoreRankingService rankingService;
    private AbstractCommunicationLayer connection;

    private static Logger logger = LoggerFactory.getLogger(RecommendedRewriteService.class);

    public RecommendedRewriteService(AbstractCommunicationLayer connection, int workerQueueSize,
                                     ModelLoader.ParseResult parseResult) {
        this.connection = connection;
        this.connection.setRrListener(this);
        this.connection.addCommunicationLayerEventListener(this);

        NlgManager.setupManager(parseResult.getNlgModel());

        userManager = new UserManager(parseResult.getDialogueModel(), workerQueueSize);
        userManager.addReplyListener(this);

        rankingService = new LoreRankingService();
        rankingService.addRankingListener(this);
    }

    @Override
    public void onNewReasoningResults(ReasoningResults results) {
        //todo; remove this
    }

    public void onNewLores(List<eu.intuitel.comm.entities.lms.Lore> lores) {
        // no reply to LOREs
    }

    public void onNewTugs(List<eu.intuitel.comm.entities.lms.Tug> tugs) {
        for (eu.intuitel.comm.entities.lms.Tug tug : tugs) {
            logger.info("Received TUG id=" + tug.getMId() + " user id=" + tug.getUId());
            connection.flagTugOnly(tug.getMId(), tug.getUId());
            userManager.pushMessage(convertTugToFeedback(tug, variableFiller));
        }
    }

    public void onNewQueryBuilderOutput(List<QueryBuilderOutput> queryBuilderOutputs) {
        for (QueryBuilderOutput output : queryBuilderOutputs) {
            logger.info("Received QBO id=" + output.getmId() + " user id=" + output.getuId());
            userManager.pushMessage(convertReasoningResultToFeedback(userManager, output, variableFiller));
            rankingService.handleReasoningResults(output);
        }
    }

    public void onNewReflexes(List<Reflex> reflexes) {
        for (Reflex reflex : reflexes) {
            logger.info("Received Reflex id=" + reflex.getmId() + " user id=" + reflex.getuId());
            connection.flagTugOnly(reflex.getmId(), reflex.getuId());
            userManager.pushMessage(convertReflexToRequest(userManager, reflex, variableFiller));
        }
    }

    public void replyReceived(ReplyMessage replyMessage) {
        try {
            logger.info("Received Reply Message id=" + replyMessage.getMessageID() +
                        " user id=" + replyMessage.getTargetUserID() + ". Sending TUG ...");
            connection.sendTug(convertReplyToTug(replyMessage, NlgManager.getInstance()));
        } catch (IntuitelException e) {
            logger.error("Error sending TUGs!", e);
        }
    }

    public void rankingResultReceived(RankingResult rankingResults) {
        try {
            logger.info("Received Ranking Message id=" + rankingResults.getMessageID() +
                        " user id=" + rankingResults.getUserID() + ". Sending LORE ...");
            connection.sendLore(convertRankingResultToLORE(rankingResults, variableFiller.dbAccessor));
        } catch (IntuitelException e) {
            logger.error("Error sending LOREs!", e);
        }
    }

    public void errorReceived(ErrorMessage errorMessage) {
        //currently the LPM --> RR protocol does not support error reporting
        logger.error(errorMessage.toString());
        connection.flagLoreOnly(errorMessage.getFeedbackMessageID(), errorMessage.getUserID());
    }

    public void updateReceived(UpdateMessage updateMessage) {
        try {
            logger.info("Received Update Message id=" + updateMessage.getMessageID() +
                        " user id=" + updateMessage.getTargetUserID() +
                        ". Updating database for params: " + updateMessage.getInputParams() +
                        ". Sending Tug Ack for the selected user ...");
            variableFiller.fillDbFromParams(updateMessage);
            connection.sendTugAck(createTugAck(updateMessage));
        } catch (IntuitelException e) {
            logger.error("Error sending Learner Update!", e);
        }
    }

    public void onClose(AbstractCommunicationLayer communicationLayer) {
        terminate();
    }

    public void onOpen(AbstractCommunicationLayer communicationLayer) {
        //do nothing when communication layer opens
    }

    public void terminate() {
        logger.info("RR service is closing ...");
        rankingService.removeRankingListener(this);
        rankingService.close();

        userManager.removeReplyListener(this);
        userManager.closeManager();

        connection.setRrListener(null);
        connection.Close();
        connection.removeCommunicationLayerEventListener(this);
        logger.info("RR service is closed ...");
    }

    public void waitForTermination() {
        userManager.waitToFinish();
    }

    public static Properties loadProperties() {
        Properties properties = new Properties();
        try {
            InputStream propertiesStream = new FileInputStream(FILENAME_PROPERTIES);
            properties.load(propertiesStream);
            propertiesStream.close();
        } catch (Throwable e) {
            logger.error("Exception while loading properties!", e);
            return null;
        }

        if (properties.getProperty(PROP_RANKING_MODEL) == null) {
            logger.error("The " + PROP_RANKING_MODEL + " property is not set.");
            return null;
        }

        File rankingModel = new File(properties.getProperty(PROP_RANKING_MODEL));
        if (!rankingModel.canRead()) {
            logger.error("Cannot read the provided ranking model: " + properties.getProperty(PROP_RANKING_MODEL));
            return null;
        }

        if (properties.getProperty(PROP_DIALOGUE_MODEL) == null) {
            logger.error("The " + PROP_DIALOGUE_MODEL + " property is not set.");
            return null;
        }

        return properties;
    }

    public static ModelLoader.ParseResult loadFeedbackModel(Properties properties) {
        ModelLoader.ParseResult parseResult;
        boolean hasErrors = false;
        try {
            LanguageCode code = LanguageCode.getCode(properties.getProperty(PROP_NLG_LANGUAGE,
                                                                            LanguageCode.DEFAULT.getLanguageCode()));
            parseResult = ModelLoader.load(code, files(properties.getProperty(PROP_DIALOGUE_MODEL, "").split(",")));
            for (String error : parseResult.getWarnings()) {
                logger.error("Warning: " + error);
            }

            for (String error : parseResult.getErrors()) {
                logger.error("Error: " + error);
                hasErrors = true;
            }

            if (hasErrors) {
                return null;
            }
        } catch (ParseException e) {
            logger.error(e.getMessage(), e);
            return null;
        }
        return parseResult;
    }

    private static boolean loadVariableFillerModel(Properties properties, UserManager userManager) {
        List<String> errors = new LinkedList<>();
        DbAccessor dbAccessor;
        try {
            if (PROP_DB_TYPE_HIBERNATE.equals(properties.getProperty(PROP_DB_TYPE))) {
                logger.info("Using the remote hibernate DB layer ...");
                dbAccessor = DbAccessor.createHibernateDbAccessor();
            } else {
                logger.info("Using the local test DB cache ...");
                dbAccessor = DbAccessor.createCachedDbAccessor();
            }
        } catch (DbAccessException e) {
            logger.error(e.getMessage(), e);
            return false;
        }
        if (dbAccessor != null && dbAccessor.isConnectionEstablished()) {
            variableFiller = new VariableFiller(userManager, dbAccessor);
            variableFiller.loadFillerModel(properties.getProperty(PROP_FILLER_MODEL), errors);
            boolean hasErrors = false;
            for (String error : errors) {
                logger.error("Error: " + error);
                hasErrors = true;
            }

            return !hasErrors;
        } else {
            return false;
        }
    }

    public static VariableFiller getVariableFiller() {
        return variableFiller;
    }

    public static boolean loadRankingModel(Properties properties) {
        File rankingModel = new File(properties.getProperty(PROP_RANKING_MODEL));
        try {
            return ConfigurationLoader.loadRankingModels(rankingModel);
        } catch (Exception e) {
            logger.error("Error loading the ranking model", e);
            return false;
        }
    }

    public static boolean startRR(AbstractCommunicationLayer connection,
                                  Properties properties, ModelLoader.ParseResult parseResult) {
        if (!ConfigurationLoader.isModelConfigured()) {
            logger.error("Apparently the ranking model is not loaded properly. Can't start the service!");
            return false;
        } else {
            logger.info("Feedback model loaded successfully !");
        }

        int queueSize = Integer.parseInt(properties.getProperty(PROP_WORKER_QUEUE_SIZE));
        final RecommendedRewriteService rr = new RecommendedRewriteService(connection, queueSize, parseResult);

        if (!loadVariableFillerModel(properties, rr.userManager)) {
            return false;
        } else {
            logger.info("Variable filler model loaded successfully !");
        }

        try {
            logger.info("Connecting to the communication library using...");

            if (properties.getProperty(PROP_COMM_BASIC_URI) != null) {
                connection.OpenBasic(URI.create(properties.getProperty(PROP_COMM_BASIC_URI)),
                                     URI.create(properties.getProperty(PROP_COMM_BASIC_GATEWAY)));
            } else if (properties.getProperty(PROP_COMM_USER) != null) {
                connection.OpenAdvanced(properties.getProperty(PROP_COMM_USER),
                                        properties.getProperty(PROP_COMM_PASS),
                                        properties.getProperty(PROP_COMM_XMPP_SERVER),
                                        Integer.parseInt(properties.getProperty(PROP_COMM_XMPP_PORT)),
                                        properties.getProperty(PROP_COMM_TSB_USER));
            } else {
                logger.error("No valid communication configuration found");
            }
        } catch (IOException e) {
            logger.error(e.getMessage(), e);
            return false;
        }

        logger.info("RR started successfully ...");

        if (connection.isContinuousResponse()) {
            Runtime.getRuntime().addShutdownHook(new Thread() {
                public void run() {
                    rr.terminate();
                }
            });
            logger.info("RR is waiting for input or interruption commands !");
            rr.waitForTermination();
        }
        return true;
    }

    public static void main(String[] args) throws Exception {
        Properties properties = loadProperties();
        if (properties == null) {
            System.exit(1);
        }

        if (!loadRankingModel(properties)) {
            System.exit(2);
        }

        ModelLoader.ParseResult parseResult = loadFeedbackModel(properties);
        if (parseResult == null) {
            System.exit(2);
        }

        AbstractCommunicationLayer communicationLayer;
        if (PROP_CONNECTION_TYPE_COMM_LAYER.equals(properties.getProperty(PROP_CONNECTION_TYPE))) {
            communicationLayer = CommunicationLayerDelegate.instance;
            logger.info("Using the remote communication library ...");
        } else {
            communicationLayer = MockupRRServiceConnection.instance;
            logger.info("Using the local mockup communication library ...");
        }

        if (!startRR(communicationLayer, properties, parseResult)) {
            System.exit(3);
        }
    }
}

