/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Version: 1.00
 * Date: 2014/7/25
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.communication.connection;

import eu.intuitel.CommLayer.CommLibrary.IntuitelService;
import eu.intuitel.CommLayer.CommLibrary.IntuitelServiceConnection;
import eu.intuitel.CommLayer.CommLibrary.services.rr.RecommendationRewriterService;
import eu.intuitel.comm.entities.exceptions.IntuitelException;
import eu.intuitel.comm.entities.lpm.Reflex;
import eu.intuitel.comm.entities.qb.QueryBuilderOutput;
import eu.intuitel.comm.entities.rr.Learner;
import eu.intuitel.comm.entities.rr.Lore;
import eu.intuitel.comm.entities.rr.Tug;
import eu.intuitel.comm.services.IRecommendationRewriterConnection;
import eu.intuitel.comm.services.IRecommendationRewriterListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.net.URI;
import java.util.*;

import static edu.uk.reading.transformer.CommunicationUtils.merge;

public abstract class AbstractCommunicationLayer extends IntuitelServiceConnection
        implements IRecommendationRewriterConnection, Runnable {
    private final List<CommunicationEventListener> listeners = new LinkedList<>();

    private final Set<String> tugOnlyMsg = new HashSet<>();
    private final Set<String> loreOnlyMsg = new HashSet<>();

    private final Set<String> ignoreMsgId = new HashSet<>();
    private final Map<String, TimeoutPair<Lore>> loreMap = new HashMap<>();
    private final Map<String, TimeoutPair<Tug>> tugMap = new HashMap<>();

    private final Logger logger = LoggerFactory.getLogger(AbstractCommunicationLayer.class);

    private Thread timeoutChecker = new Thread(this);

    protected volatile boolean opened = false;

    private static final long TIMEOUT = 5 * 1000;

    public void OpenBasic(URI uri, URI intuitelGateway) throws IOException {
        if (timeoutChecker != null) {
            timeoutChecker.start();
        }

        if (!opened) {
            OpenBasicDelegate(uri, intuitelGateway);
            opened = true;
            fireOpenEvent(this);
        }
    }

    protected abstract void OpenBasicDelegate(URI uri, URI intuitelGateway) throws IOException;

    public void OpenAdvanced(String username, String password, String xmppServer,
                             int xmppServerPort, String tsbUsername) throws IOException {
        if (!opened) {
            OpenAdvancedDelegate(username, password, xmppServer, xmppServerPort, tsbUsername);
            opened = true;

            if (timeoutChecker != null) {
                timeoutChecker.start();
            }
            fireOpenEvent(this);
        }
    }

    protected abstract void OpenAdvancedDelegate(String username, String password, String xmppServer,
                                                 int xmppServerPort, String tsbUsername) throws IOException;

    @Override
    public void Close() {
        if (timeoutChecker != null) {
            timeoutChecker.interrupt();
        }

        if (opened) {
            CloseDelegate();
            opened = false;
            fireCloseEvent(this);
        }
    }

    protected abstract void CloseDelegate();

    public void addCommunicationLayerEventListener(CommunicationEventListener listener) {
        synchronized (listeners) {
            listeners.add(listener);
        }
    }

    public void removeCommunicationLayerEventListener(CommunicationEventListener listener) {
        synchronized (listeners) {
            listeners.remove(listener);
        }
    }

    protected void fireOpenEvent(AbstractCommunicationLayer instance) {
        synchronized (listeners) {
            for (CommunicationEventListener listener : listeners) {
                listener.onOpen(instance);
            }
        }
    }

    protected void fireCloseEvent(AbstractCommunicationLayer instance) {
        synchronized (listeners) {
            for (CommunicationEventListener listener : listeners) {
                listener.onClose(instance);
            }
        }
    }

    @Override
    protected final java.util.logging.Logger getLogger() {
        return java.util.logging.Logger.getLogger(this.getClass().getName());
    }

    @Override
    protected final Class<? extends IntuitelService> getServiceClass() {
        return RecommendationRewriterService.class;
    }

    public abstract void setRrListener(IRecommendationRewriterListener listener);

    @Deprecated
    @Override
    public void sendLores(List<Lore> lores) throws IntuitelException {
        throw new IntuitelException("Invalid operation", new NoSuchMethodException());
    }

    @Deprecated
    @Override
    public void sendTugs(List<Tug> tugs) throws IntuitelException {
        throw new IntuitelException("Invalid operation", new NoSuchMethodException());
    }

    public void flagTugOnly(String mid, String uid) {
        tugOnlyMsg.add(key(mid, uid));
        logger.debug("Reflex Only flagged. Adding mid=" + mid + " uid=" + uid);
    }

    public void flagLoreOnly(String mid, String uid) {
        if (isLoreMessageAvailable(key(mid, uid))) {
            try {
                sendLearnerUpdate(merge(null, getLore(key(mid, uid))));
            } catch (IntuitelException e) {
                logger.error("Intuitel exception mid=" + mid + " uid=" + uid, e);
            }
            logger.debug("Lore Only flagged. Sending mid=" + mid + " uid=" + uid);
        } else {
            loreOnlyMsg.add(key(mid, uid));
            logger.debug("Lore Only flagged. Addign mid=" + mid + " uid=" + uid);
        }
    }

    @Override
    public void sendLore(Lore lore) throws IntuitelException {
        synchronized (ignoreMsgId) {
            if (ignoreMsgId.contains(key(lore.getMId(), lore.getUId()))) {
                logger.debug("Incoming LORE, TUG is already sent. Ignoring mid=" + lore.getMId() + " uid=" + lore.getUId());
                ignoreMsgId.remove(key(lore.getMId(), lore.getUId()));
            }
        }

        if (loreOnlyMsg.contains(key(lore.getMId(), lore.getUId()))) {
            logger.debug("This is a LORE Only reply. Sending mid=" + lore.getMId() + " uid=" + lore.getUId());
            sendLearnerUpdate(merge(null, lore));
        } else if (isTugMessageAvailable(key(lore.getMId(), lore.getUId()))) {
            logger.debug("Incoming LORE, TUG is available. Sending mid=" + lore.getMId() + " uid=" + lore.getUId());
            sendLearnerUpdate(merge(getTug(key(lore.getMId(), lore.getUId())), lore));
        } else {
            logger.debug("Incoming LORE, TUG is NOT available. Queuing mid=" + lore.getMId() + " uid=" + lore.getUId());
            queueMessage(lore);
        }
    }

    @Override
    public void sendTug(Tug tug) throws IntuitelException {
        synchronized (ignoreMsgId) {
            if (ignoreMsgId.contains(key(tug.getMId(), tug.getUId()))) {
                logger.debug("Incoming TUG, LORE is already sent. Ignoring mid=" + tug.getMId() + " uid=" + tug.getUId());
                ignoreMsgId.remove(key(tug.getMId(), tug.getUId()));
            }
        }

        if (tugOnlyMsg.contains(key(tug.getMId(), tug.getUId()))) {
            logger.debug("This is a TUG Only reply (Reflex Msg). Sending mid=" + tug.getMId() + " uid=" + tug.getUId());
            sendLearnerUpdate(merge(tug, null));
        } else if (isLoreMessageAvailable(key(tug.getMId(), tug.getUId()))) {
            logger.debug("Incoming TUG, LORE is available. Sending mid=" + tug.getMId() + " uid=" + tug.getUId());
            sendLearnerUpdate(merge(tug, getLore(key(tug.getMId(), tug.getUId()))));
        } else {
            logger.debug("Incoming TUG, LORE is NOT available. Queuing mid=" + tug.getMId() + " uid=" + tug.getUId());
            queueMessage(tug);
        }
    }

    private String key(String mid, String uid) {
        return mid + "~" + uid;
    }

    private boolean isTugMessageAvailable(String key) {
        synchronized (tugMap) {
            return tugMap.containsKey(key);
        }
    }

    private Tug getTug(String key) {
        synchronized (tugMap) {
            return tugMap.get(key).getItem();
        }
    }

    private void queueMessage(Tug tug) {
        synchronized (tugMap) {
            tugMap.put(key(tug.getMId(), tug.getUId()), new TimeoutPair<>(tug));
        }
    }

    private boolean isLoreMessageAvailable(String key) {
        synchronized (loreMap) {
            return loreMap.containsKey(key);
        }
    }

    private Lore getLore(String key) {
        synchronized (loreMap) {
            return loreMap.get(key).getItem();
        }
    }

    private void queueMessage(Lore lore) {
        synchronized (loreMap) {
            loreMap.put(key(lore.getMId(), lore.getUId()), new TimeoutPair<>(lore));
        }
    }

    @Override
    public void run() {
        while (opened) {
            synchronized (tugMap) {
                for (Map.Entry<String, TimeoutPair<Tug>> tugs : tugMap.entrySet()) {
                    if (System.currentTimeMillis() - tugs.getValue().getTimestamp() > TIMEOUT) {
                        Tug tug = tugs.getValue().getItem();
                        logger.debug("TUG Timeout reached. Sending TUG only message. Sending mid=" + tug.getMId() +
                                     " uid=" + tug.getUId());

                        try {
                            sendLearnerUpdate(merge(tug, null));
                            String key = key(tug.getMId(), tug.getUId());
                            synchronized (ignoreMsgId) {
                                ignoreMsgId.add(key);
                            }
                            tugMap.remove(key);
                        } catch (IntuitelException e) {
                            logger.error("Intuitel exception mid=" + tug.getMId() + " uid=" + tug.getUId(), e);
                        }
                    }
                }
            }

            synchronized (loreMap) {
                for (Map.Entry<String, TimeoutPair<Lore>> lores : loreMap.entrySet()) {
                    if (System.currentTimeMillis() - lores.getValue().getTimestamp() > TIMEOUT) {
                        Lore lore = lores.getValue().getItem();
                        logger.debug("LORE Timeout reached. Sending LORE only message. Sending mid=" + lore.getMId() +
                                     " uid=" + lore.getUId());

                        try {
                            sendLearnerUpdate(merge(null, lore));
                        } catch (IntuitelException e) {
                            logger.error("Intuitel exception mid=" + lore.getMId() + " uid=" + lore.getUId(), e);
                        }
                    }
                }
            }

            try {
                Thread.sleep(TIMEOUT);
            } catch (InterruptedException e) {
                //-- ignore
            }
        }
    }

    @Override
    public abstract void sendLearnerUpdate(List<Learner> learners) throws IntuitelException;

    @Override
    public abstract void sendLearnerUpdate(Learner learner) throws IntuitelException;

    @Override
    public abstract void onNewReflexes(List<Reflex> reflexes);

    public abstract void onNewLores(List<eu.intuitel.comm.entities.lms.Lore> lores);

    public abstract void onNewTugs(List<eu.intuitel.comm.entities.lms.Tug> tugs);

    public abstract void onNewQueryBuilderOutput(List<QueryBuilderOutput> queryBuilderOutputs);

    public abstract boolean isContinuousResponse();
}
