/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Version: 1.00
 * Date: 2014/7/25
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.db;

import edu.uk.reading.db.fillers.DbFiller;
import edu.uk.reading.db.fillers.DbFillerModel;
import edu.uk.reading.dm.common.LanguageCode;
import edu.uk.reading.dm.common.param.Param;
import edu.uk.reading.dm.common.param.ParamType;
import edu.uk.reading.dm.common.param.ParamValue;
import edu.uk.reading.dm.common.util.StringUtils;
import edu.uk.reading.dm.manager.UserManager;
import edu.uk.reading.dm.messages.dialogue.FeedbackMessage;
import edu.uk.reading.dm.messages.dialogue.UpdateMessage;
import edu.uk.reading.user.UserProperties;
import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.types.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.bind.*;
import javax.xml.bind.util.ValidationEventCollector;
import java.io.FileReader;
import java.io.Reader;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import static edu.uk.reading.dm.common.param.Param.joinParams;

public class VariableFiller {
    public static final String DB_PROP_CURRENT_CC = "currentCC";
    public static final String DB_PROP_CURRENT_KO = "currentKO";
    public static final String DB_PROP_LAST_CC = "lastCC";

    public static final String DB_PROP_CURRENT_DIFFICULTY_LVL = "difficultyLevel";
    public static final String DB_PROP_CURRENT_EQF_LVL = "eqfLevel";

    public static final String DB_PROP_CURRENT_DEAF_STATE = "deaf";
    public static final String DB_PROP_CURRENT_MUTE_STATE = "mute";

    public static final String DB_PROP_DELETE_LP = "deleteLP";
    public static final String DB_PROP_COMPLETE_ALL_KO = "completeAllKOs";
    public static final String DB_PROP_RESTART_CURRENT_CC = "restartCurrentCC";

    public static final String DB_PROP_INCREASE_DIFFICULTY = "increaseDifficultyLevel";
    public static final String DB_PROP_DECREASE_DIFFICULTY = "decreaseDifficultyLevel";

    public static final String DB_PROP_CURRENT_MICRO_LP = "microLP";
    public static final String DB_PROP_CURRENT_MACRO_LP = "macroLP";
    public static final String DB_PROP_CURRENT_AGE = "age";
    public static final String DB_PROP_CURRENT_GENDER = "gender";
    public static final String DB_PROP_CURRENT_CULTURE = "culture";
    public static final String DB_PROP_CURRENT_LANGUAGE = "language";

    private static Logger logger = LoggerFactory.getLogger(VariableFiller.class);

    private UserManager userManager;
    private DbFillerModel fillerModel;

    public final DbAccessor dbAccessor;

    public VariableFiller(UserManager userManager, DbAccessor dbAccessor) {
        this.userManager = userManager;
        this.dbAccessor = dbAccessor;
    }

    public void loadFillerModel(String filename, List<String> errors) {
        try {
            FileReader reader = new FileReader(filename);
            fillerModel = parseModel(reader, DbFillerModel.class, errors, VariableFiller.class.getName());
            reader.close();
        } catch (Throwable e) {
            errors.add("[" + VariableFiller.class.getName() + "] Exception type: " + e.getClass().getName() +
                       " message: " + e.getMessage());
        }
    }

    public void fillParamsFromDb(FeedbackMessage message) {
        if (message != null && ((message.getOutputParams() != null && !message.getOutputParams().isEmpty())
                                || (message.getInputParams() != null && !message.getInputParams().isEmpty()))) {
            if (fillerModel != null && fillerModel.getDbGetters() != null && !fillerModel.getDbGetters().isEmpty()) {
                Set<Param> allParams = joinParams(message.getOutputParams(), message.getInputParams());
                for (DbFiller filler : fillerModel.getDbGetters()) {
                    Param result = null;
                    if (areOutputParamsTriggered(filler.getTriggerParams(), allParams)) {
                        String learnerId = message.getTargetUserID();
                        String courseId = userManager.getUserProperty(learnerId, UserProperties.PROP_COURSE_ID);
                        switch (filler.getVariable()) {
                            case DB_PROP_CURRENT_CC:
                            case DB_PROP_LAST_CC:
                                result = wrapCC(filler.getVariable(), dbAccessor.getLastConceptContainerIri(learnerId));
                                break;
                            case DB_PROP_CURRENT_KO:
                                String koIri = dbAccessor.getLastKnowledgeObjectIri(learnerId, courseId);
                                result = wrapKO(filler.getVariable(), koIri);
                                break;
                            case DB_PROP_CURRENT_DIFFICULTY_LVL:
                                int level = dbAccessor.getPreferredDifficultyLevel(learnerId, courseId).getLevel();
                                result = wrapParam(filler.getVariable(), ParamType.number, level);
                                break;
                            default:
                                logger.info("Error: Unable to fill " + filler.getVariable() +
                                            " param, for the message: " + message);
                        }
                    }
                    if (result != null) {
                        if (message.getOutputParams() == null) {
                            message.setOutputParams(new HashSet<Param>());
                        }
                        message.getOutputParams().add(result);
                    }
                }
            }
        }
    }

    public void fillDbFromParams(UpdateMessage message) {
        if (fillerModel != null && fillerModel.getDbSetters() != null && !fillerModel.getDbSetters().isEmpty()
            && message.getInputParams() != null && !message.getInputParams().isEmpty()) {
            for (DbFiller filler : fillerModel.getDbSetters()) {
                for (Param triggerParam : filler.getTriggerParams()) {
                    for (Param inputParam : message.getInputParams()) {
                        if (inputParam.equals(triggerParam)) {
                            if (triggerParam.getValue() == null
                                || Param.equalValues(triggerParam.getValue(), inputParam.getValue())) {
                                setDbParam(message.getTargetUserID(), filler.getVariable(), inputParam);
                            }
                        }
                    }
                }
            }
        }
    }

    private boolean areOutputParamsTriggered(List<Param> triggerParams, Set<Param> outputParams) {
        for (Param triggerParam : triggerParams) {
            for (Param outputParam : outputParams) {
                if (triggerParam.equals(outputParam)
                    && Param.equalValues(triggerParam.getValue(), outputParam.getValue())) {
                    return true;
                }
            }
        }
        return false;
    }

    private void setDbParam(String learnerID, String dbVariableName, Param param) {
        String courseId = userManager.getUserProperty(learnerID, UserProperties.PROP_COURSE_ID);
        switch (dbVariableName) {
            // complex db actions
            case DB_PROP_DELETE_LP:
                dbAccessor.restartMacroLpSelection(learnerID, courseId);
                break;
            case DB_PROP_COMPLETE_ALL_KO:
                dbAccessor.markAllKoComplete(learnerID, courseId);
                break;
            case DB_PROP_RESTART_CURRENT_CC:
                dbAccessor.resetCcScore(learnerID, courseId, dbAccessor.getLastConceptContainerIri(learnerID));
                break;
            case DB_PROP_INCREASE_DIFFICULTY:
                dbAccessor.increasePreferredDifficulty(learnerID, courseId);
                break;
            case DB_PROP_DECREASE_DIFFICULTY:
                dbAccessor.decreasePreferredDifficulty(learnerID, courseId);
                break;
            // simple setters
            // LP + CC setters
            case DB_PROP_CURRENT_MACRO_LP:
                dbAccessor.storeMacroLpSelection(learnerID, getValue(param), courseId);
                break;
            case DB_PROP_CURRENT_MICRO_LP:
                dbAccessor.storeMicroLpSelection(learnerID, getValue(param), courseId);
                break;
            case DB_PROP_CURRENT_CC:
                dbAccessor.storeLearningStep(learnerID, courseId, null, getValue(param), System.currentTimeMillis());
                break;
            // Learner profile
            case DB_PROP_CURRENT_AGE:
                dbAccessor.setLearnerAge(learnerID, getIntValue(param));
                break;
            case DB_PROP_CURRENT_GENDER:
                dbAccessor.setLearnerGender(learnerID, Gender.getGender(getValue(param)));
                break;
            case DB_PROP_CURRENT_CULTURE:
                dbAccessor.setLearnerCulture(learnerID, Culture.getCulture(getValue(param)));
                break;
            case DB_PROP_CURRENT_LANGUAGE:
                dbAccessor.setLearnerLanguages(learnerID, Language.getLanguage(getValue(param)));
                break;
            // difficulty
            case DB_PROP_CURRENT_DIFFICULTY_LVL:
                dbAccessor.setPreferredDifficultyLevel(learnerID, courseId,
                                                       DifficultyLevel.getLevel(getIntValue(param)));
                break;
            case DB_PROP_CURRENT_EQF_LVL:
                dbAccessor.setPreferredEqfLevel(learnerID, courseId, EqfLevel.getLevel(getIntValue(param)));
                break;
            // accessibility
            case DB_PROP_CURRENT_MUTE_STATE:
                dbAccessor.setLearnerMuteState(learnerID, Boolean.valueOf(getValue(param)));
                break;
            case DB_PROP_CURRENT_DEAF_STATE:
                dbAccessor.setLearnerDeafState(learnerID, Boolean.valueOf(getValue(param)));
                break;
        }
    }

    private String getValue(Param param) {
        Object value = StringUtils.getParamValue(param, LanguageCode.DEFAULT);
        if (value == null) {
            return null;
        } else {
            return value.toString();
        }
    }

    private Integer getIntValue(Param param) {
        Object value = StringUtils.getParamValue(param, LanguageCode.DEFAULT);
        if (value == null) {
            return null;
        } else {
            return Integer.parseInt(value.toString());
        }
    }

    private Param wrapKO(String paramName, String value) {
        return wrapParam(paramName, ParamType.string, value, dbAccessor.getKoTitle(value));
    }

    private Param wrapCC(String paramName, String value) {
        return wrapParam(paramName, ParamType.string, value, dbAccessor.getCcTitle(value));
    }

    private Param wrapParam(String paramName, ParamType paramType, String value, String title) {
        if (value != null) {
            if (title != null) {
                Param param = new Param(paramName, paramType);
                param.setValue(new ParamValue(value, title));
                return param;
            } else {
                return new Param(paramName, paramType, value);
            }
        } else {
            return null;
        }
    }

    private Param wrapParam(String paramName, ParamType paramType, Integer value) {
        if (value != null) {
            return new Param(paramName, paramType, value.toString());
        } else {
            return null;
        }
    }

    @SuppressWarnings("unchecked")
    private static <T> T parseModel(Reader sourceCode, Class<T> modelClass, List<String> errors, String source) {
        try {
            JAXBContext context = JAXBContext.newInstance(modelClass);
            Unmarshaller unmarshaller = context.createUnmarshaller();
            ValidationEventCollector vec = new ValidationEventCollector();
            unmarshaller.setEventHandler(vec);
            T model = (T) unmarshaller.unmarshal(sourceCode);
            if (vec.hasEvents()) {
                for (ValidationEvent ve : vec.getEvents()) {
                    String msg = ve.getMessage();
                    ValidationEventLocator vel = ve.getLocator();
                    int line = vel.getLineNumber();
                    int column = vel.getColumnNumber();
                    errors.add("[" + source + "](line = " + line + ", column = " + column + "): " + msg);
                }
            }
            return model;
        } catch (JAXBException e) {
            errors.add("[" + source + "] Exception: " + e.getMessage());
            return null;
        }
    }
}
