/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Șerban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Contributor: Andrea Zielinski, IOSB Fraunhofer, Germany, http://www.iosb.fraunhofer.de/
 * Version: 1.00
 * Date: 2014/7/3
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.ranking;

import com.google.common.collect.Maps;
import com.google.common.collect.Ordering;

import eu.intuitel.comm.entities.qb.QueryBuilderOutput;
import eu.intuitel.config.ConfigurationLoader;
import eu.intuitel.util.*;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

public class LoreRankingService extends Thread {
	public static final String PROPERTY_RECOMMENDATION_STRATEGY = "RecommendationStrategy";
	public static final String PROPERTY_RECOMMENDATION_AXIOM = "Axiom";

	public static final String VALUE_CRITERIA_SOFT = "SoftCriteria";
	public static final String VALUE_CRITERIA_HARD = "HardCriteria";
	public static final String VALUE_CRITERIA_Mixed = "MixedCriteria";

	public static final String VALUE_AXIOM_1 = "Recommendation Axiom 1_1";
	public static final String VALUE_AXIOM_2 = "Recommendation Axiom 1_2";
	public static final String VALUE_AXIOM_3 = "Recommendation Axiom 1_3";

	private static Logger logger = LoggerFactory
			.getLogger(LoreRankingService.class);

	private final List<RankingListener> listeners = new LinkedList<>();

	private volatile boolean running = true;

	public LoreRankingService() {
		super("Ranking Service Thread");

		start();
	}

	public void addRankingListener(RankingListener listener) {
		synchronized (listeners) {
			listeners.add(listener);
		}
	}

	public void removeRankingListener(RankingListener listener) {
		synchronized (listeners) {
			listeners.remove(listener);
		}
	}

	private void notifyRankingResult(RankingResult rankingResult) {
		synchronized (listeners) {
			for (RankingListener listener : listeners) {
				listener.rankingResultReceived(rankingResult);
			}
		}
	}

	private final Queue<QueryBuilderOutput> processingQueue = new LinkedList<>();

	public void handleReasoningResults(QueryBuilderOutput reasoningResults) {
		synchronized (processingQueue) {
			processingQueue.add(reasoningResults);
			processingQueue.notifyAll();
		}
	}

	private Map<String, Integer> processReasoningResult(
			QueryBuilderOutput reasoningResults) {
		Map<String, Integer> result = new HashMap<>();

		List<String> DirectSuccs = ParseQBResult
				.getDirectSuccessorslmsID(reasoningResults);
		List<String> AllSuccs = ParseQBResult
				.getAllSuccessorslmsID(reasoningResults);
		List<String> PrevSuccs = ParseQBResult
				.getSuccessorsOfPreviouslmsID(reasoningResults);

		String propRecommendationStrategy = ConfigurationLoader.properties
				.getProperty(PROPERTY_RECOMMENDATION_STRATEGY, "").trim();
		String propRecommendationAxiom = ConfigurationLoader.properties
				.getProperty(PROPERTY_RECOMMENDATION_AXIOM, "").trim();

		logger.info("Using {} for the Recommendation Strategy",
				propRecommendationStrategy);
		logger.info("Using {} for the Axiom", propRecommendationAxiom);

		if (VALUE_CRITERIA_HARD.equalsIgnoreCase(propRecommendationStrategy)) {
			HashMap<String, Double> dictI = null;
			if (VALUE_AXIOM_1.equalsIgnoreCase(propRecommendationAxiom)) {
				dictI = HardRanking.getHardResults(reasoningResults,
						DirectSuccs);
			} else if (VALUE_AXIOM_2.equalsIgnoreCase(propRecommendationAxiom)) {
				dictI = HardRanking.getHardResults(reasoningResults, AllSuccs);
			} else if (VALUE_AXIOM_3.equalsIgnoreCase(propRecommendationAxiom)) {
				dictI = HardRanking.getHardResults(reasoningResults, PrevSuccs);
			}

			if (dictI != null) {
				for (String key : dictI.keySet()) {
					result.put(key, 100);
				}
			}
		} else if (VALUE_CRITERIA_SOFT
				.equalsIgnoreCase(propRecommendationStrategy)) {
			HashMap<String, Integer> dict = null;
			if (VALUE_AXIOM_1.equalsIgnoreCase(propRecommendationAxiom)) {
				dict = SoftRanking.getSoftResults(reasoningResults, DirectSuccs);
			} else if (VALUE_AXIOM_2.equalsIgnoreCase(propRecommendationAxiom)) {
				dict = SoftRanking.getSoftResults(reasoningResults, AllSuccs);
			} else if (VALUE_AXIOM_3.equalsIgnoreCase(propRecommendationAxiom)) {
				dict = SoftRanking.getSoftResults(reasoningResults, PrevSuccs);
			}

			if (dict != null) {
				 result = SoftRanking.sortByComparator(dict);
			}  
		}
		
		 else if (VALUE_CRITERIA_Mixed
					.equalsIgnoreCase(propRecommendationStrategy)) {
				HashMap<String, Integer> dict = null;
				if (VALUE_AXIOM_1.equalsIgnoreCase(propRecommendationAxiom)) {
					dict = MixedRanking.getMixedResults(reasoningResults, DirectSuccs);
						
				} else if (VALUE_AXIOM_2.equalsIgnoreCase(propRecommendationAxiom)) {
					dict = MixedRanking.getMixedResults(reasoningResults, AllSuccs);
				} else if (VALUE_AXIOM_3.equalsIgnoreCase(propRecommendationAxiom)) {
					dict = MixedRanking.getMixedResults(reasoningResults, PrevSuccs);
				}

				if (dict != null) {
					 result = SoftRanking.sortByComparator(dict);
				}  
			}
			
		

		logger.info("Returning the ranking result: " + result);

		return result;
	}

	public void run() {
		while (running) {
			QueryBuilderOutput message;
			synchronized (processingQueue) {
				message = processingQueue.poll();
			}
			if (message != null) {
				try {
					Map<String, Integer> result = processReasoningResult(message);
					RankingResult rankingResult = new RankingResult(
							message.getmId(), message.getuId());
					rankingResult.setRankingResults(result);
					notifyRankingResult(rankingResult);
				} catch (Throwable t) {
					// super-safe catch
					logger.error("Failed to process the message: " + message, t);
				}
			} else {
				synchronized (processingQueue) {
					try {
						processingQueue.wait();
					} catch (InterruptedException e) {
						// --
						running = false;
					}
				}
			}
		}
	}

	public void close() {
		running = false;
		interrupt();
	}
}
