/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Version: 1.00
 * Date: 2014/7/3
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.transformer;

import edu.uk.reading.db.VariableFiller;
import edu.uk.reading.dm.common.RequestType;
import edu.uk.reading.dm.common.param.Param;
import edu.uk.reading.dm.common.param.ParamValue;
import edu.uk.reading.dm.manager.UserManager;
import edu.uk.reading.dm.messages.dialogue.FeedbackMessage;
import edu.uk.reading.dm.messages.dialogue.ReplyMessage;
import edu.uk.reading.dm.messages.dialogue.UpdateMessage;
import edu.uk.reading.dm.nlg.NlgDecoratedResult;
import edu.uk.reading.dm.nlg.NlgManager;
import edu.uk.reading.ranking.RankingResult;
import edu.uk.reading.user.UserProperties;
import eu.intuitel.comm.entities.lms.Data;
import eu.intuitel.comm.entities.lms.RetVal;
import eu.intuitel.comm.entities.lpm.Reflex;
import eu.intuitel.comm.entities.lpm.ReflexOption;
import eu.intuitel.comm.entities.qb.DidacticFactor;
import eu.intuitel.comm.entities.qb.QueryBuilderOutput;
import eu.intuitel.comm.entities.rr.Learner;
import eu.intuitel.comm.entities.rr.Lore;
import eu.intuitel.comm.entities.rr.Tug;
import eu.intuitel.comm.entities.rr.TugAck;
import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.db.model.KnowledgeObject;

import java.util.*;

import static edu.uk.reading.dm.common.StringUtils.encodeHtml;
import static edu.uk.reading.tools.CommonUtils.tset;

public class CommunicationUtils {
    public static final int TUG_MSG_TYPE_SIMPLE_TEXT = 1;
    public static final int TUG_MSG_TYPE_REQUEST_TEXT = 4;

    public static eu.intuitel.comm.entities.rr.Tug convertReplyToTug(ReplyMessage replyMessage, NlgManager nlgManager) {
        eu.intuitel.comm.entities.rr.Tug result = new eu.intuitel.comm.entities.rr.Tug();
        result.setMId(replyMessage.getFeedbackMessageID());
        result.setRId(replyMessage.getMessageID());
        result.setUId(replyMessage.getTargetUserID());

        NlgDecoratedResult nlgDecoratedResult = nlgManager.decorateMessage(replyMessage);
        result.setMData(htmlMsgFormat(nlgDecoratedResult.getPattern(), nlgDecoratedResult.getAnswers()));

        if (replyMessage.getRequestType() == RequestType.inform) {
            result.setMType(TUG_MSG_TYPE_SIMPLE_TEXT);
        } else if (replyMessage.getRequestType() == RequestType.request) {
            result.setMType(TUG_MSG_TYPE_REQUEST_TEXT);
        }
        return result;
    }

    private static String htmlMsgFormat(String pattern, List<String> answers) {
        StringBuilder sb = new StringBuilder();
        sb.append("<![CDATA[");
        sb.append("<h4>").append(encodeHtml(pattern)).append("</h4>");
        for (String answer : answers) {
            sb.append("\n").append(answer);
        }
        sb.append("]]>");
        return sb.toString();
    }

    public static FeedbackMessage convertTugToFeedback(eu.intuitel.comm.entities.lms.Tug tug,
                                                       VariableFiller filler) {
        FeedbackMessage feedbackMessage = new FeedbackMessage();

        feedbackMessage.setMessageID(tug.getMId());
        feedbackMessage.setTargetUserID(tug.getUId());

        feedbackMessage.setTimestamp(System.currentTimeMillis());

        if (tug.getRetVal() == RetVal.OK) {
            feedbackMessage.setRequestType(RequestType.reply);
        } else if (tug.getRetVal() == RetVal.PAUSE) {
            feedbackMessage.setRequestType(RequestType.cancel);
        }

        if (tug.getData() != null) {
            Set<Param> inputParams = new HashSet<>();
            for (Data data : tug.getData()) {
                inputParams.add(new Param(data.getName(), data.getValue()));
            }
            feedbackMessage.setInputParams(inputParams);
        }

        filler.fillParamsFromDb(feedbackMessage);
        return feedbackMessage;
    }

    public static FeedbackMessage convertReasoningResultToFeedback(UserManager userManager,
                                                                   QueryBuilderOutput reasoningResults,
                                                                   VariableFiller filler) {
        FeedbackMessage feedbackMessage = new FeedbackMessage();

        feedbackMessage.setMessageID(reasoningResults.getmId());
        feedbackMessage.setTargetUserID(reasoningResults.getuId());

        feedbackMessage.setTimestamp(System.currentTimeMillis());
        feedbackMessage.setRequestType(RequestType.inform);
        userManager.setUserProperty(reasoningResults.getuId(),
                                    UserProperties.PROP_COURSE_ID,
                                    reasoningResults.getCourseId());

        if (reasoningResults.getDidacticFactors() != null
            && reasoningResults.getDidacticFactors().getDidacticFactor() != null) {

            Set<Param> inputParams = new HashSet<>();
            for (DidacticFactor df : reasoningResults.getDidacticFactors().getDidacticFactor()) {
                inputParams.add(new Param(trimUriToValue(df.getName()),
                                          trimUriToValue(df.getValue())));
            }
            feedbackMessage.setOutputParams(inputParams);
        }

        filler.fillParamsFromDb(feedbackMessage);
        return feedbackMessage;
    }

    private static String trimUriToValue(String uri) {
        if (uri == null || uri.trim().length() == 0 || !uri.contains("#")) {
            return uri;
        } else {
            return uri.substring(uri.indexOf('#') + 1);
        }
    }

    public static FeedbackMessage convertReflexToRequest(UserManager userManager, Reflex reflex,
                                                         VariableFiller filler) {
        FeedbackMessage feedbackMessage = new FeedbackMessage();

        feedbackMessage.setMessageID(reflex.getmId());
        feedbackMessage.setTargetUserID(reflex.getuId());

        feedbackMessage.setTimestamp(System.currentTimeMillis());
        feedbackMessage.setRequestType(RequestType.request);
        userManager.setUserProperty(reflex.getuId(), UserProperties.PROP_COURSE_ID, reflex.getCourseId());

        if (reflex.getReflexReason() != null) {
            feedbackMessage.setInputParams(tset(new Param(reflex.getReflexReason())));
        }

        if (reflex.getOptions() != null) {
            Set<Param> outputParams = new HashSet<>();
            for (Map.Entry<String, List<Map.Entry<String, String>>> data : extract(reflex.getOptions()).entrySet()) {
                if (data.getValue() != null) {
                    Param param = new Param(data.getKey());
                    for (Map.Entry<String, String> value : data.getValue()) {
                        param.setValue(new ParamValue(value.getKey(), value.getValue()));
                    }
                    outputParams.add(param);
                }
            }
            feedbackMessage.setOutputParams(outputParams);
        }

        filler.fillParamsFromDb(feedbackMessage);
        return feedbackMessage;
    }

    private static Map<String, List<Map.Entry<String, String>>> extract(List<ReflexOption> options) {
        Map<String, List<Map.Entry<String, String>>> result = new HashMap<>();

        for (ReflexOption option : options) {
            List<Map.Entry<String, String>> values = result.get(option.getName());
            if (values == null) {
                values = new LinkedList<>();
                result.put(option.getName(), values);
            }
            values.add(new AbstractMap.SimpleEntry<>(option.getValue(), option.getTitle()));
        }

        return result;
    }

    public static Lore convertRankingResultToLORE(RankingResult rankingResult, DbAccessor dbAccessor) {
        Lore result = new Lore();
        result.setMId(rankingResult.getMessageID());
        result.setUId(rankingResult.getUserID());

        for (Map.Entry<String, Integer> item : rankingResult.getRankingResults().entrySet()) {
            Lore.LorePrio element = new Lore.LorePrio();
            element.setLoId(getLoLmsId(dbAccessor, item.getKey()));
            element.setValue(item.getValue());

            result.getLorePrio().add(element);
        }
        return result;
    }

    public static String getLoLmsId(DbAccessor dbAccessor, String loID) {
        KnowledgeObject ko = dbAccessor.getKoData(loID);
        if (ko != null) {
            if (ko.getLmsId() != null && !ko.getLmsId().trim().equals("")) {
                return ko.getLmsId();
            }
        }
        return loID;
    }

    public static TugAck createTugAck(UpdateMessage updateMessage) {
        TugAck tugAck = new TugAck();
        tugAck.setUId(updateMessage.getTargetUserID());
        tugAck.setRId(updateMessage.getMessageID());
        tugAck.setMId(updateMessage.getFeedbackMessageID());
        return tugAck;
    }

    public static Learner merge(Tug tug, Lore lore) {
        Learner learner = new Learner();
        learner.setTime(System.currentTimeMillis());

        if (tug != null) {
            learner.setMId(tug.getMId());
            learner.setUId(tug.getUId());
        } else if (lore != null) {
            learner.setMId(lore.getMId());
            learner.setUId(lore.getUId());
        }

        if (lore != null) {
            learner.setLore(lore);
        }

        if (tug != null) {
            learner.setTug(Collections.singletonList(tug));
        }

        return learner;
    }
}
