/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Version: 1.00
 * Date: 2014/7/25
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.component;

import edu.uk.reading.db.VariableFiller;
import edu.uk.reading.dm.common.LanguageCode;
import edu.uk.reading.dm.common.RequestType;
import edu.uk.reading.dm.common.param.Param;
import edu.uk.reading.dm.manager.UserManager;
import edu.uk.reading.dm.messages.dialogue.FeedbackMessage;
import edu.uk.reading.dm.messages.dialogue.UpdateMessage;
import edu.uk.reading.user.UserProperties;
import eu.intuitel.db.access.DbAccessException;
import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.types.*;
import org.junit.Assert;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import java.util.*;

import static edu.uk.reading.dm.common.util.StringUtils.getParamValue;
import static org.junit.Assert.*;

public class VariableFillerTest {
    private static final String DB_USER = "testUser1";
    private static final String DB_COURSE_ID = "course134";

    private static final String DB_KO1 = "ko1";
    private static final String DB_CC1 = "cc1";
    private static final String DB_CC2 = "cc2";

    private static final String DB_MACRO_LP = "macroLP1";
    private static final String DB_MICRO_LP = "microLP1";

    private static final String DB_NOT_DELETE = "notDeleted";

    private static VariableFiller variableFiller;

    @BeforeClass
    public static void loadModel() {
        UserManager userManager = new UserManagerMockup();
        userManager.setUserProperty(DB_USER, UserProperties.PROP_COURSE_ID, DB_COURSE_ID);

        List<String> errors = new LinkedList<>();
        try {
            variableFiller = new VariableFiller(userManager, DbAccessor.createCachedDbAccessor());
        } catch (DbAccessException e) {
            e.printStackTrace();
        }
        variableFiller.loadFillerModel("dbFiller.xml", errors);
        if (!errors.isEmpty()) {
            for (String error : errors) {
                System.err.println(error);
            }
            System.exit(1);
        }
    }

    @Before
    public void setupTest() {
        variableFiller.dbAccessor.clearDataStorage();
        variableFiller.dbAccessor.storeLearner(DB_USER, DB_USER, null, 22, Culture.Unknown, false, false, false);
        variableFiller.dbAccessor.loginUser(DB_USER, DB_COURSE_ID);
        variableFiller.dbAccessor.storeLearnerPreference(DB_USER, DB_COURSE_ID,
                                                         EqfLevel.Unknown,
                                                         DifficultyLevel.Unknown);

        variableFiller.dbAccessor.storeCcCompletion(DB_USER, DB_COURSE_ID, DB_CC1, false);
        variableFiller.dbAccessor.storeKoCompletionAccessed(DB_USER, DB_COURSE_ID, DB_KO1, true);
        variableFiller.dbAccessor.storeLearningStep(DB_USER, DB_COURSE_ID, DB_KO1, DB_CC1, System.currentTimeMillis());

        variableFiller.dbAccessor.storeMacroLpData(DB_MACRO_LP, DB_COURSE_ID, DB_MACRO_LP, "");
        variableFiller.dbAccessor.storeMacroLpSelection(DB_USER, DB_MACRO_LP, DB_COURSE_ID);

        variableFiller.dbAccessor.storeMicroLpData(DB_MICRO_LP, DB_COURSE_ID, DB_MICRO_LP, "");
        variableFiller.dbAccessor.storeMicroLpSelection(DB_USER, DB_MICRO_LP, DB_COURSE_ID);
    }

    @Test
    public void testActions() {
        testAction(DB_NOT_DELETE, new Param("invalidateLpSelection", "N"));
        testAction(VariableFiller.DB_PROP_DELETE_LP, new Param("invalidateLpSelection", "Y"));
        testAction(VariableFiller.DB_PROP_COMPLETE_ALL_KO, new Param("markKoAsComplete", "Y"));
        testAction(VariableFiller.DB_PROP_RESTART_CURRENT_CC, new Param("restartCurrentCC", "Y"));

        testAction(VariableFiller.DB_PROP_INCREASE_DIFFICULTY, new Param("increaseDifficulty", "Y"));
        testAction(VariableFiller.DB_PROP_DECREASE_DIFFICULTY, new Param("decreaseDifficulty", "Y"));
    }

    @Test
    public void testSetters() {
        testSetParam(VariableFiller.DB_PROP_CURRENT_MACRO_LP, new Param("currentMacroLP", DB_MACRO_LP));
        testSetParam(VariableFiller.DB_PROP_CURRENT_MICRO_LP, new Param("currentMicroLP", DB_MICRO_LP));
        testSetParam(VariableFiller.DB_PROP_CURRENT_CC, new Param("MultipleCurrentCc", DB_CC2));

        testSetParam(VariableFiller.DB_PROP_CURRENT_GENDER, new Param("gender", "f"));
        testSetParam(VariableFiller.DB_PROP_CURRENT_AGE, new Param("age", "23"));
        testSetParam(VariableFiller.DB_PROP_CURRENT_CULTURE, new Param("culture", "european"));
        testSetParam(VariableFiller.DB_PROP_CURRENT_LANGUAGE, new Param("language", "de"));
    }

    @Test
    public void testGetters() {
        testParam(VariableFiller.DB_PROP_CURRENT_CC, true,
                  new Param("DfCurrentCcCompletion", "DfCurrentCcCompletionValue_ThirdThird"), 2);
        testParam(VariableFiller.DB_PROP_CURRENT_CC, true,
                  new Param("DfCurrentCcCompletion", "DfCurrentCcCompletionValue_SecondThird"), 2);
        testParam(VariableFiller.DB_PROP_CURRENT_CC, true,
                  new Param("DfTimePeriodSinceLastLogin", "DfTimePeriodSinceLastLoginValue_LongAgo"), 2);

        testParam(VariableFiller.DB_PROP_LAST_CC, true,
                  new Param("DfLastCcCompletion", "DfLastCcCompletionValue_High"), 2);
        testParam(VariableFiller.DB_PROP_LAST_CC, false,
                  new Param("DfLastCcCompletionD", "DfLastCcCompletionValue_High"), 1);

        testParam(VariableFiller.DB_PROP_CURRENT_DIFFICULTY_LVL, true,
                  new Param("increaseDifficulty", "Y"), 2);
        testParam(VariableFiller.DB_PROP_CURRENT_DIFFICULTY_LVL, false,
                  new Param("increaseDifficulty", "N"), 1);
        testParam(VariableFiller.DB_PROP_CURRENT_DIFFICULTY_LVL, true,
                  new Param("decreaseDifficulty", "Y"), 2);
        testParam(VariableFiller.DB_PROP_CURRENT_DIFFICULTY_LVL, false,
                  new Param("decreaseDifficulty", "N"), 1);
    }

    private static void testAction(String dbParam, Param param) {
        UpdateMessage msg = new UpdateMessage();
        msg.setMessageID(UUID.randomUUID().toString());
        msg.setTargetUserID(DB_USER);
        msg.setTimestamp(System.currentTimeMillis());
        msg.setRequestType(RequestType.request);
        msg.setInputParams(new HashSet<>(Arrays.asList(param)));

        variableFiller.fillDbFromParams(msg);

        boolean solved = false;
        switch (dbParam) {
            case VariableFiller.DB_PROP_DELETE_LP:
                Assert.assertNull(variableFiller.dbAccessor.getMacroLpOfLearner(DB_USER, DB_COURSE_ID));
                Assert.assertNull(variableFiller.dbAccessor.getMicroLpOfLearner(DB_USER, DB_COURSE_ID));
                solved = true;
                break;
            case DB_NOT_DELETE:
                Assert.assertNotNull(variableFiller.dbAccessor.getMacroLpOfLearner(DB_USER, DB_COURSE_ID));
                Assert.assertNotNull(variableFiller.dbAccessor.getMicroLpOfLearner(DB_USER, DB_COURSE_ID));
                solved = true;
                break;
            case VariableFiller.DB_PROP_COMPLETE_ALL_KO:
                assertEquals(variableFiller.dbAccessor.getKoCompletionStatus(DB_USER, DB_KO1).intValue(), 100);
                solved = true;
                break;
            case VariableFiller.DB_PROP_RESTART_CURRENT_CC:
                Integer completion = variableFiller.dbAccessor.getKoCompletionStatus(DB_USER, DB_KO1);
                assertNull(completion);
                solved = true;
                break;

            case VariableFiller.DB_PROP_INCREASE_DIFFICULTY:
                assertEquals(variableFiller.dbAccessor.getPreferredDifficultyLevel(DB_USER, DB_COURSE_ID),
                             DifficultyLevel.Intermediate);
                solved = true;
                break;
            case VariableFiller.DB_PROP_DECREASE_DIFFICULTY:
                assertEquals(variableFiller.dbAccessor.getPreferredDifficultyLevel(DB_USER, DB_COURSE_ID),
                             DifficultyLevel.Beginner);
                solved = true;
                break;
        }
        Assert.assertTrue("No test has been conducted for this action!", solved);
    }

    private static void testSetParam(String dbParam, Param param) {
        UpdateMessage msg = new UpdateMessage();
        msg.setMessageID(UUID.randomUUID().toString());
        msg.setTargetUserID(DB_USER);
        msg.setTimestamp(System.currentTimeMillis());
        msg.setRequestType(RequestType.request);
        msg.setInputParams(new HashSet<>(Arrays.asList(param)));

        variableFiller.fillDbFromParams(msg);

        boolean solved = false;
        switch (dbParam) {
            case VariableFiller.DB_PROP_CURRENT_MACRO_LP:
                assertEquals(variableFiller.dbAccessor.getMacroLpOfLearner(DB_USER, DB_COURSE_ID),
                             getParamValue(param, LanguageCode.DEFAULT));
                solved = true;
                break;
            case VariableFiller.DB_PROP_CURRENT_MICRO_LP:
                assertEquals(variableFiller.dbAccessor.getMicroLpOfLearner(DB_USER, DB_COURSE_ID),
                             getParamValue(param, LanguageCode.DEFAULT));
                solved = true;
                break;
            case VariableFiller.DB_PROP_CURRENT_CC:
                assertEquals(variableFiller.dbAccessor.getLastConceptContainerIri(DB_USER),
                             getParamValue(param, LanguageCode.DEFAULT));
                solved = true;
                break;
            case VariableFiller.DB_PROP_CURRENT_AGE:
                assertEquals(variableFiller.dbAccessor.getLearnerAge(DB_USER),
                             Integer.valueOf(getParamValue(param, LanguageCode.DEFAULT).toString()));
                solved = true;
                break;
            case VariableFiller.DB_PROP_CURRENT_GENDER:
                assertEquals(variableFiller.dbAccessor.getLearnerGender(DB_USER),
                             Gender.getGender(getParamValue(param, LanguageCode.DEFAULT).toString()));
                solved = true;
                break;
            case VariableFiller.DB_PROP_CURRENT_CULTURE:
                assertEquals(variableFiller.dbAccessor.getLearnerCulture(DB_USER),
                             Culture.getCulture(getParamValue(param, LanguageCode.DEFAULT).toString()));
                solved = true;
                break;
            case VariableFiller.DB_PROP_CURRENT_LANGUAGE:
                Collection<Language> languages = variableFiller.dbAccessor.getLearnerLanguages(DB_USER);
                assertNotNull(languages);
                Language paramLanguage = Language.getLanguage(getParamValue(param, LanguageCode.DEFAULT).toString());
                assertTrue(languages.contains(paramLanguage));
                solved = true;
                break;
        }
        Assert.assertTrue("No test has been conducted for this action!", solved);
    }

    private static void testParam(String dbParam, boolean shouldContain,
                                  Param param, int paramSize) {
        FeedbackMessage msg = new FeedbackMessage();
        msg.setMessageID(UUID.randomUUID().toString());
        msg.setTargetUserID(DB_USER);
        msg.setTimestamp(System.currentTimeMillis());
        msg.setRequestType(RequestType.inform);
        msg.setOutputParams(new HashSet<>(Arrays.asList(param)));

        variableFiller.fillParamsFromDb(msg);

        assertEquals(msg.getOutputParams().contains(new Param(dbParam)), shouldContain);
        assertEquals(msg.getOutputParams().size(), paramSize);
    }
}

