/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Version: 1.00
 * Date: 2014/7/25
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.model;

import edu.uk.reading.component.UserManagerMockup;
import edu.uk.reading.db.VariableFiller;
import edu.uk.reading.dm.common.RequestType;
import edu.uk.reading.dm.common.param.Param;
import edu.uk.reading.dm.manager.UserManager;
import edu.uk.reading.dm.messages.GenericMessage;
import edu.uk.reading.dm.messages.dialogue.ErrorMessage;
import edu.uk.reading.dm.messages.dialogue.FeedbackMessage;
import edu.uk.reading.dm.messages.dialogue.ReplyMessage;
import edu.uk.reading.dm.messages.dialogue.UpdateMessage;
import edu.uk.reading.dm.model.ModelLoader;
import edu.uk.reading.dm.model.ParseException;
import edu.uk.reading.dm.nlg.NlgDecoratedResult;
import edu.uk.reading.dm.nlg.NlgManager;
import edu.uk.reading.dm.session.UserSession;
import edu.uk.reading.user.UserProperties;
import eu.intuitel.db.access.DbAccessException;
import eu.intuitel.db.access.DbAccessor;
import eu.intuitel.types.Culture;
import eu.intuitel.types.DifficultyLevel;
import eu.intuitel.types.EqfLevel;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import java.util.*;

import static edu.uk.reading.dm.common.LanguageCode.DEFAULT;
import static edu.uk.reading.dm.messages.MessageUtils.*;
import static edu.uk.reading.tools.CommonUtils.*;
import static org.junit.Assert.*;

public class IntuitelModelInformDbFillerTest {
    private static final String DB_USER = "testUser1";
    private static final String DB_COURSE_ID = "course134";
    private static final String DB_MACRO_LP = "maLP";
    private static final String DB_MICRO_LP = "miLP";
    private static final String DB_CC = "cc";
    private static final String DB_KO = "ko";
    private static final String DB_KO_PREVIOUS = "ko-prev";

    private static UserSession userSession;
    private static NlgManager nlgManager;
    private static VariableFiller variableFiller;

    @BeforeClass
    public static void setup() {
        try {
            UserManager userManager = new UserManagerMockup();

            ModelLoader.ParseResult parseResult = ModelLoader.load(DEFAULT, file("inform-dialogue-model.xml"));
            if (!parseResult.getErrors().isEmpty()) {
                System.err.println(parseResult.getErrors());
                fail();
            }
            userSession = new UserSession(UUID.randomUUID().toString(), parseResult.getDialogueModel());
            nlgManager = NlgManager.getInstance(parseResult.getNlgModel());

            List<String> errors = new LinkedList<>();
            try {
                variableFiller = new VariableFiller(userManager, DbAccessor.createCachedDbAccessor());
            } catch (DbAccessException e) {
                e.printStackTrace();
            }
            variableFiller.loadFillerModel("../dbFiller.xml", errors);
            assertTrue("Failed to load Db Filler model", errors.isEmpty());

            userManager.setUserProperty(DB_USER, UserProperties.PROP_COURSE_ID, DB_COURSE_ID);
        } catch (ParseException e) {
            e.printStackTrace();
        }
    }

    @Before
    public void setupTest() {
        variableFiller.dbAccessor.clearDataStorage();
        variableFiller.dbAccessor.storeLearner(DB_USER, DB_USER, null, 22, Culture.Unknown, false, false, false);
        variableFiller.dbAccessor.loginUser(DB_USER, DB_COURSE_ID);
        variableFiller.dbAccessor.storeLearnerPreference(DB_USER, DB_COURSE_ID,
                                                         EqfLevel.Unknown,
                                                         DifficultyLevel.Unknown);

        variableFiller.dbAccessor.storeCcCompletion(DB_USER, DB_COURSE_ID, DB_CC, false);
        variableFiller.dbAccessor.storeKoCompletionAccessed(DB_USER, DB_COURSE_ID, DB_KO_PREVIOUS, false);
        variableFiller.dbAccessor.storeKoCompletionAccessed(DB_USER, DB_COURSE_ID, DB_KO, false);

        variableFiller.dbAccessor
                .storeLearningStep(DB_USER, DB_COURSE_ID, DB_KO_PREVIOUS, DB_CC, System.currentTimeMillis());
        variableFiller.dbAccessor.storeLearningStep(DB_USER, DB_COURSE_ID, DB_KO, DB_CC, System.currentTimeMillis());

        variableFiller.dbAccessor.storeMacroLpData(DB_MACRO_LP, DB_COURSE_ID, DB_MACRO_LP, "");
        variableFiller.dbAccessor.storeMacroLpSelection(DB_USER, DB_MACRO_LP, DB_COURSE_ID);

        variableFiller.dbAccessor.storeMicroLpData(DB_MICRO_LP, DB_COURSE_ID, DB_MICRO_LP, "");
        variableFiller.dbAccessor.storeMicroLpSelection(DB_USER, DB_MICRO_LP, DB_COURSE_ID);
    }

    @Test
    public void testInformTotalCcCompletionMessages() {
        testSimpleInform("nlg.learning.total.cc.completion.full",
                         "DfTotalCcCompletion", "DfTotalCcCompletionValue_Full",
                         "invalidateLpSelection", "N");
        assertEquals(DB_MACRO_LP, variableFiller.dbAccessor.getMacroLpOfLearner(DB_USER, DB_COURSE_ID));
        assertEquals(DB_MICRO_LP, variableFiller.dbAccessor.getMicroLpOfLearner(DB_USER, DB_COURSE_ID));

        testSimpleInform("nlg.learning.total.cc.completion.full",
                         "DfTotalCcCompletion", "DfTotalCcCompletionValue_Full",
                         "invalidateLpSelection", "Y");
        assertNull(variableFiller.dbAccessor.getMacroLpOfLearner(DB_USER, DB_COURSE_ID));
        assertNull(variableFiller.dbAccessor.getMicroLpOfLearner(DB_USER, DB_COURSE_ID));
    }

    @Test
    public void testInformLastCcCompletionMessages() {
        testSimpleInform("nlg.learning.last.cc.completion.high",
                         "DfLastCcCompletion", "DfLastCcCompletionValue_High",
                         "markKoAsComplete", "N");
        assertTrue(variableFiller.dbAccessor.getCompletedKOs(DB_USER, DB_COURSE_ID).isEmpty());

        testSimpleInform("nlg.learning.last.cc.completion.high",
                         "DfLastCcCompletion", "DfLastCcCompletionValue_High",
                         "markKoAsComplete", "Y");
        assertFalse(variableFiller.dbAccessor.getCompletedKOs(DB_USER, DB_COURSE_ID).isEmpty());
        assertTrue(variableFiller.dbAccessor.getCompletedKOs(DB_USER, DB_COURSE_ID).contains(DB_KO));
    }

    @Test
    public void testInformTimeStatisticsMessages() {
        testSimpleInform("nlg.learning.session.length.long",
                         "DfComparativeSessionLength", "DfComparativeSessionLengthValue_Long",
                         "troubleWithLearningMaterial", "Y");

        testSimpleInform("nlg.learning.session.length.short",
                         "DfComparativeSessionLength", "DfComparativeSessionLengthValue_Short",
                         "increaseDifficulty", "Y");
        assertEquals(DifficultyLevel.DEFAULT,
                     variableFiller.dbAccessor.getPreferredDifficultyLevel(DB_USER, DB_COURSE_ID));
    }

    @Test
    public void testInformLoginTimeStatisticsMessages() {
        testSimpleInform("nlg.login.last.long.ago",
                         "DfTimePeriodSinceLastLogin", "DfTimePeriodSinceLastLoginValue_LongAgo",
                         "restartCurrentCC", "N");
        assertEquals(variableFiller.dbAccessor.getLastConceptContainerIri(DB_USER), DB_CC);
        assertEquals(variableFiller.dbAccessor.getLastKnowledgeObjectIri(DB_USER, DB_COURSE_ID), DB_KO);

        testSimpleInform("nlg.login.last.long.ago",
                         "DfTimePeriodSinceLastLogin", "DfTimePeriodSinceLastLoginValue_LongAgo",
                         "restartCurrentCC", "Y");
        assertEquals(variableFiller.dbAccessor.getLastConceptContainerIri(DB_USER), DB_CC);
        assertEquals(variableFiller.dbAccessor.getLastKnowledgeObjectIri(DB_USER, DB_COURSE_ID), DB_KO_PREVIOUS);
    }

    private FeedbackMessage createInformFeedbackMessage(RequestType type,
                                                        Set<Param> inputParams,
                                                        Set<Param> outputParams) {
        FeedbackMessage feedbackMessage = new FeedbackMessage();
        feedbackMessage.setMessageID(UUID.randomUUID().toString());
        feedbackMessage.setTargetUserID(DB_USER);
        feedbackMessage.setTimestamp(System.currentTimeMillis());
        feedbackMessage.setRequestType(type);
        feedbackMessage.setInputParams(inputParams);
        feedbackMessage.setOutputParams(outputParams);
        return feedbackMessage;
    }

    private void testSimpleInform(String expectedNlgPattern,
                                  String param, String paramValues,
                                  String answer, String answerValue) {
        testInform(expectedNlgPattern, ttuple(new Param(param, paramValues)));
        testAnswer(expectedNlgPattern, ttuple(new Param(answer, answerValue)));
    }

    private void testAnswer(String expectedNlgPattern, Param[] inputParams) {
        FeedbackMessage feedbackMessage = createInformFeedbackMessage(RequestType.reply, tset(inputParams), null);
        variableFiller.fillParamsFromDb(feedbackMessage);

        List<GenericMessage> genericMessages = userSession.processFeedbackMessage(feedbackMessage);
        assertFalse(String.format("Error reached to the recommendation: NLG Pattern=%s, Params=%s",
                                  expectedNlgPattern, Arrays.toString(inputParams)),
                    contains(genericMessages, ErrorMessage.class));
        assertNotNull(String.format("The engine returned an empty message : NLG Pattern=%s, Params=%s",
                                    expectedNlgPattern, Arrays.toString(inputParams)), genericMessages);
        if (contains(genericMessages, UpdateMessage.class)) {
            variableFiller.fillDbFromParams(getUpdate(genericMessages));
        }
    }

    private void testInform(String expectedNlgPattern, Param[] outputParams) {
        FeedbackMessage feedbackMessage = createInformFeedbackMessage(RequestType.inform,
                                                                      null,
                                                                      tset(outputParams));
        variableFiller.fillParamsFromDb(feedbackMessage);

        List<GenericMessage> genericMessages = userSession.processFeedbackMessage(feedbackMessage);
        assertFalse(String.format("Error reached to the recommendation: NLG Pattern=%s, Params=%s",
                                  expectedNlgPattern, Arrays.toString(outputParams)),
                    contains(genericMessages, ErrorMessage.class));
        assertNotNull(String.format("The engine returned an empty message : NLG Pattern=%s, Params=%s",
                                    expectedNlgPattern, Arrays.toString(outputParams)), genericMessages);
        assertTrue(String.format("Return result is invalid : NLG Pattern=%s, Params=%s",
                                 expectedNlgPattern, Arrays.toString(outputParams)),
                   contains(genericMessages, ReplyMessage.class));

        NlgDecoratedResult nlgDecoratedResult = nlgManager.decorateMessage(getReply(genericMessages));
        assertNotNull(String.format("The NLG Engine returned an empty message : NLG Pattern=%s, Params=%s",
                                    expectedNlgPattern, Arrays.toString(outputParams)), nlgDecoratedResult);
        assertNotNull(String.format("The NLP Engine returned an empty pattern : NLG Pattern=%s, Params=%s",
                                    expectedNlgPattern, Arrays.toString(outputParams)),
                      nlgDecoratedResult.getPattern());
        assertNotNull(String.format("The NLP Engine returned an empty answer : NLG Pattern=%s, Params=%s",
                                    expectedNlgPattern, Arrays.toString(outputParams)),
                      nlgDecoratedResult.getAnswers());
    }
}
