/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Version: 1.00
 * Date: 2014/7/25
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.model;

import edu.uk.reading.dm.common.RequestType;
import edu.uk.reading.dm.common.param.Param;
import edu.uk.reading.dm.messages.GenericMessage;
import edu.uk.reading.dm.messages.dialogue.ErrorMessage;
import edu.uk.reading.dm.messages.dialogue.FeedbackMessage;
import edu.uk.reading.dm.messages.dialogue.ReplyMessage;
import edu.uk.reading.dm.model.ModelLoader;
import edu.uk.reading.dm.model.ParseException;
import edu.uk.reading.dm.nlg.NlgDecoratedResult;
import edu.uk.reading.dm.nlg.NlgManager;
import edu.uk.reading.dm.session.UserSession;
import org.junit.Before;
import org.junit.Test;

import java.util.*;

import static edu.uk.reading.dm.common.LanguageCode.DEFAULT;
import static edu.uk.reading.dm.messages.MessageUtils.contains;
import static edu.uk.reading.dm.messages.MessageUtils.getReply;
import static edu.uk.reading.tools.CommonUtils.*;
import static org.junit.Assert.*;

public class IntuitelModelInformTest {
    private UserSession userSession;
    private NlgManager nlgManager;

    @Before
    public void setup() {
        try {
            ModelLoader.ParseResult parseResult = ModelLoader.load(DEFAULT, file("inform-dialogue-model.xml"));
            if (!parseResult.getErrors().isEmpty()) {
                System.err.println(parseResult.getErrors());
            }
            userSession = new UserSession(UUID.randomUUID().toString(), parseResult.getDialogueModel());
            nlgManager = NlgManager.getInstance(parseResult.getNlgModel());
        } catch (ParseException e) {
            e.printStackTrace();
        }
    }

    @Test
    public void testInformTotalCcCompletionMessages() {
        //-------------- total CC completion -------------------------------------------------------
        testSimpleInform("nlg.learning.total.cc.completion.full", 1,
                         "DfTotalCcCompletion", "DfTotalCcCompletionValue_Full");
        //--------------- total CC completion positive feedback-------------------------------------
        testComplexInform("nlg.learning.total.cc.completion.almost.full.positive",
                          "DfTotalCcCompletion",
                          new String[]{"DfTotalCcCompletionValue_AlmostFull"},
                          "DfAbsoluteCourseTimeExposure",
                          new String[]{"DfAbsoluteCourseTimeExposureValue_VeryLong",
                                              "DfAbsoluteCourseTimeExposureValue_Long",
                                              "DfAbsoluteCourseTimeExposureValue_HalfTime",
                                              "DfAbsoluteCourseTimeExposureValue_Short",
                                              "DfAbsoluteCourseTimeExposureValue_VeryShort"});
        testComplexInform("nlg.learning.total.cc.completion.almost.full.positive",
                          "DfTotalCcCompletion",
                          new String[]{"DfTotalCcCompletionValue_AlmostFull"},
                          "DfAbsoluteCourseTimeExposure",
                          new String[]{"DfAbsoluteCourseTimeExposureValue_HalfTime",
                                              "DfAbsoluteCourseTimeExposureValue_Short",
                                              "DfAbsoluteCourseTimeExposureValue_VeryShort"});
        //--------------- total CC completion negative feedback-------------------------------------
        testRandomInform(tlist("nlg.learning.total.cc.completion.almost.full.negative",
                               "nlg.learning.total.cc.completion.almost.full.negative.diagnostic"),
                         "DfTotalCcCompletion",
                         new String[]{"DfTotalCcCompletionValue_AlmostFull"},
                         "DfAbsoluteCourseTimeExposure",
                         new String[]{"DfAbsoluteCourseTimeExposureValue_CompleteOrOver"});
        testRandomInform(tlist("nlg.learning.total.cc.completion.half.time.negative",
                               "nlg.learning.total.cc.completion.half.time.negative.diagnostic"),
                         "DfTotalCcCompletion",
                         new String[]{"DfTotalCcCompletionValue_Half"},
                         "DfAbsoluteCourseTimeExposure",
                         new String[]{"DfAbsoluteCourseTimeExposureValue_CompleteOrOver",
                                             "DfAbsoluteCourseTimeExposureValue_VeryLong",
                                             "DfAbsoluteCourseTimeExposureValue_Long"});
    }

    @Test
    public void testInformCurrentCcCompletionMessages() {
        //--------------- current CC completion positive feedback------------------------------------
        testComplexInform("nlg.learning.total.cc.completion.third.third.positive", 0,
                          "DfCurrentCcCompletion",
                          new String[]{"DfCurrentCcCompletionValue_ThirdThird"},
                          "DfAbsoluteCourseTimeExposure",
                          new String[]{"DfAbsoluteCourseTimeExposureValue_HalfTime",
                                              "DfAbsoluteCourseTimeExposureValue_Short",
                                              "DfAbsoluteCourseTimeExposureValue_VeryShort"},
                          new Param("currentCC", "cc1"));
        testComplexInform("nlg.learning.total.cc.completion.second.third.positive", 0,
                          "DfCurrentCcCompletion",
                          new String[]{"DfCurrentCcCompletionValue_SecondThird"},
                          "DfAbsoluteCourseTimeExposure",
                          new String[]{"DfAbsoluteCourseTimeExposureValue_Short",
                                              "DfAbsoluteCourseTimeExposureValue_VeryShort"},
                          new Param("currentCC", "cc1"));
        //--------------- current CC completion negative feedback------------------------------------
        testRandomInform(tlist("nlg.learning.total.cc.completion.third.third.negative",
                               "nlg.learning.total.cc.completion.third.third.negative.diagnostic"),
                         "DfCurrentCcCompletion",
                         new String[]{"DfCurrentCcCompletionValue_ThirdThird"},
                         "DfAbsoluteCourseTimeExposure",
                         new String[]{"DfAbsoluteCourseTimeExposureValue_VeryLong",
                                             "DfAbsoluteCourseTimeExposureValue_Long",
                                             "DfAbsoluteCourseTimeExposureValue_CompleteOrOver"},
                         new Param("currentCC", "cc1"));
        testRandomInform(tlist("nlg.learning.total.cc.completion.second.third.negative",
                               "nlg.learning.total.cc.completion.second.third.negative.diagnostic"),
                         "DfCurrentCcCompletion",
                         new String[]{"DfCurrentCcCompletionValue_SecondThird"},
                         "DfAbsoluteCourseTimeExposure",
                         new String[]{"DfAbsoluteCourseTimeExposureValue_HalfTime",
                                             "DfAbsoluteCourseTimeExposureValue_CompleteOrOver",
                                             "DfAbsoluteCourseTimeExposureValue_VeryLong",
                                             "DfAbsoluteCourseTimeExposureValue_Long"},
                         new Param("currentCC", "cc1"));
    }

    @Test
    public void testInformLastCcCompletionMessages() {
        testSimpleInform("nlg.learning.last.cc.completion.full", 0,
                         "DfLastCcCompletion", "DfLastCcCompletionValue_Full",
                         "lastCC", "cc1");
        testSimpleInform("nlg.learning.last.cc.completion.high", 1,
                         "DfLastCcCompletion", "DfLastCcCompletionValue_High",
                         "lastCC", "cc1");
        testSimpleInform("nlg.learning.last.cc.completion.medium.to.low", 0,
                         "DfLastCcCompletion", "DfLastCcCompletionValue_Medium",
                         "lastCC", "cc1");
        testSimpleInform("nlg.learning.last.cc.completion.medium.to.low", 0,
                         "DfLastCcCompletion", "DfLastCcCompletionValue_Low",
                         "lastCC", "cc1");
        testSimpleInform("nlg.learning.last.cc.completion.medium.to.low", 0,
                         "DfLastCcCompletion", "DfLastCcCompletionValue_VeryLow",
                         "lastCC", "cc1");
    }

    @Test
    public void testInformTimeStatisticsMessages() {
        testSimpleInform("nlg.learning.comparative.course.time.longer", 0,
                         "DfComparativeCourseTime", "DfComparativeCourseTimeValue_Longer");
        testSimpleInform("nlg.learning.comparative.course.time.average", 0,
                         "DfComparativeCourseTime", "DfComparativeCourseTimeValue_Average");
        testSimpleInform("nlg.learning.comparative.course.time.shorter", 0,
                         "DfComparativeCourseTime", "DfComparativeCourseTimeValue_Shorter");

        testSimpleInform("nlg.learning.recent.learning.pace.slow", 0,
                         "DfRecentLearningPace", "DfRecentLearningPaceValue_Slow");
        testSimpleInform("nlg.learning.recent.learning.pace.fast", 0,
                         "DfRecentLearningPace", "DfRecentLearningPaceValue_Fast");

        testSimpleInform("nlg.learning.comparative.ko.time.slower", 0,
                         "DfComparativeKoTime", "DfComparativeKoTimeValue_Slower");
        testSimpleInform("nlg.learning.comparative.ko.time.faster", 0,
                         "DfComparativeKoTime", "DfComparativeKoTimeValue_Faster");

        testSimpleInform("nlg.learning.session.length.long", 1,
                         "DfComparativeSessionLength", "DfComparativeSessionLengthValue_Long");
        testSimpleInform("nlg.learning.session.length.short", 1,
                         "DfComparativeSessionLength", "DfComparativeSessionLengthValue_Short");
    }

    @Test
    public void testInformLoginTimeStatisticsMessages() {
        testSimpleInform("nlg.login.last.long.ago", 1,
                         "DfTimePeriodSinceLastLogin", "DfTimePeriodSinceLastLoginValue_LongAgo",
                         "currentCC", "cc1");
        testSimpleInform("nlg.login.last.recently",
                         "DfTimePeriodSinceLastLogin", "DfTimePeriodSinceLastLoginValue_Recently");
    }

    @Test
    public void testInformEnvironmentFeedbackMessages() {
        testSimpleInform("nlg.session.connectivity.poor",
                         "DfConnectivity", "DfConnectivityValue_Poor");
        testSimpleInform("nlg.session.noise.loud",
                         "DfNoise", "DfNoiseValue_Loud");
        testSimpleInform("nlg.session.noise.loud",
                         "DfNoise", "DfNoiseValue_VeryLoud");
    }

    private FeedbackMessage createInformFeedbackMessage(Set<Param> inputParams, Set<Param> outputParams) {
        FeedbackMessage feedbackMessage = new FeedbackMessage();
        feedbackMessage.setMessageID(UUID.randomUUID().toString());
        feedbackMessage.setTargetUserID(UUID.randomUUID().toString());
        feedbackMessage.setTimestamp(System.currentTimeMillis());
        feedbackMessage.setRequestType(RequestType.inform);
        feedbackMessage.setInputParams(inputParams);
        feedbackMessage.setOutputParams(outputParams);
        return feedbackMessage;
    }

    private void testSimpleInform(String expectedNlgPattern, String param, String paramValues) {
        testSimpleInform(expectedNlgPattern, 0, param, paramValues);
    }

    private void testSimpleInform(String expectedNlgPattern, int answerListSize, String param, String paramValues) {
        testInform(tlist(expectedNlgPattern), answerListSize, new Param(param, paramValues));
    }

    private void testSimpleInform(String expectedNlgPattern, int answerListSize,
                                  String param1, String param1Values,
                                  String param2, String param2Values) {
        testInform(tlist(expectedNlgPattern), answerListSize,
                   new Param(param1, param1Values), new Param(param2, param2Values));
    }

    private void testComplexInform(String expectedNlgPattern,
                                   String param1, String[] param1Values,
                                   String param2, String[] param2Values) {
        testComplexInform(expectedNlgPattern, 0, param1, param1Values, param2, param2Values);
    }

    private void testComplexInform(String expectedNlgPattern, int answerListSize,
                                   String param1, String[] param1Values,
                                   String param2, String[] param2Values) {
        for (String param1Value : param1Values) {
            for (String param2Value : param2Values) {
                testInform(tlist(expectedNlgPattern), answerListSize,
                           new Param(param1, param1Value), new Param(param2, param2Value));
            }
        }
    }

    private void testComplexInform(String expectedNlgPattern, int answerListSize,
                                   String param1, String[] param1Values,
                                   String param2, String[] param2Values,
                                   Param additionalParams) {
        for (String param1Value : param1Values) {
            for (String param2Value : param2Values) {
                testInform(tlist(expectedNlgPattern), answerListSize,
                           new Param(param1, param1Value), new Param(param2, param2Value), additionalParams);
            }
        }
    }

    private void testRandomInform(List<String> expectedNlgPattern,
                                  String param1, String[] param1Values,
                                  String param2, String[] param2Values) {
        for (String param1Value : param1Values) {
            for (String param2Value : param2Values) {
                testInform(expectedNlgPattern, -1,
                           new Param(param1, param1Value), new Param(param2, param2Value));
            }
        }
    }

    private void testRandomInform(List<String> expectedNlgPattern,
                                  String param1, String[] param1Values,
                                  String param2, String[] param2Values,
                                  Param additionalParams) {
        for (String param1Value : param1Values) {
            for (String param2Value : param2Values) {
                testInform(expectedNlgPattern, -1,
                           new Param(param1, param1Value), new Param(param2, param2Value), additionalParams);
            }
        }
    }

    private void testInform(List<String> expectedNlgPattern, int answerListSize, Param... outputParams) {
        FeedbackMessage feedbackMessage = createInformFeedbackMessage(null, new HashSet<>(Arrays.asList(outputParams)));

        String patternString = expectedNlgPattern.toString();
        String outString = Arrays.toString(outputParams);

        List<GenericMessage> genericMessages = userSession.processFeedbackMessage(feedbackMessage);
        assertFalse(String.format("Error reached to the recommendation: NLG Pattern=%s, Params=%s",
                                  patternString, outString), contains(genericMessages, ErrorMessage.class));
        assertNotNull(String.format("The engine returned an empty message : NLG Pattern=%s, Params=%s",
                                    patternString, outString), genericMessages);
        assertTrue(String.format("Return result is invalid : NLG Pattern=%s, Params=%s",
                                 patternString, outString), contains(genericMessages, ReplyMessage.class));

        NlgDecoratedResult nlgDecoratedResult = nlgManager.decorateMessage(getReply(genericMessages));
        assertNotNull(String.format("The NLG Engine returned an empty message : NLG Pattern=%s, Params=%s",
                                    patternString, outString), nlgDecoratedResult);
        assertNotNull(String.format("The NLP Engine returned an empty pattern : NLG Pattern=%s, Params=%s",
                                    patternString, outString), nlgDecoratedResult.getPattern());
        assertNotNull(String.format("The NLP Engine returned an empty answer : NLG Pattern=%s, Params=%s",
                                    patternString, outString), nlgDecoratedResult.getAnswers());
        if (answerListSize >= 0) {
            assertEquals(String.format("The answer list size does not match : NLG Pattern=%s, Params=%s",
                                       patternString, outString),
                         answerListSize, nlgDecoratedResult.getAnswers().size());
        }

        boolean containsAny = containsAny(getReply(genericMessages).getNlgTemplateIDs(), expectedNlgPattern);
        String templateId = getReply(genericMessages).getNlgTemplateIDs().toString();
        assertTrue(String.format("The NLG Engine returned an invalid pattern : NLG Pattern=%s, Params=%s, " +
                                 "Actual patterns=%s", patternString, outString, templateId), containsAny);
    }
}
