/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Version: 1.00
 * Date: 2014/7/25
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.model;

import edu.uk.reading.dm.common.LanguageCode;
import edu.uk.reading.dm.common.param.Param;
import edu.uk.reading.dm.model.ModelLoader;
import edu.uk.reading.dm.model.ParseException;
import edu.uk.reading.dm.nlg.*;
import edu.uk.reading.dm.states.DMState;
import edu.uk.reading.dm.states.DialogueModel;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;

import java.io.File;
import java.util.*;

import static edu.uk.reading.tools.CommonUtils.*;
import static org.junit.Assert.*;

@RunWith(Parameterized.class)
public class IntuitelModelValidation {
    private DialogueModel dialogueModel;
    private NlgModel nlgModel;
    private boolean hasErrors = false;

    private File[] currentModel;
    private LanguageCode currentLanguage;

    public IntuitelModelValidation(File[] model, LanguageCode language) {
        this.currentModel = model;
        this.currentLanguage = language;
    }

    @Parameterized.Parameters
    public static Collection<Object[]> primeNumbers() {
        return list(tuple(files("inform-dialogue-model.xml"), LanguageCode.EN),
                    tuple(files("request-dialogue-model.xml"), LanguageCode.EN),
                    tuple(files("inform-dialogue-model.xml"), LanguageCode.DE),
                    tuple(files("request-dialogue-model.xml"), LanguageCode.DE));
    }

    @Before
    public void setup() {
        try {
            ModelLoader.ParseResult parseResult = ModelLoader.load(currentLanguage, true, currentModel);
            if (!parseResult.getErrors().isEmpty()) {
                System.err.println(parseResult.getErrors());
                hasErrors = true;
            }
            dialogueModel = parseResult.getDialogueModel();
            nlgModel = parseResult.getNlgModel();
        } catch (ParseException e) {
            e.printStackTrace();
        }
    }

    @Test
    public void testHasErrors() {
        assertFalse(hasErrors);
    }

    @Test
    public void testExistingStartStates() {
        assertNotNull(dialogueModel.getStates());
        assertFalse(dialogueModel.getStates().isEmpty());
    }

    @Test
    public void testExistingEndStates() {
        assertTrue(hasFinal(dialogueModel.getStates(), new HashSet<DMState>()));
    }

    @Test
    public void testModelConsistency() {
        Set<DMState> visitedStates = new HashSet<>();
        Set<DMState> invalidStates = new HashSet<>();
        visitModel(dialogueModel.getStates(), visitedStates, invalidStates);
        Set<DMState> all = new HashSet<>(dialogueModel.getAllStates());
        all.removeAll(visitedStates);

        for (DMState state : all) {
            System.err.println("Non visited state: " + state);
        }
        for (DMState state : invalidStates) {
            System.err.println("Invalid state: " + state);
        }
        assertTrue(all.isEmpty());
        assertTrue(invalidStates.isEmpty());
    }

    @Test
    public void testModelValidity() {
        assertTrue(areStatesValid(dialogueModel.getStates(), new HashSet<DMState>()));
    }

    @Test
    public void testLanguageConsistency() {
        if (!LanguageCode.DEFAULT.equals(currentLanguage)) {
            Map<String, Map<String, Set<String>>> defaultNlgModel = loadDefaultNlgModel();

            Assert.assertEquals(String.format("Model=%s, Language=%s", Arrays.toString(currentModel), currentLanguage),
                                size(defaultNlgModel), size(nlgModel.getNlgPatterns()));

            for (NlgPattern pattern : nlgModel.getNlgPatterns()) {
                Assert.assertNotNull(String.format("Model=%s, Language=%s, Pattern=%s",
                                                   Arrays.toString(currentModel),
                                                   currentLanguage, pattern.getPatternID()),
                                     defaultNlgModel.get(pattern.getPatternID()));
                Map<String, Set<String>> defaultModelAnswers = defaultNlgModel.get(pattern.getPatternID());
                Assert.assertEquals(String.format("Model=%s, Language=%s, Pattern=%s",
                                                  Arrays.toString(currentModel),
                                                  currentLanguage, pattern.getPatternID()),
                                    size(defaultModelAnswers), size(pattern.getAnswerList()));

                if (pattern.getAnswerList() != null) {
                    for (NlgAnswer nlgAnswer : pattern.getAnswerList()) {
                        Assert.assertNotNull(String.format("Model=%s, Language=%s, Pattern=%s",
                                                           Arrays.toString(currentModel),
                                                           currentLanguage, pattern.getPatternID()),
                                             defaultModelAnswers.get(nlgAnswer.getParamName()));
                        Set<String> defaultModelOptions = defaultModelAnswers.get(nlgAnswer.getParamName());
                        Set<String> newModelOptions = new HashSet<>();

                        if (nlgAnswer.getOptions() != null) {
                            for (NlgAnswerOption nlgAnswerOption : nlgAnswer.getOptions()) {
                                newModelOptions.add(nlgAnswerOption.getValue());
                            }
                        } else if (nlgAnswer.getSourceParam() != null) {
                            newModelOptions.add(nlgAnswer.getSourceParam());
                        } else if (nlgAnswer.getType() == NlgAnswerType.freeInput) {
                            newModelOptions.add(NlgAnswerType.freeInput.name());
                        }
                        Assert.assertEquals(String.format("Model=%s, Language=%s, Pattern=%s",
                                                          Arrays.toString(currentModel),
                                                          currentLanguage, pattern.getPatternID()),
                                            size(defaultModelOptions), size(newModelOptions));
                        Assert.assertTrue(String.format("Model=%s, Language=%s, Pattern=%s",
                                                        Arrays.toString(currentModel),
                                                        currentLanguage, pattern.getPatternID()),
                                          defaultModelOptions.containsAll(newModelOptions));
                        Assert.assertTrue(String.format("Model=%s, Language=%s, Pattern=%s",
                                                        Arrays.toString(currentModel),
                                                        currentLanguage, pattern.getPatternID()),
                                          newModelOptions.containsAll(defaultModelOptions));
                        defaultModelAnswers.remove(nlgAnswer.getParamName());
                    }
                }
                Assert.assertTrue(String.format("Model=%s, Language=%s, Pattern=%s",
                                                Arrays.toString(currentModel),
                                                currentLanguage, pattern.getPatternID()),
                                  defaultModelAnswers.isEmpty());
                defaultNlgModel.remove(pattern.getPatternID());
            }
            Assert.assertTrue(String.format("Model=%s, Language=%s", Arrays.toString(currentModel), currentLanguage),
                              defaultNlgModel.isEmpty());
        }
    }

    private Map<String, Map<String, Set<String>>> loadDefaultNlgModel() {
        try {
            ModelLoader.ParseResult parseResult = ModelLoader.load(LanguageCode.DEFAULT, true, currentModel);
            if (!parseResult.getErrors().isEmpty()) {
                return null;
            }

            Map<String, Map<String, Set<String>>> result = new HashMap<>();
            for (NlgPattern pattern : parseResult.getNlgModel().getNlgPatterns()) {
                Map<String, Set<String>> answers = result.get(pattern.getPatternID());
                if (answers == null) {
                    answers = new HashMap<>();
                    result.put(pattern.getPatternID(), answers);
                }

                if (pattern.getAnswerList() != null && !pattern.getAnswerList().isEmpty()) {
                    for (NlgAnswer nlgAnswer : pattern.getAnswerList()) {
                        Set<String> options = answers.get(nlgAnswer.getParamName());
                        if (options == null) {
                            options = new HashSet<>();
                            answers.put(nlgAnswer.getParamName(), options);
                        }
                        if (nlgAnswer.getOptions() != null && !nlgAnswer.getOptions().isEmpty()) {
                            for (NlgAnswerOption nlgAnswerOption : nlgAnswer.getOptions()) {
                                options.add(nlgAnswerOption.getValue());
                            }
                        } else if (nlgAnswer.getSourceParam() != null) {
                            options.add(nlgAnswer.getSourceParam());
                        } else if (nlgAnswer.getType() == NlgAnswerType.freeInput) {
                            options.add(NlgAnswerType.freeInput.name());
                        }
                    }
                }
            }
            return result;
        } catch (ParseException e) {
            e.printStackTrace();
            return null;
        }
    }

    private boolean hasFinal(List<DMState> states, Set<DMState> visitedStates) {
        if (states == null || states.isEmpty()) {
            return false;
        } else {
            for (DMState state : states) {
                if (!visitedStates.contains(state)) {
                    visitedStates.add(state);
                    if (state.isFinal()) {
                        return true;
                    }
                    if (hasFinal(state.getNextStates(), visitedStates)) {
                        return true;
                    }
                    if (state.getDefaultNextStates() != null) {
                        for (DMState defState : state.getDefaultNextStates()) {
                            if (!visitedStates.contains(defState)) {
                                visitedStates.add(defState);
                                if (defState.isFinal()) {
                                    return true;
                                }
                            }
                        }
                    }
                }
            }
            return false;
        }
    }

    private void visitModel(List<DMState> states, Set<DMState> visitedStates, Set<DMState> invalidStates) {
        if (states != null && !states.isEmpty()) {
            for (DMState state : states) {
                if (!visitedStates.contains(state)) {
                    visitedStates.add(state);
                    List<DMState> children = children(state);
                    if (!children.isEmpty()) {
                        visitModel(children, visitedStates, invalidStates);
                    } else if (!state.isFinal()) {
                        invalidStates.add(state);
                    }
                }
            }
        }
    }

    private List<DMState> children(DMState state) {
        List<DMState> children = new LinkedList<>();
        if (state.getNextStates() != null) {
            children.addAll(state.getNextStates());
        }
        if (state.getDefaultNextStates() != null) {
            children.addAll(state.getDefaultNextStates());
        }
        return children;
    }

    private boolean areStatesValid(List<DMState> states, Set<DMState> visitedStates) {
        if (states == null || states.isEmpty()) {
            return true;
        } else {
            for (DMState state : states) {
                if (!visitedStates.contains(state)) {
                    visitedStates.add(state);
                    if (state.getRequestType() == null) {
                        System.err.println("Error: Invalid Request Type on stateID=" + state.getStateID());
                        return false;
                    } else if (!areParamsValid(state.getOutputParams())) {
                        System.err.println("Error: Invalid Output params on stateID=" + state.getStateID());
                        return false;
                    } else if (!areParamsValid(state.getInputParams())) {
                        System.err.println("Error: Invalid Input params on stateID=" + state.getStateID());
                        return false;
                    } else if (!areParamsValid(state.getOptionalParams())) {
                        System.err.println("Error: Invalid Optional params on stateID=" + state.getStateID());
                        return false;
                    } else if (!areNlgPatternsValid(state.getNlgPatterns())) {
                        System.err.println("Error: Invalid NLG Patterns on stateID=" + state.getStateID());
                        return false;
                    } else if (!areStatesValid(children(state), visitedStates)) {
                        return false;
                    }
                }
            }
            return true;
        }
    }

    private boolean areParamsValid(Set<Param> params) {
        if (params == null || params.isEmpty()) {
            return true;
        } else {
            for (Param param : params) {
                if (param.getValue() != null) {
                    return false;
                }
            }
            return true;
        }
    }

    private boolean areNlgPatternsValid(List<NlgExpression> nlgList) {
        if (nlgList == null || nlgList.isEmpty()) {
            return false;
        } else {
            for (NlgExpression nlgID : nlgList) {
                if (nlgModel.lookupByID(nlgID.getPattern()) == null) {
                    return false;
                }
            }
            return true;
        }
    }
}
