/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Version: 1.00
 * Date: 2014/7/25
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.service;

import eu.intuitel.comm.entities.lms.Tug;
import eu.intuitel.comm.entities.lpm.Reflex;
import eu.intuitel.comm.entities.qb.QueryBuilderOutput;

import javax.xml.bind.*;
import javax.xml.bind.util.ValidationEventCollector;
import java.io.*;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

public class MessageUtils {
    @SuppressWarnings("unchecked")
    public static <T> T parseMessage(Reader sourceCode, Class<T> messageClass, List<String> errors, String source) {
        try {
            JAXBContext context = JAXBContext.newInstance(messageClass);
            Unmarshaller unmarshaller = context.createUnmarshaller();
            ValidationEventCollector vec = new ValidationEventCollector();
            unmarshaller.setEventHandler(vec);
            T model = (T) unmarshaller.unmarshal(sourceCode);
            if (vec.hasEvents()) {
                for (ValidationEvent ve : vec.getEvents()) {
                    String msg = ve.getMessage();
                    ValidationEventLocator vel = ve.getLocator();
                    int line = vel.getLineNumber();
                    int column = vel.getColumnNumber();
                    errors.add("[" + source + "](line = " + line + ", column = " + column + "): " + msg);
                }
            }
            return model;
        } catch (JAXBException e) {
            errors.add("[" + source + "] Exception: " + e.getMessage());
            return null;
        }
    }

    public static List<Reflex> parseReflexMessage(File file, String mid, List<String> errors) {
        try {
            Reader input = new FileReader(file);
            Reflex reflex = parseMessage(input, Reflex.class, errors, file.getName());
            reflex.setmId(mid);
            List<Reflex> result = Collections.singletonList(reflex);
            input.close();
            return result;
        } catch (IOException e) {
            errors.add("[" + file.getName() + "] Exception: " + e.getMessage());
            return Collections.emptyList();
        }
    }

    public static List<Tug> parseTugMessage(File file, String mid, List<String> errors) {
        try {
            Reader input = new FileReader(file);
            Tug tug = parseMessage(input, Tug.class, errors, file.getName());
            tug.setMId(mid);
            List<Tug> result = Collections.singletonList(tug);
            input.close();
            return result;
        } catch (IOException e) {
            errors.add("[" + file.getName() + "] Exception: " + e.getMessage());
            return Collections.emptyList();
        }
    }

    public static List<QueryBuilderOutput> parseQueryBuilderOutput(File file, String mid, List<String> errors) {
        try {
            Reader input = new FileReader(file);
            QueryBuilderOutput qbo = parseMessage(input, QueryBuilderOutput.class, errors, file.getName());
            qbo.setmId(mid);
            List<QueryBuilderOutput> result = Collections.singletonList(qbo);
            input.close();
            return result;
        } catch (IOException e) {
            errors.add("[" + file.getName() + "] Exception: " + e.getMessage());
            return Collections.emptyList();
        }
    }

    public static String toXML(Object message) {
        return toXML(message, new LinkedList<String>(), "toXML");
    }

    public static String toXML(Object message, List<String> errors, String source) {
        try {
            JAXBContext context = JAXBContext.newInstance(message.getClass());
            Marshaller marshaller = context.createMarshaller();
            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
            ValidationEventCollector vec = new ValidationEventCollector();
            marshaller.setEventHandler(vec);

            StringWriter sw = new StringWriter();
            marshaller.marshal(message, sw);
            if (vec.hasEvents()) {
                for (ValidationEvent ve : vec.getEvents()) {
                    String msg = ve.getMessage();
                    ValidationEventLocator vel = ve.getLocator();
                    int line = vel.getLineNumber();
                    int column = vel.getColumnNumber();
                    errors.add("[" + source + "](line = " + line + ", column = " + column + "): " + msg);
                }
            }

            String result = sw.toString();
            sw.close();
            return result;
        } catch (JAXBException e) {
            errors.add("[" + source + "] Exception: " + e.getMessage());
            return null;
        } catch (IOException e) {
            errors.add("[" + source + "] Exception: " + e.getMessage());
            return null;
        }
    }
}
