/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Version: 1.00
 * Date: 2014/9/22
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.service;

import edu.uk.reading.communication.RecommendedRewriteService;
import edu.uk.reading.communication.connection.mockup.MockupConnectionListener;
import edu.uk.reading.communication.connection.mockup.MockupRRServiceConnection;
import edu.uk.reading.db.VariableFiller;
import edu.uk.reading.dm.model.ModelLoader;
import eu.intuitel.comm.entities.qb.QueryBuilderOutput;
import eu.intuitel.comm.entities.rr.Learner;
import eu.intuitel.comm.entities.rr.Lore;
import eu.intuitel.comm.entities.rr.Tug;
import eu.intuitel.comm.entities.rr.TugAck;
import eu.intuitel.config.ConfigurationLoader;
import eu.intuitel.types.Culture;
import eu.intuitel.types.DifficultyLevel;
import eu.intuitel.types.EqfLevel;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

import java.io.File;
import java.util.*;

import static edu.uk.reading.service.MessageUtils.parseQueryBuilderOutput;
import static edu.uk.reading.service.MessageUtils.parseTugMessage;
import static edu.uk.reading.tools.CommonUtils.*;
import static org.junit.Assert.*;

public class RRQbResultServiceTester implements MockupConnectionListener {
    private static final String TEST_USER = "test-User1";
    private static final String TEST_COURSE = "cc1";
    private static final String TEST_KO = "ko1";
    private static final String TEST_CC = "cc1";

    private static final String TEST_MACRO_LP = "#myMacroLp1";
    private static final String TEST_MICRO_LP = "#myMicroLp1";

    private static final Map<String, String> actionParams = new HashMap<>();

    static {
        actionParams.put("reply-ir0", VariableFiller.DB_PROP_DECREASE_DIFFICULTY);
        actionParams.put("ir1", VariableFiller.DB_PROP_DECREASE_DIFFICULTY);
    }

    private static final Map<String, List<Lore.LorePrio>> loreResults = new HashMap<>();

    static {
        loreResults
                .put("inform-ir0", tlist(lorePrio("90c82185-6732-4e21-4", 70), lorePrio("90c82185-6732-4e21-5", 70)));
        loreResults
                .put("inform-ir1", tlist(lorePrio("90c82185-6732-4e21-4", 70), lorePrio("90c82185-6732-4e21-5", 90)));
    }

    private static final File[][] INFORM_REPLY_FILES =
            ttuple(files("../src/test/DM/resources/mockup-messages/reasoner/queryBuilderOutput-scenario1.xml",
                         "../src/test/DM/resources/mockup-messages/reasoner/queryBuilderOutput-scenario1-reply.xml"),
                   files("../src/test/DM/resources/mockup-messages/reasoner/queryBuilderOutput-scenario2.xml"),
                   files("../src/test/DM/resources/mockup-messages/reasoner/queryBuilderOutput-scenario3.xml"));

    private static String rrName = RRQbResultServiceTester.class.getSimpleName();

//    private static final int TEST_WAIT_TIME = 15 * 1000;

    private static final int TEST_COUNTER = size(INFORM_REPLY_FILES);

    private static MockupRRServiceConnection communicationLayer = MockupRRServiceConnection.instance;
    private static VariableFiller variableFiller;

    private static final Object learnerUpdateLock = new Object();
    private static boolean learnerUpdateTestPerformed = false;

    @BeforeClass
    public static void setup() {
        Properties properties = RecommendedRewriteService.loadProperties();
        Assert.assertNotNull(properties);
        assertTrue(RecommendedRewriteService.loadRankingModel(properties));
        assertNotNull(ConfigurationLoader.properties);
        assertFalse(ConfigurationLoader.properties.isEmpty());
        ModelLoader.ParseResult parseResult = RecommendedRewriteService.loadFeedbackModel(properties);
        Assert.assertNotNull(parseResult);
        Assert.assertTrue(RecommendedRewriteService.startRR(communicationLayer, properties, parseResult));

        variableFiller = RecommendedRewriteService.getVariableFiller();
        communicationLayer.setMockupConnectionListener(new RRQbResultServiceTester());

        System.out.println("Test list size = " + TEST_COUNTER);
        mockupLearner();
    }

    private static void mockupLearner() {
        variableFiller.dbAccessor.storeLearner(TEST_USER, TEST_USER, null, 22, Culture.Unknown, false, false, false);
        variableFiller.dbAccessor.loginUser(TEST_USER, TEST_COURSE);
        variableFiller.dbAccessor.storeLearnerPreference(TEST_USER, TEST_COURSE,
                                                         EqfLevel.Unknown,
                                                         DifficultyLevel.Unknown);

        variableFiller.dbAccessor.storeCcCompletion(TEST_USER, TEST_COURSE, TEST_CC, false);
        variableFiller.dbAccessor.storeKoCompletionAccessed(TEST_USER, TEST_COURSE, TEST_KO, true);

        variableFiller.dbAccessor.storeMacroLpSelection(TEST_USER, TEST_MACRO_LP, TEST_COURSE);
        variableFiller.dbAccessor.storeMicroLpSelection(TEST_USER, TEST_MICRO_LP, TEST_COURSE);
    }

    @Test
    public void informWithReplyTest() {
        List<String> errors = new LinkedList<>();

        for (int i = 0; i < INFORM_REPLY_FILES.length; i++) {
            File[] files = INFORM_REPLY_FILES[i];
            synchronized (learnerUpdateLock) {
                learnerUpdateTestPerformed = false;
                List<QueryBuilderOutput> queryBuilderOutputs = parseQueryBuilderOutput(files[0], "inform-ir" + i,
                                                                                       errors);
                assertFalse(printErrors(errors));
                communicationLayer.onNewQueryBuilderOutput(queryBuilderOutputs);
                errors.clear();

                System.out.println("Waiting for the Learner Update result to perform ...");
                try {
//                    learnerUpdateLock.wait(TEST_WAIT_TIME);
                    learnerUpdateLock.wait();
                } catch (InterruptedException e) {
                    //-- ignore
                }
                assertTrue("Learner Result not performed ! Test file=" + files[0], learnerUpdateTestPerformed);

                if (files.length > 1) {
                    learnerUpdateTestPerformed = false;
                    List<eu.intuitel.comm.entities.lms.Tug> tugs = parseTugMessage(files[1], "reply-ir" + i, errors);
                    assertFalse(printErrors(errors));
                    communicationLayer.onNewTugs(tugs);
                    errors.clear();

                    System.out.println("Waiting for Learner test to perform ...");
                    try {
//                        learnerUpdateLock.wait(TEST_WAIT_TIME);
                        learnerUpdateLock.wait();
                    } catch (InterruptedException e) {
                        //ignore
                    }
                    assertTrue("Learner Test not performed ! Test file=" + files[1], learnerUpdateTestPerformed);
                }
            }
        }
    }

    public void loreUpdated(Lore lore) {
        System.out.println("LORE Update: " + lore.getMId());
        Assert.assertNotNull(lore);
        List<String> errors = new LinkedList<>();
        MessageUtils.toXML(lore, errors, rrName);
        printErrors(errors);
        assertTrue(errors.isEmpty());
        testLorePriorities(lore.getMId(), lore.getLorePrio());
    }

    public void tugUpdated(Tug tug) {
        System.out.println("TUG Update: " + tug.getMId());
        Assert.assertNotNull(tug);
        List<String> errors = new LinkedList<>();
        MessageUtils.toXML(tug, errors, rrName);
        printErrors(errors);
        assertTrue(errors.isEmpty());
    }

    public void learnerUpdated(Learner learner) {
        System.out.println("Learner Update: " + learner.getMId());
        Assert.assertNotNull(learner);
        List<String> errors = new LinkedList<>();
        MessageUtils.toXML(learner, errors, rrName);
        printErrors(errors);
        assertTrue(errors.isEmpty());
        testInformActions(learner.getMId());

        if (learner.getLore() != null) {
            loreUpdated(learner.getLore());
        }

        if (learner.getTug() != null) {
            for (Tug tug : learner.getTug()) {
                tugUpdated(tug);
            }
        }

        fireLearnerTestPerformed();
    }

    public void tugAckUpdated(TugAck tug) {
    }

    private static void testLorePriorities(String testID, List<Lore.LorePrio> lorePriorities) {
        List<Lore.LorePrio> testPriorities = loreResults.get(testID);
        if (testPriorities != null) {
            assertTrue("Lore Priorities are not equal for the test=" + testID + " . Expected=" + toString(testPriorities) +
                       " got=" + toString(lorePriorities), equalLists(lorePriorities, testPriorities));
        } else {
            fail("No test running for LORE template:" + testID);
        }
    }

    private static boolean equalLists(List<Lore.LorePrio> l1, List<Lore.LorePrio> l2) {
        return hasEmptyDiff(l1, l2) && hasEmptyDiff(l2, l1);
    }

    private static boolean hasEmptyDiff(List<Lore.LorePrio> l1, List<Lore.LorePrio> l2) {
        List<Lore.LorePrio> tempCollection = new LinkedList<>(l1);
        for (Lore.LorePrio item : l2) {
            Iterator<Lore.LorePrio> iterator = tempCollection.iterator();
            while (iterator.hasNext()) {
                Lore.LorePrio element = iterator.next();
                if (item.getLoId().equals(element.getLoId()) && item.getValue() == element.getValue()) {
                    iterator.remove();
                }
            }
        }
        return tempCollection.isEmpty();
    }

    private static void testInformActions(String actionID) {
        String params = actionParams.get(actionID);
        if (params != null) {
            testAction(params);
        }
    }

    private static void testAction(String dbParam) {
        if (!dbParam.trim().equals("")) {
            boolean solved = false;
            switch (dbParam) {
                case VariableFiller.DB_PROP_DELETE_LP:
                    Assert.assertNull(variableFiller.dbAccessor.getMacroLpOfLearner(TEST_USER, TEST_COURSE));
                    Assert.assertNull(variableFiller.dbAccessor.getMicroLpOfLearner(TEST_USER, TEST_COURSE));
                    solved = true;
                    break;
                case VariableFiller.DB_PROP_COMPLETE_ALL_KO:
                    assertEquals(variableFiller.dbAccessor.getKoCompletionStatus(TEST_USER, TEST_KO).intValue(), 100);
                    solved = true;
                    break;
                case VariableFiller.DB_PROP_RESTART_CURRENT_CC:
                    assertEquals(variableFiller.dbAccessor.getKoCompletionStatus(TEST_USER, TEST_KO).intValue(), 0);
                    solved = true;
                    break;

                case VariableFiller.DB_PROP_INCREASE_DIFFICULTY:
                    assertEquals(variableFiller.dbAccessor.getPreferredDifficultyLevel(TEST_USER, TEST_COURSE),
                                 DifficultyLevel.Intermediate);
                    solved = true;
                    break;
                case VariableFiller.DB_PROP_DECREASE_DIFFICULTY:
                    assertEquals(variableFiller.dbAccessor.getPreferredDifficultyLevel(TEST_USER, TEST_COURSE),
                                 DifficultyLevel.Intermediate);
                    solved = true;
                    break;
            }
            Assert.assertTrue("No test has been conducted for this action!", solved);
        }
    }

    private static boolean printErrors(List<String> errors) {
        if (!errors.isEmpty()) {
            for (String error : errors) {
                System.err.println(error);
            }
            return true;
        }
        return false;
    }

    private static void fireLearnerTestPerformed() {
        synchronized (learnerUpdateLock) {
            learnerUpdateTestPerformed = true;
            System.out.println("Learner Test performed. Notifying locks ...");
            learnerUpdateLock.notifyAll();
        }
    }

    private static Lore.LorePrio lorePrio(String iri, int priority) {
        Lore.LorePrio result = new Lore.LorePrio();
        result.setLoId(iri);
        result.setValue(priority);
        return result;
    }

    private static String toString(List<Lore.LorePrio> prios) {
        StringBuilder sb = new StringBuilder();
        for (Lore.LorePrio prio : prios) {
            if (sb.length() > 0) {
                sb.append(",");
            }
            sb.append("(").append(prio.getLoId()).append(",").append(prio.getValue()).append(")");
        }
        return "[" + sb.toString() + "]";
    }
}
