/*
 * Creator: INTUITEL Consortium, www.intuitel.eu
 * Contributor: Ovidiu Serban, ISR Laboratory, University of Reading, UK, http://www.isr.reading.ac.uk/
 * Version: 1.00
 * Date: 2014/7/25
 * Copyright: Copyright (C) INTUITEL Consortium
 * License: The license that applies to this file has not yet been determined
 */

package edu.uk.reading.service;

import edu.uk.reading.communication.RecommendedRewriteService;
import edu.uk.reading.communication.connection.mockup.MockupConnectionListener;
import edu.uk.reading.communication.connection.mockup.MockupRRServiceConnection;
import edu.uk.reading.db.VariableFiller;
import edu.uk.reading.dm.model.ModelLoader;
import eu.intuitel.comm.entities.lpm.Reflex;
import eu.intuitel.comm.entities.rr.Learner;
import eu.intuitel.comm.entities.rr.Lore;
import eu.intuitel.comm.entities.rr.Tug;
import eu.intuitel.comm.entities.rr.TugAck;
import eu.intuitel.config.ConfigurationLoader;
import eu.intuitel.types.Culture;
import eu.intuitel.types.DifficultyLevel;
import eu.intuitel.types.EqfLevel;
import eu.intuitel.types.Gender;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

import java.io.File;
import java.util.*;

import static edu.uk.reading.service.MessageUtils.parseReflexMessage;
import static edu.uk.reading.service.MessageUtils.parseTugMessage;
import static edu.uk.reading.tools.CommonUtils.*;
import static org.junit.Assert.*;

public class RRReflexServiceTester implements MockupConnectionListener {
    private static final String TEST_USER = "test-User1";
    private static final String TEST_COURSE = "cc1";
    private static final String TEST_KO = "ko1";
    private static final String TEST_CC = "cc1";

    private static final String TEST_MACRO_LP = "#myMacroLp1";
    private static final String TEST_MICRO_LP = "#myMicroLp1";

    private static final Map<String, String[]> reflexParams = new HashMap<>();

    static {
        reflexParams.put("r-r0", ttuple(VariableFiller.DB_PROP_CURRENT_MICRO_LP, TEST_MICRO_LP));
        reflexParams.put("r-r1", ttuple(VariableFiller.DB_PROP_CURRENT_MACRO_LP, TEST_MACRO_LP));
        reflexParams.put("r-r2", ttuple(VariableFiller.DB_PROP_CURRENT_AGE, "23"));
        reflexParams.put("r-r3", ttuple(VariableFiller.DB_PROP_CURRENT_GENDER, "m"));
        reflexParams.put("r-r4", ttuple(VariableFiller.DB_PROP_CURRENT_DEAF_STATE, "false"));
        reflexParams.put("r-r5", ttuple(VariableFiller.DB_PROP_CURRENT_DIFFICULTY_LVL, "1"));
    }

    private static final File[][] REFLEX_FILES =
            ttuple(files("../src/test/DM/resources/mockup-messages/reflex/reflex-microLP.xml",
                         "../src/test/DM/resources/mockup-messages/reflex/reflex-microLP-reply.xml"),
                   files("../src/test/DM/resources/mockup-messages/reflex/reflex-macroLP.xml",
                         "../src/test/DM/resources/mockup-messages/reflex/reflex-macroLP-reply.xml"),
                   files("../src/test/DM/resources/mockup-messages/reflex/reflex-age.xml",
                         "../src/test/DM/resources/mockup-messages/reflex/reflex-age-reply.xml"),
                   files("../src/test/DM/resources/mockup-messages/reflex/reflex-gender.xml",
                         "../src/test/DM/resources/mockup-messages/reflex/reflex-gender-reply.xml"),
                   files("../src/test/DM/resources/mockup-messages/reflex/reflex-deaf.xml",
                         "../src/test/DM/resources/mockup-messages/reflex/reflex-deaf-reply.xml"),
                   files("../src/test/DM/resources/mockup-messages/reflex/reflex-difficulty.xml",
                         "../src/test/DM/resources/mockup-messages/reflex/reflex-difficulty-reply.xml"));

    private static String rrName = RRReflexServiceTester.class.getSimpleName();

    private static final int TEST_WAIT_TIME = 15 * 1000;

    private static final int TEST_COUNTER = size(REFLEX_FILES);

    private static MockupRRServiceConnection communicationLayer = MockupRRServiceConnection.instance;
    private static VariableFiller variableFiller;

    private static final Object learnerUpdateLock = new Object();
    private static boolean learnerUpdateTestPerformed = false;

    @BeforeClass
    public static void setup() {
        Properties properties = RecommendedRewriteService.loadProperties();
        Assert.assertNotNull(properties);
        assertTrue(RecommendedRewriteService.loadRankingModel(properties));
        assertNotNull(ConfigurationLoader.properties);
        assertFalse(ConfigurationLoader.properties.isEmpty());
        ModelLoader.ParseResult parseResult = RecommendedRewriteService.loadFeedbackModel(properties);
        Assert.assertNotNull(parseResult);
        Assert.assertTrue(RecommendedRewriteService.startRR(communicationLayer, properties, parseResult));

        variableFiller = RecommendedRewriteService.getVariableFiller();
        communicationLayer.setMockupConnectionListener(new RRReflexServiceTester());

        System.out.println("Test list size = " + TEST_COUNTER);
        mockupLearner();
    }

    private static void mockupLearner() {
        variableFiller.dbAccessor.storeLearner(TEST_USER, TEST_USER, null, 22, Culture.Unknown, false, false, false);
        variableFiller.dbAccessor.loginUser(TEST_USER, TEST_COURSE);
        variableFiller.dbAccessor.storeLearnerPreference(TEST_USER, TEST_COURSE,
                                                         EqfLevel.Unknown,
                                                         DifficultyLevel.Unknown);

        variableFiller.dbAccessor.storeCcCompletion(TEST_USER, TEST_COURSE, TEST_CC, false);
        variableFiller.dbAccessor.storeKoCompletionAccessed(TEST_USER, TEST_COURSE, TEST_KO, true);

        variableFiller.dbAccessor.storeMacroLpSelection(TEST_USER, TEST_MACRO_LP, TEST_COURSE);
        variableFiller.dbAccessor.storeMicroLpSelection(TEST_USER, TEST_MICRO_LP, TEST_COURSE);
    }

    @Test
    public void reflexTest() {
        List<String> errors = new LinkedList<>();

        for (int i = 0; i < REFLEX_FILES.length; i++) {
            File[] files = REFLEX_FILES[i];
            synchronized (learnerUpdateLock) {
                learnerUpdateTestPerformed = false;
                List<Reflex> reflexes = parseReflexMessage(files[0], "r" + i, errors);
                assertFalse(printErrors(errors));
                communicationLayer.onNewReflexes(reflexes);
                errors.clear();

                System.out.println("Waiting for Learner test to perform ...");
                try {
                    learnerUpdateLock.wait(TEST_WAIT_TIME);
                } catch (InterruptedException e) {
                    //ignore
                }

                assertTrue("Learner Test not performed ! Test file=" + files[0], learnerUpdateTestPerformed);

                if (files.length > 1) {
                    learnerUpdateTestPerformed = false;
                    List<eu.intuitel.comm.entities.lms.Tug> tugs = parseTugMessage(files[1], "r-r" + i, errors);
                    assertFalse(printErrors(errors));
                    communicationLayer.onNewTugs(tugs);
                    errors.clear();

                    System.out.println("Waiting for Learner test to perform ...");
                    try {
                        learnerUpdateLock.wait(TEST_WAIT_TIME);
                    } catch (InterruptedException e) {
                        //ignore
                    }

                    assertTrue("Learner Test not performed ! Test file=" + files[1], learnerUpdateTestPerformed);
                }
            }
        }
    }

    public void learnerUpdated(Learner learner) {
        System.out.println("Learner Update: " + learner.getMId());
        Assert.assertNotNull(learner);
        List<String> errors = new LinkedList<>();
        MessageUtils.toXML(learner, errors, rrName);
        printErrors(errors);
        assertTrue(errors.isEmpty());
        testReflexSetter(learner.getMId());
        fireLearnerTestPerformed();
    }

    public void tugAckUpdated(TugAck tug) {
        System.out.println("TugAck: " + tug.getMId());

        testReflexSetter(tug.getMId());
        fireLearnerTestPerformed();
    }

    private static void testReflexSetter(String actionID) {
        String[] params = reflexParams.get(actionID);
        if (params != null) {
            testSetParam(params[0], params[1]);
        }
    }

    private static void testSetParam(String dbParam, String param) {
        if (!dbParam.trim().equals("")) {
            boolean solved = false;
            switch (dbParam) {
                case VariableFiller.DB_PROP_CURRENT_MACRO_LP:
                    assertEquals(variableFiller.dbAccessor.getMacroLpOfLearner(TEST_USER, TEST_COURSE),
                                 param);
                    solved = true;
                    break;
                case VariableFiller.DB_PROP_CURRENT_MICRO_LP:
                    assertEquals(variableFiller.dbAccessor.getMicroLpOfLearner(TEST_USER, TEST_COURSE),
                                 param);
                    solved = true;
                    break;
                case VariableFiller.DB_PROP_CURRENT_AGE:
                    assertEquals(variableFiller.dbAccessor.getLearnerAge(TEST_USER), Integer.valueOf(param));
                    solved = true;
                    break;
                case VariableFiller.DB_PROP_CURRENT_GENDER:
                    assertEquals(variableFiller.dbAccessor.getLearnerGender(TEST_USER), Gender.getGender(param));
                    solved = true;
                    break;
                case VariableFiller.DB_PROP_CURRENT_DEAF_STATE:
                    assertEquals(variableFiller.dbAccessor.isLearnerDeaf(TEST_USER), Boolean.getBoolean(param));
                    solved = true;
                    break;
                case VariableFiller.DB_PROP_CURRENT_DIFFICULTY_LVL:
                    assertEquals(variableFiller.dbAccessor.getPreferredDifficultyLevel(TEST_USER, TEST_COURSE),
                                 DifficultyLevel.getLevel(Integer.parseInt(param)));
                    solved = true;
                    break;
            }
            Assert.assertTrue("No test has been conducted for this action!", solved);
        }
    }

    private static boolean printErrors(List<String> errors) {
        if (!errors.isEmpty()) {
            for (String error : errors) {
                System.err.println(error);
            }
            return true;
        }
        return false;
    }

    private static void fireLearnerTestPerformed() {
        synchronized (learnerUpdateLock) {
            learnerUpdateTestPerformed = true;
            System.out.println("Learner Test performed. Notifying locks ...");
            learnerUpdateLock.notifyAll();
        }
    }
}
