/*
 * Copyright (C) 2013, FZI Forschungszentrum Informatik
 *
 * Modifications to the initial code base are copyright of their
 * respective authors, or their employers as appropriate. Authorship
 * of the modifications may be determined from the ChangeLog placed at
 * the end of this file.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package de.fzi.herakles.analyser.impl;

import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owlapi.model.OWLOntology;

import de.fzi.herakles.analyser.Analyser;
import de.fzi.herakles.registry.ServerRegistry;
import de.fzi.owllink.reasoner.server.entities.ReasonerType;

/**
 * Abstract Analyser class that provides the {@link OWLOntology} and the
 * {@link ServerRegistry} package private get-methods for the
 * {@link AnalyserFacotry} class.
 * 
 * @author Marcel Weigel
 * 
 */
public abstract class AnalyserAbst implements Analyser {

	private final static Log LOG = LogFactory.getLog(AnalyserAbst.class);

	protected OWLOntology ontology;

	protected ServerRegistry serverRegistry;
	
	private Set<String> excludedReasonerTypeIds;

	AnalyserAbst(OWLOntology ontology, ServerRegistry serverRegistry, Set<String> excludedReasonerTypeIds) {
		guard_AnalyserAbst(ontology, serverRegistry, excludedReasonerTypeIds);

		this.ontology = ontology;
		this.serverRegistry = serverRegistry;
		this.excludedReasonerTypeIds = excludedReasonerTypeIds;
	}

	private void guard_AnalyserAbst( OWLOntology ontology, ServerRegistry serverRegistry , Set<String> excludedReasonerTypeIds) {
		if (ontology == null) {
			final String errStr = "OWLOntology is null.";
			LOG.fatal(errStr);
			throw new NullPointerException(errStr);
		}
		if (serverRegistry == null) {
			final String errStr = "ServerRegistry is null.";
			LOG.fatal(errStr);
			throw new NullPointerException(errStr);
		}
		if (excludedReasonerTypeIds == null) {
			final String errStr = "Excluded ReasonerTypeIds set is null.";
			LOG.fatal(errStr);
			throw new NullPointerException(errStr);
		}
		if (serverRegistry.isEmpty()) {
			LOG.warn("ServerRegistry is empty.");
		}
		if (ontology.isEmpty()) {
			LOG.info("OWLOntology is empty.");
		}
	}

	OWLOntology getOntology() {
		return ontology;
	}

	ServerRegistry getServerRegistry() {
		return serverRegistry;
	}
	
	protected boolean isReasonerTypeExcluded(ReasonerType reasonerType) {
		for(String reasonerTypeId : this.excludedReasonerTypeIds) {
			if(reasonerTypeId.equals(reasonerType.getId())) {
				return true;
			}
		}
		return false;
	}

	public abstract Set<ReasonerType> getSuitableReasonerTypes();

}
