/*
 * Copyright (C) 2013, FZI Forschungszentrum Informatik
 *
 * Modifications to the initial code base are copyright of their
 * respective authors, or their employers as appropriate. Authorship
 * of the modifications may be determined from the ChangeLog placed at
 * the end of this file.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package de.fzi.herakles.job;

import java.util.List;

import de.fzi.herakles.job.impl.ExecutionState;
import de.fzi.herakles.job.impl.JobState;
import de.fzi.herakles.job.request.Request;
import de.fzi.herakles.job.response.Response;
import de.fzi.herakles.registry.ReasonerAdapter;

/**
 * 
 * The Job interface contains all functionality to managed requests to multiple {@link ReasonerAdapter}.
 * It provides to create new requests, check the condition and query the results.
 * 
 * @author Marcel Weigel
 * 
 */
public interface Job<T> {

	/**
	 * Get the {@link Request} from the job.
	 * 
	 * @return Request
	 */
	Request getRequest();

	/**
	 * Get the {@link JobState} from the job.
	 * 
	 * @return JobState
	 */
	JobState getJobState();

	/**
	 * Execute a job with a given {@link ReasonerAdpater}.
	 * 
	 * @param reasonerAdapter ReasonerAdapter
	 */
	void execute( ReasonerAdapter reasonerAdapter );

	/**
	 * Check if the job has finished responses.
	 * 
	 * @return True if one response has finished.
	 */
	boolean hasResponses();
	
	/**
	 * At least one response have finished successfully.
	 * 
	 * @return True if one response have finished successfully.
	 */
	boolean successfulResponseAvailable();
	
	/**
	 * Check if all responses have finished.
	 * 
	 * @return True if all responses have finished.
	 */
	boolean allResponsesAvailable();

	/**
	 * All responses have failed.
	 * 
	 * @return True if all responses have failed.
	 */
	boolean allResponsesFailed();

	/**
	 * Get a list of all responses. Sorted by response times.
	 * 
	 * @return List of Responses.
	 */
	List<Response<T>> getResponses();

	/**
	 * Get a list of all successfully completed responses. Sorted by response time.
	 * 
	 * @return List of successfully responses.
	 */
	List<Response<T>> getSuccessfullResponses();
	
	/**
	 * Get the associated {@link Response} to {@link ReasonerAdapter}.
	 * 
	 * @param reasonerAdapter ReasonerAdapter
	 * @return Response, returns null if no request has finished.
	 */
	Response<T> getResponse( ReasonerAdapter reasonerAdapter );
	
	/**
	 * Get the {@link ExecutionState} from the given {@link ReasonerAdapter}.
	 * 
	 * @param reasonerAdapter ReasonerAdapter
	 * @return ExecutionState, return null if reasonerAdapter was not found.
	 */
	ExecutionState getExcutionState( ReasonerAdapter reasonerAdapter );

	/**
	 * Create a new successfully completed {@link Response} with a given {@link ReasonerAdapter} and a result of generic type.
	 * 
	 * @param reasonerAdapter ReasonerAdapter
	 * @param result Result Value
	 */
	void createResponse( ReasonerAdapter reasonerAdapter, T result );

	/**
	 * Create a new failed {@link Response} with a given {@link ReasonerAdapter} and a exception
	 * 
	 * @param reasonerAdapter ReasonerAdapter
	 * @param exception Throwable
	 */
	void createResponse( ReasonerAdapter reasonerAdapter, Throwable exception );
	
	/**
	 * Get the number of successfully finished responses.
	 * 
	 * @return Number of successfully responses.
	 */
	int getResponsesSuccessfullyFinishedCount();
	
	/**
	 * Get the number of failed finished responses.
	 * 
	 * @return Number of failed responses.
	 */
	int getResponsesFailedFinishedCount();

}
