/*
 * Copyright (C) 2013, FZI Forschungszentrum Informatik
 *
 * Modifications to the initial code base are copyright of their
 * respective authors, or their employers as appropriate. Authorship
 * of the modifications may be determined from the ChangeLog placed at
 * the end of this file.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package de.fzi.herakles.job.impl;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import de.fzi.herakles.job.Job;
import de.fzi.herakles.job.request.Request;
import de.fzi.herakles.job.request.impl.RequestFactory;
import de.fzi.herakles.job.response.Response;
import de.fzi.herakles.job.response.impl.ResponseFactory;
import de.fzi.herakles.registry.ReasonerAdapter;

/**
 * 
 * @author Marcel Weigel
 * 
 */
class JobImpl<T> implements Job<T> {
	
	private final static Log LOG = LogFactory.getLog(JobImpl.class);

	private JobListener jobListener;

	private JobState jobState;

	private Map<ReasonerAdapter, ExecutionState> reasoerAdapterExecutionStateMap;

	private Request request;

	private List<Response<T>> responses;

	JobImpl(JobListener jobListener, Method method, Object[] param) {
		guard_JobImpl(jobListener, method, param);
		
		this.jobState = JobState.NEW;

		this.jobListener = jobListener;
		this.responses = new ArrayList<Response<T>>();
		this.reasoerAdapterExecutionStateMap = new HashMap<ReasonerAdapter, ExecutionState>();

		this.request = RequestFactory.getInstance().getRequest(method, param);
	}
	
	private void guard_JobImpl( JobListener jobListener, Method method, Object[] param ) {
		if (jobListener == null) {
			final String errStr = "JobListener is null.";
			LOG.fatal(errStr);
			throw new NullPointerException(errStr);
		}
		if (method == null) {
			final String errStr = "Method is null.";
			LOG.fatal(errStr);
			throw new NullPointerException(errStr);
		}
	}

	public void execute( ReasonerAdapter reasonerAdapter ) {
		if (hasNoSubjobStarted()) {
			this.jobState = JobState.PROCESSING;
			this.jobListener.executeCalled(this);
		}
		this.reasoerAdapterExecutionStateMap.put(reasonerAdapter, ExecutionState.PROCESSING);

		reasonerAdapter.invoke(this);
	}

	public List<Response<T>> getResponses() {
		return responses;
	}

	public List<Response<T>> getSuccessfullResponses() {
		List<Response<T>> responses = new ArrayList<Response<T>>();
		for (Response<T> response : this.responses) {
			if (!response.isFailed()) {
				responses.add(response);
			}
		}
		return responses;
	}

	public Response<T> getResponse( ReasonerAdapter reasonerAdapter ) {
		for (Response<T> response : this.responses) {
			if (response.getSource().equals(reasonerAdapter)) {
				return response;
			}
		}
		return null;
	}

	public boolean hasResponses() {
		return !this.responses.isEmpty();
	}

	public boolean successfulResponseAvailable() {
		return this.jobState == JobState.PROCESSING_SOME_SUCCESS 
				|| this.jobState == JobState.PROCESSING_SOME_SUCCESS_SOME_FAILED 
				|| this.jobState == JobState.COMPLETED_ALL_SUCCESS;
	}

	public boolean allResponsesAvailable() {
		return this.jobState == JobState.COMPLETED_SOME_FAILED
				|| this.jobState == JobState.COMPLETED_ALL_SUCCESS
				|| this.jobState == JobState.COMPLETED_ALL_FAILED;
	}

	public boolean allResponsesFailed() {
		return this.jobState == JobState.COMPLETED_ALL_FAILED;
	}

	public Request getRequest() {
		return request;
	}

	public JobState getJobState() {
		return jobState;
	}

	public ExecutionState getExcutionState( ReasonerAdapter reasonerAdapter ) {
		if (this.reasoerAdapterExecutionStateMap.containsKey(reasonerAdapter)) {
			return this.reasoerAdapterExecutionStateMap.get(reasonerAdapter);
		}
		return null;
	}

	public synchronized void createResponse( ReasonerAdapter reasonerAdapter, T result ) {
		this.reasoerAdapterExecutionStateMap.put(reasonerAdapter, ExecutionState.SUCCESS);
		subjobFinishedSuccess();

		Response<T> response = ResponseFactory.getInstance().getResponse(reasonerAdapter, result);

		this.responses.add(response);
	}

	public synchronized void createResponse( ReasonerAdapter reasonerAdapter, Throwable exception ) {
		this.reasoerAdapterExecutionStateMap.put(reasonerAdapter, ExecutionState.FAILED);
		subjobFinishedFailed();

		Response<T> response = ResponseFactory.getInstance().getResponse(reasonerAdapter, exception);

		this.responses.add(response);
	}

	private boolean hasNoSubjobStarted() {
		return this.reasoerAdapterExecutionStateMap.isEmpty();
	}

	private boolean haveAllSubjobsFinished() {
		for (ExecutionState executionState : this.reasoerAdapterExecutionStateMap.values()) {
			if (executionState == ExecutionState.PROCESSING) {
				return false;
			}
		}
		return true;
	}

	private void subjobFinishedSuccess() {
		if (haveAllSubjobsFinished()) {
			if (this.jobState == JobState.PROCESSING_SOME_SUCCESS) {
				this.jobState = JobState.COMPLETED_ALL_SUCCESS;
			} else if (this.jobState == JobState.PROCESSING_SOME_FAILED) {
				this.jobState = JobState.COMPLETED_SOME_FAILED;
			} else if (this.jobState == JobState.PROCESSING_SOME_SUCCESS_SOME_FAILED) {
				this.jobState = JobState.COMPLETED_SOME_FAILED;
			} else if (this.jobState == JobState.PROCESSING) {
				this.jobState = JobState.COMPLETED_ALL_SUCCESS;
			}
		} else {
			if (this.jobState == JobState.PROCESSING_SOME_SUCCESS) {
				this.jobState = JobState.PROCESSING_SOME_SUCCESS;
			} else if (this.jobState == JobState.PROCESSING_SOME_FAILED) {
				this.jobState = JobState.PROCESSING_SOME_SUCCESS_SOME_FAILED;
			} else if (this.jobState == JobState.PROCESSING_SOME_SUCCESS_SOME_FAILED) {
				this.jobState = JobState.PROCESSING_SOME_SUCCESS_SOME_FAILED;
			} else if (this.jobState == JobState.PROCESSING) {
				this.jobState = JobState.PROCESSING_SOME_SUCCESS;
			}
		}
	}

	private void subjobFinishedFailed() {
		if (haveAllSubjobsFinished()) {
			if (this.jobState == JobState.PROCESSING_SOME_SUCCESS) {
				this.jobState = JobState.COMPLETED_SOME_FAILED;
			} else if (this.jobState == JobState.PROCESSING_SOME_FAILED) {
				this.jobState = JobState.COMPLETED_ALL_FAILED;
			} else if (this.jobState == JobState.PROCESSING_SOME_SUCCESS_SOME_FAILED) {
				this.jobState = JobState.COMPLETED_SOME_FAILED;
			} else if (this.jobState == JobState.PROCESSING) {
				this.jobState = JobState.COMPLETED_ALL_FAILED;
			}
		} else {
			if (this.jobState == JobState.PROCESSING_SOME_SUCCESS) {
				this.jobState = JobState.PROCESSING_SOME_SUCCESS_SOME_FAILED;
			} else if (this.jobState == JobState.PROCESSING_SOME_FAILED) {
				this.jobState = JobState.PROCESSING_SOME_FAILED;
			} else if (this.jobState == JobState.PROCESSING_SOME_SUCCESS_SOME_FAILED) {
				this.jobState = JobState.PROCESSING_SOME_SUCCESS_SOME_FAILED;
			} else if (this.jobState == JobState.PROCESSING) {
				this.jobState = JobState.PROCESSING_SOME_FAILED;
			}
		}
	}

	public int getResponsesSuccessfullyFinishedCount() {
		int successfullyFinishedCount = 0;
		for (ExecutionState executionState : this.reasoerAdapterExecutionStateMap.values()) {
			if (executionState == ExecutionState.SUCCESS) {
				successfullyFinishedCount++;
			}
		}
		return successfullyFinishedCount;
	}

	public int getResponsesFailedFinishedCount() {
		int failedFinishedCount = 0;
		for (ExecutionState executionState : this.reasoerAdapterExecutionStateMap.values()) {
			if (executionState == ExecutionState.FAILED) {
				failedFinishedCount++;
			}
		}
		return failedFinishedCount;
	}

	@SuppressWarnings("unchecked")
	@Override
	public boolean equals( Object obj ) {
		if (obj == null)
			return false;
		if (obj == this)
			return true;
		if (!(obj instanceof JobImpl<?>))
			return false;

		JobImpl<T> job = (JobImpl<T>) obj;
		return (this.request == job.request || this.request != null && this.request.equals(job.request))
				&& (this.responses == job.responses || this.responses != null && this.responses.equals(job.responses))
				&& this.jobState == job.jobState
				&& (this.reasoerAdapterExecutionStateMap == job.reasoerAdapterExecutionStateMap || this.reasoerAdapterExecutionStateMap != null
						&& this.reasoerAdapterExecutionStateMap.equals(job.reasoerAdapterExecutionStateMap))
				&& (this.jobListener == job.jobListener || this.jobListener != null && this.jobListener.equals(job.jobListener));
	}

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 31 * hash + (this.request == null ? 0 : this.request.hashCode());
		hash = 31 * hash + (this.responses == null ? 0 : this.responses.hashCode());
		hash = 31 * hash + (this.jobState == null ? 0 : this.jobState.hashCode());
		hash = 31 * hash + (this.reasoerAdapterExecutionStateMap == null ? 0 : this.reasoerAdapterExecutionStateMap.hashCode());
		hash = 31 * hash + (this.jobListener == null ? 0 : this.jobListener.hashCode());
		return hash;
	}

	@Override
	public String toString() {
		String[] classPath = JobImpl.class.getName().split("\\.");
		return String.format("%s(%s, jobState=%s)", classPath[classPath.length-1], this.request, this.jobState);
	}
	
	
}
