/*
 * Copyright (C) 2013, FZI Forschungszentrum Informatik
 *
 * Modifications to the initial code base are copyright of their
 * respective authors, or their employers as appropriate. Authorship
 * of the modifications may be determined from the ChangeLog placed at
 * the end of this file.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package de.fzi.herakles.reasoner;

import java.lang.invoke.WrongMethodTypeException;
import java.lang.reflect.Method;
import java.net.URL;
import java.nio.file.FileSystemNotFoundException;
import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.TimeoutException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owlapi.model.AxiomType;
import org.semanticweb.owlapi.model.OWLAxiom;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLDataProperty;
import org.semanticweb.owlapi.model.OWLDataPropertyExpression;
import org.semanticweb.owlapi.model.OWLLiteral;
import org.semanticweb.owlapi.model.OWLNamedIndividual;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.owllink.OWLlinkReasoner;
import org.semanticweb.owlapi.owllink.Request;
import org.semanticweb.owlapi.owllink.Response;
import org.semanticweb.owlapi.owllink.builtin.response.OWLlinkErrorResponseException;
import org.semanticweb.owlapi.owllink.builtin.response.ResponseMessage;
import org.semanticweb.owlapi.reasoner.AxiomNotInProfileException;
import org.semanticweb.owlapi.reasoner.BufferingMode;
import org.semanticweb.owlapi.reasoner.ClassExpressionNotInProfileException;
import org.semanticweb.owlapi.reasoner.FreshEntitiesException;
import org.semanticweb.owlapi.reasoner.InconsistentOntologyException;
import org.semanticweb.owlapi.reasoner.InferenceType;
import org.semanticweb.owlapi.reasoner.Node;
import org.semanticweb.owlapi.reasoner.NodeSet;
import org.semanticweb.owlapi.reasoner.ReasonerInternalException;
import org.semanticweb.owlapi.reasoner.ReasonerInterruptedException;
import org.semanticweb.owlapi.reasoner.TimeOutException;
import org.semanticweb.owlapi.reasoner.UnsupportedEntailmentTypeException;
import org.semanticweb.owlapi.reasoner.impl.OWLReasonerBase;
import org.semanticweb.owlapi.util.Version;

import de.fzi.herakles.analyser.Analyser;
import de.fzi.herakles.analyser.impl.AnalyserFactory;
import de.fzi.herakles.executor.Executor;
import de.fzi.herakles.executor.impl.DefaultExecutorFactory;
import de.fzi.herakles.executor.impl.ExecutorException;
import de.fzi.herakles.job.Job;
import de.fzi.herakles.job.impl.JobManager;
import de.fzi.herakles.loader.Loader;
import de.fzi.herakles.loader.impl.LoaderFactory;
import de.fzi.herakles.registry.ReasonerAdapter;
import de.fzi.herakles.registry.ReasonerRegistry;
import de.fzi.herakles.registry.ServerAdapter;
import de.fzi.herakles.registry.ServerRegistry;
import de.fzi.herakles.registry.impl.ReasonerRegistryFactory;
import de.fzi.herakles.registry.impl.ServerRegistryFactory;
import de.fzi.owllink.reasoner.server.entities.Owner;
import de.fzi.owllink.reasoner.server.entities.ReasonerType;

/**
 * 
 * @author Marcel Weigel
 * 
 */
public class HeraklesReasoner extends OWLReasonerBase implements OWLlinkReasoner {

	private final static Log LOG = LogFactory.getLog(HeraklesReasoner.class);

	private final static String METHODE_REFENCE_CLASS = "org.semanticweb.owlapi.owllink.OWLlinkReasoner";

	private Analyser analyser;

	private Loader loader;

	private Executor executor;

	private ServerRegistry serverRegistry;

	private ReasonerRegistry reasonerRegistry;

	private JobManager jobManager;

	private Owner owner;

	private HeraklesReasonerConfiguration heraklesReasonerConfiguration;

	HeraklesReasoner(OWLOntology ontology, HeraklesReasonerConfiguration heraklesReasonerConfiguration,
			BufferingMode bufferingMode) {
		super(ontology, heraklesReasonerConfiguration, bufferingMode);

		guard_HeraklesReasoner(ontology, heraklesReasonerConfiguration, bufferingMode);

		this.heraklesReasonerConfiguration = heraklesReasonerConfiguration;
		this.serverRegistry = ServerRegistryFactory.getInstance().getServerRegistry();
		this.reasonerRegistry = ReasonerRegistryFactory.getInstance().getReasonerRegistry(ontology, bufferingMode);
		this.owner = new Owner("Herakles");

		initRegistries();

		this.analyser = AnalyserFactory.getInstance().getAnalyser(ontology, this.serverRegistry, heraklesReasonerConfiguration.getExcludedReasonerTypeIds());
		this.loader = LoaderFactory.getInstance().getLoader(this.serverRegistry, this.reasonerRegistry, this.owner);
		this.executor = DefaultExecutorFactory.getInstance().getExecutor(this.reasonerRegistry,
				this.heraklesReasonerConfiguration.getTimeOut());
		this.jobManager = JobManager.getInstance();

		initReasoner();
	}

	private void guard_HeraklesReasoner( OWLOntology ontology, HeraklesReasonerConfiguration heraklesReasonerConfiguration,
			BufferingMode bufferingMode ) {
		if (ontology == null) {
			final String errStr = "OWLOntology is null.";
			LOG.fatal(errStr);
			throw new NullPointerException(errStr);
		}
		if (heraklesReasonerConfiguration == null) {
			final String errStr = "HeraklesReasonerConfiguration is null.";
			LOG.fatal(errStr);
			throw new NullPointerException(errStr);
		}
		if (bufferingMode == null) {
			final String errStr = "BufferingMode is null.";
			LOG.fatal(errStr);
			throw new NullPointerException(errStr);
		}
		if (ontology.isEmpty()) {
			LOG.info("OWLOntology is empty.");
		}
	}

	private void fillUpServerRegistry( Set<URL> serverUrls ) {
		for (URL url : serverUrls) {
			if (!this.serverRegistry.isServerAdapterRegistered(url.getHost(), url.getPort())) {
				this.serverRegistry.createAndAddServerAdatper(url.getHost(), url.getPort());
			}
		}
	}

	private void initRegistries() {
		fillUpServerRegistry(this.heraklesReasonerConfiguration.getServerUrls());
	}

	private void initReasoner() {
		Set<ReasonerType> reasonerTypes = this.analyser.getSuitableReasonerTypes();

		if (reasonerTypes.isEmpty()) {
			LOG.warn("No reasoner types were selected.");
		}

		this.loader.updateReasonerRegistry(reasonerTypes);
	}

	@Override
	protected void handleChanges( Set<OWLAxiom> arg0, Set<OWLAxiom> arg1 ) {
		// TODO Chang ontology
		System.out.println("Ontology has changed!");
	}

	private <T> T invoke( String methodName ) {
		return invoke(methodName, null, null);
	}

	private <T> T invoke( String methodName, Class<?>[] paramTypes, Object[] params ) {
		guard_invoke(methodName);

		Method method = getMethod(methodName, paramTypes);
		
		if(LOG.isInfoEnabled())
			LOG.info(String.format("Method whith name %s() was called.", method.getName()));

		Job<T> job = this.jobManager.getJob(method, params);

		T result;
		try {
			result = this.executor.execute(job);
		} catch (ExecutorException e) {
			throw new ReasonerInternalException(e);
		} catch (TimeoutException e) {
			final String errStr = "Get a time out exception, while executing a job.";
			LOG.fatal(errStr);
			throw new ReasonerInternalException(e);
		}

		return result;
	}

	private void guard_invoke( String methodName ) {
		if (methodName == null) {
			final String errStr = "Method name is null.";
			LOG.fatal(errStr);
			throw new NullPointerException(errStr);
		}
		if (methodName.isEmpty()) {
			final String errStr = "Method name is empty.";
			LOG.fatal(errStr);
			throw new IllegalArgumentException(errStr);
		}
	}

	private Method getMethod( String methodName, Class<?>[] paramTypes ) {
		Method method = null;
		try {
			Class<?> clazz = Class.forName(METHODE_REFENCE_CLASS);

			method = clazz.getMethod(methodName, paramTypes);

		} catch (ClassNotFoundException e) {
			final String errStr = String.format("Could not find %s class.", METHODE_REFENCE_CLASS);
			LOG.fatal(errStr);
			throw new FileSystemNotFoundException(errStr);
		} catch (NoSuchMethodException e) {
			final String errStr = String.format("No such method in %s class.", METHODE_REFENCE_CLASS);
			LOG.fatal(errStr);
			throw new WrongMethodTypeException(errStr);
		}
		return method;
	}

	public Node<OWLClass> getBottomClassNode() {
		return invoke("getBottomClassNode");
	}

	public Node<OWLDataProperty> getBottomDataPropertyNode() {
		return invoke("getBottomDataPropertyNode");
	}

	public Node<OWLObjectPropertyExpression> getBottomObjectPropertyNode() {
		return invoke("getBottomObjectPropertyNode");
	}

	public NodeSet<OWLClass> getDataPropertyDomains( OWLDataProperty arg0, boolean arg1 ) throws InconsistentOntologyException,
			FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getDataPropertyDomains", new Class<?>[] { OWLDataProperty.class, boolean.class }, new Object[] { arg0,
				arg1 });
	}

	public Set<OWLLiteral> getDataPropertyValues( OWLNamedIndividual arg0, OWLDataProperty arg1 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getDataPropertyValues", new Class<?>[] { OWLNamedIndividual.class, OWLDataProperty.class }, new Object[] {
				arg0, arg1 });
	}

	public NodeSet<OWLNamedIndividual> getDifferentIndividuals( OWLNamedIndividual arg0 ) throws InconsistentOntologyException,
			FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getDifferentIndividuals", new Class<?>[] { OWLNamedIndividual.class }, new Object[] { arg0 });
	}

	public NodeSet<OWLClass> getDisjointClasses( OWLClassExpression arg0 ) throws ReasonerInterruptedException, TimeOutException,
			FreshEntitiesException, InconsistentOntologyException {
		return invoke("getDisjointClasses", new Class<?>[] { OWLClassExpression.class }, new Object[] { arg0 });
	}

	public NodeSet<OWLDataProperty> getDisjointDataProperties( OWLDataPropertyExpression arg0 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getDisjointDataProperties", new Class<?>[] { OWLDataPropertyExpression.class }, new Object[] { arg0 });
	}

	public NodeSet<OWLObjectPropertyExpression> getDisjointObjectProperties( OWLObjectPropertyExpression arg0 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getDisjointObjectProperties", new Class<?>[] { OWLObjectPropertyExpression.class }, new Object[] { arg0 });
	}

	public Node<OWLClass> getEquivalentClasses( OWLClassExpression arg0 ) throws InconsistentOntologyException,
			ClassExpressionNotInProfileException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getEquivalentClasses", new Class<?>[] { OWLClassExpression.class }, new Object[] { arg0 });
	}

	public Node<OWLDataProperty> getEquivalentDataProperties( OWLDataProperty arg0 ) throws InconsistentOntologyException,
			FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getEquivalentDataProperties", new Class<?>[] { OWLDataProperty.class }, new Object[] { arg0 });
	}

	public Node<OWLObjectPropertyExpression> getEquivalentObjectProperties( OWLObjectPropertyExpression arg0 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getEquivalentObjectProperties", new Class<?>[] { OWLObjectPropertyExpression.class },
				new Object[] { arg0 });
	}

	public NodeSet<OWLNamedIndividual> getInstances( OWLClassExpression ce, boolean direct )
			throws InconsistentOntologyException, ClassExpressionNotInProfileException, FreshEntitiesException,
			ReasonerInterruptedException, TimeOutException {
		return invoke("getInstances", new Class<?>[] { OWLClassExpression.class, boolean.class }, new Object[] { ce, direct });
	}

	public Node<OWLObjectPropertyExpression> getInverseObjectProperties( OWLObjectPropertyExpression arg0 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getInverseObjectProperties", new Class<?>[] { OWLObjectPropertyExpression.class },
				new Object[] { arg0 });
	}

	public NodeSet<OWLClass> getObjectPropertyDomains( OWLObjectPropertyExpression arg0, boolean arg1 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getObjectPropertyDomains", new Class<?>[] { OWLObjectPropertyExpression.class, boolean.class },
				new Object[] { arg0, arg1 });
	}

	public NodeSet<OWLClass> getObjectPropertyRanges( OWLObjectPropertyExpression arg0, boolean arg1 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getObjectPropertyRanges", new Class<?>[] { OWLObjectPropertyExpression.class, boolean.class },
				new Object[] { arg0, arg1 });
	}

	public NodeSet<OWLNamedIndividual> getObjectPropertyValues( OWLNamedIndividual arg0, OWLObjectPropertyExpression arg1 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getObjectPropertyValues", new Class<?>[] { OWLNamedIndividual.class, OWLObjectPropertyExpression.class },
				new Object[] { arg0, arg1 });
	}

	public Set<InferenceType> getPrecomputableInferenceTypes() {
		return invoke("getPrecomputableInferenceTypes");
	}

	public String getReasonerName() {
		return invoke("getReasonerName");
	}

	public Version getReasonerVersion() {
		return invoke("getReasonerVersion");
	}

	public Node<OWLNamedIndividual> getSameIndividuals( OWLNamedIndividual arg0 ) throws InconsistentOntologyException,
			FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getSameIndividuals", new Class<?>[] { OWLNamedIndividual.class }, new Object[] { arg0 });
	}

	public NodeSet<OWLClass> getSubClasses( OWLClassExpression arg0, boolean arg1 ) throws ReasonerInterruptedException,
			TimeOutException, FreshEntitiesException, InconsistentOntologyException, ClassExpressionNotInProfileException {
		return invoke("getSubClasses", new Class<?>[] { OWLClassExpression.class, boolean.class }, new Object[] { arg0, arg1 });
	}

	public NodeSet<OWLDataProperty> getSubDataProperties( OWLDataProperty arg0, boolean arg1 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getSubDataProperties", new Class<?>[] { OWLDataProperty.class, boolean.class },
				new Object[] { arg0, arg1 });
	}

	public NodeSet<OWLObjectPropertyExpression> getSubObjectProperties( OWLObjectPropertyExpression arg0, boolean arg1 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getSubObjectProperties", new Class<?>[] { OWLObjectPropertyExpression.class, boolean.class },
				new Object[] { arg0, arg1 });
	}

	public NodeSet<OWLClass> getSuperClasses( OWLClassExpression arg0, boolean arg1 ) throws InconsistentOntologyException,
			ClassExpressionNotInProfileException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getSuperClasses", new Class<?>[] { OWLClassExpression.class, boolean.class }, new Object[] { arg0, arg1 });
	}

	public NodeSet<OWLDataProperty> getSuperDataProperties( OWLDataProperty arg0, boolean arg1 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getSuperDataProperties", new Class<?>[] { OWLDataProperty.class, boolean.class }, new Object[] { arg0,
				arg1 });
	}

	public NodeSet<OWLObjectPropertyExpression> getSuperObjectProperties( OWLObjectPropertyExpression arg0, boolean arg1 )
			throws InconsistentOntologyException, FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getSuperObjectProperties", new Class<?>[] { OWLObjectPropertyExpression.class, boolean.class },
				new Object[] { arg0, arg1 });
	}

	public Node<OWLClass> getTopClassNode() {
		return invoke("getTopClassNode");
	}

	public Node<OWLDataProperty> getTopDataPropertyNode() {
		return invoke("getTopDataPropertyNode");
	}

	public Node<OWLObjectPropertyExpression> getTopObjectPropertyNode() {
		return invoke("getTopObjectPropertyNode");
	}

	public NodeSet<OWLClass> getTypes( OWLNamedIndividual arg0, boolean arg1 ) throws InconsistentOntologyException,
			FreshEntitiesException, ReasonerInterruptedException, TimeOutException {
		return invoke("getTypes", new Class<?>[] { OWLNamedIndividual.class, boolean.class }, new Object[] { arg0, arg1 });
	}

	public Node<OWLClass> getUnsatisfiableClasses() throws ReasonerInterruptedException, TimeOutException,
			InconsistentOntologyException {
		return invoke("getUnsatisfiableClasses");
	}

	public void interrupt() {
		stopExecutor();
		stopAllReasonerServer();
	}

	private void stopExecutor() {
		if (this.executor != null) {
			this.executor.interrupt();
		}
	}

	private void stopAllReasonerServer() {
		if (this.reasonerRegistry != null) {
			Iterator<ReasonerAdapter> interator = this.reasonerRegistry.getIdleReasonerAdapters().iterator();
			while(interator.hasNext()) {
				try {
					ReasonerAdapter reasonerAdapter = interator.next();

					ServerAdapter serverAdapter = reasonerAdapter.getServerAdapter();
					serverAdapter.stopReasoner(reasonerAdapter, this.reasonerRegistry);
					
				} catch(ConcurrentModificationException e) {
					LOG.error("Could not stop reasoner besouse it is already in process.");
				}
			}
		}
	}

	public boolean isConsistent() throws ReasonerInterruptedException, TimeOutException {
		return invoke("isConsistent");
	}

	public boolean isEntailed( OWLAxiom arg0 ) throws ReasonerInterruptedException, UnsupportedEntailmentTypeException,
			TimeOutException, AxiomNotInProfileException, FreshEntitiesException, InconsistentOntologyException {
		return invoke("isEntailed", new Class<?>[] { OWLAxiom.class }, new Object[] { arg0 });
	}

	public boolean isEntailed( Set<? extends OWLAxiom> arg0 ) throws ReasonerInterruptedException,
			UnsupportedEntailmentTypeException, TimeOutException, AxiomNotInProfileException, FreshEntitiesException,
			InconsistentOntologyException {
		return invoke("isEntailed", new Class<?>[] { Set.class }, new Object[] { arg0 });
	}

	public boolean isEntailmentCheckingSupported( AxiomType<?> arg0 ) {
		return invoke("isEntailmentCheckingSupported", new Class<?>[] { AxiomType.class }, new Object[] { arg0 });
	}

	public boolean isPrecomputed( InferenceType arg0 ) {
		return invoke("isPrecomputed", new Class<?>[] { InferenceType.class }, new Object[] { arg0 });
	}

	public boolean isSatisfiable( OWLClassExpression arg0 ) throws ReasonerInterruptedException, TimeOutException,
			ClassExpressionNotInProfileException, FreshEntitiesException, InconsistentOntologyException {
		return invoke("isSatisfiable", new Class<?>[] { OWLClassExpression.class }, new Object[] { arg0 });
	}

	public void precomputeInferences( InferenceType... arg0 ) throws ReasonerInterruptedException, TimeOutException,
			InconsistentOntologyException {
		invoke("precomputeInferences", new Class<?>[] { InferenceType[].class }, new Object[] { arg0 });
	}

	public <R extends Response> R answer( Request<R> arg0 ) throws OWLlinkErrorResponseException {
		// TODO Auto-generated method stub
		return null;
	}

	public ResponseMessage answer( Request... arg0 ) {
		// TODO Auto-generated method stub
		return null;
	}

}
