/*
 * Copyright (C) 2013, FZI Forschungszentrum Informatik
 *
 * Modifications to the initial code base are copyright of their
 * respective authors, or their employers as appropriate. Authorship
 * of the modifications may be determined from the ChangeLog placed at
 * the end of this file.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package de.fzi.herakles.reasoner;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.reasoner.BufferingMode;
import org.semanticweb.owlapi.reasoner.IllegalConfigurationException;
import org.semanticweb.owlapi.reasoner.OWLReasonerConfiguration;
import org.semanticweb.owlapi.reasoner.OWLReasonerFactory;

/**
 * 
 * @author Marcel Weigel
 * 
 */
public class HeraklesReasonerFactory implements OWLReasonerFactory {

	private final static Log LOG = LogFactory.getLog(HeraklesReasonerFactory.class);

	private static HeraklesReasonerFactory heraklesReasonerFactory = new HeraklesReasonerFactory();

	private HeraklesReasonerFactory() {
	}

	public String getReasonerName() {
		return "HeraklesReasoner";
	}

	public static HeraklesReasonerFactory getInstance() {
		return heraklesReasonerFactory;
	}

	public HeraklesReasoner createNonBufferingReasoner( OWLOntology ontology ) {
		final String errStr = "Herakles requires a HeraklesReasonerConfiguration.";
		LOG.error(errStr);
		throw new UnsupportedOperationException(errStr);
	}

	public HeraklesReasoner createReasoner( OWLOntology ontology ) {
		final String errStr = "Herakles requires a HeraklesReasonerConfiguration.";
		LOG.error(errStr);
		throw new UnsupportedOperationException(errStr);
	}

	public HeraklesReasoner createNonBufferingReasoner( OWLOntology ontology, OWLReasonerConfiguration config )
			throws IllegalConfigurationException {
		guard_createReasoner(ontology, config);

		return new HeraklesReasoner(ontology, (HeraklesReasonerConfiguration) config, BufferingMode.NON_BUFFERING);
	}

	public HeraklesReasoner createReasoner( OWLOntology ontology, OWLReasonerConfiguration config )
			throws IllegalConfigurationException {
		guard_createReasoner(ontology, config);

		return new HeraklesReasoner(ontology, (HeraklesReasonerConfiguration) config, BufferingMode.BUFFERING);
	}

	private void guard_createReasoner( OWLOntology ontology, OWLReasonerConfiguration config ) {
		if (ontology == null) {
			final String errStr = "Ontology is null.";
			LOG.fatal(errStr);
			throw new NullPointerException(errStr);
		}
		if (config == null) {
			final String errStr = "OWLReasonerConfiguration is null.";
			LOG.fatal(errStr);
			throw new NullPointerException(errStr);
		}
		if (!(config instanceof HeraklesReasonerConfiguration)) {
			final String errStr = "Configuration must be a HeraklesReasonerConfiguration.";
			LOG.fatal(errStr);
			throw new IllegalConfigurationException(errStr, config);
		}
		HeraklesReasonerConfiguration heraklesReasonerConfiguration = (HeraklesReasonerConfiguration) config;
		if (heraklesReasonerConfiguration.getServerUrls().isEmpty()) {
			final String errStr = "No server urls configured.";
			LOG.fatal(errStr);
			throw new IllegalConfigurationException(errStr, config);
		}
	}
}
