/*
 * Copyright (C) 2013, FZI Forschungszentrum Informatik
 *
 * Modifications to the initial code base are copyright of their
 * respective authors, or their employers as appropriate. Authorship
 * of the modifications may be determined from the ChangeLog placed at
 * the end of this file.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package de.fzi.herakles.registry;

import java.util.Set;

import de.fzi.owllink.reasoner.server.entities.Owner;
import de.fzi.owllink.reasoner.server.entities.ReasonerType;

/**
 * The ReasonerRegistry registered all {@link ReasonerAdapter} which was started
 * from a {@link ServerAdapter}.
 * 
 * @author Marcel Weigel
 * 
 */
public interface ReasonerRegistry {

	/**
	 * Get a set from all ReasonerAdapters which are registered on this
	 * {@link ReasonerRegistry}.
	 * 
	 * @return Set of ReasonerAdapters.
	 */
	Set<ReasonerAdapter> getReasonerAdapters();

	/**
	 * Get a set from all ReasonerAdapters which are registered on this
	 * {@link ReasonerRegistry} and they are idle.
	 * 
	 * @return Set of idle ReasonerAdapters.
	 */
	Set<ReasonerAdapter> getIdleReasonerAdapters();

	/**
	 * Get a set from all ReasonerAdapters which are registered on this
	 * {@link ReasonerRegistry} and was started on the {@link ServerAdapter}.
	 * 
	 * @param serverAdapter
	 *            ServerAdapter
	 * @return Set of ReasonerAdapters which are registered on ServerAdapter.
	 */
	Set<ReasonerAdapter> getReasonerAdapters( ServerAdapter serverAdapter );

	/**
	 * Get one ReasonerAdapter which was started on a {@link ServerAdapter} and
	 * started on port.
	 * 
	 * @param serverAdapter
	 *            ServerAdapter
	 * @param port
	 *            Port of the ReasonerAdapter
	 * @return The ReasonerAdapter.
	 */
	ReasonerAdapter getReasonerAdapter( ServerAdapter serverAdapter, int port );

	/**
	 * Check if a ReasonerAdapter was registered on a {@link ServerAdapter} on a
	 * special port.
	 * 
	 * @param serverAdapter
	 *            ServerAdapter
	 * @param port
	 *            Port of the ReasonerAdapter
	 * @return True if the ReasonerAdapter was registered.
	 */
	boolean existsReasonerAdapter( ServerAdapter serverAdapter, int port );

	/**
	 * Get the number of registered ReasonerAdapters on the registry.
	 * 
	 * @return Number of ReasonerAdapter.
	 */
	int countReasonerAdapters();

	/**
	 * Create a new {@link ReasonerAdapter} with port, owner and reasoner type.
	 * As host it get the same as the {@link ServerAdapter}.
	 * 
	 * @param serverAdapter
	 *            SeverAdapter
	 * @param port
	 *            port
	 * @param owner
	 *            Owner
	 * @param reasonerType
	 *            ReasonerType
	 * @return ReasonerAdapter
	 */
	ReasonerAdapter createAndAddReasonerAdapter( ServerAdapter serverAdapter, int port, Owner owner, ReasonerType reasonerType );

	/**
	 * Remove the {@link ReasonerAdapter} from the {@link ReasonerRegistry}.
	 * 
	 * @param reasonerAdapter
	 *            ReasonerAdapter
	 */
	void removeReasonerAdapter( ReasonerAdapter reasonerAdapter );

	/**
	 * Remove all {@link ReasonerAdapters} from {@link ReasonerRegistry}.
	 * 
	 * @param reasonerAdapters
	 *            Set of ReasonerAdapters.
	 */
	void removeAllReasonerAdapters( Set<ReasonerAdapter> reasonerAdapters );

	/**
	 * Check if {@link ReasonerRegistry} is empty.
	 * 
	 * @return True if the ReasonerRegistry is empty.
	 */
	boolean isEmpty();

}
