/*
 * Copyright (C) 2013, FZI Forschungszentrum Informatik
 *
 * Modifications to the initial code base are copyright of their
 * respective authors, or their employers as appropriate. Authorship
 * of the modifications may be determined from the ChangeLog placed at
 * the end of this file.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package de.fzi.herakles.registry;

import java.util.Set;

import de.fzi.owllink.reasoner.server.owllinkserver.OwlLinkServerMetadata;
import de.fzi.owllink.reasoner.server.entities.Owner;
import de.fzi.owllink.reasoner.server.entities.ReasonerType;
import de.fzi.owllink.reasoner.server.entities.ServerState;

/**
 * A ServerAdapter maps a rest client which control the OwlLinkServers.
 * 
 * @author Marcel Weigel
 * 
 */
public interface ServerAdapter {

	/**
	 * Get the name of the host from the {@link ServerAdapter}.
	 * 
	 * @return The name of the host.
	 */
	String getHost();

	/**
	 * Get the port of the {@link ServerAdapter}.
	 * 
	 * @return The port.
	 */
	int getPort();

	/**
	 * Get all reasoner types which are available on the {@link ServerAdapter}.
	 * 
	 * @return Set of all available reasoner types.
	 */
	Set<ReasonerType> getAvailableReasonerTypes();

	/**
	 * Start a new OwlLinkServer on the rest client and add a new
	 * {@link ReasonerAdapter} on the given {@link ReasonerRegistry}.
	 * 
	 * @param reasonerType
	 *            ReasonerType
	 * @param owner
	 *            Owner
	 * @param reasonerRegistry
	 *            ReasonerRegistry
	 * @return The created ReasonerAdapter.
	 */
	ReasonerAdapter startReasoner( ReasonerType reasonerType, Owner owner, ReasonerRegistry reasonerRegistry );

	/**
	 * Stop the OwlLinkServer which belongs to the {@link ReasonerAdapter} on
	 * the rest client. And Remove the {@link ReasonerAdapter} from the given
	 * {@link ReasonerRegistry}.
	 * 
	 * @param reasonerAdapter
	 *            ReasonerAdapter
	 * @param reasonerRegistry
	 *            ReasonerRegistry
	 */
	void stopReasoner( ReasonerAdapter reasonerAdapter, ReasonerRegistry reasonerRegistry );

	/**
	 * Stop all OwlLinkServer which was created from the owner on the rest
	 * client. Remove all {@link ReasonerAdapters} from the
	 * {@link ReasonerRegistry}.
	 * 
	 * @param owner
	 *            Owner
	 * @param reasonerRegistry
	 *            ReasonerRegistry
	 */
	void stopAllReasoner( Owner owner, ReasonerRegistry reasonerRegistry );

	/**
	 * Get the number of started reasoner.
	 * 
	 * @return Number of started reasoner.
	 */
	int getStartedReasonerCount();

	/**
	 * Get the number of started reasoner which belongs to the owner.
	 * 
	 * @param owner
	 *            Owner
	 * @return Number of started reasoner.
	 */
	int getStartedReasonerCount( Owner owner );

	/**
	 * Get all OwlLinkServerMetadata from all reasoner.
	 * 
	 * @return Set of OwlLinkServerMetadata.
	 */
	Set<OwlLinkServerMetadata> getAllOwlLinkServerMetadata();

	/**
	 * 
	 * Get all OwlLinkServerMetadata from all reasoner which belongs to the
	 * owner.
	 * 
	 * @return Set of OwlLinkServerMetadata.
	 */
	Set<OwlLinkServerMetadata> getAllOwlLinkServerMetadata( Owner owner );

	/**
	 * Get one OwlLinkServerMetadata witch belongs to the ReasonerAdapter.
	 * 
	 * @param reasonerAdapter
	 *            ReasonerAdpater
	 * @return The OwlLinkServerMetadata.
	 */
	OwlLinkServerMetadata getOwlLinkServerMetadata( ReasonerAdapter reasonerAdapter );

	/**
	 * Get the ServerState of the given ReasonerAdapter.
	 * 
	 * @param reasonerAdapter
	 *            ReasonerAdapter
	 * @return The server state of the ReasonerAdapter.
	 */
	ServerState getReasonerState( ReasonerAdapter reasonerAdapter );

	/**
	 * Check if the given ReasonerAdapter is started.
	 * 
	 * @param reasonerAdapter
	 *            ReasonerAdapter.
	 * @return True if the reasoner was started.
	 */
	boolean isReasonerStarted( ReasonerAdapter reasonerAdapter );

}
