/*
 * Copyright (C) 2013, FZI Forschungszentrum Informatik
 *
 * Modifications to the initial code base are copyright of their
 * respective authors, or their employers as appropriate. Authorship
 * of the modifications may be determined from the ChangeLog placed at
 * the end of this file.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package de.fzi.herakles.registry.impl;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.owllink.OWLlinkHTTPXMLReasonerFactory;
import org.semanticweb.owlapi.owllink.OWLlinkReasoner;
import org.semanticweb.owlapi.owllink.OWLlinkReasonerConfiguration;
import org.semanticweb.owlapi.owllink.OWLlinkReasonerIOException;
import org.semanticweb.owlapi.owllink.OWLlinkReasonerRuntimeException;
import org.semanticweb.owlapi.reasoner.BufferingMode;
import org.semanticweb.owlapi.util.Version;

import de.fzi.herakles.registry.ReasonerAdapter;
import de.fzi.herakles.registry.ServerAdapter;
import de.fzi.owllink.reasoner.server.entities.Owner;
import de.fzi.owllink.reasoner.server.entities.ReasonerType;

/**
 * 
 * @author Marcel Weigel
 * 
 */
public class ReasonerAdapterFactory {

	private final static Log LOG = LogFactory.getLog(ReasonerAdapterFactory.class);

	private static final int MIN_PORT_NUMBER = 1024;
	private static final int MAX_PORT_NUMBER = 65535;

	private static ReasonerAdapterFactory reasonerAdapterFactory;

	private Set<ReasonerAdapter> reasonerAdapters;

	private OWLlinkHTTPXMLReasonerFactory owlReasonerFactory;

	private ReasonerAdapterFactory() {
		this.owlReasonerFactory = new OWLlinkHTTPXMLReasonerFactory();
		this.reasonerAdapters = new HashSet<ReasonerAdapter>();
	}

	static ReasonerAdapterFactory getInstance() {
		if (reasonerAdapterFactory == null) {
			reasonerAdapterFactory = new ReasonerAdapterFactory();
		}
		return reasonerAdapterFactory;
	}

	ReasonerAdapter getReasonerAdapter( ServerAdapter serverAdapter, int port, Owner owner, ReasonerType reasonerType,
			OWLOntology ontology, BufferingMode bufferingMode ) {
		guard_createOwlLinkReasoner(serverAdapter, port, owner, reasonerType, ontology, bufferingMode);

		ReasonerAdapter reasonerAdapter = null;
		for (ReasonerAdapter reasonerAdapterTmp : this.reasonerAdapters) {
			if (reasonerAdapterTmp.getServerAdapter().equals(serverAdapter) && reasonerAdapterTmp.getPort() == port
					&& reasonerAdapterTmp.getOwner().equals(owner) && reasonerAdapterTmp.getReasonerType().equals(reasonerType)) {
				reasonerAdapter = reasonerAdapterTmp;
			}
		}
		if (reasonerAdapter == null) {
			OWLlinkReasoner owlLinkReasoner = createOwlLinkReasoner(serverAdapter.getHost(), port, ontology, bufferingMode);
			reasonerAdapter = new ReasonerAdapterImpl(serverAdapter, port, owner, reasonerType, owlLinkReasoner);
			this.reasonerAdapters.add(reasonerAdapter);
		}
		return reasonerAdapter;
	}

	private void guard_createOwlLinkReasoner( ServerAdapter serverAdapter, int port, Owner owner, ReasonerType reasonerType,
			OWLOntology ontology, BufferingMode bufferingMode ) {
		if (serverAdapter == null) {
			final String errStr = "ServerAdapter is null.";
			LOG.error(errStr);
			throw new NullPointerException(errStr);
		}
		if (serverAdapter.getHost() == null) {
			final String errStr = "Host name is null.";
			LOG.error(errStr);
			throw new NullPointerException(errStr);
		}
		if (owner == null) {
			final String errStr = "Owner is null.";
			LOG.error(errStr);
			throw new NullPointerException(errStr);
		}
		if (reasonerType == null) {
			final String errStr = "ReasonerType is null.";
			LOG.error(errStr);
			throw new NullPointerException(errStr);
		}
		if (ontology == null) {
			final String errStr = "OWLOntology is null.";
			LOG.error(errStr);
		}
		if (bufferingMode == null) {
			final String errStr = "BufferingMode is null.";
			LOG.error(errStr);
		}
		if (serverAdapter.getHost().isEmpty()) {
			final String errStr = "Host name is empty.";
			LOG.error(errStr);
			throw new IllegalArgumentException(errStr);
		}
		if (!(MIN_PORT_NUMBER <= serverAdapter.getPort() && serverAdapter.getPort() <= MAX_PORT_NUMBER)) {
			final String errStr = String.format("Port %04d has wrong range.", serverAdapter.getPort());
			LOG.error(errStr);
			throw new IllegalArgumentException(errStr);
		}
		if (!(MIN_PORT_NUMBER <= port && port <= MAX_PORT_NUMBER)) {
			final String errStr = String.format("Port %04d has wrong range.", port);
			LOG.error(errStr);
			throw new IllegalArgumentException(errStr);
		}
	}

	private OWLlinkReasoner createOwlLinkReasoner( String host, int port, OWLOntology ontology, BufferingMode bufferingMode ) {
		URL reasonerUrl = getUrl(host, port);

		OWLlinkReasonerConfiguration reasonerConfiguration = new OWLlinkReasonerConfiguration(reasonerUrl);

		OWLlinkReasoner owlLinkReasoner;
		try {
			switch (bufferingMode) {
			case BUFFERING:
				owlLinkReasoner = this.owlReasonerFactory.createReasoner(ontology, reasonerConfiguration);
				break;
			case NON_BUFFERING:
				owlLinkReasoner = this.owlReasonerFactory.createNonBufferingReasoner(ontology, reasonerConfiguration);
				break;
			default:
				final String errStr = String.format("BufferingMode not suppored: %s.", bufferingMode);
				LOG.error(errStr);
				throw new OWLlinkReasonerRuntimeException(errStr);
			}
			if (LOG.isInfoEnabled()) {
				Version v = owlLinkReasoner.getReasonerVersion();
				LOG.info(String.format("Created Reasoner(%s) with Version(major=%d, minor=%d, patch=%d, build=%d).", owlLinkReasoner.getReasonerName(), v.getMajor(), v.getMinor(), v.getPatch(), v.getBuild()));
			}

		} catch (OWLlinkReasonerIOException e) {
			final String errStr = String.format("Can not connect to URL %s.", reasonerUrl);
			LOG.error(errStr);
			throw new OWLlinkReasonerIOException(errStr);
		} catch (OWLlinkReasonerRuntimeException e) {
			final String errStr = String.format("Perhaps the libraries have diffenrent versions. Error Message: %s.",
					e.getCause());
			LOG.error(errStr);
			throw new OWLlinkReasonerRuntimeException(errStr);
		} catch(NoSuchMethodError e) {
			final String errStr = String.format("Perhaps the OWLApi from server and clint have different versions. Error message: %s", e.getCause());
			LOG.fatal(errStr);
			throw new OWLlinkReasonerRuntimeException(errStr);
		}

		return owlLinkReasoner;
	}

	private URL getUrl( String address, int port ) {
		URL reasonerUrl;
		try {
			reasonerUrl = new URL(String.format("http://%s:%04d", address, port));
		} catch (MalformedURLException e) {
			final String errStr = String.format("Clound not parsed address %s and port %04d to URL.", address, port);
			LOG.error(errStr);
			throw new IllegalArgumentException(errStr);
		}
		return reasonerUrl;
	}

}
