/*
* Copyright (C) 2013, FZI Forschungszentrum Informatik
*
* Modifications to the initial code base are copyright of their
* respective authors, or their employers as appropriate. Authorship
* of the modifications may be determined from the ChangeLog placed at
* the end of this file.
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 3.0 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/
package de.fzi.herakles.registry.impl;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.owllink.OWLlinkReasoner;
import org.semanticweb.owlapi.reasoner.BufferingMode;

import de.fzi.herakles.job.Job;
import de.fzi.herakles.registry.ReasonerAdapter;
import de.fzi.herakles.registry.ServerAdapter;
import de.fzi.owllink.reasoner.server.entities.Owner;
import de.fzi.owllink.reasoner.server.entities.ReasonerType;

/**
 * 
 * @author Marcel Weigel
 *
 */
class ReasonerAdapterImpl implements ReasonerAdapter {
	
	private final static Log LOG = LogFactory.getLog(ReasonerAdapterImpl.class);

	private ServerAdapter serverAdapter;
	
	private int port;
	
	private Owner owner;
	
	private ReasonerType reasonerType;
	
	private OWLlinkReasoner owLlinkReasoner;
	
	private boolean isBusy;

	ReasonerAdapterImpl(ServerAdapter serverAdapter, int port, Owner owner, ReasonerType reasonerType, OWLlinkReasoner owlLinkReasoner) {
		this.serverAdapter = serverAdapter;
		this.port = port;
		this.owner = owner;
		this.reasonerType = reasonerType;
		
		this.owLlinkReasoner = owlLinkReasoner;
	}

	public ServerAdapter getServerAdapter() {
		return serverAdapter;
	}

	public int getPort() {
		return port;
	}

	public Owner getOwner() {
		return owner;
	}
	
	public ReasonerType getReasonerType() {
		return reasonerType;
	}

	public OWLOntology getOntology() {
		return this.owLlinkReasoner.getRootOntology();
	}

	public BufferingMode getBufferingMode() {
		return this.owLlinkReasoner.getBufferingMode();
	}
	
	public boolean isBusy() {
		return this.isBusy;
	}

	@SuppressWarnings("unchecked")
	public <T> void invoke(Job<T> job) {
		this.isBusy = true;

		Method method = job.getRequest().getMethod();
		Object[] params = job.getRequest().getParams();
		
		T returnValue = null;
		if(this.owLlinkReasoner == null) {
			final String meg = "No OwlLinkReasoner found!";
			LOG.info(meg);
			job.createResponse(this, new Exception(meg));
		} else {
			try {
				if(LOG.isInfoEnabled())
					LOG.info(String.format("Invoke methode %s() on reasoner %s.", method.getName(), this.reasonerType.getName()));
				
				returnValue = (T) method.invoke(this.owLlinkReasoner, params);
				
				if(LOG.isInfoEnabled())
					LOG.info(String.format("Invoke methode %s() on reasoner %s has successfilly finished.", method.getName(), this.reasonerType.getName()));
				job.createResponse(this, returnValue);
			} catch (IllegalAccessException e) {
				final String meg = String.format("Invoke methode %s() on reasoner %s has failied becouse no acess rights.", method.getName(), this.reasonerType.getName());
				LOG.fatal(meg);
				throw new AssertionError(meg);
			} catch (IllegalArgumentException e) {
				final String meg = String.format("Invoke methode %s() on reasoner %s has failied becouse wrong arguments.", method.getName(), this.reasonerType.getName());
				LOG.fatal(meg);
				throw new AssertionError(meg);
			} catch (InvocationTargetException e) {
				if(LOG.isInfoEnabled()) 
					LOG.info(String.format("Invoke methode %s() on reasoner %s has failied becouse exception throw: %s.", method.getName(), this.reasonerType.getName(), e.getTargetException().getMessage()));
				
				job.createResponse(this, e.getTargetException());
			}
		}

		this.isBusy = false;
	}

	@Override
	public boolean equals(Object obj) {
		if(obj == null)
			return false;
		if(obj == this)
			return true;
		if(!(obj instanceof ReasonerAdapterImpl)) 
			return false;
		
		ReasonerAdapterImpl rhs = (ReasonerAdapterImpl) obj;
		return (this.reasonerType == rhs.reasonerType || this.reasonerType != null && this.reasonerType.equals(rhs.reasonerType)) &&
				(this.serverAdapter == rhs.serverAdapter || this.serverAdapter != null && this.serverAdapter.equals(rhs.serverAdapter)) &&
				this.port == rhs.port &&
				(this.owner == rhs.owner || this.owner != null && this.owner.equals(rhs.owner)) &&
				(this.owLlinkReasoner == rhs.owLlinkReasoner || this.owLlinkReasoner != null && this.owLlinkReasoner.equals(rhs.owLlinkReasoner));
	}

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 31 * hash + (this.reasonerType == null ? 0 : this.reasonerType.hashCode());
		hash = 31 * hash + (this.serverAdapter == null ? 0 : this.serverAdapter.hashCode());
		hash = 31 * hash + this.port;
		hash = 31 * hash + (this.owner == null ? 0 : this.owner.hashCode());
		hash = 31 * hash + (this.owLlinkReasoner == null ? 0 : this.owLlinkReasoner.hashCode());
		return hash;
	}
	
	@Override
	public String toString() {
		String[] classPath = ReasonerAdapterImpl.class.getName().split("\\.");
		return String.format("%s(host=%s, port=%04d, reasonerType=%s)", classPath[classPath.length-1], this.serverAdapter.getHost(), this.port, this.reasonerType);
	}
	
}
