/*
* Copyright (C) 2013, FZI Forschungszentrum Informatik
*
* Modifications to the initial code base are copyright of their
* respective authors, or their employers as appropriate. Authorship
* of the modifications may be determined from the ChangeLog placed at
* the end of this file.
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 3.0 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/
package de.fzi.herakles.registry.impl;

import java.util.HashSet;
import java.util.Set;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.fzi.herakles.registry.ReasonerAdapter;
import de.fzi.herakles.registry.ReasonerRegistry;
import de.fzi.herakles.registry.ServerAdapter;
import de.fzi.owllink.reasoner.server.owllinkserver.OwlLinkServerMetadata;
import de.fzi.owllink.reasoner.server.entities.Owner;
import de.fzi.owllink.reasoner.server.entities.ReasonerType;
import de.fzi.owllink.reasoner.server.entities.ServerState;
import de.fzi.owllink.rest.client.OwlLinkRestClient;
import de.fzi.owllink.rest.client.OwlLinkRestClientImpl;

/**
 * 
 * @author Marcel Weigel
 *
 */
class ServerAdapterImpl implements ServerAdapter {
	
	private static final Log LOG = LogFactory.getLog(ServerAdapterImpl.class);

	private String host;
	
	private int port;
	
	private OwlLinkRestClient owlLinkRestClient;
	
	ServerAdapterImpl(String host, int port) {
		this.host = host;
		this.port = port;
		
		this.owlLinkRestClient = new OwlLinkRestClientImpl(host, port);
	}

	public String getHost() {
		return host;
	}

	public int getPort() {
		return port;
	}

	public Set<ReasonerType> getAvailableReasonerTypes() {
		Set<ReasonerType> availableReasonerTypes = new HashSet<ReasonerType>();
		try {
			availableReasonerTypes.addAll(this.owlLinkRestClient.getAvailableReasonerTypes());
		} catch (Exception e) {
			LOG.error(String.format("Could not get available reasoner types from server %s:%04d.", this.host, this.port));				
		}
		
		return availableReasonerTypes;
	}

	public ReasonerAdapter startReasoner(ReasonerType reasonerType, Owner owner, ReasonerRegistry reasonerRegistry) {		
		if(LOG.isInfoEnabled())
			LOG.info(String.format("Starting reasoner of type %s.", reasonerType.getName()));
		if(LOG.isDebugEnabled())
			LOG.debug(String.format("owner: %s, server: %s, port: %04d.", owner, this.host, this.port));
		ReasonerAdapter reasonerAdapter = null;
		int port = -1;
		try {
			port = this.owlLinkRestClient.registerOwlLinkServer(owner, reasonerType);
		} catch (Exception e) {
			LOG.error(String.format("Could not registered OWLLinkServer on server %s:%04d.", this.host, this.port));
		}
		try {
			this.owlLinkRestClient.startOwlLinkServer(port);
		} catch(Exception e) {
			LOG.error(String.format("Could not started OwlLinkServer on server %s:%04d.", this.host, port));
		} 
		if(LOG.isInfoEnabled())
			LOG.info(String.format("Succsessfully started Reasoner(%s) with %s.", reasonerType.getName(), reasonerType.getVersion()));
		reasonerAdapter = reasonerRegistry.createAndAddReasonerAdapter(this, port, owner, reasonerType);
		return reasonerAdapter;
	}

	public void stopReasoner(ReasonerAdapter reasonerAdapter, ReasonerRegistry reasonerRegistry) {
		try {
			this.owlLinkRestClient.stopOwlLinkServer(reasonerAdapter.getPort());
		} catch (Exception e) {
			LOG.error(String.format("Could not stop OwlLinkServer on server %s:%04d.", this.host, this.port));
		}
		try {
			this.owlLinkRestClient.deregistreOwlLinkServer(reasonerAdapter.getPort());			
		} catch (Exception e) {
			LOG.error(String.format("Could not deregistered OwlLinkServer on server %s:%04d.", this.host, this.port));
		}
		reasonerRegistry.removeReasonerAdapter(reasonerAdapter);
	}
	
	public void stopAllReasoner(Owner owner, ReasonerRegistry reasonerRegistry) {
		Set<OwlLinkServerMetadata> owlLinkServerMetadatas = this.owlLinkRestClient.getAllOwlLinkServerMetadatas(owner);
		Set<ReasonerAdapter> reasonerAdapters = new HashSet<ReasonerAdapter>();
		for(OwlLinkServerMetadata owlLinkServerMetadata : owlLinkServerMetadatas) {
			ReasonerAdapter reasonerAdapter = reasonerRegistry.getReasonerAdapter(this, owlLinkServerMetadata.getPort());
			reasonerAdapters.add(reasonerAdapter);
		}
		
		try {
			this.owlLinkRestClient.stopAllOwlLinkServers(owner);		
		} catch (Exception e) {
			LOG.error(String.format("Could not stop all OwlLinkServers on server %s:%04d.", this.host, this.port));
		}
		try {
			this.owlLinkRestClient.deregisteredAllOwlLinkServers(owner);	
		} catch (Exception e) {
			LOG.error(String.format("Could not deregistered all OwlLinkServers on server %s:%04d.", this.host, this.port));
		}
		
		reasonerRegistry.removeAllReasonerAdapters(reasonerAdapters);
	}

	public int getStartedReasonerCount() {
		int startedOwlLinkServerCount = 0;
		try {
			startedOwlLinkServerCount = this.owlLinkRestClient.getStartedOwlLinkServerCount();
		} catch (Exception e) {
			LOG.error(String.format("Could not count started OwlLinkServers on server %s:%04d.", this.host, this.port));
		}
		return startedOwlLinkServerCount;
	}
	
	public int getStartedReasonerCount(Owner owner) {
		int startedOwlLinkServerCount = 0;
		try {
			startedOwlLinkServerCount = this.owlLinkRestClient.getStartedOwlLinkServerCount(owner);
		} catch (Exception e) {
			LOG.error(String.format("Could not count started OwlLinkServers on server %s:%04d.", this.host, this.port));
		}
		return startedOwlLinkServerCount;
	}
	
	public Set<OwlLinkServerMetadata> getAllOwlLinkServerMetadata() {
		Set<OwlLinkServerMetadata> owlLinkServerMetadatas = new HashSet<OwlLinkServerMetadata>();
		try {
			owlLinkServerMetadatas = this.owlLinkRestClient.getAllOwlLinkServerMetadatas();
		} catch (Exception e) {
			LOG.error(String.format("Could not get all OwlLinkServerMetadata on server %s:%04d.", this.host, this.port));
		}
		return owlLinkServerMetadatas;
	}
	
	public Set<OwlLinkServerMetadata> getAllOwlLinkServerMetadata(Owner owner) {
		Set<OwlLinkServerMetadata> owlLinkServerMetadatas = new HashSet<OwlLinkServerMetadata>();
		try {
			owlLinkServerMetadatas = this.owlLinkRestClient.getAllOwlLinkServerMetadatas(owner);
		} catch (Exception e) {
			LOG.error(String.format("Could not get all OwlLinkServerMetadata on server %s:%04d.", this.host, this.port));
		}
		return owlLinkServerMetadatas;
	}
	
	public OwlLinkServerMetadata getOwlLinkServerMetadata(ReasonerAdapter reasonerAdapter) {
		OwlLinkServerMetadata owlLinkServerMetadata = null;
		try {
			owlLinkServerMetadata = this.owlLinkRestClient.getOwlLinkServerMetadata(reasonerAdapter.getPort());
		} catch (Exception e) {
			LOG.error(String.format("Could not get OwlLinkServerMetadata on server %s:%04d.", this.host, this.port));
		}
		return owlLinkServerMetadata;
	}
	
	public ServerState getReasonerState(ReasonerAdapter reasonerAdapter) {
		ServerState serverState = null;
		try {
			serverState = this.owlLinkRestClient.getOwlLinkServerState(reasonerAdapter.getPort());
		} catch (Exception e) {
			LOG.error(String.format("Could not get ServerState from OwlLinkServer running on port: %04d on server %s:%04d", reasonerAdapter.getPort(), this.host, this.port));
		}
		return serverState;
	}

	public boolean isReasonerStarted(ReasonerAdapter reasonerAdapter) {
		return getReasonerState(reasonerAdapter) == ServerState.Started;
	}
	
	@Override
	public boolean equals(Object obj) {
		if(obj == null)
			return false;
		if(obj == this)
			return true;
		if(!(obj instanceof ServerAdapterImpl)) 
			return false;
		
		ServerAdapterImpl sa = (ServerAdapterImpl) obj;
		return (this.host == sa.host || this.host != null && this.host.equals(sa.host)) &&
				this.port == sa.port &&
				(this.owlLinkRestClient == sa.owlLinkRestClient || this.owlLinkRestClient != null && this.owlLinkRestClient.equals(sa.owlLinkRestClient));
	}

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 31 * hash + (this.host == null ? 0 : this.host.hashCode());
		hash = 31 * hash + this.port;
		hash = 31 * hash + (this.owlLinkRestClient == null ? 0 : this.owlLinkRestClient.hashCode());
		return hash;
	}
}
