package de.fzi.herakles.integration;

import static org.junit.Assert.*;

import java.io.File;
import java.util.HashSet;
import java.util.Set;

import org.junit.Test;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.semanticweb.owlapi.reasoner.BufferingMode;

import de.fzi.herakles.registry.ReasonerAdapter;
import de.fzi.herakles.registry.ReasonerRegistry;
import de.fzi.herakles.registry.ServerAdapter;
import de.fzi.herakles.registry.impl.ReasonerRegistryFactory;
import de.fzi.herakles.registry.impl.ServerAdapterFactory;
import de.fzi.owllink.reasoner.server.owllinkserver.OwlLinkServerMetadata;
import de.fzi.owllink.reasoner.server.entities.Owner;
import de.fzi.owllink.reasoner.server.entities.ReasonerType;
import de.fzi.owllink.reasoner.server.entities.ServerState;

public class IntegrationTest {
	
	private String host = "localhost";
	private int port = 8080;
	
	private Owner owner = new Owner("ServerAdapterTest", this.host);
	
	private OWLOntologyManager owlOntologyManager;
	
	private OWLOntology loadOntologyTestFile() {
		String path = IntegrationTest.class.getProtectionDomain().getCodeSource().getLocation().getPath() + "de/fzi/herakles/serverregistry/pizza.owl";
		OWLOntology owlOntology = null;
		File file = new File(path);
		
		if(!file.exists()) {
			System.err.printf("#Error[ServerAdapterTest]: Cannt find File: %s \n", path);
			return null;
		}

		try {
			owlOntology = this.owlOntologyManager.loadOntologyFromOntologyDocument(file);
		} catch (OWLOntologyCreationException e) {
			System.err.printf("#Error[ServerAdapterTest]: Cannt load Ontology: %s \n", file);
		}
		
		return owlOntology;
	}
	
	@Test
	public void reasonerServerTest() throws OWLOntologyCreationException {

		ServerAdapter serverAdapter = ServerAdapterFactory.getInstance().getServerAdapter(this.host, this.port);
		
		assertNotNull(serverAdapter);
		assertNotNull(this.owner);

//		OWLOntology ontology = this.owlOntologyManager.createOntology(IRI.create("TestOntology1"));
		OWLOntology ontology = loadOntologyTestFile();
		BufferingMode bufferingMode = BufferingMode.BUFFERING;
		
		assertNotNull(ontology);
		assertNotNull(bufferingMode);
		
		ReasonerRegistry reasonerRegistry = ReasonerRegistryFactory.getInstance().getReasonerRegistry(ontology, bufferingMode);
		
		serverAdapter.stopAllReasoner(this.owner, reasonerRegistry);
		
		assertNotNull(reasonerRegistry);
		
		Set<ReasonerType> reasonerTypes = serverAdapter.getAvailableReasonerTypes();
		
		assertNotNull(reasonerTypes);
		assertFalse(reasonerTypes.isEmpty());
		
		ReasonerType reasonerType = reasonerTypes.iterator().next();
		
		assertNotNull(reasonerType);
		
		ReasonerAdapter reasonerAdapter = serverAdapter.startReasoner(reasonerType, this.owner, reasonerRegistry);
		
		assertNotNull(reasonerAdapter);
		
		OwlLinkServerMetadata olsm = serverAdapter.getOwlLinkServerMetadata(reasonerAdapter);
		
		assertNotNull(olsm);
		
		assertEquals(reasonerAdapter.getPort(), olsm.getPort());
		assertEquals(this.owner, olsm.getOwner());
		assertEquals(ServerState.Started, olsm.getServerState());
		
		ServerState serverStatus = serverAdapter.getReasonerState(reasonerAdapter);
		
		assertNotNull(serverStatus);
		assertEquals(ServerState.Started, serverStatus);
		
		boolean isStarted = serverAdapter.isReasonerStarted(reasonerAdapter);
		
		assertEquals(true, isStarted);
		
		serverAdapter.stopReasoner(reasonerAdapter, reasonerRegistry);
	}
	
	@Test
	public void restServerTest() throws OWLOntologyCreationException {
		ServerAdapter serverAdapter = ServerAdapterFactory.getInstance().getServerAdapter(this.host, this.port);
		
		assertNotNull(serverAdapter);
		assertNotNull(this.owner);
		
		OWLOntology ontology = this.owlOntologyManager.createOntology(IRI.create("TestOntology1"));
		BufferingMode bufferingMode = BufferingMode.BUFFERING;
		
		assertNotNull(ontology);
		assertNotNull(bufferingMode);
		
		ReasonerRegistry reasonerRegistry = ReasonerRegistryFactory.getInstance().getReasonerRegistry(ontology, bufferingMode);
		
		assertNotNull(reasonerRegistry);
		
		assertEquals(this.host, serverAdapter.getHost());
		assertEquals(this.port, serverAdapter.getPort());
		
		Set<ReasonerType> reasonerTypes = serverAdapter.getAvailableReasonerTypes();
		
		assertNotNull(reasonerTypes);
		assertNotSame(0, reasonerTypes.size());
		
		assertEquals(0, serverAdapter.getStartedReasonerCount(this.owner));
		
		Set<ReasonerAdapter> reasonerAdapters = new HashSet<ReasonerAdapter>();
		for(ReasonerType reasonerType : reasonerTypes) {
			ReasonerAdapter reasonerAdapter = serverAdapter.startReasoner(reasonerType, this.owner, reasonerRegistry);
			reasonerAdapters.add(reasonerAdapter);
		}
		assertEquals(reasonerTypes.size(), reasonerAdapters.size());
		assertEquals(reasonerTypes.size(), serverAdapter.getStartedReasonerCount(this.owner));
		
		Set<OwlLinkServerMetadata> olsms = serverAdapter.getAllOwlLinkServerMetadata(this.owner);
		
		assertNotNull(olsms);
		assertEquals(reasonerTypes.size(), olsms.size());
		
		for(ReasonerAdapter reasonerAdapter : reasonerAdapters) {
			assertTrue(serverAdapter.isReasonerStarted(reasonerAdapter));
		}
		
		for(ReasonerAdapter reasonerAdapter : reasonerAdapters) {
			serverAdapter.stopReasoner(reasonerAdapter, reasonerRegistry);
		}
		
		assertEquals(0, serverAdapter.getStartedReasonerCount(this.owner));
	}
	
	@Test
	public void restServerOwnerTest() throws OWLOntologyCreationException {
		ServerAdapter serverAdapter = ServerAdapterFactory.getInstance().getServerAdapter(this.host, this.port);
		Owner owner1 = new Owner("ServerAdapterOwnerTest", this.host);
		Owner owner2 = new Owner("ANotherOwner", "192.168.2.1");
		
		assertNotNull(serverAdapter);
		assertNotNull(owner1);
		assertNotNull(owner2);
		
		OWLOntology ontology = this.owlOntologyManager.createOntology(IRI.create("TestOntology1"));
		BufferingMode bufferingMode = BufferingMode.BUFFERING;
		
		assertNotNull(ontology);
		assertNotNull(bufferingMode);
		
		ReasonerRegistry reasonerRegistry = ReasonerRegistryFactory.getInstance().getReasonerRegistry(ontology, bufferingMode);
		
		serverAdapter.stopAllReasoner(this.owner, reasonerRegistry);
		
		assertNotNull(reasonerRegistry);
		
		assertNotNull(reasonerRegistry);
		
		assertEquals(this.host, serverAdapter.getHost());
		assertEquals(this.port, serverAdapter.getPort());
		
		Set<ReasonerType> reasonerTypes = serverAdapter.getAvailableReasonerTypes();
		
		assertNotNull(reasonerTypes);
		assertNotSame(0, reasonerTypes.size());
		
		assertEquals(0, serverAdapter.getStartedReasonerCount(this.owner));
		
		Set<ReasonerAdapter> reasonerAdapters = new HashSet<ReasonerAdapter>();
		for(ReasonerType reasonerType : reasonerTypes) {
			ReasonerAdapter reasonerAdapter = serverAdapter.startReasoner(reasonerType, owner1, reasonerRegistry);
			reasonerAdapters.add(reasonerAdapter);
		}
		assertEquals(reasonerTypes.size(), reasonerAdapters.size());
		assertEquals(reasonerTypes.size(), serverAdapter.getStartedReasonerCount(owner1));
		
		Set<OwlLinkServerMetadata> olsms = serverAdapter.getAllOwlLinkServerMetadata(owner1);
		
		assertNotNull(olsms);
		assertEquals(reasonerTypes.size(), olsms.size());
		
		for(ReasonerType reasonerType : reasonerTypes) {
			ReasonerAdapter reasonerAdapter = serverAdapter.startReasoner(reasonerType, owner2, reasonerRegistry);
			reasonerAdapters.add(reasonerAdapter);
		}
		
		assertEquals(reasonerTypes.size() * 2, reasonerAdapters.size());
		assertEquals(reasonerTypes.size() * 2, serverAdapter.getStartedReasonerCount());
		assertEquals(reasonerTypes.size(), serverAdapter.getStartedReasonerCount(owner1));
		
		olsms = serverAdapter.getAllOwlLinkServerMetadata(owner1);
		
		assertNotNull(olsms);
		assertEquals(reasonerTypes.size(), olsms.size());
		
		olsms = serverAdapter.getAllOwlLinkServerMetadata(owner2);
		
		assertNotNull(olsms);
		assertEquals(reasonerTypes.size(), olsms.size());
		
		olsms = serverAdapter.getAllOwlLinkServerMetadata();
		
		assertNotNull(olsms);
		assertEquals(reasonerTypes.size() * 2, olsms.size());
		
		serverAdapter.stopAllReasoner(owner1, reasonerRegistry);
		
		assertEquals(reasonerTypes.size(), serverAdapter.getStartedReasonerCount());
		assertEquals(reasonerTypes.size(), serverAdapter.getStartedReasonerCount(owner2));
		assertEquals(0, serverAdapter.getStartedReasonerCount(owner1));
		
		serverAdapter.stopAllReasoner(owner2, reasonerRegistry);
		
		assertEquals(0, serverAdapter.getStartedReasonerCount());
		assertEquals(0, serverAdapter.getStartedReasonerCount(owner2));
		assertEquals(0, serverAdapter.getStartedReasonerCount(owner1));
	}

}
