/*
* Copyright (C) 2013, FZI Forschungszentrum Informatik
*
* Modifications to the initial code base are copyright of their
* respective authors, or their employers as appropriate. Authorship
* of the modifications may be determined from the ChangeLog placed at
* the end of this file.
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 3.0 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

package de.fzi.owllink.reasoner.server.classloader;

import java.util.HashMap;
import java.util.Properties;
import java.util.Set;

import de.fzi.owllink.reasoner.server.manager.ConfigurationException;
import de.fzi.owllink.reasoner.server.properties.PropertyReader;
import de.fzi.owllink.reasoner.server.reasoner.ExtendedOWLlinkServerFactory;

/**
 * 
 * @author Marcel Weigel
 *
 */
public class Classloader {

	private PropertyReader propertyReader;
	
	private HashMap<String, Class<ExtendedOWLlinkServerFactory>> classToReasonerNameMap = new HashMap<String, Class<ExtendedOWLlinkServerFactory>>();
	
	public Classloader(Properties properties) throws ConfigurationException {
		this.propertyReader = new PropertyReader(properties);
		
		loadAllClasses();
	}
	
	public Set<String> getAllReasonerIds() {
		return this.propertyReader.getProperties();
	}
	
	@SuppressWarnings("unchecked")
	private void loadAllClasses() {
		Set<String> reasonerIds = this.propertyReader.getProperties();
		
		for (String reasonerId : reasonerIds) {
			try {
			String className = this.propertyReader.getProperty(reasonerId);
			
			Class<ExtendedOWLlinkServerFactory> owlLinkServerFacotryClass = null;
			try {
				owlLinkServerFacotryClass = (Class<ExtendedOWLlinkServerFactory>) Class.forName(className);
				
				if(!ExtendedOWLlinkServerFactory.class.isAssignableFrom(owlLinkServerFacotryClass)) {
					throw new ConfigurationException(String.format("The class: %s from reasoner: %s is not a instance from ExtendedOWLlinkServerFactory.", className, reasonerId));
				}
				
			} catch (ClassNotFoundException e) {
				throw new ConfigurationException(String.format("Cannot find class: %s for reasoner: %s.", className, reasonerId));
			}
			
			this.classToReasonerNameMap.put(reasonerId, owlLinkServerFacotryClass);
			} catch(ConfigurationException e) {
				System.err.println("Warning: " + e);
			}
		}
	}
	
	public ExtendedOWLlinkServerFactory getInstance(String reasonerId) throws ConfigurationException {
		if(!this.classToReasonerNameMap.containsKey(reasonerId)) {
			throw new ConfigurationException(String.format("Cannot find reasoner with name: %s.", reasonerId));
		}
		Class<ExtendedOWLlinkServerFactory> reasonerFactoryClass = this.classToReasonerNameMap.get(reasonerId);
		
		ExtendedOWLlinkServerFactory owlLinkServerFactory = null;
		
		try {
			owlLinkServerFactory = (ExtendedOWLlinkServerFactory) reasonerFactoryClass.newInstance();

		} catch (InstantiationException e) {
			throw new ConfigurationException(String.format("Cannot instantiat class for reasoner: %s.", reasonerId));
		} catch (IllegalAccessException e) {
			throw new ConfigurationException(String.format("You don't have access to create a instance from class of reasoner: %s.", reasonerId));
		}
		
		return owlLinkServerFactory;
	}
}
