/*
 * Copyright (C) 2013, FZI Forschungszentrum Informatik
 *
 * Modifications to the initial code base are copyright of their
 * respective authors, or their employers as appropriate. Authorship
 * of the modifications may be determined from the ChangeLog placed at
 * the end of this file.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
package de.fzi.owllink.reasoner.server.manager;

import java.io.IOException;
import java.net.DatagramSocket;
import java.net.ServerSocket;
import java.util.HashSet;
import java.util.Properties;
import java.util.Random;
import java.util.Set;

import org.semanticweb.owlapi.owllink.server.serverfactory.OWLlinkServerFactory;

import de.fzi.owllink.reasoner.server.classloader.Classloader;
import de.fzi.owllink.reasoner.server.classloader.ReasonerTypeLoader;
import de.fzi.owllink.reasoner.server.entities.Owner;
import de.fzi.owllink.reasoner.server.entities.ReasonerType;
import de.fzi.owllink.reasoner.server.entities.ServerState;
import de.fzi.owllink.reasoner.server.owllinkserver.OwlLinkServerAdapter;
import de.fzi.owllink.reasoner.server.owllinkserver.OwlLinkServerAdapterImpl;
import de.fzi.owllink.reasoner.server.owllinkserver.OwlLinkServerMetadata;

/**
 * 
 * @author Marcel Weigel
 * 
 */
public class OwlLinkServerManagerImpl implements OwlLinkServerManager {

	private static final int MIN_PORT_NUMBER = 1024;
	private static final int MAX_PORT_NUMBER = 65535;

	private Random random = new Random();

	private Set<OwlLinkServerAdapter> owlLinkServers;

	private Classloader classloader;
	
	private ReasonerTypeLoader reasonerTypeLoader;

	private Set<ReasonerType> reasonerTypes;

	private Properties properties;
	
	private String hostAddress;

	OwlLinkServerManagerImpl(String hostAddress, Properties properties) {
		this.owlLinkServers = new HashSet<OwlLinkServerAdapter>();

		this.properties = properties;
		this.hostAddress = hostAddress;

		try {
			this.classloader = new Classloader(properties);
		} catch (ConfigurationException e) {
			Error(e.getMessage());
		}
		this.reasonerTypeLoader = new ReasonerTypeLoader(this.classloader);
		
		this.reasonerTypes = this.reasonerTypeLoader.getReasonerTypes();
	}

	Set<ReasonerType> getReasonerTypes() {
		return reasonerTypes;
	}

	Properties getProperties() {
		return properties;
	}

	private OwlLinkServerAdapter getOwlLinkServer( int port ) {
		for (OwlLinkServerAdapter owlLinkServer : this.owlLinkServers) {
			if (owlLinkServer.getOwlLinkServerMetadata().getPort() == port) {
				return owlLinkServer;
			}
		}
		return null;
	}

	private int getAvaiablePort() {
		int port = -1;
		do {
			port = this.random.nextInt(MAX_PORT_NUMBER - MIN_PORT_NUMBER) + MIN_PORT_NUMBER;
		} while (!isServerPortAvailable(port));
		return port;
	}

	public Set<OwlLinkServerMetadata> getAllOwlLinkServerMetadatas() {
		return getAllOwlLinkServerMetadatas(this.owlLinkServers);
	}

	public Set<OwlLinkServerMetadata> getAllOwlLinkServerMetadatas( Owner owner ) {
		return getAllOwlLinkServerMetadatas(getOwlLinkServerFromOwner(owner));
	}

	private Set<OwlLinkServerMetadata> getAllOwlLinkServerMetadatas( Set<OwlLinkServerAdapter> owlLinkServers ) {
		Set<OwlLinkServerMetadata> owlLinkServerMetadata = new HashSet<OwlLinkServerMetadata>();
		for (OwlLinkServerAdapter owlLinkServer : owlLinkServers) {
			owlLinkServerMetadata.add(owlLinkServer.getOwlLinkServerMetadata());
		}
		return owlLinkServerMetadata;
	}

	public OwlLinkServerMetadata getOwlLinkServerMetadata( int port ) {
		OwlLinkServerAdapter owlLinkReasonerServer = getOwlLinkServer(port);
		if (owlLinkReasonerServer != null) {
			return owlLinkReasonerServer.getOwlLinkServerMetadata();
		}
		return null;
	}

	public Set<ReasonerType> getAvailableReasonerTypes() {
		return this.reasonerTypes;
	}

	public boolean isReasonerTypeAvailable( ReasonerType reasonerType ) {
		for (ReasonerType reasonerTypeTmp : this.reasonerTypes) {
			if (reasonerTypeTmp.equals(reasonerType)) {
				return true;
			}
		}
		return false;
	}

	public int registerOwlLinkServer( ReasonerType reasonerType, Owner owner ) throws ConfigurationException {
		if (owner == null || !owner.isValid())
			throw new ConfigurationException("The Owner parameter is null or not valid.");

		if (reasonerType == null || !reasonerType.isValid())
			throw new ConfigurationException("The ReasonerType parameter is null or not valid.");

		if (!isReasonerTypeAvailable(reasonerType))
			throw new ConfigurationException("The ReasonerType is not on the server avaiable.");

		int port = getAvaiablePort();
		if (!isServerPortAvailable(port)) {
			OwlLinkServerAdapter serverByPort = getOwlLinkServer(port);
			if (serverByPort == null) {
				throw new ConfigurationException(String.format("Server port: '%d' is already in used from a other application.",
						port));
			} else {
				throw new ConfigurationException(String.format(
						"Server port: '%d' is already in used from reasoner with id: '%s' and created by owner: '%s'.", port,
						serverByPort.getOwlLinkServerMetadata().getReasonerType().getId(), serverByPort
								.getOwlLinkServerMetadata().getOwner()));
			}

		} else {
			OwlLinkServerAdapter owlLinkServer = createOwlLinkServer(port, reasonerType, owner);

			this.owlLinkServers.add(owlLinkServer);

			wirteActionInfo(owlLinkServer, "registred");

			return port;
		}
	}

	public void deregisterOwlLinkServer( int port ) {
		OwlLinkServerAdapter owlLinkServer = getOwlLinkServer(port);
		if (owlLinkServer == null) {
			return;
		}

		if (owlLinkServer.isStarted()) {
			stopOwlLinkServer(port);
		}

		this.owlLinkServers.remove(owlLinkServer);

		wirteActionInfo(owlLinkServer, "deregistred");
	}

	public void deregisterAllOwlLinkServers() {
		for (OwlLinkServerMetadata olsm : getAllOwlLinkServerMetadatas()) {
			deregisterOwlLinkServer(olsm.getPort());
		}
	}

	public void deregisterAllOwlLinkServers( Owner owner ) {
		for (OwlLinkServerMetadata olsm : getAllOwlLinkServerMetadatas(owner)) {
			deregisterOwlLinkServer(olsm.getPort());
		}
	}

	public void startOwlLinkServer( int port ) {
		OwlLinkServerAdapter owlLinkServer = getOwlLinkServer(port);
		if (owlLinkServer == null) {
			Log("No Server on port: '%d' is initialized.", port);
		} else {
			if (!owlLinkServer.isStarted()) {
				owlLinkServer.start();

				if (owlLinkServer.isStarted()) {
					wirteActionInfo(owlLinkServer, "started");
				}
			}
		}
	}

	public void stopOwlLinkServer( int port ) {
		OwlLinkServerAdapter owlLinkServer = getOwlLinkServer(port);
		if (owlLinkServer == null) {
			return;
		}

		if (owlLinkServer.isStarted()) {
			owlLinkServer.stop();

			if (!owlLinkServer.isStarted()) {
				wirteActionInfo(owlLinkServer, "stopped");
			}
		}
	}

	public void stopAllOwlLinkServers() {
		for (OwlLinkServerMetadata olsm : getAllOwlLinkServerMetadatas()) {
			stopOwlLinkServer(olsm.getPort());
		}
	}

	public void stopAllOwlLinkServers( Owner owner ) {
		for (OwlLinkServerMetadata olsm : getAllOwlLinkServerMetadatas(owner)) {
			stopOwlLinkServer(olsm.getPort());
		}
	}

	public boolean isOwlLinkServerRegistered( int port ) {
		for (OwlLinkServerAdapter owlLinkServer : this.owlLinkServers) {
			if (owlLinkServer.getOwlLinkServerMetadata().getPort() == port) {
				return true;
			}
		}
		return false;
	}

	public boolean isOwlLinkServerStarted( int port ) {
		OwlLinkServerAdapter owlLinkServer = getOwlLinkServer(port);
		if (owlLinkServer == null) {
			return false;
		}

		return owlLinkServer.isStarted();
	}

	public boolean isServerPortAvailable( int port ) {
		if (!isPortValide(port))
			return false;

		for (OwlLinkServerAdapter owlLinkServer : this.owlLinkServers) {
			if (owlLinkServer.getOwlLinkServerMetadata().getPort() == port) {
				return false;
			}
		}

		return checkOtherApplications(port);
	}

	private static boolean checkOtherApplications( int port ) {
		if (port < MIN_PORT_NUMBER || port > MAX_PORT_NUMBER) {
			throw new IllegalArgumentException("Invalid start port: " + port);
		}

		ServerSocket ss = null;
		DatagramSocket ds = null;
		try {
			ss = new ServerSocket(port);
			ss.setReuseAddress(true);
			ds = new DatagramSocket(port);
			ds.setReuseAddress(true);
			return true;
		} catch (IOException e) {
		} finally {
			if (ds != null) {
				ds.close();
			}

			if (ss != null) {
				try {
					ss.close();
				} catch (IOException e) {
				}
			}
		}

		return false;
	}

	public ReasonerType getOwlLinkServerReasonerType( int port ) {
		OwlLinkServerAdapter owlLinkServer = getOwlLinkServer(port);
		if (owlLinkServer == null) {
			return null;
		}

		return owlLinkServer.getOwlLinkServerMetadata().getReasonerType();
	}

	public ServerState getOwlLinkServerState( int port ) {
		OwlLinkServerAdapter owlLinkServer = getOwlLinkServer(port);
		if (owlLinkServer == null) {
			return null;
		}

		return owlLinkServer.getOwlLinkServerMetadata().getServerState();
	}

	public int countOwlLinkServersRegistered() {
		return this.owlLinkServers.size();
	}

	public int countOwlLinkServersRegistered( Owner owner ) {
		return getOwlLinkServerFromOwner(owner).size();
	}

	public int countOwlLinkServersStarted() {
		return countOwlLinkServersStarted(this.owlLinkServers);
	}

	public int countOwlLinkServersStarted( Owner owner ) {
		return countOwlLinkServersStarted(getOwlLinkServerFromOwner(owner));
	}

	private int countOwlLinkServersStarted( Set<OwlLinkServerAdapter> owlLinkServers ) {
		int startedOwlLinkServersCount = 0;

		for (OwlLinkServerAdapter owlLinkServer : owlLinkServers) {
			if (owlLinkServer.isStarted())
				startedOwlLinkServersCount++;
		}

		return startedOwlLinkServersCount;
	}

	private Set<OwlLinkServerAdapter> getOwlLinkServerFromOwner( Owner owner ) {
		Set<OwlLinkServerAdapter> owlLinkServers = new HashSet<OwlLinkServerAdapter>();
		for (OwlLinkServerAdapter owlLinkServer : this.owlLinkServers) {
			if (owlLinkServer.getOwlLinkServerMetadata().getOwner().equals(owner)) {
				owlLinkServers.add(owlLinkServer);
			}
		}
		return owlLinkServers;
	}

	private OwlLinkServerAdapterImpl createOwlLinkServer( int port, ReasonerType reasonerType, Owner owner )
			throws ConfigurationException {
		OWLlinkServerFactory owlLinkServerFactory = this.classloader.getInstance(reasonerType.getId());

		return new OwlLinkServerAdapterImpl(this.hostAddress, port, reasonerType, owner, owlLinkServerFactory.createServer(port));
	}

	private boolean isPortValide( int port ) {
		return MIN_PORT_NUMBER <= port && port <= MAX_PORT_NUMBER;
	}

	private void wirteActionInfo( OwlLinkServerAdapter owlLinkServer, String action ) {
		OwlLinkServerMetadata olsm = owlLinkServer.getOwlLinkServerMetadata();
		Info("OwlLinkReaserServer with %s from %s on port:%04d was %s.", olsm
				.getReasonerType(), olsm.getOwner(), olsm.getPort(), action);
	}

	private void Log( String message, Object... objects ) {
		System.out.printf("Log: " + message + "\n", objects);
	}

	private void Info( String message, Object... objects ) {
		System.out.printf("Info: " + message + "\n", objects);
	}

	private void Error( String message, Object... objects ) {
		System.err.printf("Error: " + message + "\n", objects);
	}
}