/*
* Copyright (C) 2013, FZI Forschungszentrum Informatik
*
* Modifications to the initial code base are copyright of their
* respective authors, or their employers as appropriate. Authorship
* of the modifications may be determined from the ChangeLog placed at
* the end of this file.
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 3.0 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/
package de.fzi.owllink.rest.client;

import java.net.URI;
import java.util.HashSet;
import java.util.Set;

import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.UriBuilder;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.sun.jersey.api.client.Client;
import com.sun.jersey.api.client.ClientHandlerException;
import com.sun.jersey.api.client.UniformInterfaceException;
import com.sun.jersey.api.client.WebResource;
import com.sun.jersey.api.client.config.ClientConfig;
import com.sun.jersey.api.client.config.DefaultClientConfig;

import de.fzi.owllink.reasoner.server.entities.Owner;
import de.fzi.owllink.reasoner.server.entities.ReasonerType;
import de.fzi.owllink.reasoner.server.entities.ServerState;
import de.fzi.owllink.reasoner.server.owllinkserver.OwlLinkServerMetadata;
import de.fzi.owllink.rest.commons.parser.OwlLinkServerMetadataParser;
import de.fzi.owllink.rest.commons.parser.OwnerParser;
import de.fzi.owllink.rest.commons.parser.ReasonerTypeParser;
import de.fzi.owllink.rest.commons.parser.ServerStateParser;
import de.fzi.owllink.rest.commons.xml.XmlOwlLinkServerMetadata;
import de.fzi.owllink.rest.commons.xml.XmlOwner;
import de.fzi.owllink.rest.commons.xml.XmlReasonerType;
import de.fzi.owllink.rest.commons.xml.XmlRegisterOwlLinkServerRequest;
import de.fzi.owllink.rest.commons.xml.XmlServerState;

/**
 * 
 * @author Marcel Weigel
 *
 */
public class OwlLinkRestClientImpl implements OwlLinkRestClient {
	
	private final static Log LOG = LogFactory.getLog(OwlLinkRestClientImpl.class);
	
	private WebResource service;
	private String id;
	private String host;
	private int port;
	
	public OwlLinkRestClientImpl(String host, int port) {
		this.host = host;
		this.port = port;
		
		this.id = String.format("%s:%04d", host, port);
		
		ClientConfig config = new DefaultClientConfig();
	    Client client = Client.create(config);
	    this.service = client.resource(getBaseURI()); 
	}
	
	public String getId() {
		return id;
	}

	public String getHost() {
		return host;
	}

	public int getPort() {
		return port;
	}
	
	public void refreshConnection() {
		ClientConfig config = new DefaultClientConfig();
	    Client client = Client.create(config);
	    this.service = client.resource(getBaseURI()); 
	}

	public boolean isConnected() {
		try {
			this.service.path("servers").path("metadata").accept(MediaType.APPLICATION_XML).get(XmlOwlLinkServerMetadata[].class);
		} catch(Exception ex) {
			return false;
		}
		return true;
	}

	public Set<OwlLinkServerMetadata> getAllOwlLinkServerMetadatas() {
		XmlOwlLinkServerMetadata[] xmlOwlLinkServerMetadatas = this.service.path("servers").path("metadata").accept(MediaType.APPLICATION_XML).get(XmlOwlLinkServerMetadata[].class);
		return getSetFromArray(xmlOwlLinkServerMetadatas);
	}
	
	public Set<OwlLinkServerMetadata> getAllOwlLinkServerMetadatas(Owner owner) {
		XmlOwner xmlOwner = OwnerParser.getXmlOwner(owner);
		XmlOwlLinkServerMetadata[] xmlOwlLinkServerMetadatas = this.service.path("servers").path("metadata").accept(MediaType.APPLICATION_XML).post(XmlOwlLinkServerMetadata[].class, xmlOwner);
		return getSetFromArray(xmlOwlLinkServerMetadatas);
	}
	
	private Set<OwlLinkServerMetadata> getSetFromArray(XmlOwlLinkServerMetadata[] xmlOwlLinkServerMetadatas) {
		Set<OwlLinkServerMetadata> owlLinkServerMetadatas = new HashSet<OwlLinkServerMetadata>();
		for(XmlOwlLinkServerMetadata xmlOwlLinkServerMetadata : xmlOwlLinkServerMetadatas) {
			owlLinkServerMetadatas.add(OwlLinkServerMetadataParser.getOwlLinkServerMetadata(xmlOwlLinkServerMetadata));
		}
		return owlLinkServerMetadatas;
	}
	
	public OwlLinkServerMetadata getOwlLinkServerMetadata(int port) {
		XmlOwlLinkServerMetadata xolsm = this.service.path("servers").path(Integer.toString(port)).accept(MediaType.APPLICATION_XML).get(XmlOwlLinkServerMetadata.class);
		
		return OwlLinkServerMetadataParser.getOwlLinkServerMetadata(xolsm);
	}

	public int registerOwlLinkServer(Owner owner, ReasonerType reasonerType) {
		if(owner == null || !owner.isValid() || reasonerType == null || !reasonerType.isValid()) {
			return -1;
		}
		XmlOwner xmlOwner = OwnerParser.getXmlOwner(owner);
		XmlReasonerType xmlReasonerType = ReasonerTypeParser.getXmlReasonerType(reasonerType);
		XmlRegisterOwlLinkServerRequest xmlRegisterOwlLinkServerRequest = new XmlRegisterOwlLinkServerRequest(xmlOwner, xmlReasonerType);
		String portStr = this.service.path("servers").accept(MediaType.APPLICATION_XML).post(String.class, xmlRegisterOwlLinkServerRequest);
		
		return Integer.parseInt(portStr);
	}
	
	public void deregistreOwlLinkServer(int port) {
		this.service.path("servers").path(Integer.toString(port)).accept(MediaType.APPLICATION_XML).delete();
	}

	public void startOwlLinkServer(int port) {
		this.service.path("servers").path(Integer.toString(port)).accept(MediaType.APPLICATION_XML).post();
	}

	public void stopOwlLinkServer(int port) {
		this.service.path("servers").path(Integer.toString(port)).path("active").accept(MediaType.APPLICATION_XML).delete();
	}

	public ServerState getOwlLinkServerState(int port) {
		XmlServerState xmlServerState = this.service.path("servers").path(Integer.toString(port)).path("status").accept(MediaType.APPLICATION_XML).get(XmlServerState.class);
		
		return ServerStateParser.getServerState(xmlServerState);
	}

	public Set<ReasonerType> getAvailableReasonerTypes() {
		Set<ReasonerType> reasonerTypes = new HashSet<ReasonerType>();
		try {
			XmlReasonerType[] xmlReasonerTypes = this.service.path("rest").path("reasoners").accept(MediaType.APPLICATION_XML).get(XmlReasonerType[].class);
			for(XmlReasonerType xmlReasonerType : xmlReasonerTypes) {
				reasonerTypes.add(ReasonerTypeParser.getReasonerType(xmlReasonerType));
			}
		} catch(ClientHandlerException e) {
			final String errStr = String.format("Cannot connect to server %s:%04d.", this.host, this.port);
			LOG.error(errStr);
			throw e;
		} catch(UniformInterfaceException e) {
			final String errStr = String.format("Cannot find serverlet on server %s:%04d.", this.host, this.port);
			LOG.error(errStr);
			throw e;
		}
		return reasonerTypes;
	}
	
	public int getRegisteredOwlLinkServerCount() {
		return Integer.parseInt(this.service.path("servers").path("countRegistered").accept(MediaType.APPLICATION_XML).get(String.class));
	}
	
	public int getRegisteredOwlLinkServerCount(Owner owner) {
		XmlOwner xmlOwner = OwnerParser.getXmlOwner(owner);
		return Integer.parseInt(this.service.path("servers").path("countRegistered").accept(MediaType.APPLICATION_XML).post(String.class, xmlOwner));
	}
	
	public int getStartedOwlLinkServerCount() {
		return Integer.parseInt(this.service.path("servers").path("countStarted").accept(MediaType.APPLICATION_XML).get(String.class));
	}
	
	public int getStartedOwlLinkServerCount(Owner owner) {
		XmlOwner xmlOwner = OwnerParser.getXmlOwner(owner);
		return Integer.parseInt(this.service.path("servers").path("countStarted").accept(MediaType.APPLICATION_XML).post(String.class, xmlOwner));
	}
	
	public void deregistredAllOwlLinkServers() {
		Set<OwlLinkServerMetadata> owlLinkServerMetadatas = getAllOwlLinkServerMetadatas();
		deregistredAllOwlLinkServers(owlLinkServerMetadatas);
	}
	
	public void deregisteredAllOwlLinkServers(Owner owner) {
		Set<OwlLinkServerMetadata> owlLinkServerMetadatas = getAllOwlLinkServerMetadatas(owner);
		deregistredAllOwlLinkServers(owlLinkServerMetadatas);
	}
	
	private void deregistredAllOwlLinkServers(Set<OwlLinkServerMetadata> owlLinkServerMetadatas) {
		for(OwlLinkServerMetadata owlLinkServerMetadata : owlLinkServerMetadatas) {
			deregistreOwlLinkServer(owlLinkServerMetadata.getPort());
		}
	}
	
	public void stopAllOwlLinkServers() {
		Set<OwlLinkServerMetadata> owlLinkServerMetadatas = getAllOwlLinkServerMetadatas();
		stopAllOwlLinkServers(owlLinkServerMetadatas);
	}
	
	public void stopAllOwlLinkServers(Owner owner) {
		Set<OwlLinkServerMetadata> owlLinkServerMetadatas = getAllOwlLinkServerMetadatas(owner);
		stopAllOwlLinkServers(owlLinkServerMetadatas);
	}
	
	private void stopAllOwlLinkServers(Set<OwlLinkServerMetadata> owlLinkServerMetadatas) {
		for(OwlLinkServerMetadata owlLinkServerMetadata : owlLinkServerMetadatas) {
			stopOwlLinkServer(owlLinkServerMetadata.getPort());
		}
	}
	
	private URI getBaseURI() {
		return UriBuilder.fromUri(String.format("http://%s:%d/rest-server", this.host, this.port)).build();
	}
	
	@Override
	public boolean equals(Object obj) {
		if(obj == null)
			return false;
		if(obj == this)
			return true;
		if(!(obj instanceof OwlLinkRestClientImpl)) 
			return false;
		
		OwlLinkRestClientImpl olrc = (OwlLinkRestClientImpl) obj;
		return (this.id == olrc.id || this.id != null && this.id.equals(olrc.id)) && 
				(this.host == olrc.host || this.host != null && this.host.equals(olrc.host)) &&
				this.port == olrc.port &&
				(this.service == olrc.service || this.service != null && this.service.equals(olrc.service));
	}

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 31 * hash + (this.id == null ? 0 : this.id.hashCode());
		hash = 31 * hash + (this.host == null ? 0 : this.host.hashCode());
		hash = 31 * hash + this.port;
		hash = 31 * hash + (this.service == null ? 0 : this.service.hashCode());
		return hash;
	}
}
