/*
* Copyright (C) 2013, FZI Forschungszentrum Informatik
*
* Modifications to the initial code base are copyright of their
* respective authors, or their employers as appropriate. Authorship
* of the modifications may be determined from the ChangeLog placed at
* the end of this file.
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 3.0 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/
package de.fzi.owllink.rest.jetty.server.controller;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.fzi.owllink.reasoner.server.entities.Owner;
import de.fzi.owllink.reasoner.server.entities.ReasonerType;
import de.fzi.owllink.reasoner.server.entities.ServerState;
import de.fzi.owllink.reasoner.server.manager.ConfigurationException;
import de.fzi.owllink.reasoner.server.manager.OwlLinkServerManager;
import de.fzi.owllink.reasoner.server.manager.OwlLinkServerManagerFactory;
import de.fzi.owllink.reasoner.server.owllinkserver.OwlLinkServerMetadata;

import de.fzi.owllink.rest.commons.xml.XmlOwlLinkServerMetadata;
import de.fzi.owllink.rest.commons.xml.XmlOwner;
import de.fzi.owllink.rest.commons.xml.XmlReasonerType;
import de.fzi.owllink.rest.commons.parser.OwlLinkServerMetadataParser;
import de.fzi.owllink.rest.commons.parser.OwnerParser;
import de.fzi.owllink.rest.commons.parser.ReasonerTypeParser;

/**
 * 
 * @author Marcel Weigel
 *
 */
public class RestServerController {
	
	private final static Log LOG = LogFactory.getLog(RestServerController.class);

	private OwlLinkServerManager owlLinkServerManager;
	
	public RestServerController() {
		String hostAddress = "0.0.0.0";
		try {
			hostAddress = InetAddress.getLocalHost().getHostAddress();
		} catch (UnknownHostException e) {
			LOG.error("Counld not find any host address.");
		}
		this.owlLinkServerManager = OwlLinkServerManagerFactory.getInstance().getOwlLinkServerManagerInstance(hostAddress);
	}
	
	public XmlReasonerType[] getAvailableReasonerTypes() {
		Set<ReasonerType> reasonerTypes = owlLinkServerManager.getAvailableReasonerTypes();
		
		XmlReasonerType[] xmlReasonerTypes = new XmlReasonerType[reasonerTypes.size()];
		int i = 0;
		for(ReasonerType reasonerType : reasonerTypes) {
			xmlReasonerTypes[i++] = ReasonerTypeParser.getXmlReasonerType(reasonerType);
		}
		return xmlReasonerTypes;
	}

	public boolean isReasonerAvailable(XmlReasonerType xmlReasonerType) {
		return owlLinkServerManager.isReasonerTypeAvailable(ReasonerTypeParser.getReasonerType(xmlReasonerType));
	}

	public int registerOwlLinkServer(XmlOwner xmlOwner, XmlReasonerType xmlReasonerType) {
		int port;
		try {
			Owner owner = OwnerParser.getOwner(xmlOwner);
			ReasonerType reasonerType = ReasonerTypeParser.getReasonerType(xmlReasonerType);
			port = this.owlLinkServerManager.registerOwlLinkServer(reasonerType, owner);
		} catch (ConfigurationException e) { 
			return -1;
		}
		return port;
	}

	public void deregistreOwlLinkServer(int port) {;
		this.owlLinkServerManager.deregisterOwlLinkServer(port);
	}
	
	public void deregistreAllOwlLinkServers() {
		this.owlLinkServerManager.deregisterAllOwlLinkServers();
	}
	
	public void startOwlLinkServer(int port) {
		this.owlLinkServerManager.startOwlLinkServer(port);
	}
	
	public void stopOwlLinkServer(int port) {
		this.owlLinkServerManager.stopOwlLinkServer(port);
	}
	
	public boolean isOwlLinkServerRegistered(int port) {
		return this.owlLinkServerManager.isOwlLinkServerRegistered(port);
	}
	
	public boolean isOwlLinkServerStarted(int port) {
		return this.owlLinkServerManager.isOwlLinkServerStarted(port);
	}
	
	public XmlReasonerType getOwlLinkServerReasonerType(int port) {
		ReasonerType reasonerType = this.owlLinkServerManager.getOwlLinkServerReasonerType(port);
		return ReasonerTypeParser.getXmlReasonerType(reasonerType);
	}
	
	public ServerState getOwlLinkServerState(int port) {
		return this.owlLinkServerManager.getOwlLinkServerState(port);
	}
	
	public int countOwlLinkServersRegistered() {
		return this.owlLinkServerManager.countOwlLinkServersRegistered();
	}
	
	public int countOwlLinkServersRegistered(XmlOwner xmlOwner) {
		Owner owner = OwnerParser.getOwner(xmlOwner);
		return this.owlLinkServerManager.countOwlLinkServersRegistered(owner);
	}
	
	public int countOwlLinkServersStarted() {
		return this.owlLinkServerManager.countOwlLinkServersStarted();
	}
	
	public int countOwlLinkServersStarted(XmlOwner xmlOwner) {
		Owner owner = OwnerParser.getOwner(xmlOwner);
		return this.owlLinkServerManager.countOwlLinkServersStarted(owner);
	}
	
	public XmlOwlLinkServerMetadata[] getXmlOwlLinkServerMetadatas() {
		Set<OwlLinkServerMetadata> owlLinkServerMetadatas = this.owlLinkServerManager.getAllOwlLinkServerMetadatas();
		XmlOwlLinkServerMetadata[] xmlOwlLinkServerMetadatas = new XmlOwlLinkServerMetadata[owlLinkServerMetadatas.size()];
		int i = 0;
		for(OwlLinkServerMetadata owlLinkServerMetadata : owlLinkServerMetadatas) {
			xmlOwlLinkServerMetadatas[i++] = OwlLinkServerMetadataParser.getXmlOwlLinkServerMetadata(owlLinkServerMetadata);
		}
		return xmlOwlLinkServerMetadatas;
	}
	
	public XmlOwlLinkServerMetadata[] getXmlOwlLinkServerMetadatas(XmlOwner xmlOwner) {
		Owner owner = OwnerParser.getOwner(xmlOwner);
		Set<OwlLinkServerMetadata> owlLinkServerMetadatas = this.owlLinkServerManager.getAllOwlLinkServerMetadatas(owner);
		XmlOwlLinkServerMetadata[] xmlOwlLinkServerMetadatas = new XmlOwlLinkServerMetadata[owlLinkServerMetadatas.size()];
		int i = 0;
		for(OwlLinkServerMetadata owlLinkServerMetadata : owlLinkServerMetadatas) {
			xmlOwlLinkServerMetadatas[i++] = OwlLinkServerMetadataParser.getXmlOwlLinkServerMetadata(owlLinkServerMetadata);
		}
		return xmlOwlLinkServerMetadatas;
	}

	public XmlOwlLinkServerMetadata getXmlOwlLinkServerMetadata(int port) {
		OwlLinkServerMetadata olsm = this.owlLinkServerManager.getOwlLinkServerMetadata(port);
		return OwlLinkServerMetadataParser.getXmlOwlLinkServerMetadata(olsm);
	}
}
