/*
* Copyright (C) 2013, FZI Forschungszentrum Informatik
*
* Modifications to the initial code base are copyright of their
* respective authors, or their employers as appropriate. Authorship
* of the modifications may be determined from the ChangeLog placed at
* the end of this file.
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 3.0 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/
package de.fzi.owllink.rest.jetty.server.resource;

import java.net.HttpURLConnection;

import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Request;
import javax.ws.rs.core.UriInfo;

import de.fzi.owllink.reasoner.server.entities.ServerState;

import de.fzi.owllink.rest.commons.xml.XmlOwlLinkServerMetadata;
import de.fzi.owllink.rest.commons.xml.XmlServerState;
import de.fzi.owllink.rest.commons.parser.ServerStateParser;
import de.fzi.owllink.rest.jetty.server.dao.ServerDao;

/**
 * 
 * @author Marcel Weigel
 *
 */
public class ServerResource {
	@Context
	UriInfo uriInfo;
	@Context
	Request request;
  
	int port;
  
	public ServerResource(UriInfo uriInfo, Request request, int port) 
	{
		this.uriInfo = uriInfo;
		this.request = request;
		this.port = port;
	}
	
	/**
	 * Diese Operation ruft einen OwlLinkServer ab. Dies gibt die Metadaten des OwlLinkServer zurck.
	 * 
	 * @return Gibt die Metadata des OwlLinkServers zurck.
	 * 
	 *  Return 404: Angeben OwlLinkServer Id existiert nicht.
	 *  Return 500: Interner Fehler.
	 */
	@GET
	@Produces({MediaType.APPLICATION_XML})
	public XmlOwlLinkServerMetadata getOwlLinkServerMetadata() 
	{
		if(!ServerDao.instance.getModel().isOwlLinkServerRegistered(this.port)) {
			throw new WebApplicationException(HttpURLConnection.HTTP_NOT_FOUND);
		}
		XmlOwlLinkServerMetadata xolsm;
		
		try {
			xolsm = ServerDao.instance.getModel().getXmlOwlLinkServerMetadata(this.port);
		} catch(Exception e) {
			throw new WebApplicationException(HttpURLConnection.HTTP_INTERNAL_ERROR);
		}
		
		return xolsm;
	}
	
	/**
	 * Diese Operation erlaubt es einen LinkOwlServer zu lschen. Smtliche Daten eine 
	 * OwlLinkServer werden gelscht und der benutze Port wird wieder freigegeben. Falls 
	 * der angegebene OwlLinkServer noch am Laufen ist kann er nicht gelscht werden.
	 * 
	 *  Return 200: Ok, der LinkOwlServer wurde gelscht
	 *  Return 400: Bentigte Parameter werden vermisst oder Wert hat falsches Format.
	 *  Return 404: Angeben LinkOwlServer Id existiert nicht.
	 *  Return 409: Der LinkOwlServer ist noch gestartet.
	 *  Return 500: Interner Fehler.
	 */
	@POST
	@Produces(MediaType.APPLICATION_XML)
	public void activeServer() 
	{  
		if(!ServerDao.instance.getModel().isOwlLinkServerRegistered(this.port)) {
			throw new WebApplicationException(HttpURLConnection.HTTP_NOT_FOUND);
		}
		if(ServerDao.instance.getModel().isOwlLinkServerStarted(this.port)) {
			throw new WebApplicationException(HttpURLConnection.HTTP_CONFLICT);
		}
		
		try {
			ServerDao.instance.getModel().startOwlLinkServer(this.port);
		} catch(Exception e) {
			throw new WebApplicationException(HttpURLConnection.HTTP_INTERNAL_ERROR);
		}
			
		if(!ServerDao.instance.getModel().isOwlLinkServerStarted(this.port)) {
			throw new WebApplicationException(HttpURLConnection.HTTP_CONFLICT);
		}
		
		throw new WebApplicationException(HttpURLConnection.HTTP_OK);
	}

	/**
	 * Diese Operation erlaubt es einen LinkOwlServer zu lschen. Smtliche Daten eine 
	 * LinkOwlServer werden gelscht und der benutze Port wird wieder freigegeben. Falls 
	 * der angegebene LinkOwlServer noch am Laufen ist kann er nicht gelscht werden.
	 * 
	 *  Return 200: Ok, der LinkOwlServer wurde gelscht
	 *  Return 404: Angeben LinkOwlServer Id existiert nicht.
	 *  Return 409: Der LinkOwlServer ist nicht registiert.
	 *  Return 500: Interner Fehler.
	 */
	@DELETE
	@Produces({MediaType.APPLICATION_XML})
	public void deleteLinkServer() 
	{
		if(!ServerDao.instance.getModel().isOwlLinkServerRegistered(this.port)) {
			throw new WebApplicationException(HttpURLConnection.HTTP_NOT_FOUND);
		}
		
		try {
			ServerDao.instance.getModel().deregistreOwlLinkServer(this.port);
		} catch(Exception e) {
			throw new WebApplicationException(HttpURLConnection.HTTP_INTERNAL_ERROR);
		}
		
		if(ServerDao.instance.getModel().isOwlLinkServerRegistered(this.port)) {
			throw new WebApplicationException(HttpURLConnection.HTTP_CONFLICT);
		}
		
		throw new WebApplicationException(HttpURLConnection.HTTP_OK);
	}
	
	@GET 
	@Path("status")
	@Produces(MediaType.APPLICATION_XML)
	public XmlServerState statusServerApplication()
	{   
		if(!ServerDao.instance.getModel().isOwlLinkServerRegistered(this.port)) {
			throw new WebApplicationException(HttpURLConnection.HTTP_NOT_FOUND);
		}
		
		ServerState serverState = null;
		try {
			serverState = ServerDao.instance.getModel().getOwlLinkServerState(this.port);
		} catch(Exception e) {
			throw new WebApplicationException(HttpURLConnection.HTTP_INTERNAL_ERROR);
		}
		if(serverState == null) {
			throw new WebApplicationException(HttpURLConnection.HTTP_INTERNAL_ERROR);
		}
		return ServerStateParser.getXmlServerState(serverState);
	}
	
	@Path("{active}")
	public ActiveResource getActive(@PathParam("active") String active) 
	{
		return new ActiveResource(uriInfo, request, this.port);
	}
} 