/*
* Copyright (C) 2013, FZI Forschungszentrum Informatik
*
* Modifications to the initial code base are copyright of their
* respective authors, or their employers as appropriate. Authorship
* of the modifications may be determined from the ChangeLog placed at
* the end of this file.
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 3.0 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/
package de.fzi.owllink.rest.jetty.server.resource;

import java.net.HttpURLConnection;

import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Request;
import javax.ws.rs.core.UriInfo;
import javax.xml.bind.JAXBElement;

import de.fzi.owllink.rest.commons.xml.XmlOwlLinkServerMetadata;
import de.fzi.owllink.rest.commons.xml.XmlOwner;
import de.fzi.owllink.rest.commons.xml.XmlReasonerType;
import de.fzi.owllink.rest.commons.xml.XmlRegisterOwlLinkServerRequest;
import de.fzi.owllink.rest.jetty.server.dao.ServerDao;

/**
 * 
 * @author Marcel Weigel
 *
 */
@Path("/servers")
public class ServersResource {
	
	@Context
	UriInfo uriInfo;
	@Context
	Request request;

	/**
	 * Diese Operation ruft alle OwlLinkServer ab die auf dem Server 
	 * registriert sind. Dieses OwlLinkServer Array wird als Json Objekt zurck gegeben.
	 * 
	 * @return Die Liste mit alle Metadatas. 
	 * 
	 *  Return 400: Bentigte Parameter werden vermisst oder Wert hat falsches Format. (HTTP_BAD_REQUEST)
	 *  Return 500: Interner Fehler. (HTTP_INTERNAL_ERROR)
	 */
	@GET
	@Path("metadata")
	@Produces({ MediaType.APPLICATION_XML })
	public XmlOwlLinkServerMetadata[] getMetadats() {
		return ServerDao.instance.getModel().getXmlOwlLinkServerMetadatas();
	}
	
	@POST
	@Path("metadata")
	@Produces({ MediaType.APPLICATION_XML })
	public XmlOwlLinkServerMetadata[] getMetadats(JAXBElement<XmlOwner> jaxbElement) {
		XmlOwner xmlOwner = jaxbElement.getValue();
		return ServerDao.instance.getModel().getXmlOwlLinkServerMetadatas(xmlOwner);
	}
	
	/**
	 * Diese Operation erlaubt es eine neuen OwlLinkReasonerServer zu registrieren. Die 
	 * OwlLinkReasonerServer Daten werden ber die Ressourcen des Xml Objekt bergeben.
	 * 
	 * @param jaxbElementOwner Beinhaltet Informationen ber den Eigentmer des OwlLinkReasonerServers.
	 * @param jaxbElementReasonerType Beinhaltet Informationen ber den Reasoner, der registiert werden soll.
	 * 
	 * @return Gibt den Port des erzeugten OwlLinkReasonerServer zurck.
	 * 
	 * 	Return 400: Bentigte Parameter werden vermisst oder Wert hat falsches Format. (HTTP_BAD_REQUEST)
	 *  Return 500: Interner Fehler. (HTTP_INTERNAL_ERROR)
	 */
	@POST
	@Consumes(MediaType.APPLICATION_XML)
	public String registerOwlLinkReasoner(JAXBElement<XmlRegisterOwlLinkServerRequest> jaxbElement) {
		XmlRegisterOwlLinkServerRequest xmlRegisterOwlLinkReasonerServerRequest = jaxbElement.getValue();
		XmlOwner xmlOwner = xmlRegisterOwlLinkReasonerServerRequest.getXmlOwner();
		XmlReasonerType xmlReasonerType = xmlRegisterOwlLinkReasonerServerRequest.getXmlReasonerType();
		
		if(!ServerDao.instance.getModel().isReasonerAvailable(xmlReasonerType)) {
			throw new WebApplicationException(HttpURLConnection.HTTP_BAD_REQUEST);
		}
		int port;
		try {
			port = ServerDao.instance.getModel().registerOwlLinkServer(xmlOwner, xmlReasonerType);
		} catch(Exception e) {
			throw new WebApplicationException(HttpURLConnection.HTTP_INTERNAL_ERROR);
		}
		return Integer.toString(port);
	}
	
	@GET
	@Consumes(MediaType.TEXT_HTML)
	public String test() {
		return "Rest Server is running!";
	}
	
	/**
	 * Diese Operation lscht alle OwlLinkServer die regesitert worden sind.
	 * 
	 *  Return 200: Es wurden alle OwlLinkServer deregisterit. (HTTP_OK)
	 *  Return 409: Es wurden nicht alle OwlLinkServer deregisterit. (HTTP_CONFLICT)
	 */
	@DELETE
	@Consumes(MediaType.APPLICATION_XML)
	public void deregistrateAllOwlLinkServers() {
		ServerDao.instance.getModel().deregistreAllOwlLinkServers();
		if(ServerDao.instance.getModel().countOwlLinkServersRegistered() != 0) {
			throw new WebApplicationException(HttpURLConnection.HTTP_CONFLICT);
		}

		throw new WebApplicationException(HttpURLConnection.HTTP_OK);
	}
	
	@GET 
	@Path("countRegistered")
	@Produces(MediaType.APPLICATION_XML)
	public String countServersRegistered()
	{   
		return Integer.toString(ServerDao.instance.getModel().countOwlLinkServersRegistered());
	}
	
	@POST 
	@Path("countRegistered")
	@Produces(MediaType.APPLICATION_XML)
	public String countServersRegistered(JAXBElement<XmlOwner> jaxbElement)
	{   
		XmlOwner xmlOwner = jaxbElement.getValue();
		return Integer.toString(ServerDao.instance.getModel().countOwlLinkServersRegistered(xmlOwner));
	}
	
	@GET 
	@Path("countStarted")
	@Produces(MediaType.APPLICATION_XML)
	public String countServersStarted()
	{   
		return Integer.toString(ServerDao.instance.getModel().countOwlLinkServersStarted());
	}
	
	@POST
	@Path("countStarted")
	@Produces(MediaType.APPLICATION_XML)
	public String countServersStarted(JAXBElement<XmlOwner> jaxbElement)
	{   
		XmlOwner xmlOwner = jaxbElement.getValue();
		return Integer.toString(ServerDao.instance.getModel().countOwlLinkServersStarted(xmlOwner));
	}

	@Path("{server}")
	public ServerResource getServer(@PathParam("server") int serverPort) {
		return new ServerResource(uriInfo, request, serverPort);
	}

}