/*
* Copyright (C) 2013, FZI Forschungszentrum Informatik
*
* Modifications to the initial code base are copyright of their
* respective authors, or their employers as appropriate. Authorship
* of the modifications may be determined from the ChangeLog placed at
* the end of this file.
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 3.0 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/
package de.fzi.owllink.rest.viewer.model.monitor;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import de.fzi.owllink.reasoner.server.owllinkserver.OwlLinkServerMetadata;
import de.fzi.owllink.rest.viewer.model.RestViewerModel;
import de.fzi.owllink.rest.viewer.view.RestViewerView;

/**
 * 
 * @author Marcel Weigel
 *
 */
public class RestServerMonitor {

	private static final int OWLLINK_SERVER_REFRESH_TIME = 100;
	private static final int REST_SERVER_REFRESH_TIME = 1000;
	
	private Map<String, Set<OwlLinkServerMetadata>> cashedMetadatas = new LinkedHashMap<String, Set<OwlLinkServerMetadata>>();
	private Map<String, RestServerState> cashedRestServerState = new LinkedHashMap<String, RestServerState>();
	
	private boolean hasAllChanges = false;
	private boolean hasSelectedRestServerChanges = false;
	
	private Thread owllinkWatcherThread;
	private Thread restWatcherThread;
	
	private boolean threadsLock = false;
	private boolean owllinkWatcherThreadWait = false;
	private boolean restWatcherThreadWait = false;
	
	private RestViewerModel model;
	private RestViewerView view;
	
	public RestServerMonitor(RestViewerView view, RestViewerModel model) {
		this.view = view;
		this.model = model;
		
		initOwllinkTimerThread();
		initRestTimerThread();
	}
	
	private void initOwllinkTimerThread() {
		this.owllinkWatcherThread = new Thread() {
			public void run() {
				while (!this.isInterrupted()) {
//					try {
						if(!threadsLock) {
							owllinkWatcherThreadWait = false;
							refreshAllOwllinkServers();
						} else {
							owllinkWatcherThreadWait = true;
						}
//					} catch (Exception e) {
//						view.error("Refresh Owllink Thread", e.toString());
//					}
					try {
						Thread.sleep(OWLLINK_SERVER_REFRESH_TIME);
					} catch (InterruptedException e) {
					}
				}
			}
		};
		this.owllinkWatcherThread.start();
	}
	
	private void initRestTimerThread() {
		this.restWatcherThread = new Thread() {
			public void run() {
				while (!this.isInterrupted()) {
					try {
						if(!threadsLock) {
							restWatcherThreadWait = false;
							refreshAllRestServers();
						} else {
							restWatcherThreadWait = true;
						}
					} catch (Exception e) {
						view.error("Refresh Rest Thread", e.toString());
					}
					try {
						Thread.sleep(REST_SERVER_REFRESH_TIME);
					} catch (InterruptedException e) {
					}
				}
			}
		};
		this.restWatcherThread.start();
	}
	
	public void lockThreadStart() {
		this.threadsLock = true;
		
		while(!this.owllinkWatcherThreadWait && !this.restWatcherThreadWait) {
		}
	}
	
	public void lockThreadEnd() {
		this.threadsLock = false;
	}
	
	private void refreshAllRestServers() {
		boolean hasRestServerChanges = false;
		
		for(String restServerId : this.model.getAllRestClientIds()) {
			if(!isRestServerCashed(restServerId)) {
				this.cashedMetadatas.put(restServerId, new LinkedHashSet<OwlLinkServerMetadata>());
				RestServerState restServerState = new RestServerState(this.model.isRestServerConnected(restServerId));
				this.cashedRestServerState.put(restServerId, restServerState);
				hasRestServerChanges = true;
			} else {
				boolean state = this.model.isRestServerConnected(restServerId);
				if(state != getCashedRestServerState(restServerId).getLastConnectionState()) {
					hasRestServerChanges = true;
					setCashedRestServerState(restServerId, state, false);
				}
			}
		} 
		for(String restClientId : this.cashedRestServerState.keySet()) {
			if(!this.model.existRestServer(restClientId)) {
				removeCashedRestServerState(restClientId);
				hasRestServerChanges = true;
			}
		}
		if(hasRestServerChanges) {
			this.model.refreshRestServerPanel();
		}
	}
	
	private boolean isRestServerCashed(String restClientId) {
		return this.cashedRestServerState.containsKey(restClientId);
	}
	
	private RestServerState getCashedRestServerState(String restClientId) {
		if(isRestServerCashed(restClientId)) {
			return this.cashedRestServerState.get(restClientId);
		}
		return null;
	}
	
	private void setCashedRestServerState(String restClientId, boolean state, boolean lastSet) {
		if(isRestServerCashed(restClientId)) {
			RestServerState restServerState = this.cashedRestServerState.get(restClientId);
			restServerState.setLastConnectionState(state);
			restServerState.setWasConnectionLastSet(lastSet);
		}
	}
	
	private void refreshAllOwllinkServers() {
		Map<String, Set<OwlLinkServerMetadata>> owlLinkServerMetatdataMap = new HashMap<String, Set<OwlLinkServerMetadata>>();
		this.hasAllChanges = false;

		if(this.model.isRestClientSelected()) {
			String selectedRestServertId = this.model.getSelectedRestClientId();

			refreshOwllinkServers(selectedRestServertId);
			
			owlLinkServerMetatdataMap.put(selectedRestServertId, getCashedMetadatas(selectedRestServertId));
		} else {
			for(String restServerId : this.model.getAllRestClientIds()) {
				refreshOwllinkServers(restServerId);
			}
			owlLinkServerMetatdataMap = this.cashedMetadatas;
		}
		for(String restClientId : owlLinkServerMetatdataMap.keySet()) {
			if(!this.model.existRestServer(restClientId)) {
				this.hasAllChanges = true;
				removeCashedMetadatas(restClientId);
			}
		}
		
		if(this.hasSelectedRestServerChanges) {
			this.hasAllChanges = true;
			this.hasSelectedRestServerChanges = false;
		}
		
		if(this.hasAllChanges) {
			this.model.refreshReasonerPanel(owlLinkServerMetatdataMap);
		}
	}
	
	private void refreshOwllinkServers(String restClientId) {
		Set<OwlLinkServerMetadata> currentMetadatas = this.model.getOwlLinkServerMetadatas(restClientId);
		if(!this.model.isRestServerConnected(restClientId)) {
			//TODO Rest Server nicht mehr zu erreichen.
			
			RestServerState restServerState = getCashedRestServerState(restClientId);
			if(restServerState == null) {
				return;
			}
			if(!restServerState.getWasConnectionLastSet()) {
				restServerState.setWasConnectionLastSet(true);
				
				currentMetadatas = new LinkedHashSet<OwlLinkServerMetadata>();
				
				setCashedMetadatas(restClientId, currentMetadatas);
				this.hasAllChanges = true;
			}
		} else {
			currentMetadatas = this.model.getOwlLinkServerMetadatas(restClientId);
			
			boolean hadChanges = false;
			
			for(OwlLinkServerMetadata currentMetadata : currentMetadatas) {
				OwlLinkServerMetadata metadata = getCashedMetadata(restClientId, currentMetadata.getPort());
				if(metadata == null) {
					this.view.printServerRegistred(currentMetadata);
					hadChanges = true;
				} else if(currentMetadata.getServerState() != metadata.getServerState()) {
					this.view.printServerStateChanged(currentMetadata);
					hadChanges = true;
				}
			}
			
			for(OwlLinkServerMetadata metadata : getCashedMetadatas(restClientId)) {
				OwlLinkServerMetadata currentMetadata = this.model.getOwlLinkServerMetadata(restClientId, metadata.getPort());
				if(currentMetadata == null) {
					this.view.printServerDeregistred(metadata);
					hadChanges = true;
				}
			}
			
			if(hadChanges) {
				setCashedMetadatas(restClientId, currentMetadatas);
				this.hasAllChanges = true;
			}
		}	
	}
	
	private Set<OwlLinkServerMetadata> getCashedMetadatas(String restClientId) {
		if(this.cashedMetadatas.containsKey(restClientId)) {
			return this.cashedMetadatas.get(restClientId);
		}
		return new LinkedHashSet<OwlLinkServerMetadata>();
	}
	
	private void setCashedMetadatas(String restClientId, Set<OwlLinkServerMetadata> metadatas) {
		this.cashedMetadatas.put(restClientId, metadatas);
	}
	
	public OwlLinkServerMetadata getCashedMetadata(String restClientId, int port) {
		Set<OwlLinkServerMetadata> metadatas = getCashedMetadatas(restClientId);
		if(metadatas != null) {
			for(OwlLinkServerMetadata metadata : metadatas) {
				if(metadata.getPort() == port) {
					return metadata;
				}
			}
		}
		return null;
	}
	
	private void removeCashedMetadatas(String restClientId) {
		if(this.cashedMetadatas.containsKey(restClientId)) {
			this.cashedMetadatas.remove(restClientId);
		}
	}
	
	private void removeCashedRestServerState(String restClientId) {
		if(this.cashedRestServerState.containsKey(restClientId)) {
			this.cashedRestServerState.remove(restClientId);
		}
	}

	public void interruptAllThreads() {
		this.owllinkWatcherThread.interrupt();
		this.restWatcherThread.interrupt();
	}

	public void selectedRestServerHasChanges() {
		this.hasSelectedRestServerChanges = true;
	}
}
